/*
 * snf_cuda_example.cu
 *
 * Simple program to demonstrate receiving packets with snf and sending them to
 * the gpu.
 *
 * The main objectives of this simple example is to demonstrate the following:
 *  - Retrieving the snf ring address and size
 *  - Registering the snf ring with cuda
 *  - Receiving a group of packets with snf_ring_recv_many
 *  - DMAing the group of packets to the gpu
 *  - Releasing the packet data back to snf
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <snf.h>
#include <cuda_runtime.h>

#define GPU_THREADS_PER_BLOCK 128
#define GPU_BLOCKS 100
#define GPU_ITERATIONS 100

#define NUM_PACKETS (GPU_THREADS_PER_BLOCK * GPU_BLOCKS)

/* some simple check macros */

#define CUDA_CHECK(value) {						\
    cudaError_t _m_cudaStat = value;					\
    if (_m_cudaStat != cudaSuccess) {					\
      fprintf(stderr, "Error '%s' (%d) at line %d in file %s\n",	\
	      cudaGetErrorString(_m_cudaStat), _m_cudaStat,		\
	      __LINE__, __FILE__);					\
      exit(1);								\
    } }

#define SNF_CHECK(value) {						\
    int _ret_val = value;						\
    if (_ret_val != 0) {						\
      fprintf(stderr, "Error '%s' (%d) at line %d in file %s\n",	\
	      strerror(_ret_val), _ret_val, __LINE__, __FILE__);	\
      exit(1);								\
    }									\
  }						

__device__ unsigned int mac_vendor_table[256][256][256];

/*
 * Trivial gpu kernel to count the number of unique vendor mac address 
 * prefixes of the packets passed in.
 */
__global__ void count_macs(struct snf_recv_req *req, uintptr_t ring_addr_h,
			   uintptr_t ring_addr_d, unsigned int *unique_cnt)
{
  unsigned int tid = blockIdx.x * blockDim.x + threadIdx.x;
  unsigned int old;
  uint8_t a, b, c;

  void *packet;
  
  packet = (void *)((uintptr_t)req[tid].pkt_addr - ring_addr_h + ring_addr_d);
  
  /* find the destination mac */
  a = *(uint8_t *)((uintptr_t) packet + 0);
  b = *(uint8_t *)((uintptr_t) packet + 1);
  c = *(uint8_t *)((uintptr_t) packet + 2);

  old = atomicExch(&mac_vendor_table[a][b][c], 1);
  if (old == 0) {
    atomicAdd(unique_cnt, 1);
  }

  /* find the source mac */
  a = *(uint8_t *)((uintptr_t) packet + 6);
  b = *(uint8_t *)((uintptr_t) packet + 7);
  c = *(uint8_t *)((uintptr_t) packet + 8);

  old = atomicExch(&mac_vendor_table[a][b][c], 1);
  if (old == 0) {
    atomicAdd(unique_cnt, 1);
  }

  return;
}


int dma_to_gpu(struct snf_recv_req *reqs, 
	       struct snf_recv_req *reqs_d,
	       int num_reqs, 
	       void *host_ring_addr_base,
	       void *gpu_ring_addr_base)
{

  void *dma_start_addr_h;
  void *dma_start_addr_d;
  size_t dma_length;
  int64_t gap;
  int i, si;

  if (num_reqs < 1 || reqs == NULL) {
    return EINVAL;
  }

  /* DMA the packet data */
  si = 0;
  do {
    dma_start_addr_h = reqs[si].pkt_addr;
    dma_start_addr_d = (void *) ((uintptr_t) dma_start_addr_h - 
				 (uintptr_t) host_ring_addr_base + 
				 (uintptr_t) gpu_ring_addr_base); 
    dma_length = reqs[si].length;
  
    for (i = si + 1; i < num_reqs; i++) {
      gap = (uintptr_t)reqs[i].pkt_addr - 
	(uintptr_t)dma_start_addr_h - dma_length;

      /* if we have wrapped the ring, or if the gap is too big, 
	 do separate dmas */
      if (gap < 0 || gap > (32*1024)) {
	break;
      }

      dma_length += reqs[i].length + gap;
    }

    CUDA_CHECK(cudaMemcpy(dma_start_addr_d, dma_start_addr_h, 
			  dma_length, cudaMemcpyHostToDevice));

    si = i;
  } while (i < num_reqs);

  /* DMA the req data */
  CUDA_CHECK(cudaMemcpy(reqs_d, reqs, sizeof(struct snf_recv_req) * num_reqs,
			cudaMemcpyHostToDevice));

  return 0;
}

int main (void)
{
  snf_handle_t hsnf;
  snf_ring_t hring;
  struct snf_ring_portinfo snf_pi;

  int portcount;
  int gpucount;
  int rc;  

  void *gpu_ring_addr_d;
  struct snf_recv_req *recv_reqs_d;
  struct snf_recv_req *recv_reqs_h;

  void *mac_vendor_table_hp;

  unsigned int *unique_vendor_macs_d;
  unsigned int *unique_vendor_macs_h;

  int i, j;
  int num_pkts;
  int size_to_free;
  int nreqs;

  /* initialize and open snf */
  SNF_CHECK(snf_init(SNF_VERSION_API));
  SNF_CHECK(snf_open(0, 0, NULL, 0, 0, &hsnf));
  SNF_CHECK(snf_ring_open(hsnf, &hring));

  SNF_CHECK(snf_ring_portinfo_count(hring, &portcount));
  if (portcount != 1) {
    printf("This example program does not support aggregate rings\n"
	   "portcount = %d\n", portcount);
    exit(1);
  }

  /* check for cuda capable device */
  CUDA_CHECK(cudaGetDeviceCount(&gpucount));
  if (gpucount < 1) {
    printf("Error, couldn't find a cuda capable device\n");
  }

  /* get the information about the port from snf, including data ring */
  SNF_CHECK(snf_ring_portinfo(hring, &snf_pi));
  /* register the snf recv ring with cuda */
  CUDA_CHECK(cudaHostRegister((void *)snf_pi.data_addr, snf_pi.data_size, 
			      cudaHostRegisterMapped));
  printf("Registered snf recv ring with cuda.\n");

  /* Allocate some memory on both the gpu device and the host */
  CUDA_CHECK(cudaMalloc((void **)&gpu_ring_addr_d, snf_pi.data_size));
  CUDA_CHECK(cudaMalloc((void **)&recv_reqs_d, 
			sizeof(struct snf_recv_req) * NUM_PACKETS));
  CUDA_CHECK(cudaHostAlloc((void **)&recv_reqs_h, 
			   sizeof(struct snf_recv_req) * NUM_PACKETS, 
			   cudaHostAllocDefault));
  CUDA_CHECK(cudaGetSymbolAddress(&mac_vendor_table_hp, mac_vendor_table));
  CUDA_CHECK(cudaMemset(mac_vendor_table_hp, 0, 
			sizeof(mac_vendor_table)));
  CUDA_CHECK(cudaMalloc((void **)&unique_vendor_macs_d,
			sizeof(unsigned int)));
  CUDA_CHECK(cudaHostAlloc((void **)&unique_vendor_macs_h,
			   sizeof(unsigned int),
			   cudaHostAllocDefault));
  CUDA_CHECK(cudaMemset(unique_vendor_macs_d, 0, 
			sizeof(unsigned int)));


  /* Start snf and start capture packets */
  SNF_CHECK(snf_start(hsnf));
  for (i = 0; i < GPU_ITERATIONS; i++) {
    num_pkts = 0;

    /* Get a bunch of packets from snf */
    while ( num_pkts < NUM_PACKETS) {
      rc = snf_ring_recv_many(hring, -1, &recv_reqs_h[num_pkts], 
			      NUM_PACKETS - num_pkts, &nreqs, NULL);
      if (rc == EINTR || rc == EAGAIN) {
	continue;
      }
      num_pkts += nreqs;
    }

    /* DMA the bunch to the gpu */
    rc = dma_to_gpu(recv_reqs_h, recv_reqs_d, num_pkts, 
		    (void *)snf_pi.data_addr, gpu_ring_addr_d);
    SNF_CHECK(rc);

    /* Start the cuda kernel to process the bunch of packets */
    count_macs <<<GPU_BLOCKS, GPU_THREADS_PER_BLOCK>>>
      (recv_reqs_d, snf_pi.data_addr, (uintptr_t)gpu_ring_addr_d, 
       unique_vendor_macs_d);

    /* Return the packets back to snf.  */
    size_to_free = 0;
    for (j = 0; j < num_pkts; j++) {
      size_to_free += recv_reqs_h[j].length_data;
    }
    SNF_CHECK(snf_ring_return_many(hring, size_to_free, NULL));
  }

  /* Copy back the final results from the gpu */
  CUDA_CHECK(cudaMemcpy(unique_vendor_macs_h,
			unique_vendor_macs_d,
			sizeof(unsigned int), 
			cudaMemcpyDeviceToHost));

  printf("Number of unique vendor mac address prefixes: %d\n", 
	 *unique_vendor_macs_h);

  /* release cuda memory and reset device */
  CUDA_CHECK(cudaHostUnregister((void *)snf_pi.data_addr));
  CUDA_CHECK(cudaFree(gpu_ring_addr_d));
  CUDA_CHECK(cudaFree(recv_reqs_d));
  CUDA_CHECK(cudaFreeHost(recv_reqs_h));
  CUDA_CHECK(cudaFreeHost((void *)unique_vendor_macs_h));
  CUDA_CHECK(cudaFree((void *)unique_vendor_macs_d));
  CUDA_CHECK(cudaDeviceReset());

  /* close snf stuff */
  SNF_CHECK(snf_ring_close(hring));
  SNF_CHECK(snf_close(hsnf));

  return 0;
}
