/*************************************************************************
 * The contents of this file are subject to the MYRICOM SNIFFER10G
 * LICENSE (the "License"); User may not use this file except in
 * compliance with the License.  The full text of the License can found
 * in LICENSE.TXT
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See
 * the License for the specific language governing rights and
 * limitations under the License.
 *
 * Copyright 2008 - 2010 by Myricom, Inc.  All rights reserved.
 * Copyright 2015 - 2016 by CSP, Inc.  All rights reserved.
 ***********************************************************************/

#ifndef _WIN32
#include <inttypes.h>
#endif
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <assert.h>
#ifndef _WIN32
#define __USE_GNU
#include <pthread.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <getopt.h>
#else
#include <winsock2.h>
#include <windows.h>
#include <signal.h>
#include "snf_win.h"
#include "getopt.h"
#endif

#include "snf.h"
#ifdef SNF_TREE
#include "mal_timing.h"
#include "mal_trace.h"
#else
#include "mal_simplified.h"
#endif


#ifdef _WIN32
#define C_API __cdecl
#else
#define C_API
#endif

#ifdef __FreeBSD__
#include <sys/param.h>
#if __FreeBSD_version < 702100
#define PTHREAD_SETAFFINITY_NP_MISSING
#else
#include <pthread_np.h>
#define cpu_set_t cpuset_t
#endif /* __FreeBSD_version */
#elif defined(__GLIBC__)
#if !__GLIBC_PREREQ(2,4)
#define PTHREAD_SETAFFINITY_NP_MISSING
#endif
#endif
#define MAX_WORKERS               32
#define MAX_CONC_RECVS            1024

#ifndef _WIN32
#define DBG1(vargs...) if (G.verbose) printf(vargs)
#define DBG2(vargs...) if (G.verbose > 1 && (wrk->num_pkts % 10000 == 0)) printf(vargs)
#define DBG3(vargs...) if (G.verbose > 2 && (wrk->num_pkts % 10000 == 0)) printf(vargs)
#else
#define DBG1(...) if (G.verbose) printf(__VA_ARGS__)
#define DBG2(...) if (G.verbose > 1) printf(__VA_ARGS__)
#define DBG3(...) if (G.verbose > 2) printf(__VA_ARGS__)
#endif

static inline int pkt_validate(void *pkt, uint32_t length);

#define TV_TO_US(tv) ((tv)->tv_sec * 1000000 + (tv)->tv_usec)

struct worker {
  int           worker_id;
  int           cpu_id;
  snf_ring_t    hring;
  snf_inject_t  hinj;
#ifndef _WIN32
  pthread_t thread_id;
#else
  HANDLE thread_id;
#endif
  int should_exit;
  uint64_t num_pkts;
  uint64_t num_bytes;
  uint64_t num_pkts_last;
  uint64_t num_bytes_last;
  uint64_t num_pkts_inject;
};
typedef struct worker worker_t;

struct G {
  snf_handle_t hsnf;
  snf_netdev_reflect_t href;
  unsigned snap_print;
  uint64_t num_pkts;
  uint64_t num_bytes[MAX_WORKERS];
  uint64_t num_pkts_expected;
  uint64_t exit_usecs;
  unsigned duration;
  unsigned custom_hash_rr;
  int64_t  delay_nsecs;
  unsigned nreqs;
  unsigned validate;
  unsigned verbose;
  int      recv_wait;
  int      reinject;
  int      copyout;
  int      recv_many;
  int      netdev_reflect;
  int      check_ooo;
  unsigned num_workers;
  unsigned num_workers_ready;
  unsigned expired;
  uint64_t cpumask;
  worker_t worker[MAX_WORKERS];
  int usingArray;
  int arrayOfCPUs[MAX_WORKERS];
#ifndef _WIN32
  pthread_mutex_t wrk_mtx;
  pthread_cond_t  wrk_cond;
#else
  HANDLE wrk_mtx;
  HANDLE wrk_sem;
#endif
  struct timeval tv_last;
  uint64_t usecs_last;
  uint64_t num_pkts_last;
  uint64_t num_bytes_last;
} G;

struct pkt_hdr {
  uint32_t length;
  uint32_t ofst;
};
typedef struct pkt_hdr pkt_hdr_t;

static void
usage(const char *progname)
{
  printf("usage: %s -w <num_workers> [options]\n\n", progname);
  printf(" options:\n");
  printf("  -p <port number>: Myri10G port number to use.\n");
  printf("     -n <num pkts>: number of packet to receive (default: 0 - infinite)\n");
  printf("                -t: show periodic stats, every second\n");
  printf("                -V: validate incoming packets\n");
  printf("                -v: verbose, or -vv very verbose\n");
  printf("      -d <ring sz>: size of the receive ring in bytes, or megabyte\n");
  printf("     -D <nanosecs>: add <nanosecs> of synthetic processing delay in packet handling\n");
  printf("        -W <msecs>: timeout of <msecs> in blocking receive calls\n");
  printf("     -S <snap len>: display first <snap len> bytes of each packet\n");
  printf("  -R <port number>: reinject every received packet on port <portnum>\n");
  printf("       -A <app_id>: set application ID\n");
  printf("-q <core>[,<core>]: thread ID to which assign the respective workers. Should match <num_workers>\n");
  printf("                -O: do not check out-of-order packet timestamps\n");
  printf("            --help: Display this help, and exit.\n");
  printf("\n");
  exit(1);
}

/*******************************************************************************
 * CPU Utilization Benchmark Abstraction:
 *
 * The following represents the means to track CPU utilization for measuring
 * some entity, like the SNF recv operation. 
 *******************************************************************************/
/*
 * To enable the CPU utilization benchmark tracking abstraction
 * enable the following define.
 */
#undef CPU_UTILIZATION_SUPPORT

struct cpu_utilization_s {
  int          cpu;

  uint32_t     max_seconds;

  mal_cycles_t cyc_per_second;
  mal_cycles_t cyc_per_nano_second;

#define CPU_UTILIZATION_STATE_INIT				0x00000000
#define CPU_UTILIZATION_STATE_WAIT_HIT				0x00000001
#define CPU_UTILIZATION_STATE_WAIT_CALL				0x00000002
#define CPU_UTILIZATION_STATE_ACTIVE				0x00000004
#define CPU_UTILIZATION_STATE_INACTIVE				0x00000008
  uint32_t     tp_state;
  mal_cycles_t tp_cyc_begin;
  mal_cycles_t tp_cyc_end;
  mal_cycles_t tp_cyc_stop;

  int          cyc_started;
  mal_cycles_t cyc_start;
  mal_cycles_t cyc_stop;

  struct cu_event_s {
    char         *name;
    uint64_t     cnt; 
    mal_cycles_t cyc_total;
#define CPU_UTILIZATION_EVENTS_MAX	3
# define CPU_UTILIZATION_EVENTS_HIT	0
# define CPU_UTILIZATION_EVENTS_MISS	1
# define CPU_UTILIZATION_EVENTS_IGNORE	2
  } events[CPU_UTILIZATION_EVENTS_MAX];

#define CPU_UTILIZATION_FLAG_DEFAULT				0x00000000
#define CPU_UTILIZATION_FLAG_BEGIN_TEST_PERIOD_ON_FIRST_HIT	0x00000001
#define CPU_UTILIZATION_FLAG_BEGIN_TEST_PERIOD_ON_FIRST_CALL	0x00000002 //default is first hit
#define CPU_UTILIZATION_FLAG_TEST_PERIOD_MAX_SECONDS		0x00000004
#define CPU_UTILIZATION_FLAG_DUMP_STDIO				0x00000008
#define CPU_UTILIZATION_FLAG_DUMP_FILE				0x00000010
#define CPU_UTILIZATION_FLAG_DEFAULT_FLAGS			(CPU_UTILIZATION_FLAG_BEGIN_TEST_PERIOD_ON_FIRST_HIT|CPU_UTILIZATION_FLAG_DUMP_FILE|CPU_UTILIZATION_FLAG_TEST_PERIOD_MAX_SECONDS)
  uint32_t     flags;

  void         (*cpu_utilization_stop_func)(struct cpu_utilization_s *cu, int event, int nrecs);
};

typedef struct cpu_utilization_s cpu_utilization_t;

cpu_utilization_t cpu_utilization[MAX_WORKERS];

#if defined(CPU_UTILIZATION_SUPPORT)
static void cpu_utilization__measurement_begin(cpu_utilization_t *cu);
static void cpu_utilization__measurement_end(cpu_utilization_t *cu);

static void
cpu_utilization__measurement_start(cpu_utilization_t *cu)
{
  cu->cyc_started = 1;
  cu->cyc_start = mal_get_cycles();
}

static void
_cpu_utilization__measurement_stop_active(cpu_utilization_t *cu, int event, int nrecs)
{
  mal_cycles_t cycles;

  if( (cu->cyc_stop = mal_get_cycles()) >= cu->tp_cyc_stop ) {
    cpu_utilization__measurement_end(cu);
    return;
  } else if (cu->cyc_started == 0) {
    return; //drop it, cycle not started
  }

  cycles = cu->cyc_stop - cu->cyc_start;

  if (event == 0) {
    cu->events[CPU_UTILIZATION_EVENTS_HIT].cnt += nrecs;
    cu->events[CPU_UTILIZATION_EVENTS_HIT].cyc_total += cycles;
  } else if (event == EAGAIN) {
    cu->events[CPU_UTILIZATION_EVENTS_MISS].cnt += nrecs;
    cu->events[CPU_UTILIZATION_EVENTS_MISS].cyc_total += cycles;
  } else {
    cu->events[CPU_UTILIZATION_EVENTS_IGNORE].cnt += nrecs;
    cu->events[CPU_UTILIZATION_EVENTS_IGNORE].cyc_total += cycles;
  }

  cu->cyc_started = 0;
}

static void
_cpu_utilization__measurement_stop_inactive(cpu_utilization_t *cu, int event, int nrecs)
{
  return;
}

static void
_cpu_utilization__measurement_stop_wait(cpu_utilization_t *cu, int event, int nrecs)
{
  if (cu->tp_state == CPU_UTILIZATION_STATE_WAIT_HIT) {
    if (event != 0) {
      return;
    }
  } else if (cu->tp_state != CPU_UTILIZATION_STATE_WAIT_CALL) {
    return;
  }

  cpu_utilization__measurement_begin(cu);
}

#define	cpu_utilization__measurement_stop(cu, event, nrecs)  (cu)->cpu_utilization_stop_func(cu, event, nrecs)

static void
cpu_utilization__measurement_begin(cpu_utilization_t *cu)
{
  if (!(cu->tp_state & (CPU_UTILIZATION_STATE_WAIT_HIT|CPU_UTILIZATION_STATE_WAIT_CALL))) {
    return;
  }

  cu->tp_state = CPU_UTILIZATION_STATE_ACTIVE;
  cu->cpu_utilization_stop_func = _cpu_utilization__measurement_stop_active;

  cu->tp_cyc_begin = mal_get_cycles();

  if (cu->flags & CPU_UTILIZATION_FLAG_TEST_PERIOD_MAX_SECONDS) {
    if (cu->max_seconds == 0) {
      cu->tp_cyc_stop = ~0;
    } else {
      cu->tp_cyc_stop = cu->tp_cyc_begin + (cu->cyc_per_second * cu->max_seconds);
    }
  } else {
    cu->tp_cyc_stop = ~0;
  }
}

static void
cpu_utilization__measurement_end(cpu_utilization_t *cu)
{
  if (!(cu->tp_state & (CPU_UTILIZATION_STATE_ACTIVE))) {
    return;
  }

  cu->tp_state = CPU_UTILIZATION_STATE_INACTIVE;
  cu->cpu_utilization_stop_func = _cpu_utilization__measurement_stop_inactive;

  cu->tp_cyc_end = mal_get_cycles();
}

static void
_cpu_utilization__dump_stats(cpu_utilization_t *cu, int evt, char *output, int output_max_size)
{
  mal_cycles_t tp_cycles;
  mal_cycles_t tp_cyc_end;
  mal_cycles_t tp_cyc_avg;
  mal_cycles_t avg_ns;
  mal_cycles_t event_ns;
  mal_cycles_t event_cyc_avg;
  mal_cycles_t app_ns;
  mal_cycles_t total_ns;
  float tp_cyc_util_p;
  float tp_cyc_total_p;
  float tp_cyc_event_p;
  struct cu_event_s *event;

  if ((output_max_size == 0) || (output == NULL)) {
    return;
  }

  event = &(cu->events[evt]);

  if (!(cu->tp_state & CPU_UTILIZATION_STATE_INACTIVE)) {
    if ( (tp_cyc_end = mal_get_cycles()) > cu->tp_cyc_stop ) {
      tp_cyc_end = cu->tp_cyc_stop;
    }
  } else {
    tp_cyc_end = cu->tp_cyc_end;
  }

  tp_cycles = tp_cyc_end - cu->tp_cyc_begin;
  
  tp_cyc_total_p = tp_cycles;
  tp_cyc_event_p = event->cyc_total;
  tp_cyc_util_p = (tp_cyc_event_p / tp_cyc_total_p) * 100.0;

  if (event->cnt == 0) {
    event_cyc_avg = tp_cyc_avg = avg_ns = event_ns = total_ns = app_ns = 0;
  } else {
    tp_cyc_avg = tp_cycles / event->cnt;
    event_cyc_avg = event->cyc_total / event->cnt;
    avg_ns = event_cyc_avg / cu->cyc_per_nano_second;
    event_ns = event->cyc_total / cu->cyc_per_nano_second;
    total_ns = tp_cycles / cu->cyc_per_nano_second;
    app_ns = total_ns - event_ns;
  }

  (void)snprintf(output, output_max_size, "%10s UTILIZATION STATS: cpu=%2d, cnt=%12lu, "
                 "capture time=%14luns, "
                 "CPU util=%6.2f%%, total cycles=%14lu, total time=%14luns, "
                 "avg cycles=%14lu, avg time=%14luns, app time=%14luns",
         event->name, cu->cpu, event->cnt,
         total_ns,
         tp_cyc_util_p, event->cyc_total, event_ns,
         event_cyc_avg, avg_ns, app_ns);
}

#define CPU_UTILIZATION_DUMP_FILE_NAME "snf_multi_recv.perf"
#define CPU_UTILIZATION_DUMP_OUTPUT_BUFFER_SIZE	4096
static char _cpu_utilization_output[CPU_UTILIZATION_DUMP_OUTPUT_BUFFER_SIZE+1];

static void
cpu_utilization__dump(void)
{
  int worker;
  int flags = 0;
  cpu_utilization_t *cu;
#ifndef _WIN32
  FILE *dump_file = NULL;
#endif /* _WIN32 */

  for (worker=0; worker<MAX_WORKERS; worker++) {
    cu = &(cpu_utilization[worker]);
    if (!(cu->tp_state & (CPU_UTILIZATION_STATE_ACTIVE|CPU_UTILIZATION_STATE_INACTIVE))) {
      continue;
    }

    if (flags == 0 ) {
      if (cu->flags & CPU_UTILIZATION_FLAG_DUMP_FILE) {
        flags |= CPU_UTILIZATION_FLAG_DUMP_FILE;
      }

      if (cu->flags & CPU_UTILIZATION_FLAG_DUMP_STDIO) {
        flags |= CPU_UTILIZATION_FLAG_DUMP_STDIO;
      }

      if (flags == 0) {
        continue;
      }

      if (flags & CPU_UTILIZATION_FLAG_DUMP_FILE) {
#ifndef _WIN32
        if ((dump_file = fopen(CPU_UTILIZATION_DUMP_FILE_NAME, "w")) == NULL) {
          flags &= ~CPU_UTILIZATION_FLAG_DUMP_FILE;
        }
#else
      /* not supported */;
#endif /* _WIN32 */
      }
    }

    _cpu_utilization__dump_stats(cu, CPU_UTILIZATION_EVENTS_HIT, _cpu_utilization_output, CPU_UTILIZATION_DUMP_OUTPUT_BUFFER_SIZE);
    if (flags & CPU_UTILIZATION_FLAG_DUMP_STDIO) {
      printf("%s\n",_cpu_utilization_output);
    }
    if (flags & CPU_UTILIZATION_FLAG_DUMP_FILE) {
#ifndef _WIN32
      fprintf(dump_file,"%s\n",_cpu_utilization_output);
#else
      /* not supported */;
#endif /* _WIN32 */
    }

    _cpu_utilization__dump_stats(cu, CPU_UTILIZATION_EVENTS_MISS, _cpu_utilization_output, CPU_UTILIZATION_DUMP_OUTPUT_BUFFER_SIZE);
    if (flags & CPU_UTILIZATION_FLAG_DUMP_STDIO) {
      printf("%s\n",_cpu_utilization_output);
    }
    if (flags & CPU_UTILIZATION_FLAG_DUMP_FILE) {
#ifndef _WIN32
      fprintf(dump_file,"%s\n",_cpu_utilization_output);
#else
      /* not supported */;
#endif /* _WIN32 */
    }

    _cpu_utilization__dump_stats(cu, CPU_UTILIZATION_EVENTS_IGNORE, _cpu_utilization_output, CPU_UTILIZATION_DUMP_OUTPUT_BUFFER_SIZE);
    if (flags & CPU_UTILIZATION_FLAG_DUMP_STDIO) {
      printf("%s\n\n",_cpu_utilization_output);
    }
    if (flags & CPU_UTILIZATION_FLAG_DUMP_FILE) {
#ifndef _WIN32
      fprintf(dump_file,"%s\n\n",_cpu_utilization_output);
#else
      /* not supported */;
#endif /* _WIN32 */
    }
  }

  if (flags & CPU_UTILIZATION_FLAG_DUMP_FILE) {
#ifndef _WIN32
    fflush(dump_file);
    fclose(dump_file);
#else
    /* not supported */;
#endif /* _WIN32 */
  }
}

static void
cpu_utilization__init(cpu_utilization_t *cu, int cpu, uint32_t flags, uint32_t max_seconds)
{
  mal_cycles_t ns_in_sec;

  memset(cu, 0, sizeof(cpu_utilization_t));

  if (CPU_UTILIZATION_FLAG_DEFAULT == flags) {
    cu->flags = CPU_UTILIZATION_FLAG_DEFAULT_FLAGS;
  } else {
    cu->flags = flags;
  }

  if (!(flags & CPU_UTILIZATION_FLAG_BEGIN_TEST_PERIOD_ON_FIRST_CALL)) {
    cu->flags |= CPU_UTILIZATION_FLAG_BEGIN_TEST_PERIOD_ON_FIRST_HIT;
    cu->tp_state = CPU_UTILIZATION_STATE_WAIT_HIT;
  } else {
    cu->flags &= ~CPU_UTILIZATION_FLAG_BEGIN_TEST_PERIOD_ON_FIRST_HIT;
    cu->tp_state = CPU_UTILIZATION_STATE_WAIT_CALL;
  }

  if (cu->flags & CPU_UTILIZATION_FLAG_TEST_PERIOD_MAX_SECONDS) {
    if ((cu->max_seconds = max_seconds) == 0) {
      cu->flags &= ~CPU_UTILIZATION_FLAG_TEST_PERIOD_MAX_SECONDS;
    }
  }

  mal_cycles_counter_init();
  cu->cyc_per_second = mal_cycles_per_second();

  ns_in_sec = (1000 * 1000 * 1000);
  cu->cyc_per_nano_second = cu->cyc_per_second / ns_in_sec;

  cu->cpu_utilization_stop_func = _cpu_utilization__measurement_stop_wait;

  cu->cpu = cpu;

  cu->events[CPU_UTILIZATION_EVENTS_HIT].name = "HIT";
  cu->events[CPU_UTILIZATION_EVENTS_MISS].name = "MISS";
  cu->events[CPU_UTILIZATION_EVENTS_IGNORE].name = "IGNORE";
}
#else /* defined(CPU_UTILIZATION_SUPPORT) */
#define cpu_utilization__init(cu, cpu, flags, max_seconds)
#define cpu_utilization__dump()
#define cpu_utilization__measurement_end(cu)
#define cpu_utilization__measurement_begin(cu)
#define	cpu_utilization__measurement_stop(cu, event, nrecs)
#define cpu_utilization__measurement_start(cu)
#endif /* defined(CPU_UTILIZATION_SUPPORT) */

static void
dump_data(unsigned char *pg, int len)
{
  int i = 0;
  while (i < len) {
    printf
      ("%04d: %02x %02x %02x %02x %02x %02x %02x %02x - %02x %02x %02x %02x %02x %02x %02x %02x\n",
       i, pg[i], pg[i + 1], pg[i + 2], pg[i + 3], pg[i + 4], pg[i + 5],
       pg[i + 6], pg[i + 7], pg[i + 8], pg[i + 9], pg[i + 10], pg[i + 11],
       pg[i + 12], pg[i + 13], pg[i + 14], pg[i + 15]);
    i += 16;
  }
}

static inline int
pkt_validate(void *pkt, uint32_t length)
{
  pkt_hdr_t *hdr = (pkt_hdr_t *)pkt;
  uint32_t *pld_ptr;
  uint32_t hdr_ofst = ntohl(hdr->ofst) / 4;
  uint32_t i;

  if (length != ntohl(hdr->length)) {
    printf("packet length missmatch, in mesg:%u, in packet:%u\n",
            length, hdr->length);
    dump_data((unsigned char *)pkt, 32);
    return 1;
  }

  pld_ptr = &((uint32_t *)pkt)[2];
  for (i = 0; i < length / 4 - 2; i++) {
    if (*pld_ptr != hdr_ofst) {
      printf("packet data missmatch at ofst %u, expected:%u, actual:%u\n",
            i, hdr_ofst, *pld_ptr);
      dump_data((unsigned char *)pkt, length);
      return 1;
    }
    pld_ptr++;
    hdr_ofst++;
  }
  return 0;
}

void
stats(void)
{
  worker_t *wrk;
  unsigned int w;
  uint64_t num_pkts = 0;
  int64_t num_bytes = 0;
  int have_nic_stats = 0;
  int have_nic_stats_inj = 0;
  struct snf_ring_stats stats;
  struct snf_inject_stats istats;
  for (w = 0; w < G.num_workers; w++) {
    wrk = &G.worker[w];
    num_pkts += wrk->num_pkts;
    num_bytes += wrk->num_bytes;
    if (wrk->hring == NULL || snf_ring_getstats(wrk->hring, &stats)) {
      printf("[%u] %" PRIu64 " pkts, %" PRIu64" bytes\n", w, wrk->num_pkts,
         wrk->num_bytes);
    }
    else {
      have_nic_stats = 1;
      printf("Ring %d Packets received,    ring:%12" PRIu64 ", app:%12" PRIu64
             ", overflow:%12" PRIu64 "\n", w, stats.ring_pkt_recv, wrk->num_pkts,
             stats.ring_pkt_overflow);
    }
    if (wrk->hinj && 0 == snf_inject_getstats(wrk->hinj, &istats)) {
      printf("Ring %d Packets reinjected,  ring:%12" PRIu64 "\n", w,
          istats.inj_pkt_send);
      have_nic_stats_inj = 1;
    }
  }

  if (have_nic_stats) {
    /* NIC-level stats */
    printf("SNF dropped overflow SW  : %12" PRIu64 "\n", stats.snf_pkt_overflow);
    printf("NIC dropped overflow HW  : %12" PRIu64 "\n", stats.nic_pkt_overflow);
    printf("NIC dropped bad PHY/CRC  : %12" PRIu64 "\n", stats.nic_pkt_bad);
  //printf("NIC dropped PKTs Filtered: %12" PRIu64 "\n", stats.nic_pkt_dropped);
  }
  if (have_nic_stats_inj)
    printf("NIC send packets       : %12" PRIu64 "\n\n", istats.nic_pkt_send);

  if (num_pkts) {
    printf("\n");
    printf("Total packets:                app:%12" PRIu64 "\n", num_pkts);
    printf("Average Packet Length:        app:%12" PRIu64 " bytes\n\n",
            num_bytes / num_pkts);
    printf("Total bytes:                  app:%12" PRIu64 " (%" PRIu64 " MB)\n",
            num_bytes, num_bytes / 1024 / 1024);
    num_bytes = stats.nic_bytes_recv;
    num_bytes -= (8 /* HW header */ + 4 /* CRC */) * stats.nic_pkt_recv;
    if (num_bytes > 0)
      printf("Total raw bytes + HW align    nic:%12" PRIu64 " (%" PRIu64 " MB)\n",
             num_bytes, num_bytes / 1024 / 1024);
  }
}

static int
cpu_getmask_ith(unsigned cpumask, int id)
{
  int i, i_found = 0;
  unsigned mask;
  for (i = 0; i < sizeof(cpumask)*8; i++) {
    mask = (1U<<i);
    if (mask & cpumask) {
      if (i_found == id) /* cpu i to be used for id */
        return i;
      i_found++;
    }
  }
  /* not found, return any cpu */
  return -1;
}

void
print_periodic_stats(void)
{
  float delta_secs;
  uint64_t delta_pkts;
  uint64_t delta_bytes;
  uint32_t pps;
  float gbps;
  float bps;
  struct timeval tv_now;
  uint64_t usecs_now;
  worker_t *wrk;
  unsigned int w;
  uintptr_t q_size;
  uint64_t num_bytes;
  uint64_t wrk_delta_pkts[MAX_WORKERS];
  struct snf_ring_qinfo qinfo;
  float q_maxused = .0, q_used;

  gettimeofday(&tv_now, NULL);
  usecs_now = TV_TO_US(&tv_now);
  delta_secs = (float)((usecs_now - G.usecs_last) / 1000000.0);

#ifndef _WIN32
  pthread_mutex_lock(&G.wrk_mtx);
#else
  WaitForSingleObject(G.wrk_mtx, INFINITE);
#endif

  /* Threads only update pkts */
  delta_pkts = G.num_pkts - G.num_pkts_last;
  G.num_pkts_last = G.num_pkts;

  num_bytes = 0;
  for (w = 0; w < G.num_workers; w++) {
    wrk = &G.worker[w];

    /* it's possible that the periodic time fires before all rings are started */
    if (wrk->hring == NULL)
      continue;

    wrk_delta_pkts[w] = wrk->num_pkts - wrk->num_pkts_last;
    wrk->num_pkts_last = wrk->num_pkts;
    num_bytes += G.num_bytes[w];
    snf_ring_recv_qinfo(wrk->hring, &qinfo);
    q_size = qinfo.q_avail + qinfo.q_borrowed + qinfo.q_free;
    q_used = (1 - (float) qinfo.q_free / q_size);
    if (q_used > q_maxused) {
      q_maxused = q_used;
    }
  }
  delta_bytes = num_bytes - G.num_bytes_last;
  if (delta_pkts) {
    /* Update the last count only if we are going to print stats below.
     * Otherwise, num_pkts and num_bytes may get out of sync.
     * For example, delta_pkts > 0 and delta_bytes = 0.
     * The same issue applies to the saved timestamps.
     */
    G.num_bytes_last = num_bytes;
    G.usecs_last = usecs_now;
    G.tv_last = tv_now;
  }

  if (G.duration && G.expired == 0) {
    if (!G.exit_usecs)
      G.exit_usecs = usecs_now + G.duration * 1000000;
    else if (usecs_now >= G.exit_usecs) {
      G.expired = 1;
    }
  }
#ifndef _WIN32
  pthread_mutex_unlock(&G.wrk_mtx);
#else
  ReleaseMutex(G.wrk_mtx);
#endif

  if (delta_pkts) {
    pps = (uint32_t)(delta_pkts / delta_secs);
    bps = ((float) delta_bytes * 8) / delta_secs;
    gbps = (float)(bps / 1000000000.0);

    for (w = 0; w < G.num_workers; w++) {
      DBG1("[%u] %" PRIu64 " ", w, wrk_delta_pkts[w]);
    }

    printf
        ("%" PRIu64 " pkts (%" PRIu64 "B) in %.3f secs (%u pps), Avg Pkt: %"
         PRIu64 ", Q-fullness=%4.1f%% BW (Gbps): %6.3f\n",
         delta_pkts, delta_bytes, delta_secs,
         pps, delta_bytes / delta_pkts, q_maxused * 100.0, gbps);
    fflush(stdout);

  }
}

void C_API
sigexit(int sig)
{
#ifndef _WIN32
  static pthread_mutex_t exit_mtx = PTHREAD_MUTEX_INITIALIZER;
  static int once = 0;

  pthread_mutex_lock(&exit_mtx);
  cpu_utilization__dump();
  if (!once) {
    once = 1;
    stats();
    fflush(stdout);
  }
  pthread_mutex_unlock(&exit_mtx);
  if (sig >= 0)
    exit(0);
#else
  static volatile LONG once = 0;
  LONG l;

  l = InterlockedIncrement(&once);
  cpu_utilization__dump();
  if (l == 1) {
    stats();
    fflush(stdout);
  }
  if (sig >= 0)
    exit(0);
#endif
}

#ifndef _WIN32
void
sigalrm(int sig)
{
  print_periodic_stats();
  alarm(1);
}
#else
DWORD
stats_thread(LPVOID lpParameter)
{
  for (;;) {
    Sleep(1000);
    print_periodic_stats();
  }
}
#endif

#ifndef _WIN32
void *
#else
DWORD WINAPI
#endif
work_thread(void *stuff)
{
  int worker_id;
  worker_t *wrk;
  int rc;
  int i;
#ifndef _WIN32
  sigset_t sigmask;
#endif
  uint64_t num_pkts = 0;
  uint64_t last_tstamp = 0;
  void *copybuf = NULL;
  struct snf_recv_req reqs[32];
  int nreqs = 1;
  uint32_t data_borrowed_len;
  struct snf_ring_qinfo qinfo;
  mal_cycles_t cyc_start, cyc_itvl;
  mal_cycles_t cycles_per_second;
  uint64_t update_threshold;
#if defined(CPU_UTILIZATION_SUPPORT)
  int worker_cpu_id = 0;
#endif /* defined(CPU_UTILIZATION_SUPPORT) */

  /* For many-ring tests, per-ring PPS may be far smaller than 100000. That
   * would make periodic stats very choppy. So reduce the threshold based on
   * the number of rings. At 32 rings, 100000 / 3 seems to work well for both
   * 60B and 1000B frames.
   */
  update_threshold = 100000 / (1 + G.num_workers / 16);

  wrk = (worker_t *)stuff;
  wrk->hring = NULL;
  worker_id = wrk->worker_id;
#ifdef _WIN32
  if (wrk->cpu_id >= 0)
  {
    SYSTEM_INFO info;
    DWORD tmp;

    GetSystemInfo(&info);
    if(G.usingArray)
      tmp = G.arrayOfCPUs[worker_id];
    else
      tmp = wrk->cpu_id % info.dwNumberOfProcessors;
    SetThreadIdealProcessor(GetCurrentThread(), tmp);
#if defined(CPU_UTILIZATION_SUPPORT)
    worker_cpu_id = tmp;
#endif /* defined(CPU_UTILIZATION_SUPPORT) */
    printf("Worker %d running on CPU %d\n", worker_id, wrk->cpu_id);
  }
#elif !defined(PTHREAD_SETAFFINITY_NP_MISSING)
  if (wrk->cpu_id >= 0)
  {
    if(G.usingArray)
      wrk->cpu_id = G.arrayOfCPUs[worker_id];

    cpu_set_t cpuset;
    CPU_ZERO(&cpuset);
    CPU_SET(wrk->cpu_id, &cpuset);
    pthread_setaffinity_np(pthread_self(), sizeof(cpuset), (void *)&cpuset);
#if defined(CPU_UTILIZATION_SUPPORT)
    worker_cpu_id = wrk->cpu_id;
#endif /* defined(CPU_UTILIZATION_SUPPORT) */
    printf("Worker %d running on CPU %d\n", worker_id, wrk->cpu_id);
  }
#endif

#ifndef _WIN32
  /* mask signals except SIGINT to allow ^C */
  sigfillset(&sigmask);
#if !defined __FreeBSD__
  sigdelset(&sigmask, SIGINT);
#endif
  rc = pthread_sigmask(SIG_BLOCK, &sigmask, NULL);
#endif
 
  cpu_utilization__init(&(cpu_utilization[worker_id]), worker_cpu_id, CPU_UTILIZATION_FLAG_DEFAULT, 30);

  mal_cycles_counter_init();
  cycles_per_second = mal_cycles_per_second();
  cyc_itvl = (mal_cycles_t) (G.delay_nsecs * cycles_per_second / 1000000000.0);

  rc = snf_ring_open(G.hsnf, &wrk->hring);
  if (rc) {
    fprintf(stderr, "[%d] Can't open ring %d: %s\n", worker_id, worker_id, strerror(rc));
    goto done_thread;
  }

  if (G.reinject >= 0) {
    rc = snf_inject_open(G.reinject, 0, &wrk->hinj);
    if (rc) {
      fprintf(stderr, "[%d] Can't inject on ring %d\n", worker_id, G.reinject);
      goto done_thread;
    }
  }
  if (G.copyout) {
    copybuf = calloc(1, 16*1024);
    if (copybuf == NULL) {
      fprintf(stderr, "[%d] Can't allocate copyout buffer\n", worker_id);
      goto done_thread;
    }
  }

#ifndef _WIN32
  pthread_mutex_lock(&G.wrk_mtx);
  if (++G.num_workers_ready == G.num_workers) {
    pthread_mutex_unlock(&G.wrk_mtx);
    pthread_cond_broadcast(&G.wrk_cond);
    rc = snf_start(G.hsnf);
    if (rc) {
      fprintf(stderr, "[%d] Can't start snf: %s\n", worker_id, strerror(rc));
      goto done_thread;
    }
  }
  else {
    pthread_cond_wait(&G.wrk_cond, &G.wrk_mtx);
    pthread_mutex_unlock(&G.wrk_mtx);
  }
#else
  WaitForSingleObject(G.wrk_mtx, INFINITE);
  if (++G.num_workers_ready == G.num_workers) {
    ReleaseMutex(G.wrk_mtx);
    ReleaseSemaphore(G.wrk_sem, G.num_workers, NULL);
    rc = snf_start(G.hsnf);
    if (rc) {
      fprintf(stderr, "[%d] Can't start snf: %s\n", worker_id, strerror(rc));
      goto done_thread;
    }
  }
  else {
    ReleaseMutex(G.wrk_mtx);
    WaitForSingleObject(G.wrk_sem, INFINITE);
  }
#endif

  if (G.recv_many) {
    int cnt;
    rc = snf_ring_portinfo_count(wrk->hring, &cnt);
    if (rc != 0) {
      fprintf(stderr, "[%d] unexpected snf_ring_portinfo_count error %d (%s)\n",
            worker_id, rc, strerror(rc));
      goto done_thread;
    }
    if (cnt != 1) {
      fprintf(stderr, "[%d] Test only supports borrow option on non-aggregated ports (cnt=%d)\n",
              worker_id, cnt);
      goto done_thread;
    }
  }

  DBG1("[%u] started\n", worker_id);

  cyc_start = 0;

  while (!wrk->should_exit) {
    cpu_utilization__measurement_start(&(cpu_utilization[worker_id]));
    if (G.recv_many)
      rc = snf_ring_recv_many(wrk->hring, G.recv_wait, reqs, 32, &nreqs, NULL);
    else {
      nreqs = 1;
      rc = snf_ring_recv(wrk->hring, G.recv_wait, &reqs[0]);
    }
    cpu_utilization__measurement_stop(&(cpu_utilization[worker_id]), rc, nreqs);

    if (rc == 0) {
      data_borrowed_len = 0;
      for (i = 0; i < nreqs; i++) {
        data_borrowed_len += reqs[i].length_data;
        wrk->num_bytes += reqs[i].length;
        wrk->num_pkts++;
        num_pkts++;

#ifdef SNF_TREE
        tracepoint(snf, code_marker, wrk->num_pkts, "worker id packet");
#endif
        if (G.check_ooo && (reqs[i].timestamp < last_tstamp ||
                            reqs[i].timestamp == 0))
          printf("out-of-order packet timestamps: req index %d last_tstamp=%llu"
                 " current_tsamp=%llu difference %lld\n", i, (long long unsigned)last_tstamp,
                 (long long unsigned)reqs[i].timestamp,(long long) last_tstamp - reqs[i].timestamp);
        last_tstamp = reqs[i].timestamp;
        if (copybuf != NULL)
          memcpy(copybuf, reqs[i].pkt_addr, reqs[i].length);
        if (G.reinject >= 0) {
          rc = snf_inject_send(wrk->hinj, 10, 0, reqs[i].pkt_addr, reqs[i].length);
          if (!rc)
            wrk->num_pkts_inject++;
        }
        DBG2("[%d] received %d bytes (timestamp %lu) (so far %llu)\n",
	     worker_id, reqs[i].length, reqs[i].timestamp,
	     (unsigned long long) wrk->num_pkts);
	if (G.netdev_reflect) {
	  rc = snf_netdev_reflect(G.href, reqs[i].pkt_addr, reqs[i].length);
	  if (rc != 0) {
	    fprintf(stderr, "[%d] unexpected snf_netdev_reflect error %d (%s)\n",
	      worker_id, rc, strerror(rc));
	    break;
	  }
	}
        if (G.snap_print > 0) {
          dump_data(reqs[i].pkt_addr,
             (reqs[i].length > G.snap_print) ? G.snap_print : reqs[i].length);
        }
      }
      if (G.delay_nsecs > 0) {
        cyc_start = mal_get_cycles();
        while (mal_get_cycles() < cyc_start + cyc_itvl);
      }
      if (G.recv_many) {
        rc = snf_ring_return_many(wrk->hring, data_borrowed_len, &qinfo);
        if (rc != 0) {
          fprintf(stderr, "[%d] unexpected snf_ring_return_many error %d (%s)\n",
            worker_id, rc, strerror(rc));
          break;
        }
#if 0
        static int first = 100;
        if (first) {
          printf("q_borrowed=%8u q_avail=%8u q_free=%8u\n",
              (unsigned) qinfo.q_borrowed, (unsigned) qinfo.q_avail, (unsigned) qinfo.q_free);
          first--;
        }
#endif
      }
      if (num_pkts >= update_threshold)
        goto update_num_pkts;
    }
    else if (rc == EINTR) {
      if (G.expired)
        goto update_num_pkts;
      else
        continue;
    }
    else if (rc == EAGAIN) { /* timed out */
update_num_pkts:
#ifndef _WIN32
      pthread_mutex_lock(&G.wrk_mtx);
#else
      WaitForSingleObject(G.wrk_mtx, INFINITE);
#endif
      G.num_pkts += num_pkts;
      G.num_bytes[worker_id] = wrk->num_bytes;
      num_pkts = 0;
      if (G.num_pkts >= G.num_pkts_expected || G.expired) {
        unsigned int w;
        DBG1("[%d] all %llu received\n", worker_id,
                                       (unsigned long long) G.num_pkts);
        /* kill off all the workers */
        for (w = 0; w < G.num_workers; w++) {
          if (worker_id != G.worker[w].worker_id) {
            DBG1("[%d] killing worker %d\n", worker_id, G.worker[w].worker_id);
#ifndef _WIN32
            pthread_cancel(G.worker[w].thread_id);
#else
	    G.worker[w].should_exit = 1;
#endif
          }
        }
#ifndef _WIN32
        pthread_mutex_unlock(&G.wrk_mtx);
#else
	ReleaseMutex(G.wrk_mtx);
#endif
        break;
      }
      else {
#ifndef _WIN32
        pthread_mutex_unlock(&G.wrk_mtx);
#else
	ReleaseMutex(G.wrk_mtx);
#endif
      }
    }
    else {
      fprintf(stderr, "[%d] unexpected snf_ring_recv error %d (%s)\n",
          worker_id, rc, strerror(rc));
      break;
    }
  }

done_thread:
  if (copybuf)
    free(copybuf);
  DBG1("[%d] exiting\n", worker_id);
  return 0;
}

int
C_API
main(int argc, char **argv)
{
  int rc = 0;
  char c;
  int periodic_stats = 0;
  int portnum = 0;
  int32_t app_id = -1;
  uint64_t dataring_sz = 0;
  int w;
  worker_t *wrk;
  int option_index = 0;
  static struct option long_options[] = {
    {"help", no_argument, 0, '?'},
    {0,0,0,0}
  };
#ifndef _WIN32
  void *thread_rc;
#endif

  /* set defaults */
  G.num_workers = 0;
  G.recv_wait = 500;
  G.num_pkts_expected = -1;
  G.duration = 0;
  G.delay_nsecs = 0;
  G.custom_hash_rr = 0;
  G.snap_print = 0;
  G.nreqs = 1;
  G.validate = 0;
  G.verbose = 0;
  G.reinject = -1;
  G.copyout = 0;
  G.cpumask = ~0ULL;
  G.recv_many = 0;
  G.usingArray = 0;
  G.check_ooo = 1;

  /* get args */
  while ((c = getopt_long(argc, argv, "p:b:tvVR:S:n:T:d:w:D:W:CMNA:q:O?", long_options, &option_index)) != -1) {
    if (c == 'b' || c == 'p') {
      portnum = strtoul(optarg, NULL, 0);
    } else if (c == 't') {
      periodic_stats = 1;
    } else if (c == 'A') {
      app_id = strtoul(optarg, NULL, 0);
    } else if (c == 'V') {
      G.validate = 1;
    } else if (c == 'n') {
      G.num_pkts_expected = strtoul(optarg, NULL, 0);
    } else if (c == 'T') {
      G.duration = strtoul(optarg, NULL, 0);
    } else if (c == 'v') {
      G.verbose++;
    } else if (c == 'R') {
      G.reinject = strtoul(optarg, NULL, 0);
    } else if (c == 'S') {
      G.snap_print = strtoul(optarg, NULL, 0);
    } else if (c == 'd') {
      dataring_sz = strtoull(optarg, NULL, 0);
    } else if (c == 'D') {
      G.delay_nsecs = strtoll(optarg, NULL, 0);
    } else if (c == 'C') {
      G.copyout = 1;
    } else if (c == 'M') {
      G.recv_many = 1;
    } else if (c == 'N') {
      G.netdev_reflect = 1;
    } else if (c == 'O') {
      G.check_ooo = 0;
    } else if (c == 'w') {
      long long cpumask;
      int n;
      if ((n = sscanf(optarg, "%u:%llx", &G.num_workers, &cpumask)) < 1)
        usage(argv[0]);
      if (G.num_workers > MAX_WORKERS) {
        printf("Invalid number of worker threads (%d, max:%d)\n",
            G.num_workers, MAX_WORKERS);
        exit(1);
      }
      if (n >= 2)
        G.cpumask = cpumask;
    } else if (c == 'q') {
      char *endptr;
      int coreID;
      int i;
      for(i= 0; i < MAX_WORKERS; ++i){
        coreID = strtoul(optarg, &endptr, 0);
        G.arrayOfCPUs[i] = coreID;
        if(*endptr == '\0')
          break;
        optarg = endptr + 1;
      }
      G.usingArray = 1;
    } else if (c == 'W') {
      G.recv_wait = (int)strtoull(optarg, NULL, 0);
    } else {
      if (c != '?')
        printf("ERROR: Unknown option: %c\n", c);
      usage(argv[0]);
    }
  }
  if (!G.num_workers) {
    printf("\nERROR: Number of workers/rings not specified with -w option:\n\n");
    usage(argv[0]);
  }

  snf_init(SNF_VERSION_API);

  if (app_id != -1)
    snf_set_app_id(app_id);

#ifdef _WIN32
  if (G.cpumask) {
    SetThreadAffinityMask(GetCurrentThread(), (DWORD_PTR)G.cpumask);
  }
#elif !defined(PTHREAD_SETAFFINITY_NP_MISSING)
  if (G.cpumask) {
    cpu_set_t cpuset;
    int i;
    CPU_ZERO(&cpuset);
    for (i = 0; i < sizeof(G.cpumask) * 8; i++) {
      if (G.cpumask & (1<<i))
        CPU_SET(i, &cpuset);
    }
    pthread_setaffinity_np(pthread_self(), sizeof(cpuset), (void *)&cpuset);
  }
#endif

  rc = snf_open(portnum, G.num_workers, NULL, dataring_sz, -1, &G.hsnf);
  if (rc) {
    errno = rc;
    perror("Can't open snf");
    exit(2);
  }

  if (G.netdev_reflect) {
    rc = snf_netdev_reflect_enable(G.hsnf, &G.href);
    if (rc) {
      errno = rc;
      perror("Can't open snf for reflecting pacekts to netdev");
      exit(2);
    }
  }

#ifndef _WIN32
  if (SIG_ERR == signal(SIGINT, sigexit))
    exit(1);
  if (SIG_ERR == signal(SIGTERM, sigexit))
    exit(1);
#else
  signal(SIGINT, sigexit);
#endif

#ifndef _WIN32
  pthread_mutex_init(&G.wrk_mtx, NULL);
  pthread_cond_init(&G.wrk_cond, NULL);
#else
  G.wrk_mtx = CreateMutex(NULL, FALSE, NULL);
  G.wrk_sem = CreateSemaphore(NULL, 0, G.num_workers, NULL);
#endif
  G.num_workers_ready = 0;

  printf("Capture using %d rings and worker threads with default RSS hashing", G.num_workers);
  if (G.delay_nsecs)
    printf(", each thread spends an %d nanosecs of additional (synthetic) processing\n",
        (int) G.delay_nsecs);
  else
    printf("\n");

  /* initialize the CPU util measurement objects */
  memset(&(cpu_utilization[0]), 0, (sizeof(cpu_utilization_t) * MAX_WORKERS));

  for (w = 0; w < (int)G.num_workers; w++) {
    wrk = &G.worker[w];
    memset(wrk, 0, sizeof(worker_t));
    wrk->worker_id = w;
    wrk->cpu_id = cpu_getmask_ith((unsigned)G.cpumask, w);
    wrk->should_exit = 0;

#ifndef _WIN32
    rc = pthread_create(&wrk->thread_id, NULL, work_thread, (void *)wrk);
    if (rc != 0) {
      perror("pthread_create");
      exit(1);
    }
#else
    wrk->thread_id = CreateThread(NULL, 0, work_thread, wrk, 0, NULL);
    if (wrk->thread_id == NULL) {
      fprintf(stderr, "CreateThread\n");
      exit(1);
    }
#endif
  }

  if (periodic_stats || G.duration) {
#ifndef _WIN32
    if (SIG_ERR == signal(SIGALRM, sigalrm)) {
      exit(1);
    }
#else
    CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE *) stats_thread, NULL, 0, NULL);
#endif
    gettimeofday(&G.tv_last, NULL);
    G.usecs_last = TV_TO_US(&G.tv_last);
#ifndef _WIN32
    alarm(1);
#endif
  }

#ifndef _WIN32
  for (w = 0; w < G.num_workers; w++)
    pthread_join(G.worker[w].thread_id, &thread_rc);
#else
  for (w = 0; w < (int)G.num_workers; w++)
    WaitForSingleObject(G.worker[w].thread_id, INFINITE);
#endif

  sigexit(-1);

  for (w = 0; w < (int)G.num_workers; w++) {
    wrk = &G.worker[w];
    if (wrk->hring) {
      snf_ring_close(wrk->hring);
      wrk->hring = NULL;
    }
    if (wrk->hinj) {
      snf_inject_close(wrk->hinj);
      wrk->hinj = NULL;
    }
  }

  rc = snf_close(G.hsnf);
  if (rc) {
    errno = rc;
    perror("Can't close snf");
  }

  exit(0);
}
