/*************************************************************************
 * The contents of this file are subject to the MYRICOM SNIFFER10G
 * LICENSE (the "License"); User may not use this file except in
 * compliance with the License.  The full text of the License can found
 * in LICENSE.TXT
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See
 * the License for the specific language governing rights and
 * limitations under the License.
 *
 * Copyright 2008 by Myricom, Inc.  All rights reserved.
 * Copyright 2015 - 2016 by CSP, Inc.  All rights reserved.
 ***********************************************************************/
/*
 * Simple program to demonstrate how to receive packets using
 * the Myricom Sniffer API.
 */

#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <signal.h>
#include <assert.h>
#ifndef _WIN32
#include <inttypes.h>
#include <pthread.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <getopt.h>
#else /* _WIN32 */
#include <winsock2.h>
#include <windows.h>
#include "getopt.h"
#include "snf_win.h"
#endif

#ifdef _WIN32
#define C_API __cdecl
#else
#define C_API
#endif

#include "snf.h"
#ifdef SNF_TREE
#include "mal_timing.h"
#else
#include "mal_simplified.h"
#endif

int getout = 0;

#define MAX_WORKERS	1

/*******************************************************************************
 * CPU Utilization Benchmark Abstraction:
 *
 * The following represents the means to track CPU utilization for measuring
 * some entity, like the SNF recv operation. 
 *******************************************************************************/
/*
 * To enable the CPU utilization benchmark tracking abstraction
 * enable the following define.
 */
#undef CPU_UTILIZATION_SUPPORT

struct cpu_utilization_s {
  int          cpu;

  uint32_t     max_seconds;

  mal_cycles_t cyc_per_second;
  mal_cycles_t cyc_per_nano_second;

#define CPU_UTILIZATION_STATE_INIT				0x00000000
#define CPU_UTILIZATION_STATE_WAIT_HIT				0x00000001
#define CPU_UTILIZATION_STATE_WAIT_CALL				0x00000002
#define CPU_UTILIZATION_STATE_ACTIVE				0x00000004
#define CPU_UTILIZATION_STATE_INACTIVE				0x00000008
  uint32_t     tp_state;
  mal_cycles_t tp_cyc_begin;
  mal_cycles_t tp_cyc_end;
  mal_cycles_t tp_cyc_stop;

  int          cyc_started;
  mal_cycles_t cyc_start;
  mal_cycles_t cyc_stop;

  struct cu_event_s {
    char         *name;
    uint64_t     cnt; 
    mal_cycles_t cyc_total;
#define CPU_UTILIZATION_EVENTS_MAX	3
# define CPU_UTILIZATION_EVENTS_HIT	0
# define CPU_UTILIZATION_EVENTS_MISS	1
# define CPU_UTILIZATION_EVENTS_IGNORE	2
  } events[CPU_UTILIZATION_EVENTS_MAX];

#define CPU_UTILIZATION_FLAG_DEFAULT				0x00000000
#define CPU_UTILIZATION_FLAG_BEGIN_TEST_PERIOD_ON_FIRST_HIT	0x00000001
#define CPU_UTILIZATION_FLAG_BEGIN_TEST_PERIOD_ON_FIRST_CALL	0x00000002 //default is first hit
#define CPU_UTILIZATION_FLAG_TEST_PERIOD_MAX_SECONDS		0x00000004
#define CPU_UTILIZATION_FLAG_DUMP_STDIO				0x00000008
#define CPU_UTILIZATION_FLAG_DUMP_FILE				0x00000010
#define CPU_UTILIZATION_FLAG_DEFAULT_FLAGS			(CPU_UTILIZATION_FLAG_BEGIN_TEST_PERIOD_ON_FIRST_HIT|CPU_UTILIZATION_FLAG_DUMP_FILE|CPU_UTILIZATION_FLAG_TEST_PERIOD_MAX_SECONDS)
  uint32_t     flags;

  void         (*cpu_utilization_stop_func)(struct cpu_utilization_s *cu, int event, int nrecs);
};

typedef struct cpu_utilization_s cpu_utilization_t;

cpu_utilization_t cpu_utilization[MAX_WORKERS];

#if defined(CPU_UTILIZATION_SUPPORT)
static void cpu_utilization__measurement_begin(cpu_utilization_t *cu);
static void cpu_utilization__measurement_end(cpu_utilization_t *cu);

static void
cpu_utilization__measurement_start(cpu_utilization_t *cu)
{
  cu->cyc_started = 1;
  cu->cyc_start = mal_get_cycles();
}

static void
_cpu_utilization__measurement_stop_active(cpu_utilization_t *cu, int event, int nrecs)
{
  mal_cycles_t cycles;

  if( (cu->cyc_stop = mal_get_cycles()) >= cu->tp_cyc_stop ) {
    cpu_utilization__measurement_end(cu);
    return;
  } else if (cu->cyc_started == 0) {
    return; //drop it, cycle not started
  }

  cycles = cu->cyc_stop - cu->cyc_start;

  if (event == 0) {
    cu->events[CPU_UTILIZATION_EVENTS_HIT].cnt += nrecs;
    cu->events[CPU_UTILIZATION_EVENTS_HIT].cyc_total += cycles;
  } else if (event == EAGAIN) {
    cu->events[CPU_UTILIZATION_EVENTS_MISS].cnt += nrecs;
    cu->events[CPU_UTILIZATION_EVENTS_MISS].cyc_total += cycles;
  } else {
    cu->events[CPU_UTILIZATION_EVENTS_IGNORE].cnt += nrecs;
    cu->events[CPU_UTILIZATION_EVENTS_IGNORE].cyc_total += cycles;
  }

  cu->cyc_started = 0;
}

static void
_cpu_utilization__measurement_stop_inactive(cpu_utilization_t *cu, int event, int nrecs)
{
  return;
}

static void
_cpu_utilization__measurement_stop_wait(cpu_utilization_t *cu, int event, int nrecs)
{
  if (cu->tp_state == CPU_UTILIZATION_STATE_WAIT_HIT) {
    if (event != 0) {
      return;
    }
  } else if (cu->tp_state != CPU_UTILIZATION_STATE_WAIT_CALL) {
    return;
  }

  cpu_utilization__measurement_begin(cu);
}

#define	cpu_utilization__measurement_stop(cu, event, nrecs)  (cu)->cpu_utilization_stop_func(cu, event, nrecs)

static void
cpu_utilization__measurement_begin(cpu_utilization_t *cu)
{
  if (!(cu->tp_state & (CPU_UTILIZATION_STATE_WAIT_HIT|CPU_UTILIZATION_STATE_WAIT_CALL))) {
    return;
  }

  cu->tp_state = CPU_UTILIZATION_STATE_ACTIVE;
  cu->cpu_utilization_stop_func = _cpu_utilization__measurement_stop_active;

  cu->tp_cyc_begin = mal_get_cycles();

  if (cu->flags & CPU_UTILIZATION_FLAG_TEST_PERIOD_MAX_SECONDS) {
    if (cu->max_seconds == 0) {
      cu->tp_cyc_stop = ~0;
    } else {
      cu->tp_cyc_stop = cu->tp_cyc_begin + (cu->cyc_per_second * cu->max_seconds);
    }
  } else {
    cu->tp_cyc_stop = ~0;
  }
}

static void
cpu_utilization__measurement_end(cpu_utilization_t *cu)
{
  if (!(cu->tp_state & (CPU_UTILIZATION_STATE_ACTIVE))) {
    return;
  }

  cu->tp_state = CPU_UTILIZATION_STATE_INACTIVE;
  cu->cpu_utilization_stop_func = _cpu_utilization__measurement_stop_inactive;

  cu->tp_cyc_end = mal_get_cycles();
}

static void
_cpu_utilization__dump_stats(cpu_utilization_t *cu, int evt, char *output, int output_max_size)
{
  mal_cycles_t tp_cycles;
  mal_cycles_t tp_cyc_end;
  mal_cycles_t tp_cyc_avg;
  mal_cycles_t avg_ns;
  mal_cycles_t event_ns;
  mal_cycles_t event_cyc_avg;
  mal_cycles_t app_ns;
  mal_cycles_t total_ns;
  float tp_cyc_util_p;
  float tp_cyc_total_p;
  float tp_cyc_event_p;
  struct cu_event_s *event;

  if ((output_max_size == 0) || (output == NULL)) {
    return;
  }

  event = &(cu->events[evt]);

  if (!(cu->tp_state & CPU_UTILIZATION_STATE_INACTIVE)) {
    if ( (tp_cyc_end = mal_get_cycles()) > cu->tp_cyc_stop ) {
      tp_cyc_end = cu->tp_cyc_stop;
    }
  } else {
    tp_cyc_end = cu->tp_cyc_end;
  }

  tp_cycles = tp_cyc_end - cu->tp_cyc_begin;
  
  tp_cyc_total_p = tp_cycles;
  tp_cyc_event_p = event->cyc_total;
  tp_cyc_util_p = (tp_cyc_event_p / tp_cyc_total_p) * 100.0;

  if (event->cnt == 0) {
    event_cyc_avg = tp_cyc_avg = avg_ns = event_ns = total_ns = app_ns = 0;
  } else {
    tp_cyc_avg = tp_cycles / event->cnt;
    event_cyc_avg = event->cyc_total / event->cnt;
    avg_ns = event_cyc_avg / cu->cyc_per_nano_second;
    event_ns = event->cyc_total / cu->cyc_per_nano_second;
    total_ns = tp_cycles / cu->cyc_per_nano_second;
    app_ns = total_ns - event_ns;
  }

  (void)snprintf(output, output_max_size, "%10s UTILIZATION STATS: cpu=%2d, cnt=%12lu, "
                 "capture time=%14luns, "
                 "CPU util=%6.2f%%, total cycles=%14lu, total time=%14luns, "
                 "avg cycles=%14lu, avg time=%14luns, app time=%14luns",
         event->name, cu->cpu, event->cnt,
         total_ns,
         tp_cyc_util_p, event->cyc_total, event_ns,
         event_cyc_avg, avg_ns, app_ns);
}

#define CPU_UTILIZATION_DUMP_FILE_NAME "snf_simple_recv.perf"
#define CPU_UTILIZATION_DUMP_OUTPUT_BUFFER_SIZE	4096
static char _cpu_utilization_output[CPU_UTILIZATION_DUMP_OUTPUT_BUFFER_SIZE+1];

static void
cpu_utilization__dump(void)
{
  int worker;
  int flags = 0;
  cpu_utilization_t *cu;
#ifndef _WIN32
  FILE *dump_file = NULL;
#endif /* _WIN32 */

  for (worker=0; worker<MAX_WORKERS; worker++) {
    cu = &(cpu_utilization[worker]);
    if (!(cu->tp_state & (CPU_UTILIZATION_STATE_ACTIVE|CPU_UTILIZATION_STATE_INACTIVE))) {
      continue;
    }

    if (flags == 0 ) {
      if (cu->flags & CPU_UTILIZATION_FLAG_DUMP_FILE) {
        flags |= CPU_UTILIZATION_FLAG_DUMP_FILE;
      }

      if (cu->flags & CPU_UTILIZATION_FLAG_DUMP_STDIO) {
        flags |= CPU_UTILIZATION_FLAG_DUMP_STDIO;
      }

      if (flags == 0) {
        continue;
      }

      if (flags & CPU_UTILIZATION_FLAG_DUMP_FILE) {
#ifndef _WIN32
        if ((dump_file = fopen(CPU_UTILIZATION_DUMP_FILE_NAME, "w")) == NULL) {
          flags &= ~CPU_UTILIZATION_FLAG_DUMP_FILE;
        }
#else
      /* not supported */;
#endif /* _WIN32 */
      }
    }

    _cpu_utilization__dump_stats(cu, CPU_UTILIZATION_EVENTS_HIT, _cpu_utilization_output, CPU_UTILIZATION_DUMP_OUTPUT_BUFFER_SIZE);
    if (flags & CPU_UTILIZATION_FLAG_DUMP_STDIO) {
      printf("%s\n",_cpu_utilization_output);
    }
    if (flags & CPU_UTILIZATION_FLAG_DUMP_FILE) {
#ifndef _WIN32
      fprintf(dump_file,"%s\n",_cpu_utilization_output);
#else
      /* not supported */;
#endif /* _WIN32 */
    }

    _cpu_utilization__dump_stats(cu, CPU_UTILIZATION_EVENTS_MISS, _cpu_utilization_output, CPU_UTILIZATION_DUMP_OUTPUT_BUFFER_SIZE);
    if (flags & CPU_UTILIZATION_FLAG_DUMP_STDIO) {
      printf("%s\n",_cpu_utilization_output);
    }
    if (flags & CPU_UTILIZATION_FLAG_DUMP_FILE) {
#ifndef _WIN32
      fprintf(dump_file,"%s\n",_cpu_utilization_output);
#else
      /* not supported */;
#endif /* _WIN32 */
    }

    _cpu_utilization__dump_stats(cu, CPU_UTILIZATION_EVENTS_IGNORE, _cpu_utilization_output, CPU_UTILIZATION_DUMP_OUTPUT_BUFFER_SIZE);
    if (flags & CPU_UTILIZATION_FLAG_DUMP_STDIO) {
      printf("%s\n\n",_cpu_utilization_output);
    }
    if (flags & CPU_UTILIZATION_FLAG_DUMP_FILE) {
#ifndef _WIN32
      fprintf(dump_file,"%s\n\n",_cpu_utilization_output);
#else
      /* not supported */;
#endif /* _WIN32 */
    }
  }

  if (flags & CPU_UTILIZATION_FLAG_DUMP_FILE) {
#ifndef _WIN32
    fflush(dump_file);
    fclose(dump_file);
#else
    /* not supported */;
#endif /* _WIN32 */
  }
}

static void
cpu_utilization__init(cpu_utilization_t *cu, int cpu, uint32_t flags, uint32_t max_seconds)
{
  mal_cycles_t ns_in_sec;

  memset(cu, 0, sizeof(cpu_utilization_t));

  if (CPU_UTILIZATION_FLAG_DEFAULT == flags) {
    cu->flags = CPU_UTILIZATION_FLAG_DEFAULT_FLAGS;
  } else {
    cu->flags = flags;
  }

  if (!(flags & CPU_UTILIZATION_FLAG_BEGIN_TEST_PERIOD_ON_FIRST_CALL)) {
    cu->flags |= CPU_UTILIZATION_FLAG_BEGIN_TEST_PERIOD_ON_FIRST_HIT;
    cu->tp_state = CPU_UTILIZATION_STATE_WAIT_HIT;
  } else {
    cu->flags &= ~CPU_UTILIZATION_FLAG_BEGIN_TEST_PERIOD_ON_FIRST_HIT;
    cu->tp_state = CPU_UTILIZATION_STATE_WAIT_CALL;
  }

  if (cu->flags & CPU_UTILIZATION_FLAG_TEST_PERIOD_MAX_SECONDS) {
    if ((cu->max_seconds = max_seconds) == 0) {
      cu->flags &= ~CPU_UTILIZATION_FLAG_TEST_PERIOD_MAX_SECONDS;
    }
  }

  mal_cycles_counter_init();
  cu->cyc_per_second = mal_cycles_per_second();

  ns_in_sec = (1000 * 1000 * 1000);
  cu->cyc_per_nano_second = cu->cyc_per_second / ns_in_sec;

  cu->cpu_utilization_stop_func = _cpu_utilization__measurement_stop_wait;

  cu->cpu = cpu;

  cu->events[CPU_UTILIZATION_EVENTS_HIT].name = "HIT";
  cu->events[CPU_UTILIZATION_EVENTS_MISS].name = "MISS";
  cu->events[CPU_UTILIZATION_EVENTS_IGNORE].name = "IGNORE";
}
#else /* defined(CPU_UTILIZATION_SUPPORT) */
#define cpu_utilization__init(cu, cpu, flags, max_seconds)
#define cpu_utilization__dump()
#define cpu_utilization__measurement_end(cu)
#define cpu_utilization__measurement_begin(cu)
#define	cpu_utilization__measurement_stop(cu, event, nrecs)
#define cpu_utilization__measurement_start(cu)
#endif /* defined(CPU_UTILIZATION_SUPPORT) */

static void
dump_data(unsigned char *pg, int len)
{
  int i = 0;
  while (i < len) {
    printf
      ("%04d: %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x\n",
       i, pg[i], pg[i + 1], pg[i + 2], pg[i + 3], pg[i + 4], pg[i + 5],
       pg[i + 6], pg[i + 7], pg[i + 8], pg[i + 9], pg[i + 10], pg[i + 11],
       pg[i + 12], pg[i + 13], pg[i + 14], pg[i + 15]);
    i += 16;
  }
}

uint64_t num_pkts = 0;
unsigned int max_received_tv_delta = 0;
uint64_t num_bytes = 0;
uint64_t num_pkts_inject = 0;
uint64_t num_pkts_netdev_reflect = 0;
snf_ring_t hring;
snf_handle_t hsnf;
int periodic_stats = 0;

#define TOGGLE(i) ((i+ 1) & 1)
#define TV_TO_US(tv) ((tv)->tv_sec * 1000000 + (tv)->tv_usec)
unsigned int itvl_idx = 0;
struct itvl_stats {
  struct timeval tv;
  uint64_t usecs;
  uint64_t num_pkts;
  uint64_t num_bytes;
} itvl[2];

float secs_delta;

void
stats()
{
  struct snf_ring_stats stats;
  uint64_t nbytes;
  int rc;
  if ((rc = snf_ring_getstats(hring, &stats))) {
    perror("nic stats failed");
  }

  printf("\n");
  if (num_pkts == stats.ring_pkt_recv) {
    printf("Packets received   in HW:        %" PRIu64 "\n", num_pkts);
  } else {
    printf("Packets received,    app:        %" PRIu64 ", ring: %" PRIu64 "\n",
           num_pkts, stats.ring_pkt_recv);
  }
  printf("Packets reinjected,    app:        %" PRIu64 "\n", num_pkts_inject);

  printf("Packets reflected to netdev:       %" PRIu64 "\n", num_pkts_netdev_reflect);

  printf("Total bytes received,  app:        %" PRIu64 " (%" PRIu64 " MB)\n",
          num_bytes, num_bytes / 1024 / 1024);
  nbytes = stats.nic_bytes_recv;
  printf("Total bytes received, HW: %" PRIu64 " (%" PRIu64 " MB)\n",
          nbytes, nbytes / 1024 / 1024);
  if (num_pkts > 0) {
    printf("Average Packet Length:    %" PRIu64 " bytes\n",
          num_bytes / num_pkts);
  }

  printf("Dropped, NIC overflow:    %" PRIu64 "\n", stats.nic_pkt_overflow);
  printf("Dropped, ring overflow:   %" PRIu64 "\n", stats.ring_pkt_overflow);
  printf("Dropped, bad:             %" PRIu64 "\n", stats.nic_pkt_bad);
  //printf("Dropped, filtered:        %" PRIu64 "\n\n", stats.nic_pkt_dropped);
}


static inline
int
get_timesource(char *tm_mode)
{
  int rc;
  enum snf_timesource_state ts_state;
  
  *tm_mode = 'X';
  rc = snf_get_timesource_state(hsnf, &ts_state);
  
  if (rc) {
    errno = rc;
    perror("Can't get timesource status\n");
    return -1;
  } else {
    switch (ts_state) {
    
    case SNF_TIMESOURCE_LOCAL:
      *tm_mode = 'L';
      break;
    case SNF_TIMESOURCE_EXT_UNSYNCED:
      *tm_mode = 'U';
      break;
    case SNF_TIMESOURCE_EXT_SYNCED:
      *tm_mode = 'S';
      break;
    case SNF_TIMESOURCE_EXT_FAILED:
      *tm_mode = 'F';
      break;
    case SNF_TIMESOURCE_ARISTA_ACTIVE:
      *tm_mode = 'A';
      break;
    case SNF_TIMESOURCE_PPS:
      *tm_mode = 'P';
      break;
    }
  }
  return 0;
}

void
print_periodic_stats(void)
{
  struct itvl_stats *this_itvl = &itvl[itvl_idx];
  struct itvl_stats *last_itvl = &itvl[TOGGLE(itvl_idx)];
  float delta_secs;
  uint64_t delta_pkts;
  uint64_t delta_bytes;
  uint32_t pps;
  float gbps;
  float bps;
  char time_mode = ' ';

  if (periodic_stats > 1)
    get_timesource(&time_mode);  

  gettimeofday(&this_itvl->tv, NULL);
  this_itvl->usecs = TV_TO_US(&this_itvl->tv);
  this_itvl->num_pkts = num_pkts;
  this_itvl->num_bytes = num_bytes;
  delta_secs = (float)((this_itvl->usecs - last_itvl->usecs) / 1000000.0);
  delta_pkts = this_itvl->num_pkts - last_itvl->num_pkts;
  delta_bytes = this_itvl->num_bytes - last_itvl->num_bytes;

  if (delta_pkts != 0) {
    pps = (uint32_t)(delta_pkts / delta_secs);
    bps = ((float) delta_bytes * 8) / delta_secs;
    gbps = (float)(bps / 1000000000.0);

    printf
      ("%" PRIu64 " pkts (%" PRIu64 "B) in %.3f secs (%u pps), Avg Pkt: %"
       PRIu64 ", BW (Gbps): %6.3f %c\n", delta_pkts, delta_bytes, delta_secs,
       pps, delta_bytes / delta_pkts, gbps, time_mode);
    fflush(stdout);
  } else if (periodic_stats > 1) {
    printf("time mode: %c\n", time_mode);
  }

  itvl_idx = TOGGLE(itvl_idx);
}

#ifdef _WIN32
DWORD WINAPI stats_thread(LPVOID lpParameter)
{
  for (;;) {
    Sleep(1000);
    print_periodic_stats();
  }
}
#endif

void C_API
sigexit(int sig)
{
  getout = 1;
}

#ifndef _WIN32
void
sigalrm(int sig)
{
  print_periodic_stats();
  alarm(1);
}
#endif

void
usage(void)
{
  printf("Usage: snf_simple_recv [-v] [-t]  [-p <port number>] [-p] [-n <num pkts>]\n");
  printf("         [-d <ring sz>] [-S <snap len>] [-m <pkts>]\n");
  printf("                  -v: verbose\n");
  printf("                  -t: print periodic statistics, -t -t: also print timestamp mode\n");
  printf("    -p <port number>: Myri10G port number to use.\n");
  printf("        -T <timeout>: Wait for at most <timeout> milliseconds between calls, 0 never blocks\n");
  printf("       -n <num pkts>: number of packet to receive (default: 0 - infinite)\n");
  printf("        -d <ring sz>: size if the recieve ring in bytes\n");
  printf("       -S <snap len>: display first <snap len> bytes of each packet\n");
  printf("                  -D: display nanosecond delta from timestamp of prior packet\n");
  printf("    -R <port number>: reinject every received packet on port <portnum>\n");
  printf("                  -N: pass every received packet to netdev\n");
  printf("                  -O: do not check out-of-order packet timestamps\n");
  printf("           -m <pkts>: print verbose output every <pkts> packets (need -v)\n");
  printf("              --help: Display this help, and exit\n");
  exit(1);
}

#ifdef _WIN32
typedef void (WINAPI *PGST)(LPFILETIME);

static LARGE_INTEGER offset;
static int initialized = 0;
static PGST lpGetSystemTime = NULL;

LARGE_INTEGER
getFILETIMEoffset()
{
  SYSTEMTIME s;
  FILETIME f;
  LARGE_INTEGER t;

  s.wYear = 1970;
  s.wMonth = 1;
  s.wDay = 1;
  s.wHour = 0;
  s.wMinute = 0;
  s.wSecond = 0;
  s.wMilliseconds = 0;
  SystemTimeToFileTime(&s, &f);
  t.QuadPart = f.dwHighDateTime;
  t.QuadPart <<= 32;
  t.QuadPart |= f.dwLowDateTime;
  return (t);
}
#endif

static
uint64_t
host_nsecs(void)
{
#ifndef _WIN32
  struct timespec ts;
  clock_gettime(CLOCK_REALTIME, &ts);
  return (uint64_t) ts.tv_sec * 1000000000 + ts.tv_nsec;
#else
  ULARGE_INTEGER t;
  FILETIME f;

  if (!initialized) {
    offset = getFILETIMEoffset();

    lpGetSystemTime = (PGST) GetProcAddress(GetModuleHandle(TEXT("kernel32.dll")), "GetSystemTimePreciseAsFileTime");
    if (lpGetSystemTime == NULL) {
      lpGetSystemTime = GetSystemTimeAsFileTime;
    }
    initialized = 1;
  }

  lpGetSystemTime(&f);

  t.HighPart = f.dwHighDateTime;
  t.LowPart = f.dwLowDateTime;
  t.QuadPart -= offset.QuadPart;
  t.QuadPart *= 100;

  return t.QuadPart;
#endif
}

int
C_API
main(int argc, char **argv)
{
  int rc;
  snf_inject_t hinj = NULL;
  snf_netdev_reflect_t hnetdev = NULL;
  struct snf_recv_req recv_req;
  char c;
  int verbose = 0;
  int snap_print = 0;
  int delta_print = 0;
  int inspect_pkt;
  int portnum = 0;
  int reinject = -1;
  int netdev_reflect = 0;
  uint64_t pkts_expected = 0xffffffffffffffffULL;
  uint64_t dataring_sz = 0;
  int timeout_ms = -1;
  int pkt_mod = 10000;
  int check_ooo = 1;
  uint64_t last_tstamp = 0;
#if defined(CPU_UTILIZATION_SUPPORT)
  int worker_id = 0;
#endif /* defined(CPU_UTILIZATION_SUPPORT) */
  int option_index = 0;
  static struct option long_options[] = {
    {"help", no_argument, 0, '?'},
    {0,0,0,0}
  };

  /* get args */
  while ((c = getopt_long(argc, argv, "T:vtb:p:n:d:S:DR:m:NO?", long_options, &option_index)) != -1) {
    if (c == 'v') {
      verbose++;
    } else if (c == 't') {
      periodic_stats++;
    } else if (c == 'p' || c == 'b') {
      portnum = strtoul(optarg, NULL, 0);
    } else if (c == 'T') {
      timeout_ms = strtoul(optarg, NULL, 0);
    } else if (c == 'n') {
      pkts_expected = strtoul(optarg, NULL, 0);
    } else if (c == 'd') {
      dataring_sz = strtoull(optarg, NULL, 0);
    } else if (c == 'S') {
      snap_print = strtoul(optarg, NULL, 0); 
    } else if (c == 'D') {
      delta_print++;
    } else if (c == 'R') {
      reinject = strtoul(optarg, NULL, 0);
    } else if (c == 'm') {
      pkt_mod = strtoul(optarg, NULL, 0);
    } else if (c == 'N') {
      netdev_reflect = 1;
    } else if (c == 'O') {
      check_ooo = 0;
    } else {
      if (c != '?')
	printf("Unknown option: %c\n", c);
      usage();
    }
  }

  snf_init(SNF_VERSION_API);
  rc = snf_open(portnum, 0, NULL, dataring_sz, -1, &hsnf);
  if (rc) {
    errno = rc;
    perror("Can't open snf for sniffing");
    return -1;
  }
  rc = snf_ring_open(hsnf, &hring);
  if (rc) {
    errno = rc;
    perror("Can't open a receive ring for sniffing");
    return -1;
  } 
  if (reinject >= 0) {
    rc = snf_inject_open(portnum, 0, &hinj);
    if (rc) {
      errno = rc;
      perror("Can't re-inject packets from ring");
      return -1;
    }
  }
  if (netdev_reflect) {
    rc = snf_netdev_reflect_enable(hsnf, &hnetdev);
    if (rc) {
      errno = rc;
      perror("Can't reflect packets to netdev\n");
      return -1;
    }
  }

  rc = snf_start(hsnf);
  if (rc) {
    errno = rc;
    perror("Can't start packet capture for sniffing");
    return -1;
  }

  printf("snf_recv ready to receive\n");

  if (SIG_ERR == signal(SIGINT, sigexit))
    exit(1);
  if (SIG_ERR == signal(SIGTERM, sigexit))
    exit(1);

  if (periodic_stats) {
#ifndef _WIN32
    if (SIG_ERR == signal(SIGALRM, sigalrm))
      exit(1);
    memset(itvl, 0, sizeof(struct itvl_stats) * 2);
    gettimeofday(&itvl[0].tv, NULL);
    gettimeofday(&itvl[1].tv, NULL);
    itvl_idx = TOGGLE(0);
    alarm(1);
#else
    itvl_idx = TOGGLE(0);
    CreateThread(NULL, 0, stats_thread, NULL, 0, NULL);
#endif
  }

  /* initialize the CPU util measurement objects */
  memset(&(cpu_utilization[0]), 0, (sizeof(cpu_utilization_t) * MAX_WORKERS));

  /* simple recv has only one recv to measure */
  cpu_utilization__init(&(cpu_utilization[worker_id]), worker_id, CPU_UTILIZATION_FLAG_DEFAULT, 30);

  inspect_pkt = (verbose || (snap_print > 0) || delta_print);
  while (num_pkts < pkts_expected && !getout) {
    cpu_utilization__measurement_start(&(cpu_utilization[worker_id]));
    rc = snf_ring_recv(hring, timeout_ms, &recv_req);
    cpu_utilization__measurement_stop(&(cpu_utilization[worker_id]), rc, 1);
    if (rc == EAGAIN || rc == EINTR)
      continue;
    else if (rc == 0) {
      if (check_ooo && (last_tstamp > recv_req.timestamp ||
                        recv_req.timestamp == 0))
        printf("out-of-order packet timestamps: last_tstamp=%llu current_tsamp=%llu\n",
               (long long unsigned)last_tstamp, (long long unsigned)recv_req.timestamp);

      num_pkts++;
      num_bytes += recv_req.length;
      if (reinject >= 0) {
        rc = snf_inject_send(hinj, 10, 0, recv_req.pkt_addr, recv_req.length);
        if (!rc)
          num_pkts_inject++;
      }
      if (netdev_reflect) {
        rc = snf_netdev_reflect(hnetdev, recv_req.pkt_addr, recv_req.length);
        if (!rc)
          num_pkts_netdev_reflect++;
      }
      if (!inspect_pkt)
        continue;
      if (verbose && (num_pkts % pkt_mod == 0)) {
        unsigned long long h_ts, recv_ts;
	signed long long diff;
        h_ts = host_nsecs();
	recv_ts = (unsigned long long) recv_req.timestamp;
	if (h_ts > recv_ts) {
	  diff = h_ts - recv_ts;
	} else {
	  diff = (recv_ts - h_ts) * -1;
	}
        printf("pkt: %llu, len: %u, ts_hw: %llu ts_host: %llu ts_diff:%lld\n",
            (unsigned long long)num_pkts,
            recv_req.length, (unsigned long long) recv_req.timestamp,
            h_ts, diff);
      }

      if (delta_print) {
        unsigned long long recv_ts;

	recv_ts = (unsigned long long) recv_req.timestamp;
        if (last_tstamp == 0)
          last_tstamp = recv_ts;  /* so first value is not huge */

        printf("%lld %lld\n",  (unsigned long long) num_pkts, (recv_ts - last_tstamp));
      }

      if (snap_print > 0) {
          dump_data(recv_req.pkt_addr,
		    (recv_req.length > (uint32_t)snap_print) ? snap_print : recv_req.length);
      }
      last_tstamp = recv_req.timestamp;
    }
    else {
      fprintf(stderr, "error: snf_recv = %d (%s)\n",
                 rc, strerror(rc));
      break;
    }
  }

  stats();

  cpu_utilization__dump();

  if (hinj)
    snf_inject_close(hinj);

  rc = snf_ring_close(hring);
  if (rc) {
    errno = rc;
    perror("Can't close receive ring");
    return -1;
  }
  rc = snf_close(hsnf);
  if (rc) {
    errno = rc;
    perror("Can't close sniffer device");
    return -1;
  }

  return 0;
}
