#ifndef __ARISTA_SWITCH_H
#define __ARISTA_SWITCH_H

/* Switch timestamps packets defined by aristanetworks */

#include "mal_timing.h"

//#define ARISTA_DEBUG 1

#if !MAL_KERNEL
#if MAL_OS_LINUX || MAL_OS_FREEBSD
#include <stdint.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/param.h>
#endif
#endif

#define ARISTA_TEST_ASIC_TIME     0x4153494354494d45  /* asictime in ascii */
#define ARISTA_TEST_DROPPED       0x44524f504652414d  /* dropfram in ascii */ 
#define ARISTA_TEST_LAST_SYNC     0x4c41535453594e43  /* lastsync in ascii */
#define ARISTA_TEST_DELAYED_TIME  ((rand() % 6 + 1) * 1000000000ULL)  /* used to emulate kf delay */

typedef enum {
  ARISTA_STATUS_DISABLED = 0, 
  ARISTA_STATUS_INITED,                /* need first keyframe */
  ARISTA_STATUS_CHANGEDMODE,           /* switch changed to unsupported mode */
  ARISTA_STATUS_FALLBACK_LOCAL,        /* temporary timetamping fallback */
  ARISTA_STATUS_ACTIVE,                /* normal arista mode */
  ARISTA_STATUS_OUTOFSYNC,             /* arista switch can't sync, user defines when to fallback */
  ARISTA_STATUS_NEED_KEYFRAME,         /* waited more than Xsec without having a 253 frame */
} arista_tsp_status;

typedef uint32_t arista_tsp_t;
#define ARISTA_GET64_H24b(nano64b) (nano64b >> 32 & 0xFFFFFF00)
#define ARISTA_GET64_L07b(nano64b) (nano64b & 0x7F)

/* switch port ASIC 31bit timestamp
31 ----------------------------------- 8 -- 7 -- 6 --------- 0
|         high 24 bits                 |   pad   |  low 7b   | */
#define ARISTA_SET31_H24b(tick64b) ((tick64b & 0x7FFFFF80) << 1)
#define ARISTA_SET31_L07b(tick64b) ((tick64b & 0x0000007F))
#define ARISTA_GET31_L31b(tick32b) (((tick32b & 0xFFFFFF00) >> 1) | (tick32b & 0x7F))

/* non standard keyframe sent by Arista switches on egress port
   having subscribed to the alternate subscribing (connecting on the
   switch, using switch commands). */
#define ARISTA_KF_IPPROTO    253
#define ARISTA_KF_TTL        64
#define ARISTA_MAX_STRING    512
#define ARISTA_ERR_STRING    40


#define ARISTA_TICK_TO_NS    (20.0/7.0)    /* ns value of tick on arista switch asic */
#define ARISTA_TICK_MAX_T    0x7FFFFFFF    /* max tick */

 /* max 5 sec. without keyframe invalidates the timestamping protocol,
    this value should be set according to the 31bit timestamp
    capability and the switch oscillator speed. Should be changed
    according to the switch datasheet. */
#define ARISTA_KF_TIMEOUT_1  3500000000ULL /* first call, 3.5s, in nsec */
#define ARISTA_KF_TIMEOUT_2  1000000000ULL /* 1 s, in nsec, and goes to fallback mode */
#define ARISTA_TS_MAX_FLIP   3             /* n flips x 6.15sec */

#define ARISTA_KF_PKCHECK  10000000        /* check every PKCHECK packets */
#define ARISTA_KF_PTP_WAIT 7000000000ULL   /* 7sec to wait for confirmation, ns */ 

typedef enum { /* encoding the status of fcs type field of the keyframe */
  ARISTA_FCS_TIMESTAMPING_DISABLED = 0,
  ARISTA_FCS_TIMESTAMPING_APPENDED,
  ARISTA_FCS_TIMESTAMPING_OVERWRITE,
  ARISTA_FCS_TIMESTAMPING_TESTING_MODE,
} arista_mode;

typedef struct arista_keyframe_skew {
  uint64_t    ak_skew_numerator;
  uint64_t    ak_skew_denominator;
} ast_kf_skew_t;

typedef struct arista_keyframe {
  uint64_t    ak_asic_time;
  uint64_t    ak_utc_time;
  uint64_t    ak_last_sync_time;
  uint64_t    ak_skew_numerator;
  uint64_t    ak_skew_denominator;
  uint64_t    ak_kf_timestamp;
  uint64_t    ak_dropped_frames;
  uint16_t    ak_device_id;
  uint16_t    ak_egress_swport;
  uint8_t     ak_fcs_type;
  uint8_t     ak_reserved;
} ast_kf_t;

typedef struct arista_stats {
  uint64_t               arst_kfrecv;
  uint64_t               arst_kfneeded;
  uint64_t               arst_kfmalformed;
  uint32_t               arst_kfmalformedlen;
  uint64_t               arst_flips;
  uint64_t               arst_ptpoos;
  uint64_t               arst_unsupported_mode;
  uint64_t               arst_dropped;
  uint64_t               arst_neg_adj_event;
} ast_stats_t;

/* each NIC port rxs in upper-layers should instanciate this */
typedef struct arista_control {  
  arista_mode            arct_mode;         /* mode in which the switch is sending the timestamps */
  arista_tsp_status      arct_status;       /* state machine */
  uint8_t                arct_kf_check_rate;/* check if kf rate is still high enough */
  uint8_t                arct_kf_check_ptp; /* is ptp activated ? */ 
  uint8_t                arct_kf_recv;      /* user can receive the keyframe as a packet */
  uint8_t                arct_kf_dupkern;   /* user can ask to duplicate kf to the kernel */
  uint32_t               arct_kf_last_ts;   /* last aligned 31bit timestamp */
  struct arista_keyframe arct_lastkf;       /* not pointer because prefetched for time computing 48B */
  uint8_t                arct_kf_hasskew;   /* keyframe has skew num/denom */
  uint8_t                pad[1];            /* 64B cache-aligned block */
  uint64_t               arct_kf_flipped;   /* number of time timestamp overflowned */
  uint64_t               arct_kf_ptp_wait;  /* time to wait before ptp is really out-of-sync */
  struct arista_stats    arct_stats; 
  char                   arct_prodname[4];  /* short name */ 
  uint32_t               arct_kf_dst_ip;    /* dst IP in the keyframe (used in DBL to filter) */
  uint64_t		 arct_last_ts;      /* last packet UTC timestamp */
} ast_ctl_t;


#define IS_ARISTA_MODE_ACTIVE(control)   (unlikely(control.arct_status >= ARISTA_STATUS_ACTIVE))
#define IS_ARISTA_MODE_FALLBACK(control) (control.arct_status == ARISTA_STATUS_FALLBACK_LOCAL)
#define IS_ARISTA_SUPPORTED_MODE(mode)   (mode == ARISTA_FCS_TIMESTAMPING_APPENDED)
#define ARISTA_DEFAULT_SWITCH_DST_IP     "1.1.1.0"        
#define ARISTA_SUPPORTED_MODE_STR        "append (before-fcs)"
#define ARISTA_SIZEOF_TIMESTAMP          sizeof(arista_tsp_t)
#define ARISTA_SIZEOF_KF                 62 /* packed sizeof(ast_kf_t) */
#define ARISTA_SIZEOF_KF_FRAME           14+20+62 /* ethh+iph+KF */

#define ARISTA_MODE_CHANGE_UNSUPPORTED 1
#define ARISTA_MODE_CHANGE_SUPPORTED   2
#define ARISTA_MODE_PTP_DISCONNECTED   4
#define ARISTA_MODE_PTP_RECONNECTED    8
#define ARISTA_MODE_KF_SHORTAGE_START 16
#define ARISTA_MODE_KF_SHORTAGE_STOP  32
#define ARISTA_MODE_KF_MALFORMED      64

#if !MAL_KERNEL
static inline uint64_t
arista_host_nsecs(void)
{
#ifndef _WIN32
  uint64_t nsecs;
  struct timespec ts;  
  clock_gettime(CLOCK_REALTIME, &ts);
  nsecs = (uint64_t) ts.tv_sec * 1000000000 + ts.tv_nsec;
  return nsecs;
#else
  FILETIME ft;
  ULARGE_INTEGER x;

  GetSystemTimeAsFileTime(&ft);
  x.LowPart = ft.dwLowDateTime;
  x.HighPart = ft.dwHighDateTime;
  x.QuadPart *= 100;
  return x.QuadPart;
#endif
}
#else /* KERNEL LAND */
static inline uint64_t
arista_host_nsecs(void)
{
  struct timeval tv;  
  mal_gettimeofday(&tv,NULL);
  return (uint64_t) tv.tv_sec * 1000000000 + tv.tv_usec * 1000;
}

#endif


static inline 
int
arista_check_kf_pkt(uint8_t *pkt_addr)
{
  uint8_t *fp = pkt_addr;
  uint8_t ip_protocol;
  uint16_t ethtype_be16;

  fp += 12; /* skip src+dst MACs */
  ethtype_be16 = *(uint16_t *)fp;
  if (ethtype_be16 == htons(0x8100)) { /* skip VLAN */
    fp += 4;
    ethtype_be16 = *(uint16_t *)fp;
  }
  fp += 2; /* skip ethtype */

  if (ethtype_be16 == htons(0x0800)) { /* IPv4 */
    fp += 9; /* protocol number */
    ip_protocol = *(uint8_t *)fp;
    if (ip_protocol == ARISTA_KF_IPPROTO) return 1;
  }
  if (ethtype_be16 == htons(0x86dd)) { /* IPv6 */
    fp += 6; /* protocol number */
    ip_protocol = *(uint8_t *)fp;
    if (ip_protocol == ARISTA_KF_IPPROTO) return 1;
  }
  return 0; /* not one */
}

/* loads a keyframe from an IP keyframe packet */
static inline int
arista_get_kf_pkt(void *pktstart, uint32_t len, ast_ctl_t *ctl)
{
  int rc = 0;
  uint8_t *fp = (uint8_t *)pktstart; 
  uint16_t ethtype_be16;
  ast_kf_t *lastkf = (ast_kf_t *)&ctl->arct_lastkf;

  /* 253 can be used to inject bad frames */
  if (len != (ARISTA_SIZEOF_KF_FRAME - (ctl->arct_kf_hasskew ? 0 : sizeof(ast_kf_skew_t)))) {
    ctl->arct_stats.arst_kfmalformed++;
    ctl->arct_stats.arst_kfmalformedlen = len;
    return ARISTA_MODE_KF_MALFORMED;
  }

  fp += 12; /* skip src+dst MACs */
  ethtype_be16 = *(uint16_t *)fp;
  fp = (uint8_t *)pktstart;
  if (ethtype_be16 == htons(0x8100))  /* if VLAN, skip a little bit more */
    fp += 18 + 20; /* we know we have a keyframe, skip all protocols */
  else
    fp += 14 + 20; /* no vlan */

  lastkf->ak_asic_time       = mx_ntohll(*(uint64_t *)fp); fp += 8;
  lastkf->ak_utc_time        = mx_ntohll(*(uint64_t *)fp); fp += 8;
  lastkf->ak_last_sync_time  = mx_ntohll(*(uint64_t *)fp); fp += 8;
  if (ctl->arct_kf_hasskew) {
    lastkf->ak_skew_numerator    = mx_ntohll(*(uint64_t *)fp); fp += 8;
    lastkf->ak_skew_denominator  = mx_ntohll(*(uint64_t *)fp); fp += 8;
  }
  lastkf->ak_kf_timestamp    = mx_ntohll(*(uint64_t *)fp); fp += 8;
  lastkf->ak_dropped_frames  = mx_ntohll(*(uint64_t *)fp); fp += 8;
  lastkf->ak_device_id       = ntohs(*(uint16_t *)fp);     fp += 2;
  lastkf->ak_egress_swport   = ntohs(*(uint16_t *)fp);     fp += 2;
  lastkf->ak_fcs_type        = (*(uint8_t *)fp);           fp += 1;
  lastkf->ak_reserved        = (*(uint8_t *)fp);           

  ctl->arct_stats.arst_kfrecv++;

  /* keyframe shortage, reset register flipping since this is a new kf */
  ctl->arct_kf_flipped = 0;
  if (ctl->arct_status == ARISTA_STATUS_NEED_KEYFRAME) {
    ctl->arct_status = ARISTA_STATUS_ACTIVE;
    rc |= ARISTA_MODE_KF_SHORTAGE_STOP;
  }
  /* mode changes */
  if (! IS_ARISTA_SUPPORTED_MODE(lastkf->ak_fcs_type) &&
      ctl->arct_status != ARISTA_STATUS_CHANGEDMODE) {
    ctl->arct_status = ARISTA_STATUS_CHANGEDMODE;
    ctl->arct_stats.arst_unsupported_mode++;
    rc |= ARISTA_MODE_CHANGE_UNSUPPORTED;
  }
  else if (ctl->arct_status == ARISTA_STATUS_CHANGEDMODE &&
	   IS_ARISTA_SUPPORTED_MODE(lastkf->ak_fcs_type)) {
    ctl->arct_status = ARISTA_STATUS_ACTIVE;
    rc |= ARISTA_MODE_CHANGE_SUPPORTED;
  }
  /* sync changes */
  if (ctl->arct_kf_check_ptp) {
    /* kf pretends the ptp sync is gone, breath for a few sec. */
    if (unlikely(lastkf->ak_last_sync_time == 0)) { 
      if (ctl->arct_status != ARISTA_STATUS_OUTOFSYNC) {
	if (ctl->arct_kf_ptp_wait == 0){ /* arm a pseudo-timer */
	  ctl->arct_kf_ptp_wait = arista_host_nsecs() + ARISTA_KF_PTP_WAIT;
	}
	if (ctl->arct_kf_ptp_wait < arista_host_nsecs()) { /* really out-of-sync */
	  ctl->arct_kf_ptp_wait = 0;
	  ctl->arct_status = ARISTA_STATUS_OUTOFSYNC;
	  ctl->arct_stats.arst_ptpoos++;
	  rc |= ARISTA_MODE_PTP_DISCONNECTED;
	}
      }
    }
    else if (lastkf->ak_last_sync_time > 0 && 
	     ctl->arct_status == ARISTA_STATUS_OUTOFSYNC) {
      ctl->arct_kf_ptp_wait = 0;
      ctl->arct_status = ARISTA_STATUS_ACTIVE;
      rc |= ARISTA_MODE_PTP_RECONNECTED;
    }
  }
  return rc;
}

/* Shortcut helper macros to handle getenv and printing */
#define ARISTA_INT_PARAM(env, value)		\
  sprintf(tmp, "%s", env);			\
  if (getenv(tmp)) {				\
    ge = getenv(tmp);				\
    if (ge != 0 && *ge != '\0') {		\
      opt = atoi(ge);				\
      if (opt == 0) value = 0;			\
      else value = 1;				\
    }						\
  }				

#define ARISTA_STRING(string, test) {		\
    sprintf(tmp, string, test ? "on":"off");	\
    desc = strcat(desc, tmp);			\
  }

#define ARISTA_STAT(string, stat) {		\
    sprintf(tmp, string, stat);			\
    desc = strcat(desc, tmp);			\
  }

/* Handle environment config variables. Return 1 if Arista timestamping
 * is enabled, 0 otherwise.
 */
static inline int
arista_init(ast_ctl_t *actl, char *desc)
{
  int opt;
  char *ge;
  char *prod      = actl->arct_prodname;
  char tmp[ARISTA_MAX_STRING]   = "";
  char switch_ip_addr[32];

  actl->arct_mode           = ARISTA_FCS_TIMESTAMPING_DISABLED; /* need to get activated */
  actl->arct_kf_recv        = 0; /* always turned off by default */
  actl->arct_kf_check_rate  = 1; /* always turned on  by default */
  actl->arct_kf_check_ptp   = 1; /* always turned on  by default */
  actl->arct_kf_dupkern     = 0; /* don't duplicate keyframe to kernel, on DBL only */
  actl->arct_kf_last_ts     = 0;
  actl->arct_kf_flipped     = 0;
  actl->arct_kf_ptp_wait    = -1; 
  actl->arct_kf_hasskew     = 0; /* kf doesn't have skew by default */
  memset(&actl->arct_lastkf, 0, sizeof(ast_kf_t));
  memset(&actl->arct_stats , 0, sizeof(ast_stats_t));

  /* check users' settings */
  strcpy(desc, "** ");

  /* dst addr in the keyframe should match the one of the switch */
  sprintf(tmp, "%s", "MYRI_ARISTA_PARAM_KF_DST_IP");
  if (getenv(tmp)) {
    ge = getenv(tmp);
    if (ge != 0 && *ge != '\0') {
      actl->arct_kf_dst_ip = inet_addr(ge);
      strncpy(switch_ip_addr, ge, MIN(strlen(ge)+1, 32));
    }
  } else {
    actl->arct_kf_dst_ip = inet_addr(ARISTA_DEFAULT_SWITCH_DST_IP);
    strncpy(switch_ip_addr, ARISTA_DEFAULT_SWITCH_DST_IP, 32);
  }

  /* check if the timestamping is enabled */
  sprintf(tmp, "%s", "MYRI_ARISTA_ENABLE_TIMESTAMPING");
  if (getenv(tmp)) {
    ge = getenv(tmp);			
    if (ge != 0 && *ge != '\0') {		
      opt = atoi(ge);				
      if (opt == 0) {
	actl->arct_status  = ARISTA_STATUS_DISABLED;       
	strcat(desc, "Arista timestamping [off] "); 
	return 0;
      } else {
	actl->arct_status  = ARISTA_STATUS_INITED;    
	strcat(desc, "Arista timestamping [on], mode: "); 
	strcat(desc, ARISTA_SUPPORTED_MODE_STR);
	strcat(desc, ", dst_ip: ");
	strcat(desc, switch_ip_addr);
      }
    }
  } else {
    actl->arct_status  = ARISTA_STATUS_DISABLED;       
    strcat(desc, "Arista timestamping [off] "); 
    return 0;
  }

  /* ---- parameters */
  /* SNF-only */
  if (strcmp("SNF", prod) == 0) {
    ARISTA_INT_PARAM("MYRI_ARISTA_PARAM_KF_RECEIVE"      , actl->arct_kf_recv);
  }
  /* DBL-only */
  if (strcmp("DBL", prod) == 0) {
    ARISTA_INT_PARAM("MYRI_ARISTA_PARAM_KF_KERNDUP"      , actl->arct_kf_dupkern);
  }
  /* common */
  ARISTA_INT_PARAM("MYRI_ARISTA_PARAM_KF_CHECK_PTP"    , actl->arct_kf_check_ptp);
  ARISTA_INT_PARAM("MYRI_ARISTA_PARAM_KF_CHECK_RATE"   , actl->arct_kf_check_rate);
  ARISTA_INT_PARAM("MYRI_ARISTA_ENABLE_SKEW"   , actl->arct_kf_hasskew);

  /* ---- print the features selected */
  desc = strcat(desc, "\n\t  * features: ");
  ARISTA_STRING("keyframe check min.rate [%s] ", actl->arct_kf_check_rate);
  ARISTA_STRING("kf check ptp [%s] "           , actl->arct_kf_check_ptp);
  /* SNF-only */
  if (strcmp("SNF", prod) == 0) {
    ARISTA_STRING("kf receive [%s] "             , actl->arct_kf_recv);
  }
  /* DBL-only */
  if (strcmp("DBL", prod) == 0) {
    ARISTA_STRING("kf kernel dup [%s] "          , actl->arct_kf_dupkern);
  }
  ARISTA_STRING("kf skew [%s] "           , actl->arct_kf_hasskew);

  return 1;
}

static inline char *
arista_fini(ast_ctl_t *actl)
{
  char tmp[128]      = "";
  char *desc         = mal_malloc(sizeof(char) * ARISTA_MAX_STRING);
  char *ret          = mal_malloc(sizeof(char) * ARISTA_MAX_STRING);

  /* dump stats */
  if (actl->arct_status != ARISTA_STATUS_DISABLED) {
    sprintf(desc,"Arista stats: ");
    ARISTA_STAT("KF [recv %ld, "             , (long int)actl->arct_stats.arst_kfrecv);
    ARISTA_STAT("flips %ld, "                , (long int)actl->arct_stats.arst_flips);
    ARISTA_STAT("shortages %ld, "            , (long int)actl->arct_stats.arst_kfneeded);
    ARISTA_STAT("malformed %ld] "            , (long int)actl->arct_stats.arst_kfmalformed);
    ARISTA_STAT("Sw [ptp out of sync. %ld, " , (long int)actl->arct_stats.arst_ptpoos);
    ARISTA_STAT("unsupp. %ld, "              , (long int)actl->arct_stats.arst_unsupported_mode);
    ARISTA_STAT("dropped frm %ld, "          , (long int)actl->arct_lastkf.ak_dropped_frames);
    ARISTA_STAT("neg_adj_events %ld]"        , (long int)actl->arct_stats.arst_neg_adj_event);
    sprintf(ret, desc, ARISTA_MAX_STRING);
    goto out;
  }
  ret = NULL;
  
 out:
  mal_free(desc);
  return ret;
}

#ifdef ARISTA_DEBUG
static inline char *
dump_keyframe(ast_kf_t *kf)
{
  char tmp[128]      = "";
  char *desc         = mal_malloc(sizeof(char) * ARISTA_MAX_STRING);
  char *ret          = mal_malloc(sizeof(char) * ARISTA_MAX_STRING);

  sprintf(desc,"KF: ");
  ARISTA_STAT("asic:%llu, ", (unsigned long long)kf->ak_asic_time);
  ARISTA_STAT("utc:%llu, ",  (unsigned long long)kf->ak_utc_time);
  ARISTA_STAT("sync:%llu, \n                 ", (unsigned long long)kf->ak_last_sync_time);
  ARISTA_STAT("skew_num:%llu, ",      (unsigned long long)kf->ak_skew_numerator);
  ARISTA_STAT("skew_denom:%llu, ",      (unsigned long long)kf->ak_skew_denominator);
  ARISTA_STAT("keyf:%llu, ",      (unsigned long long)kf->ak_kf_timestamp);
  ARISTA_STAT("drops:%llu, ",   (unsigned long long)kf->ak_dropped_frames);
  ARISTA_STAT("id:%u, ",        kf->ak_device_id);
  ARISTA_STAT("swport:%u, ",    kf->ak_egress_swport);
  ARISTA_STAT("fcs_type:%u",    kf->ak_fcs_type);
  sprintf(ret, desc, ARISTA_MAX_STRING);
  return ret;
}
#endif

static inline uint32_t
arista_retrieve_frame_timestamp(arista_tsp_t ts)
{
  uint32_t t = ARISTA_GET31_L31b(ts);
  return t;
}

/* Survive to kf drought. In the case of heavy weight traffic,
 basically, unless patched, the Arista switch will stop injecting kf
 when queues are full. We count the flipping of the timestamp to keep
 a relative timing despite the issue generated by Arista. */ 
static inline uint64_t
arista_compute_timestamp(uint32_t ts, ast_ctl_t *ctl)
{
  float d;
  uint64_t pkt_timestamp;

  /* flipped the counter, could be overlap or kf shortage */
  if (unlikely(ts < ctl->arct_kf_last_ts)) {
    /* check if not false alarm, an packet order inversion */
    if (ctl->arct_kf_last_ts > 0xFFFFF000 && ts < 0xFFF) {
      ctl->arct_kf_flipped++;
      ctl->arct_stats.arst_flips++;
    }
  }

  d = (float)( (ts + (ARISTA_TICK_MAX_T * ctl->arct_kf_flipped) -
	       (ctl->arct_lastkf.ak_asic_time & ARISTA_TICK_MAX_T)) * 
               (ctl->arct_kf_hasskew ? ARISTA_TICK_TO_NS *
                  ((double)ctl->arct_lastkf.ak_skew_denominator /
                   (double)ctl->arct_lastkf.ak_skew_numerator) : 
                  ARISTA_TICK_TO_NS));
  ctl->arct_kf_last_ts = ts; 

  pkt_timestamp = ctl->arct_lastkf.ak_utc_time + (uint64_t)(d);

  if (unlikely(pkt_timestamp < ctl->arct_last_ts)) {
    pkt_timestamp = ctl->arct_last_ts;
    ctl->arct_stats.arst_neg_adj_event++;
  }
  ctl->arct_last_ts = pkt_timestamp;

  return (pkt_timestamp);
}

/* Try to survive to MAX_FLIP kf drought */
static inline int
arista_need_new_keyframe(ast_ctl_t *ctl)
{
  if (ctl->arct_kf_flipped == ARISTA_TS_MAX_FLIP && 
      ctl->arct_status != ARISTA_STATUS_NEED_KEYFRAME) {
    return ARISTA_MODE_KF_SHORTAGE_START;
  }
  return 0;
}

#endif /* __ARISTA_SWITCH_H */
