/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2011 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mal.h"
#if !MAL_OS_WINDOWS
#include <sys/errno.h>
#endif

#ifndef offsetof
#define offsetof(type, member) ((size_t) &(((type *) 0)->member))
#endif

#ifdef MAL_KERNEL
#error Cannot compile mal_filt.c into kernel
#endif
#if MAL_OS_MACOSX || MAL_OS_FREEBSD
#error Cannot compile mal_filt.c on macos/freebsd yet
#endif

struct mal_hash_entry
{
  struct mal_hash_entry *next;
  uint32_t key_high;
  uint32_t key_low;
  uint32_t data;
};

struct mal_hash
{
  struct mal_hash_entry **index;
  uint32_t mask;
  uint32_t used_cnt;
  uint32_t total_cnt;
  struct mal_hash_entry *free;
  struct mal_hash_entry *table;
  struct mal_hash_entry *lru_entry;
  struct mal_hash_entry lru_dummy;
};


#include "mal_filt.h"

#define hash_func(a, b, m)   (((a) ^ (b)) & (m))  // XXX - straw man

static inline
mal_hash_entry_t
hash_insert(mal_hash_t hash, 
	    const uint32_t data, 
	    const uint32_t key_high, 
	    const uint32_t key_low)
{
  mal_hash_entry_t entry;
  uint32_t i;
  
  /* get free hash entry */
  entry = hash->free;
  
  /* compute the hash index */
  i = hash_func(key_high, key_low, hash->mask);
  
  if (entry) {
    /* consume entry */
    assert(hash->used_cnt < hash->total_cnt);
    hash->used_cnt++;
    hash->free = entry->next;

    /* init entry */
    entry->next = hash->index[i];
    entry->data = data;
    entry->key_high = key_high;
    entry->key_low = key_low;
    
    /* insert the entry in the hash list */
    hash->index[i] = entry;
    
    /* reset LRU */
    hash->lru_entry = &hash->lru_dummy;
  }
  
  return entry;
}


static inline
void
hash_remove(mal_hash_t hash, 
	    const mal_hash_entry_t entry, 
	    const uint32_t key_high, 
	    const uint32_t key_low)
{
  uint32_t i;
  mal_hash_entry_t current, prev, next;
  
  /* compute the hash index */
  i = hash_func(key_high, key_low, hash->mask);
  
  /* start at the beginning of the hash list */
  assert(hash->used_cnt > 0);
  current = hash->index[i];
  prev = (mal_hash_entry_t)(uintptr_t)(&hash->index[i]);

  /* reset LRU */
  hash->lru_entry = &hash->lru_dummy;
  
  /* walk the hash list and remove entry when found */
  while (current) {
    next = current->next;
    if (likely(current == entry)) {
      prev->next = next;
      current->next = hash->free;
      hash->free = current;
      hash->used_cnt--;
      return;
    }
    prev = current;
    current = next;
  }
}


void
mal_hash_stats(mal_hash_t hash, int verbose)
{
  uint32_t i;
  mal_hash_entry_t current;
  int cnt = 0;
  int bkts;
  int max_bkts = 0;

  for (i = 0; i < hash->total_cnt * 2; i++) {
    if (verbose) printf("%d: ", i);
    current = hash->index[i];
    bkts = 0;
    while (current) {
      bkts++;
      if (verbose) printf("x");
      current = current->next;
      cnt++;
    }
    if (verbose) printf("\n");
    if (bkts > max_bkts)
      max_bkts = bkts;
  }
  printf("Total entries:%d, max bucket/entry:%d\n", cnt, max_bkts);
}


static inline
mal_hash_entry_t
mal_hash_lookup(mal_hash_t hash, const uint32_t key_high, const uint32_t key_low)
{
  uint32_t i;
  mal_hash_entry_t entry;
  
  /* LRU */
  entry = hash->lru_entry;
  if (likely((entry->key_high == key_high) && (entry->key_low == key_low))) {
    return entry;
  }

  /* compute the hash index */
  i = hash_func(key_high, key_low, hash->mask);
  
  /* start at the beginning of the hash list */
  entry = hash->index[i];
  
  /* look for the matching entry in the hash list */
  while (entry) {
    
    /* test if it's a match */
    if (likely((entry->key_high == key_high) && (entry->key_low == key_low))) {
      hash->lru_entry = entry;
      return entry;
    }
    
    /* try the next entry in the list */
    entry = entry->next;
  }
  
  return NULL;
}


uint32_t
mal_hash_tag_lookup(mal_hash_t hash, const uint32_t key_high, const uint32_t key_low)
{
  mal_hash_entry_t e;
  e = mal_hash_lookup(hash, key_high, key_low);
  if (e)
    return e->data;
  else
    return 0xffffffff;
}

int
mal_hash_init(mal_hash_t *hash, const int cnt)
{
  int i;
  struct mal_hash *hash_o;

  /* cnt should be a power of 2 */
  assert((cnt & (cnt-1)) == 0);

  /* Assert next is first in struct for silly prev ptr trick */
  assert(offsetof(struct mal_hash_entry, next) == 0);

  hash_o = (struct mal_hash *)malloc(sizeof(struct mal_hash));
  if (hash == NULL)
    return -ENOMEM;
  
  hash_o->mask = (cnt * 2) - 1;
  hash_o->used_cnt = 0;
  hash_o->total_cnt = cnt;
  hash_o->lru_entry = &hash_o->lru_dummy;
  hash_o->lru_dummy.key_high = 0;
  hash_o->lru_dummy.key_low = 0;

  hash_o->table = (mal_hash_entry_t)malloc(cnt * sizeof(struct mal_hash_entry));
  if (hash_o->table == NULL) {
    return -ENOMEM;
  }
  hash_o->index = (mal_hash_entry_t *)malloc(cnt * 2 * sizeof(hash_o->index[0]));
  if (hash_o->index == NULL) {
    return -ENOMEM;
  }
  memset(hash_o->index, 0, cnt * 2 * sizeof(hash_o->index[0]));

  /* init free list */
  for (i = 0; i < cnt; i++)
    hash_o->table[i].next = &hash_o->table[i+1];
  hash_o->table[cnt-1].next = 0;
  hash_o->free = &hash_o->table[0];

  *hash = hash_o;
  return 0;
}


static inline
int
pkt_parse(uint8_t *pkt, uint32_t *key_high, uint32_t *key_low)
{
  uint32_t ip_hlen, ip_proto, ip_dest_16h, ip_dest_16l, ip_port, tcp;
  uint16_t pkt_proto;
  uint8_t *frame;
  uint16_t vlan;
  uint32_t dest_mac;

  dest_mac = *(uint32_t *)pkt;
  pkt_proto = *(uint16_t *)(pkt + 12);
  frame = pkt + 14;
  vlan = 0;
  if (unlikely(pkt_proto == 0x8100)) {
    vlan = *(uint16_t *)(pkt + 12 + 2);
    pkt_proto = *(uint16_t *)(pkt + 12 + 4);
    frame += 4;
  }

  ip_hlen = *(uint8_t *)frame;
  ip_proto = *(uint8_t *)(frame + 9);
  ip_dest_16l = *(uint16_t *)(frame + 16);
  ip_dest_16h = *(uint16_t *)(frame + 18);

  /* check if UDP */
  tcp = 0x10000;
  if (unlikely(ip_proto == 17)) {
    ip_proto = 6;
    tcp = 0;
  }

  /* interesting vlan bits are the last 12 */
  vlan = vlan << 20;

  /* if not TCP or UDP, bail out */
  if (unlikely(ip_proto != 6))
    return 0;

  ip_hlen = (ip_hlen & 0x0000000f) << 2;
  ip_port = *(uint16_t *)(frame + ip_hlen + 2);

  /* compose hash keys */
  *key_high = (ip_dest_16h << 16) + ip_dest_16l;

  /* If dest mac is 0xffffffff, use IP dest broadcast */
  if (unlikely(dest_mac == 0xffffffff))
    *key_high = 0xffffffff;
  *key_low = ip_port | tcp | vlan;
  return 1;
}


void
mal_add_filter(mal_hash_t hash, const uint32_t tag, const uint32_t key_high,
                             const uint32_t key_low)
{
  hash_insert(hash, tag, key_high, key_low);
  return;
}


void
mal_del_filter(mal_hash_t hash, const uint32_t tag, const uint32_t key_high, const uint32_t key_low)
{
  mal_hash_entry_t entry;
  if (likely(entry = mal_hash_lookup(hash, key_high, key_low))) {
    do {
       if (likely(entry->data == tag)) {
         hash_remove(hash, entry, key_high, key_low);
         break;
       }
       entry = entry->next;
     } while (entry);
   }
}


int
mal_get_filter_tag(mal_hash_t hash, uint8_t *pkt)
{
  uint32_t key_high, key_low;
  mal_hash_entry_t entry;
  if (likely(pkt_parse(pkt, &key_high, &key_low))) {
    if (likely(entry = mal_hash_lookup(hash, key_high, key_low))) {
      return entry->data;
    }
  }
  return -1;
}
