/*************************************************************************
 * The contents of this file are subject to the MYRICOM ABSTRACTION      *
 * LAYER (MAL) SOFTWARE AND DOCUMENTATION LICENSE (the "License");       *
 * User may not use this file except in compliance with the License.     *
 * The full text of the License can found in LICENSE.TXT                 *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2009 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

#ifndef MAL_KERNEL
#include <stdio.h>
#include <stdarg.h>
#if MAL_OS_LINUX
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <dlfcn.h>
#elif MAL_OS_WINDOWS
#include <windows.h>
#elif MAL_OS_FREEBSD || MAL_OS_MACOSX
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#elif MAL_OS_SOLARIS
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stropts.h>
#include <sys/mman.h>
#include <sys/errno.h>
#elif MAL_OS_UDRV
#include <sys/socket.h>
#include <sys/un.h>
#include <alloca.h>
#include <unistd.h>
#include <sys/mman.h>
#include "mx_arch_comm.h"

static uint32_t mx_address_space;
char *mx__init_brk;
#else
#error
#endif

#include "mal.h"
#include "mal_io.h"
#include "mal_valgrind.h"


#if MAL_OS_WINDOWS
/* Windows error values are not the same as errno. Do our best to map
 * from one to the other. If the customer bit (1 << 29) is set the
 * return value came from us and all we have to do is strip off the
 * top bits. Otherwise, it's a Windows value and we try to convert it
 * into an errno.
 */
static DWORD translate_last_error(DWORD last_error)
{
  if (last_error & (1 << 29)) {
    last_error &= ~((1 << 29) | (1 << 30) | (1 << 31));
  }
  else {
    switch (last_error) {
    case ERROR_PATH_NOT_FOUND:
      last_error = ENODEV;
      break;
    default:
      last_error = EINVAL;
      break;
    }
  }

  return last_error;
}
#endif

/* For systems w/o cloning devices, we use the convenction that:
 *  endpoint == -1 means open the generic board device (/dev/myrip0)
 *  endpoint == -2 means open the raw board device (/dev/myrir0) 
 */
MAL_FUNC(int)
mal_open(int unit, int endpoint, mal_handle_t *handle, uint8_t product)
{
#if MAL_OS_LINUX || MAL_OS_FREEBSD || MAL_OS_MACOSX || MAL_OS_SOLARIS
  char buf[80];
  int rc;
  extern int errno;

  unsigned char prodchar = 'A' + product;

  mal_init();

  if (product >= MAL_MAX_PRODUCTS) {
    return (EINVAL);
  }

#if MAL_OS_SOLARIS
    sprintf(buf, "/dev/myri%ccp%d", prodchar,unit);
#else
    sprintf(buf, "/dev/myri%cp%d", prodchar,unit);
#endif

  if (MAL_OS_MACOSX)
    switch (endpoint) {
    case -2: 
      sprintf(buf, "/dev/myri%cr%d", prodchar,unit); break;
    case -1: 
      sprintf(buf, "/dev/myri%cp%d", prodchar,unit); break;
    default: 
      sprintf(buf, "/dev/myri%c%de%d", prodchar, unit, endpoint); 
      break;
    }

  rc = open(buf, O_RDWR);
  if (rc != -1) {
    *handle = rc;
    return 0;
  }

  /* try again with unprivledged devices */
  if (MAL_OS_MACOSX && endpoint >= 0) {
    sprintf(buf, "/dev/myri%c%de%d", prodchar, unit, endpoint);
  } else {
#if MAL_OS_SOLARIS
    sprintf(buf, "/dev/myri%cc%d", prodchar, unit);
#else
    sprintf(buf, "/dev/myri%c%d", prodchar, unit);
#endif
  }

  rc = open(buf, O_RDWR);
  if (rc != -1) {
    *handle = rc;
    return 0;
  }

  return errno;

#elif MAL_OS_WINDOWS
  char buf[80];

  unsigned char prodchar = 'A' + product;

  mal_init();

  /* TODO: Error handling. */
  if (endpoint == -1) {
    sprintf(buf, "\\\\.\\myri%c%d\\p", prodchar, unit);
    *handle = CreateFile(buf, FILE_READ_ATTRIBUTES | SYNCHRONIZE, 0, NULL,
			 OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL);
    if (*handle != INVALID_HANDLE_VALUE)
      return 0;
    else {
      sprintf(buf, "\\\\.\\myri%d", unit);
      *handle = CreateFile(buf, FILE_READ_ATTRIBUTES | SYNCHRONIZE, 0, NULL,
			   OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL);
      if (*handle != INVALID_HANDLE_VALUE)
	return 0;
      else {
	errno = translate_last_error(GetLastError());
	return errno;
      }
    }
  }
  else if (endpoint == -2) {
    sprintf(buf, "\\\\.\\myri%c%d\\r", prodchar, unit);
    *handle = CreateFile(buf, FILE_READ_ATTRIBUTES | SYNCHRONIZE, 0, NULL,
			 OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL);
    if (*handle != INVALID_HANDLE_VALUE)
      return 0;
    else {
      errno = translate_last_error(GetLastError());
      return errno;
    }
  }
  else {
    sprintf(buf, "\\\\.\\myri%c%d\\e%d", prodchar, unit, endpoint);
    *handle = CreateFile(buf, FILE_READ_ATTRIBUTES | SYNCHRONIZE, 0, NULL,
			 OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL);
    if (*handle != INVALID_HANDLE_VALUE)
      return 0;
    else {
      errno = translate_last_error(GetLastError());
      return errno;
    }
  }

#elif MAL_OS_UDRV
  int rc, sock;
  struct sockaddr_un un;
  
  mal_init();

  if (unit > 0)
    return EINVAL;
  sprintf(un.sun_path, MX_IOCTL_SOCK);
  sock = socket(AF_UNIX, SOCK_STREAM, 0);
  if (sock < 0)
    return errno;
  un.sun_family = AF_UNIX;
  rc = connect(sock,(struct sockaddr*)&un,sizeof(un));
  if (rc < 0) {
    close(sock);
    return errno;
  }
  *handle = sock;
  return 0;
#else
#error
#endif
}

MAL_FUNC(int)
mal_open_any_board(mal_handle_t *handle, uint8_t product)
{
  int i, rc;
  int chrdev_exists = 0, driver_loaded = 0, perms_ok = 0;
  for (i = 0; i <= 32; i++) {
    if (!(rc = mal_open_board(i, handle, product)))
      return 0;
    else if (rc == ENOENT)
      break;
    else {
      chrdev_exists = 1;
      if (rc == ENODEV || rc == ENXIO)
	break;
      else {
	driver_loaded = 1;
	if ((rc != EACCES && rc != EPERM))
	  perms_ok = 1;
      }
    }
  }

  if (!chrdev_exists)
    return ENOENT;
  else if (!driver_loaded)
    return ENODEV;
  else if (!perms_ok)
    return EACCES;
  else
    return EBUSY;
}

MAL_FUNC(int)
mal_close(mal_handle_t handle)
{
#if MAL_OS_LINUX || MAL_OS_FREEBSD || MAL_OS_MACOSX || MAL_OS_SOLARIS
  extern int errno;
  
  if (__mal_close(handle))
    return errno;
  else
    return 0;
    
#elif MAL_OS_WINDOWS
  /* TODO: Error handling. */
  CloseHandle(handle);
  return 0;
#elif MAL_OS_UDRV
  shutdown(handle, SHUT_RDWR);
  if (close(handle))
    return errno;
  else
    return 0;
#else
#error
#endif
}

MAL_FUNC(int)
mal_ioctl(mal_handle_t handle, int cmd, void *buf, size_t bufsize)
{
#if MAL_OS_LINUX || MAL_OS_FREEBSD || MAL_OS_MACOSX || MAL_OS_SOLARIS
  {
    int ret;
    MAL_VALGRIND_PRE_IOCTL_CHECK(cmd, buf);
    ret = __mal_ioctl(handle, cmd, (unsigned long) (uintptr_t) buf);
    if (ret == 0) {
      MAL_VALGRIND_POST_IOCTL_CHECK(cmd, buf);
      return 0;
    } else
      return errno;
  }
  
#elif MAL_OS_WINDOWS
  DWORD bytes;

  errno = 0;
  if (DeviceIoControl(handle, cmd, buf, (DWORD)bufsize, buf,
		      (DWORD)bufsize, &bytes, NULL) == 0) {
    errno = translate_last_error(GetLastError());
    return errno;
  }
  else
    return 0;
  
#elif MAL_OS_UDRV
  struct iobuf {
    uint32_t cmd_status;
    char data[0];
  } *iobuf = 0;
  mx_nic_id_hostname_t *nic_host = buf;
  mx_set_hostname_t *set_host = buf;
  myri_raw_send_t *rsend = buf;
  myri_raw_next_event_t *rnext = buf;
  myri_get_logging_t *glog = buf;
  mx_get_eeprom_string_t *geeprom = buf;
  mx_reg_t rdma;
  int isize;
  int rc;

  /* handle scatter/gather cases of the driver/lib interface */
  switch (cmd) {
  case MX_REGISTER:
    isize = bufsize;
    rdma = *(mx_reg_t *)buf;
    mal_always_assert(rdma.nsegs == 1);
    mal_always_assert(rdma.segs.vaddr >= (size_t)mx__init_brk);
    rdma.segs.vaddr = ((uint64_t)mx_address_space << 32) +
      0x80000000 + rdma.segs.vaddr - (size_t)mx__init_brk;
    buf = &rdma;
    break;
  case MX_GET_PRODUCT_CODE:
  case MX_GET_PART_NUMBER:
    isize = bufsize;
    bufsize = bufsize + 256;
    break;
  case MYRI_GET_COUNTERS_VAL:
    isize = bufsize;
    bufsize = 16384;
    break;
  case MYRI_GET_LOGGING:
    isize = bufsize;
    bufsize += glog->size;
    break;
  case MX_GET_PEER_TABLE:
    bufsize = 16384;
    isize = 0;
    break;
  case MX_NIC_ID_TO_HOSTNAME:
  case MX_HOSTNAME_TO_NIC_ID:
    bufsize += nic_host->len;
    isize = bufsize;
    iobuf = alloca(bufsize + sizeof(*iobuf));
    memcpy(iobuf->data, buf, sizeof(*nic_host));
    memcpy(iobuf->data + sizeof(*nic_host), 
	   (void*)(uintptr_t)nic_host->va, nic_host->len);
    break;
  case MX_SET_HOSTNAME:
    bufsize += set_host->len;
    isize = bufsize;
    iobuf = alloca(bufsize + sizeof(*iobuf));
    memcpy(iobuf->data, buf, sizeof(*set_host));
    memcpy(iobuf->data + sizeof(*set_host), 
	   (void*)(uintptr_t)set_host->va, set_host->len);
    break;
  case MYRI_RAW_SEND:
    bufsize += rsend->data_length + rsend->route_length;
    isize = bufsize;
    iobuf = alloca(bufsize + sizeof(*iobuf));
    memcpy(iobuf->data, buf, sizeof(*rsend));
    memcpy(iobuf->data + sizeof(*rsend), 
	   (void*)(uintptr_t)rsend->route_pointer, rsend->route_length);
    memcpy(iobuf->data + sizeof(*rsend) + rsend->route_length,
	   (void*)(uintptr_t)rsend->data_pointer, rsend->data_length);
    break;
  case MYRI_RAW_GET_NEXT_EVENT:
    isize = bufsize;
    bufsize += 2048;
    break;
  case MYRI_GET_COUNTERS_STR:
    isize = bufsize;
    bufsize = 32768;
    break;
  default:
    isize = bufsize;
  }
  if (!iobuf) {
    iobuf = alloca(bufsize + sizeof(*iobuf));
    memcpy(iobuf->data, buf, isize);
  }
  iobuf->cmd_status = cmd;
  rc = udrv_send(handle, iobuf,bufsize + sizeof(*iobuf), 0);
  mal_always_assert(rc == bufsize + sizeof(*iobuf));
  rc = udrv_recv(handle, iobuf,bufsize + sizeof(*iobuf), 0);
  mal_always_assert(rc >= sizeof(*iobuf) && rc <= bufsize + sizeof(*iobuf));
  /* handle wierd driver/lib interface */
  if (iobuf->cmd_status == 0) {
    uint64_t old_field;
    switch (cmd) {
    case MX_GET_PRODUCT_CODE:
    case MX_GET_PART_NUMBER:
      strncpy((void*)(uintptr_t)geeprom->buffer, 
	      iobuf->data + sizeof(*geeprom), MYRI_MAX_STR_LEN);
      memcpy(geeprom, iobuf->data, sizeof(*geeprom));
      break;
    case MX_NIC_ID_TO_HOSTNAME:
    case MX_HOSTNAME_TO_NIC_ID:
      memcpy((void*)(uintptr_t)nic_host->va, 
	     iobuf->data + sizeof(*nic_host), nic_host->len);
      memcpy(nic_host, iobuf->data, sizeof(*nic_host));
      break;
    case MX_SET_HOSTNAME:
      memcpy((void*)(uintptr_t)set_host->va, 
	     iobuf->data + sizeof(*set_host), set_host->len);
      memcpy(set_host, iobuf->data, sizeof(*set_host));
      break;
    case MYRI_RAW_GET_NEXT_EVENT:
      old_field = rnext->recv_buffer;
      memcpy(rnext, iobuf->data, sizeof(*rnext));
      rnext->recv_buffer = old_field;
      if (rnext->status == MYRI_RAW_RECV_COMPLETE)
	memcpy((void*)(uintptr_t)rnext->recv_buffer, 
	       iobuf->data + sizeof(*rnext), rnext->recv_bytes);
      break;
    case MYRI_GET_LOGGING:
      memcpy((char*)(uintptr_t)glog->buffer, iobuf->data + sizeof(*glog), 
	     rc - sizeof(*iobuf) - sizeof(*glog));
      break;
    default:
      memcpy(buf, iobuf->data, rc - sizeof(*iobuf));
    }
  } else {
    fprintf(stderr,"command %x returned errno=%d:%s\n", cmd, 
	    iobuf->cmd_status, strerror(iobuf->cmd_status));
  }
  return iobuf->cmd_status;
#else
#error
#endif
}

MAL_FUNC(void *)
mal_mmap(void *start, size_t length, mal_handle_t handle, uintptr_t offset, int read_only)
{
#if MAL_OS_LINUX || MAL_OS_FREEBSD || MAL_OS_SOLARIS
  unsigned flags = PROT_READ;
  if (!read_only)
    flags |= PROT_WRITE;

  return mmap(start, length, flags, MAP_SHARED, handle, offset);
#elif MAL_OS_MACOSX || MAL_OS_WINDOWS
  myri_mmap_t x;

  x.offset = (uint32_t)offset;
  x.len = (uint32_t)length;
  x.va = (uint64_t)(uintptr_t)start;
  x.requested_permissions = -1;  /* XXX does windows need this?? */
  if (mal_ioctl(handle, MYRI_MMAP, &x, sizeof(x)) == 0)
    return (void *)(uintptr_t)x.va;
  else
    return (void *)-1;

#elif MAL_OS_UDRV
  mx_mmapio_t x;
  int fd;
  char *ptr;
  unsigned long pad;

  memset(&x,0,sizeof(x));
  x.offset = (uint32_t)offset;
  x.len = (uint32_t)length;
  x.as = mx_address_space;
  if (mal_ioctl(handle, MYRI_MMAP, &x, sizeof(x)))
    return (void *)-1;
  mx_address_space = x.as;
  fd = open(x.outpath,O_RDWR);
  if (fd < 0)
    return (void *)-1;
  pad = 0;
  mal_always_assert(!start || pad == 0);
  ptr = mmap(start, x.len, PROT_READ | (read_only ? 0 : PROT_WRITE),
	     MAP_SHARED | MAP_FILE | (start ? MAP_FIXED : 0), 
	     fd, x.offset - pad);
  close(fd);
  return (uintptr_t)ptr == -1 ? ptr : ptr + pad;
#else
#error
#endif
}

MAL_FUNC(int)
mal_munmap(void *start, size_t length)
{
#if MAL_OS_LINUX || MAL_OS_FREEBSD || MAL_OS_SOLARIS || MAL_OS_UDRV
  return munmap(start, length);
#elif MAL_OS_MACOSX || MAL_OS_WINDOWS
  return 0;
#else
#error what OS !?
#endif
}

/*
 * provide functionality to real system calls so we can bypass 
 * the versions that our various products may overload
 */

mal_socket_functions_t _mal_fops;

#define _STRINGIFY_HELPER(x) #x
#define _STRINGIFY(x) _STRINGIFY_HELPER(x)

#if MAL_OS_LINUX && defined(MAL_HAVE_LIBDL)
#ifndef RTLD_DEFAULT
#define RTLD_DEFAULT      ((void *) 0)
#endif
#endif

/*
 * Initialize function pointers to default library functions in case
 * this code is run in a preloaded environment
 */
void
mal_init_fops()
{
#ifdef MAL_HAVE_LIBDL
  void (*func)(mal_socket_functions_t *);

  /* If we have libdl and mtcp_setup_socket_funcs() exists, call it
   * to set up our sockets functions
   */
  func = (typeof(func)) dlsym(RTLD_DEFAULT, "mtcp_setup_socket_funcs");
  if (func != NULL) {
    func(&_mal_fops);
    return;
  }
#endif

#define assign_sym(fns, sym) do { (fns).sym = sym; } while (0)
#if !MAL_OS_WINDOWS
  assign_sym(_mal_fops, dup);
  assign_sym(_mal_fops, dup2);
  assign_sym(_mal_fops, ioctl);
  assign_sym(_mal_fops, close);
#endif
  assign_sym(_mal_fops, socket);
  assign_sym(_mal_fops, bind);
  assign_sym(_mal_fops, accept);
  assign_sym(_mal_fops, listen);
  assign_sym(_mal_fops, connect);
  assign_sym(_mal_fops, recv);
  assign_sym(_mal_fops, send);
  assign_sym(_mal_fops, recvfrom);
  assign_sym(_mal_fops, sendto);
  assign_sym(_mal_fops, setsockopt);
  assign_sym(_mal_fops, getsockopt);
  assign_sym(_mal_fops, getsockname);
  assign_sym(_mal_fops, getpeername);
  assign_sym(_mal_fops, select);

  return;
}

/*
 * ioctl is special since mal_ioctl already does something
 * slightly different than regular ioctl.  It will call __mal_ioctl
 * when it wants to actually perform the ioctl
 */
#if !MAL_OS_WINDOWS
MAL_FUNC(int)
__mal_ioctl(mal_handle_t handle, int cmd, ...)
{
  va_list args;
  long arg;

  va_start(args, cmd);
  arg = va_arg(args, long);
  va_end(args);

  return _mal_fops.ioctl(handle, cmd, arg);
}

MAL_FUNC(int)
__mal_close(int s)
{
  return _mal_fops.close(s);
}
#endif

MAL_FUNC(int)
mal_closesocket(mal_socket_t s)
{
#if MAL_OS_WINDOWS
  return closesocket(s);
#else
  return _mal_fops.close(s);
#endif
}

MAL_FUNC(int)
mal_getsockopt(mal_socket_t s, int level, int optname, void *optval, mal_socklen_t *optlen)
{
  return _mal_fops.getsockopt(s, level, optname, optval, optlen);
}

MAL_FUNC(int)
mal_setsockopt(mal_socket_t s, int level, int optname, const void *optval,
    mal_socklen_t optlen)
{
  return _mal_fops.setsockopt(s, level, optname, optval, optlen);
}

MAL_FUNC(mal_ret_len_t)
mal_sendto(mal_socket_t s, const void *msg, mal_len_t len, int flags,
       const struct sockaddr *to, mal_socklen_t tolen)
{
  return _mal_fops.sendto(s, msg, len, flags, to, tolen);
}

MAL_FUNC(mal_socket_t)
mal_socket(int af, int type, int protocol)
{
  mal_init();
  return _mal_fops.socket(af, type, protocol);
}

MAL_FUNC(int)
mal_bind(mal_socket_t s, const struct sockaddr *addr, mal_socklen_t addrlen)
{
  return _mal_fops.bind(s, addr, addrlen);
}

MAL_FUNC(int)
mal_getsockname(mal_socket_t s, struct sockaddr *name, mal_socklen_t *namelen)
{
  return _mal_fops.getsockname(s, name, namelen);
}

MAL_FUNC(int)
mal_select(int nfds, fd_set *readfds, fd_set *writefds,
               fd_set *exceptfds, struct timeval *timeout)
{
  return _mal_fops.select(nfds, readfds, writefds, exceptfds, timeout);
}

#endif /* MAL_KERNEL */
