
/* Used by the  myri_* tools so that board numbers can span our 
   different products */
MAL_FUNC(int)
mal_boardnum_to_product(int num, uint8_t *product, int *unit) {
  uint8_t prod;
  int unum;
  uint32_t bcount, bmax;
  uint32_t total = 0;
  mal_handle_t handle = (mal_handle_t)-1;
  int rc;
  int i;

  for (prod=0; prod < MAL_MAX_PRODUCTS; prod++) {
    if (mal_open_any_board(&handle, prod)) {
      continue;
    }
    rc = (mal_ioctl(handle, MYRI_GET_BOARD_COUNT, &bcount, sizeof(bcount)) ||
	  mal_ioctl(handle, MYRI_GET_BOARD_MAX, &bmax, sizeof(bmax)));
    mal_close(handle);
    if (rc) {
      continue;
    }

    if (total + bcount > (uint32_t)num) {
      /* Because of linux pci hotplug, and windows driver handling, we need
	 to handle 'holes' in the driver's instance state.  We have to 
	 open each device before we can count it */
      unum = 0;
      for (i=0; i<(int)bmax; i++) {
	if (mal_open_board(i, &handle, prod) == 0) {
	  unum++;
	  mal_close(handle);
	}
	if (total + (unum-1) == num) {   /* unit number counts 0..x */ 
	  *product = prod;
	  *unit = unum-1;                /* unum will be 1 when we find unit 0 */
	  return 0;
	}
      }
      return EINVAL; 
    }
    total += bcount;
  }
  /* Board outside the range of existing boards */
  return EINVAL;
}


MAL_FUNC(int)
mal_product_to_boardnum(uint8_t product, int unit, int *num) {
  int prod;
  uint32_t bcount, bmax;
  int total = 0;
  mal_handle_t handle = (mal_handle_t)-1;
  int rc;
  int i;

  /* count the number of boards in the products preceeding us */
  for(prod=0; prod<product; prod++) {
    if (mal_open_any_board(&handle, prod)) {
      continue;
    }
    rc = mal_ioctl(handle, MYRI_GET_BOARD_COUNT, &bcount, sizeof(bcount));
    mal_close(handle);
    if (rc) {
      continue;
    }
    total += bcount;
  }

  /* now open the product we care about and do some sanity checking */
  rc = mal_open_any_board(&handle, product);
  if ( rc ) {
    return rc;
  }
  rc = (mal_ioctl(handle, MYRI_GET_BOARD_COUNT, &bcount, sizeof(bcount)) ||
	mal_ioctl(handle, MYRI_GET_BOARD_MAX, &bmax, sizeof(bmax)));	
  mal_close(handle);
  if (rc)
    return rc;

  if (unit > (int)bmax) {
    return EINVAL;
  }
  /* Because of possible 'holes' in the driver instances, we have to count up 
     to our unit number to see which are openable */
  for (i=0; i<unit; i++) {
    if (mal_open_board(i, &handle, product)) {
      continue;
    }
    total++;
  }
  *num = total;
  return 0; 
}

/* the mac supplied to this function can be a 'short' mac that just
   specifies the last few characters.  The only thing we enforce is 
   that it be at least 3 characters long to encourage people to include
   at least one ':' */
MAL_FUNC(int)
mal_mac_to_product(char *mac, uint8_t *product, int *unit) {
  uint8_t prod;
  mal_handle_t handle = (mal_handle_t)-1;
  uint32_t bmax;
  int rc;
  myri_get_nic_id_t nicid;
  int mac_len = 0;
  char nic_mac[18];
  int i;

  if (mac == NULL) {
    return EINVAL;
  }
  mac_len = (int)strlen(mac);
  if (mac_len < 3 || mac_len > 17) {
    return EINVAL;
  }

  for (prod=0; prod < MAL_MAX_PRODUCTS; prod++) {
    if (mal_open_any_board(&handle, prod)) {
      continue;
    }
    rc = (mal_ioctl(handle, MYRI_GET_BOARD_MAX, &bmax, sizeof(bmax)));
    mal_close(handle);
    if (rc) {
      continue;
    }
    for (i=0; i<(int)bmax; i++) {
      if (mal_open_board(i, &handle, prod)) {
	continue;
      }
      nicid.board=i;
      rc = mal_ioctl(handle, MYRI_GET_NIC_ID, &nicid, sizeof(nicid));
      mal_close(handle);
      if (rc) {
	continue;
      }
      mal_nic_id_to_str(nic_mac, nicid.nic_id, sizeof(nic_mac));
      /* a mac string should have up to 17 characters.  Compare the mac we have
	 been given with the end of the mac string we have retrieved from the 
	 nic.  This allows shorter macs consisting of the end bytes*/
      if ( strncasecmp(mac, nic_mac+(17-mac_len),mac_len) == 0 ) {
	*product = prod;
	*unit = i;
	return 0;
      }
    }
  }
  /* Couldn't find the mac address */
  return EINVAL;
}

/* helper function that takes either a board number or a mac address as
   the board identifier */
MAL_FUNC(int)
mal_translate_to_product(char *board, uint8_t *product, int *unit) {
#if !MAL_OS_WINDOWS
  uint8_t mac_addr[6];
  uint64_t nic_id;
  char mac_str[18];
#endif
  char *b;
  char *endptr;
  long int bnum;

  /* If the board is NULL just return the info for board 0 */
  if (board == NULL) {
    return mal_boardnum_to_product(0, product, unit);
  }
#if MAL_OS_WINDOWS
  /* don't do mal_get_macaddr on windows, as interface names on windows
     don't work quite the same way */ 
  b = board;
#else
  /* first assume board is an interface and try converting to mac*/
  if (mal_get_macaddr(board, mac_addr)) {
    b = board;
  } else {
    mal_macaddr_to_nic_id(mac_addr, &nic_id);
    mal_nic_id_to_str(mac_str, nic_id, sizeof(mac_str));
    b = mac_str;
  }
#endif
  /* if it has a : in it, assume it is a mac address */
  if (strchr(b, ':') != NULL) {
    return mal_mac_to_product(b, product, unit);
  }

  /* otherwise if it converts cleanly to a number use it */
  bnum = strtol(b, &endptr, 10);
  if (*endptr != '\0') {
    return EINVAL;
  } else {
    return mal_boardnum_to_product(bnum, product, unit);
  }
}
