/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2011 by Myricom, Inc.  All rights reserved.          *
 * Copyright 2015 by CSP, Inc.  All rights reserved.                     *
 *************************************************************************/

#include "mal.h"
#include "mal_io.h"
#include "mal_valgrind.h"
#include "myri_license_defs.h"
#include "mcp_config.h"
#ifndef MYRI_PHOENIX
#include "mcp_global.h"
#endif

#ifdef MAL_KERNEL
#error Cannot compile mal_utils.c into kernel
#endif

#if !MAL_OS_WINDOWS
#if MAL_OS_MACOSX || MAL_OS_FREEBSD
# include <sys/sysctl.h>
# include <sys/socket.h>
# include <net/if_dl.h>
# include <net/route.h>
#endif
#include <string.h>
#include <netinet/in.h>
#include <net/if.h>
#include <sys/param.h>
#else /* MAL_OS_WINDOWS */
#include <winsock2.h>
#include <ws2ipdef.h>
#include <iphlpapi.h>
#endif

#ifdef HAVE_SYS_SOCKIO_H
#  include <sys/sockio.h>
#endif

#ifdef HAVE_IFADDRS_H
#  include <ifaddrs.h>
#endif

MAL_FUNC(void)
mal_init()
{
  static int init_done=0;
  if (init_done)
    return;

  mal_init_fops();
  init_done = 1;
}

#if MAL_OS_LINUX
int
mal_get_macaddr(const char *ifname, uint8_t mac_addr[6])
{
  struct ifreq ifr;
  int fd;
  int rc = 0;
  
  fd = mal_socket(AF_INET, SOCK_DGRAM, 0);
  if (fd < 0) {
    rc = errno;
  } else {
    strncpy(ifr.ifr_name, ifname, MIN(strlen(ifname)+1, IFNAMSIZ));
    rc = mal_ioctl(fd, SIOCGIFHWADDR, &ifr, sizeof(ifr));
    if (rc != 0) {
      rc = errno;
    } else {
      memcpy(mac_addr, ifr.ifr_hwaddr.sa_data, 6);
    }
    close(fd);
  }
  return rc;
}
#elif MAL_OS_SOLARIS /* Solaris */
#include <net/if_arp.h>
int
mal_get_macaddr(const char *ifname, uint8_t mac_addr[6])
{
  struct ifreq ifr;
  struct xarpreq ar;
  int fd;
  int rc = 0;

  if ((fd = mal_socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    rc = errno;
  else {
    strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
    if (ioctl(fd, SIOCGIFADDR, &ifr))
      rc = errno;
    else {
      memset(&ar, 0, sizeof (ar));
      memcpy(&ar.xarp_pa, &ifr.ifr_addr, sizeof(ifr.ifr_addr));
      ar.xarp_ha.sdl_family = AF_LINK;
      if (ioctl(fd, SIOCGXARP, (caddr_t)&ar) < 0)
        rc = errno;
      else if ((ar.xarp_flags & ATF_AUTHORITY) == 0)
        rc = ENODEV;
      else
        memcpy(mac_addr, (uint8_t *)LLADDR(&ar.xarp_ha), 6);
    }
    close(fd);
  }
  return rc;
}
#elif MAL_OS_MACOSX || MAL_OS_FREEBSD
int
mal_get_macaddr(const char *ifname, uint8_t mac_addr[6])
{
  struct if_msghdr *ifm;
  struct sockaddr_dl *sdl;
  size_t mib_len;
  char *buf, *p;
  uint8_t *macp = NULL;
  int rc = 0;
  int mib[] = { CTL_NET, AF_ROUTE, 0, AF_LINK, NET_RT_IFLIST, 0 };
  size_t ifname_len, len;

  if (sysctl(mib, 6, NULL, &mib_len, NULL, 0) < 0)
    rc = errno;
  else {
    ifname_len = strlen(ifname);
    if ((buf = mal_malloc(mib_len)) == NULL)
      rc = ENOMEM;
    else {
      if (sysctl(mib, 6, buf, &mib_len, NULL, 0) < 0) {
	rc = errno;
	mal_free(buf);
	return rc;
      }
      for (p = buf; !macp && p < buf + mib_len; p += ifm->ifm_msglen) {
        ifm = (struct if_msghdr *)p;
        sdl = (struct sockaddr_dl *)(ifm + 1);
        if (ifm->ifm_type != RTM_IFINFO || (ifm->ifm_addrs & RTA_IFP) == 0)
          continue;
        if (sdl->sdl_family != AF_LINK || !sdl->sdl_nlen)
          continue;
        len = sdl->sdl_nlen > ifname_len ? sdl->sdl_nlen : ifname_len;
        if (memcmp(sdl->sdl_data, ifname, len) == 0)
          macp = (uint8_t *)LLADDR(sdl);
      }
      if (macp)
        memcpy(mac_addr, macp, 6);
      else
        rc = ENODEV;
      mal_free(buf);
    }
  }
  return rc;
}
#elif MAL_OS_WINDOWS
/* Returns non-zero if ipaddress matches any of the strings in
 * ip_addr_string. */
static
int
ipaddress_matches_ip_addr_string(const char *ipaddress,
				 PIP_ADDR_STRING ip_addr_string)
{
  int ret;

  ret = 0;

  while (ip_addr_string) {
    if (strcmp(ipaddress, ip_addr_string->IpAddress.String) == 0) {
      ret = 1;
      break;
    }

    ip_addr_string = ip_addr_string->Next;
  }

  return ret;
}

/* Return a pointer to the adaper info that matches ifname. NULL if not
 * found. Normally ifname is something like eth0 or myri0 but on Windows
 * it's an undecipherable GUID. Rather than deal with that mess we use
 * the convention that ifname will be the string form of the dotted ip
 * address. */
static
PIP_ADAPTER_INFO
lookup_ifname_in_adapter_info(const char *ifname, PIP_ADAPTER_INFO info)
{
  while (info != NULL) {
    if (ipaddress_matches_ip_addr_string(ifname, &info->IpAddressList)) {
      break;
    }

    info = info->Next;
  }

  return info;
}

/* Allocate and return infromation on all the adapters. Returns NULL if
 * unable to for any reason. Caller is responsible for freeing the
 * memory when done. */
static
PIP_ADAPTER_INFO
get_adapters_info(void)
{
  PIP_ADAPTER_INFO info;
  DWORD dw;
  ULONG buf_len;

  info = NULL;

  buf_len = 0;
  dw = GetAdaptersInfo(NULL, &buf_len);
  if (dw == ERROR_BUFFER_OVERFLOW) {
    info = mal_malloc(buf_len);
    if (info != NULL) {
      dw = GetAdaptersInfo(info, &buf_len);
      if (dw != ERROR_SUCCESS) {
	mal_free(info);
	info = NULL;
      }
    }
  }

  return info;
}

MAL_FUNC(int)
mal_get_macaddr(const char *ifname, uint8_t mac_addr[6])
{
  PIP_ADAPTER_INFO info;
  PIP_ADAPTER_INFO matched_info;
  int ret;

  ret = -1;

  info = get_adapters_info();
  matched_info = lookup_ifname_in_adapter_info(ifname, info);
  if (matched_info != NULL && matched_info->AddressLength == 6) {
    memcpy(mac_addr, matched_info->Address, 6);
    ret = 0;
  }

  if (info != NULL) {
    mal_free(info);
  }

  return ret;
}
#elif MAL_OS_UDRV
int
mal_get_macaddr(const char *ifname, uint8_t mac_addr[6])
{
  return ENODEV;
}
#else
#warning mal_get_macaddr() not implemented on this platform
#endif

static
int
mal_match_macaddr(const char *ifname,
                  const struct mal_nic_info *nic_info, int num_nics)
{
  int i;
  uint64_t nic_id;
  uint8_t mac_addr[6];

  memset(mac_addr, 0, 6);

  if (mal_get_macaddr(ifname, mac_addr))
    return -1;
  mal_macaddr_to_nic_id(mac_addr, &nic_id);

  for (i = 0; i < num_nics; i++) {
    if (nic_info[i].curr_nic_id == nic_id)
	return i;
  }
  return -1;
}

MAL_FUNC(int)
mal_getifaddrs(struct mal_ifaddrs **ifaddrs_o, int inet_only, int *num_ifs)
{
  struct mal_ifaddrs *ifaddrs_mal = NULL;
  int num_if_myri = 0, num_if_mal = 0;
  int i, rc = 0;
  struct mal_nic_info *nic_info = NULL;
#if HAVE_IFADDRS_H || MAL_OS_SOLARIS
  struct mal_nic_info *ni;
#endif

  if ((rc = mal_get_nicinfo(&nic_info, &num_if_myri)) || !num_if_myri)
    goto bail;

#if HAVE_IFADDRS_H
  /* Get list of all addresses on this node */
  {
    struct ifaddrs *ifaddr, *ifa;
    int idx;
    int k35snf_listed, k35snf_total;

    if (getifaddrs(&ifaddr) == -1) {
      rc = errno;
      goto bail;
    }

    k35snf_listed = 0;
    k35snf_total = 0;
    for (ifa = ifaddr; ifa != NULL; ifa = ifa->ifa_next) {
      ifa->ifa_data = NULL;
      
      /* make sure there is an address field */
      if (ifa->ifa_addr == NULL) continue;

      if (inet_only && ifa->ifa_addr->sa_family != AF_INET)
        ;
      else if ((idx = mal_match_macaddr(ifa->ifa_name,
                      nic_info, num_if_myri)) >= 0) {
        num_if_mal++;
        ifa->ifa_data = &nic_info[idx];
        if (nic_info[idx].nic_type == MAL_BOARD_TYPE_PHX_K35SNF) {
          k35snf_listed++;
          k35snf_total += nic_info[idx].num_ports;
        }
      }
    }
    if (k35snf_listed > 0)
      num_if_mal += k35snf_total - k35snf_listed;

    if (num_if_mal) {
      struct sockaddr_in *sin;
      char *ifname;
      ifaddrs_mal = mal_malloc((sizeof(struct mal_ifaddrs)+IFNAMSIZ) * num_if_mal);
      if (ifaddrs_mal == NULL) {
        rc = ENOMEM;
      }
      else {
        ifname = (char *)(ifaddrs_mal+num_if_mal);
        for (ifa = ifaddr, i = 0; ifa != NULL; ifa = ifa->ifa_next) {
          ni = (struct mal_nic_info *) ifa->ifa_data;
          if (ni == NULL)
            continue;
          ifaddrs_mal[i].mal_ifa_next = &ifaddrs_mal[i+1];
          strncpy(ifname, ifa->ifa_name, IFNAMSIZ);
          ifname[IFNAMSIZ-1] = '\0';
          ifaddrs_mal[i].mal_ifa_name = ifname;
          ifname += IFNAMSIZ;
          ifaddrs_mal[i].mal_ifa_flags = 0;
          if (ifa->ifa_flags & IFF_BROADCAST)
            ifaddrs_mal[i].mal_ifa_flags |= MAL_IFA_FLAG_BROADCAST;
          if (ifa->ifa_flags & IFF_MULTICAST)
            ifaddrs_mal[i].mal_ifa_flags |= MAL_IFA_FLAG_MULTICAST;
          sin = (struct sockaddr_in *) ifa->ifa_addr;
          if (sin)
            ifaddrs_mal[i].mal_ifa_ipaddr = sin->sin_addr.s_addr;
          sin = (struct sockaddr_in *) ifa->ifa_netmask;
          if (sin)
            ifaddrs_mal[i].mal_ifa_netmask = sin->sin_addr.s_addr;
          sin = (struct sockaddr_in *) ifa->ifa_broadaddr;
          if (sin)
            ifaddrs_mal[i].mal_ifa_bcastaddr = sin->sin_addr.s_addr;
          ifaddrs_mal[i].mal_ifa_boardnum = ni->boardnum;
          ifaddrs_mal[i].mal_ifa_nicid = ni->nic_id;
          mal_nic_id_to_macaddr(ni->nic_id, ifaddrs_mal[i].mal_ifa_macaddr);
          ifaddrs_mal[i].mal_ifa_max_endpoints = ni->nic_max_endpoints;
          ifaddrs_mal[i].mal_ifa_link_state = ni->nic_link_state;
          ifaddrs_mal[i].mal_ifa_snf_maxring = ni->snf_maxring;
          ifaddrs_mal[i].mal_ifa_snf_maxinject = ni->snf_maxinject;
          ifaddrs_mal[i].mal_ifa_link_speed = ni->nic_link_speed;
          i++;

          if (ni->nic_type == MAL_BOARD_TYPE_PHX_K35SNF) {
            int port, num_ports = ni->num_ports;
            for (port = 1; port < num_ports; port++) {
              /* ifa corresponds to port 0. Ports 1, 2, and 3 follow right
               * after in the nic_info array.
               */
              ni++;
              ifaddrs_mal[i].mal_ifa_next = &ifaddrs_mal[i+1];
              snprintf(ifname, IFNAMSIZ, "%s-snf%d", ifa->ifa_name, port);
              ifname[IFNAMSIZ-1] = '\0';
              ifaddrs_mal[i].mal_ifa_name = ifname;
              ifname += IFNAMSIZ;
              ifaddrs_mal[i].mal_ifa_flags = 0;
              ifaddrs_mal[i].mal_ifa_boardnum = ni->boardnum;
              ifaddrs_mal[i].mal_ifa_nicid = ni->nic_id;
              mal_nic_id_to_macaddr(ni->nic_id, ifaddrs_mal[i].mal_ifa_macaddr);
              ifaddrs_mal[i].mal_ifa_max_endpoints = ni->nic_max_endpoints;
              ifaddrs_mal[i].mal_ifa_link_state = ni->nic_link_state;
              ifaddrs_mal[i].mal_ifa_snf_maxring = ni->snf_maxring;
              ifaddrs_mal[i].mal_ifa_snf_maxinject = ni->snf_maxinject;
              ifaddrs_mal[i].mal_ifa_link_speed = ni->nic_link_speed;
              i++;
            }
          }
        }
        ifaddrs_mal[i-1].mal_ifa_next = NULL;
      }
    }
    freeifaddrs(ifaddr);
  }
#elif MAL_OS_SOLARIS
#define _lifr_to_u32(lifrp) (uint32_t)(  \
	    ((struct sockaddr_in *) &((lifrp)->lifr_addr))->sin_addr.s_addr)
  {
    struct lifnum lifn;
    struct lifconf lifc;
    struct lifreq *lifrp = NULL;
    uint64_t nic_id;
    int j, fd, idx;
    int num_if_lifc;
    struct mal_ifaddrs *ifm;
    size_t slen;
    char *ifname;
    uint8_t mac_addr[6];
    struct mal_nic_info **idx_map_nic = NULL;

    if ((fd = mal_socket(AF_INET, SOCK_DGRAM, 0)) < 0)
      rc = errno;
    else {
      lifn.lifn_family = AF_UNSPEC;
      lifn.lifn_flags = 0;
      if (mal_ioctl(fd, SIOCGLIFNUM, &lifn))
        lifn.lifn_count = 16;
      else
        lifn.lifn_count += 16;

      for (;;) {
        lifc.lifc_len = lifn.lifn_count * sizeof(*lifrp);
        lifrp = mal_malloc(lifc.lifc_len);
        if (lifrp == NULL) {
          rc = ENOMEM;
          goto bail;
        }
        lifc.lifc_family = AF_UNSPEC;
        lifc.lifc_flags = 0;
        lifc.lifc_buf = (char *)lifrp;
        if (mal_ioctl(fd, SIOCGLIFCONF, &lifc)) {
          mal_free(lifrp);
          if (errno == EINVAL) {
            lifn.lifn_count <<= 1;
            continue;
          }
          mal_closesocket(fd);
          rc = errno;
          goto bail;
        }
        if (lifc.lifc_len < (lifn.lifn_count - 1) * sizeof(*lifrp))
          break;
        mal_free(lifrp);
        lifn.lifn_count <<= 1;
      }
      /* lifrp contains interfaces */
      num_if_lifc = lifc.lifc_len / sizeof(*lifrp);
      if (!(idx_map_nic = mal_calloc(num_if_lifc, sizeof(struct mal_nic_info *)))) {
        mal_free(lifrp);
        rc = ENOMEM;
        goto bail;
      }

      for (i = 0; i < num_if_lifc; i++) {
        idx_map_nic[i] = NULL;
        if ((idx = mal_match_macaddr(lifrp[i].lifr_name, 
                      nic_info, num_if_myri)) >= 0) {
          idx_map_nic[i] = &nic_info[idx];
          num_if_mal++;
        }
      }

      if (!num_if_mal)
	goto bail;

      ifm = ifaddrs_mal = 
        mal_calloc(1, (sizeof(struct mal_ifaddrs)+IFNAMSIZ) * num_if_mal);
      if (ifaddrs_mal == NULL) {
        mal_free(lifrp);
        mal_free(idx_map_nic);
        rc = ENOMEM;
        goto bail;
      }
      ifname = (char *)(ifaddrs_mal+num_if_mal);
      for (i = 0; i < num_if_lifc; i++) {
        ni = (struct mal_nic_info *) idx_map_nic[i];
        if (ni == NULL)
          continue;
        ifm->mal_ifa_next = ifm + 1;
        strncpy(ifname, lifrp[i].lifr_name, IFNAMSIZ);
        ifname[IFNAMSIZ-1] = '\0';
        ifm->mal_ifa_name = ifname;
        ifname += IFNAMSIZ;
        ifm->mal_ifa_flags = 0;
        if (!mal_ioctl(fd, SIOCGLIFFLAGS, lifrp)) {
          if (lifrp[i].lifr_flags & IFF_BROADCAST)
            ifm->mal_ifa_flags |= MAL_IFA_FLAG_BROADCAST;
          if (lifrp[i].lifr_flags & IFF_MULTICAST)
            ifm->mal_ifa_flags |= MAL_IFA_FLAG_MULTICAST;
        }
        if (!mal_ioctl(fd, SIOCGLIFADDR, &lifrp[i]))
          ifm->mal_ifa_ipaddr = _lifr_to_u32(&lifrp[i]);
        if (!mal_ioctl(fd, SIOCGLIFNETMASK, &lifrp[i]))
          ifm->mal_ifa_netmask = _lifr_to_u32(&lifrp[i]);
        if (ifm->mal_ifa_flags & MAL_IFA_FLAG_BROADCAST)
          if (!mal_ioctl(fd, SIOCGLIFBRDADDR, &lifrp[i]))
            ifm->mal_ifa_bcastaddr = _lifr_to_u32(&lifrp[i]);
        ifm->mal_ifa_boardnum = ni->boardnum;
        ifm->mal_ifa_nicid = ni->nic_id;
        mal_nic_id_to_macaddr(ni->nic_id, ifm->mal_ifa_macaddr);
        ifm->mal_ifa_max_endpoints = ni->nic_max_endpoints;
        ifm->mal_ifa_link_state = ni->nic_link_state;
        ifm->mal_ifa_link_speed = ni->nic_link_speed;
        ifm++;
      }
      if (ifm != ifaddrs_mal)
	ifm[-1].mal_ifa_next = NULL;
      mal_free(lifrp);
      mal_free(idx_map_nic);
      mal_closesocket(fd);
    }
  }
#elif MAL_OS_WINDOWS
  {
    INTERFACE_INFO if_info[32];
    SOCKET s;
    DWORD bytes_returned;
    int num_if;
    int loop;
    struct sockaddr_in *psin;
    IN_ADDR in_addr;
    char *ifname;
    char *tmp_ifname;
    int idx;
    unsigned long ip_addr, netmask, broadcast_addr;

    /* Populate if_info. */
    s = WSASocket(AF_INET, SOCK_DGRAM, 0, NULL, 1, 0);
    if (s == INVALID_SOCKET) {
      rc = -1;
      goto bail;
    }
    if (WSAIoctl(s, SIO_GET_INTERFACE_LIST, NULL, 0, &if_info,
		 sizeof (if_info), &bytes_returned, NULL, NULL) ==
	SOCKET_ERROR) {
      closesocket(s);
      rc = -1;
      goto bail;
    }
    closesocket(s);
    num_if = bytes_returned / sizeof (INTERFACE_INFO);

    /* Count the number of mal compatible interfaces. */
    for (loop = 0; loop < num_if; ++loop) {
      psin = (SOCKADDR_IN*)&if_info[loop].iiAddress;
      in_addr.S_un.S_addr = psin->sin_addr.S_un.S_addr;
      if ((idx = mal_match_macaddr(inet_ntoa(in_addr), nic_info,
				   num_if_myri)) >= 0) {
	num_if_mal++;
      }
    }

    if (num_if_mal == 0) {
      /* rc is not set because zero interfaces is not an error. */
      goto bail;
    }

    ifaddrs_mal = mal_malloc((sizeof (struct mal_ifaddrs) + 16) *
			     num_if_mal);
    if (ifaddrs_mal == NULL) {
      rc = ENOMEM;
      goto bail;
    }

    i = 0;
    ifname = (char*)(ifaddrs_mal+num_if_mal);
    for (loop = 0; loop < num_if; ++loop) {
      psin = (SOCKADDR_IN*)&if_info[loop].iiAddress;
      ip_addr = psin->sin_addr.S_un.S_addr;
      in_addr.S_un.S_addr = psin->sin_addr.S_un.S_addr;
      tmp_ifname = inet_ntoa(in_addr);
      if ((idx = mal_match_macaddr(tmp_ifname, nic_info,
				   num_if_myri)) >= 0) {
	ifaddrs_mal[i].mal_ifa_next = &ifaddrs_mal[i+1];
	strcpy(ifname, tmp_ifname);
	ifaddrs_mal[i].mal_ifa_name = ifname;
	ifname += 16;
	ifaddrs_mal[i].mal_ifa_flags = 0;
	if (if_info[loop].iiFlags & IFF_BROADCAST) {
	  ifaddrs_mal[i].mal_ifa_flags |= MAL_IFA_FLAG_BROADCAST;
	}
	if (if_info[loop].iiFlags & IFF_MULTICAST) {
	  ifaddrs_mal[i].mal_ifa_flags |= MAL_IFA_FLAG_MULTICAST;
	}
	ifaddrs_mal[i].mal_ifa_ipaddr = psin->sin_addr.S_un.S_addr;
	psin = (SOCKADDR_IN*)&if_info[loop].iiNetmask;
	netmask = psin->sin_addr.S_un.S_addr;
	ifaddrs_mal[i].mal_ifa_netmask = psin->sin_addr.S_un.S_addr;
	/*
	  iiBroadcastAddress seems to be broken. Calculate it ourselves.
	psin = (SOCKADDR_IN*)&if_info[loop].iiBroadcastAddress;
	ifaddrs_mal[i].mal_ifa_bcastaddr = psin->sin_addr.S_un.S_addr;
	*/
	broadcast_addr = ip_addr | ~netmask;
	ifaddrs_mal[i].mal_ifa_bcastaddr = broadcast_addr;
	ifaddrs_mal[i].mal_ifa_boardnum = nic_info[idx].boardnum;
	ifaddrs_mal[i].mal_ifa_nicid = nic_info[idx].nic_id;
	mal_nic_id_to_macaddr(nic_info[idx].nic_id,
			      ifaddrs_mal[i].mal_ifa_macaddr);
        ifaddrs_mal[i].mal_ifa_max_endpoints = nic_info[idx].nic_max_endpoints;
        ifaddrs_mal[i].mal_ifa_link_state = nic_info[idx].nic_link_state;
        ifaddrs_mal[i].mal_ifa_link_speed = nic_info[idx].nic_link_speed;
	i++;
      }
    }
    ifaddrs_mal[i-1].mal_ifa_next = NULL;
  }
#else
  fprintf(stderr, "mal_getifaddrs() not available\n");
  rc = ENXIO;
#endif

bail:
  if (!rc) {
    *ifaddrs_o = ifaddrs_mal;
    *num_ifs = num_if_mal;
  }
  if (nic_info != NULL)
    mal_free_nicinfo(nic_info);

  return rc;
}

MAL_FUNC(void)
mal_freeifaddrs(struct mal_ifaddrs *ifaddrs)
{
  mal_free(ifaddrs);
}

#if MAL_OS_WINDOWS
static uint64_t
read_vlanid_from_registry(const char *adapter_name)
{
  HKEY key_adapter_list;
  LONG result;
  DWORD disp;
  
  result = RegCreateKeyExA(HKEY_LOCAL_MACHINE,
    "System\\CurrentControlSet\\Control\\Class\\{4D36E972-E325-11CE-BFC1-08002BE10318}",
    0,
    NULL,
    REG_OPTION_NON_VOLATILE,
    KEY_READ,
    NULL,
    &key_adapter_list,
    &disp);
  if (result == ERROR_SUCCESS) {
    char adapter_index[128+1];
    DWORD dwIndex, dwBytes;
    
    result = RegEnumKeyA(key_adapter_list, 0, adapter_index,
      sizeof(adapter_index));
    for (dwIndex=1; result == ERROR_SUCCESS; ++dwIndex) {
      HKEY key_adapter;

      if (ERROR_SUCCESS == RegCreateKeyExA(key_adapter_list,
          adapter_index,
          0,
          NULL,
          REG_OPTION_NON_VOLATILE,
          KEY_READ,
          NULL,
          &key_adapter,
          &disp)) {
        
        char instanceid[256];

        dwBytes = sizeof(instanceid);
        memset(instanceid, 0, sizeof(instanceid));
        if (ERROR_SUCCESS == RegQueryValueExA(
            key_adapter,
            "NetCfgInstanceId",
            NULL,
            NULL,
            (LPBYTE)instanceid,
            &dwBytes)) {

          //printf("netcfginstanceid %s %s\n", instanceid, adapter_name);
          if (strcmp(instanceid, adapter_name) == 0) {
            DWORD vlanid, pcp, type;
            char vlanid_str[16];

            vlanid = 0;
            pcp = 0;

            if (ERROR_SUCCESS == RegQueryValueExA(
                key_adapter,
                "VlanId",
                NULL,
                &type,
                NULL,
                NULL)) {
              if (type != REG_DWORD && type != REG_SZ) {
                // Not expected
                RegCloseKey(key_adapter);
                RegCloseKey(key_adapter_list);
                return 0;
              }
            }
            else {
              // No VlanId
              RegCloseKey(key_adapter);
              RegCloseKey(key_adapter_list);
              return 0;
            }

            if (type == REG_DWORD) {
              dwBytes = sizeof(vlanid);
              if (ERROR_SUCCESS == RegQueryValueExA(
                  key_adapter,
                  "VlanId",
                  NULL,
                  NULL,
                  (LPBYTE)&vlanid,
                  &dwBytes)) {
                //printf("vlanid=%d\n", vlanid);
                //return vlanid;
              }
              else {
                // No VlanId
                RegCloseKey(key_adapter);
                RegCloseKey(key_adapter_list);
                return 0;
              }
            }
            else {
              dwBytes = sizeof(vlanid_str);
              if (ERROR_SUCCESS == RegQueryValueExA(
                  key_adapter,
                  "VlanId",
                  NULL,
                  NULL,
                  (LPBYTE)vlanid_str,
                  &dwBytes)) {
                vlanid = atoi(vlanid_str);
                //printf("vlanid=%s %d\n", vlanid_str, vlanid);
                //return vlanid;
                // The built-in team on Windows 2012 uses
                // VlanId=-1 to indicate no vlan id.
                if ((int)vlanid < 0) {
                  RegCloseKey(key_adapter);
                  RegCloseKey(key_adapter_list);
                  return 0;
                }
              }
              else {
                // No VlanId
                RegCloseKey(key_adapter);
                RegCloseKey(key_adapter_list);
                return 0;
              }
            }

            dwBytes = sizeof(pcp);
            if (ERROR_SUCCESS == RegQueryValueExA(
                key_adapter,
                "PCP",
                NULL,
                NULL,
                (LPBYTE)&pcp,
                &dwBytes)) {
              //printf("pcp=%d\n", pcp);
            }
            else {
              // No PCP
              pcp = 0;
            }

            vlanid = ((pcp << 13) | vlanid);
            RegCloseKey(key_adapter);
            RegCloseKey(key_adapter_list);
            return vlanid;
          }
        }

        RegCloseKey(key_adapter);
      }
      
      result = RegEnumKeyA(key_adapter_list, dwIndex, adapter_index,
        sizeof(adapter_index));
    }

    RegCloseKey(key_adapter_list);
  }

  return 0;
}

static uint16_t
get_vlan_tag_windows(const char *ipstr)
{
  PIP_ADAPTER_ADDRESSES addr, cur;
  PIP_ADAPTER_UNICAST_ADDRESS unicast_addr;
  ULONG buflen;
  DWORD err;
  uint16_t vlanid = 0;
  unsigned int ip;

  ip = inet_addr(ipstr);
  if (ip == INADDR_NONE) {
    /* Invalid IPv4 address? */
    /* XXX log, warning, or error message */
    return 0;
  }
  
  buflen = sizeof(IP_ADAPTER_ADDRESSES) * 64;
  addr = mal_malloc(buflen);
  if (addr == NULL) {
    /* Failed to malloc */
    /* XXX log, warning, or error message */
    return 0;
  }
  memset(addr, 0, buflen);
  err = GetAdaptersAddresses(AF_INET, 0, NULL, addr, &buflen);
  if (err != NO_ERROR) {
    if (err == ERROR_BUFFER_OVERFLOW) {
      /* Try again */
      mal_free(addr);
      addr = mal_malloc(buflen);
      if (addr == NULL) {
        /* Failed to malloc */
        /* XXX log, warning, or error message */
        return 0;
      }
      memset(addr, 0, buflen);
      if (ERROR_SUCCESS != GetAdaptersAddresses(AF_INET, 0, NULL, addr,
            &buflen)) {
        /* Failed twice, give up */
        /* XXX log, warning, or error message */
        goto exit;
      }
    }
    else {
      /* Unexpected error */
      /* XXX log, warning, or error message */
      goto exit;
    }
  }

  if (buflen < sizeof(IP_ADAPTER_ADDRESSES)) {
    /* Unexpected error */
    /* XXX log, warning, or error message */
    goto exit;
  }

  // Find the adapter whose primary IP address is
  // the IP we are looking for
  cur = addr;
  while (cur) {
    if (cur->IfType == IF_TYPE_ETHERNET_CSMACD &&
      cur->FirstUnicastAddress != NULL) {
      unicast_addr = cur->FirstUnicastAddress;
      while (unicast_addr) {
        struct sockaddr_in *in;
        unsigned int ifip;
        
        in = (struct sockaddr_in *)
          unicast_addr->Address.lpSockaddr;
        //printf("FirstUnicastAddress=%x\n", in->sin_addr.S_un.S_addr);
        
        ifip = in->sin_addr.S_un.S_addr;
        if (ifip == ip) {
          // AdapterName is the NetCfgInstanceId
          // Search the registry and get the VLAN ID if it exists
          vlanid = (uint16_t)read_vlanid_from_registry(cur->AdapterName);
          goto exit;
        }

        unicast_addr = unicast_addr->Next;
      }
    }

    cur = cur->Next;
  }

exit:
  if (addr)
    mal_free(addr);
  return vlanid;
}
#endif /* MAL_OS_WINDOWS */

MAL_FUNC(uint16_t)
mal_get_vlan_tag(const char *ifname)
{
  uint16_t tag = 0;
  const uint16_t tag_prio = 0xe000;

#if MAL_OS_LINUX || MAL_OS_FREEBSD
  char *p;
  /* first check for the vlan names like vlan0001 and vlan2 */
  if (sscanf(ifname, "vlan%hu", &tag) == 1 && tag < 4096) {
    return htons(tag_prio | tag);
  } 
  /* next check for the ethX.Y format */ 
  p = strchr(ifname, '.');
  if (p == NULL) return 0;
  ++p;
  if (*p < '0' || *p > '9') return 0;
  tag = atoi(p);
  if (tag < 4096) {
    return htons(tag_prio | tag);
  }
#elif MAL_OS_MACOSX
  if (sscanf(ifname, "vlan%hu", &tag) == 1 && tag < 4096)
    return htons(tag_prio | tag);
#elif MAL_OS_SOLARIS
  if (sscanf(ifname, "myri%hu", &tag) == 1 && tag >= 10000)
    return htons(tag_prio | tag/1000);
#elif MAL_OS_WINDOWS
  return htons(get_vlan_tag_windows(ifname));
#else
#warning update get_vlan_tag for OS
#endif

  return 0;
}

static
mal_wcsum
wcsum_copy_odd(uint8_t *dst, const uint8_t *src, int len, mal_wcsum wcsum)
{
  mal_assert(len > 0);
  mal_assert(len < sizeof(mal_wcsum));
  mal_assert(((uintptr_t) dst & (sizeof(uintptr_t)-1)) == 
            ((uintptr_t) src & (sizeof(uintptr_t)-1)));

  if (1 & (uintptr_t) src) {
    *dst = *src;
    wcsum += (*src << 8);
    dst++;
    src++;
    len--;
  }
  if (len >= 2) { /* at least 2 bytes */
    if (2 & (uintptr_t) src) {
      wcsum += *(uint16_t *)src;
      *(uint16_t *)dst = *(uint16_t *)src;
      dst += 2;
      src += 2;
      len -= 2;
    }
    if (len >= 4) { /* at least 4 bytes */
        mal_assert(((uintptr_t) src & 0x3) == 0);
        wcsum += *(uint32_t *)src;
        *(uint32_t *)dst = *(uint32_t *) src;
        dst += 4;
        src += 4;
        len -= 4;
    }
    if (len & 2) {
      wcsum += *(uint16_t *) src;
      *(uint16_t *)dst = *(uint16_t *)src;
      dst += 2;
      src += 2;
      len -= 2;
    }
  }
  if (len) {
    wcsum += *src;
    *dst = *src;
  }

  return wcsum;
}

uint32_t
mal_piocopy_csum(uintptr_t *piobuf, const uint8_t *src, uint32_t len, uint32_t *offp)
{
  mal_wcsum wcsum = 0;
  uint32_t csum;
  uintptr_t tmp = 0, *bufp = (uintptr_t *) src;
  uint32_t ubytes = 0;
  uint32_t off = (uint32_t)((uintptr_t) src & (MAL_PIOWORDSIZE-1));

  mal_assert(len > 0 && (((uintptr_t) piobuf & (MAL_PIOWORDSIZE-1)) == 0));

  if (unlikely(off)) {
    ubytes = MAL_PIOWORDSIZE - off;
    if (len < ubytes)
      ubytes = len;
    wcsum = wcsum_copy_odd((uint8_t *)&tmp + off, src, ubytes, wcsum);
    bufp = (uintptr_t *)(src + ubytes);
    len -= ubytes;
    *piobuf++ = tmp;
  }

  while (len > (MAL_PIOWORDSIZE-1)) {
    tmp = *bufp++;
#if MAL_PIOWORDSIZE == 8
    add_carry64(wcsum, tmp);
#else
    add_carry32(wcsum, tmp);
#endif
    *piobuf++ = tmp;
    len -= MAL_PIOWORDSIZE;
  }

  if (unlikely(len)) {
    uint8_t *pdst = (uint8_t *)&tmp;
    uint8_t *psrc = (uint8_t *)bufp;
    if (len & 4) {
#if MAL_PIOWORDSIZE == 8
      add_carry64(wcsum, *(uint32_t *)psrc);
#else
      add_carry32(wcsum, *(uint32_t *)psrc);
#endif
      *(uint32_t *)pdst = *(uint32_t *)psrc;
      pdst += 4; psrc += 4;
    }
    if (len & 2) {
#if MAL_PIOWORDSIZE == 8
      add_carry64(wcsum, *(uint16_t *)psrc);
#else
      add_carry32(wcsum, *(uint16_t *)psrc);
#endif
      *(uint16_t *)pdst = *(uint16_t *)psrc;
      pdst += 2; psrc += 2;
    }
    if (len & 1) {
#if MAL_PIOWORDSIZE == 8
      add_carry64(wcsum, *(uint8_t *)psrc);
#else
      add_carry32(wcsum, *(uint8_t *)psrc);
#endif
      *(uint8_t *)pdst = *(uint8_t *)psrc;
    }
    *piobuf = tmp;
  }

  MAL_STBAR();

#if MAL_PIOWORDSIZE == 8
  csum = do_add_carry32((wcsum>>32), wcsum & 0xffffffff);
#else
  csum = wcsum;
#endif

  /* When 1-byte mis-aligned, the 16-bit checksum needs to be reversed */
  if (unlikely(off & 1)) {
    csum = (csum >> 16) + (csum & 0xffff);
    csum += (csum >> 16);
    csum = ((csum >> 8) & 0xff) | ((csum & 0xff) << 8);
  }

  *offp = off;
  return csum;
}

/*
 * Check license validity and print a message if not properly
 * licensed
 */
int
mal_check_license(mal_handle_t handle, int boardnum, char *product)
{
  uint32_t license_valid;
  static char license_once[128];
  int rc;

  /* query liceense validity */
  license_valid = boardnum;
  rc = mal_ioctl(handle, MYRI_CHECK_LICENSE,
                 &license_valid, sizeof(license_valid));
  if (rc != 0) goto out;

  /* Print a useful message about the license if not valid */
  if (license_valid != MYRI_KEYCHECK_SUCCESS) {
    if (license_once[boardnum] == 0) {
      mx_get_eeprom_string_t ee;
      char serial[MYRI_MAX_SERIAL_LENGTH+1];

      /* do not print again for this board */
      license_once[boardnum] = 1;

      /* get the serial number if we can */
      ee.board_number = boardnum;
      ee.buffer = (uintptr_t)serial;
      rc = mal_ioctl(handle, MX_GET_SERIAL_STRING, &ee, sizeof(ee));
      if (rc) {
	strcpy(serial, "<unknown>");
      }
      if (license_valid == MYRI_KEYCHECK_FAIL_TIME_BUILD) {
        fprintf(stderr, "Please contact your software provider or visit their website for "
                "an updated %s product build: %s\n", product, 
		myri_keycheck_str[license_valid]);
      } else if (license_valid < MYRI_KEYCHECK_MAXVAL) {
        fprintf(stderr, "License check failed on board %d, sn=%s: %s\n",
		boardnum, serial, myri_keycheck_str[license_valid]);
        if (license_valid != MYRI_KEYCHECK_UNSUPPORTED_HW) {
	  fprintf(stderr, "Please contact your software provider to obtain an appropriate %s key\n", product);
        } else {
	  fprintf(stderr, "%s does not support this adapter, please contact your software provider\n", product);
        }
        if (!MCP_ENFORCE_LICENSE) {
	  fprintf(stderr, "  (warning only)\n");
        }
      }
    }

    /* return "Permission denied" if enforcing */
    rc = MCP_ENFORCE_LICENSE ? EACCES : 0;

  }

 out:
  return rc;
}

MAL_FUNC(int)
mal_ifaddr_to_macaddr(uint32_t ipaddr_be32, uint8_t *mac)
{
  int rc;
  struct mal_ifaddrs *mal_ifs = NULL, *mif;
  int num_if_mal = 0;


  rc = mal_getifaddrs(&mal_ifs, 1, &num_if_mal);
  if (rc != 0) {
    return rc;
  }

  mif = mal_ifs;
  while (mif) {
    if (mif->mal_ifa_ipaddr == ipaddr_be32) {
      break;
    }
    mif = mif->mal_ifa_next;
  }
  if (mif == NULL) {
#ifndef _WIN32
      rc = ENODEV;
#else
      rc = ERROR_OPEN_FAILED;
#endif
    goto bail;
  }

  /* copy out the mac */
  memcpy(mac, mif->mal_ifa_macaddr, 6);

bail:
  if (mal_ifs != NULL) mal_freeifaddrs(mal_ifs);
  return rc;

}
