/*************************************************************************
 * The contents of this file are subject to the MYRICOM ABSTRACTION      *
 * LAYER (MAL) SOFTWARE AND DOCUMENTATION LICENSE (the "License").       *
 * User may not use this file except in compliance with the License.     *
 * The full text of the License can found in LICENSE.TXT                 *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2009 by Myricom, Inc.  All rights reserved.                 *
 *************************************************************************/
#ifndef _myri_clksync_h
#define _myri_clksync_h

#include "mal_stbar.h"

/* Accumulate ticks into nanoseconds.
 * There's no function to initialize the myri_clksync_nticks_t state, its up 
 * to the user to initialize nsecs to either be representative of system host 
 * time or just 0 if the nsecs are to be monotonic from 0.
 */
typedef struct {
  uint64_t  ticks;
  uint64_t  ticks_mask;
  uint64_t  nsecs;
  uint32_t  nt_numerator;
  uint32_t  nt_denominator;
} myri_clksync_nticks_t;

/*
 * Converting between two clocks in nanoseconds, in this case from a NIC clock
 * to the host clock.  The expected usage is for a background process to
 * periodically update the clockparams.
 */
#define MYRI_CLKSYNC_MULTIPLIER_SHIFT 30
typedef struct {
  uint64_t nsecs_base;
  int64_t  hostnic_drift_nsecs;
  int64_t  multiplier;
  int64_t  error_nsecs;
} myri_clksync_params_t;

typedef struct {
  uint32_t	        seqnum;
  myri_clksync_params_t clkp;
  myri_clksync_nticks_t nticks;
} myri_clksync_seq_t;

/* myri_timesync_ticks2nsecs is different from the 'clksync' functions
   later in this file in that when operating in timesync mode, the information 
   in the myri_clksync_nticks_t in the kernel window can change.  In clksync
   mode, it didn't change, which is why (for now) it gets its own parallel-ish
   implementation, even though it is in practice, it is very similar to other
   functions already implemented later in this file */

static inline uint64_t
myri_timesync_ticks2nsecs(volatile const myri_clksync_seq_t *clksync, 
			  uint64_t ticks) 
{
  uint64_t nsecs;
  uint64_t delta_ticks;
  uint32_t seqnum;
  volatile const myri_clksync_nticks_t *ntss;

  do {
  try_again:
    seqnum = clksync->seqnum;
    MAL_READBAR();

    if (unlikely(seqnum & 1)) { /* back off, hit middle of update */
      MAL_CPU_RELAX();
      goto try_again;
    }
    ntss = &clksync->nticks;
    delta_ticks = (ticks - ntss->ticks);
    nsecs = (((delta_ticks * ntss->nt_numerator) / ntss->nt_denominator) 
	     + ntss->nsecs);
    MAL_READBAR();
  } while (seqnum != clksync->seqnum);
  
  return nsecs;
}

/* Utility to convert/transform NIC-level nanoseconds to host-level nanoseconds
 * and account for drift
 */
static inline uint64_t
__myri_clksync_convert_nsecs(volatile const myri_clksync_params_t *clkp, uint64_t nsecs_in)
{
  uint64_t nsecs = nsecs_in + clkp->hostnic_drift_nsecs;
  nsecs += ((int64_t)(nsecs_in - clkp->nsecs_base)
               * clkp->multiplier) >> MYRI_CLKSYNC_MULTIPLIER_SHIFT;
  return nsecs;
}

/*
 * Convert nic nsec to host nsec.  Makes the approximation that for
 * very small c,  1/(1+c) == (1-c)
 */
static inline uint64_t
__myri_clksync_convert_host_nsecs(volatile const myri_clksync_params_t *clkp, uint64_t host_nsecs)
{
  uint64_t nic_nsecs;

  nic_nsecs = host_nsecs - clkp->hostnic_drift_nsecs;

  nic_nsecs += ((int64_t)(clkp->nsecs_base + clkp->hostnic_drift_nsecs -
			  host_nsecs)
               * clkp->multiplier) >> MYRI_CLKSYNC_MULTIPLIER_SHIFT;

  return nic_nsecs;
}

static inline uint64_t
myri_timesync_nsecs2ticks(volatile const myri_clksync_seq_t *clksync, 
			  uint64_t nsecs) 
{
  uint64_t ticks;
  uint64_t delta_nsecs;
  uint32_t seqnum;
  uint64_t nic_nsecs;
  volatile const myri_clksync_nticks_t *ntss;
  volatile const myri_clksync_params_t *params;
  
  do {
  try_again:
    seqnum = clksync->seqnum;
    MAL_READBAR();
    
    if (unlikely(seqnum & 1)) { /* back off, hit middle of update */
      MAL_CPU_RELAX();
      goto try_again;
    }
    ntss = &clksync->nticks;
    params = &clksync->clkp;
    
    /* get delta nsecs according to host */
    nic_nsecs = __myri_clksync_convert_host_nsecs(params, nsecs);
    delta_nsecs = (nic_nsecs - ntss->nsecs);

    /* convert this to NIC ticks */
    ticks = (((delta_nsecs * ntss->nt_denominator) / ntss->nt_numerator) 
	     + ntss->ticks);
    MAL_READBAR();
  } while (seqnum != clksync->seqnum);
  
  return ticks;
}

static inline uint64_t
myri_clksync_nticks_update(myri_clksync_nticks_t *ntss, uint64_t ticks)
{
  uint64_t delta_ticks;
  uint64_t nsecs;

  delta_ticks = (ticks - ntss->ticks) & ntss->ticks_mask;
  nsecs = (((delta_ticks * ntss->nt_numerator) / ntss->nt_denominator) + ntss->nsecs);
  /*Since we are no longer doing a multiply and shift above, we don't need to 
    update these two fields.  In fact, doing so causes drift as the division
    above produces an inexact answer.
    ntss->ticks = ticks;
    ntss->nsecs = nsecs;*/
  return nsecs;
}

/* Reader needs to ensure that 'ntss' parameters will not be updated while
 * this function is being used */
static inline uint64_t
__myri_clksync_nticks_to_nsecs(volatile const myri_clksync_nticks_t *ntss,
                               uint64_t ticks)
{
  uint64_t delta_ticks;
  uint64_t nsecs;
  
  delta_ticks = (ticks - ntss->ticks) & ntss->ticks_mask;
  if (delta_ticks > ntss->ticks_mask/2) {
    delta_ticks = (ntss->ticks - ticks) & ntss->ticks_mask;
    nsecs = ntss->nsecs - ((delta_ticks * ntss->nt_numerator) / ntss->nt_denominator);
  }
  else
    nsecs = ntss->nsecs + ((delta_ticks * ntss->nt_numerator) / ntss->nt_denominator);
  return nsecs;
}

/* Given NIC-level nanoseconds, ensures correct conversion to host-level
 * nanosecs -- the seq locks ensure that reads are consitent while parameters
 * may be updating */
static inline uint64_t
myri_clksync_convert_nsecs(volatile const myri_clksync_seq_t *clksync, uint64_t nsecs_in)
{
  uint64_t nsecs;
  uint32_t seqnum;

  do {
try_again:
    seqnum = clksync->seqnum;
    MAL_READBAR();

    if (unlikely(seqnum & 1)) { /* back off, hit middle of update */
      MAL_CPU_RELAX();
      goto try_again;
    }
    nsecs = __myri_clksync_convert_nsecs(&clksync->clkp, nsecs_in);
    MAL_READBAR();
  } while (seqnum != clksync->seqnum);

  return nsecs;
}

/* Same as myri_clksync_convert_nsecs but accepts ticks instead of nanosecs.
 * The function makes the extra step of converting to nic nanosecs before
 * converting to host-level nanosecs */
static inline uint64_t
myri_clksync_convert_nticks_to_nsecs(volatile const myri_clksync_seq_t *clksync, uint64_t nticks)
{
  uint64_t nsecs;
  uint32_t seqnum;

  do {
try_again:
    seqnum = clksync->seqnum;
    MAL_READBAR();

    if (unlikely(seqnum & 1)) { /* back off, hit middle of update */
      MAL_CPU_RELAX();
      goto try_again;
    }
    nsecs = __myri_clksync_nticks_to_nsecs(&clksync->nticks, nticks);
    nsecs = __myri_clksync_convert_nsecs(&clksync->clkp, nsecs);
    MAL_READBAR();
  } while (seqnum != clksync->seqnum);

  return nsecs;
}

static inline void
myri_clksync_copy_nticks_params(volatile const myri_clksync_seq_t *clksync,
                                myri_clksync_nticks_t *ntss)
{
  uint32_t seqnum;

  do {
try_again:
    seqnum = clksync->seqnum;
    MAL_READBAR();

    if (unlikely(seqnum & 1)) { /* back off, hit middle of update */
      MAL_CPU_RELAX();
      goto try_again;
    }

#if MAL_OS_MACOSX
    memcpy(ntss, (const void *)&clksync->nticks, sizeof (*ntss));
#else
    *ntss = clksync->nticks;
#endif

    MAL_READBAR();
  } while (seqnum != clksync->seqnum);

}

#endif /* _myri_clksync_h */
