/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2014 - 2016 by CSP, Inc.  All rights reserved.              *
 *************************************************************************/

#ifndef _phx_hw_h_
#define _phx_hw_h_

/* The Phoenix controller uses little endian.  Every multi-byte integer
 * field is in little endian.
 */

typedef struct {
  uint32_t low;
  uint32_t high;
} phx_dma_addr_t;

typedef phx_dma_addr_t nic_dma_addr_t;

/* Interrupt Descriptor and flag values */
typedef struct {
  uint8_t endpt_bitmap[8];  /*  8 x 8 = 64 bits for 64 endpoints */
  uint32_t flags;
  uint8_t pad[3];
  uint8_t seqnum;      /* 1..255 looping */
} phx_intr_t;

typedef phx_intr_t nic_intr_t;

#define PHX_INTR_LINK_UP	(1 << 0)
#define PHX_INTR_LINK_DOWN	(1 << 1)
#define PHX_INTR_ENDPT_CLOSED	(1 << 2)
#define PHX_INTR_ENDPT_WAKE_RX	(1 << 3)
#define PHX_INTR_ENDPT_WAKE_TX	(1 << 4)
#define PHX_INTR_XCVR_IN	(1 << 5)
#define PHX_INTR_XCVR_OUT	(1 << 6)
#define PHX_INTR_DMABENCH	(1 << 7)

/* Registers */

/* APU (Address Prefetch Unit) supports 2-level page table.
 * 1st level:
 * Alpha pages.  These contain phx_apu_flp_t's.  This structure points to
 * 2nd level pages (phx_apu_flp_t.beta).
 *
 * 2nd level:
 * Beta pages.  These contain phx_dma_addr_t's.  These addresses point to
 * 3rd level pages.  Each beta page is completely filled (i.e. 512 addresses).
 *
 * 3rd level:
 * Gamma pages.  These pages contain packet data or descriptors.
 */

#define PHX_APU_GAMMA_PAGES_PER_BETA   (PAGE_SIZE / sizeof(phx_dma_addr_t))
#define PHX_APU_BETA_PAGES_PER_FLP     6
#define PHX_APU_FLPS_PER_ALPHA         (PAGE_SIZE / sizeof(phx_apu_flp_t))

/* APU fist-level (a.k.a. alpha) descriptor.  Its size is 64B, and it is 64B
 * aligned.
 */
typedef struct {
  uint64_t opcode;       /* Bits 0-3 encode the number of FLP addresses */
  phx_dma_addr_t next;    /* DMA address to the next apu_flp */
  phx_dma_addr_t beta[PHX_APU_BETA_PAGES_PER_FLP];  /* 0-6 beta addresses */
} phx_apu_flp_t;

#define PHX_APU_FLP_OPCODE_STOP    (1 << 63)  /* Not set: continue */
#define PHX_APU_FLP_OPCODE_GET_NUM_BETA(op) (op & 0xf)
#define PHX_APU_FLP_OPCODE_SET_NUM_BETA(op, n) (op |= n)

/* Beta pages must be completely filled.  So, the receive ring must be
 * a multiple of 2MB (4KB x 4KB / 8B).
 */
#define PHX_APU_RING_SIZE_MULTIPLE     (1 << 21)

/* 512B receive descriptor */
#define PHX_RX_DESC_PKT_MAX 63
#define MCP_DBL_BATCH_CNT PHX_RX_DESC_PKT_MAX
#define PHX_RX_DESC_SEQNUM_MAX 255

#define PHX_DESC_FLAG_EP_LB_NORMAL     (1 << 0)
#define PHX_DESC_FLAG_EP_LB_TIMESTAMP  (1 << 1)
#define PHX_DESC_FLAG_MAC_ERROR        (1 << 2)

typedef struct {
  struct phx_rx_desc_pkt {
    uint16_t length;
    uint16_t tag;
    uint32_t nano_timestamp;
  } pkts[PHX_RX_DESC_PKT_MAX - 1];
  uint16_t pkt1_length;
  uint16_t pkt1_tag;
  uint32_t timestamp_high; /* seconds */
  uint32_t timestamp_low;  /* sub-second value in nanoseconds */
  uint16_t desc_flags;
  uint8_t pkt_cnt;
  uint8_t seqnum;
} phx_rx_desc_t;

#define PHX_TX_DMA_PAGES_PER_ENDPOINT  32    /* 128KB FIXME depends on PHX_TX_THRESHOLD value */

#define PHX_TX_DESC_HEADER_LEN  56
/* Transmit descriptor, used for both PIO and DMA.  FIXME.  Temporary */
typedef struct {
  uint32_t data_offset;
  uint16_t data_length;
  uint8_t header[PHX_TX_DESC_HEADER_LEN];
  uint8_t header_length;
  uint8_t flags;
} phx_tx_desc_t;

/* If the DMA flag is off, then it signifies PIO */
#define PHX_TX_DESC_FLAG_DMA                 (1 << 0)
#define PHX_TX_DESC_FLAG_LOOPBACK_NORMAL     (1 << 1)
#define PHX_TX_DESC_FLAG_LOOPBACK_TIMESTAMP  (1 << 2)
#define PHX_TX_DESC_FLAG_SEND_PORT0          (1 << 3)
#define PHX_TX_DESC_FLAG_SEND_PORT1          (1 << 4)
#define PHX_TX_DESC_FLAG_SEND_PORT2          (1 << 5)
#define PHX_TX_DESC_FLAG_SEND_PORT3          (1 << 6)

#define PHX_PHYS_PORT_NUM_TO_FLAG(num) ((1<<3)<<(num))

/* Send completion. */
typedef struct {
  /* FIXME.  Alignment may change. */
  struct {
    uint32_t sent;
    uint32_t unused;
  } endpoint[64];
} phx_tx_completion_t;

/*** Filters ***/
enum phx_dbl_filter_mode {
  PHX_DBL_FILTER_NORMAL = 0, /* default DBL mode */
  PHX_DBL_FILTER_ETHER_MULTI = 1, /* all ethernet multicast packets to one endpoint */
  PHX_DBL_FILTER_ALL = 2, /* all packets to one endpoint */
  PHX_DBL_FILTER_NONE = 255, /* reserved, not exposed to API */
};

#define PHX_RX_FILTER_CNT    2048  /* FIXME */
#define PHX_MAC_FILTER_CNT    128

#define PHX_BBSYSCTL_BASE                 0x0
#define PHX_BBSYSCTL_R0                   (PHX_BBSYSCTL_BASE + 0x000)
#define PHX_BBSYSCTL_R4                   (PHX_BBSYSCTL_BASE + 0x004)
#define PHX_BBSYSCTL_R8                   (PHX_BBSYSCTL_BASE + 0x008)
#define PHX_BBSYSCTL_RC                   (PHX_BBSYSCTL_BASE + 0x00C)
#define PHX_BBSYSCTL_CONST0               (PHX_BBSYSCTL_BASE + 0x010)
#define PHX_BBSYSCTL_CONST1               (PHX_BBSYSCTL_BASE + 0x014)
#define PHX_BBSYSCTL_CP_RD_DW             (PHX_BBSYSCTL_BASE + 0x0B0)
#define PHX_BBSYSCTL_CP_RD_HEADER         (PHX_BBSYSCTL_BASE + 0x0B4)
#define PHX_BBSYSCTL_CP_WR_DW             (PHX_BBSYSCTL_BASE + 0x0B8)
#define PHX_BBSYSCTL_CP_WR_HEADER         (PHX_BBSYSCTL_BASE + 0x0BC)

/* RX/TX TIME offsets to BB BASE */
#define PHX_TIME_STATUS                   (PHX_BBSYSCTL_BASE + 0x030)
#define PHX_TIME_CONTROL                  (PHX_BBSYSCTL_BASE + 0x034)
#define PHX_TIME_LS                       (PHX_BBSYSCTL_BASE + 0x038)
#define PHX_TIME_MS                       (PHX_BBSYSCTL_BASE + 0x03C)
#define PHX_TIME_COMPARE_LS               (PHX_BBSYSCTL_BASE + 0x040)
#define PHX_TIME_COMPARE_MS               (PHX_BBSYSCTL_BASE + 0x044)
#define PHX_TIME_REFPERPPS                (PHX_BBSYSCTL_BASE + 0x048)

/* K35 stores time status in a different location */
#define K35_PPS_STATUS                    (PHX_BBSYSCTL_BASE + 0x034)
#define K35_10MHZ_CONTROL                 (PHX_BBSYSCTL_BASE + 0x088)
#define K35_PANEL_PPS                     (PHX_BBSYSCTL_BASE + 0x10905c)
#define K35_EDGE_PPS                      (PHX_BBSYSCTL_BASE + 0x109060)
#define K35_10MHZ_STATUS                  (PHX_BBSYSCTL_BASE + 0x109058)

/* These are the 'old' time base specific registers */ 
#define PHX_TIME_V1_TRIM_PPB              (PHX_BBSYSCTL_BASE + 0x04C)
#define PHX_TIME_V1_SKIP_NS               (PHX_BBSYSCTL_BASE + 0x050)

/* These are the 'new' time base specific registers */
/* Note that the new time servo reuses offset 0x04c and 0x050*/
#define PHX_TIME_V2_SERVO_GAINS           (PHX_BBSYSCTL_BASE + 0x04C)
#define PHX_TIME_V2_HIT_DELTA_PHASE_LS    (PHX_BBSYSCTL_BASE + 0x050)
#define PHX_TIME_V2_HIT_DELTA_PHASE_MS    (PHX_BBSYSCTL_BASE + 0x054)
#define PHX_TIME_V2_SET_CFREQ_LS          (PHX_BBSYSCTL_BASE + 0x058)
#define PHX_TIME_V2_SET_CFREQ_MS          (PHX_BBSYSCTL_BASE + 0x05C)
#define PHX_TIME_V2_GET_FILT_ERROR_LS     (PHX_BBSYSCTL_BASE + 0x060)
#define PHX_TIME_V2_GET_FILT_ERROR_MS     (PHX_BBSYSCTL_BASE + 0x064)

/* the following values are scaled up so we can do the math
   with integers without losing too much precision */
#define PHX_TIME_V2_INC_NOMINAL_scaled_up 1844674407370955161LL
#define PHX_TIME_V2_INC_PER_PPB_scaled_up 1844674407
#define PHX_TIME_V2_INC_scale_factor      10000000 
#define PHX_TIME_V2_NS_MAX_ADJUSTMENT     7629
#define PHX_TIME_V2_NS_ONE                18446744074

#define PHX_GPIO_DIR	                  (PHX_BBSYSCTL_BASE + 0x080)
#define PHX_GPIO_OUT	                  (PHX_BBSYSCTL_BASE + 0x088)
#define PHX_GPIO_IN	                  (PHX_BBSYSCTL_BASE + 0x090)
#define PHX_TIME_TX_FNS_ADJUSTMENT        (PHX_BBSYSCTL_BASE + 0x408)
#define PHX_TIME_TX_NS_ADJUSTMENT         (PHX_BBSYSCTL_BASE + 0x410)
#define PHX_TIME_RX_FNS_ADJUSTMENT        (PHX_BBSYSCTL_BASE + 0x488)
#define PHX_TIME_RX_NS_ADJUSTMENT         (PHX_BBSYSCTL_BASE + 0x490)

/* the lower part of the status is a rolling count of pps pulses */
#define PHX_TIME_STATUS_MASK              0xFFFFFF00
#define PHX_TIME_STATUS_PpsLostSticky     0x80000000
#define PHX_TIME_STATUS_GpsInSticky       0x40000000
#define PHX_TIME_STATUS_PpsInSticky       0x20000000
#define PHX_TIME_STATUS_TimeSetSticky     0x10000000
#define PHX_TIME_STATUS_PpsOK             0x08000000
#define PHX_TIME_STATUS_PpsLost           0x04000000

#define PHX_TIME_CONTROL_DrivePpsOut      0x0003
#define K35_10MHZ_MASK                    0x0002
#define PHX_TIME_CONTROL_PpsInDisable     0x0004
#define PHX_TIME_CONTROL_GpsDisable       0x0008
#define PHX_TIME_CONTROL_ServoDisable     0x0010
#define PHX_TIME_CONTROL_ProportDisable   0x0020
#define PHX_TIME_CONTROL_InvertPps        0x0040
#define PHX_TIME_CONTROL_JamPhasePps      0x0080
#define PHX_TIME_CONTROL_IntegralDisable  0x0100
#define PHX_TIME_CONTROL_ClearSticky      0x80000000

#define PHX_TIME_GPIO_OUT_10MHz_Enable        0x8000
#define PHX_TIME_GPIO_IN_10MHz_PLL           0x10000

/* APU: 0x02_0000 - 0x02_ffff */
#define PHX_APU_BASE                       0x20000
#define PHX_APU_ID_VERSION                 0x000
#define PHX_APU_NUM_CHANNEL                0x004 /* 2B */
#define PHX_APU_ENDPOINTS                  0x006 /* 1B */
#define PHX_APU_PORTS                      0x007 /* 1B */
#define PHX_APU_ALPHA_FETCHES              0x008
#define PHX_APU_BETA_FETCHES               0x00C
#define PHX_APU_GAMMA_FETCHES              0x010
#define PHX_APU_LAST_GAMMA_CH              0x014 /* 1B */
#define PHX_APU_LOOP_CH                    0x015 /* 1B */
#define PHX_APU_ALPHA_NEXT_LS              0x020
#define PHX_APU_ALPHA_NEXT_MS              0x024
#define PHX_APU_ALPHA_COUNT                0x028
#define PHX_APU_ALPHA_TAKE                 0x02C
#define PHX_APU_BETA_NEXT_LS               0x040
#define PHX_APU_BETA_NEXT_MS               0x044
#define PHX_APU_BETA_COUNT                 0x048
#define PHX_APU_GAMMA_NEXT_LS              0x060
#define PHX_APU_GAMMA_NEXT_MS              0x064
#define PHX_APU_GAMMA_COUNT                0x068
#define PHX_APU_GAMMA_TAKE                 0x06C
#define PHX_APU_CHANNEL_RESET              0x07C
/* APU has 256 channels.
 *
 * Channel = (data_stream ? 0 : 1) + (2 * endpoint) + (2 * EP * port)
 * EP = endpoints per port, round up to the nearest power of 2.
 *
 * Use phx_apu_channel() in phx_common.h to compute channel IDs.  Each
 * channel uses one 256B block, including all of the above registers.
 */
#define PHX_APU_CHANNEL_SIZE               0x100

struct phx_apu_config {
  uint32_t id_version;
  uint16_t num_channel;
  uint8_t endpoints;
  uint8_t ports;
  uint32_t channels_per_port;
};

/* NCPP
 * The base address is same as Sniffer RX.  The board has either NCPP
 * or Sniffer.  Normally, it is Sniffer.  So, do not use these NCPP
 * registers.
 */
#if 0
#define PHX_NCPP_BASE                     0x5000
#define PHX_NCPP_R0                        0x000
#define PHX_NCPP_CONST0                    0x00C
#define PHX_NCPP_MACADDR_3_0               0x050
#define PHX_NCPP_MACADDR_5_4               0x054
#define PHX_NCPP_RX_FILTER_SELECT          0x058
#endif /* 0 */

/* DblRx: 0x03_0000 - 0x03_ffff */
#define PHX_DBL_RX_BASE                    0x30000
#define PHX_DBL_RX_R0                      0x000
#define PHX_DBL_RX_R4                      0x004
#define PHX_DBL_RX_CYCLE_COUNT             0x008
#define PHX_DBL_RX_CONST0                  0x00C /* 0xBACECACE */
#define PHX_DBL_RX_NUM_DW_PER_PUSH         0x010
#define PHX_DBL_RX_COMMAND                 0x014
#define PHX_DBL_RX_MAX_LENGTH_LIMIT        0x018
#define PHX_DBL_RX_BYTE_COUNT              0x020 /* 64 bit */
#define PHX_DBL_RX_PACKET_COUNT            0x028 /* 64 bit */
#define PHX_DBL_RX_FF_BYTE_COUNT           0x030
#define PHX_DBL_RX_FF_PACKET_COUNT         0x034
#define PHX_DBL_RX_FD_BYTE_COUNT           0x040
#define PHX_DBL_RX_FD_PACKET_COUNT         0x044
#define PHX_DBL_RX_PACKET_OVERFLOW         0x050
#define PHX_DBL_RX_DESC_OVERFLOW           0x054
#define PHX_DBL_RX_PACKET_COUNT_LIMIT      0x060
#define PHX_DBL_RX_LENGTH_LIMIT            0x064
#define PHX_DBL_RX_TIME_LIMIT              0x068
#define PHX_DBL_RX_BP_RESET                0x0A0 /* zeroes the six counters below */ 
#define PHX_DBL_RX_BP_EMPTY                0x0A4
#define PHX_DBL_RX_BP_Q1                   0x0A8
#define PHX_DBL_RX_BP_Q2                   0x0Ac
#define PHX_DBL_RX_BP_Q3                   0x0B0
#define PHX_DBL_RX_BP_Q4                   0x0B4
#define PHX_DBL_RX_BP_FULL                 0x0B8
/* The following registers are per endpoint. */
#define PHX_DBL_RX_DATA_RING_SIZE          0x100
#define PHX_DBL_RX_DATA_CONSR_ADVANCE_N    0x104
#define PHX_DBL_RX_DATA_CONSR_POSITION_N   0x108
#define PHX_DBL_RX_DATA_PRODW_POSITION_N   0x10C
#define PHX_DBL_RX_DESC_RING_SIZE          0x110
#define PHX_DBL_RX_DESC_CONSR_ADVANCE_N    0x114
#define PHX_DBL_RX_DESC_CONSR_POSITION_N   0x118
#define PHX_DBL_RX_DESC_PRODW_POSITION_N   0x11C
#define PHX_DBL_RX_BYTE_COUNT_N            0x120 /* 64 bit */
#define PHX_DBL_RX_PACKET_COUNT_N          0x128 /* 64 bit */
#define PHX_DBL_RX_ENDPOINT_COMMAND        0x130 /* 0x1=start, 0x2=stop 0x3=reset*/
#define PHX_DBL_RX_DESC_COUNT              0x134
#define PHX_DBL_RX_DESC_WAKE_REQUEST       0x138
#define PHX_DBL_RX_STOP_FLUSHED            0x13C

/* All of the registers above are one set.  There is one set for each
 * endpoint.  The start/base address for endpoint ep[n] is 
 * 0x30000 + (PHX_DBL_RX_ENDPOINT_SIZE x n).
 */
#define PHX_DBL_RX_ENDPOINT_SIZE           0x400

/* Packet alignment */
#define PHX_DBL_RX_PACKET_ALIGNMENT        16

#define PHX_ALIGN_DBL_RX_PACKET_LENGTH(len) \
  ((len + PHX_DBL_RX_PACKET_ALIGNMENT - 1) & ~(PHX_DBL_RX_PACKET_ALIGNMENT - 1))

/* Filters: 0x01_0000 - 0x01_ffff */
#define PHX_FLTR_BASE                     0x10000
#define PHX_FLTR_SOFT_RESET               0x0000
#define PHX_FLTR_ID_VERSION               0x0004
#define PHX_FLTR_MAX_LATENCY              0x0100
#define PHX_FLTR_BACK_PRESSURE            0x0104
#define PHX_FLTR_PORT_RESET               0x0200
#define PHX_FLTR_FILTER_FLAGS             0x0204
#define PHX_FLTR_KERNEL_ENDPOINT          0x0208
#define PHX_FLTR_INCOMING_FILTER          0x020C
#define PHX_FLTR_EP_DISABLE_LO            0x0210
#define PHX_FLTR_EP_DISABLE_HI            0x0214
#define PHX_FLTR_CLASS_D_DISCARD          0x0218
#define PHX_FLTR_PORT                     0x021C
#define PHX_FLTR_F_SRC_IP                 0x0220
#define PHX_FLTR_F_DST_IP                 0x0224
#define PHX_FLTR_F_SRC_PORT               0x0228
#define PHX_FLTR_F_DST_PORT               0x022A
#define PHX_FLTR_F_VLAN                   0x022C
#define PHX_FLTR_F_TAG                    0x022E
#define PHX_FLTR_F_PROTO                  0x0230
#define PHX_FLTR_F_ADD_RM                 0x0231
#define PHX_FLTR_F_ENDPOINT_TARGET        0x0232
#define PHX_FLTR_F_FLAG                   0x0233
#define PHX_FLTR_ADD_SLOTS                0x0400
#define PHX_FLTR_FREE_SLOTS               0x0404
#define PHX_FLTR_USED_SLOTS               0x0406
#define PHX_FLTR_PKTS_RECEIVED            0x0410
#define PHX_FLTR_BYTES_RECEIVED           0x0414
#define PHX_FLTR_PKTS_SENT                0x0418
#define PHX_FLTR_BYTES_SENT               0x041C
#define PHX_FLTR_PKTS_DROPPED             0x0420
#define PHX_FLTR_PKTS_ABORTED             0x0424
#define PHX_FLTR_INC_PACKETS_DROPPED      0x0428
#define PHX_FLTR_INC_BYTES_DROPPED        0x042C
#define PHX_FLTR_PRT_MAX_LATENCY          0x0430
#define PHX_FLTR_PRT_EGRESS_STALLED       0x0434
#define PHX_FLTR_PKTS_FILTERED            0x0438
#define PHX_FLTR_PKTS_UNFILTERED          0x043C
#define PHX_FLTR_PKTS_SKIPPED_FILTER      0x0440
#define PHX_FLTR_PKTS_DROP_FILTER         0x0444
#define PHX_FLTR_PKTS_SENT_WITHOUT_ERROR  0x0448
#define PHX_FLTR_PKTS_DISCARD             0x044C
#define PHX_FLTR_PKTS_CLASS_D             0x0450
#define PHX_FLTR_PKTS_DISCARD_CLASS_D     0x0454
#define PHX_FLTR_LOOKUP_CNT               0x0500
#define PHX_FLTR_LOOKUP_FAIL              0x0504
#define PHX_FLTR_LOOKUP_SEARCH            0x0508
#define PHX_FLTR_ADD_ERROR                0x050C
#define PHX_FLTR_REMOVE_ERROR             0x0510
#define PHX_FLTR_RP_PKTS_RECV             0x0560
#define PHX_FLTR_RP_PKTS_SENT             0x0564
#define PHX_FLTR_RP_BYTES_SENT            0x0568
#define PHX_FLTR_RP_PKTS_DROPPED          0x056C
#define PHX_FLTR_RP_SENT_DUP              0x0570
#define PHX_FLTR_RP_SENT_ABORT            0x0574
#define PHX_FLTR_RP_DROPPED_ABORT         0x0578
#define PHX_FLTR_RP_DROPPED_CONTROL       0x057C
#define PHX_FLTR_RP_DROPPED_META          0x0580
#define PHX_FLTR_RP_CONTROL_FULL          0x0584
#define PHX_FLTR_RP_RECV_ABORT            0x0588
#define PHX_FLTR_RP_META_ERROR            0x058C

#define PHX_FLTR_EP_PKTS_SENT             0x0800 
#define PHX_FLTR_EP_BYTES_SENT            0x0804
#define PHX_FLTR_STAT_PKTS_SENT(ep)  (0x800 + 8 * (ep))
#define PHX_FLTR_STAT_BYTES_SENT(ep) (0x804 + 8 * (ep))

#define PHX_FLTR_FLAG_BIT_KERNEL_DUP        0x80
#define PHX_FLTR_FLAG_BIT_IGNORE_VLAN       0x20
#define PHX_FLTR_FLAG_BIT_IGNORE_PROTO      0x10
#define PHX_FLTR_FLAG_BIT_IGNORE_SRC_IP     0x08
#define PHX_FLTR_FLAG_BIT_IGNORE_SRC_PORT   0x04
#define PHX_FLTR_FLAG_BIT_IGNORE_DST_IP     0x02
#define PHX_FLTR_FLAG_BIT_IGNORE_DST_PORT   0x01


/* DblTx: 0x04_0000 - 0x05_ffff */
#define PHX_REG_TX_BASE                   0x40000
#define PHX_REG_TX_R0                     0x0000
#define PHX_REG_TX_R4                     0x0004
#define PHX_REG_TX_NUM_DW_PER_PUSH        0x0010
#define PHX_REG_TX_AUTO_READ_REQUEST      0x0014
#define PHX_REG_TX_THRESHOLD              0x0020
#define PHX_REG_TX_WAKE_REQUEST           0x0024 /* Do we still have this?  What does it do? */
#define PHX_REG_TX_HOST_COMPLETION_LSB    0x0028
#define PHX_REG_TX_HOST_COMPLETION_MSB    0x002C

/* Per-endpoint registers.  These are offsets within each endpoint region. */
#define PHX_EP_REG_TX_ADDR_LOAD_LSB       0x0050
#define PHX_EP_REG_TX_ADDR_LOAD_MSB       0x0054
#define PHX_EP_REG_TX_ADDR_BUFFER_FINISH  0x0058
#define PHX_EP_REG_TX_ADDR_BUFFER_CLEAR   0x005C
#define PHX_EP_REG_TX_DESC_BUFFER_SIZE    0x0060
#define PHX_EP_REG_TX_COMPLETION_COUNT    0x0064
#define PHX_EP_REG_TX_STATUS              0x0070
#define PHX_EP_REG_TX_WAKE_REQUEST        0x0110
#define PHX_EP_REG_TX_DESCRIPTOR_DATA     0x0FC0 /* ... + 64B */

/* RxSelector merge core before the filter */
#define PHX_RX_SEL_BASE                   0x80000
#define PHX_RX_SEL_LOOPBACK_RESET         0x0000
#define PHX_RX_SEL_BYPASS_DISABLE         0x003c
#define PHX_RX_SEL_RESET                  0x0100
#define PHX_RX_SEL_PKTS_RX                0x0104
#define PHX_RX_SEL_PKTS_DROP              0x0108
#define PHX_RX_SEL_PKTS_ABT               0x010c
#define PHX_RX_SEL_PKTS_SENT              0x0110
#define PHX_RX_SEL_PKTS_NO_META           0x0114
#define PHX_RX_SEL_PKTS_META_DROP         0x0118
#define PHX_RX_SEL_PKTS_EXTRA_META        0x011c
#define PHX_RX_SEL_PKTS_CTRL_DROP         0x0120
#define PHX_RX_SEL_BYTES_SENT             0x0124
#define PHX_RX_SEL_PKTS_BYPASSED          0x0128
#define PHX_RX_SEL_BYTES_BYPASSED         0x012c

/* Interrupt control unit (ICU) */
#define PHX_ICU_BASE                      0x90000
#define PHX_ICU_ENDPOINTS                 0x0000
#define PHX_ICU_OPERATION_MODE            0x0004
#define PHX_ICU_VECTORS_ALLOC             0x0008
#define PHX_ICU_GLOBAL_MASK               0x000C
#define PHX_ICU_ENABLE_LO                 0x0010
#define PHX_ICU_ENABLE_MID                0x0014
#define PHX_ICU_ENABLE_HI                 0x0018
#define PHX_ICU_ENABLE_IR                 0x001C
#define PHX_ICU_DISABLE_IR                0x0020
#define PHX_ICU_INTR_INTERVAL             0x0024
#define PHX_ICU_LINK_STATUS               0x0040
#define PHX_ICU_REQ_LINK_IR               0x0044
/*#define PHX_ICU_REQ_LINK_IR_2           0x0048 mirror */
#define PHX_ICU_CLR_LINK_IR               0x004C
/*#define PHX_ICU_CLR_LINK_IR_2           0x0050 mirror */
#define PHX_ICU_EP_RESET                  0x0060
/* ICU debug operations */
#define PHX_ICU_CP_LINK_UP_IR             0x1000
#define PHX_ICU_CP_LINK_DN_IR             0x1004
#define PHX_ICU_CP_EP_WAKE_RX_IR          0x1010
#define PHX_ICU_CP_EP_WAKE_TX_IR          0x1014
#define PHX_ICU_LAST_IR                   0x1020
#define PHX_ICU_LAST_IR_DESC_W0           0x1030
#define PHX_ICU_LAST_IR_DESC_W1           0x1034
#define PHX_ICU_LAST_IR_DESC_W2           0x1038
#define PHX_ICU_LAST_IR_DESC_W3           0x103C
#define PHX_ICU_DISABLE_IR_DESC           0x1040
/* ICU End Point Status */
#define PHX_ICU_END_POINT_STATUS_OFFSET   0x2000
/* (1+endpoints) DWORDS of status */

/* Interrupt descriptor addresses */
#define PHX_ICU_DESC_ADDRESS_OFFSET       0x4000
/* (1+endpoints) many pairs of <LS, MS> */

/* stuff from the pcie spec */
#define PHX_ICU_VECTOR_TABLE_OFFSET          0xffc000
#define PHX_ICU_PENDING_BIT_ARRAY_OFFSET     0xffe000

/* DMA bench command support */
#define PHX_DMABENCH_BASE                 0xb0000
#define PHX_DMABENCH_R0                   (PHX_DMABENCH_BASE + 0x000)
#define PHX_DMABENCH_R4                   (PHX_DMABENCH_BASE + 0x004)
#define PHX_DMABENCH_CONST0               (PHX_DMABENCH_BASE + 0x008)
#define PHX_DMABENCH_CONST1               (PHX_DMABENCH_BASE + 0x00C)
#define PHX_DMABENCH_DMA_LS               (PHX_DMABENCH_BASE + 0x010)
#define PHX_DMABENCH_DMA_MS               (PHX_DMABENCH_BASE + 0x014)
#define PHX_DMABENCH_EXTENT               (PHX_DMABENCH_BASE + 0x018)
#define PHX_DMABENCH_ITER                 (PHX_DMABENCH_BASE + 0x01C)
#define PHX_DMABENCH_WRITE_VAL            (PHX_DMABENCH_BASE + 0x020)
#define PHX_DMABENCH_READ_VAL             (PHX_DMABENCH_BASE + 0x024)
#define PHX_DMABENCH_FREE                 (PHX_DMABENCH_BASE + 0x028)
#define PHX_DMABENCH_LAST                 (PHX_DMABENCH_BASE + 0x02C)
#define PHX_DMABENCH_COMMAND              (PHX_DMABENCH_BASE + 0x030)
#define PHX_DMABENCH_STATUS               (PHX_DMABENCH_BASE + 0x034)
#define PHX_DMABENCH_ITER_REM             (PHX_DMABENCH_BASE + 0x038)
#define PHX_DMABENCH_THIS_MPS             (PHX_DMABENCH_BASE + 0x03C)
#define PHX_DMABENCH_THIS_RRS             (PHX_DMABENCH_BASE + 0x040)

struct phx_icu_link_status {
  uint8_t status;         /* Set bit N indicates port N's link is up. */
  uint8_t intr_req;       /* Ports requesting interrupt for link change */
  uint8_t unused;
  uint8_t last_intr_status; /* Bit mask of prior link change interrupts sent*/
};

/* AB Arbiter
 * See the documentation for A/B Arbitration in /nfs/common/PHOENIX/ab/ABAritration.pdf
 */

#define PHX_AB_BASE                       0xB0000
#define PHX_AB_MODULE_ID                  0x000
#define PHX_AB_UNIT_COUNT                 0x004
#define PHX_AB_HISTORY                    0x008
#define PHX_AB_PORTS                      0x00c
#define PHX_AB_AUTORESET_ENABLE           0x010
#define PHX_AB_AUTORESET_BOUND            0x014
#define PHX_AB_AUTORESET_COUNT            0x018
#define PHX_AB_GLOBAL_BYPASS              0x020
#define PHX_AB_GLOBAL_RESET               0x024
#define PHX_AB_RESET_COUNTS               0x040
#define PHX_AB_BYPASS                     0x044
#define PHX_AB_PKT_COUNT                  0x050
#define PHX_AB_PKT_PASS                   0x054
#define PHX_AB_PRE_HISTORY                0x058
#define PHX_AB_PKT_A_PASSED               0x05c
#define PHX_AB_PKT_A_REMOVED              0x060
#define PHX_AB_PKT_B_PASSED               0x064
#define PHX_AB_PKT_B_REMOVED              0x068
#define PHX_AB_PKT_C_PASSED               0x06c
#define PHX_AB_PKT_C_REMOVED              0x070
#define PHX_AB_PKT_D_PASSED               0x074
#define PHX_AB_PKT_D_REMOVED              0x078
#define PHX_AB_LAST_UNIT                  0x080
#define PHX_AB_LAST_COUNT                 0x084
#define PHX_AB_LAST_SEQ                   0x088   

/* There are AB_UNIT_COUNT instances of each of the items below */
/* Sequence Number for this unit (8 bytes) */
#define PHX_AB_UNIT_SEQ_NUMBER            (PHX_AB_BASE + 0x4000)
/* 32-bit counter for the number of times a reset sequence packet was processed */
#define PHX_AB_UNIT_STATS_AUTORESET       (PHX_AB_BASE + 0x5000)
/* 64-bit counter for the number of packets passed because seq number was < (currentSeq - abHistory) */
#define PHX_AB_UNIT_STATS_PRE_HISTORY     (PHX_AB_BASE + 0x7000)
/* Unit statistics 8 counters of 8 bytes each */
#define PHX_AB_UNIT_STATS_BASE            (PHX_AB_BASE + 0x8000)
#define PHX_AB_STATS_A_PASSED             0x000
#define PHX_AB_STATS_A_DROPPED            0x008
#define PHX_AB_STATS_B_PASSED             0x010
#define PHX_AB_STATS_B_DROPPED            0x018
#define PHX_AB_STATS_C_PASSED             0x020
#define PHX_AB_STATS_C_DROPPED            0x028
#define PHX_AB_STATS_D_PASSED             0x030
#define PHX_AB_STATS_D_DROPPED            0x038

/* Altera MAC
 * See the documentation for the low latency MAC from Altera.
 * http://www.altera.com/literature/ug/ug_32b_10g_ethernet_mac.pdf
 */
#define PHX_MAC_BASE                      (0x100000+0x1000)
#define PHX_MAC_PRIMARY_MAC_ADDR0         (0x0010 * 4)
#define PHX_MAC_PRIMARY_MAC_ADDR1         (0x0011 * 4)
#define PHX_MAC_TX_CONTROL                (0x0020 * 4)
#define PHX_MAC_TX_MAX_FRAME_LENGTH       (0x002C * 4)
#define PHX_MAC_RX_FRAME_CONTROL          (0x00AC * 4)
#define PHX_MAC_TX_PAUSE_MACCTRL          (0x014a * 4)
#define PHX_MAC_RX_PAUSE_MACCTRL          (0x01ca * 4)
#define PHX_MAC_RX_FRAME_CONTROL_EN_ALLUCAST  (1 << 0)
#define PHX_MAC_RX_FRAME_CONTROL_EN_ALLMCAST  (1 << 1)
#define PHX_MAC_RX_CONTROL                (0x00A0 * 4)
#define PHX_MAC_RX_MAX_FRAME_LENGTH       (0x00AE * 4)

/* Xilinx MAC
 * See the documentation for the Xilinx MAC used on DBL2X.
 * http://www.xilinx.com/support/documentation/ip_documentation/ten_gig_eth_mac/v15_0/pg072-ten-gig-eth-mac.pdf
 */
#define PHX_XLX_MAC_BASE                      0x100000
#define PHX_XLX_MAC_TX_CONTROL                0x0408
#define PHX_XLX_MAC_TX_MTU_CONFIG             0x0418
#define PHX_XLX_MAC_RX_CONTROL                0x0404
#define PHX_XLX_MAC_RX_MTU_CONFIG             0x0414
#define PHX_XLX_MDIO_CFG0                     0x0500
#define PHX_XLX_MDIO_CFG1                     0x0504
#define PHX_XLX_MDIO_TX                       0x0508
#define PHX_XLX_MDIO_RX                       0x050C

/* These enable jumbo frames and enable the transceiver */
#define PHX_XLX_TX_ENABLE                     0x50000000
#define PHX_XLX_TX_DISABLE                    0x0
#define PHX_XLX_RX_ENABLE                     0x50000000
#define PHX_XLX_RX_DISABLE                    0x0

/* The NIC does not drop oversized frames (> MTU) right now.
 * Instead it puts the receive path in a bad state.  As a workaround, set
 * rx_max_frame_length=10K and check/drop frames in the driver.
 */
#define PHX_WORKAROUND_MAC_OVERSIZE_FRAMES 1

/* Hardware does an unsigned comparison, not a sequence
 * number comparison.  That is, "wake_request_index < completion_count",
 * not "(int)(completion_count - wake_request_index) > 0".
 *
 * Avoid writing the wake request past the send completion count.
 * E.g. writing 4294967295 when the send completion is 3.
 * The actual implementaion is quite messy.  We basically try to
 * go over the wraparound point without using interrupts.
 *
 * PHX-238
 */
#define PHX_WORKAROUND_TX_WAKE_WRAP_AROUND 0

/* PHXSNF-26 workaround */
#define PHX_WORKAROUND_PHXSNF_26 1

/* K35 has a SERDES preemphasis register for each port */
#define K35_SERDES_PREEMPHASIS_BASE           0x106000
#define K35_SERDES_PREEMPHASIS_SIZE           4
#define K35_SERDES_10G_COPPER                 0xf1205
#define K35_SERDES_10G_FIBER                  0xc0600
#define K35_SERDES_1G_COPPER                  0x80000
#define K35_SERDES_1G_FIBER                   0x80000

/* K35 has a MAC comparitor core
 * If the mac is not promiscuous, then only matching packets
 *  make it to the default endpoint.
 */
#define K35_MAC_FILTER_BASE                   0x107000
#define K35_MAC_FILTER_HIGH                   0x0000
#define K35_MAC_FILTER_LOW                    0x0004
#define K35_MAC_FILTER_MODE                   0x0080
#define K35_MAC_FILTER_CTRL                   0x0084
#define K35_MAC_FILTER_DISCARD_PKTS           0x0088
#define K35_MAC_FILTER_DISCARD_BYTES          0x008c
#define K35_MAC_FILTER_FWD_PKTS               0x0090
#define K35_MAC_FILTER_FWD_BYTES              0x0094

#define K35_MAC_FILTER_SIZE                   8
#define K35_NUM_MAC_FILTERS                   16

/* The mode bits represent "allow all" for each type */
#define K35_MAC_FILTER_MODE_PROMISC           0x0001
#define K35_MAC_FILTER_MODE_ALL_UNICAST       0x0002
#define K35_MAC_FILTER_MODE_ALL_MULTICAST     0x0004
#define K35_MAC_FILTER_MODE_ALL_BROADCAST     0x0008

typedef struct {
  uint32_t high;  /* msb justified 0060dd43 */
  uint32_t low;   /* xxyy0000 */
  uint32_t used;
} phx_mac_filter_t;

/* I2C cores */
#define PHX_I2C_BASE                      (0x100000+0x8000)
#define PHX_I2C_MUTEX                     0x200

#define PHX_LOS_MUTEX_CLAIM_VAL           0x00001
#define PHX_I2C_MUTEX_CLAIM_VAL           0x10001
#define PHX_I2C_MUTEX_RELEASE_VAL         0x10000

/* Xilinx (K35) General Registers relative to I2C offset */
#define K35_PORT_LINK_LEDS         (0x1000) // Link LEDs (FPGA controlled)
#define K35_PORT_ACTV_LEDS         (0x1004) // Activity LEDs (FPGA controlled)
#define K35_STATUS_LEDS            (0x1008) // Green Status LEDs
#define K35_SFP_PRESENT            (0x100C) // SFP presence detected register
#define K35_SFP_LOS                (0x1010) // SFP LOS state register
#define K35_SFP_TX_FAULT           (0x1014) // SFP TXFAULT condition register
#define K35_SFP_TX_DISABLE         (0x1018) // SFP TX disable control register
#define K35_FAN_PWM_PERCENT        (0x101C) // Control heatsink fan speed
#define K35_FAN_PWM                (0x1020) // Read heatsink pwm fan speed
#define K35_I2C_MUX_SELECT         (0x1024) // Allows remote i2c control on P9 connector (Aardvark)
#define K35_NOR_FLASH_WP           (0x1028) // Flash write protection (1 = unprotected)
#define K35_FPGA_PROG              (0x102C) // Force FPGA load from Flash
#define K35_ASCII_ID_0             (0x1030) // "CSPi"
#define K35_ASCII_ID_1             (0x1034) // "KU35" or "KU40"
#define K35_ASCII_ID_2             (0x1038) // "DBLX" or "SNFX"
#define K35_ASCII_ID_3             (0x103C) // "2P10"
#define K35_ASCII_ID_4             (0x1040) // "G..."
#define K35_FPGA_REV               (0x1044) // Board and Firmware Revisions
#define K35_FPGA_REV_BRANCH        (0x1048) // Firmware Branch
#define K35_MONTH_DAY              (0x104C) // Bitstream creation date
#define K35_TIME                   (0x1050) // Bitstream creation time

#define K35_MAC_RESET              (0x1064) // Reset MAC bits

#define K35_MAC_STRIDE             (0x1000) // Difference between MAC cores for multiport

/* perhaps above registers are not really relative to I2C */
#define K35_LTSSM_NOT_L0           (0x10906C) // Link Training and Status not in L0 state (cycles)

/* K35SNF receive packet alignment */
#define K35SNF_RX_PACKET_ALIGNMENT 4
#define K35SNF_ALIGN_RX_PACKET_LENGTH(len) \
  ((len + K35SNF_RX_PACKET_ALIGNMENT - 1) & ~(K35SNF_RX_PACKET_ALIGNMENT - 1))

/* K35SNF send packet alignment */
#define K35SNF_TX_PACKET_ALIGNMENT 4
#define K35SNF_ALIGN_TX_PACKET_LENGTH(len) \
  ((len + K35SNF_TX_PACKET_ALIGNMENT - 1) & ~(K35SNF_TX_PACKET_ALIGNMENT - 1))

typedef struct {
  uint64_t timestamp;
  uint32_t tag;
  uint16_t pktlen;
  uint8_t flags;
  uint8_t seqnum;
} k35snf_rx_desc_t;

typedef struct {
  uint64_t delta_ns;
  uint32_t tag; /* unused */
  uint16_t pktlen;
  uint8_t flags; /* unused */
  uint8_t seqnum; /* unused */
} k35snf_tx_desc_t;

/* A page starts with this structure.  These counts are deltas since
 * the last update.
 */
typedef struct {
  uint32_t packets;
  uint32_t bytes; /* a multiple of 4 */
} k35snf_tx_completion_t;

typedef struct {
  uint8_t linkStatus;
  uint8_t linkChangeIRReq;
  uint8_t unused;
  uint8_t linkLastIRStatus;
} k35snf_icu_link_status_t;

typedef struct {
  uint8_t presStatus;
  uint8_t presChangeIRReq;
  uint8_t unused;
  uint8_t presLastIRStatus;
} k35snf_icu_pres_status_t;

typedef struct {
  uint8_t endpt_bitmap[8];  /*  8 x 8 = 64 bits for 64 endpoints */
  uint32_t flags;
  uint8_t pad[3];
  uint8_t seqnum;      /* 1..255 looping */
} k35snf_intr_t;

typedef struct {
  uint64_t pattern;
  uint64_t mask;
} k35snf_L2_filter_pattern_mask_t;

typedef struct {
  uint16_t offset;
  uint16_t data;
} k35snf_L2_filter_lookup_table_t;

#define K35SNF_INTR_FLAG_WAKE_TX     (1 << 4)
#define K35SNF_INTR_FLAG_WAKE_RX     (1 << 5)

#define K35SNF_SYS_BASE                    0x00000
#define K35SNF_SYS_R0                      0x000
#define K35SNF_SYS_R4                      0x004
#define K35SNF_SYS_R8                      0x008
#define K35SNF_SYS_RC                      0x00C
#define K35SNF_SYS_CONST0                  0x010
#define K35SNF_SYS_CONST1                  0x014
#define K35SNF_SYS_FCSYS0                  0x01C
#define K35SNF_SYS_COMMIT_ID0              0x020
#define K35SNF_SYS_COMMIT_ID1              0x024
#define K35SNF_SYS_COMMIT_ID2              0x028
#define K35SNF_SYS_COMMIT_ID3              0x02C
#define K35SNF_SYS_TIME_STATUS             0x030
#define K35SNF_SYS_TIME_CONTROL            0x034
#define K35SNF_SYS_TIME_LS                 0x038
#define K35SNF_SYS_TIME_MS                 0x03C
#define K35SNF_SYS_TIME_COMPARE_LS         0x040
#define K35SNF_SYS_TIME_COMPARE_MS         0x044
#define K35SNF_SYS_TIME_REFPERPPS          0x048
#define K35SNF_SYS_TIME_TRIM_PPB           0x04C
#define K35SNF_SYS_TIME_SKIP_NS            0x050
#define K35SNF_SYS_LED_DRIVE               0x068
#define K35SNF_SYS_GPIO_DIR                0x080
#define K35SNF_SYS_GPIO_OUT                0x088
#define K35SNF_SYS_GPIO_IN                 0x090
#define K35SNF_SYS_GPIO_ERISING            0x098
#define K35SNF_SYS_GPIO_EFALLING           0x0A0

/* K35SNF has 4 APUs one for each port each APU has 4 channels */
#define K35SNF_RX_APU_BASE                 0x20000
#define K35SNF_RX_APU_SIZE                 0x4000
#define K35SNF_RX_APU_CHANNEL_SIZE         0x100
#define K35SNF_RX_APU_ID_VERSION           0x000
#define K35SNF_RX_APU_NUM_CHANNEL          0x004 /* 2B */
#define K35SNF_RX_APU_ENDPOINTS            0x006 /* 1B */
#define K35SNF_RX_APU_PORTS                0x007 /* 1B */
#define K35SNF_RX_APU_ALPHA_FETCHES        0x008
#define K35SNF_RX_APU_BETA_FETCHES         0x00C
#define K35SNF_RX_APU_GAMMA_FETCHES        0x010
#define K35SNF_RX_APU_LAST_GAMMA_CH        0x014 /* 1B */
#define K35SNF_RX_APU_LOOP_CH              0x015 /* 1B */
#define K35SNF_RX_APU_PCI_DEVID            0x018
#define K35SNF_RX_APU_ALPHA_NEXT_LS        0x020
#define K35SNF_RX_APU_ALPHA_NEXT_MS        0x024
#define K35SNF_RX_APU_ALPHA_COUNT          0x028
#define K35SNF_RX_APU_ALPHA_TAKE           0x02C
#define K35SNF_RX_APU_BETA_NEXT_LS         0x040
#define K35SNF_RX_APU_BETA_NEXT_MS         0x044
#define K35SNF_RX_APU_BETA_COUNT           0x048
#define K35SNF_RX_APU_GAMMA_NEXT_LS        0x060
#define K35SNF_RX_APU_GAMMA_NEXT_MS        0x064
#define K35SNF_RX_APU_GAMMA_COUNT          0x068
#define K35SNF_RX_APU_GAMMA_TAKE           0x06C
#define K35SNF_RX_APU_CHANNEL_RESET        0x07C

/* K35SNF has 4 SnifferRX Lite blocks, one for each port.
 * Each block occupies 0x1000.
 */
#define K35SNF_RX_SIZE                     0x1000
#define K35SNF_RX_BASE                     0x30000
#define K35SNF_RX_RX_R0                    0x000
#define K35SNF_RX_RX_R4                    0x004
#define K35SNF_RX_RX_CYCLE_COUNT           0x008
#define K35SNF_RX_RX_CONST0                0x00C /* 0xBACECACE */
#define K35SNF_RX_COMMAND                  0x014
#define K35SNF_RX_BYTE_COUNT               0x020 /* 48 bit rolling */
#define K35SNF_RX_PACKET_COUNT             0x028 /* 48 bit rolling */
#define K35SNF_RX_FF_PACKET_COUNT          0x038 /* 48 bit rolling */
#define K35SNF_RX_TIME_LIMIT               0x068
#define K35SNF_RX_DATA_TLP_Q1              0x078
#define K35SNF_RX_DATA_TLP_Q2              0x07C
#define K35SNF_RX_DATA_TLP_Q3              0x080
#define K35SNF_RX_DATA_TLP_Q4              0x084
#define K35SNF_RX_DATA_TLP_FULL            0x088
#define K35SNF_RX_DESC_TLP_Q1              0x08C
#define K35SNF_RX_DESC_TLP_Q2              0x090
#define K35SNF_RX_DESC_TLP_Q3              0x094
#define K35SNF_RX_DESC_TLP_Q4              0x098
#define K35SNF_RX_DESC_TLP_FULL            0x09C
#define K35SNF_RX_PCIBP_RESET              0x0A0
#define K35SNF_RX_PCIBP_EMPTY              0x0A4
#define K35SNF_RX_PCIBP_Q1                 0x0A8
#define K35SNF_RX_PCIBP_Q2                 0x0AC
#define K35SNF_RX_PCIBP_Q3                 0x0B0
#define K35SNF_RX_PCIBP_Q4                 0x0B4
#define K35SNF_RX_PCIBP_FULL               0x0B8
#define K35SNF_RX_TLP_STATS_CLEAR          0x0BC
#define K35SNF_RX_DATA_TLP_PS_MIN          0x0C0
#define K35SNF_RX_DATA_TLP_PS_MAX          0x0C4
#define K35SNF_RX_DATA_TLP_FULL_PS_MIN     0x0C8
#define K35SNF_RX_DATA_TLP_FULL_PS_MAX     0x0CC
#define K35SNF_RX_DATA_TLP_DW_PS_MIN       0x0D0
#define K35SNF_RX_DATA_TLP_DW_PS_MAX       0x0D4
#define K35SNF_RX_DATA_TLP_PLDW_PS_MIN     0x0D8
#define K35SNF_RX_DATA_TLP_PLDW_PS_MAX     0x0DC
#define K35SNF_RX_DESC_TLP_PS_MIN          0x0E0
#define K35SNF_RX_DESC_TLP_PS_MAX          0x0E4
#define K35SNF_RX_DESC_TLP_FULL_PS_MIN     0x0E8
#define K35SNF_RX_DESC_TLP_FULL_PS_MAX     0x0EC
#define K35SNF_RX_DESC_TLP_DW_PS_MIN       0x0F0
#define K35SNF_RX_DESC_TLP_DW_PS_MAX       0x0F4
#define K35SNF_RX_DESC_TLP_PLDW_PS_MIN     0x0F8
#define K35SNF_RX_DESC_TLP_PLDW_PS_MAX     0x0FC
/* Only one RX endpoint per port */
#define K35SNF_RX_DATA_RING_SIZE           0x100
#define K35SNF_RX_DATA_CONSR_ADVANCE       0x104
#define K35SNF_RX_DATA_CONSR_POSITION      0x108
#define K35SNF_RX_DATA_PRODW_POSITION      0x10C
#define K35SNF_RX_DESC_RING_SIZE           0x110
#define K35SNF_RX_DESC_CONSR_ADVANCE       0x114
#define K35SNF_RX_DESC_CONSR_POSITION      0x118
#define K35SNF_RX_DESC_PRODW_POSITION      0x11C
#define K35SNF_RX_EP_BYTE_COUNT            0x120 /* 48 bit = RX_BYTE_COUNT */
#define K35SNF_RX_EP_PACKET_COUNT          0x128 /* 48 bit = RX_PACKET_COUNT */
#define K35SNF_RX_EP_DESC_COUNT            0x134
#define K35SNF_RX_EP_DESC_WAKE_REQUEST     0x138 /* FIXME: Seems to be write only */
#define K35SNF_RX_STOP_FLUSHED             0x13C
#define K35SNF_RX_DATA_PG_AVAIL            0x140
#define K35SNF_RX_DESC_PG_AVAIL            0x144
#define K35SNF_RX_DATA_PG_AVAIL_MIN        0X148
#define K35SNF_RX_DESC_PG_AVAIL_MIN        0x14C

#define K35SNF_RSS_SIZE                  0x1000
#define K35SNF_RSS_BASE                  0x50000
#define K35SNF_RSS_MASK                  0x080
#define K35SNF_RSS_SELECT                0x084
#define K35SNF_RSS_RESET                 0x088
#define K35SNF_RSS_PKT_SENT              0x100
#define K35SNF_RSS_BYTES_SENT            0x104
#define K35SNF_RSS_PKT_DROP              0x108
#define K35SNF_RSS_BYTES_DROP            0x10c
#define K35SNF_RSS_PKT_ERR               0x110
#define K35SNF_RSS_BYTES_ERR             0x114
#define K35SNF_RSS_DISP_PKT_FWD          0x124
#define K35SNF_RSS_DISP_PKT_DROP         0x128
/* RSS hash selector options */
#define K35SNF_RSS_MODE_TUNNELED         (1 << 0)
#define K35SNF_RSS_MODE_OUTER_IP         (1 << 1)
#define K35SNF_RSS_MODE_LINK             (1 << 2)

/* RSS mask bits, 1 = ignore field */
#define K35SNF_RSS_MASK_PORT            (1 << 0)
#define K35SNF_RSS_MASK_MPLS            (1 << 1)
#define K35SNF_RSS_MASK_VLAN            (1 << 2)
#define K35SNF_RSS_MASK_DST_IP          (1 << 3)
#define K35SNF_RSS_MASK_SRC_IP          (1 << 4)
#define K35SNF_RSS_MASK_TEID            (1 << 5)
#define K35SNF_RSS_MASK_DST_PORT        (1 << 6)
#define K35SNF_RSS_MASK_SRC_PORT        (1 << 7)

/* Bit descriptions of the RSS reset register. */
#define K35SNF_MERGE_NEXT_RING_CLEAR_RSS_STATS     0x1
#define K35SNF_MERGE_NEXT_RING_RESET_RSS_BLOCK     0x2
#define K35SNF_MERGE_NEXT_RING_DISABLE_SEQUENCER   0x4

/* Next ring extraction from descriptor definitions */
#define K35SNF_MERGE_NEXT_RING_ID_SHIFT  24
#define K35SNF_MERGE_NEXT_RING_ID_MASK   0xf

#define K35SNF_TX_SIZE                   0x1000
#define K35SNF_TX_BASE                   0x60000
#define K35SNF_TX_R0                     0x000
#define K35SNF_TX_TLP_STATS_CLEAR        0x004
#define K35SNF_TX_CONST0                 0x008
#define K35SNF_TX_COMMAND                0x010
#define K35SNF_TX_STOP_FLUSHED           0x014
#define K35SNF_TX_TIME_LIMIT             0x018
#define K35SNF_TX_BYTE_COUNT             0x020
#define K35SNF_TX_PACKET_COUNT           0x028
#define K35SNF_TX_COMPL_ADDR_LOW         0x030
#define K35SNF_TX_COMPL_ADDR_HIGH        0x034
#define K35SNF_TX_COMPL_COUNT            0x038
#define K35SNF_TX_COMPL_COUNT_ACK        0x03C
#define K35SNF_TX_DATA_MRD_Q1            0x040
#define K35SNF_TX_DATA_MRD_Q2            0x044
#define K35SNF_TX_DATA_MRD_Q3            0x048
#define K35SNF_TX_DATA_MRD_Q4            0x04C
#define K35SNF_TX_DATA_MRD_FULL          0x050
#define K35SNF_TX_DESC_MRD_Q1            0x054
#define K35SNF_TX_DESC_MRD_Q2            0x058
#define K35SNF_TX_DESC_MRD_Q3            0x05C
#define K35SNF_TX_DESC_MRD_Q4            0x060
#define K35SNF_TX_DESC_MRD_FULL          0x064
#define K35SNF_TX_DATA_CPLD_Q1           0x068
#define K35SNF_TX_DATA_CPLD_Q2           0x06C
#define K35SNF_TX_DATA_CPLD_Q3           0x070
#define K35SNF_TX_DATA_CPLD_Q4           0x074
#define K35SNF_TX_DATA_CPLD_FULL         0x078
#define K35SNF_TX_DESC_CPLD_Q1           0x07C
#define K35SNF_TX_DESC_CPLD_Q2           0x080
#define K35SNF_TX_DESC_CPLD_Q3           0x084
#define K35SNF_TX_DESC_CPLD_Q4           0x088
#define K35SNF_TX_DESC_CPLD_FULL         0x08C
#define K35SNF_TX_DATA_MRD_PS_MIN        0x090
#define K35SNF_TX_DATA_MRD_PS_MAX        0x094
#define K35SNF_TX_DATA_MRD_FULL_PS_MIN   0x098
#define K35SNF_TX_DATA_MRD_FULL_PS_MAX   0x09C
#define K35SNF_TX_DATA_MRD_DW_PS_MIN     0x0A0
#define K35SNF_TX_DATA_MRD_DW_PS_MAX     0x0A4
#define K35SNF_TX_DESC_MRD_PS_MIN        0x0A8
#define K35SNF_TX_DESC_MRD_PS_MAX        0x0AC
#define K35SNF_TX_DESC_MRD_FULL_PS_MIN   0x0B0
#define K35SNF_TX_DESC_MRD_FULL_PS_MAX   0x0B4
#define K35SNF_TX_DESC_MRD_DW_PS_MIN     0x0B8
#define K35SNF_TX_DESC_MRD_DW_PS_MAX     0x0BC
#define K35SNF_TX_DATA_CPLD_PS_MIN       0x0C0
#define K35SNF_TX_DATA_CPLD_PS_MAX       0x0C4
#define K35SNF_TX_DATA_CPLD_FULL_PS_MIN  0x0C8
#define K35SNF_TX_DATA_CPLD_FULL_PS_MAX  0x0CC
#define K35SNF_TX_DATA_CPLD_DW_PS_MIN    0x0D0
#define K35SNF_TX_DATA_CPLD_DW_PS_MAX    0x0D4
#define K35SNF_TX_DESC_CPLD_PS_MIN       0x0D8
#define K35SNF_TX_DESC_CPLD_PS_MAX       0x0DC
#define K35SNF_TX_DESC_CPLD_FULL_PS_MIN  0x0E0
#define K35SNF_TX_DESC_CPLD_FULL_PS_MAX  0x0E4
#define K35SNF_TX_DESC_CPLD_DW_PS_MIN    0x0E8
#define K35SNF_TX_DESC_CPLD_DW_PS_MAX    0x0EC
#define K35SNF_TX_DATA_RING_SIZE         0x100
#define K35SNF_TX_DATA_CONSR_POSITION    0x10C
#define K35SNF_TX_DESC_RING_SIZE         0x110
#define K35SNF_TX_DESC_PRODW_ADVANCE     0x114
#define K35SNF_TX_DESC_PRODW_POSITION    0x118
#define K35SNF_TX_DESC_CONSR_POSITION    0x11C

#define K35SNF_XMAC_RX_SIZE                  0x1000
#define K35SNF_XMAC_RX_BASE                  0x80000
#define K35SNF_XMAC_RX_R0                    0x000
#define K35SNF_XMAC_RX_STATS_CLEAR           0x004
#define K35SNF_XMAC_RX_PKTS_RCVD             0x008
#define K35SNF_XMAC_RX_PKTS_ABORT            0x010
#define K35SNF_XMAC_RX_PKTS_DROPPED          0x018
#define K35SNF_XMAC_RX_PKTS_NO_META          0x020
#define K35SNF_XMAC_RX_PKTS_NO_OUT_CTRL      0x028
#define K35SNF_XMAC_RX_PKTS_DROP_ABORT       0x030
#define K35SNF_XMAC_RX_PKTS_SENT             0x038 /* 64 bit */
#define K35SNF_XMAC_RX_BYTES_SENT            0x040 /* 64 bit */
#define K35SNF_XMAC_RX_SEND_DUP              0x048
#define K35SNF_XMAC_RX_SENT_WITH_ERR         0x050
#define K35SNF_XMAC_RX_EXTRA_META            0x058
#define K35SNF_XMAC_RX_CTRL_DROPPED          0x060
#define K35SNF_XMAC_RX_DISCARDED             0x068

/* L2 filter configuration. */
#define K35SNF_XMAC_RX_FILTER_PATTERN_MASK   0x200  /* Array of 16 k35snf_L2_filter_pattern_mask_t structures. */
#define K35SNF_XMAC_RX_FILTER_MATCH_COUNTS   0x300  /* Count of matches. */
#define K35SNF_XMAC_RX_FILTER_LOOKUP_TABLE   0x380  /* 32 keyhole addresses for programming MTAU. */
#define K35SNF_XMAC_RX_FILTER_MU_ENABLE      0x400  /* 16 bit register, each bit enables its respective match unit. */
#define K35SNF_XMAC_RX_FILTER_MU_COUNT       16     /* There are 16 match units. */
#define K35SNF_XMAC_RX_FILTER_LT_ENTRY_COUNT (64 * 1024)

/* 4KB per Xilinx port */
#define K35SNF_XMAC_SIZE                     0x1000
#define K35SNF_XMAC_BASE                     0x100000
#define K35SNF_XMAC_RX_BYTES             0x200
#define K35SNF_XMAC_TX_BYTES             0x208
#define K35SNF_XMAC_RX_FRAMES_RUNT       0x210
#define K35SNF_XMAC_RX_FRAMES_FRAG       0x218
#define K35SNF_XMAC_RX_FRAMES_64         0x220
#define K35SNF_XMAC_RX_FRAMES_65_127     0x228
#define K35SNF_XMAC_RX_FRAMES_128_255    0x230
#define K35SNF_XMAC_RX_FRAMES_256_511    0x238
#define K35SNF_XMAC_RX_FRAMES_512_1023   0x240
#define K35SNF_XMAC_RX_FRAMES_1024_MAX   0x248
#define K35SNF_XMAC_RX_FRAMES_OVERSIZE   0x250
#define K35SNF_XMAC_TX_FRAMES_64         0x258
#define K35SNF_XMAC_TX_FRAMES_65_127     0x260
#define K35SNF_XMAC_TX_FRAMES_128_255    0x268
#define K35SNF_XMAC_TX_FRAMES_256_511    0x270
#define K35SNF_XMAC_TX_FRAMES_512_1023   0x278
#define K35SNF_XMAC_TX_FRAMES_1024_MAX   0x280
#define K35SNF_XMAC_TX_FRAMES_OVERSIZE   0x288
#define K35SNF_XMAC_RX_FRAMES_OK         0x290
#define K35SNF_XMAC_RX_FCS_BAD           0x298
#define K35SNF_XMAC_RX_BCAST_OK          0x2A0
#define K35SNF_XMAC_RX_MCAST_OK          0x2A8
#define K35SNF_XMAC_RX_MAC_CTRL          0x2B0
#define K35SNF_XMAC_RX_LT_RANGE          0x2B8
#define K35SNF_XMAC_RX_VLAN              0x2C0
#define K35SNF_XMAC_RX_PAUSE             0x2C8
#define K35SNF_XMAC_RX_MAC_UNSUP         0x2D0
#define K35SNF_XMAC_TX_FRAMES_OK         0x2D8
#define K35SNF_XMAC_TX_BCAST_OK          0x2E0
#define K35SNF_XMAC_TX_MCAST_OK          0x2E8
#define K35SNF_XMAC_TX_UNDERRUN          0x2F0
#define K35SNF_XMAC_RX_CONFIG_WORD_0         0x400
#define K35SNF_XMAC_RX_CONFIG_WORD_1         0x404
#define K35SNF_XMAC_TX_CONFIG_WORD           0x408
#define K35SNF_XMAC_FLOW_CTRL_CONFIG         0x40C
#define K35SNF_XMAC_RECONL_CONFIG_WORD       0x410
#define K35SNF_XMAC_RX_MTU_CONFIG_WORD       0x414
#define K35SNF_XMAC_TX_MTU_CONFIG_WORD       0x418
#define K35SNF_XMAC_VERSION                  0x4F8
#define K35SNF_XMAC_CAPABILITY               0x4FC
#define K35SNF_XMAC_MDIO_CFG0                0x500
#define K35SNF_XMAC_MDIO_CFG1                0x504
#define K35SNF_XMAC_MDIO_TX                  0x508
#define K35SNF_XMAC_MDIO_RX                  0x50C


#define K35SNF_MDIO_OP_ADDR                  0
#define K35SNF_MDIO_OP_WRITE                 1
#define K35SNF_MDIO_OP_READ                  2
#define K35SNF_MDIO_OP_EXT_READ              3

/* These are the reported value types in the Version register */
#define K35SNF_MAC_1G                        0x09000000
#define K35SNF_MAC_10G                       0x0f010000   /* aka XMAC */

/* One per-NIC ICU */
#define K35SNF_ICU_BASE                      0x70000
#define K35SNF_ICU_ENDPOINTS                 0x0000
#define K35SNF_ICU_OPERATION_MODE            0x0004
#define K35SNF_ICU_VECTORS_ALLOC             0x0008
#define K35SNF_ICU_GLOBAL_MASK               0x000C
#define K35SNF_ICU_ENABLE                    0x0010
#define K35SNF_ICU_ENABLE_IR                 0x001C
#define K35SNF_ICU_DISABLE_IR                0x0020
#define K35SNF_ICU_INTR_INTERVAL             0x0024
#define K35SNF_ICU_PCI_ID                    0x0030
#define K35SNF_ICU_LINK_STATUS               0x0040
#define K35SNF_ICU_REQ_LINK_CHANGE_IR        0x0044
#define K35SNF_ICU_CLR_LINK_CHANGE_IR        0x004C
#define K35SNF_ICU_SET_STATUS_UP             0x0054
#define K35SNF_ICU_SET_STATUS_DOWN           0x0058
#define K35SNF_ICU_EP_RESET                  0x0060
#define K35SNF_ICU_PRES_STATUS               0x0064
#define K35SNF_ICU_REQ_PRES_CHANGE_IR        0x0068
#define K35SNF_ICU_CLR_PRES_CHANGE_IR        0x0070
#define K35SNF_ICU_SET_PRES_IN               0x0078
#define K35SNF_ICU_SET_PRES_OUT              0x007C
#define K35SNF_ICU_CP_LINK_UP_IR             0x1000
#define K35SNF_ICU_CP_LINK_DN_IR             0x1004
#define K35SNF_ICU_CP_PRES_IN_IR             0x1008
#define K35SNF_ICU_CP_PRES_OUT_IR            0x100C
#define K35SNF_ICU_CP_EP_WAKE_RX_IR          0x1010
#define K35SNF_ICU_CP_EP_WAKE_TX_IR          0x1014
#define K35SNF_ICU_LAST_IR                   0x1020
/* Each word of the 16B descriptor */
#define K35SNF_ICU_LAST_IR_DESC_W0           0x1030
#define K35SNF_ICU_LAST_IR_DESC_W1           0x1034
#define K35SNF_ICU_LAST_IR_DESC_W2           0x1038
#define K35SNF_ICU_LAST_IR_DESC_W3           0x103C
#define K35SNF_ICU_DISABLE_IR_DESC           0x1040
/* EP_STATUS + i x 4 = endpoint i's status */
#define K35SNF_ICU_EP_STATUS                 0x2000
#define K35SNF_ICU_EP_STATUS_RX_REQUESTED_IR (1 << 3)
#define K35SNF_ICU_EP_STATUS_TX_REQUESTED_IR (1 << 4)
/* IDT_ADDRESS + i x 8 = DMA address of the interrupt i's descriptor */
#define K35SNF_ICU_IDT_ADDRESS               0x4000

/* These are opaque and are not defined anywhere.  We only use pointers.
 * They are used to catch mistakes where we use global instead of endpoint,
 * and vice versa.
 */

struct phx_nic_mem_global;
struct phx_nic_mem_endpoint;

#endif /* _phx_hw_h_ */
