/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2009 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

#include "mx_arch.h"
#include "mx_instance.h"
#include "mcp_wrapper_common.h"

#include "mcp_array_10g.c"

#define CTA(x) do { char (*a)[(x) ? 1 : -1] = 0; (void) a; } while (0)

int
myri_select_mcp(int board_type, const unsigned char **mcp,
		int *len, unsigned int *parity_critical_start,
		unsigned int *parity_critical_end)
{
  switch (board_type) {
  case MAL_BOARD_TYPE_ZOE:
    *mcp = mcp_array_ze;
    *len = mcp_array_ze_length;
    *parity_critical_start = 0;
    *parity_critical_end = 0;
    break;

  case MAL_BOARD_TYPE_ZOES:
    *mcp = mcp_array_zes;
    *len = mcp_array_zes_length;
    *parity_critical_start = 0;
    *parity_critical_end = 0;
    break;
  
  default:
    /* we should never get here */
    *mcp = NULL;
    *len = 0;
    break;
  }

  if (*len == 0)
    return 1;
  
  return 0;
}


int
myri_counters(int board_type, const char ***counters, uint32_t *count, const uint32_t **cclass)
{
  switch (board_type) {
  case MAL_BOARD_TYPE_ZOE:
  case MAL_BOARD_TYPE_ZOES:
    *counters = mcp_counters_ze_str;
    *count = sizeof(struct mcp_counters_ze) / sizeof(uint32_t);
    *cclass = mcp_counters_ze_class;
    CTA(sizeof(struct mcp_counters_ze) <= MX_MAX_COUNTER_SIZE);
    break;
  
  default:
    /* we should never get here */
    return 1;
  }
  return 0;
}

static inline
uint64_t
__u64_accum(mx_instance_state_t *is, int idx)
{
  return is->counters.accum[idx] + 
          (uint64_t)((uint32_t) ntohl(*((uint32_t *) is->counters.addr + idx))
           - is->counters.accum_last[idx]);
}

#define _MCP_GET_COUNTER_IDX(type,field) \
        offsetof(type, field) / sizeof(__typeof__( ((type *)0)->field ))
#define MCP_GET_ZE_IDX(field) \
          _MCP_GET_COUNTER_IDX(struct mcp_counters_ze, field)

/*
 * This function uses a read-copy mechanism to update 64-bit counters.  The
 * caller is responsible for ensuring that the output counter array is
 * sufficiently large to accomodate all updates.
 */
int
myri_counters64_get(mx_instance_state_t *is, uint64_t *c_o, int num_counters,
                    enum myri_counters_get_type type)
{
  uint32_t seqnum;
  int i;

  do {
 try_again:
    seqnum = is->counters.seqnum;
    MAL_READBAR();
    if (unlikely(seqnum & 1)) { /* back off, hit middle of update */
      MAL_CPU_RELAX();
      goto try_again;
    }
    switch (type) {
      case MYRI_COUNTERS_GET_SNF:
        if (num_counters != 7)
          return EINVAL;
        c_o[0] = __u64_accum(is, MCP_GET_ZE_IDX(snf_recv));
        c_o[1] = __u64_accum(is, MCP_GET_ZE_IDX(snf_drop_flow));
        c_o[2] = __u64_accum(is, MCP_GET_ZE_IDX(net_overflow_drop));
        c_o[3] = __u64_accum(is, MCP_GET_ZE_IDX(net_bad_crc32));
        c_o[4] = __u64_accum(is, MCP_GET_ZE_IDX(net_recv_kbytes));
        c_o[5] = __u64_accum(is, MCP_GET_ZE_IDX(snf_send));
        c_o[6] = __u64_accum(is, MCP_GET_ZE_IDX(net_send_kbytes));
        break;
      case MYRI_COUNTERS_GET_ALL:
      default:
        num_counters = num_counters < is->counters.cnt ? 
                       num_counters : is->counters.cnt;
        for (i = 0; i < num_counters; i++)
          c_o[i] = __u64_accum(is, i);
        break;
    }
    MAL_READBAR();
  } while (seqnum != is->counters.seqnum);
  return 0;
}

static
void
__myri_counters64_modify(mx_instance_state_t *is, int is_clear)
{
  uint32_t *stat_n = (uint32_t *) is->counters.addr;
  uint32_t val_prev, val_next;
  int i;

  mal_assert((offsetof(struct mcp_counters_ze, lanai_uptime) == 0));

  mx_spin_lock(&is->counters.spinlock);
  is->counters.seqnum++;
  MAL_WRITEBAR();
  for (i = 0; i < is->counters.cnt; i++) {
    if (is_clear) {
      /* When clearing counters, we always skip the first counter which is
       * lanai_uptime  -- see assertion above */
      if (i == 0)
        continue;
      is->counters.accum[i] = 0;
      is->counters.accum_last[i] = 0;
      *((uint32_t *)is->counters.addr + i) = 0;
    }
    else {
      val_prev = is->counters.accum_last[i];
      val_next = ntohl(stat_n[i]);
      is->counters.accum[i] += (val_next - val_prev);
      is->counters.accum_last[i] = val_next;
    }
  }
  MAL_WRITEBAR();
  is->counters.seqnum++;
  mx_spin_unlock(&is->counters.spinlock);
}

void
myri_counters64_update(mx_instance_state_t *is)
{
  __myri_counters64_modify(is, 0 /* false, not clearing */);
}

void
myri_counters64_clear(mx_instance_state_t *is)
{
  __myri_counters64_modify(is, 1 /* true, yes clearing */);
}
