/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2011 by Myricom, Inc.  All rights reserved.          *
 * Copyright 2014 - 2016 by CSP, Inc.  All rights reserved.              *
 *************************************************************************/

#ifndef _mx_instance_h_
#define _mx_instance_h_

#include "mal.h"
#include "mal_io.h"
#include "mx_pin.h"
#include "mx_misc.h"
#include "mx_pci.h"
#ifndef MYRI_PHOENIX
#include "mcp_requests.h"
#include "mcp_wrapper_common.h"
#endif

#ifndef MX_DISABLE_ETHERNET
#define MX_DISABLE_ETHERNET 0
#endif
#ifndef MX_HAS_PCIE_LINK_RESET
#define MX_HAS_PCIE_LINK_RESET 0
#endif

/* A static map of driver endpoints and phoenix endpoints.
 *
 * SNF bistream internally partitions endpoints.  On E7, each port gets
 * 4 endpoints.  0-3 for port 0, 4-7 for port 1, and so on.
 * So, we reserve the kernel and SNF endpoints during init.  Only TX
 * endpoints are allocated from the pool.
 *
 * type          driver endpoints    phoenix endpoints
 * Kernel        none                E7: 0, 4, 8, 12 // E1: 0, 8
 * SNF_TX        0 ... 31            0 ... 31 except kernel and SNF_RX
 * SNF_RX        32 ... 63           E7: 1, 5, 9, 13 // E1: 1, 9
 * SNF_RX_RING   64 ... 95           none
 * SNF_RX_GROUP  96 ... 127          none
 * SNF_HANDLE    128 ... 159         none
 * SNF_RX_PQ     160 ... 191         none
 *
 * We special case kernel and SNF_RX endpoints when handling
 * interrupts so that they either do not need driver endpoints
 * (kernel) or driver endpoints and phoenix endpoints can differ
 * (SNF_RX).
 */
#define PHX_SNF_MAX_ENDPOINTS  32


/* forward declaration */
typedef struct mx_instance_state mx_instance_state_t;
typedef struct mx_endpt_state mx_endpt_state_t;
typedef struct mx_waitq mx_waitq_t;
typedef struct mx_page_pin mx_page_pin_t;
typedef struct mx_mcp_dma_chunk mx_mcp_dma_chunk_t;
typedef struct mx_host_dma_win mx_host_dma_win_t;
typedef struct myri_rx_filterq_entry myri_rx_filterq_entry_t;

#ifdef MYRI_PHOENIX
struct phx_ether;
#else
struct mx_ether;
#endif

#ifdef MYRI_PHOENIX
typedef struct {
  mx_page_pin_t *pages;
  int n_pages;
} phx_pages_t;

/* 3-level receive ring.
 * Allocation/free code, common to DBL and SNF.
 */
typedef struct {
  uint64_t size;          /* all of gamma pages in bytes */
  phx_pages_t gamma;      /* contains DMA data from NIC */
  phx_pages_t beta;       /* 2nd level pages */
  phx_pages_t alpha;      /* 1st level pages */
  uint32_t preamble_size;
  uint32_t preamble_n_pages;
#if MAL_OS_WINDOWS
  /* Gamma pages are allocated differently on Windows. */
  void *gamma_mdl;
#endif
} phx_rx_ring_t;

int phx_alloc_rx_ring(mx_instance_state_t *is, phx_rx_ring_t *ring,
                      uint64_t size, uint32_t preamble_size);
void phx_free_rx_ring(mx_instance_state_t *is, phx_rx_ring_t *ring);

int phx_alloc_pages(mx_instance_state_t *is, phx_pages_t *p, int n_pages);
void phx_free_pages(mx_instance_state_t *is, phx_pages_t *p);

#if MAL_OS_WINDOWS
/* FIXME.  Move to a proper header file. */
int phx_win_alloc_gamma_pages(mx_instance_state_t *is, phx_pages_t *p,
                              int n_pages, void **gamma_mdl);
void phx_win_free_gamma_pages(mx_instance_state_t *is, phx_pages_t *p,
                              void *gamma_mdl);
#endif

extern mx_sync_t mx_fpga_mutex;
#endif

#define MYRI_HOST_DMA32(x)  htonl(x)

#define MYRI_COPYBLOCK_F_STATIC  0x1
#define MYRI_COPYBLOCK_F_KALLOC  0x2
#define MYRI_COPYBLOCK_F_HUGE    0x4
typedef struct mx_copyblock
{
  void *alloc_addr;
  void *addr;
  uint32_t size;
  uint32_t nvpages;
  int flags;
  mx_page_pin_t *pins;
  nic_dma_addr_t *dmas;
#if MAL_OS_WINDOWS
  PMDL mdl;
#endif
} mx_copyblock_t;

/* mx_huge_copyblock is 3-level page table/ring for Lanai.
 * 1. Data pages in host memory
 * 2. Pages in host memory that store (1)'s DMA addresses
 * 3. DMA addresses of (2) in NIC memory
 *
 * Phoenix 3-level page table is a bit different and needs pages
 * that hold additional descriptors.  FIXME.
 * 1. Data pages in host memory
 * 2. Pages in host memory that store (1)'s DMA addresses
 * 3. Pages in host memory that store phx_apu_flp_t's.
 *    These structures form a linked list and contain DMA addresses of (2).
 * 4. DMA address of the first phx_apu_flp_t in NIC memory.
 */
typedef struct mx_huge_copyblock
{
#if MAL_OS_LINUX || MAL_OS_UDRV
  struct page **pages;
#elif MAL_OS_WINDOWS
  PMDL mdl;
#endif
  void *kva;
  int nvpages;
  uint64_t size;
  uint64_t size_requested;
  mx_page_pin_t *pins;
  mx_copyblock_t cb;
  unsigned long pinned_pages;
} mx_huge_copyblock_t;



#define mx_cb_to_huge_to_copyblock(cbptr) \
  (struct mx_huge_copyblock *)((uintptr_t)(cbptr) - offsetof(mx_huge_copyblock_t, cb))

typedef struct myri_copyblock
{
  enum { MYRI_CB_K = 0, MYRI_CB_K_HUGE } type;
  int cursize;
  int defsize;
  int flags;
  union {
    mx_huge_copyblock_t hcb;
    mx_copyblock_t cb;
  } u;
} myri_copyblock_t;

/* ops support poll mechanism */
#define MYRI_EOP_F_POLL 0x1

typedef struct myri_endpt_ops {
  /* es->is already set and is lock is held, do necessary to set the type of
   * endpoint being allocated */
  int  flags;
  int  (*alloc)(mx_endpt_state_t *es, mx_instance_state_t *is,
                myri_set_endpt_t *set_endpt);
  int  (*enable)(mx_endpt_state_t *es, uaddr_t arg);
  int  (*ioctl)(mx_endpt_state_t *es, uint32_t cmd, uaddr_t in);
  int  (*mmap)(mx_endpt_state_t *es, unsigned long req, unsigned long *off,
               void **kva, int *mem_type, int *readonly, mx_page_pin_t **pin);
  int  (*poll_events_pending)(mx_endpt_state_t *es);
  void (*poll_prep_wait)(mx_endpt_state_t *es);
  int  (*poll_wait)(mx_endpt_state_t *es, int timeout_ms, int events);
  void (*interrupt_is)(mx_instance_state_t *is, void *context);
  int  (*interrupt_ep)(mx_endpt_state_t *es, int type);
  void (*close)(mx_endpt_state_t *es);
} myri_endpt_ops_t;

typedef struct mx_intr
{
  uint8_t seqnum;
  uint32_t spurious;
  uint32_t cnt;
  nic_intr_t *desc;
  mx_page_pin_t pin;
} mx_intr_t;

#ifdef MYRI_PHOENIX
#include "phx_snf_io.h"
#include "phx_snf_fwd.h"
#else /* MYRI_PHOENIX */
#include "myri_snf_io.h"
#include "myri_snf_fwd.h"
#endif /* !defined(MYRI_PHOENIX) */

typedef int (mx_init_board_t)(mx_instance_state_t *is);
typedef int (mx_map_board_t)(mx_instance_state_t *is);
typedef void (mx_unmap_board_t)(mx_instance_state_t *is);
typedef void (mx_claim_interrupt_t)(mx_instance_state_t *is, uint8_t endpt_irq, uint8_t endpt);
typedef int (mx_read_irq_level_t)(mx_instance_state_t *is);
typedef void (mx_enable_interrupt_t)(mx_instance_state_t *is);
typedef void (mx_disable_interrupt_t)(mx_instance_state_t *is);
typedef void (mx_park_board_t)(mx_instance_state_t *is);
typedef int (mx_detect_parity_t)(mx_instance_state_t *is);
typedef void (mx_get_freq_t)(mx_instance_state_t *is);
typedef uint32_t (mx_get_rtc_t)(mx_instance_state_t *is);

struct phx_counter;
struct phx_fpga_counter;
struct phx_fpga_endpoint_pool;
struct phx_fpga_instance;

typedef int (mx_get_counter_definitions_t)(mx_instance_state_t *is,
                                           const struct phx_counter **counters,
                                           uint32_t *count);
typedef int (mx_counter_accum_index_t)(mx_instance_state_t *is,
                                       struct phx_fpga_counter *ctr, int nth_endpoint);
typedef int (mx_init_endpoint_pool_t)(mx_instance_state_t *is,
                                      struct phx_fpga_endpoint_pool *pool);
typedef struct phx_physical_ep* (mx_ether_endpoint_t)(mx_instance_state_t *is);
typedef struct phx_physical_ep* (mx_sniffer_endpoint_t)(mx_instance_state_t *is);
typedef void (mx_sniffer_read_stats_t)(mx_instance_state_t *is, phx_snf_stats_t *x);
typedef void (mx_enable_hw_hash_t)(mx_instance_state_t *is, uint32_t rss_flags);
typedef int (mx_enable_next_ring_merge_t)(mx_instance_state_t *is, uint32_t portmask);
typedef void (mx_setup_hw_caps_t)(mx_instance_state_t *is);
typedef struct phx_physical_ep* (mx_physical_endpoint_t)(mx_instance_state_t *is,
                                                         int nth_endpoint);
typedef int (mx_set_intr_coal_t)(mx_instance_state_t *is, int microseconds);
typedef void (mx_set_mac_addr_t)(mx_instance_state_t *is, uint8_t mac[6]);
typedef int (mx_set_promisc_t)(mx_instance_state_t *is, int promisc);
typedef int (mx_enable_all_multicast_t)(mx_instance_state_t *is);
typedef int (mx_disable_all_multicast_t)(mx_instance_state_t *is);
typedef int (mx_remove_filter_t)(mx_instance_state_t *is,  myri_rx_filterq_entry_t *entry);
typedef int (mx_add_filter_t)(mx_instance_state_t *is,  myri_rx_filterq_entry_t *entry);

typedef void (mx_mac_set_addr_t)(mx_instance_state_t *is, uint8_t mac[6]);
typedef void (mx_mac_set_promisc_t)(mx_instance_state_t *is, int promisc);
typedef void (mx_mac_enable_t)(mx_instance_state_t *is);
typedef void (mx_mac_disable_t)(mx_instance_state_t *is);
typedef void (mx_mac_set_max_frame_length_t)(mx_instance_state_t *is, int mtu);

#ifdef MYRI_PHOENIX
typedef struct mx_board_ops {
  mx_init_board_t               *init;
  mx_map_board_t                *map;
  mx_unmap_board_t              *unmap;
  mx_enable_interrupt_t         *enable_interrupt;
  mx_disable_interrupt_t        *disable_interrupt;
  mx_park_board_t               *park;
  mx_get_counter_definitions_t  *get_counter_definitions;
  mx_counter_accum_index_t      *counter_accum_index;
  mx_init_endpoint_pool_t       *init_endpoint_pool;
  mx_ether_endpoint_t           *ether_endpoint;
  mx_sniffer_endpoint_t         *sniffer_endpoint;
  mx_sniffer_read_stats_t       *sniffer_read_stats;
  mx_enable_hw_hash_t           *enable_hw_hash;
  mx_enable_next_ring_merge_t   *enable_next_ring_merge;
  mx_setup_hw_caps_t            *setup_hw_caps;
  mx_physical_endpoint_t        *physical_endpoint;
  mx_set_intr_coal_t            *set_intr_coal;
  mx_set_mac_addr_t             *set_mac_addr;
  mx_set_promisc_t              *set_promisc;
  mx_enable_all_multicast_t     *enable_all_multicast;
  mx_disable_all_multicast_t    *disable_all_multicast;
  mx_remove_filter_t		*remove_filter;
  mx_add_filter_t		*add_filter;
} mx_board_ops_t;

extern mx_board_ops_t phx_e1_ops;
extern mx_board_ops_t phx_dbl4p_ops;
extern mx_board_ops_t phx_k35_ops;
extern mx_board_ops_t phx_k35snf_ops;

void phx_e1_init_board_ops(void);
void phx_dbl4p_init_board_ops(void);
void phx_k35_init_board_ops(void);
void phx_k35snf_init_board_ops(void);

int mx_select_board_type(mx_instance_state_t *is, int vendor_id, int device_id, int pci_revision);

typedef struct mx_mac_ops {
//  mx_get_counter_definitions_t  *get_counter_definitions;
//  mx_counter_accum_index_t      *counter_accum_index;
  mx_mac_set_addr_t             *mac_set_hwaddr;
  mx_mac_set_promisc_t          *mac_set_promisc;
  mx_mac_enable_t               *mac_enable;
  mx_mac_disable_t              *mac_disable;
  mx_mac_set_max_frame_length_t *mac_set_max_frame_length;
} mx_mac_ops_t;

extern mx_mac_ops_t phx_altera_mac_ops;
extern mx_mac_ops_t phx_xilinx_xmac_ops;
void phx_altera_init_mac_ops(void);
void phx_xilinx_init_xmac_ops(void);

#else /* MYRI_PHOENIX */
typedef struct mx_board_ops {
  mx_init_board_t 		*init;
  mx_map_board_t 		*map;
  mx_unmap_board_t 		*unmap;
  mx_claim_interrupt_t 		*claim_interrupt;
  mx_read_irq_level_t		*read_irq_level;
  mx_enable_interrupt_t 	*enable_interrupt;
  mx_disable_interrupt_t 	*disable_interrupt;
  mx_park_board_t		*park;
  mx_detect_parity_t		*detect_parity_error;
  mx_get_freq_t		        *get_freq;
  mx_get_rtc_t			*get_rtc;
} mx_board_ops_t;

extern mx_board_ops_t mx_lx_ops;
extern mx_board_ops_t mx_lz_ops;
void mx_lx_init_board_ops(void);
void mx_lz_init_board_ops(void);
void mx_lz_read_mac(mx_instance_state_t *is, char mac[18]);
void mx_lz_phys_id(mx_instance_state_t *is, int id_on);
#endif /* !defined(MYRI_PHOENIX) */
int mx_maccmp(const char *mac_a, const char *list_of_mac_b_substr);

#define MX_ADDRS_PER_VPAGE  (MX_VPAGE_SIZE / sizeof(nic_dma_addr_t))
#define MX_ADDRS_PER_VPAGE_SHIFT  (MX_VPAGE_SHIFT - MX_DMA_ADDR_SHIFT)

#define MX_NUM_DIRECT_DMA  4  /* XXXX move me elsewhere!!!! */


/* An mx_dma_table contains one page of DMA addresses.
   Lengths for each DMA are implicitly assumed to be VPAGE size.
   The MCP accesses this structure, so it must be contiguous in DMA space.  
   The table can be a pte table or a higher-level table */

struct mx_dma_table
{
  mx_page_pin_t pin;
  nic_dma_addr_t *desc;		/* pointer to DMA page */
  union {
    mx_page_pin_t *pins; /* DMA pin info for data pages */
    struct mx_dma_table **tables; /* tables of inferor lvel  */
  } u;
  int nb_entries;
  int nb_subtables;
  int log_level; /* 0=> table of data pages , n*MX_ADDRS_PER_VPAGE_SHIFT => n-higher-level table */
};

typedef struct mx_dma_table mx_dma_table_t;

/* An mx_host_dma_win_t is a container for one rdma window,
   it contains the page table associated with this window. */

struct mx_host_dma_win
{
  uint32_t flags;
  uint32_t length;
  uint32_t seqnum;
  struct mx_dma_table table;
};

#ifndef MYRI_PHOENIX
typedef struct myri_raw_tx_buf
{
  STAILQ_ENTRY(myri_raw_tx_buf) next;
  nic_dma_addr_t dma;
  void *buf;
  uint32_t id;
  uint64_t context;
} myri_raw_tx_buf_t;

STAILQ_HEAD(myri_raw_tx_bufq, myri_raw_tx_buf);

/* HW clock functionality */
int  myri_clock_init(mx_instance_state_t *is);
void myri_clock_destroy(mx_instance_state_t *is);
int  myri_clock_sync_now(mx_instance_state_t *is);
void myri_clock_intr(mx_instance_state_t *is);
#endif /* !defined(MYRI_PHOENIX) */

int myri_secs_to_year(uint32_t seconds);
uint32_t myri_year_to_secs(int year);

#if MYRI_ENABLE_PTP
enum myri_ptp_state {
  MYRI_PTP_STOPPED  = -1,
  MYRI_PTP_STARTING = 0,
  MYRI_PTP_RUNNING  = 1
};
#endif

/*
 * Definitions for RX filters
 */
struct myri_rx_filterq_entry
{
  TAILQ_ENTRY(myri_rx_filterq_entry) entries;
  myri_rx_filter_t filt;
  /* Each DBL filter may lead to 2 PHX filters (tstamp). */
  phx_public_filter_t phx_filt;
  phx_public_filter_t phx_filt2;
  uint32_t data[2];              /* dest MAC address */
};

TAILQ_HEAD(myri_rx_filterq_head, myri_rx_filterq_entry);

#ifdef MYRI_PHOENIX
int  phx_add_rx_filter(mx_endpt_state_t *es, myri_rx_filter_t *f);
int  phx_remove_rx_filter(mx_endpt_state_t *es, myri_rx_filter_t *f,
                          struct myri_rx_filterq_entry *entry);
#else
int  myri_add_rx_filter(mx_endpt_state_t *es, myri_rx_filter_t *f);
int  myri_remove_rx_filter(mx_endpt_state_t *es, myri_rx_filter_t *f,
                           struct myri_rx_filterq_entry *entry);
#endif

#define MX_IS_DEAD        (1 << 0)
#define MX_IS_LOGGING     (1 << 1)
#define MX_HOSTNAME_SET   (1 << 2)
#define MX_PARITY_RECOVERY (1 << 3)

#define MYRI_WAIT_RX_EVENT  1
#define MYRI_WAIT_TX_EVENT  2

struct ep_irq {
  mx_intr_t intr;
  struct mx_instance_state *is;
  mx_endpt_state_t **ep;
  int phx_endpt; /* n'th endpoint within the function */
};

#define MX_IS_MAX_CHILD_PORTS  3

struct mx_instance_state
{
  unsigned int id;            /* unit number.         */
  unsigned int flags;         /* Flags, defined above */
  mx_board_ops_t board_ops;

#ifdef MYRI_PHOENIX
  mx_mac_ops_t mac_ops;
  struct mx_instance_phx_state {
    /* Memory mapped Phoenix board */
    struct phx_nic_mem_global *nic_mem;
    char *i2c_base_mem; /* pointer to the start of the i2c cores */
    char *eeprom_strings;
    char *product_code;
    char *part_number;
    char *serial;
    /* FIXME.  Does this really belong here? */
    struct phx_apu_config apu_config;
    mx_sync_t filter_sync;
    struct {
      uint32_t bus;
      uint32_t dev;
      uint32_t func;
    } pci_info;
    uint32_t num_functions;
    struct phx_fpga_instance *pfi;
    phx_mac_filter_t l2_filters[PHX_MAC_FILTER_CNT];
    uint32_t num_l2_filters;
    /* phx_{start,stop}_rx_endpoint */
    uint32_t rx_desc_pages;
    uint32_t rx_data_pages;
    uint32_t port_merge;
  } phx;
  int discard_multicast_enabled;
#if MAL_OS_LINUX
  phx_class_d_discard_whitelist_t *class_d_discard_whitelist;
#endif
  int intr_coal_usec;
#else
  /*************
   * LANai access fields
   *************/

  /* Record of lanai mappings */
    
  struct
  {
    /* This pointer is opaque to ensure that the special registers
       are referenced only through the pio.h interface. */
    void *special_regs;
    volatile uint32_t *control_regs;
    volatile uint8_t *sram;
    
#if MX_DMA_DBG
    volatile uint32_t *dma_dbg_bitmap;
    volatile uint32_t dma_dbg_lowest_addr;
#endif
    uint32_t running;
    uint32_t public_off;
    char *eeprom_strings;
    char *product_code;
    char *part_number;
    char *serial;
  } lanai;

  /* Pointer to LANai interface card, mapped into kernel memory. */
#endif

  uint32_t board_span;
  uint32_t sram_size;
  uint32_t specials_size;
#ifdef MYRI_PHOENIX
  int irq_one_to_one_enabled;
  int irq_shared_enabled;
#else
  uint32_t using_msi:1;
  uint32_t msi_enabled:1;
  uint32_t msix_enabled:1;
#endif
  uint32_t cpu_freq;
  uint32_t pci_freq;
  mx_arch_instance_info_t arch; /* Architecture-specific field. */
  mx_sync_t sync;
  mx_endpt_state_t **es;
  int es_count;
  char is_name_default[8];
  const char *is_name;
  uint8_t mac_addr[8];
  uint8_t perm_mac_addr[8];
  char *mac_addr_string;

#ifdef MYRI_MSI
  struct {
    char *name;
    void *vectors;
  } msix;
#endif

  mx_intr_t intr;
  struct ep_irq *ep_irq;

  mx_spinlock_t cmd_spinlock;
  struct {
    mx_sync_t sync;
    uint32_t size;
  } logging;
  mx_sync_t init_sync;
  mx_atomic_t ref_count;
#ifdef MYRI_PHOENIX
  /* A skeleton raw endpoint for now */
  struct {
    mx_endpt_state_t *es;
    int use_count;
  } raw;
#else
  struct {
    mx_endpt_state_t *es;
    int use_count;
    mx_spinlock_t spinlock;
    mx_sync_t sync;
    int wakeup_needed;
    myri_raw_tx_buf_t *tx_bufs;
    struct myri_raw_tx_bufq tx_bufq;
    mx_spinlock_t tx_spinlock;
    mx_copyblock_t tx_cb;
    uint32_t tx_req;
    uint32_t tx_done;
    mcp_raw_desc_t *descs;
    mx_page_pin_t descs_pin;
    uint32_t desc_cnt;
  } raw;
#endif
  uint32_t num_ports;
  uint8_t pci_rev;
  uint16_t pci_devid;
  uint32_t bar0_low;
  uint32_t bar0_high;
  uint32_t bar2_low;
  uint32_t bar2_high;
  uint32_t msi_addr_low;
  uint32_t msi_addr_high;
  uint16_t msi_data;
  uint16_t msi_ctrl;
  uint16_t exp_devctl;
  uint32_t ze_print_pos;
  int board_type;
  uint32_t mcp_print_len;
  uint32_t mcp_print_addr;
  uint32_t mcp_print_idx;
#ifndef MYRI_PHOENIX
  char *mcp_print_buffer;
  uint32_t old_lanai_uptime;
  struct mx_ether *ether;
#else
  struct phx_ether *ether;
#endif
  int ether_is_open;
  struct {
    mx_sync_t wait_sync;
    uint32_t cycles;
    uint32_t count;
    int busy;
  } dmabench;
  int parity_errors_detected;
  int parity_errors_corrected;
  uintptr_t parity_critical_start;
  uintptr_t parity_critical_end;
  mx_sync_t ready_for_recovery_sync;
  struct {
    int command_tweaked;
    uint16_t command_offset;
    uint16_t command_orig_val;
  } pci_cap;
  uint32_t link_state;
  struct {
    int reason;
    int arg;
  } saved_state;
  uint32_t xcvr_present_last;
  uint32_t xcvr_present_state;
  struct {
    mx_page_pin_t pin;
    char *addr;
  } tx_done;
  struct {
    mx_page_pin_t pin;
    char *addr;
  } sysbox;
  /* All PHX counters are in the board-global FPGA structure. */
#ifndef MYRI_PHOENIX
  struct  {
    uint32_t seqnum;
    mx_spinlock_t spinlock; /* For accumulation */
    mx_sync_t sync;
    mx_page_pin_t pin;
    char *addr;      /* host DMA copy */
    uint32_t *mcp;   /* SRAM access (pio read) */
    uint64_t *accum; /* Accumulated counts in driver */
    uint32_t *accum_last;  /* Keep track of last value read */
    uint32_t cnt;
    uint32_t ep_cnt; /* counters that are repeated per endpoint */
  } counters;
#endif
  struct {
    mx_page_pin_t pin;
    char *addr;
  } bogus;
  mx_kernel_window_t *kernel_window;
  uint32_t *dma_dbg_bitmap_host;
  int manual_dispersion;
  uint32_t crc32_table[16];


  myri_snf_rx_state_t   snf;
#ifndef MYRI_PHOENIX
  void                 *snf_rx_doorbell;
#endif


#ifdef MYRI_PHOENIX
  /* FIXME.  Always pretend we use the local source.
   * No support for other time sources.
   */
  enum myri_timesource_state clk_ts_state;
#else /* MYRI_PHOENIX */
  /* NIC and host clock synchronization */
  mx_spinlock_t           clk_spinlock;
  enum myri_timesource_state clk_ts_state;
  uint32_t                clk_db_offset;
  int                     clk_hwclock_seqno;
  int                     clk_sync_init_cnt;
  uint64_t                clk_sync_init_nsecs;
  mcp_hwclock_update_t   *clk_update;
  myri_clksync_params_t   clkp_latest;
#if MYRI_ENABLE_PTP
  enum myri_ptp_state     ptp_state;
  mx_spinlock_t           ptp_state_lock;
  struct myri_ptp        *ptp;
#endif

  /* enable/disable use of timesource */
  int                     timesource_enable;
  
  /* data used in the synching of the timesource */
  struct {
    uint64_t nsecs;
    uint64_t cycles;
    int year;
    int attempts;
  } timesource_data;

#if MAL_OS_LINUX && defined MYRI_PPS_SUPPORT
  /* pps stuff */
  MYRI_PPS_TYPE pps_dev;
  uint64_t pps_last_second;
  int pps_timesource_good;
  struct timer_list pps_timer;
#endif /* MAL_OS_LINUX && MYRI_PPS_SUPPORT */
#endif /* !defined(MYRI_PHOENIX) */

  struct mx_instance_state *child_ports[MX_IS_MAX_CHILD_PORTS];
  int num_child_ports;
  int port_id;
  struct mx_instance_state *parent_port;
  uint64_t link_speed;
};

#define MX_IS_PARENT(is) ((is)->parent_port == NULL)
#define MX_IS_CHILD(is) ((is)->parent_port != NULL)
#define MX_IS_PORT_ID(is) ((is)->parent_port ? is->port_id : is->phx.pci_info.func)
#define MX_IS_PORT(is, port_id) (port_id ? (is)->child_ports[port_id - 1] : (is))
#define MX_IS_K35SNF(is) ((is)->board_type == MAL_BOARD_TYPE_PHX_K35SNF)

#define MX_RDMA_WIN_PER_PAGE (MX_VPAGE_SIZE/sizeof(mcp_rdma_win_t))

#define MX_ES_SOFT      (1 << 0)
#define MX_ES_CLOSING	(1 << 1)
#define MX_ES_RECOVERY	(1 << 2)
#define MX_ES_MCP_SESSION (1 << 3)
#define MX_ES_MCP_ENABLED (1 << 4)
#define MX_ES_EARLY_ENABLE (1 << 5) /* XXX: will be removed */

struct mx_endpt_state
{
  mx_instance_state_t *is;
  enum myri_endpt_type es_type;
  uint32_t endpt;
  uint32_t flags;
  uint32_t ref_count;  /* access protected by sync */
  mx_sync_t sync;
  myri_endpt_ops_t *eops;

  mx_sync_t wait_sync;
  mx_sync_t close_sync;
  mx_sync_t vq_sync;  /* prehash */

  myri_endpt_flow_t *flow_window;
  mx_page_pin_t flow_window_pin;	

  /* MD handle */
  mx_arch_endpt_info_t arch;

#ifdef MYRI_PHOENIX
  struct phx_physical_ep *phx_ep; /* RX and TX endpoints */

  /* The common queues for DBL/Lanai.
   *
   * user_mmaped_sram:
   *   ummap: mcp_endpt_ummap_t.  PIO buffer and flow control vars on NIC.
   *   It is 16KB per endpoint.  These flow control vars are for receive
   *   (eventq_host_flow and recvq_host_flow).
   *
   *   flow_window above is in kernel memory.
   *   - rx_flow: latest eventq_host_flow and recvq_host_flow
   *   - tx_flow: latest tx_compl_host
   *   These are here to support polling (i.e. sleep till tx or rx happens
   *   in the kernel).  The kernel needs to know about these values, so we
   *   create a page shared between the kernel and the user.
   *
   *   mx_instance_t has tx_done, which contains send completions for
   *   all endpoints.  Phoenix also needs flow_window and tx_done.
   *
   * user_mmaped_zreq:
   *   requests: mcp_ureq_t.  Send requests on NIC.
   * sendq:
   *   Send DMA buffer.  Many pages.
   * eventq:
   *   Not used by the user land.
   * recvq:
   *   Receive buffer.  mcp_uevt_dbl_t and packets all get stored here.
   *   This is mx_huge_copyblock_t, the Lanai's 3-level page table.
   *   Phoenix uses slightly different 3-level page table.  See comments above
   *   about mx_huge_copyblock_t.
   *
   * For Phoenix, we do not need eventq at all.  It is an MX thing.
   * All necessary queues (DMA buffer, PIO buffer, receive data, receive
   * descriptor) are in Phoenix DBL endpoint state to isolate the code.
   *
   * Affects:
   * mx_mmap_off_to_kva(), MYRI_GET_COPYBLOCKS, MYRI_SET_COPYBLOCK
   * mx_dma_map_copyblock(), mx_huge_copyblock functions.
   */
#else
  /* copyblocks */
  mx_copyblock_t user_mmapped_sram;  /* sram, not kernel mem */
  mx_copyblock_t user_mmapped_zreq;  /* mmio, not kernel mem */
  mx_copyblock_t sendq;
  mx_copyblock_t eventq;
  mx_huge_copyblock_t recvq;
  uint64_t recvq_size; /* requested size */
#endif

  uint32_t privileged;
  uint32_t is_kernel;
#ifdef _LINUX_UIDGID_H
  kuid_t euid;
#else
  uint32_t euid;
#endif
  myri_endpt_opener_t opener;
  int endpt_error;
  uint32_t session_id;

  /* If endpoint is pinned to a cpu, keep track of its cpuid and nodeid, if
   * applicable */
  int cpu_id;
  int node_id;

  /* filter queue */
  struct myri_rx_filterq_head rx_filterq;


#ifdef MYRI_PHOENIX
  union snf_endpt_state snf;
#endif

#ifndef MYRI_PHOENIX
  union snf_endpt_state snf;
#endif /* !defined(MYRI_PHOENIX) */

};

int mx_init_driver(void);
int mx_finalize_driver(void);

int  mx_instance_init (mx_instance_state_t *is, int32_t unit);
int  mx_instance_finalize (mx_instance_state_t *is);
#ifndef MYRI_PHOENIX
void mx_lanai_print(mx_instance_state_t *is, int idx);
int mx_load_mcp(mx_instance_state_t *is, void * img, int limit, int *actual_len);
#endif

void *mx_map_pci_space (mx_instance_state_t * is, int bar, uint32_t offset, uint32_t len);
void *mx_map_io_space (mx_instance_state_t * is, uint32_t offset, uint32_t len);
void  mx_unmap_io_space (mx_instance_state_t * is, uint32_t len, void *kaddr);

/* Return the number of child instances/ports belonging to this parent instance. */
int mx_instance_child_ports(mx_instance_state_t *parent);
int mx_instance_init_child(mx_instance_state_t *parent, mx_instance_state_t *child,
                           int unit);

extern mx_instance_state_t **mx_instances;
extern uint32_t myri_max_instance;
extern uint32_t mx_num_instances;
extern uint32_t mx_num_child_instances;
extern uint32_t myri_mx_small_message_threshold;
extern uint32_t myri_mx_medium_message_threshold;
extern uint32_t myri_security_disabled;
extern uint32_t myri_z_loopback;
extern uint32_t myri_serdes_mode;
extern int mx_has_xm;
extern char * myri_mac;
extern uint32_t myri_pbkey;
extern uint32_t myri_mx_counters_64bit;

#ifdef MYRI_PHOENIX
/* Module param to disable the kernel ethernet endpoint */
extern int myri_disable_eth;

/* Module param to enable the ipv4 multicast discard feature */
/* now used as a bitmask to steer individual ports , 0 = off*/
extern uint32_t myri_enable_ipv4_multicast_discard_bitmask;

/* Module parameter to make vsc7224/vsc7111 setup flexible */
extern uint32_t myri_phx_serdes_mode;

/* Module param to push the host clock to the NIC periodically */
extern int myri_host_clock_sync_to_nic;

/* Send completion threshold */
extern int myri_tx_threshold;

/* Coalesing maximum defaults */
extern int myri_rx_coalesce_pkt;
extern int myri_rx_coalesce_time;
extern int myri_rx_coalesce_len;
extern int myri_k35snf_rx_timelimit;

/* Limit # of packets in phx_ether_rx_done() */
extern int myri_ether_rx_done_limit;

/* Maximum read request size (512 by default) */
extern int myri_pcie_mrr;

/* Kernel ethernet endpoint's ring sizes in MB */
extern unsigned myri_rx_data_ring_mb;
extern unsigned myri_rx_desc_ring_mb;
#endif

#if MAL_OS_MACOSX || MAL_OS_SOLARIS
#define MX_MAX_INSTANCE_DEFAULT 4
#else
#define MX_MAX_INSTANCE_DEFAULT 32
#endif

extern uint32_t myri_mx_max_nodes;
extern uint32_t myri_recvq_vpage_cnt;
extern uint32_t myri_eventq_vpage_cnt;

extern uint32_t myri_intr_coal_delay;

extern mx_atomic_t myri_max_user_pinned_pages;
extern uint32_t mx_cacheline_size;
extern uint32_t myri_override_e_to_f;
extern uint32_t myri_parity_recovery;
extern uint32_t myri_recover_from_all_errors;
extern uint32_t myri_pcie_down;
extern uint32_t myri_pcie_down_on_error;
extern uint32_t myri_dma_pfn_max;
extern uint32_t myri_force;
extern uint32_t myri_jtag;
extern uint32_t myri_mx_max_macs;
extern uint32_t myri_msi_level;
extern uint32_t myri_ether_pause;
extern uint32_t myri_idt_war;
extern uint32_t myri_clksync_period;
extern uint32_t myri_clksync_error;
extern uint32_t myri_clksync_error_count;
extern uint32_t myri_clksync_verbose;
extern uint32_t myri_timesource_loss_mode;
extern uint32_t myri_clk_enable_10mhz;
extern uint32_t myri_clk_enable_pps;
extern uint32_t myri_clk_invert_pps;

extern uint32_t myri_snf_rings;
extern uint32_t myri_snf_flags;
extern uint32_t myri_snf_dmasend_threshold;

extern mx_spinlock_t mx_lanai_print_spinlock;

extern char mx_default_hostname[MYRI_MAX_STR_LEN];

/* sleeping and waking up */

#define MX_SLEEP_INTR    1
#define MX_SLEEP_NOINTR  2

/* for the watchdog to properly detect parity error (rather than such
   error ending up as some MX_DEAD_xxx_TIMEOUT), the final timeout
   should be:
   final-timeout >= (min-desired-timeout + 2*Watchdog)
   in addition to sram parity error, this avoid confusing root cause
   and secondary effects for any HW check done in the watchdog_body
*/

#define MX_SMALL_WAIT    100
#define MX_CLOSE_WAIT	 1000*30 /* 30 seconds */
#define MX_LOGGING_WAIT  MX_MAX_WAIT

#if MAL_OS_UDRV
#define MCP_INIT_TIMEOUT	(mx_lxgdb ? 600*1000 : 12*1000)
#define MCP_COMMAND_TIMEOUT	(mx_lxgdb ? 600*1000 : 10)
#define MX_WATCHDOG_TIMEOUT	(mx_lxgdb ? 10 : 5)
#define MX_LXGDB		mx_lxgdb
#else
/* FIXME: do we need 10s timeout ? */
#define MCP_INIT_TIMEOUT	12*1000
#define MCP_COMMAND_TIMEOUT	10
#define MX_WATCHDOG_TIMEOUT	1
#define MX_LXGDB		0
#endif


void mx_wake(mx_sync_t *);
int mx_wake_once(mx_sync_t *); /* returns 1 if notify happened */
int mx_sleep(mx_sync_t *, int, int flags);
mx_waitq_t *mx_new_waitq(void);
void mx_free_waitq(mx_waitq_t *);

/* syncs.. need to move elsewhere */

void mx_sync_init (mx_sync_t *s, mx_instance_state_t *is, int unique, 
		   char *string);
void mx_sync_reset (mx_sync_t *s);
void mx_sync_destroy (mx_sync_t *s);

#ifndef mx_spin_lock_destroy
void mx_spin_lock_destroy(mx_spinlock_t *s);
#endif

void mx_spin_lock_init(mx_spinlock_t *s, mx_instance_state_t *is, int unique, 
		       char *string);

#ifndef mx_mutex_enter
void mx_mutex_enter (mx_sync_t *s);
int  mx_mutex_try_enter (mx_sync_t *s); /* returns 0 if successful */
void mx_mutex_exit (mx_sync_t *s);
#endif

extern mx_sync_t mx_global_mutex;

/* open/close/ioctl .. need to move elsewhere */


int mx_common_ioctl(mx_endpt_state_t *es, uint32_t cmd, const uaddr_t in);

int mx_endptless_ioctl(uint32_t cmd, const uaddr_t in,
		       uint32_t privileged, uint32_t is_kernel);

int myri_common_open_args(int cmd, uaddr_t arg, myri_set_endpt_t *);

int myri_common_open(int32_t unit, mx_endpt_state_t *es, 
                     myri_set_endpt_t *set_endpt, uaddr_t arg);

void mx_common_close (mx_endpt_state_t *es);

void mx_unpin_page(mx_instance_state_t *, mx_page_pin_t *, int);
void mx_unpin_vpages(mx_instance_state_t *, mx_page_pin_t *, int, int);
int mx_pin_page(mx_instance_state_t *, mx_page_pin_t *, int, uint64_t);
int mx_pin_vpages(mx_instance_state_t *, mx_page_pin_t *, nic_dma_addr_t *, int, int, uint64_t);

/* PCI accessor functions */
int mx_read_pci_config_8 (mx_instance_state_t *is,
			  uint32_t offset,
			  uint8_t *val);

int mx_read_pci_config_16 (mx_instance_state_t *is,
			   uint32_t offset,
			   uint16_t *val);

int mx_read_pci_config_32 (mx_instance_state_t *is,
			   uint32_t offset,
			   uint32_t *val);

int mx_write_pci_config_8 (mx_instance_state_t *is,
			   uint32_t offset,
			   uint8_t val);

int mx_write_pci_config_16 (mx_instance_state_t *is,
			    uint32_t offset,
			    uint16_t val);

int mx_write_pci_config_32 (mx_instance_state_t *is,
			    uint32_t offset,
			    uint32_t val);


#if MAL_OS_UDRV
void mx_write_lanai_isr(struct mx_instance_state *is, uint32_t val);
void mx_write_lanai_ctrl_bit(struct mx_instance_state *is, int val);
#else
#define mx_write_lanai_isr(is,val) mx_write_lanai_special_reg_u32(is,lx.ISR, val)
#define mx_write_lanai_ctrl_bit(is,bit) mx_write_pci_config_32(is, 0x40, 1U << (bit))
/* or equivalent: is->lanai.control_regs[0x10] = htonl(1 << MX_LX_CTRL(bit)))) */
#endif

#define MX_MEM_HOSTMEM 0
#define MX_MEM_SRAM 1
#define MX_MEM_CONTROL 2
#define MX_MEM_SPECIAL 3
#define MX_MEM_HOSTMEM_UNMAPPED 4

int mx_mmap_off_to_kva(mx_endpt_state_t *, unsigned long, void **, int *, int *,
		       mx_page_pin_t **);

#ifdef MYRI_PHOENIX
void k35snf_common_interrupt(mx_instance_state_t *is);
void k35snf_common_msix_ep_irq(struct ep_irq *emi);
int phx_endptless_ioctl(uint32_t cmd, const uaddr_t in,
                        uint32_t privileged, uint32_t is_kernel);
int phx_common_interrupt(mx_instance_state_t *is);
#ifdef MYRI_MSI
int phx_common_msix_ep_irq(struct ep_irq *emi);
#endif
#else /* MYRI_PHOENIX */
int mx_mcp_command(mx_instance_state_t * is, enum mcp_cmd_type type, 
		   uint32_t data0, uint32_t data1, uint32_t data2, 
		   uint32_t *result);
  
int mx_common_interrupt(mx_instance_state_t *is);

#ifdef MYRI_MSI
int mx_common_msix_ep_irq(struct ep_irq *emi);
#endif
#endif /* !defined(MYRI_PHOENIX) */

/* memory registration related functionality */
void mx_free_dma_win (mx_instance_state_t *is, mx_host_dma_win_t *hc);
mx_host_dma_win_t *mx_new_dma_win(mx_instance_state_t *is, int nsegs);

static inline int
mx_is_dead(mx_instance_state_t *is)
{
  return (is->flags & (MX_IS_DEAD | MX_PARITY_RECOVERY));
}
#define mx_digit(c) (((c) >= '0' && (c) <= '9') ? ((c) - '0') : \
                 (((c) >= 'A' && (c) <= 'F') ? (10 + (c) - 'A') : \
                  (((c) >= 'a' && (c) <= 'f') ? (10 + (c) - 'a') : -1)))

int mx_alloc_dma_page(mx_instance_state_t *is, char **addr, mx_page_pin_t *pin);

void mx_free_dma_page(mx_instance_state_t *is, char **addr, mx_page_pin_t *pin);
int mx_alloc_zeroed_dma_page(mx_instance_state_t *is, char **addr, mx_page_pin_t *pin);
#if MAL_OS_LINUX
int mx_alloc_zeroed_dma_page_node(mx_instance_state_t *is, char **addr, mx_page_pin_t *pin, int node_id);
#else
#define mx_alloc_zeroed_dma_page_node mx_alloc_zeroed_dma_page
#endif
int mx_alloc_zeroed_dma_page_relax_order(mx_instance_state_t *is, char **addr, mx_page_pin_t *pin);

#ifdef MX_HAS_RELAX_ORDER_DMA
int mx_alloc_dma_page_relax_order(mx_instance_state_t *is, char **addr, mx_page_pin_t *pin);
#else
#define mx_alloc_dma_page_relax_order mx_alloc_dma_page
#define mx_alloc_copyblock_relax_order mx_alloc_copyblock
#endif

int mx_rand(void);
void mx_watchdog_body(void);


int mx_dma_map_copyblock(mx_endpt_state_t *es, mx_copyblock_t *cb, uint32_t offset);

int mx_common_alloc_copyblock(mx_instance_state_t *is, mx_copyblock_t *cb);
void mx_common_free_copyblock(mx_instance_state_t *is, mx_copyblock_t *cb);
#ifndef MYRI_PHOENIX
void mx_parse_mcp_error(mx_instance_state_t *is);
#endif
#if MAL_OS_LINUX
void *mx_map_huge_copyblock(mx_huge_copyblock_t *hcb);
void mx_unmap_huge_copyblock(mx_huge_copyblock_t *hcb);
#else
#define mx_map_huge_copyblock(hcb) NULL
#define mx_unmap_huge_copyblock(hcb) do { } while (0)
#endif


/* copyin and out */
static inline int
mx_copyin(uaddr_t what, void *where, size_t amount, uint32_t is_kernel)
{
  if (is_kernel) {
    bcopy((void *)(uintptr_t)what, where, amount);
    return 0;
  } else
    return mx_arch_copyin(what, where, amount);
}

static inline int
mx_copyout(void *what, uaddr_t where, size_t amount, uint32_t is_kernel)
{
  if (is_kernel) {
    bcopy(what, (void *)(uintptr_t)where, amount);
    return 0;
  } else
    return mx_arch_copyout(what, where, amount);
}

int mx_instance_status_string(int unit, char **str, int *len);

void mx_set_default_hostname(void);

#if MAL_OS_LINUX
void myri_update_numa_config(mx_endpt_state_t *es);
#else
static inline void myri_update_numa_config(mx_endpt_state_t *es)
{
  es->cpu_id = -1;
  es->node_id = -1;
}
#endif


mx_endpt_state_t *mx_get_endpoint(mx_instance_state_t *is, int endpt);
void mx_put_endpoint(mx_endpt_state_t *es);
int mx_direct_get(mx_endpt_state_t *dst_es, mx_shm_seg_t *dst_segs, uint32_t dst_nsegs,
		  mx_endpt_state_t *src_es, mx_shm_seg_t *src_segs, uint32_t src_nsegs,
		  uint32_t length);
mx_instance_state_t *mx_get_instance(uint32_t unit);
void mx_release_instance(mx_instance_state_t *is);
#ifndef MYRI_PHOENIX
void mx_mark_board_dead(mx_instance_state_t *is, int reason, int arg);
#endif
void mx_parse_eeprom_strings(mx_instance_state_t *is);

int mx_direct_get_common(mx_shm_seg_t *dst_segs, uint32_t dst_nsegs,
			 void * src_space, mx_shm_seg_t *src_segs, uint32_t src_nsegs,
			 uint32_t length);

/* OS specific callback for direct get, copying from another process
 * user-space to current process user-space.
 */
int mx_arch_copy_user_to_user(uaddr_t udst, uaddr_t usrc, void * src_space,
			      uint32_t len);


int mx_pcie_link_reset(mx_instance_state_t *is);
int mx_intr_recover(mx_instance_state_t *is);
uint16_t mx_bridge_pci_sec_status(mx_instance_state_t *is);
uint32_t mx_bridge_id(mx_instance_state_t *is);
int mx_find_capability(mx_instance_state_t *is, unsigned cap_id);

int myri_select_mcp(int board_type, const unsigned char **mcp,
		    int *len, unsigned int *parity_critical_start,
		    unsigned int *parity_critical_end);
#ifndef MYRI_PHOENIX  
int myri_counters(int board_type, const char ***counters, uint32_t *count,
                  const uint32_t **cclass);
int  myri_counters64_get(mx_instance_state_t *is, uint64_t *c_o, int num_counters,
                         enum myri_counters_get_type type);
#endif

void myri_counters64_update(mx_instance_state_t *is);
void myri_counters64_clear(mx_instance_state_t *is);
void myri_counters_uptime_init(mx_instance_state_t *is);

int myri_check_license(mx_instance_state_t *is, uint32_t *valid);

int
mx_alloc_huge_copyblock(mx_instance_state_t *is, mx_huge_copyblock_t *hcb,
			unsigned long lib_offset, int node_id);

void
mx_free_huge_copyblock(mx_instance_state_t *is, mx_huge_copyblock_t *hcb);


int myri_atoi(char *s);
#ifndef MYRI_PHOENIX
int myri_endpt_get_common_copyblocks(const mx_endpt_state_t *es, mx_get_copyblock_t *desc,
                                     uint32_t *offp);
int myri_endpt_enable_common_copyblocks(mx_endpt_state_t *es);
int myri_endpt_enable_mcp(mx_endpt_state_t *es);
#endif

#if MX_KERNEL_LIB
#define mx_endpt_alloc_copyblock(es,cb)					\
  ((es)->is_kernel ? mx_common_alloc_copyblock((es)->is,cb)		\
   : mx_alloc_copyblock((es)->is,cb))
#define mx_endpt_alloc_copyblock_relax_order(es,cb)			\
  ((es)->is_kernel ? mx_common_alloc_copyblock((es)->is,cb)		\
   : mx_alloc_copyblock_relax_order((es)->is,cb))
#define mx_endpt_free_copyblock(es,cb)			   \
  ((es)->is_kernel ? mx_common_free_copyblock((es)->is,cb) \
   : mx_free_copyblock((es)->is,cb))
#define mx_endpt_alloc_huge_copyblock(es,cb)				\
   mx_alloc_huge_copyblock((es)->is,cb,0,-1)
#define mx_endpt_alloc_huge_copyblock_relax_order(es,cb)		\
   mx_alloc_huge_copyblock_relax_order((es)->is,cb,0)
#define mx_endpt_free_huge_copyblock(es,cb)			\
   mx_free_huge_copyblock((es)->is,cb)
#else
#define mx_endpt_alloc_copyblock(es,cb) mx_alloc_copyblock((es)->is,cb)
#define mx_endpt_alloc_copyblock_relax_order(es,cb) mx_alloc_copyblock_relax_order((es)->is,cb)
#define mx_endpt_free_copyblock(es,cb) mx_free_copyblock((es)->is,cb)
#if MAL_OS_WINDOWS
int mx_win_alloc_huge_copyblock(mx_instance_state_t *is, 
				mx_huge_copyblock_t *hcb,
				unsigned long lib_offset,
				int node_id);
void mx_win_free_huge_copyblock(mx_instance_state_t *is,
				mx_huge_copyblock_t *hcb);
#define mx_endpt_alloc_huge_copyblock(es,cb) mx_win_alloc_huge_copyblock((es)->is,cb,0,0)
#define mx_endpt_alloc_huge_copyblock_relax_order(es,cb) mx_win_alloc_huge_copyblock_relax_order((es)->is,cb,0,0)
#define mx_endpt_free_huge_copyblock(es,cb) mx_win_free_huge_copyblock((es)->is,cb)
#else
#define mx_endpt_alloc_huge_copyblock(es,cb) mx_alloc_huge_copyblock((es)->is,cb,0,-1)
#define mx_endpt_alloc_huge_copyblock_relax_order(es,cb) mx_alloc_huge_copyblock_relax_order((es)->is,cb,0)
#define mx_endpt_free_huge_copyblock(es,cb) mx_free_huge_copyblock((es)->is,cb)
#endif
#endif



#if MAL_OS_WINDOWS
void phx_get_miniport_stats(mx_instance_state_t *is, phx_miniport_stats_t *stats);
#endif

int myri_ether_soft_rx(mx_instance_state_t *is, myri_soft_rx_t *r);

#endif /* _mx_instance_h_ */
