/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2011 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

#include "mx_arch.h"
#include "mx_instance.h"
#include "kraw.h"

myri_endpt_ops_t myri_ops_kraw;

static
int
kraw_eop_alloc(mx_endpt_state_t *es, mx_instance_state_t *is, myri_set_endpt_t *set_endpt)
{
  int status;

  /* can have just one open of raw message api */
  if (is->raw.use_count != 0)
    return EBUSY;

  /* and the endpoint must be privileged */
  if ((es->privileged | myri_security_disabled) == 0)
    return EPERM;
  es->is = is;
  is->raw.use_count++;

  mx_mutex_exit(&is->sync);
  status = myri_kraw_init(is);
  mx_mutex_enter(&is->sync);

  if (status) {
    MX_WARN(("%s: failed to setup raw interface. Status %d\n", 
	     is->is_name, status));
    is->raw.es = 0;
    es->is = 0;
    is->raw.use_count--;
    return status;
  }

  /* Raw is a soft endpoint */
  es->flags |= MX_ES_SOFT;
  is->raw.es = es;
  es->session_id = 0;
  return 0;
}

static
int
kraw_eop_enable(mx_endpt_state_t *es, uaddr_t arg)
{
  /* Raw is enabled at allocation time */
  return 0;
}

static
void
kraw_eop_intr_is(mx_instance_state_t *is, void *context)
{
  if (is->raw.es != NULL)
    myri_kraw_intr(is);
}

static
int
kraw_eop_ioctl(mx_endpt_state_t *es, uint32_t cmd, uaddr_t in)
{
  uaddr_t out = in;
  int status;

  if (es->es_type != MYRI_ES_RAW)
    return EINVAL;

  switch (cmd) {

  case MYRI_RAW_GET_NEXT_EVENT:
    {
      myri_raw_next_event_t x;
      if ((status = mx_copyin(in, &x, sizeof(x), es->is_kernel)))
        break;
      status = myri_kraw_next_event(es, &x);
      if (!status)
	mx_copyout(&x, out, sizeof(x), es->is_kernel);
    }
    break;

  case MYRI_RAW_SEND:
    {
      myri_raw_send_t x;
      if ((status = mx_copyin(in, &x, sizeof(x), es->is_kernel)))
        break;
      status = myri_kraw_send(es, &x);
    }
    break;

  default:
    status = ENOTTY;
  }

  return status;
}

static
void
kraw_eop_close(mx_endpt_state_t *es)
{
  mx_instance_state_t *is = es->is;

  myri_kraw_tx_wait_pending(is);

  mx_mutex_enter(&is->sync);
  is->raw.es = NULL;
  mx_mutex_exit(&is->sync);

  myri_kraw_destroy(is);

  mx_mutex_enter(&is->sync);
  is->raw.use_count--;
  mx_mutex_exit(&is->sync);
}

void
myri_kraw_init_ops(void)
{
  bzero(&myri_ops_kraw, sizeof(myri_ops_kraw));
  myri_ops_kraw.alloc = kraw_eop_alloc;
  myri_ops_kraw.interrupt_is = kraw_eop_intr_is;
  myri_ops_kraw.enable = kraw_eop_enable;
  myri_ops_kraw.ioctl = kraw_eop_ioctl;
  myri_ops_kraw.close = kraw_eop_close;
}

