/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2014-2016 by CSP, Inc.  All rights reserved.                *
 *************************************************************************/

#include "mx_arch.h"
#include "mx_misc.h"
#include "mx_instance.h"
#include "mx_malloc.h"
#include "mx_pio.h"
#include "mal_stbar.h"
#include "myri_version.h"
#include "mal_cpu.h"
#include "phx_hw.h"
#include "phx_common.h"
#include "phx_avl.h"

/* Temporary until we have a more general logging/verbosity control */
#if MAL_DEBUG
#define _VERBOSE 1
#else
#define _VERBOSE 0
#endif

mx_mac_ops_t phx_altera_mac_ops;

void altera_mac_disable(mx_instance_state_t *is)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + PHX_MAC_BASE;

#if _VERBOSE
  MX_INFO(("%s: Turning off MAC tx/rx transfer control\n", __FUNCTION__));
#endif
  PHX_REG_WRITE_4((r + PHX_MAC_TX_CONTROL), 1);
  PHX_REG_WRITE_4((r + PHX_MAC_RX_CONTROL), 1);
}

void altera_mac_enable(mx_instance_state_t *is)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + PHX_MAC_BASE;

#if _VERBOSE
  MX_INFO(("%s: Turning on MAC tx/rx transfer control\n", __FUNCTION__));
#endif
  PHX_REG_WRITE_4((r + PHX_MAC_TX_CONTROL), 0);
  PHX_REG_WRITE_4((r + PHX_MAC_RX_CONTROL), 0);
}

void altera_mac_set_max_frame_length(mx_instance_state_t *is, int mtu)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + PHX_MAC_BASE;
#if _VERBOSE
  uint32_t tx_max, rx_max;

  MX_INFO(("%s:\n", __FUNCTION__));
  tx_max = PHX_REG_READ_4((r + PHX_MAC_TX_MAX_FRAME_LENGTH));
  rx_max = PHX_REG_READ_4((r + PHX_MAC_RX_MAX_FRAME_LENGTH));
  MX_INFO(("current: tx_max=%u rx_max=%u\n", tx_max, rx_max));
#endif

#ifdef PHX_WORKAROUND_MAC_OVERSIZE_FRAMES
  /* Set lengths to be some large so the NIC never sees
   * oversized frames.
   */
  PHX_REG_WRITE_4((r + PHX_MAC_TX_MAX_FRAME_LENGTH), 10000);
  PHX_REG_WRITE_4((r + PHX_MAC_RX_MAX_FRAME_LENGTH), 10000);
#else
  /* "Length" in the Altera MAC covers [Ethernet header ... CRC32].
   * The MAC understands VLAN tagged packets and uses length + 4 when
   * checking the VLAN packet's length.
   *
   * TX length does not affect transmission at all.  It is only used
   * for statistics in the MAC.
   */
  /* Add 14 (ethernet header) and 4 (crc32) to the OS MTU. */
  PHX_REG_WRITE_4((r + PHX_MAC_TX_MAX_FRAME_LENGTH), (mtu + 18));
  PHX_REG_WRITE_4((r + PHX_MAC_RX_MAX_FRAME_LENGTH), (mtu + 18));
#endif

#if _VERBOSE
  tx_max = PHX_REG_READ_4((r + PHX_MAC_TX_MAX_FRAME_LENGTH));
  rx_max = PHX_REG_READ_4((r + PHX_MAC_RX_MAX_FRAME_LENGTH));
  MX_INFO(("current: tx_max=%u rx_max=%u\n", tx_max, rx_max));
#endif
}

void altera_mac_set_promisc_mode(mx_instance_state_t *is, int enable)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + PHX_MAC_BASE;
  uint32_t control;

  if (enable) {
    control = (PHX_MAC_RX_FRAME_CONTROL_EN_ALLUCAST
               | PHX_MAC_RX_FRAME_CONTROL_EN_ALLMCAST);
  } else {
    /* We do not do multicast filtering using the Altera MAC.
     * So always accept multicast packets.
     */
    control = PHX_MAC_RX_FRAME_CONTROL_EN_ALLMCAST;
  }
  PHX_REG_WRITE_4((r + PHX_MAC_RX_FRAME_CONTROL), control);
}

void altera_mac_set_address(mx_instance_state_t *is, uint8_t mac[6])
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + PHX_MAC_BASE;
  uint32_t addr0, addr1;
 
#if _VERBOSE
  MX_INFO(("%s: %02x:%02x:%02x:%02x:%02x:%02x\n", __FUNCTION__,
           mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]));
#endif
  addr1 = ((uint32_t)mac[0] << 8) | ((uint32_t)mac[1]);
  addr0 = ((uint32_t)mac[2] << 24) | ((uint32_t)mac[3] << 16)
      | ((uint32_t)mac[4] << 8) | ((uint32_t)mac[5]);
  PHX_REG_WRITE_4((r + PHX_MAC_PRIMARY_MAC_ADDR0), addr0);
  PHX_REG_WRITE_4((r + PHX_MAC_PRIMARY_MAC_ADDR1), addr1);
}

void 
phx_altera_init_mac_ops(void)
{
//  mx_get_counter_definitions_t  *get_counter_definitions;
//  //  mx_counter_accum_index_t      *counter_accum_index;
  phx_altera_mac_ops.mac_set_hwaddr             = altera_mac_set_address;
  phx_altera_mac_ops.mac_set_promisc            = altera_mac_set_promisc_mode;
  phx_altera_mac_ops.mac_enable                 = altera_mac_enable;
  phx_altera_mac_ops.mac_disable                = altera_mac_disable;
  phx_altera_mac_ops.mac_set_max_frame_length   = altera_mac_set_max_frame_length;
}
