/******************************************************************************
*                                                                             *
* License Agreement                                                           *
*                                                                             *
* Copyright (c) 2003 Altera Corporation, San Jose, California, USA.           *
* All rights reserved.                                                        *
*                                                                             *
* Permission is hereby granted, free of charge, to any person obtaining a     *
* copy of this software and associated documentation files (the "Software"),  *
* to deal in the Software without restriction, including without limitation   *
* the rights to use, copy, modify, merge, publish, distribute, sublicense,    *
* and/or sell copies of the Software, and to permit persons to whom the       *
* Software is furnished to do so, subject to the following conditions:        *
*                                                                             *
* The above copyright notice and this permission notice shall be included in  *
* all copies or substantial portions of the Software.                         *
*                                                                             *
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  *
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    *
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE *
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER      *
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     *
* FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         *
* DEALINGS IN THE SOFTWARE.                                                   *
*                                                                             *
* This agreement shall be governed in all respects by the laws of the State   *
* of California and by the laws of the United States of America.              *
*                                                                             */
/**
 * Library:
 *
 * File:        i2c.c
 *
 * Desc:        I2C functions for OpenCores I2C IP
 *
 *  Provided by
 *  -----------
 *      BittWare, Inc.
 *      45 South Main St. Suite L100
 *      Concord, NH 03301
 *      Ph:     (603) 226 0404
 *      Fax:    (603) 226 6667
 *      WWW:    http://www.bittware.com
 *      Email:  support@bittware.com
 *
 * Copyright ? 2011, BittWare, Inc.
 *
 * License:
 *
 * The user is hereby granted a non-exclusive license to use
 * and or modify this software provided that it runs on BittWare
 * hardware. Usage of this software on non-BittWare hardware
 * without the express written permission of BittWare is strictly
 * prohibited.
 *
**/

/*
* Modified:                                                                   *
*   9/29/14 G. Hannon (ghannon@cspi.com) incorporated all necessary altera    *
*   routines into this one file for kernel inclusion.                         *
*   12/29/15 G. Hannon (ghannon@cspi.com) incorporated routines for Xilinx    *
*   fpga based board differences.                                             *
******************************************************************************/


#include "mal_auto_config.h"
#include "mx_arch.h"
#include "mx_misc.h"
#include "mx_instance.h"
#include "mal_stbar.h"
#include "phx_avl.h"
#include "phx_common.h"
#include "phx_firmware_versions.h"

/* FIXME... */
#if MAL_OS_WINDOWS
extern void phx_win_usleep(int usec);
#endif

static void
microsecond_sleep(int usec)
{
#if MAL_OS_LINUX
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 0, 0)
	usleep_range(usec, usec + 200);
#else
	udelay(usec);
#endif
#elif MAL_OS_WINDOWS
	phx_win_usleep(usec);
#else
#error "Not implemented yet"
#endif
}

/*
 * SPI Interface Routines
 */  

/* This is a very simple routine which performs one SPI master transaction.
 * It would be possible to implement a more efficient version using interrupts
 * and sleeping threads but this is probably not worthwhile initially.
 */

int avl_spi_command(char* base, uint32_t slave,
                           uint32_t write_length, const uint8_t * write_data,
                           uint32_t read_length, uint8_t * read_data,
                           uint32_t flags)
{
	const uint8_t * write_end = write_data + write_length;
	uint8_t * read_end = read_data + read_length;

	uint32_t write_zeros = read_length;
	uint32_t read_ignore = write_length;
	uint32_t status;

	/* We must not send more than two bytes to the target before it has
	* returned any as otherwise it will overflow. */	
	/* Unfortunately the hardware does not seem to work with credits > 1,
	* leave it at 1 for now. */
	
	int32_t credits = 1;

	/* Warning: this function is not currently safe if called in a multi-threaded
	* environment, something above must perform locking to make it safe if more
	* than one thread intends to use it.
	*/
	IOWR_ALTERA_AVALON_SPI_SLAVE_SEL(base, 1 << slave);
  
	/* Set the SSO bit (force chipselect) only if the toggle flag is not set */
	if ((flags & ALT_AVALON_SPI_COMMAND_TOGGLE_SS_N) == 0) {
		IOWR_ALTERA_AVALON_SPI_CONTROL(base, ALTERA_AVALON_SPI_CONTROL_SSO_MSK);
	}

	/*
	* Discard any stale data present in the RXDATA register, in case
	* previous communication was interrupted and stale data was left
	* behind.
	*/
	IORD_ALTERA_AVALON_SPI_RXDATA(base);
    
	/* Keep clocking until all the data has been processed. */
	for ( ; ; )
	{
		do
		{
			status = IORD_ALTERA_AVALON_SPI_STATUS(base);
		}
		while (((status & ALTERA_AVALON_SPI_STATUS_TRDY_MSK) == 0 || credits == 0) &&
		       (status & ALTERA_AVALON_SPI_STATUS_RRDY_MSK) == 0);

		if ((status & ALTERA_AVALON_SPI_STATUS_TRDY_MSK) != 0 && credits > 0)
		{
			credits--;

			if (write_data < write_end)
				IOWR_ALTERA_AVALON_SPI_TXDATA(base, *write_data++);
			else if (write_zeros > 0)
			{
				write_zeros--;
				IOWR_ALTERA_AVALON_SPI_TXDATA(base, 0);
			}
			else
				credits = -1024;
		};

		if ((status & ALTERA_AVALON_SPI_STATUS_RRDY_MSK) != 0)
		{
			uint32_t rxdata = IORD_ALTERA_AVALON_SPI_RXDATA(base);

			if (read_ignore > 0)
				read_ignore--;
			else
				*read_data++ = (uint8_t)rxdata;
			credits++;

			if (read_ignore == 0 && read_data == read_end)
				break;
		}
	}

	/* Wait until the interface has finished transmitting */
	do
	{
		status = IORD_ALTERA_AVALON_SPI_STATUS(base);
	}
	while ((status & ALTERA_AVALON_SPI_STATUS_TMT_MSK) == 0);

	/* Clear SSO (release chipselect) unless the caller is going to
	* keep using this chip
	*/
	if ((flags & ALT_AVALON_SPI_COMMAND_MERGE) == 0)
		IOWR_ALTERA_AVALON_SPI_CONTROL(base, 0);

	return read_length;
}


/*
 * PIO interface routines
 */

int avl_pio_init(char* base)
{
        // This is not needed unless the cores were built as bi-directional 
        // inputs 
        IOWR_ALTERA_AVALON_PIO_DIRECTION(base +  AVL_BASE_PIO_I, 0x0000);
        // outputs
        IOWR_ALTERA_AVALON_PIO_DIRECTION(base + AVL_BASE_PIO_O, 0xffff);

        return 0;
}

int avl_pio_read(char* base, uint32_t mask, uint32_t* value)
{
        uint32_t gpio;

        gpio = IORD_ALTERA_AVALON_PIO_DATA(base + AVL_BASE_PIO_I);
        *value = (gpio & mask);

        return 0;
}

int avl_pio_write(char* base, uint32_t mask, uint32_t value)
{
        // I would like to use set and clear bits, 
        // but I don't know that these are built into the core
        IOWR_ALTERA_AVALON_PIO_DATA(base + AVL_BASE_PIO_O, (value & mask));
        return 0;
}


/*
 * I2C routines
 */

// Initialize one channel of I2C, setting up the prescale
int avl_i2c_init(char* base, uint32_t avl_base, uint8_t high, uint8_t low)
{
        uint8_t cmd;
        uint32_t pre_lo, pre_hi;

        fpga_read32s(base, avl_base + I2C_CTR_LO, &pre_lo, 1);
        fpga_read32s(base, avl_base + I2C_CTR_HI, &pre_hi, 1);

        if ((pre_lo != low) && (pre_hi != high)) {  
                /* disable the core */
                cmd = I2C_DISABLE;
                fpga_write8s(base, avl_base + I2C_CONTROL, &cmd, 1);

                 /* set the prescale */
                cmd = low;
                fpga_write8s(base, avl_base + I2C_CTR_LO, &cmd, 1);
                cmd = high;
                fpga_write8s(base, avl_base + I2C_CTR_HI, &cmd, 1);

        }

        cmd = I2C_ENABLE;
        fpga_write8s(base, avl_base + I2C_CONTROL, &cmd, 1);

        return 0;
}

int avl_i2c_write(char* base, uint32_t avl_base, uint8_t i2c_dat, uint8_t i2c_cmd)
{
        int timeout = 400;
        uint8_t stat;

        fpga_write8s(base, avl_base + I2C_RX_TX, &i2c_dat, 1);
        fpga_write8s(base, avl_base + I2C_CMD_STAT, &i2c_cmd, 1);

        do {
                microsecond_sleep(100);
                fpga_read8s(base, avl_base + I2C_CMD_STAT, &stat, 1);
                if(timeout-- <= 0)
			return -1;
//                        fatal_error("Timeout in avl_i2c_write address 0x%x stat = 0x%x\n", 
//                                    (i2c_core_offset+avl_base), stat);
        } while((stat & 0x2) == 0x2);

        return 0;
}

int avl_i2c_read_bytes(char* base, uint32_t avl_base, uint8_t device, uint8_t reg_ofs, uint8_t* bytes, uint8_t count)
{
        uint8_t i;

        avl_i2c_write(base, avl_base, device, I2C_START_WR);   // I2C Device Address
        avl_i2c_write(base, avl_base, reg_ofs, I2C_STOP_WR);     // Register Address
        avl_i2c_write(base, avl_base, device + 1, I2C_START_WR); // I2C Device Address + READ
        for(i = 0; i < count; i++)
        {
                if( i == (count-1))
                        avl_i2c_write(base, avl_base, 0x00, I2C_STOP_RD_NACK);  //last read  
                else
                        avl_i2c_write(base, avl_base, 0x00, I2C_RD);

                fpga_read8s(base, avl_base + I2C_RX_TX, &bytes[i], 1);
        }

        return count;
}

int at24c64_i2c_read_bytes(char *base, uint32_t avl_base, uint8_t device, uint16_t reg_ofs, uint8_t* bytes, uint8_t count)
{
        uint8_t i;
        uint8_t high, low;

        high = reg_ofs / 256;
        low = reg_ofs % 256;

        avl_i2c_write(base, avl_base, device, I2C_START_WR);   // I2C Device Address  
        avl_i2c_write(base, avl_base, high, I2C_WR);           // Register Address
        avl_i2c_write(base, avl_base, low, I2C_STOP_WR);       // Register Address
        avl_i2c_write(base, avl_base, device + 1, I2C_START_WR); // I2C Device Address + READ
        for(i = 0; i < count; i++)
        {
                if( i == (count-1))
                        avl_i2c_write(base, avl_base, 0x00, I2C_STOP_RD_NACK);  //last read
                else
                        avl_i2c_write(base, avl_base, 0x00, I2C_RD);

                fpga_read8s(base, avl_base + I2C_RX_TX, &bytes[i], 1);
        }
        return count;
}

int avl_i2c_write_bytes(char* base, uint32_t avl_base, uint8_t device, uint8_t reg_ofs, uint8_t* bytes, uint8_t count)
{
        uint8_t i;

        avl_i2c_write(base, avl_base, device, I2C_START_WR);   // I2C Device Address
        if (count == 0)
                avl_i2c_write(base, avl_base, reg_ofs, I2C_STOP_WR);     // Used for 8 way MUX
        else
                avl_i2c_write(base, avl_base, reg_ofs, I2C_WR);          // Register Address

        for(i = 0; i < count; i++)
        {
                if( i == (count-1))
                        avl_i2c_write(base, avl_base, bytes[i], I2C_STOP_WR);    // Last
                else
                        avl_i2c_write(base, avl_base, bytes[i], I2C_WR);
        }
        return count;
}

int avl_i2c_set_mux(char* base, uint8_t device, int bus)
{
        uint8_t mux_ctrl;
        int rc = 0;

        if (device == KCU105_DEV_MUXx4){
                if (bus == 0)
                        mux_ctrl = 0;   // disable all
                else
                        mux_ctrl = ((bus-1) & 0x03) | 0x04; // enable downstream bus
                rc = avl_i2c_write_bytes(base, AVL_BASE_I2C_0, device, 0, &mux_ctrl, 1);
        }
        else if ((device == KCU105_DEV_MUXx8) || (device == K35_DEV_MUXx8)){
                if (bus == 0)
                        mux_ctrl = 0;   // disable all
                else
                        mux_ctrl = 1 << (bus-1);  // enable downstream bus
                rc = avl_i2c_write_bytes(base, AVL_BASE_I2C_0, device, mux_ctrl, &mux_ctrl, 0);
        }

        microsecond_sleep(100);
        return rc;
}

int phx_sha1_to_fw_version(uint32_t sha1){
    int i;

    for (i=0; i < MAX_FIRMWARE_VERSIONS; i++){
      if (known_versions[i].fw_sha1 == sha1)
        return i+1;
    }
    return MAX_FIRMWARE_VERSIONS;
}

int phx_avl_init(char* base, uint32_t* version)
{
        int rtn=0;
        uint32_t avl_reg;
        uint32_t fw_sha1;

        /* convert the sha1 in the firmware version register to an index */
        fpga_read32s(base, AVL_SHA1_VER, &fw_sha1, 1);
        *version = (uint32_t) phx_sha1_to_fw_version(ntohl(fw_sha1));

        fpga_read32s(base, AVL_BASE_ID, &avl_reg, 1);

        /* ensure no one is using i2c channels as init may change speeds */
        do {} while (!phx_i2c_mutex_claim(base)); 

        if (avl_reg == AVL_ID_REGVAL){
		/*  Enable I2C Interface Cores in FPGA */
		avl_i2c_init(base, AVL_BASE_I2C_0, 0x00, 0x4e);
		avl_i2c_init(base, AVL_BASE_I2C_1, 0x00, 0x4e);
		avl_i2c_init(base, AVL_BASE_I2C_2, 0x00, 0x4e);
		avl_i2c_init(base, AVL_BASE_I2C_3, 0x00, 0x4e);
		avl_i2c_init(base, AVL_BASE_I2C_4, 0x00, 0x4e);
		avl_i2c_init(base, AVL_BASE_I2C_5, 0x00, 0x4e);
		avl_i2c_init(base, AVL_BASE_I2C_6, 0x00, 0x4e);
	}
 
        phx_i2c_mutex_release(base);

        return(rtn);
}

int phx_xilinx_init(char* base, uint32_t* version, uint32_t* xlx_board)
{
        int rtn=0;
        uint32_t fpga_rev;
	uint32_t board_arch;

        fpga_read32s(base, AVL_SHA1_VER, &fpga_rev, 1);
        *version = (uint32_t) phx_sha1_to_fw_version(ntohl(fpga_rev));

        fpga_read32s(base, XLX_ASCII_ID_1, &board_arch, 1);
	*xlx_board = board_arch;

        /* ensure no one is using i2c channels as init may change speeds */
        do {} while (!phx_i2c_mutex_claim(base)); 

        /* set the prescale to 100KHz (156.25M / (5 * 100K) - 1 = 0x0137 */
        avl_i2c_init(base, AVL_BASE_I2C_0, 0x01, 0x37);

	if (board_arch == K35_ID_REGVAL){
                avl_i2c_set_mux(base, K35_DEV_MUXx8, 0);
	}
        else if (board_arch == KCU105_ID_REGVAL)
	{
                avl_i2c_set_mux(base, KCU105_DEV_MUXx4, 0);
                avl_i2c_set_mux(base, KCU105_DEV_MUXx8, 0);
	}

        phx_i2c_mutex_release(base);

        return(rtn);
}


/*
  This file uses MX driver indentation.

  Local Variables:
  c-file-style:"linux"
  tab-width:8
  indent-tabs-mode:t
  End:
*/

