/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2014-2016 by CSP, Inc.  All rights reserved.                *
 *************************************************************************/

#ifndef _phx_common_h_
#define _phx_common_h_

#define myri_phx(tx) tx.phx


#define PHX_REG_READ_8(addr) (*((volatile uint64_t*)addr))
#define PHX_REG_READ_4(addr) (*((volatile uint32_t*)addr))
#define PHX_REG_READ_2(addr) (*((volatile uint16_t*)addr))
#define PHX_REG_READ_1(addr) (*((volatile uint8_t*)addr))

#define PHX_REG_WRITE_1(addr, data) do {        \
    (*((volatile uint8_t*)addr) = data);        \
    MAL_STBAR();                                \
  } while(0)

#define PHX_REG_WRITE_2(addr, data) do {        \
    (*((volatile uint16_t*)addr) = data);       \
    MAL_STBAR();                                \
  } while(0)

#define PHX_REG_WRITE_4(addr, data) do {        \
    (*((volatile uint32_t*)addr) = data);       \
    MAL_STBAR();                                \
  } while(0)

#define PHX_REG_WRITE_8(addr, data) do {        \
    (*((volatile uint64_t*)addr) = data);       \
    MAL_STBAR();                                \
  } while(0)

/* These define which firmware version a feature was first introduced in.
 * See the known versions table in phx_firmware_versions.h for
 * the mapping of SHA1 checksums to firmware versions.
 */
#define PHX_FEATURE_MULTICAST_DISCARD 24
#define PHX_FEATURE_TIMEBASE_V2       28
#define PHX_FEATURE_INTR_INTERVAL     30
#define PHX_FEATURE_SP_REGISTERS      35
#define PHX_FEATURE_8K_PIO            36
#define PHX_FEATURE_MAC_FILTER        37
#define PHX_FEATURE_PRESENCE_DETECT   58
#define PHX_FEATURE_FPGA_MERGE_NEXT_RING 60

#define PHX_HAS_FEATURE(is, feat_rev) (is->phx.pfi->firmware_rev >= feat_rev)

/* The next two values need to be adjusted together.  If we want the
   number of packets between dma of completion counts to be greater
   than 1, then we need to take into account how the number of packets
   will affect how many PIOs will fit in the 4096 bytes in the pio
   buffer.  For example, if we set the TX_THRESHOLD to 8, then only 7
   512-byte packets can fit in the pio buffer (due to overhead and
   alignment issues an 8th one won't fit) which can cause us to lockup
   because we are waiting for a completion update to tell us when we
   can send the 8th packet but the completion count will only update
   after an 8th packet has been sent.  We need to transition to using
   DMA at a size that will guarantee we don't lock up.  

   Addendum:  The number of pages used for tx dma also needs to be 
   adjusted to allow at least PHX_TX_THRESHOLD number of packets
   of max size to fit.  So don't forget to look at 
   PHX_TX_DMA_PAGES_PER_ENDPOINT.
*/

/* number of packets before dma completion count will update */
#define PHX_TX_THRESHOLD 7
#define K35_TX_THRESHOLD 4
/* the size after which we trasition from pio to dma sends */
#define PHX_DMASEND_THRESHOLD_SIZE 512
#define K35_DMASEND_THRESHOLD_SIZE 2048

#define PHX_RX_COALESCE_PKTS   15
#define PHX_RX_COALESCE_TIME   10000
#define PHX_RX_COALESCE_LENGTH (32*1024)  /* libsnf assumes at most 64KB per descriptor */
#define K35SNF_RX_TIME_LIMIT_DEFAULT 1024

/* structure for defining counters for phoenix */

typedef struct phx_counter {
  const char *str;
  uint32_t offset;
  uint32_t class;   /* Global, aliased (per-port), or endpoint */
  uint32_t stride;  /* For endpoint counters */
  uint32_t port_stride;  /* K35SNF, per-port counters */
} phx_counter_t;

/*
 * The Phoenix card has a set number of sharable A/B arbitration units.
 * If A/B is enabled, then duplicate packets are dropped by the hardware. 
 * Each unit can take a number of IP address/port combinations, which provide redundant
 * copies of the same UDP packet streams.  The unit attempts to only send up one copy of
 * each packet, using something called a tablex, that records which sequence numbers have been
 * received
 *
 * The ip addresses and port numbers are stored in the kernel so that programs can retrieve
 * the information even after an application has closed the endpoints and free'd the filters
 * used by the arbitration units.
 *
 */
/* structure to hold information regarding A/B TableX unit usage */

typedef struct phx_tablex {
  struct myri_rx_filterq_entry *filters[PHX_STREAM_MAX];
  uint32_t ip_addr[PHX_STREAM_MAX];
  uint16_t port[PHX_STREAM_MAX];
  int32_t feed_unit;     /* -1 means available */
  uint32_t filter_count;
} phx_tablex_t;

struct phx_fpga_counter {
  int id;
  uint32_t class;
  uint32_t offset;
  uint32_t stride;
  uint32_t port_stride;     /* Per-port registers (K35SNF) */
  uint64_t *accum;          /* Accumulated counts in driver */
  uint64_t *accum_last_64;  /* Keep track of last value read for 64 bit counters */
};

typedef struct phx_physical_ep {
  int id;  /* Logical ID used within the driver*/
  mx_instance_state_t *owner_is; /* current owner of this endpoint */
  uint32_t reserved;  /* reserved for a specific purpose */
  uint32_t rx_ep;     /* core endpoint number for DBLRx and associated cores */
  uint32_t tx_ep;     /* core endpoint number for DBLTx core */
  uint32_t func;      /* which function(s) may use it */
  uint32_t icu;       /* ICU vector */
  int port_id;        /* current port using this endpoint (K35SNF) */
  int kernel;         /* K35SNF port 0's endpoint being used by the kernel */
} phx_physical_ep_t;

#define INVALID_EP 0xffffffffUL

struct phx_fpga_endpoint_pool {
  int num_function_endpoints; /* Max endpoints per function (port) */
  int num_function_not_reserved_endpoints;
  int num_total_endpoints;  /* FPGA total endpoints */
  struct phx_physical_ep *endpoints;
};

struct phx_fpga_instance {
  int id;
  /* solely used to identify the fpga, so something else, like serial
     could also be used here */
  uint32_t bus;
  uint32_t dev;
  int num_functions; /* PCI functions (ports) */
  int attach_count;
  int initialized; /* Initialized the FPGA? */
#if MAL_OS_WINDOWS
  /* We do phx_avl_init() early on, before init_board().
   * See phx_win_init_fpga().
   */
  int avl_initialized;
#endif
  int firmware_rev;

  /* HW hash capabilities and current status. These may be manipulated with
   * mx_fpga_mutex acquired, as SNF may need to enable HW hash on multiple
   * FPAGs (NICs).
   */
  struct {
    uint32_t supported_flags; /* HW hash features: snf.h rss_flags */
    uint32_t *enabled_flags; /* Hash flags (per-port) */
    int *in_use; /* SNF_RX endpoints using this port's HW hash (per-port) */
  } hw_hash;

  /* Port merging mode. */
  struct {
    uint32_t mode_flags;
  } port_merge;

  /* stuff that needs to be shared between instances */
  struct {
    char *addr;
    mx_page_pin_t pin;
    int ref_count;
  } tx_cpage;    /* dma page for tx completions */

#if PHX_ENABLE_PTP
  struct ptp_clock_info      ptp_info;
  struct ptp_clock   *ptp_clock;
#endif

  mx_instance_state_t *is[MX_MAX_INSTANCE_DEFAULT];

  phx_tablex_t tablex_entries[PHX_TABLEX_MAX];  /* table_X cores are shared by all ports */

  char *eeprom_strings;         /* common copy of eeprom strings */
  uint32_t xilinx_board_type;   /* saved board type used for i2c routing */

  struct phx_fpga_endpoint_pool ep_pool;

  struct  {
    int ref_count;
    uint32_t seqnum;
    mx_spinlock_t spinlock; /* For accumulation */
    uint32_t  defined_cnt;
    struct phx_fpga_counter *counters;
    uint64_t *storage;

    /* Per-endpoint software counters for NIC endpoints.
     * Array size = myri_max_endpoints
     */
    phx_endpt_flow_tx_stats_t *ep_tx_stats;

    /* Endpoints with live counters (i.e. opened but not closed).
     * These are protected by counters.es_spinlock to avoid using
     * counters.spinlock or is->sync while updating software counters.
     */
    mx_spinlock_t es_spinlock;
    mx_endpt_state_t **es; /* Array size = total endpoints */
  } counters;

  /* The K35 has a bigger FPGA with more RAM, so we have increased the size of the PIO buffer, 
   * and moved it and made the register space smaller - to handle 32 endpoints someday. */
  struct {
    uint32_t dbltx_core_base;
    uint32_t dbltx_core_stride;
    uint32_t dbltx_core_length;
    uint32_t dbltx_pio_base;
    uint32_t dbltx_pio_stride;
    uint32_t dbltx_pio_length;
    uint32_t dmasend_threshold;
  } tx_config;

  /* information relating to adapter clock state */
  struct {
    uint32_t last_pps_state;
    uint32_t last_10mhz_state;
    uint64_t last_update_host_time;
    uint64_t pps_lost_host_time;
    uint64_t pps_good_host_time;
    mx_spinlock_t clk_spinlock;
  } clk;

};

extern struct phx_fpga_instance *phx_fpga_instances;
extern uint32_t phx_fpga_num_instances;

enum phx_clk_source {
  PHX_CLK_PPS,
  PHX_CLK_10MHZ,
  PHX_CLK_FPGA
};

/* This probably belongs in a mal_ header in common rather than duplicated  */
#define COUNTER_TYPE_MASK	0x0000ff00
#define COUNTER_TYPE_SHIFT	8

#define COUNTER_TYPE_GLOBAL     0x0000
#define COUNTER_TYPE_SW         0x0001
#define COUNTER_TYPE_DRV_UPTIME 0x0002
#define COUNTER_TYPE_UPTIME     0x0004
#define COUNTER_TYPE_2_32BIT    0x0008 /* 64 bits from 2 x 32-bit counters */
#define COUNTER_TYPE_48BIT_ROLL 0x0010 /* 48 bits rolling */
#define COUNTER_TYPE_ENDPOINT   0x0100
#define COUNTER_TYPE_64BIT      0x0200
#define COUNTER_TYPE_SNAPSHOT   0x0400
#define COUNTER_TYPE_SELFCLEAR  0x0800
#define COUNTER_TYPE_16_HI      0x1000
#define COUNTER_TYPE_16_LO      0x2000
#define COUNTER_TYPE_ALIASED    0x4000
#define COUNTER_TYPE_TX         0x8000  /* TX endpoint counters */

/* Software counters in phx_endpt_flow_tx_stats_t.
 * The offset indicates the n'th field.
 */
#define COUNTER_TYPE_TX_FLOW_STATS  0x800

#define COUNTER_DISPLAY_MASK	0x00ff0000
#define COUNTER_DISPLAY_SHIFT   16

#define COUNTER_DISPLAY_DECIMAL 0x00000
#define COUNTER_DISPLAY_HEX     0x10000
#define COUNTER_DISPLAY_VERBOSE 0x20000
#define COUNTER_DISPLAY_VB2     0x40000
#define COUNTER_DISPLAY_UPTIME  0x80000

#define CNT_GLOBAL       (COUNTER_TYPE_GLOBAL | COUNTER_DISPLAY_DECIMAL)
#define CNT_GLOBAL_64    (COUNTER_TYPE_GLOBAL | COUNTER_DISPLAY_DECIMAL | COUNTER_TYPE_64BIT)
#define CNT_GLOBAL_HEX   (COUNTER_TYPE_GLOBAL | COUNTER_DISPLAY_HEX)
#define CNT_ENDPOINT     (COUNTER_TYPE_ENDPOINT | COUNTER_DISPLAY_DECIMAL)
#define CNT_ENDPOINT_64  (COUNTER_TYPE_ENDPOINT | COUNTER_TYPE_64BIT)
#define CNT_ENDPOINT_HEX (COUNTER_TYPE_ENDPOINT | COUNTER_DISPLAY_HEX)
#define CNT_SW_ENDPOINT  (COUNTER_TYPE_ENDPOINT | COUNTER_TYPE_TX_FLOW_STATS)
#define CNT_DRV_UPTIME   (COUNTER_TYPE_GLOBAL | COUNTER_TYPE_SW | COUNTER_TYPE_DRV_UPTIME | COUNTER_DISPLAY_UPTIME)
#define CNT_UPTIME       (COUNTER_TYPE_GLOBAL | COUNTER_TYPE_SW | COUNTER_TYPE_UPTIME | COUNTER_DISPLAY_UPTIME)

/* 
 * Aliased versions - means same address is read, but in the FPGA it is a different 
 * register based on which PCI "function" of the card is accessing it.
 */
#define CNT_ALIASED      (COUNTER_TYPE_ALIASED | COUNTER_DISPLAY_DECIMAL)
#define CNT_ALIASED_64   (COUNTER_TYPE_ALIASED | COUNTER_DISPLAY_DECIMAL | COUNTER_TYPE_64BIT)
#define CNT_ALIASED_48   (COUNTER_TYPE_ALIASED | COUNTER_DISPLAY_DECIMAL | COUNTER_TYPE_48BIT_ROLL)
#define CNT_ALIASED_HEX  (COUNTER_TYPE_ALIASED | COUNTER_DISPLAY_HEX)
#define CNT_ALIASED_ENDPOINT  (COUNTER_TYPE_ALIASED | COUNTER_TYPE_ENDPOINT | COUNTER_DISPLAY_DECIMAL)

/* PIO copy */
void phx_pio_memcpy(void *dst, const void *src, int len);

/* NIC register accessors */

void phx_bbsysctl_pio_print_debug_info(struct phx_nic_mem_global *nic);

/* Multi-channel APU */
void phx_apu_config(struct phx_nic_mem_global *nic,
                    struct phx_apu_config *conf);
int phx_apu_channel_id(mx_instance_state_t *is, int data, int endpoint, int tx);
void phx_apu_channel_reset(mx_instance_state_t *is, int channel);
void phx_apu_channel_start(mx_instance_state_t *is, int channel,
                           uint32_t low, uint32_t high);
void phx_apu_channel_print_debug(struct phx_nic_mem_global *nic, int channel);

/* RX */
void phx_dbl_rx_packet_count_limit(mx_instance_state_t *is, uint32_t pkt);
void phx_dbl_rx_time_limit(mx_instance_state_t *is, uint32_t nanosecs);
void phx_dbl_rx_length_limit(mx_instance_state_t *is, uint32_t bytes);
void phx_dbl_rx_global_start(mx_instance_state_t *is);
void phx_dbl_rx_global_stop(mx_instance_state_t *is);
void phx_rx_data_cons_advance(mx_instance_state_t *is, int endpoint, uint32_t pages);
void phx_rx_desc_cons_advance(mx_instance_state_t *is, int endpoint, uint32_t pages);
uint32_t phx_dbl_rx_const0(struct phx_nic_mem_global *nic);
void phx_dbl_rx_print_debug(struct phx_nic_mem_global *nic);
void phx_start_rx_endpoint(mx_instance_state_t *is, int endpoint, uint32_t data_pages,
                           uint32_t desc_pages);
void phx_stop_rx_endpoint(mx_instance_state_t *is, int endpoint);
void phx_reset_rx_endpoint(mx_instance_state_t *is, int endpoint);

/* Filters */
void phx_dbl_filter_init(mx_instance_state_t *is, uint32_t kernel_endpoint);
void phx_dbl_filter_set_kernel_endpoint(mx_instance_state_t *is, uint32_t kernel_endpoint);
void phx_dbl_filter_discard_multicast(mx_instance_state_t *is, int discard);
void phx_dbl_filter_print_debug(struct phx_nic_mem_global *nic);

void phx_get_filter_state(mx_instance_state_t *is, phx_public_filter_state_t *filters);
void phx_get_mac_filters(mx_instance_state_t *is, phx_get_mac_filters_t *filters);

/* A/B Tables */
void phx_dbl_ab_init(struct phx_nic_mem_global *nic);
void phx_dbl_ab_local_init(struct phx_nic_mem_global *nic);
#ifdef MYRI_DBL_AB
void phx_ab_tablex_init(mx_instance_state_t *is);
int  phx_ab_seq_set(mx_endpt_state_t *es, myri_dbl_ab_seq_set_t ss);
int  phx_ab_tablex_get(mx_instance_state_t *is, myri_dbl_ab_tablex_t *tablex);
int  phx_ab_autoreset_init(mx_endpt_state_t *es, myri_dbl_ab_autoreset_t reset);
#endif

/* DBL TX */
uint32_t phx_read_ep_reg_tx_desc_buffer_size(struct phx_nic_mem_endpoint *ep);
void phx_write_ep_reg_tx_addr_buffer_clear(struct phx_nic_mem_endpoint *ep);
void phx_write_ep_reg_tx_addr_load(struct phx_nic_mem_endpoint *ep, uint32_t low, uint32_t high);
void phx_write_ep_reg_tx_addr_buffer_finish(struct phx_nic_mem_endpoint *ep);
void phx_tx_wake_request(struct phx_nic_mem_endpoint *ep, uint32_t host_count);

void phx_write_reg_tx_host_completion(struct phx_nic_mem_global *nic, uint32_t low, uint32_t high);

void phx_tx_numdwperpush(struct phx_nic_mem_global *nic, uint32_t dw);
void phx_tx_threshold(struct phx_nic_mem_global *nic, uint32_t pkts);
uint32_t phx_ep_tx_status(struct phx_nic_mem_endpoint *ep);
uint32_t phx_ep_tx_completion_count(struct phx_nic_mem_endpoint *ep);
void phx_tx_enable_auto_read_request(struct phx_nic_mem_global *nic);

/* Endpoint region */
struct phx_nic_mem_endpoint *phx_endpoint_nic_mem(mx_instance_state_t *is, int endpoint);

/* PIO buffer address */
void *phx_ep_pio_fifo(mx_instance_state_t *is, int endpoint);

/* Descript buffer address */
phx_tx_desc_t *phx_ep_desc_fifo(struct phx_nic_mem_endpoint *ep);

/* eeprom accessors */
void phx_read_specs(char *base, char *specs, size_t len, uint32_t xlx_board);
void phx_read_license(char* base, char *lic, size_t len, uint32_t xlx_board);
int phx_start_dmabench(mx_instance_state_t *is, mx_dmabench_t *x, mx_page_pin_t *pin);
int phx_finish_dmabench(mx_instance_state_t *is, mx_dmabench_t *x);

int myri_counters(int board_type, const phx_counter_t **counters, uint32_t *count,
                  const uint32_t **cclass);

int phx_counters64_get(mx_instance_state_t *is, uint64_t *c_o, int start_off,
                       int num_counters, int type);
void myri_counters64_ep_close(mx_instance_state_t *is, int endpoint);
void phx_counters_tx_fixup(mx_instance_state_t *is);

/* Attach the endpoint's software counters to the instance */
void phx_sw_counters_ep_open(mx_endpt_state_t *es);

/* Detach the endpoint's software counters from the instance */
void phx_sw_counters_ep_close(mx_endpt_state_t *es);

/* Rx Selector */
void phx_rx_selector_reset(struct phx_nic_mem_global *nic);
void phx_rx_selector_bypass_disable(struct phx_nic_mem_global *nic, uint32_t flags);

/* ICU */
uint8_t phx_icu_get_last_interrupt_seqnum(struct phx_nic_mem_global *nic);
/* For endpoints, the interrupt number is endpoint number + 1.
 * For global events, the interrupt number is 0.
 */
void phx_icu_enable_interrupt(mx_instance_state_t *is, uint32_t intr_num);
void phx_icu_disable_interrupt(mx_instance_state_t *is, uint32_t intr_num);
void phx_icu_reset_endpt(mx_instance_state_t *is, const phx_physical_ep_t *ep);
void phx_icu_request_link_change_interrupt(struct phx_nic_mem_global *nic);
uint8_t phx_icu_read_link_status(struct phx_nic_mem_global *nic);
void phx_print_intr_desc(phx_intr_t *desc);

/* 128-bit bitstream version */
void phx_read_bitstream_sha1(struct phx_nic_mem_global *nic, uint32_t val[4]);

void phx_clk_init(mx_instance_state_t *is, int invert_pps, int disable_pps, 
		  int disable_10mhz);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0)
void phx_clk_set(mx_instance_state_t *is, const struct timespec64 *ts);
void phx_clk_set_with_pps(mx_instance_state_t *is, const struct timespec64 *ts);
void phx_clk_get(mx_instance_state_t *is, struct timespec64 *ts);
#else
void phx_clk_set(mx_instance_state_t *is, const struct timespec *ts);
void phx_clk_set_with_pps(mx_instance_state_t *is, const struct timespec *ts);
void phx_clk_get(mx_instance_state_t *is, struct timespec *ts);
#endif
void phx_clk_adjtime(mx_instance_state_t *is, int64_t delta);
void phx_clk_adjfreq(mx_instance_state_t *is, int32_t ppb);
void phx_clk_monitor(mx_instance_state_t *is);
int phx_clk_is_pps_active(mx_instance_state_t *is);
int phx_clk_is_10mhz_active(mx_instance_state_t *is);
uint32_t phx_clk_get_pps_status(mx_instance_state_t *is);

void phx_attach_fpga_instance(mx_instance_state_t *is);
void phx_detach_fpga_instance(mx_instance_state_t *is);
int phx_alloc_fpga_tx_cpage(mx_instance_state_t *is);
void phx_free_fpga_tx_cpage(mx_instance_state_t *is);
int phx_alloc_fpga_counters(mx_instance_state_t *is);
void phx_free_fpga_counters(mx_instance_state_t *is);
void phx_get_fpga_ports(mx_instance_state_t *is, myri_get_all_ports_t *units);
uint32_t phx_get_dmasend_threshold(mx_instance_state_t *is);
/* the following used to maintain status of who is using which
   endpoint from the fpga endpoint pool */
int phx_claim_fpga_endpoint(mx_instance_state_t *is, int endpoint, int reserved);
void phx_release_fpga_endpoint(mx_instance_state_t *is, int endpoint);
int phx_fpga_endpoint_owner_instance_id(mx_instance_state_t *is, int endpoint);
int __phx_write_filter_regs(struct phx_nic_mem_global *nic, mx_sync_t *filter_sync,
			    myri_rx_filter_t *f, uint32_t endpt, uint8_t flags,
                            phx_public_filter_t *save);

int phx_i2c_mutex_claim(uint8_t *i2c_base);
void phx_i2c_mutex_release(uint8_t *i2c_base);

/* specialized handling for Xilinx MACs */
uint32_t phx_get_transceiver_type(mx_instance_state_t *is, uint32_t port);
void phx_reset_marvell_88E1111(mx_instance_state_t *is, uint32_t port);
void phx_set_serdes_preemphasis(mx_instance_state_t *is, uint32_t port, uint32_t value);

#if MAL_OS_WINDOWS
/* Really hack to read the MAC address from string specs, before we do
 * mx_instance_init.
 */
int phx_win_init_fpga(mx_instance_state_t *is);
#endif

int phx_parse_eeprom_strings(mx_instance_state_t *is);
void phx_icu_enable(mx_instance_state_t *is);

/* phx_vitesse.c */
int phx_vitesse_init(mx_instance_state_t *is, int first_call);
int phx_vitesse_exit(mx_instance_state_t *is);

#if MAL_OS_LINUX
/* driver/linux/phx.c */
void phx_ptp_init(struct phx_fpga_instance *fpga);
void phx_k35_check_pps(mx_instance_state_t *is);
#endif

/* Common board ops */
int phx_common_map_board(mx_instance_state_t *is);
void phx_common_unmap_board(mx_instance_state_t *is);
void phx_common_setup_hw_caps(mx_instance_state_t *is);
void phx_common_int_enable(mx_instance_state_t *is);
void phx_common_int_disable(mx_instance_state_t *is);
void phx_common_park_board(mx_instance_state_t *is);
int phx_sample_hwclock(mx_instance_state_t *is,
                    uint64_t *nic_ticks, uint64_t *host_nsecs,
                    uint64_t *host_nsecs_delay);
void phx_common_set_mac_addr(mx_instance_state_t *is, uint8_t mac[6]);

volatile uint32_t *phx_rx_desc_consr_advance(mx_instance_state_t *is, int endpoint);
volatile uint32_t *phx_rx_data_consr_advance(mx_instance_state_t *is, int endpoint);
int phx_rx_desc_size(mx_instance_state_t *is);
void phx_rx_wake_request(mx_instance_state_t *is, int endpoint, uint32_t host_count);

void *k35snf_tx_base(mx_instance_state_t *is);
void k35snf_tx_stop(mx_instance_state_t *is);
void k35snf_tx_start(mx_instance_state_t *is, uint32_t compl_low, uint32_t compl_high,
                     uint32_t data_pages, uint32_t num_desc);
uint32_t k35snf_tx_get_scale_factor(mx_instance_state_t *is);

void k35snf_icu_enable(mx_instance_state_t *is);
uint8_t k35snf_icu_read_link_status(mx_instance_state_t *is);
void k35snf_icu_request_link_change_interrupt(mx_instance_state_t *is);
void k35snf_icu_set_last_link_status_up(mx_instance_state_t *is);
void k35snf_icu_set_last_link_status_down(mx_instance_state_t *is);
uint8_t k35snf_icu_read_pres_status(mx_instance_state_t *is);
void k35snf_icu_request_pres_change_interrupt(mx_instance_state_t *is);
void k35snf_icu_set_last_pres_status_in(mx_instance_state_t *is);
void k35snf_icu_set_last_pres_status_out(mx_instance_state_t *is);
uint8_t k35snf_read_pres_status(mx_instance_state_t *is);
uint64_t k35snf_xmac_get_link_speed(mx_instance_state_t *is);
void k35snf_xmac_configure(mx_instance_state_t *is, uint32_t port_id);
void k35snf_rx_wake_request(mx_instance_state_t *is, uint32_t host_count);
/* We do not use TxWakeRequest.  Only use TxComplCountAck. */
/* void k35snf_tx_wake_request(mx_instance_state_t *is, uint32_t host_count); */
void k35snf_tx_compl_count_ack(mx_instance_state_t *is, uint32_t packets);
uint32_t k35snf_tx_compl_count(mx_instance_state_t *is);
void k35snf_xmac_rx_enable(mx_instance_state_t *is, uint32_t port_id);
void k35snf_xmac_rx_disable(mx_instance_state_t *is, uint32_t port_id);

/* Take/return the port 0 endpoints from the ethernet interface */
int k35snf_ether_remove_endpoints(mx_instance_state_t *is);
void k35snf_ether_return_endpoint(mx_instance_state_t *is);

/* Altera MAC mac_ops functions */
void altera_mac_set_max_frame_length(mx_instance_state_t *is, int mtu);
void altera_mac_set_promisc_mode(mx_instance_state_t *is, int enable);
void altera_mac_set_address(mx_instance_state_t *is, uint8_t mac[6]);
void altera_mac_disable(mx_instance_state_t *is);
void altera_mac_enable(mx_instance_state_t *is);

/* Xilinx XMAC mac_ops functions */
void xilinx_xmac_set_max_frame_length(mx_instance_state_t *is, int mtu);
void xilinx_xmac_set_promisc_mode(mx_instance_state_t *is, int enable);
void xilinx_xmac_set_address(mx_instance_state_t *is, uint8_t mac[6]);
void xilinx_xmac_disable(mx_instance_state_t *is);
void xilinx_xmac_enable(mx_instance_state_t *is);

#endif /* _phx_common_h_ */
