/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2011 by Myricom, Inc.  All rights reserved.          *
 * Copyright 2014 - 2015 by CSP, Inc.  All rights reserved.              *
 *************************************************************************/

#include "mx_arch.h"
#include "mx_instance.h"
#include "phx_hw.h"
#include "phx_common.h"

/* Phoenix:
 * Much of the code in this file is same as myri_dbl_ops.c, as soft endpoints
 * and sleep/wakeups are mostly independent of the NIC.
 */

/* static void dbl_recvq_init(mx_endpt_state_t *es); */
static int dbl_eop_poll_wait(mx_endpt_state_t *es, int timeout_ms, int events);
static int dbl_set_recvq_filter_mode(mx_endpt_state_t *es, enum phx_dbl_filter_mode mode);

/* uint32_t is limited to 4GB.  FIXME */
static uint32_t
fixup_rx_ring_size(uint32_t pages)
{
  uint32_t size, min;

  min = PHX_APU_GAMMA_PAGES_PER_BETA * PAGE_SIZE; /* 2MB */
  size = pages * PAGE_SIZE;
  size = ((size + min - 1) / min) * min; /* a multiple of 2MB */
  if (size <= min)
    size = min;
  return size;
}

static
int
dbl_eop_alloc(mx_endpt_state_t *es, mx_instance_state_t *is, myri_set_endpt_t *set_endpt)
{
  int endpoint = set_endpt->endpoint;
  enum myri_endpt_type es_type;
  struct phx_endpt_state_dbl *es_dbl;
  struct phx_physical_ep *phx_ep;

  if (endpoint < 0 || endpoint >= is->es_count)
    return ENODEV;

  es_type = set_endpt->endpt_type;
  es_dbl = &es->phx_dbl;

  phx_ep = NULL; /* initialize to an invalid endpoint */

  /* for hard endpoint, require an unused epid and claim it */
  if (es_type == MYRI_ES_DBL) {
     /* The kernel endpoint is reserved in the kernel pool and so will not
      * be returned by the claim below.  It does not use the DBL endpoint structures
     * (endpt_state_t and so on) yet.
     * Probably want to unify at some point.  FIXME.
     */
    if (is->es[endpoint] != 0) {
      return EBUSY;
    }

    phx_ep = is->board_ops.physical_endpoint(is, endpoint);
    
    if (phx_claim_fpga_endpoint(is, phx_ep->id, 0) != 0) {
      return EBUSY;
    }

    es->phx_dbl.tag2es = mx_kmalloc(PHX_RX_FILTER_CNT*sizeof(mx_endpt_state_t *),
      MX_WAITOK|MX_MZERO);
    if (es->phx_dbl.tag2es == NULL){
      phx_release_fpga_endpoint(is, phx_ep->id);
      return ENOMEM;
    }
    es->phx_dbl.soft_intr_req_es = mx_kmalloc(PHX_RX_FILTER_CNT*sizeof(mx_endpt_state_t *),
      MX_WAITOK|MX_MZERO);
    if (es->phx_dbl.soft_intr_req_es == NULL){
      phx_release_fpga_endpoint(is, phx_ep->id);
      mx_kfree(es->phx_dbl.tag2es);
      es->phx_dbl.tag2es = NULL;
      return ENOMEM;
    }

    is->es[endpoint] = es;

  /* for soft endpoint, just reference the parent endpoint */
  } else if (es_type == MYRI_ES_DBL_SOFT) {

    /* parent must exist */
    if (is->es[endpoint] == 0) {
      return ENODEV;
    }
    phx_ep = is->board_ops.physical_endpoint(is, endpoint);

    es->flags |= MX_ES_SOFT;
    es->phx_dbl.parent_es = is->es[endpoint];
  } else {
    return EINVAL;
  }

  es->is = is;
  es->endpt = endpoint;
  es->phx_ep = phx_ep;
  es->session_id = 1; /* Must be neither 0 nor -1 */
  es_dbl->dmasend_threshold = phx_get_dmasend_threshold(is);
  es_dbl->recvq_filter_mode = PHX_DBL_FILTER_NORMAL;
  mx_spin_lock_init(&es_dbl->sleep_lock, is, endpoint, "es->phx_dbl.sleep_lock");
  TAILQ_INIT(&es->rx_filterq);

  /* Initialize buffer sizes and NIC pointers.
   * We allocate ring in enable().
   */
  es_dbl->ep_mem = phx_endpoint_nic_mem(is, es->phx_ep->tx_ep);
  es_dbl->ep_mem_size = is->phx.pfi->tx_config.dbltx_core_length;

  es_dbl->tx_desc_buf = phx_ep_desc_fifo(es_dbl->ep_mem);

  es_dbl->tx_pio_buf = phx_ep_pio_fifo(is, es->phx_ep->tx_ep);
  es_dbl->tx_pio_buf_size = is->phx.pfi->tx_config.dbltx_pio_length;

  es_dbl->tx_dma_buf.n_pages = PHX_TX_DMA_PAGES_PER_ENDPOINT;

  /* desc ring = half the size of the data ring */
  es_dbl->rx_desc_ring.size = fixup_rx_ring_size(myri_recvq_vpage_cnt / 2);
  es_dbl->rx_data_ring.size = fixup_rx_ring_size(myri_recvq_vpage_cnt);

#if MAL_DEBUG
  MX_INFO(("%s: tx_pio_buf=%p tx_desc_buf=%p\n",
           __FUNCTION__, es_dbl->tx_pio_buf, es_dbl->tx_desc_buf));
#endif
  return 0;
}

static
int
dbl_recvq_filter_isvalid(mx_instance_state_t *is, enum phx_dbl_filter_mode mode)
{
  /* is->sync should be held ! */
  if (mode == PHX_DBL_FILTER_NORMAL)
    return 0;
  else if (mode == PHX_DBL_FILTER_ETHER_MULTI || mode == PHX_DBL_FILTER_ALL) {
    if (is->phx_dbl.es_filter_owner == NULL)
      return 0;
    else
      return EBUSY;
  }
  else
    return EINVAL;
}

/* Based on free_phx_ether_pages in phx_ether_common.c.
 * Factor out later on.  FIXME
 */
static void
dbl_eop_free_phx_dbl_pages(mx_endpt_state_t *es, phx_dbl_pages_t *p)
{
  mx_page_pin_t *pages;
  char *addr;
  int i;
  struct phx_endpt_state_dbl *es_dbl;

  es_dbl = &es->phx_dbl;

  if (p && p->pages) {
#if MAL_DEBUG
    MX_INFO(("%s %d: free dbl pages. is_name=%s endpt=%d pages=%p n_pages=%d\n",
             __FUNCTION__, __LINE__, es->is->is_name, es->endpt, p->pages, 
	     p->n_pages));
#endif
    pages = p->pages;
    for (i = 0; i < p->n_pages; i++) {
      if (pages[i].va) {
        addr = (char*)pages[i].va;
        mx_free_dma_page(es->is, &addr, &pages[i]);
        memset(&pages[i], 0, sizeof(pages[i]));
      }
    }
    mx_kfree(pages);
    memset(p, 0, sizeof(*p));
  }
}

/* Based on alloc_phx_ether_pages in phx_ether_common.c.
 * Factor out later on.  FIXME
 */
static int
dbl_eop_alloc_phx_dbl_pages(mx_endpt_state_t *es, phx_dbl_pages_t *p, int n_pages)
{
  mx_page_pin_t *pages;
  char *addr;
  int i;
  struct phx_endpt_state_dbl *es_dbl;

  es_dbl = &es->phx_dbl;

#if MAL_DEBUG
  MX_WARN(("%s %d: allocate phx dbl pages. is_name=%s endpt=%d n_pages=%d\n",
           __FUNCTION__, __LINE__, es->is->is_name, es->endpt, n_pages));
#endif
  pages = mx_kmalloc(sizeof(mx_page_pin_t) * n_pages, MX_MZERO | MX_NOWAIT);
  if (!pages)
    goto abort;
  p->pages = pages;
  p->n_pages = n_pages;
  for (i = 0; i < n_pages; i++) {
    if (mx_alloc_dma_page(es->is, &addr, &pages[i]))
      goto abort;
    memset(addr, 0, PAGE_SIZE);
  }
  return 0;

abort:
  dbl_eop_free_phx_dbl_pages(es, p);
  return -ENOMEM;
}

/* Based on setup_apu_alpha_beta_pages in phx_ether_common.c.
 * Factor out later on.  FIXME
 */
static void
dbl_eop_setup_apu_alpha_beta_pages(mx_endpt_state_t *es, phx_dbl_rx_ring_t *ring)
{
  int n_beta, n_alpha, i, j, k, rem;
  mx_page_pin_t *page, *alpha_page;
  phx_dma_addr_t *addr;
  phx_apu_flp_t *flp, *prev_flp;
  struct phx_endpt_state_dbl *es_dbl;

  es_dbl = &es->phx_dbl;

  /* Fill in beta pages */
  n_beta = ring->beta.n_pages;
  page = ring->gamma.pages;
  for (i = 0; i < n_beta; i++) {
    addr = (phx_dma_addr_t *)ring->beta.pages[i].va;
    for (j = 0; j < PHX_APU_GAMMA_PAGES_PER_BETA; j++) {
      addr->low = page->dma.low;
      addr->high = page->dma.high;
      addr++;
      page++;
    }
  }

  /* Fill in alpha pages */
  n_alpha = ring->alpha.n_pages;
  page = ring->beta.pages;
  rem = n_beta;
  prev_flp = NULL;
  for (i = 0; rem > 0 && i < n_alpha; i++) {
    uint64_t flp_dma_addr;

    alpha_page = &ring->alpha.pages[i];
    flp = (phx_apu_flp_t *)alpha_page->va;
    flp_dma_addr = (((uint64_t)alpha_page->dma.high) << 32) | (uint64_t)alpha_page->dma.low;
    for (j = 0; rem > 0 && j < PHX_APU_FLPS_PER_ALPHA; j++) {
      for (k = 0; rem > 0 && k < PHX_APU_BETA_PAGES_PER_FLP; k++) {
        flp->beta[k].low = page->dma.low;
        flp->beta[k].high = page->dma.high;
        rem--;
        page++;
      }
      /* Link FLPs.  The previous FLP points to the current FLP. */
      if (prev_flp) {
        prev_flp->next.low = (uint32_t)flp_dma_addr;
        prev_flp->next.high = (uint32_t)(flp_dma_addr >> 32);
      }
      flp->opcode = 0;
      PHX_APU_FLP_OPCODE_SET_NUM_BETA(flp->opcode, k);
      prev_flp = flp;
      flp++;
      flp_dma_addr += sizeof(phx_apu_flp_t);
    }
  }
  /* The last/final FLP points back to the first FLP. */
  alpha_page = &ring->alpha.pages[0];
  prev_flp->next.low = alpha_page->dma.low;
  prev_flp->next.high = alpha_page->dma.high;
}

static void
dbl_eop_free_phx_dbl_rx_ring(mx_endpt_state_t *es, phx_dbl_rx_ring_t *ring)
{
#if MAL_OS_WINDOWS
  phx_win_free_gamma_pages(es->is, &ring->gamma, ring->gamma_mdl);
  ring->gamma_mdl = NULL;
#else
  dbl_eop_free_phx_dbl_pages(es, &ring->gamma);
#endif
  dbl_eop_free_phx_dbl_pages(es, &ring->beta);
  dbl_eop_free_phx_dbl_pages(es, &ring->alpha);
}

/* Based on alloc_rx_ring in phx_ether_common.c.
 * Factor out later on.  FIXME
 */
static int
dbl_eop_alloc_phx_dbl_rx_ring(mx_endpt_state_t *es, phx_dbl_rx_ring_t *ring)
{
  int n_gamma, n_beta, n_alpha;
  int flp_per_page = PAGE_SIZE / sizeof(phx_apu_flp_t);
  struct phx_endpt_state_dbl *es_dbl;

  es_dbl = &es->phx_dbl;

  /* First, allocate pages for the data ring */
  n_gamma = ring->size / PAGE_SIZE;
  n_beta = (n_gamma + PHX_APU_GAMMA_PAGES_PER_BETA - 1) / PHX_APU_GAMMA_PAGES_PER_BETA;
  n_alpha = (n_beta + flp_per_page - 1) / flp_per_page;
#if MAL_DEBUG
  MX_WARN(("%s %d: allocate phx dbl rx ring. is_name=%s endpt=%d size=%d(%d/%d/%d)\n",
           __FUNCTION__, __LINE__, es->is->is_name, es->endpt, ring->size,
           n_gamma, n_beta, n_alpha));
#endif
#if MAL_OS_WINDOWS
  if (phx_win_alloc_gamma_pages(es->is, &ring->gamma, n_gamma, &ring->gamma_mdl))
    goto abort;
#else
  if (dbl_eop_alloc_phx_dbl_pages(es, &ring->gamma, n_gamma))
    goto abort;
#endif
  if (dbl_eop_alloc_phx_dbl_pages(es, &ring->beta, n_beta))
    goto abort;
  if (dbl_eop_alloc_phx_dbl_pages(es, &ring->alpha, n_alpha))
    goto abort;
  dbl_eop_setup_apu_alpha_beta_pages(es, ring);
  return 0;

abort:
  /* These check pointers internally.  It is okay to call them without checking
   * NULL pointers.
   */
#if MAL_OS_WINDOWS
  phx_win_free_gamma_pages(es->is, &ring->gamma, ring->gamma_mdl);
  ring->gamma_mdl = NULL;
#else
  dbl_eop_free_phx_dbl_pages(es, &ring->gamma);
#endif
  dbl_eop_free_phx_dbl_pages(es, &ring->beta);
  dbl_eop_free_phx_dbl_pages(es, &ring->alpha);
  return -ENOMEM;
}

static
int
dbl_phx_init(mx_endpt_state_t *es)
{
  struct phx_nic_mem_global *nic;
  struct phx_endpt_state_dbl *es_dbl;
  mx_page_pin_t *page;
#if MAL_DEBUG
  mx_page_pin_t *bpage, *gpage;
  phx_apu_flp_t *flp;
#endif
  int i;

  es_dbl = &es->phx_dbl;
  nic = es->is->phx.nic_mem;

  /* init_board() does per-instance filter initialization */
  /* phx_dbl_filter_init_hack(nic); */

  /* Set up the APU channels: data and descriptor */
  es_dbl->apu_data_ch = phx_apu_channel_id(es->is, 1, es->endpt, 0);
  es_dbl->apu_desc_ch = phx_apu_channel_id(es->is, 0, es->endpt, 0);
#if MAL_DEBUG
  MX_INFO(("APU channels: data=%d desc=%d\n",
           es_dbl->apu_data_ch, es_dbl->apu_desc_ch));
#endif
  phx_apu_channel_reset(es->is, es_dbl->apu_data_ch);
  phx_apu_channel_reset(es->is, es_dbl->apu_desc_ch);

  page = es_dbl->rx_data_ring.alpha.pages;
#if MAL_DEBUG
  flp =  ((phx_apu_flp_t *)page->va);
  MX_INFO(("apu 0 opcode is 0x%llx\n", flp->opcode));
  MX_INFO(("apu 0 next is high 0x%x low 0x%x\n", flp->next.high, flp->next.low));
  MX_INFO(("apu 0 first beta is high 0x%x low 0x%x\n", flp->beta[0].high, flp->beta[0].low));
  bpage = es_dbl->rx_data_ring.beta.pages;
  MX_INFO(("first beta is va 0x%llx phys 0x%x 0x%08x\n", bpage->va, bpage->dma.high, bpage->dma.low));
  MX_INFO(("apu 0 first gamma is 0x%llx\n", ((uint64_t *)bpage->va)[0]));
  gpage = es_dbl->rx_data_ring.gamma.pages;
  MX_INFO(("first gamma is va 0x%llx phys 0x%x 0x%08x\n", gpage->va, gpage->dma.high, gpage->dma.low));
#endif

  phx_apu_channel_start(es->is, es_dbl->apu_data_ch,
                        page->dma.low, page->dma.high);

  page = es_dbl->rx_desc_ring.alpha.pages;
#if MAL_DEBUG
  flp =  ((phx_apu_flp_t *)page->va);
  MX_INFO(("apu 1 opcode is 0x%llx\n", flp->opcode));
  MX_INFO(("apu 1 next is high 0x%x low 0x%x\n", flp->next.high, flp->next.low));
  MX_INFO(("apu 1 first beta is high 0x%x low 0x%x\n", flp->beta[0].high, flp->beta[0].low));
#endif

  phx_apu_channel_start(es->is, es_dbl->apu_desc_ch,
                        page->dma.low, page->dma.high);
  phx_start_rx_endpoint(es->is, es->endpt,
                        es_dbl->rx_data_ring.gamma.n_pages,
                        es_dbl->rx_desc_ring.gamma.n_pages);

  /* Transmit initialization. */

  /* Push all DMA page addresses to the NIC. */
  phx_write_ep_reg_tx_addr_buffer_clear(es_dbl->ep_mem);
  for (i = 0; i < es_dbl->tx_dma_buf.n_pages; i++) {
    phx_write_ep_reg_tx_addr_load(es_dbl->ep_mem,
                                  es_dbl->tx_dma_buf.pages[i].dma.low,
                                  es_dbl->tx_dma_buf.pages[i].dma.high);
  }
  phx_write_ep_reg_tx_addr_buffer_finish(es_dbl->ep_mem);

  /* init_board() has already set the completion page DMA address. */

  /* init_board() also sets the completion threshold. */
  /* Update the completion page every 8 packets (HW default) */
  /* phx_tx_threshold(nic, 8); */

  /* HACK, read in the tx completion count from the endpoint
     and set it in the right spot. This will go away once we can 
     reset the tx logic. */
  {
    uint32_t compl_cnt;
    uint32_t *compl;
    compl_cnt = phx_ep_tx_completion_count(es->phx_dbl.ep_mem);
    compl = (uint32_t *)((uintptr_t)es->is->tx_done.addr);
    compl[es->phx_ep->tx_ep * 2] = compl_cnt;
#if MAL_DEBUG
    MX_INFO(("Setting endpt completion count page addr %p to %d\n", compl, compl_cnt));
#endif
  }

  return 0;
}

static
int
dbl_eop_enable(mx_endpt_state_t *es, uaddr_t arg)
{
  int status = EINVAL;
  mx_instance_state_t *is = es->is;
  struct phx_endpt_state_dbl *es_dbl;

  es_dbl = &es->phx_dbl;

#if MAL_DEBUG
  MX_INFO(("%s %d: is_name=%s endpt=%d\n", __FUNCTION__, __LINE__, es->is->is_name, es->endpt));
#endif

  /* cannot enable soft endpoint */
  if (es->es_type != MYRI_ES_DBL) {
    return EINVAL;
  }

  /* We are not an "early enabler" so acquire the is->sync lock */
  mx_mutex_enter(&is->sync);

  /* Anticipate EBUSY before we try to set the filter after allocating all the
   * copyblocks */
  if ((status = dbl_recvq_filter_isvalid(is, es->phx_dbl.recvq_filter_mode)))
    goto abort;

  /* Allocate rings */
  if ((status = dbl_eop_alloc_phx_dbl_rx_ring(es, &es_dbl->rx_data_ring)))
    goto abort;
  if ((status = dbl_eop_alloc_phx_dbl_rx_ring(es, &es_dbl->rx_desc_ring)))
    goto abort;
#if MAL_OS_WINDOWS
  if ((status = phx_win_alloc_gamma_pages(es->is, &es_dbl->tx_dma_buf, es_dbl->tx_dma_buf.n_pages,
                                              &es_dbl->tx_dma_buf_mdl)))
    goto abort;
#else
  if ((status = dbl_eop_alloc_phx_dbl_pages(es, &es_dbl->tx_dma_buf, es_dbl->tx_dma_buf.n_pages)))
    goto abort;
#endif

  /* FIXME.  Clear the receive descriptor ring. */
  /* dbl_recvq_init(es); */

  dbl_phx_init(es);

  es->flags |= MX_ES_MCP_ENABLED;

  /* Now go set filter -- if this fails, the user that calls enable is
   * responsible for closing the endpoint */
  status = dbl_set_recvq_filter_mode(es, es->phx_dbl.recvq_filter_mode);

  /* the interrupt vector for each endpoint is the endpoint number + 1 */
#if MAL_DEBUG
  MX_INFO(("Enabling interrupt vector %d for endpt %d for instance %d\n",
           es->endpt + 1, es->endpt, is->id));
#endif
  phx_icu_enable_interrupt(is, es->endpt + 1);

abort:
  if (status) {
    dbl_eop_free_phx_dbl_rx_ring(es, &es_dbl->rx_data_ring);
    dbl_eop_free_phx_dbl_rx_ring(es, &es_dbl->rx_desc_ring);
#if MAL_OS_WINDOWS
    phx_win_free_gamma_pages(es->is, &es_dbl->tx_dma_buf, es_dbl->tx_dma_buf_mdl);
    es_dbl->tx_dma_buf_mdl = NULL;
#else
    dbl_eop_free_phx_dbl_pages(es, &es_dbl->tx_dma_buf);
#endif
  }

  mx_mutex_exit(&is->sync);

  phx_sw_counters_ep_open(es);

  return status;
}

static
int
dbl_set_recvq_filter_mode(mx_endpt_state_t *es, enum phx_dbl_filter_mode mode)
{
  int status;
  struct phx_endpt_state_dbl *es_dbl;

  es_dbl = &es->phx_dbl;

  /* cannot set mode on soft endpoint */
  if (es->es_type != MYRI_ES_DBL) {
    return EINVAL;
  }

  /* The endpoint must be fully enabled before we try to set the filter */
  if (!(es->flags & MX_ES_MCP_ENABLED))
    return EINVAL;

  if ((status = dbl_recvq_filter_isvalid(es->is, mode)))
    return status;

  switch (mode) {
    case PHX_DBL_FILTER_NONE:
    case PHX_DBL_FILTER_NORMAL:
    case PHX_DBL_FILTER_ALL:
    case PHX_DBL_FILTER_ETHER_MULTI:
      if (mode != es->is->phx_dbl.recvq_filter_mode) {
        MX_WARN(("%s %d: Phoenix: unimpl mode=%u\n", __FUNCTION__, __LINE__, mode));
        MX_WARN(("%s: endpt %d, unable to set filter mode to %d (err=%d)\n",
            es->is->is_name, es->endpt, mode, status));
        status = EINVAL;
        /* FIXME */
#if 0
        status = mx_mcp_command(es->is, MCP_CMD_DBL_SET_FILTER_MODE, es->endpt,
                                mode, 0, NULL);
        if (status == 0) {
          es->is->phx_dbl.recvq_filter_mode = mode;
          es->phx_dbl.recvq_filter_mode = mode;
          if (mode == MCP_DBL_FILTER_NORMAL || mode == MCP_DBL_FILTER_NONE)
            es->is->phx_dbl.es_filter_owner = NULL;
          else
            es->is->phx_dbl.es_filter_owner = es;
        }
        else {
          MX_WARN(("%s: endpt %d, MCP unable to set filter mode to %d (err=%d)\n",
                   es->is->is_name, es->endpt, mode, status));
        }
#endif
      }
      break;

    default:
      status = EINVAL;
      break;
  }
  return status;
}

static int
dbl_assign_tag(mx_endpt_state_t *es, int tag)
{
  mx_endpt_state_t *parent_es;

  /* validate tag */
  if (tag == 0 || tag >= PHX_RX_FILTER_CNT) {
    return EINVAL;
  }

  /* there must be a parent with an array */
  parent_es = es->phx_dbl.parent_es;
  if (parent_es == NULL || parent_es->phx_dbl.tag2es == NULL) {
    return ENXIO;
  }

  /* save tag value and cross-reference */
  es->phx_dbl.de_tag = tag;
  parent_es->phx_dbl.tag2es[tag] = es;

  return 0;
}

static int
dbl_post_soft_data(mx_endpt_state_t *es)
{
  /* mark data pending and perform a wakeup */
  es->phx_dbl.soft_data_pending = 1;
  myri_pollwake(es);

  return 0;
}

/* Phoenix version of myri_endpt_enable_common_copyblocks. */
static int
dbl_get_copyblocks(const mx_endpt_state_t *es, phx_get_copyblock_t *desc, uint32_t *offp)
{
  const struct phx_endpt_state_dbl *es_dbl;
  uint32_t off = 0;
  mx_instance_state_t *is;

  es_dbl = &es->phx_dbl;
  memset(desc, 0, sizeof(*desc));

  is = es->is;

  /* First the A mappings (exist at open time) */

  /* A.1.1: DblTx pages for this endpoint */
  desc->nic_mem_tx_offset = off;
  desc->nic_mem_tx_len = is->phx.pfi->tx_config.dbltx_core_length;
  off += desc->nic_mem_tx_len;

  /* A.1.1b: DblTx PIO pages for this endpoint */
  desc->nic_mem_tx_pio_offset = off;
  desc->nic_mem_tx_pio_len = is->phx.pfi->tx_config.dbltx_pio_length;
  off += desc->nic_mem_tx_pio_len;

  /* A.1.2: 1 DblRx page for this endpoint */
  desc->nic_mem_rx_offset = off;
  desc->nic_mem_rx_len = PAGE_SIZE;
  off += desc->nic_mem_rx_len;

  /* A.2: kernel window */
  desc->kernel_window_offset = off;
  desc->kernel_window_len = (es->is->kernel_window ? MX_PAGE_SIZE : 0);
  off += desc->kernel_window_len;

  /* A.3: per-endpoint flow window */
  desc->flow_window_offset = off;
  desc->flow_window_len = (es->flow_window ? MX_PAGE_SIZE : 0);
  off += desc->flow_window_len;

  /* A.4: PIO per-endpoint TX completion counters */
  desc->send_compcnt_offset = off;
  desc->send_compcnt_len = (es->is->tx_done.addr ? MX_PAGE_SIZE : 0);
  off += desc->send_compcnt_len;

  /* A.5: BBSYS Ctrl for accessing clock from userspace (Read Only) */
  desc->nic_mem_bbsys_offset = off;
  desc->nic_mem_bbsys_len = PAGE_SIZE;
  off += desc->nic_mem_bbsys_len;

  /* Then the B mappings (exist at enable time) */
  /* B.1: TX DMA buffer */
  desc->tx_dmabuf_offset = off;
  desc->tx_dmabuf_len = es_dbl->tx_dma_buf.n_pages * PAGE_SIZE;
  off += desc->tx_dmabuf_len;

  /* B.2: RX data ring */
  desc->rx_data_ring_offset = off;
  desc->rx_data_ring_len = es_dbl->rx_data_ring.size;
  off += desc->rx_data_ring_len;

  /* B.3: RX descriptor ring */
  desc->rx_desc_ring_offset = off;
  desc->rx_desc_ring_len = es_dbl->rx_desc_ring.size;
  off += desc->rx_desc_ring_len;

  if (offp)
    *offp = off;

  return 0;
}

/* mx_mmap_off_to_kva in mx_instance.c simply calls this function to get
 * the address for all mappings.
 */
static int
dbl_eop_mmap(mx_endpt_state_t *es, unsigned long req,
             unsigned long *off, void **kva,
	     int *mem_type, int *readonly, mx_page_pin_t **pin)
{
  void *tmpkva;
  mx_instance_state_t *is = es->is;
  int index;
  const struct phx_endpt_state_dbl *es_dbl;
  int endpoint;
  int tx_ep_reg_offset, tx_ep_pio_offset;

  es_dbl = &es->phx_dbl;

  /*
   * 'A' Mappings are available once the endpoint is opened.
   * 'B' Mappings are available once the endpoint is enabled.
   */
  endpoint = es->phx_ep->tx_ep;

  if (es->is->board_span) {
    /* mapping A.1.1: DblTx pages */
    tx_ep_reg_offset = is->phx.pfi->tx_config.dbltx_core_base +
                        (endpoint * is->phx.pfi->tx_config.dbltx_core_stride);
    tmpkva = (void *)((char *)es->is->phx.nic_mem + tx_ep_reg_offset +
                      (req - *off));
    *off += is->phx.pfi->tx_config.dbltx_core_length;
    if (req < *off) {
      *kva = tmpkva;
      *mem_type = MX_MEM_SRAM;
      *readonly = 0;
      return 0;
    }

    /* mapping A.1.1b: DblTx PIO pages */
    tx_ep_pio_offset = is->phx.pfi->tx_config.dbltx_pio_base +
                        (endpoint * is->phx.pfi->tx_config.dbltx_pio_stride);
    tmpkva = (void *)((char *)es->is->phx.nic_mem + tx_ep_pio_offset +
                      (req - *off));
    *off += is->phx.pfi->tx_config.dbltx_pio_length;
    if (req < *off) {
      *kva = tmpkva;
      *mem_type = MX_MEM_SRAM;
      *readonly = 0;
      return 0;
    }

    /* mapping A.1.2: DblRx pages */
    tmpkva = (void *)((char *)es->is->phx.nic_mem + PHX_DBL_RX_BASE +
                      (PHX_DBL_RX_ENDPOINT_SIZE * es->endpt) +
                      (req - *off));
    *off += PAGE_SIZE;
    if (req < *off) {
      /* PHX_DBL_RX_ENDPOINT_SIZE is not a page multiple, so tmpkva
       * is usually not page aligned.  Align it manually so we see
       * the same behavior on both Linux and Windows.
       */
      tmpkva = (void*)((uintptr_t)tmpkva & ~(PAGE_SIZE - 1));
      *kva = tmpkva;
      *mem_type = MX_MEM_SRAM;
      *readonly = 0;
      return 0;
    }
  }

  /* mapping A.2: the kernel window (jiffies) */
  *off += is->kernel_window ? MX_PAGE_SIZE : 0;
  if (req + MX_PAGE_SIZE == *off && is->kernel_window) {
    *kva = is->kernel_window;
    *mem_type = MX_MEM_HOSTMEM;
    *readonly = 1;
    /* MX_WARN(("%s %d: Phoenix: kernel window\n", __FUNCTION__, __LINE__)); */
    return 0;
  }

  /* mapping A.3: kernel per-endpoint flow page */
  *off += es->flow_window ? MX_PAGE_SIZE : 0;
  if (req + MX_PAGE_SIZE == *off && es->flow_window) {
    *kva = es->flow_window;
    *pin = &es->flow_window_pin;
    *mem_type = MX_MEM_HOSTMEM;
    *readonly = 0;
    /* MX_WARN(("%s %d: Phoenix: flow window\n", __FUNCTION__, __LINE__)); */
    return 0;
  }

  /* mapping A.4: PIO per-endpoint TX completion counters */
  if (is->tx_done.addr) {
    *off += is->tx_done.addr ? MX_PAGE_SIZE : 0;
    if (req + MX_PAGE_SIZE == *off && is->tx_done.addr) {
      *kva = is->tx_done.addr;
      *pin = &is->tx_done.pin;
      *mem_type = MX_MEM_HOSTMEM;
      *readonly = 1;
      /* MX_WARN(("%s %d: Phoenix: tx_done\n", __FUNCTION__, __LINE__)); */
      return 0;
    }
  }

  /* mapping A.5: bbsysctrl */
  if (es->is->board_span) {
    tmpkva = (void *)((char *)es->is->phx.nic_mem + (req - *off));
    *off += PAGE_SIZE;
    if (req < *off) {
      *kva = tmpkva;
      *mem_type = MX_MEM_SRAM;
      *readonly = 1;
      return 0;
    }
  }

  /* mapping B.1: TX DMA buffer */
  if (es_dbl->tx_dma_buf.n_pages) {
    index = (req - *off) / PAGE_SIZE;
    *off += (unsigned long)es_dbl->tx_dma_buf.n_pages * PAGE_SIZE;
    if (req < *off) {
      *mem_type = MX_MEM_HOSTMEM;
      *kva = (void *)(uintptr_t)es_dbl->tx_dma_buf.pages[index].va;
      *pin = &es_dbl->tx_dma_buf.pages[index];
      *readonly = 0;
      //MX_WARN(("%s %d: Phoenix: tx dma buffer\n", __FUNCTION__, __LINE__));
      return 0;
    }
  }

  /* mapping B.2: RX data ring */
  if (es_dbl->rx_data_ring.size) {
    index = (req - *off) / PAGE_SIZE;
    *off += (unsigned long)es_dbl->rx_data_ring.size;
    if (req < *off) {
      *mem_type = MX_MEM_HOSTMEM;
      *kva = (void *)(uintptr_t)es_dbl->rx_data_ring.gamma.pages[index].va;
      *pin = &es_dbl->rx_data_ring.gamma.pages[index];
      *readonly = 0;
      //MX_WARN(("%s %d: Phoenix: rx data ring\n", __FUNCTION__, __LINE__));
      return 0;
    }
  }

  /* mapping B.3: RX descriptor ring */
  if (es_dbl->rx_desc_ring.size) {
    index = (req - *off) / PAGE_SIZE;
    *off += (unsigned long)es_dbl->rx_desc_ring.size;
    if (req < *off) {
      *mem_type = MX_MEM_HOSTMEM;
      *kva = (void *)(uintptr_t)es_dbl->rx_desc_ring.gamma.pages[index].va;
      *pin = &es_dbl->rx_desc_ring.gamma.pages[index];
      *readonly = 0;
      //MX_WARN(("%s %d: Phoenix: rx descriptor ring\n", __FUNCTION__, __LINE__));
      return 0;
    }
  }

  return ENOENT;
}

static
int
dbl_eop_ioctl(mx_endpt_state_t *es, uint32_t cmd, uaddr_t in)
{
  uaddr_t out = in;
  int status = 0;
  mx_instance_state_t *is = es->is;

  switch (cmd) {

  /* This ioctl is endptless for Lanai, but it is not for Phoenix.
   * We read the queue length from the NIC.
   */
  case MX_GET_MAX_SEND_HANDLES:
    {
      int cnt = phx_read_ep_reg_tx_desc_buffer_size(es->phx_dbl.ep_mem);
      status = mx_copyout(&cnt, out, sizeof(cnt), es->is_kernel);
    }
    break;

  case MX_GET_COPYBLOCKS:
  {
    phx_get_copyblock_t desc;

    if (!(es->flags & MX_ES_MCP_ENABLED)) {
      status = EINVAL;
      break;
    }

    if ((status = dbl_get_copyblocks(es, &desc, NULL)))
      break;
    mx_copyout(&desc, out, sizeof(desc), es->is_kernel);
  }
  break;

  case MYRI_GET_PHX_PHYS_PORT_NUM:
    { 
      int pnum = es->is->phx.pci_info.func;
      status = mx_copyout(&pnum, out, sizeof(pnum), es->is_kernel);
    }
    break;

  case MYRI_GET_PHX_TX_ENDPOINT:
   {
     int tx_ep = es->phx_ep->tx_ep;
     status = mx_copyout(&tx_ep, out, sizeof(tx_ep), es->is_kernel);
   }
   break;

  case MYRI_RX_FILTER:
    {
      myri_rx_filter_t x;

      /* cannot put filter on soft endpoint */
      if (es->es_type != MYRI_ES_DBL) {
        status = EINVAL;
        break;
      }

      if (!(es->flags & MX_ES_MCP_ENABLED)) {
        status = EINVAL;
        break;
      }

      status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
      if (status)
        break;

      if (x.remove)
        status = phx_remove_rx_filter(es, &x, NULL);
      else
        status = phx_add_rx_filter(es, &x);
    }
    break;

  case MYRI_DBL_SET_FILTER_MODE:
    {
      uint32_t x;

      status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
      if (status)
        break;

      if (!(es->flags & MX_ES_MCP_ENABLED)) {
        status = EINVAL;
        break;
      }

      mx_mutex_enter(&is->sync);
      status = dbl_set_recvq_filter_mode(es, x);
      mx_mutex_exit(&is->sync);
    }
    break;

  case MYRI_DBL_WAIT_RX:
  case MYRI_DBL_WAIT_TX:
    {
      uint32_t ms;
      uint32_t events =
        (cmd == MYRI_DBL_WAIT_RX) ? MYRI_WAIT_RX_EVENT : MYRI_WAIT_TX_EVENT;

      /* cannot wait like this on soft endpoint */
      if (es->es_type != MYRI_ES_DBL) {
        status = EINVAL;
        break;
      }

      if (!(es->flags & MX_ES_MCP_ENABLED)) {
        status = EINVAL;
        break;
      }
      status = mx_copyin(in, &ms, sizeof(ms), es->is_kernel);
      if (status)
        break;
      status = dbl_eop_poll_wait(es, ms, events);
    }
    break;

#if MAL_OS_LINUX
  case MYRI_DBL_SINK_SOCKET:
    {
      myri_dbl_sink_t x;

      status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
      if (status)
        break;

      status = myri_sink_sock(x.sockfd, &x.bytes_sunk);
      if (status)
        break;
      status = mx_copyout(&x, out, sizeof(x), es->is_kernel);
      if (status)
        break;
    }
    break;
#endif /* MAL_OS_LINUX */

  case MYRI_DBL_GET_PARAMS:
    {
      struct myri_dbl_params x;

      /* cannot get params on soft endpoint */
      if (es->es_type != MYRI_ES_DBL) {
        status = EINVAL;
        break;
      }

      mx_mutex_enter(&es->sync);
      /* fill in params */
      x.dmasend_threshold = es->phx_dbl.dmasend_threshold;
      /* FIXME */
#if 0
      x.recvq_size = (uint32_t) es->recvq.size_requested;
#endif
      x.recvq_filter_mode = es->phx_dbl.recvq_filter_mode;
      mx_mutex_exit(&es->sync);

      status = mx_copyout(&x, out, sizeof(x), es->is_kernel);
    }
    break;

  case MYRI_DBL_SET_PARAMS:
    {
      struct myri_dbl_params x;

      /* cannot set params on soft endpoint */
      if (es->es_type != MYRI_ES_DBL) {
        status = EINVAL;
        break;
      }

      status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
      if (status)
        break;

      mx_mutex_enter(&es->sync);

      /* Always override this default */
      es->phx_dbl.dmasend_threshold = x.dmasend_threshold;
      /* FIXME */
#if 0
      /* See what we can do about recvq */
      if (!(es->flags & MX_ES_MCP_ENABLED)) {
        uint64_t meta_size = MX_VPAGE_SIZE * MX_VPAGE_SIZE
                             / sizeof(mcp_dma_addr_t);
        uint64_t recvq_size = (uint64_t) x.recvq_size;
        uint64_t recvq_size_min = meta_size;

        /* If user requests recvq_size to be 0, they want the smallest value
         * possible.  If their request turns out not to be at least the
         * minimum, it they will be given the minimum. */
        recvq_size /= meta_size;
        if (recvq_size == 0)
          recvq_size = recvq_size_min;
        else
          recvq_size *= meta_size;
        es->recvq.size_requested = recvq_size;
      }
      else {
        /* All the valid post-enable changes */
        if (x.recvq_filter_mode != es->dbl.recvq_filter_mode) {
          mx_mutex_enter(&es->is->sync);
          status = dbl_set_recvq_filter_mode(es, x.recvq_filter_mode);
          mx_mutex_exit(&es->is->sync);
          if (status)
            MX_WARN(("%s: endpt %d, unable to change filter mode to %d (err=%d)\n",
                     es->is->is_name, es->endpt, x.recvq_filter_mode, status));
        }
      }
#endif
      mx_mutex_exit(&es->sync);
    }
    break;

  case MYRI_DBL_ASSIGN_TAG:
    {
      uint32_t x;

      status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
      if (status)
        break;

      status = dbl_assign_tag(es, x);
    }
    break;

  case MYRI_DBL_POST_SOFT_DATA:
    {
      status = dbl_post_soft_data(es);
    }
    break;

#ifdef MYRI_DBL_AB
  case MYRI_DBL_AB_SET_SEQ:
    {
      myri_dbl_ab_seq_set_t  x;
      
      status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
      if (status)
        break;

      status = phx_ab_seq_set(es, x);
    }
    break;

  case MYRI_DBL_AB_AUTORESET:
    {
      myri_dbl_ab_autoreset_t  x;
      
      status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
      if (status)
        break;

      status = phx_ab_autoreset_init(es, x);
    }
    break;

#endif

  default:
    status = ENOTTY;
    break;
  }
  return status;
}

static inline
void
dbl_pollwake(mx_endpt_state_t *es, int flags)
{
  uintptr_t offset;
  long pindex, poff;
  uint32_t flow_desc;
  phx_rx_desc_t *e;
  phx_dbl_pages_t *gamma;
  myri_endpt_flow_t *flow;
  mx_endpt_state_t *softes;
  int req, i, tag;

  softes = 0;
  req = 0;

  /*
   * If receive wakeup, take a look at next pending packet and pull
   * out the tag to see if we need to awaken a soft endpoint
   */
  if (flags & PHX_INTR_ENDPT_WAKE_RX &&
      es->flow_window != NULL && es->phx_dbl.tag2es != NULL) {

    flow = es->flow_window;
    flow_desc = (uint32_t)flow->rx_flow.flow_desc;
    offset = (uintptr_t)flow_desc * sizeof(phx_rx_desc_t);
    pindex = offset / MX_PAGE_SIZE;
    poff = offset % MX_PAGE_SIZE;
    gamma = &es->phx_dbl.rx_desc_ring.gamma;
    pindex = pindex % gamma->n_pages;
    e = (phx_rx_desc_t*)((uintptr_t)gamma->pages[pindex].va + poff);
    if (e->seqnum == flow->rx_flow.flow_seqnum) {
      /* hykim@:
       * Why do we only check the first packet?
       * myri_dbl_ops.c:dbl_pollwake() only checks the first packet's
       * tag too.  Not sure why this is okay.  When the descriptor
       * only contains one packet, which is true now, it is correct.
       */
      tag = e->pkt1_tag;
      if (tag < PHX_RX_FILTER_CNT) {
        softes = es->phx_dbl.tag2es[tag];
        if (softes != NULL) {
          req = softes->phx_dbl.rx_intr_requested;
        }
      }
    }
  }

  /* Clear all soft waiters */
  for (i=0; i<es->phx_dbl.soft_intr_req_cnt; ++i) {
    es->phx_dbl.soft_intr_req_es[i]->phx_dbl.rx_intr_requested = 0;
  }
  es->phx_dbl.soft_intr_req_cnt = 0;

  /* if soft waiter found, wake him, else parent endpoint */
  if (req != 0) {
    softes->phx_dbl.soft_data_pending = 1;
    myri_pollwake(softes);
  } else {
    myri_pollwake(es);
  }
}

static
int
dbl_eop_intr(mx_endpt_state_t *es, int flags)
{
  int intr_requested = 0;

  mx_spin_lock(&es->phx_dbl.sleep_lock);

  if (flags & PHX_INTR_ENDPT_WAKE_TX) {
    intr_requested = es->phx_dbl.tx_intr_requested;
    es->phx_dbl.tx_intr_requested = 0;
  }
  if (flags & PHX_INTR_ENDPT_WAKE_RX) {
    intr_requested = es->phx_dbl.rx_intr_requested;
    es->phx_dbl.rx_intr_requested = 0;
  }

  mx_spin_unlock(&es->phx_dbl.sleep_lock);
  if (intr_requested & MYRI_DBL_POLLWAIT)
    dbl_pollwake(es, flags);
  if (intr_requested & MYRI_DBL_MXWAIT)
    mx_wake(&es->wait_sync);

  return 0;
}

static
void
dbl_eop_close(mx_endpt_state_t *es)
{
  struct myri_rx_filterq_entry *rx_filterq_entry;
  mx_instance_state_t *is = es->is;
  struct phx_endpt_state_dbl *es_dbl;

  es_dbl = &es->phx_dbl;

#if MAL_DEBUG
  MX_WARN(("%s %d: is_name=%s endpt=%d\n", __FUNCTION__, __LINE__, es->is->is_name, es->endpt));
#endif

  myri_counters64_ep_close(is, es->endpt);   /* accounting for 64-bit counters before the endpoint reset */

  if (es->es_type == MYRI_ES_DBL) {
    phx_sw_counters_ep_close(es);

    mx_mutex_enter(&is->sync);
    mal_always_assert(es->endpt >= 0 && es->endpt < is->es_count);
    is->es[es->endpt] = 0;
    mx_mutex_exit(&is->sync);

#if MAL_DEBUG
    MX_INFO(("Disabling interrupt for endpoint %d vector %d for instance %d\n", 
             es->endpt, es->endpt + 1, is->id));
#endif
    phx_icu_disable_interrupt(is, es->endpt + 1);
    /* might be overdoing it with both disable and reset, but why not? */
    phx_icu_reset_endpt(is, es->phx_ep);

#if MAL_DEBUG
    MX_INFO(("Clearing tx page addresses in fpga\n"));
#endif
    phx_write_ep_reg_tx_addr_buffer_clear(es->phx_dbl.ep_mem);

    /* stop dblRx */
    phx_stop_rx_endpoint(is, es->endpt);

    /* reset the apu channels */
    phx_apu_channel_reset(is, es_dbl->apu_data_ch);
    phx_apu_channel_reset(is, es_dbl->apu_desc_ch);

    /* reset dblRx */
    phx_reset_rx_endpoint(is, es->endpt);

    /* Free the rings */
    dbl_eop_free_phx_dbl_rx_ring(es, &es_dbl->rx_data_ring);
    dbl_eop_free_phx_dbl_rx_ring(es, &es_dbl->rx_desc_ring);
#if MAL_OS_WINDOWS
    phx_win_free_gamma_pages(es->is, &es_dbl->tx_dma_buf, es_dbl->tx_dma_buf_mdl);
    es_dbl->tx_dma_buf_mdl = NULL;
#else
    dbl_eop_free_phx_dbl_pages(es, &es_dbl->tx_dma_buf);
#endif

    /* tx filter is a rx filter, should we give it a separate queue? */
    while (!TAILQ_EMPTY(&es->rx_filterq)) {
      rx_filterq_entry = TAILQ_FIRST(&es->rx_filterq);
      phx_remove_rx_filter(es, NULL, rx_filterq_entry);
    }
  }

  phx_release_fpga_endpoint(is, es->phx_ep->id);
  if (es->phx_dbl.tag2es != NULL)
    mx_kfree(es->phx_dbl.tag2es);
  if (es->phx_dbl.soft_intr_req_es != NULL)
    mx_kfree(es->phx_dbl.soft_intr_req_es);

  mx_spin_lock_destroy(&es->phx_dbl.sleep_lock);
}

/* FIXME.  No "event" queue for Phoenix yet.
 * We will need to reset the receive descriptor ring for Phoenix.
 */
#if 0
static inline void
eventq_clear(mcp_uevt_dbl_t *evt)
{
  evt->batch_len = 0;
  evt->offset = 0;
  evt->pkt_cnt = 0;
  *((uint8_t *) evt + (MCP_CACHELINE_SIZE-1)) = -1;
}

static void
dbl_recvq_init(mx_endpt_state_t *es)
{
  uintptr_t ev_off;
  long pindex, poff;

  for (ev_off = 0; ev_off < es->recvq.size; ev_off += MCP_CACHELINE_SIZE) {
    pindex = (long)(ev_off / MX_PAGE_SIZE);
    poff = (long)(ev_off % MX_PAGE_SIZE);
    eventq_clear((mcp_uevt_dbl_t *) ((uintptr_t) es->recvq.pins[pindex].va + poff));
  }
}
#endif

/*
 * Use the same logic as libdbl, and peek into the user's
 * ring to see if events are pending.
 */
static
int
dbl_eop_poll_events_pending(mx_endpt_state_t *es)
{
  long pindex, poff;
  phx_rx_desc_t *e;
  phx_dbl_pages_t *gamma;
  uint32_t flow_desc;
  uintptr_t offset;
  myri_endpt_flow_t *flow;

  /* regular endpoint */
  if (es->es_type == MYRI_ES_DBL) {

    if (es->flow_window == NULL)
      return 0;

    flow = es->flow_window;
    flow_desc = (uint32_t)flow->rx_flow.flow_desc;
    offset = (uintptr_t)flow_desc * sizeof(phx_rx_desc_t);
    pindex = offset / MX_PAGE_SIZE;
    poff = offset % MX_PAGE_SIZE;
    gamma = &es->phx_dbl.rx_desc_ring.gamma;
    pindex = pindex % gamma->n_pages;
    e = (phx_rx_desc_t*)((uintptr_t)gamma->pages[pindex].va + poff);
    if (e->seqnum == flow->rx_flow.flow_seqnum)
      return MYRI_WAIT_RX_EVENT;

  /* soft endpoint */
  } else {

    /* check for and consume event.  we consume the event because we know that
     * check is always followed by consumption on soft endpoints
     */
    if (es->phx_dbl.soft_data_pending) {
      es->phx_dbl.soft_data_pending = 0;
      return MYRI_WAIT_RX_EVENT;
    }
  }

  return 0;
}

static
void
dbl_intr_req(mx_endpt_state_t *es, int req_type, int events)
{
  unsigned long flags;
  int req_cnt = 0;
  mx_endpt_state_t *softes;
  uint32_t flow_val;
  int i;
  struct phx_endpt_state_dbl *es_dbl;

  es_dbl = &es->phx_dbl;

  flags = 0; /* -Wunused on non-linux */

  /* If this is a soft endpoint, find the parent */
  if (es->es_type == MYRI_ES_DBL_SOFT) {
    softes = es;
    es = softes->phx_dbl.parent_es;
  } else {
    softes = NULL;
  }

  mx_spin_lock_irqsave(&es->phx_dbl.sleep_lock, flags);

  if (events & MYRI_WAIT_RX_EVENT) {
    if (es->phx_dbl.rx_intr_requested == 0) {
      /* Ensure that we have at most one interrupt request per endpoint */
      flow_val = es->flow_window->rx_flow.flow_desc;
      phx_rx_wake_request(es->is, es->endpt, flow_val);
      req_cnt++;
    }
    es->phx_dbl.rx_intr_requested |= req_type;

    /* If soft endpoint, mark interrupt requested here also */
    if (softes != NULL && softes->phx_dbl.rx_intr_requested == 0) {
      softes->phx_dbl.rx_intr_requested = req_type;
      i = es->phx_dbl.soft_intr_req_cnt;
      es->phx_dbl.soft_intr_req_es[i] = softes;
      i++;
      es->phx_dbl.soft_intr_req_cnt = i;
    }

  }
  if (events & MYRI_WAIT_TX_EVENT) {
    if (es->phx_dbl.tx_intr_requested == 0) {
      flow_val = es->flow_window->tx_flow.flow_desc;
#if PHX_WORKAROUND_TX_WAKE_WRAP_AROUND
      if (0xffffffff - flow_val < 9) {
        /* The wake request sequence is too close to the wraparound point.
         * The send completion might already have crossed it.  E.g. 0xfffffffe and 3.
         * Write 0 to the wake request instead.  If the send completion count
         * still is <= 0xffffffff, we would get an interrupt right away.  That is okay.
         * FIXME.  9 is a magic number.
         */
        flow_val = 0;
      }
#endif
      /* use the dbltx mapped pointer  */
      phx_tx_wake_request(es_dbl->ep_mem, flow_val);
      req_cnt++;
    }
    es->phx_dbl.tx_intr_requested |= req_type;
  }
  mx_spin_unlock_irqrestore(&es->phx_dbl.sleep_lock, flags);
  if (req_cnt)
    MAL_STBAR();
}

static
void
dbl_eop_poll_prep_wait(mx_endpt_state_t *es)
{
  /* Only RX interrupts in poll, i.e. only POLLIN */
  dbl_intr_req(es, MYRI_DBL_POLLWAIT, MYRI_WAIT_RX_EVENT);
}

static
int
dbl_eop_poll_wait(mx_endpt_state_t *es, int timeout_ms, int events)
{
  int status;

  if (timeout_ms < 0)
    timeout_ms = MX_MAX_WAIT;

  if (timeout_ms == 0) {
    /* We're not going to block below in sleep, but claim the interrupt if it
     * is available.  This is mostly useful when poll_wait is called with
     * timeout_ms==0 when the endpoint is used with poll() */
    status = mx_sleep(&es->wait_sync, 0, MX_SLEEP_INTR);
    if (status == 0)
      return status;
    else {
      if (status != EAGAIN)
        MX_WARN(("unexpected return from poll_wait: %d\n", status));
      return status;
    }
  }

  dbl_intr_req(es, MYRI_DBL_MXWAIT, events);
  status = mx_sleep(&es->wait_sync, timeout_ms, MX_SLEEP_INTR);
  return status;
}

myri_endpt_ops_t phx_dbl_ops;

void
phx_dbl_init_ops(void)
{
  bzero(&phx_dbl_ops, sizeof(phx_dbl_ops));
  phx_dbl_ops.flags = MYRI_EOP_F_POLL;
  phx_dbl_ops.alloc = dbl_eop_alloc;
  phx_dbl_ops.enable = dbl_eop_enable;
  phx_dbl_ops.ioctl = dbl_eop_ioctl;
  phx_dbl_ops.interrupt_ep = dbl_eop_intr;
  phx_dbl_ops.interrupt_is = NULL;
  phx_dbl_ops.mmap  = dbl_eop_mmap;
  phx_dbl_ops.close = dbl_eop_close;
  phx_dbl_ops.poll_events_pending = dbl_eop_poll_events_pending;
  phx_dbl_ops.poll_prep_wait = dbl_eop_poll_prep_wait;
  phx_dbl_ops.poll_wait = dbl_eop_poll_wait;
}
