/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2014-2016 by CSP, Inc.  All rights reserved.                *
 *************************************************************************/

#include "mx_arch.h"
#include "mx_misc.h"
#include "mx_instance.h"
#include "mx_malloc.h"
#include "mx_pio.h"
#include "mal_stbar.h"
#include "myri_version.h"
#include "mal_cpu.h"
#include "phx_hw.h"
#include "phx_common.h"
#include "phx_avl.h"
#include "phx_ether_common.h"

mx_board_ops_t phx_e1_ops = { NULL };

/* Identical to phx_dbl4p_set_intr_coal at this time */
static int
phx_e1_set_intr_coal(mx_instance_state_t *is, int microseconds)
{
  if (PHX_HAS_FEATURE(is, PHX_FEATURE_INTR_INTERVAL)) {
    /* 6.4 nsec per cycle */
    uint32_t cycles = ((microseconds * 10000) / 64);
    uint8_t *r = (uint8_t*)is->phx.nic_mem + PHX_ICU_BASE;
    r += PHX_ICU_INTR_INTERVAL;
    PHX_REG_WRITE_4(r, cycles);
    is->intr_coal_usec = microseconds;
  }
  return is->intr_coal_usec;
}

static int
phx_e1_init_board(mx_instance_state_t *is)
{
  uint32_t version;
  int init_fpga;
  int attempts = 0;

  /* The NIC has been mapped earlier.  So it is safe to access NIC
   * registers via nic_mem.
   */

  mx_mutex_enter(&mx_fpga_mutex);
  init_fpga = !is->phx.pfi->initialized;

  /* Disable the MAC TX/RX */
  is->mac_ops.mac_disable(is);

  if (init_fpga)
    phx_dbl_rx_global_stop(is);

  /* On Windows, the driver calls phx_win_init_fpga() early on
   * prior to this function.
   */
#if !MAL_OS_WINDOWS
  if (init_fpga) {

    MX_INFO(("Initializing FPGA: start\n"));
    phx_avl_init((char*)is->phx.i2c_base_mem, &version);

    /* firmware monotonic version for compatibility checks */
    is->phx.pfi->firmware_rev = version;
    MX_INFO(("Initializing FPGA: end, firmware version %d\n", version));

#if PHX_ENABLE_PTP
    /* initialize the PTP clock */
    phx_ptp_init(is->phx.pfi);

    is->phx.pfi->ptp_clock = ptp_clock_register(&is->phx.pfi->ptp_info,
           &is->arch.pci_dev->dev);
    if(IS_ERR(is->phx.pfi->ptp_clock)){
       is->phx.pfi->ptp_clock = NULL;
    }
#endif

    /* read the eeprom strings from the onboard eeprom */
    if (is->phx.pfi->eeprom_strings)
      phx_read_specs((char*)is->phx.i2c_base_mem, is->phx.pfi->eeprom_strings, 256, 0);
  }
#endif

  if (init_fpga) {
    phx_clk_init(is, myri_clk_invert_pps, !myri_clk_enable_pps,
		 !myri_clk_enable_10mhz);
  }

  /* private copy the eeprom strings gets modified by parsing  */
  if ((is->phx.pfi->eeprom_strings) && (is->phx.eeprom_strings))
    memcpy(is->phx.eeprom_strings, is->phx.pfi->eeprom_strings, 256);

  /* This sets MAC address (is->mac_addr).
   * PHX-310 retry reading if copy if not OK.
   */
  while (phx_parse_eeprom_strings(is) && (attempts++ < 10)){
    MX_INFO(("FATAL error could not process eeprom strings\n"));
    phx_read_specs((char*)is->phx.i2c_base_mem, is->phx.eeprom_strings, 256, 0);
  }

  MX_INFO(("Initializing PHY\n"));
  phx_vitesse_init(is, init_fpga);

  /* Set the default MTU.  The max frame length seems to be sticky.
   * The setting survives across reboots.  So explicitly set it to
   * the default value.
   */
  is->mac_ops.mac_set_max_frame_length(is, MX_DEFAULT_ETHER_MTU - 14 /* 1500 */);
  is->mac_ops.mac_set_hwaddr(is, is->mac_addr);
  is->mac_ops.mac_set_promisc(is, 0);
  is->phx.num_l2_filters = 0;
  is->link_speed = 10000000000;   /* 10^10 bps */
 
  /* Tell the NIC to use the MRR in the PCI configuration.  Otherwise,
   * it always uses 128B, which leads to ~4Gbps maximum DMA send
   * throughput.
   */
  phx_tx_enable_auto_read_request(is->phx.nic_mem);

  if (init_fpga) {

    /* global tx stuff */
    is->phx.pfi->tx_config.dbltx_core_base = PHX_REG_TX_BASE;
    is->phx.pfi->tx_config.dbltx_core_stride = 2 * PAGE_SIZE;
    is->phx.pfi->tx_config.dbltx_core_length = 1 * PAGE_SIZE;
    is->phx.pfi->tx_config.dbltx_pio_base = PHX_REG_TX_BASE + 0x1000;
    is->phx.pfi->tx_config.dbltx_pio_stride = 2 * PAGE_SIZE;
    is->phx.pfi->tx_config.dbltx_pio_length = 1 * PAGE_SIZE;
    is->phx.pfi->tx_config.dmasend_threshold = PHX_DMASEND_THRESHOLD_SIZE;

    phx_counters_tx_fixup(is);     /* Tx endpoint counters depend on stride above */

    phx_rx_selector_reset(is->phx.nic_mem);

    /* global A/B stuff */
    phx_dbl_ab_init(is->phx.nic_mem);

    /* Set the send completion page address for DBL TX.  This is a
     * global setting, not per endpoint.  Both the kernel endpoint and
     * libdbl endpoints use this same page.
     */
    phx_write_reg_tx_host_completion(is->phx.nic_mem,
				     is->tx_done.pin.dma.low,
				     is->tx_done.pin.dma.high);
    /* Completion update threshold */
    phx_tx_threshold(is->phx.nic_mem, PHX_TX_THRESHOLD);

    /* Rx packet coalescing thresholds */
    phx_dbl_rx_packet_count_limit(is, myri_rx_coalesce_pkt);
    phx_dbl_rx_time_limit(is, myri_rx_coalesce_time);
    phx_dbl_rx_length_limit(is, myri_rx_coalesce_len);
  }

  /* This filter init is not global, so do it here. */
  phx_dbl_filter_init(is, is->phx.pci_info.func);

  /* There is a local A/B bypass as well */
  phx_dbl_ab_local_init(is->phx.nic_mem);

  /* this will only take effect if enabled and supported in bitstream */
  phx_dbl_filter_discard_multicast(is, 1);

  /* Enable interrupts */
  phx_icu_enable(is);

  if (init_fpga) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0)
    struct timespec64 ts;
#else
    struct timespec ts;
#endif
    mx_getnstimeofday(&ts);
    if (phx_clk_is_pps_active(is)) {
      phx_clk_set_with_pps(is, &ts);
      MX_INFO(("PPS is active, wrote the seconds to the adapter\n"));
      is->clk_ts_state = MYRI_CLK_TS_EXTERNAL_PPS;
    } else {
      phx_clk_set(is, &ts);
      MX_INFO(("PPS is inactive, wrote the time to the adapter\n"));
      is->clk_ts_state = MYRI_CLK_TS_INTERNAL_TCXO;
    }

    /* global rx core start */
    phx_dbl_rx_global_start(is);
  }

  /* Clear the counters for the first time.  We will read all the registers
   * and save them in the process.
   */
  myri_counters64_clear(is);

  if (init_fpga)
    is->phx.pfi->initialized = 1;

  /* Enable the local MAC TX/RX */
  is->mac_ops.mac_enable(is);

  /* Interrupt coalescing */
  is->intr_coal_usec = 0;
  if (PHX_HAS_FEATURE(is, PHX_FEATURE_INTR_INTERVAL))
    phx_e1_set_intr_coal(is, myri_intr_coal_delay);

  mx_mutex_exit(&mx_fpga_mutex);
  return 0;
}

static int
phx_e1_init_endpoint_pool(mx_instance_state_t *is, struct phx_fpga_endpoint_pool *pool)
{
  int i;
  int num_functions = is->phx.num_functions;
  struct phx_physical_ep *ep;

  pool->num_function_endpoints = is->phx.apu_config.endpoints;
  pool->num_total_endpoints = is->phx.apu_config.endpoints;
  pool->endpoints = mx_kmalloc(sizeof(struct phx_physical_ep) * pool->num_total_endpoints, MX_MZERO);
  if (!pool->endpoints)
    return -1;
  /* DBL */
  for (i = 0; i < pool->num_total_endpoints; i++) {
    ep = &pool->endpoints[i];
    ep->id = i;
    ep->owner_is = NULL;
    ep->reserved = 0;
    ep->rx_ep = i;
    ep->tx_ep = i;  /* RX and TX numbers are the same */
    ep->icu = i + 1;
    ep->func = -1;  /* Can be used by any port */
    /* Kernel endpoints.  Function 0 uses endpoint 0, 1 and 1, and so on. */
    if ((i < num_functions) && (myri_disable_eth == 0)) {
      ep->reserved = 1;
      ep->func = i;
    }
  }
  pool->num_function_not_reserved_endpoints = pool->num_function_endpoints - num_functions;
  return 0;
}

static struct phx_physical_ep *
phx_e1_ether_endpoint(mx_instance_state_t *is)
{
  return &is->phx.pfi->ep_pool.endpoints[is->phx.pci_info.func];
}

/* n'th within the function */
static struct phx_physical_ep *
phx_e1_physical_endpoint(mx_instance_state_t *is, int nth)
{
  /* on E1 the per-port and physical endpoints are 1 to 1 */
  return &is->phx.pfi->ep_pool.endpoints[nth];
}

/*
 * phx_counters table
 *
 *   myri_counters will display counters in the order below.
 *     - name is the string printed by myri_counters
 *     - offset is the register offset in the FPGA address space
 *     - class - has size, whether it counts or is a static value, how/when to display it
 *     - stride is used for endpoint counters, to get FPGA address
 *            = offset of the register for endpoint 1, relative to endpoint 0 register
 */
static phx_counter_t phx_e1_counters[] = {
  { "Cycle Count", PHX_DBL_RX_BASE + PHX_DBL_RX_CYCLE_COUNT,               CNT_GLOBAL_HEX,     0},
  { "Driver Uptime", 0,                                                    CNT_DRV_UPTIME, 0},
  { "Counters Uptime", 0,                                                  CNT_UPTIME, 0},
  { "Packets Received", PHX_DBL_RX_BASE + PHX_DBL_RX_PACKET_COUNT,         CNT_GLOBAL_64,     0},
  { "Bytes Received", PHX_DBL_RX_BASE + PHX_DBL_RX_BYTE_COUNT,             CNT_GLOBAL_64,     0},
  { "Filter Packets Rx", PHX_FLTR_BASE + PHX_FLTR_PKTS_SENT,               CNT_ALIASED,     0},
  { "Filter Bytes Rx", PHX_FLTR_BASE + PHX_FLTR_BYTES_SENT,                CNT_ALIASED,     0},
  { "Packets Drop Filter", PHX_FLTR_BASE + PHX_FLTR_PKTS_DROPPED,          CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "MAC Receive Errors", PHX_FLTR_BASE + PHX_FLTR_PKTS_ABORTED,           CNT_ALIASED,     0},
  { "Packets Filtered", PHX_FLTR_BASE + PHX_FLTR_PKTS_FILTERED,            CNT_ALIASED,     0},
  { "Packets Unfiltered", PHX_FLTR_BASE + PHX_FLTR_PKTS_UNFILTERED,        CNT_ALIASED,     0},
  { "Max Filter Latency", PHX_FLTR_BASE + PHX_FLTR_PRT_MAX_LATENCY,        CNT_ALIASED | COUNTER_TYPE_SNAPSHOT,     0},
  { "Packets Skip Filter HW", PHX_FLTR_BASE + PHX_FLTR_PKTS_SKIPPED_FILTER,   CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Packets Drop Filter HW", PHX_FLTR_BASE + PHX_FLTR_PKTS_DROP_FILTER,   CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Packets Rx Filter", PHX_FLTR_BASE + PHX_FLTR_PKTS_RECEIVED,           CNT_ALIASED,     0},
  { "Packets Drop Incoming", PHX_FLTR_BASE + PHX_FLTR_INC_PACKETS_DROPPED, CNT_ALIASED,     0},
  { "Intentional Discard Pkts", PHX_FLTR_BASE +  PHX_FLTR_PKTS_DISCARD,    CNT_ALIASED | COUNTER_DISPLAY_VERBOSE,     0},
  { "Multicast Packets Rx", PHX_FLTR_BASE + PHX_FLTR_PKTS_CLASS_D,         CNT_ALIASED | COUNTER_DISPLAY_VERBOSE,     0},
  { "Multicast Pkts Discarded", PHX_FLTR_BASE + PHX_FLTR_PKTS_DISCARD_CLASS_D,    CNT_ALIASED | COUNTER_DISPLAY_VERBOSE,     0},
  { "FF Packets Received", PHX_DBL_RX_BASE + PHX_DBL_RX_FF_PACKET_COUNT,          CNT_GLOBAL | COUNTER_DISPLAY_VB2,     0},
  { "FF Bytes Received", PHX_DBL_RX_BASE + PHX_DBL_RX_FF_BYTE_COUNT,              CNT_GLOBAL | COUNTER_DISPLAY_VB2,     0},
  { "FD Packets Discarded", PHX_DBL_RX_BASE + PHX_DBL_RX_FD_PACKET_COUNT,         CNT_GLOBAL | COUNTER_DISPLAY_VB2,     0},
  { "FD Bytes Discarded", PHX_DBL_RX_BASE + PHX_DBL_RX_FD_BYTE_COUNT,             CNT_GLOBAL | COUNTER_DISPLAY_VB2,     0},
  { "PCIE FIFO empty", PHX_DBL_RX_BASE + PHX_DBL_RX_BP_EMPTY,                     CNT_GLOBAL | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "PCIE FIFO < 1/4", PHX_DBL_RX_BASE + PHX_DBL_RX_BP_Q1,                        CNT_GLOBAL | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "PCIE FIFO 1/4 - 1/2", PHX_DBL_RX_BASE + PHX_DBL_RX_BP_Q2,                    CNT_GLOBAL | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "PCIE FIFO 1/2 - 3/4", PHX_DBL_RX_BASE + PHX_DBL_RX_BP_Q3,                    CNT_GLOBAL | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "PCIE FIFO > 3/4", PHX_DBL_RX_BASE + PHX_DBL_RX_BP_Q4,                        CNT_GLOBAL | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "PCIE FIFO full", PHX_DBL_RX_BASE + PHX_DBL_RX_BP_FULL,                       CNT_GLOBAL | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "Filter Kernel Endpoint", PHX_FLTR_BASE + PHX_FLTR_KERNEL_ENDPOINT,           CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "Filter Lookup Count", PHX_FLTR_BASE + PHX_FLTR_LOOKUP_CNT,                   CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Filter Lookup Fail", PHX_FLTR_BASE + PHX_FLTR_LOOKUP_FAIL,                   CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Filter Lookup Search", PHX_FLTR_BASE + PHX_FLTR_LOOKUP_SEARCH,               CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Filter Add Error", PHX_FLTR_BASE + PHX_FLTR_ADD_ERROR,                       CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "Filter Remove Error", PHX_FLTR_BASE + PHX_FLTR_REMOVE_ERROR,                 CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Packets Received", PHX_FLTR_BASE + PHX_FLTR_RP_PKTS_RECV,      CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Packets Sent", PHX_FLTR_BASE + PHX_FLTR_RP_PKTS_SENT,          CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Bytes Sent", PHX_FLTR_BASE + PHX_FLTR_RP_BYTES_SENT,           CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Packets Dropped", PHX_FLTR_BASE + PHX_FLTR_RP_PKTS_DROPPED,    CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Sent Duplicate", PHX_FLTR_BASE + PHX_FLTR_RP_SENT_DUP,         CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Sent Abort", PHX_FLTR_BASE + PHX_FLTR_RP_SENT_ABORT,           CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Dropped Abort", PHX_FLTR_BASE + PHX_FLTR_RP_DROPPED_ABORT,     CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Dropped Control", PHX_FLTR_BASE + PHX_FLTR_RP_DROPPED_CONTROL, CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Dropped Meta Data", PHX_FLTR_BASE + PHX_FLTR_RP_DROPPED_META,  CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Control Full", PHX_FLTR_BASE + PHX_FLTR_RP_CONTROL_FULL,       CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Receive Abort", PHX_FLTR_BASE + PHX_FLTR_RP_RECV_ABORT,        CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Meta Error", PHX_FLTR_BASE + PHX_FLTR_RP_META_ERROR,           CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Filter Add Slots",  PHX_FLTR_BASE + PHX_FLTR_ADD_SLOTS,               CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE,     0},
  { "Filter Used Slots",  PHX_FLTR_BASE + PHX_FLTR_FREE_SLOTS,
                 CNT_ALIASED | COUNTER_TYPE_16_HI | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE, 0},
  { "Filter Free Slots",  PHX_FLTR_BASE + PHX_FLTR_FREE_SLOTS,
                 CNT_ALIASED | COUNTER_TYPE_16_LO | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE, 0},
  { "RX Packet Count Limit", PHX_DBL_RX_BASE + PHX_DBL_RX_PACKET_COUNT_LIMIT,     CNT_GLOBAL | COUNTER_TYPE_SNAPSHOT  | COUNTER_DISPLAY_VERBOSE,     0},
  { "RX Length Limit", PHX_DBL_RX_BASE + PHX_DBL_RX_LENGTH_LIMIT,                 CNT_GLOBAL_HEX | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE, 0},
  { "RX Time Limit", PHX_DBL_RX_BASE + PHX_DBL_RX_TIME_LIMIT,                     CNT_GLOBAL | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE,     0},
  { "Packet Overflow", PHX_DBL_RX_BASE + PHX_DBL_RX_PACKET_OVERFLOW,              CNT_GLOBAL | COUNTER_DISPLAY_VERBOSE,     0},
  { "Descriptor Overflow", PHX_DBL_RX_BASE + PHX_DBL_RX_DESC_OVERFLOW,            CNT_GLOBAL | COUNTER_DISPLAY_VERBOSE,     0},
  { "TX Pause MAC CTRL frames", PHX_MAC_BASE + PHX_MAC_TX_PAUSE_MACCTRL,          CNT_ALIASED | COUNTER_TYPE_SELFCLEAR | COUNTER_DISPLAY_VERBOSE,     0},
  { "RX Pause MAC CTRL frames", PHX_MAC_BASE + PHX_MAC_RX_PAUSE_MACCTRL,          CNT_ALIASED | COUNTER_TYPE_SELFCLEAR | COUNTER_DISPLAY_VERBOSE,     0},
  { "APU Version ", PHX_APU_BASE + PHX_APU_ID_VERSION,                            CNT_GLOBAL_HEX | COUNTER_TYPE_SNAPSHOT  | COUNTER_DISPLAY_VB2, 0},
  { "APU Configuration", PHX_APU_BASE + PHX_APU_NUM_CHANNEL,                      CNT_GLOBAL_HEX | COUNTER_TYPE_SNAPSHOT  | COUNTER_DISPLAY_VB2, 0},
  { "EP Packets Received", PHX_DBL_RX_BASE + PHX_DBL_RX_PACKET_COUNT_N,           CNT_ENDPOINT | COUNTER_TYPE_64BIT,   PHX_DBL_RX_ENDPOINT_SIZE},
  { "EP Bytes Received", PHX_DBL_RX_BASE + PHX_DBL_RX_BYTE_COUNT_N,               CNT_ENDPOINT | COUNTER_TYPE_64BIT,   PHX_DBL_RX_ENDPOINT_SIZE},
  { "EP Filter Packets Sent", PHX_FLTR_BASE + PHX_FLTR_EP_PKTS_SENT,              CNT_ALIASED_ENDPOINT | COUNTER_DISPLAY_VERBOSE,   8},
  { "EP Filter Bytes Sent", PHX_FLTR_BASE + PHX_FLTR_EP_BYTES_SENT,               CNT_ALIASED_ENDPOINT | COUNTER_DISPLAY_VERBOSE,   8},
  { "EP TX PIO Packets", 0,                                                       CNT_SW_ENDPOINT, 0},
  { "EP TX PIO Bytes", 1,                                                         CNT_SW_ENDPOINT, 0},
  { "EP TX DMA Packets", 2,                                                       CNT_SW_ENDPOINT, 0},
  { "EP TX DMA Bytes", 3,                                                         CNT_SW_ENDPOINT, 0},
  { "EP Host RX Packets", 4,                                                      CNT_SW_ENDPOINT, 0},
  { "EP Host RX Packets with MAC Errors", 5,                                      CNT_SW_ENDPOINT, 0},
  { "EP TX Completions", PHX_REG_TX_BASE + PHX_EP_REG_TX_COMPLETION_COUNT,        CNT_ENDPOINT | COUNTER_TYPE_TX, 0},
  { "EP Data Pages", PHX_DBL_RX_BASE + PHX_DBL_RX_DATA_RING_SIZE,
                 CNT_ENDPOINT | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE,   PHX_DBL_RX_ENDPOINT_SIZE},
  { "EP Desc Pages", PHX_DBL_RX_BASE + PHX_DBL_RX_DESC_RING_SIZE,
                 CNT_ENDPOINT | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE,   PHX_DBL_RX_ENDPOINT_SIZE},
  { "EP Data Consumer Position",  PHX_DBL_RX_BASE + PHX_DBL_RX_DATA_CONSR_POSITION_N,
		 CNT_ENDPOINT | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE, PHX_DBL_RX_ENDPOINT_SIZE },
  { "EP Data Producer Position",  PHX_DBL_RX_BASE + PHX_DBL_RX_DATA_PRODW_POSITION_N,
		 CNT_ENDPOINT | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE, PHX_DBL_RX_ENDPOINT_SIZE },
  { "EP Desc Consumer Position",  PHX_DBL_RX_BASE + PHX_DBL_RX_DESC_CONSR_POSITION_N,
		 CNT_ENDPOINT | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE, PHX_DBL_RX_ENDPOINT_SIZE },
  { "EP Desc Producer Position",  PHX_DBL_RX_BASE + PHX_DBL_RX_DESC_PRODW_POSITION_N,
		 CNT_ENDPOINT | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE, PHX_DBL_RX_ENDPOINT_SIZE },
  { "Rx Sel Packets Rx", PHX_RX_SEL_BASE + PHX_RX_SEL_PKTS_RX,                   CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Rx Sel Packets Dropped", PHX_RX_SEL_BASE + PHX_RX_SEL_PKTS_DROP,            CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Rx Sel Packets Aborted", PHX_RX_SEL_BASE + PHX_RX_SEL_PKTS_ABT,             CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Rx Sel Packets Ok", PHX_RX_SEL_BASE + PHX_RX_SEL_PKTS_SENT,                 CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Rx Sel Packets No Meta", PHX_RX_SEL_BASE + PHX_RX_SEL_PKTS_NO_META,         CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Rx Sel Packets Meta Dropped", PHX_RX_SEL_BASE + PHX_RX_SEL_PKTS_META_DROP,  CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Rx Sel Packets Extra Meta", PHX_RX_SEL_BASE + PHX_RX_SEL_PKTS_EXTRA_META,   CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Rx Sel Packets Ctrl Dropped", PHX_RX_SEL_BASE + PHX_RX_SEL_PKTS_CTRL_DROP,  CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Rx Sel Bytes OK", PHX_RX_SEL_BASE + PHX_RX_SEL_BYTES_SENT,                  CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "PIO CP Read Dword", PHX_BBSYSCTL_BASE + PHX_BBSYSCTL_CP_RD_DW,              CNT_GLOBAL,     0},
  { "PIO CP Read Header", PHX_BBSYSCTL_BASE + PHX_BBSYSCTL_CP_RD_HEADER,         CNT_GLOBAL,     0},
  { "PIO CP Write Dword", PHX_BBSYSCTL_BASE + PHX_BBSYSCTL_CP_WR_DW,             CNT_GLOBAL,     0},
  { "PIO CP Write Header", PHX_BBSYSCTL_BASE + PHX_BBSYSCTL_CP_WR_HEADER,        CNT_GLOBAL,     0},
  { "IRQ Endpoints", PHX_ICU_BASE + PHX_ICU_ENDPOINTS,                           CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "IRQ Vectors Allocated", PHX_ICU_BASE + PHX_ICU_VECTORS_ALLOC,               CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "IRQ Operation Mode", PHX_ICU_BASE + PHX_ICU_OPERATION_MODE,                 CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "IRQ Global Mask", PHX_ICU_BASE + PHX_ICU_GLOBAL_MASK,                       CNT_ALIASED_HEX | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2, 0},
  { "IRQ Enable Lo", PHX_ICU_BASE + PHX_ICU_ENABLE_LO,                           CNT_ALIASED_HEX | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2, 0},
  { "IRQ Enable Mid", PHX_ICU_BASE + PHX_ICU_ENABLE_MID,                         CNT_ALIASED_HEX | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2, 0},
  { "IRQ Enable Hi", PHX_ICU_BASE + PHX_ICU_ENABLE_HI,                           CNT_ALIASED_HEX | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2, 0},
#ifdef MYRI_DBL_AB
  { "AB Module ID", PHX_AB_BASE + PHX_AB_MODULE_ID,				 CNT_GLOBAL_HEX | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2, 0 },
  { "AB Unit Count", PHX_AB_BASE + PHX_AB_UNIT_COUNT,				 CNT_GLOBAL | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2, 0 },
  { "AB History Depth", PHX_AB_BASE + PHX_AB_HISTORY,				 CNT_GLOBAL | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2, 0 },
  { "AB Ports", PHX_AB_BASE + PHX_AB_PORTS,					 CNT_GLOBAL | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2, 0 },
  { "AB Global Bypass", PHX_AB_BASE + PHX_AB_GLOBAL_BYPASS,			 CNT_GLOBAL | COUNTER_TYPE_SNAPSHOT, 0 },
  { "AB Bypass", PHX_AB_BASE + PHX_AB_BYPASS,           			 CNT_ALIASED | COUNTER_TYPE_SNAPSHOT, 0 },
  { "AB Auto Reset Seq Enable", PHX_AB_BASE + PHX_AB_AUTORESET_ENABLE,           CNT_ALIASED | COUNTER_TYPE_SNAPSHOT, 0 },
  { "AB Auto Reset Boundary", PHX_AB_BASE + PHX_AB_AUTORESET_BOUND,              CNT_ALIASED | COUNTER_TYPE_SNAPSHOT, 0 },
  { "AB Packet Count", PHX_AB_BASE + PHX_AB_PKT_COUNT,				 CNT_ALIASED, 0 },
  { "AB Packet Passed", PHX_AB_BASE + PHX_AB_PKT_PASS,  			 CNT_ALIASED, 0 },
  { "AB Packet Passed Pre-history", PHX_AB_BASE + PHX_AB_PRE_HISTORY, 		 CNT_ALIASED, 0 },
  { "AB A Packets Passed", PHX_AB_BASE + PHX_AB_PKT_A_PASSED,			 CNT_ALIASED, 0 },
  { "AB A Packets Removed", PHX_AB_BASE + PHX_AB_PKT_A_REMOVED,			 CNT_ALIASED, 0 },
  { "AB B Packets Passed", PHX_AB_BASE + PHX_AB_PKT_B_PASSED,			 CNT_ALIASED, 0 },
  { "AB B Packets Removed", PHX_AB_BASE + PHX_AB_PKT_B_REMOVED,			 CNT_ALIASED, 0 },
  { "AB Auto Reset Count", PHX_AB_BASE + PHX_AB_AUTORESET_COUNT,                 CNT_ALIASED, 0 },
  { "AB Last Unit", PHX_AB_BASE + PHX_AB_LAST_UNIT,			         CNT_ALIASED | COUNTER_TYPE_SNAPSHOT, 0 },
  { "AB Last Count", PHX_AB_BASE + PHX_AB_LAST_COUNT,			         CNT_ALIASED | COUNTER_TYPE_SNAPSHOT, 0 },
  { "AB Last Seq", PHX_AB_BASE + PHX_AB_LAST_SEQ,			         CNT_ALIASED | COUNTER_TYPE_SNAPSHOT, 0 },
#endif
};

static int
phx_e1_get_counter_definitions(mx_instance_state_t *is, const phx_counter_t **counters,
                               uint32_t *count)
{
  *counters = phx_e1_counters;
  *count = sizeof(phx_e1_counters) / sizeof(struct phx_counter);
  return 0;
}

static int
phx_e1_counter_accum_index(mx_instance_state_t *is, struct phx_fpga_counter *ctr,
                           int nth_endpoint)
{
  int idx = 0;
  uint32_t bits = ctr->class & (COUNTER_TYPE_ENDPOINT | COUNTER_TYPE_ALIASED |
                                COUNTER_TYPE_GLOBAL);
  if (bits == (COUNTER_TYPE_ENDPOINT | COUNTER_TYPE_ALIASED))
    idx = is->phx.pci_info.func * is->phx.pfi->ep_pool.num_function_endpoints
        + nth_endpoint;
  else if (bits == COUNTER_TYPE_ALIASED)
    idx = is->phx.pci_info.func;
  else if (bits == COUNTER_TYPE_ENDPOINT)
    idx = nth_endpoint;
  return idx;
}

int
phx_e1_set_promisc(mx_instance_state_t *is, int promisc)
{
  is->mac_ops.mac_set_promisc(is, promisc);
  return 0;
}

int
phx_e1_enable_all_multicast(mx_instance_state_t *is)
{
  phx_dbl_filter_discard_multicast(is, 0);
  return 0;
}

int
phx_e1_disable_all_multicast(mx_instance_state_t *is)
{
  phx_dbl_filter_discard_multicast(is, 1);
  return 0;
}

int
phx_e1_remove_filter(mx_instance_state_t *is,  myri_rx_filterq_entry_t *entry)
{
  entry->filt.remove = 1;
  return __phx_write_filter_regs(is->phx.nic_mem,
				&is->phx.filter_sync,
				&entry->filt,
				is->phx.pci_info.func,
				PHX_FLTR_FLAG_BIT_IGNORE_VLAN |
				PHX_FLTR_FLAG_BIT_IGNORE_DST_PORT,
				NULL);
}

int
phx_e1_add_filter(mx_instance_state_t *is,  myri_rx_filterq_entry_t *entry)
{
  uint8_t l2_addr[6];
  uint32_t ipv4_addr;
  int rc = -EINVAL;

  /* check if the l2 address is for an ipv4 multicast
   * (starts with 01:00:5E)
   *  and that it might be on a multicast special network
   *  224.0.0.0 to 224.0.0.255 or 224.0.1.0 to 224.0.1.255
   * (ends with :00:00:XX or :00:01:XX)
   *
   * If so, add a filter with an IPV4 multicast version of the address
  */
  memcpy(l2_addr, &entry->data[0], 6);

#if 0
  MX_INFO(("add_filter %02x:%02x:%02x:%02x:%02x:%02x\n", l2_addr[0], l2_addr[1], l2_addr[2], l2_addr[3], l2_addr[4], l2_addr[5]));
#endif

  if (l2_addr[0] == 0x01 &&
      l2_addr[1] == 0x00 &&
      l2_addr[2] == 0x5E &&
      l2_addr[3] == 0x0  &&
      (l2_addr[4] == 0x0 || l2_addr[4] == 0x1)) {

    ipv4_addr = ((0xe0 << 24) |
                 ((l2_addr[3] & 0x7f) << 16) |
                 (l2_addr[4] << 8) |
                 (l2_addr[5]));

    /* Fill in filter */
    entry->filt.ip_addr = ipv4_addr;
    entry->filt.port = 0; /* don't care */
    entry->filt.vlan = 0;
    entry->filt.tag = 0;
    entry->filt.remove = 0;
    entry->filt.kernel = 0;
    entry->filt.tcp = 0;
    entry->filt.txtstamp = 0;
    entry->filt.ab.enabled = 0;

    rc = __phx_write_filter_regs(is->phx.nic_mem,
				&is->phx.filter_sync,
				&entry->filt,
				is->phx.pci_info.func,
				PHX_FLTR_FLAG_BIT_IGNORE_VLAN |
				PHX_FLTR_FLAG_BIT_IGNORE_DST_PORT,
				NULL);
  }
  return rc;
}

void
phx_e1_init_board_ops(void)
{
  phx_e1_ops.init                  = phx_e1_init_board;
  phx_e1_ops.map                   = phx_common_map_board;
  phx_e1_ops.unmap                 = phx_common_unmap_board;
  phx_e1_ops.setup_hw_caps         = phx_common_setup_hw_caps;
  phx_e1_ops.enable_interrupt      = phx_common_int_enable;
  phx_e1_ops.disable_interrupt     = phx_common_int_disable;
  phx_e1_ops.park                  = phx_common_park_board;
  phx_e1_ops.get_counter_definitions  = phx_e1_get_counter_definitions;
  phx_e1_ops.counter_accum_index   = phx_e1_counter_accum_index;
  phx_e1_ops.init_endpoint_pool    = phx_e1_init_endpoint_pool;
  phx_e1_ops.ether_endpoint        = phx_e1_ether_endpoint;
  phx_e1_ops.physical_endpoint     = phx_e1_physical_endpoint;
  phx_e1_ops.set_intr_coal         = phx_e1_set_intr_coal;
  phx_e1_ops.set_mac_addr          = phx_common_set_mac_addr;
  phx_e1_ops.set_promisc           = phx_e1_set_promisc;
  phx_e1_ops.enable_all_multicast  = phx_e1_enable_all_multicast;
  phx_e1_ops.disable_all_multicast = phx_e1_disable_all_multicast;
  phx_e1_ops.remove_filter         = phx_e1_remove_filter;
  phx_e1_ops.add_filter            = phx_e1_add_filter;
}
