/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2014-2016 by CSP, Inc.  All rights reserved.                *
 *************************************************************************/

#ifndef _phx_ether_common_h_
#define _phx_ether_common_h_

#include "mal_auto_config.h"

/* Default receive ring sizes in MB */
#define MYRI_RX_DATA_RING_MB_MIN  (2<<20)  /* power of 2 */
#define MYRI_RX_DATA_RING_MB_MAX  (16<<20)
#define MYRI_RX_DESC_RING_MB_MIN  (2<<20)  /* power of 2 */
#define MYRI_RX_DESC_RING_MB_MAX  (16<<20)

#define MX_MAX_ETHER_MTU 9014

#define MX_DEFAULT_ETHER_MTU MX_MAX_ETHER_MTU

#if MYRI_ENABLE_PTP
#include "myri_ptp_common.h"
#endif

#define MX_ETH_STOPPED 0
#define MX_ETH_STOPPING 1
#define MX_ETH_STARTING 2
#define MX_ETH_RUNNING 3
#define MX_ETH_OPEN_FAILED 4

#include "phx_ether.h"
#include "mx_pio.h"
#include "mal_io.h"
#include "phx_hw.h"

typedef phx_pages_t phx_ether_pages_t;

/** Receive rings **/

typedef struct {
	unsigned int size; /* all of gamma pages in bytes */
	phx_ether_pages_t gamma; /* contains DMA data from NIC */
	phx_ether_pages_t beta;  /* 2nd level pages */
	phx_ether_pages_t alpha; /* 1st level pages */
} phx_ether_rx_ring_t;

typedef struct
{
	phx_ether_rx_ring_t ring;
	uint32_t data_off; /* where the next packet starts */
	uint32_t cons_off; /* consumer pointer last written to the NIC */
} phx_ether_rx_data_t;

typedef struct
{
	phx_ether_rx_ring_t ring;
	uint32_t desc_idx; /* where the next descriptor starts */
	uint32_t cons_idx; /* consumer pointer last written to the NIC */
	uint8_t seqnum; /* next sequence number to expect */
	uint32_t stat_wake_req_idx;
} phx_ether_rx_desc_t;

/** Send rings **/

typedef struct {
	uint32_t end_offset; /* last byte+1 including any padding */
	uint32_t flags;
	phx_tx_desc_t save_desc; /* for testing/debugging only */
} phx_ether_tx_queued_packet_t;

typedef struct {
	phx_ether_pages_t dma; /* DMA buffer */
	uint32_t dma_buffer_size;
	uint32_t dma_offset_host;    /* producer/head */
	uint32_t dma_offset_nic;     /* consumer/tail */
	/* | --- | ----- | ---- |
	 * S     nic     host   E
	 * S-nic: free space
	 * nic-host: queued but not sent packets
	 * host-E: free space
	 */

	uint32_t pio_buffer_size;
	uint32_t pio_offset_host;    /* producer/head */
	uint32_t pio_offset_nic;     /* consumer/tail */
	char *pio_fifo;     /* on the NIC */

	/* Descriptor ring on the NIC.  desc_count entries. */
	phx_tx_desc_t *desc_fifo;   /* on the NIC */
	uint32_t desc_idx_host;      /* producer/head */
	uint32_t desc_idx_nic;       /* consumer/tail */
	uint32_t desc_count;         /* # of descriptors */

	/* Send completion page */
	volatile phx_tx_completion_t *completion;

	/* Keep queued OS packets until the NIC notifies the driver. */
	phx_ether_tx_queued_packet_t *queued;

	/* 64B aligned descriptor */
	phx_tx_desc_t *aligned_desc;
	uint8_t desc_buf[64*2];

	/* FIXME.  Zero-filled bytes to deal with <64B packets */
	uint8_t zeros[64];

	/* 64B aligned, 64B lines buffer for PIO data.
	 * The size is the maximum PIO packet size.
	 */
	uint8_t *aligned_data;
	uint8_t data_buf[64 + MX_MAX_ETHER_MTU + 4];
} phx_ether_tx_t;

typedef struct {
	uint32_t data_ring_size;
	uint32_t data_offset_host;    /* producer/head */
	uint32_t data_offset_nic;     /* consumer/tail */
	/* | --- | ----- | ---- |
	 * S     nic     host   E
	 * S-nic: free space
	 * nic-host: queued but not sent packets
	 * host-E: free space
	 */

	uint32_t desc_idx_host;      /* producer/head */
	uint32_t desc_idx_nic;       /* consumer/tail */
	uint32_t desc_count;         /* # of descriptors */
	volatile uint32_t *prod_doorbell; /* K35SNF_TX_DESC_PRODW_ADVANCE */

        phx_rx_ring_t desc_ring;
        phx_rx_ring_t data_ring;
        struct {
		mx_page_pin_t pin;
		char *addr;
		volatile k35snf_tx_completion_t *count;
        } completion;
        int apu_data_ch;
        int apu_desc_ch;
} k35snf_ether_tx_t;

struct phx_ether {
	int board_type;
	struct phx_ether_arch arch;	/* OS dependent 	*/
	int running;			/* running? 		*/
	int rx_csum_offload;		/* rx_csums? 		*/
	int pause;			/* flow-control?        */
	mx_instance_state_t *is;	/* Hardware state 	*/
	uint8_t	current_mac[6];		/* software mac address */
	int mtu;                        /* OS MTU               */

	union {
		const struct phx_physical_ep *phx;
		struct {
			const struct phx_physical_ep *tx;
			const struct phx_physical_ep *rx;
			int stolen;
		} k35snf;
	} ep;

	struct phx_nic_mem_endpoint *ep_nic_mem;
	char *ep_pio_mem;
	int apu_data_ch;
	int apu_desc_ch;
	phx_ether_rx_data_t rx_data;
	phx_ether_rx_desc_t rx_desc;
	union {
		phx_ether_tx_t phx;
		k35snf_ether_tx_t k35snf;
	} tx;
	int use_interrupt;
	/* max number of descriptors to process per ether_rx_done().
	 * 0 means as many as possible.
	 */
	int rx_done_limit;
	uint32_t stat_irq;
	uint32_t stat_rx_irq;
	uint32_t stat_tx_irq;
	uint32_t stat_icu_enable;

	mx_spinlock_t spinlock;

	uint64_t stats_dropped_oversized;
	phx_endpt_flow_tx_stats_t *ep_tx_stats;

	/* PHX, K35 */
	uint32_t debug_last_tx_wake_request;
	uint32_t debug_last_tx_completion;
	/* K35SNF */
	uint32_t debug_last_tx_completion_packets;
	uint32_t debug_last_tx_completion_bytes;

	/* timestamp */
	int want_tx_ts;
	int want_rx_ts;

	TAILQ_HEAD(, myri_rx_filterq_entry) ether_mc_filter_list;
};

#define ETHER_K35SNF(eth) ((eth)->board_type == MAL_BOARD_TYPE_PHX_K35SNF)

int phx_ether_open_common(mx_instance_state_t *is, int use_interrupt, int mtu,
			  int rx_data_ring_size, int rx_desc_ring_size);
void phx_ether_set_mtu(struct phx_ether *eth, int mtu);
void phx_ether_close_common(struct phx_ether *eth);
int phx_ether_set_promisc_common(struct phx_ether *sc, int on);
void phx_ether_set_mac_address_common(struct phx_ether *sc, uint8_t *addr);
int  phx_ether_set_pause_common(struct phx_ether *eth, int pause);
int phx_ether_start_common(struct phx_ether *eth);

int phx_ether_attach(mx_instance_state_t *is);
void phx_ether_detach(mx_instance_state_t *is);
void phx_ether_irq(mx_instance_state_t *is, uint32_t intr_flags);
void phx_ether_link_change_notify(mx_instance_state_t *is);
void phx_ether_watchdog(mx_instance_state_t *is);
void phx_ether_tx_done(mx_instance_state_t *is);
void phx_ether_clean_sent_packets(struct phx_ether *eth);
int phx_ether_rx_done(mx_instance_state_t *is);
void phx_ether_receive_packet(struct phx_ether *eth, uint16_t pkt_len,
			      uint32_t ts_seconds, uint32_t ts_nanos);
void phx_ether_process_rx_desc(struct phx_ether *eth, phx_rx_desc_t *rx_desc);

#define myri_ptp_process_rx(is, pkt, nic_ticks)

#if MYRI_ENABLE_PTP
#undef myri_ptp_process_rx
void myri_ptp_process_rx(struct mx_instance_state *is, uint8_t *pkt, uint64_t nic_ticks);
#endif


void phx_copy_from_page_ring(phx_ether_pages_t *pages, uint32_t start_offset,
			     uint32_t len, char *dst);

static inline phx_rx_desc_t *
phx_ether_get_rx_desc_from_ring(struct phx_ether *eth)
{
	phx_ether_pages_t *pages = &eth->rx_desc.ring.gamma;
	uint32_t idx, page_idx, off;
	phx_rx_desc_t *desc;

	idx = eth->rx_desc.desc_idx;
	page_idx = (idx / (PAGE_SIZE / sizeof(phx_rx_desc_t))) % pages->n_pages;
	off = idx % (PAGE_SIZE / sizeof(phx_rx_desc_t));
	desc = ((phx_rx_desc_t*)pages->pages[page_idx].va) + off;
	return desc;
}

static inline void
phx_ether_consumed_rx_desc(struct phx_ether *eth)
{
	uint8_t seqnum = eth->rx_desc.seqnum;
	eth->rx_desc.desc_idx++;
	eth->rx_desc.seqnum = (seqnum == PHX_RX_DESC_SEQNUM_MAX ? 1 : seqnum + 1);
}

static inline uint32_t
phx_ether_rx_advance_data_offset(struct phx_ether *eth, uint32_t pkt_len)
{
	uint32_t aligned_len = (pkt_len + PHX_DBL_RX_PACKET_ALIGNMENT - 1)
		& ~(PHX_DBL_RX_PACKET_ALIGNMENT - 1);
	eth->rx_data.data_off += aligned_len;
	return aligned_len;
}

static inline void
phx_ether_pio_tx_desc(struct phx_ether *eth)
{
	/* 64B */
	uint64_t *d;
	uint64_t *s;

	/* Update software counters first */
	if ((eth->tx.phx.aligned_desc->flags & PHX_TX_DESC_FLAG_DMA)) {
		eth->ep_tx_stats->tx_dma_packets++;
		eth->ep_tx_stats->tx_dma_bytes
			+= eth->tx.phx.aligned_desc->data_length
			+ eth->tx.phx.aligned_desc->header_length;
	} else {
		eth->ep_tx_stats->tx_pio_packets++;
		eth->ep_tx_stats->tx_pio_bytes
			+= eth->tx.phx.aligned_desc->data_length
			+ eth->tx.phx.aligned_desc->header_length;
	}
	d = (uint64_t*)eth->tx.phx.desc_fifo;
	s = (uint64_t*)eth->tx.phx.aligned_desc;

	*d++ = *s++;
	*d++ = *s++;
	*d++ = *s++;
	*d++ = *s++;
	*d++ = *s++;
	*d++ = *s++;
	*d++ = *s++;
	*d++ = *s++;
	MAL_STBAR();
}

typedef struct {
	char *vaddr;
	int len;
} phx_ether_pkt_frag_t;

/* Copy len bytes from the head into dst.  If frags have less than
 * len bytes, this function copies as many bytes as available.
 */
void phx_ether_copy_pkthdr(phx_ether_pkt_frag_t frags[], int n_frags, char *dst, int len);

/* Allocate len bytes in the PIO buffer.  start_offset will contain the offset
 * within the PIO buffer where the packet should start.  Return 0 if success
 * and -1 if out of space.
 */
int phx_ether_tx_alloc_piobuf(struct phx_ether *eth, int len, uint32_t *start_offset);

/* Copy len bytes starting at data_offset into the PIO buffer starting at buf_offset */
int phx_ether_tx_copy_pkt_piobuf(struct phx_ether *eth, phx_ether_pkt_frag_t frags[],
				 int n_frags, int data_offset, int len, int buf_offset);

/* Just like the PIO version, allocate len bytes in the DMA buffer. */
int phx_ether_tx_alloc_dmabuf(struct phx_ether *eth, int len, uint32_t *start_offset);

/* Copy len bytes starting at data_offset into the DMA buffer starting at buf_offset */
int phx_ether_tx_copy_pkt_dmabuf(struct phx_ether *eth, phx_ether_pkt_frag_t frags[],
				 int n_frags, int data_offset, int len, int buf_offset);

void phx_ether_rx_advance_cons(struct phx_ether *eth);

/* Enable all multicast packets */
int phx_ether_enable_all_multicast(struct phx_ether *eth);

/* Disable all multicast packets = allow matched */
int phx_ether_disable_all_multicast(struct phx_ether *eth);

/* Remove all multicast filters on our list */
int phx_ether_leave_all_multicast(struct phx_ether *eth);

/* Add a multicast filter to our list and define a filter for it */
int phx_ether_join_multicast(struct phx_ether *eth, uint32_t addr_hi, uint32_t addr_lo);

#if MAL_OS_LINUX
int phx_ether_set_discard_whitelist(mx_instance_state_t *is,
				    phx_class_d_discard_whitelist_t *list);
#endif

void k35snf_ether_irq(mx_instance_state_t *is, uint32_t intr_flags);
int k35snf_ether_open_common(mx_instance_state_t *is, int mtu,
			     int rx_data_ring_size, int rx_desc_ring_size);
void k35snf_ether_close_common(struct phx_ether *eth);
int k35snf_ether_start_common(struct phx_ether *eth);
void k35snf_ether_tx_done(mx_instance_state_t *is);
void k35snf_ether_clean_sent_packets(struct phx_ether *eth);
void k35snf_ether_process_rx_desc(struct phx_ether *eth, k35snf_rx_desc_t *rx_desc);
void k35snf_ether_receive_packet(struct phx_ether *eth, uint16_t pkt_len, uint64_t ts_nanos);
int k35snf_ether_tx_copy_pkt(struct phx_ether *eth, phx_ether_pkt_frag_t frags[],
			     int n_frags, uint32_t len, uint32_t buf_offset);

static inline k35snf_tx_desc_t *
k35snf_ether_next_tx_desc(k35snf_ether_tx_t *tx)
{
	phx_ether_pages_t *pages = &tx->desc_ring.gamma;
	uint32_t idx, page_idx, off;

	idx = tx->desc_idx_host % tx->desc_count;
	page_idx = (idx / (PAGE_SIZE / sizeof(k35snf_tx_desc_t))) % pages->n_pages;
	off = idx % (PAGE_SIZE / sizeof(k35snf_tx_desc_t));
	return ((k35snf_tx_desc_t*)pages->pages[page_idx].va) + off;
}

static inline k35snf_rx_desc_t *
k35snf_ether_get_rx_desc_from_ring(struct phx_ether *eth)
{
	phx_ether_pages_t *pages = &eth->rx_desc.ring.gamma;
	uint32_t idx, page_idx, off;
	k35snf_rx_desc_t *desc;

	idx = eth->rx_desc.desc_idx;
	page_idx = (idx / (PAGE_SIZE / sizeof(k35snf_rx_desc_t))) % pages->n_pages;
	off = idx % (PAGE_SIZE / sizeof(k35snf_rx_desc_t));
	desc = ((k35snf_rx_desc_t*)pages->pages[page_idx].va) + off;
	return desc;
}

static inline uint32_t
k35snf_ether_rx_advance_data_offset(struct phx_ether *eth, uint32_t pkt_len)
{
	uint32_t aligned_len = K35SNF_ALIGN_TX_PACKET_LENGTH(pkt_len);
	eth->rx_data.data_off += aligned_len;
	return aligned_len;
}

#endif /* _phx_ether_common_h_ */

/*
  This file uses MX driver indentation.

  Local Variables:
  c-file-style:"linux"
  tab-width:8
  indent-tabs-mode:t
  End:
*/
