/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2014-2016 by CSP, Inc.  All rights reserved.                *
 *************************************************************************/

#include "mx_arch.h"
#include "mx_misc.h"
#include "mx_instance.h"
#include "mx_malloc.h"
#include "mx_pio.h"
#include "mal_stbar.h"
#include "myri_version.h"
#include "mal_cpu.h"
#include "phx_hw.h"
#include "phx_common.h"
#include "phx_avl.h"
#include "phx_ether_common.h"

/* Temporary until we have a more general logging/verbosity control */
#if MAL_DEBUG
#define _VERBOSE 1
#else
#define _VERBOSE 0
#endif

mx_board_ops_t phx_k35_ops = { NULL };

/*
 * The xilinx mac powers up with it's receiver enabled.
 * DBL uses an FPGA image that has a inhibit which allows incoming traffic to be tossed.
 */
void phx_k35_mac_disable(mx_instance_state_t *is)
{
  uint8_t *r;
  uint32_t rx_inhibit;
  uint32_t port;

  /* have the mac disable traffic */
  if (is->mac_ops.mac_disable)
    is->mac_ops.mac_disable(is);

  /* what port is this? */
  r = (uint8_t*)is->phx.nic_mem + PHX_FLTR_BASE;
  port = PHX_REG_READ_4((r + PHX_FLTR_PORT));

  /* the inhibit is a bit in a global FPGA register, set it */
  r = (uint8_t*)is->phx.nic_mem + PHX_I2C_BASE;
#if _VERBOSE
  MX_INFO(("%s: Disable port %d for receiving\n", __FUNCTION__, port));
#endif
  rx_inhibit = PHX_REG_READ_4((r + K35_MAC_RESET));

  /* inhibit at the gearbox and hold rx/tx in reset */
  PHX_REG_WRITE_4((r + K35_MAC_RESET), (rx_inhibit | (0x10011 << port)));
  rx_inhibit = PHX_REG_READ_4((r + K35_MAC_RESET));
}

void phx_k35_mac_enable(mx_instance_state_t *is)
{
  uint8_t *r;
  uint32_t rx_inhibit;
  uint32_t port;

  /* what port is this? */
  r = (uint8_t*)is->phx.nic_mem + PHX_FLTR_BASE;
  port = PHX_REG_READ_4((r + PHX_FLTR_PORT));

  /* the inhibit is a bit in a global FPGA register, clear it */
  r = (uint8_t*)is->phx.nic_mem + PHX_I2C_BASE;
#if _VERBOSE
  MX_INFO(("%s: Enable port %d for receiving\n", __FUNCTION__, port));
#endif
  rx_inhibit = PHX_REG_READ_4((r + K35_MAC_RESET));

  /* uninhibit at the gearbox and release rx/tx reset */
  PHX_REG_WRITE_4((r + K35_MAC_RESET), (rx_inhibit & ~(0x10011 << port)));
  rx_inhibit = PHX_REG_READ_4((r + K35_MAC_RESET));

  /* allow the MAC to receive traffic */
  if (is->mac_ops.mac_enable)
    is->mac_ops.mac_enable(is);
}

#define PLUG_COPPER_PIGTAIL 0x21
#define PLUG_FIBER 0x7

void phx_k35_mac_set_preemphasis(mx_instance_state_t *is)
{
  uint32_t preemphasis=K35_SERDES_10G_FIBER;
  uint32_t connector;
  uint32_t port = is->port_id;

  /* determine the connector type */
  connector = phx_get_transceiver_type(is, port);

  if (connector == PLUG_COPPER_PIGTAIL)
    preemphasis=K35_SERDES_10G_COPPER;

  phx_set_serdes_preemphasis(is, port, preemphasis);
}

void phx_k35_mac_set_max_frame_length(struct phx_nic_mem_global *nic, int mtu)
{
  uint8_t *r = (uint8_t*)nic + PHX_XLX_MAC_BASE;
  uint32_t tx_max, rx_max;

#if _VERBOSE
  MX_INFO(("%s:\n", __FUNCTION__));
  tx_max = PHX_REG_READ_4((r + PHX_XLX_MAC_TX_MTU_CONFIG));
  tx_max &= 0x7fff;
  rx_max = PHX_REG_READ_4((r + PHX_XLX_MAC_RX_MTU_CONFIG));
  rx_max &= 0x7fff;
  MX_INFO(("current: tx_max=%u rx_max=%u\n", tx_max, rx_max));
#endif

#ifdef PHX_WORKAROUND_MAC_OVERSIZE_FRAMES
  /* Set lengths to be some large so the NIC never sees
   * oversized frames.
   */
  tx_max = 0x10000 + 10000;  /* enable for MTU reg, and 10000 limit */
  PHX_REG_WRITE_4((r + PHX_XLX_MAC_TX_MTU_CONFIG), tx_max);
  rx_max = 0x10000 + 10000;
  PHX_REG_WRITE_4((r + PHX_XLX_MAC_RX_MTU_CONFIG), rx_max);
#else
  /* "Length" in the Altera MAC covers [Ethernet header ... CRC32].
   * The MAC understands VLAN tagged packets and uses length + 4 when
   * checking the VLAN packet's length.
   *
   * TX length does not affect transmission at all.  It is only used
   * for statistics in the MAC.
   */
  /* Add 14 (ethernet header) and 4 (crc32) to the OS MTU. */
  if (mtu > 1500) {
    mtu |= 0x10000;  /* enable bit for MTU reg */
    PHX_REG_WRITE_4((r + PHX_XLX_MAC_TX_MTU_CONFIG), (mtu + 18));
    PHX_REG_WRITE_4((r + PHX_XLX_MAC_RX_MTU_CONFIG), (mtu + 18));
  }
#endif

#if _VERBOSE
  tx_max = PHX_REG_READ_4((r + PHX_XLX_MAC_TX_MTU_CONFIG));
  tx_max &= 0x7fff;
  rx_max = PHX_REG_READ_4((r + PHX_XLX_MAC_RX_MTU_CONFIG));
  rx_max &= 0x7fff;
  MX_INFO(("current: tx_max=%u rx_max=%u\n", tx_max, rx_max));
#endif
}

/*
 * Check that both PPS inputs aren't active on the K35
 * and issue a warning if they are.
 */
void phx_k35_check_pps(mx_instance_state_t *is)
{
  uint32_t edge_pps, panel_pps;
  uint8_t *edge_pps_reg, *panel_pps_reg;

  edge_pps_reg  = (uint8_t *)is->phx.nic_mem + K35_EDGE_PPS;
  panel_pps_reg = (uint8_t *)is->phx.nic_mem + K35_PANEL_PPS;

  edge_pps = PHX_REG_READ_4(edge_pps_reg);
  panel_pps = PHX_REG_READ_4(panel_pps_reg);

  if ((edge_pps > 0) && (panel_pps > 0))
  {
    MX_WARN(("Both PPS connectors are active!\n"));
  }

}

static int
phx_k35_set_intr_coal(mx_instance_state_t *is, int microseconds)
{
  if (PHX_HAS_FEATURE(is, PHX_FEATURE_INTR_INTERVAL)) {
    /* 6.4 nsec per cycle */
    uint32_t cycles = ((microseconds * 10000) / 64);
    uint8_t *r = (uint8_t*)is->phx.nic_mem + PHX_ICU_BASE;
    r += PHX_ICU_INTR_INTERVAL;
    PHX_REG_WRITE_4(r, cycles);
    is->intr_coal_usec = microseconds;
  }
  return is->intr_coal_usec;
}
    
void phx_k35_mac_set_default_mode(struct phx_nic_mem_global *nic)
{
  uint8_t *r = (uint8_t*)nic + K35_MAC_FILTER_BASE;
  uint32_t control = 0;

  /* control |= K35_MAC_FILTER_MODE_ALL_UNIICAST;  No, only our unicast */ 
  control |= K35_MAC_FILTER_MODE_ALL_BROADCAST;
  control |= K35_MAC_FILTER_MODE_ALL_MULTICAST;    /* FIXME - remove when done K35DBL-48 */

  PHX_REG_WRITE_4((r + K35_MAC_FILTER_MODE), control);
}

void phx_k35_mac_set_promisc_mode(struct phx_nic_mem_global *nic, int enable)
{
  uint8_t *r = (uint8_t*)nic + K35_MAC_FILTER_BASE;
  uint32_t control;

  control = PHX_REG_READ_4((r + K35_MAC_FILTER_MODE));
  if (enable)
    control |= K35_MAC_FILTER_MODE_PROMISC;
  else
    control &= ~K35_MAC_FILTER_MODE_PROMISC;

  PHX_REG_WRITE_4((r + K35_MAC_FILTER_MODE), control);
}

void phx_k35_mac_set_multicast_mode(struct phx_nic_mem_global *nic, int enable)
{
  uint8_t *r = (uint8_t*)nic + K35_MAC_FILTER_BASE;
  uint32_t control = 0;

  control = PHX_REG_READ_4((r + K35_MAC_FILTER_MODE));
  if (enable)
    control |= K35_MAC_FILTER_MODE_ALL_MULTICAST;
  else
    control &= ~K35_MAC_FILTER_MODE_ALL_MULTICAST;

  PHX_REG_WRITE_4((r + K35_MAC_FILTER_MODE), control);
}

int phx_k35_set_mac_filter(struct phx_nic_mem_global *nic, uint32_t addr_high, uint32_t addr_low, int entry)
{
  uint8_t *r = (uint8_t*)nic + K35_MAC_FILTER_BASE;
  uint8_t *high, *low;

  if (entry < K35_NUM_MAC_FILTERS){
    high = r + (entry * K35_MAC_FILTER_SIZE) + K35_MAC_FILTER_HIGH;
    low = r + (entry * K35_MAC_FILTER_SIZE) + K35_MAC_FILTER_LOW;
    PHX_REG_WRITE_4(high, addr_high);
    PHX_REG_WRITE_4(low, addr_low);
    return 0;
  }
  return -1;
}

void phx_k35_set_mac_addr(mx_instance_state_t *is, uint8_t mac[6])
{
  struct phx_nic_mem_global *nic= is->phx.nic_mem;
  uint32_t addr0, addr1;

/*#if _VERBOSE*/
#if 1
  MX_INFO(("%s: %02x:%02x:%02x:%02x:%02x:%02x\n", __FUNCTION__,
           mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]));
#endif

  addr0 = ((uint32_t)mac[0] << 24) | ((uint32_t)mac[1] << 16)
      | ((uint32_t)mac[2] << 8) | ((uint32_t)mac[3]);
  addr1 = ((uint32_t)mac[4] << 24) | ((uint32_t)mac[5] << 16);

  /* slot 0 used for our unicast MAC address, by convention */
  phx_k35_set_mac_filter(nic, addr0, addr1, 0);

  is->phx.l2_filters[0].high = addr0;
  is->phx.l2_filters[0].low = addr1;
  is->phx.l2_filters[0].used = 1;
}

/* clear all mac filter registers */
int phx_k35_clear_mac_filters(mx_instance_state_t *is)
{
  int i;
  phx_mac_filter_t *l2_filt;

  /* find this filter */
  for (i = 0; i < is->phx.num_l2_filters; i++){
    l2_filt = &is->phx.l2_filters[i];
    l2_filt->high = l2_filt->low = 0;
    l2_filt->used = 0;
    phx_k35_set_mac_filter(is->phx.nic_mem, l2_filt->high, l2_filt->low, i);
  }

  return 0;
}

const unsigned char *hardcoded_eeprom_strings="MAC=00:60:dd:aa:bb:40\0SN=ENE141203006790\0PC=10G-PCIE3-8D-2X\0PN=09-04642\0BOM=1.5.1.2.0\0";

void phx_read_hardcoded_specs(char *base, char *specs, size_t len)
{
        memcpy(specs, hardcoded_eeprom_strings, 256);  /* FIXME - no eeprom yet */
}

static int
phx_k35_init_board(mx_instance_state_t *is)
{
  uint32_t version=0;
  uint32_t board_arch=0;
  int init_fpga;
  int attempts = 0;

  /* The NIC has been mapped earlier.  So it is safe to access NIC
   * registers via nic_mem.
   */

  mx_mutex_enter(&mx_fpga_mutex);
  init_fpga = !is->phx.pfi->initialized;

  /* Disable the MAC TX/RX */
  phx_k35_mac_disable(is);

  /* stop DblRx */
  phx_dbl_rx_global_stop(is);

#if !MAL_OS_WINDOWS
  /* On Windows, the driver calls phx_win_init_fpga() early on
   * prior to this function.
   */
  if (init_fpga) {
    /* read the specs from the EEPROM to get MAC address */
    MX_INFO(("Initializing FPGA: start\n"));

    phx_xilinx_init((char*)is->phx.i2c_base_mem, &version, &board_arch);

    /* firmware monotonic version for compatibility checks */
    is->phx.pfi->firmware_rev = version;
    is->phx.pfi->xilinx_board_type = board_arch;
    MX_INFO(("Initializing FPGA: end, firmware version %d board arch %c%c%c%c\n", version, 
              ((board_arch & 0xff000000) >> 24),
              ((board_arch & 0xff0000) >> 16),
              ((board_arch & 0xff00) >> 8),
              (board_arch & 0xff)));

#if PHX_ENABLE_PTP
    /* initialize the PTP clock */
    phx_ptp_init(is->phx.pfi);

    is->phx.pfi->ptp_clock = ptp_clock_register(&is->phx.pfi->ptp_info,
           &is->arch.pci_dev->dev);
    if(IS_ERR(is->phx.pfi->ptp_clock)){
       is->phx.pfi->ptp_clock = NULL;
    }
    phx_k35_check_pps(is);
#endif

    /* read the eeprom strings from the onboard eeprom */
    if (is->phx.pfi->eeprom_strings)
      phx_read_specs((char*)is->phx.i2c_base_mem, is->phx.pfi->eeprom_strings, 256,
                   is->phx.pfi->xilinx_board_type);
  }
#endif

  if (init_fpga) {
    phx_clk_init(is, myri_clk_invert_pps, !myri_clk_enable_pps,
		 !myri_clk_enable_10mhz);
  }

  /* private copy the eeprom strings gets modified by parsing  */
  if ((is->phx.pfi->eeprom_strings) && (is->phx.eeprom_strings))
    memcpy(is->phx.eeprom_strings, is->phx.pfi->eeprom_strings, 256);

  /* This sets MAC address (is->mac_addr).
   * PHX-310 retry reading if copy if not OK.
   */
  while (phx_parse_eeprom_strings(is) && (attempts++ < 10)){
    MX_INFO(("FATAL error could not process eeprom strings\n"));
    phx_read_specs((char*)is->phx.i2c_base_mem, is->phx.eeprom_strings, 256,
                   is->phx.pfi->xilinx_board_type);
  }

  MX_INFO(("Initializing MAC/PHY\n"));
  /* Set the default MTU.  The max frame length seems to be sticky.
   * The setting survives across reboots.  So explicitly set it to
   * the default value.
   */
  phx_k35_mac_set_max_frame_length(is->phx.nic_mem,
                               MX_DEFAULT_ETHER_MTU - 14 /* 1500 */);

  is->phx.num_l2_filters = K35_NUM_MAC_FILTERS;
  phx_k35_clear_mac_filters(is);
  phx_k35_set_mac_addr(is, is->mac_addr);
  phx_k35_mac_set_default_mode(is->phx.nic_mem);
  phx_k35_mac_set_promisc_mode(is->phx.nic_mem, 0);
  is->link_speed = 10000000000;   /* 10^10 bps */
  phx_k35_mac_set_preemphasis(is);

  /* Tell the NIC to use the MRR in the PCI configuration.  Otherwise,
   * it always uses 128B, which leads to ~4Gbps maximum DMA send
   * throughput.
   */
  phx_tx_enable_auto_read_request(is->phx.nic_mem);

  phx_rx_selector_reset(is->phx.nic_mem);

  if (init_fpga) {
    /* global tx stuff */
    if (PHX_HAS_FEATURE(is, PHX_FEATURE_8K_PIO)) {
      is->phx.pfi->tx_config.dbltx_core_base = PHX_REG_TX_BASE;
      is->phx.pfi->tx_config.dbltx_core_stride = 1 * PAGE_SIZE;
      is->phx.pfi->tx_config.dbltx_core_length = 1 * PAGE_SIZE;
      is->phx.pfi->tx_config.dbltx_pio_base = 0x180000;
      is->phx.pfi->tx_config.dbltx_pio_stride = 4 * PAGE_SIZE;
      is->phx.pfi->tx_config.dbltx_pio_length = 2 * PAGE_SIZE;
      is->phx.pfi->tx_config.dmasend_threshold = myri_snf_dmasend_threshold;
      /* Completion update threshold */
      phx_tx_threshold(is->phx.nic_mem, myri_tx_threshold);
   } else {
      is->phx.pfi->tx_config.dbltx_core_base = PHX_REG_TX_BASE;
      is->phx.pfi->tx_config.dbltx_core_stride = 2 * PAGE_SIZE;
      is->phx.pfi->tx_config.dbltx_core_length = 1 * PAGE_SIZE;
      is->phx.pfi->tx_config.dbltx_pio_base = PHX_REG_TX_BASE + 0x1000;
      is->phx.pfi->tx_config.dbltx_pio_stride = 2 * PAGE_SIZE;
      is->phx.pfi->tx_config.dbltx_pio_length = 1 * PAGE_SIZE;
      is->phx.pfi->tx_config.dmasend_threshold = PHX_DMASEND_THRESHOLD_SIZE;

      /* Completion update threshold */
      phx_tx_threshold(is->phx.nic_mem, PHX_TX_THRESHOLD);
   }
   phx_counters_tx_fixup(is);    /* set the stride for Tx endpoint counters */

   /* global rx stuff */
   /* phx_dbl_rx_r0_enable_push(is->phx.nic_mem); */

#ifdef MYRI_DBL_AB
    /* global A/B stuff */
    phx_dbl_ab_init(is->phx.nic_mem);
#endif

    /* Set the send completion page address for DBL TX.  This is a
     * global setting, not per endpoint.  Both the kernel endpoint and
     * libdbl endpoints use this same page.
     */
    phx_write_reg_tx_host_completion(is->phx.nic_mem,
				     is->tx_done.pin.dma.low,
				     is->tx_done.pin.dma.high);
  }

  /* Rx packet coalescing thresholds */
  phx_dbl_rx_packet_count_limit(is, myri_rx_coalesce_pkt);
  phx_dbl_rx_time_limit(is, myri_rx_coalesce_time);
  phx_dbl_rx_length_limit(is, myri_rx_coalesce_len);

  /* Filter is local, per-function */
  phx_dbl_filter_init(is, 0);

#ifdef MYRI_DBL_AB
  phx_dbl_ab_local_init(is->phx.nic_mem);
#endif

  if (PHX_HAS_FEATURE(is, PHX_FEATURE_8K_PIO))
    phx_rx_selector_bypass_disable(is->phx.nic_mem, 0);
  else
    phx_rx_selector_bypass_disable(is->phx.nic_mem, 1);


  /* Enable interrupts */
  phx_icu_enable(is);

  if (init_fpga) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0)
    struct timespec64 ts;
#else
    struct timespec ts;
#endif
    mx_getnstimeofday(&ts);
    if (phx_clk_is_pps_active(is)) {
      phx_clk_set_with_pps(is, &ts);
      MX_INFO(("PPS is active, wrote the seconds to the adapter\n"));
      is->clk_ts_state = MYRI_CLK_TS_EXTERNAL_PPS;
    } else {
      phx_clk_set(is, &ts);
      MX_INFO(("PPS is inactive, wrote the time to the adapter\n"));
      is->clk_ts_state = MYRI_CLK_TS_INTERNAL_TCXO;
    }
  }

  /* Clear the counters for the first time.  We will read all the registers
   * and save them in the process.
   */
  myri_counters64_clear(is);

  if (init_fpga)
    is->phx.pfi->initialized = 1;

  /* Start DblRx before enabling the MAC */
  phx_dbl_rx_global_start(is);

  /* Enable the local MAC */
  phx_k35_mac_enable(is);

  /* Interrupt coalescing */
  /* Does this board support interrupt coal? */
  is->intr_coal_usec = 0;
  if (PHX_HAS_FEATURE(is, PHX_FEATURE_INTR_INTERVAL))
    phx_k35_set_intr_coal(is, myri_intr_coal_delay);

  mx_mutex_exit(&mx_fpga_mutex);
  return 0;
}

void
phx_k35_park_board(mx_instance_state_t *is)
{
  phx_k35_mac_disable(is);    /* disable the MAC */
}

static int
phx_k35_init_endpoint_pool(mx_instance_state_t *is, struct phx_fpga_endpoint_pool *pool)
{
  int i;
  struct phx_physical_ep *ep;

  pool->num_function_endpoints = pool->num_function_not_reserved_endpoints = is->phx.apu_config.endpoints;
  pool->num_total_endpoints = is->phx.apu_config.endpoints * is->phx.num_functions;
  pool->endpoints = mx_kmalloc(sizeof(struct phx_physical_ep) * pool->num_total_endpoints, MX_MZERO);
  if (!pool->endpoints)
    return -1;
  for (i = 0; i < pool->num_total_endpoints; i++) {
    ep = &pool->endpoints[i];
    ep->id = i;
    ep->owner_is = NULL;
    ep->reserved = 0;
    ep->rx_ep = i % pool->num_function_endpoints;  /* 0 ... N-1 */
    ep->tx_ep = i;
    ep->icu = ep->rx_ep + 1;
    ep->func = i / pool->num_function_endpoints;   /* N endpoints each */
    /* reserve Kernel endpoint (0) */
    if ((ep->rx_ep == 0) && (myri_disable_eth == 0)){
      ep->reserved = 1;
      pool->num_function_not_reserved_endpoints--;
    }
    /* reserve sniffer endpoint (1) */
    if (ep->rx_ep == 1){
      ep->reserved = 1;
      pool->num_function_not_reserved_endpoints--;
    }
  }

  return 0;
}

static struct phx_physical_ep *
phx_k35_ether_endpoint(mx_instance_state_t *is)
{
  struct phx_fpga_endpoint_pool *pool = &is->phx.pfi->ep_pool;
  return &pool->endpoints[is->phx.pci_info.func * pool->num_function_endpoints];
}

static struct phx_physical_ep *
phx_k35_sniffer_endpoint(mx_instance_state_t *is)
{
  struct phx_fpga_endpoint_pool *pool = &is->phx.pfi->ep_pool;
  return &pool->endpoints[is->phx.pci_info.func * pool->num_function_endpoints + 1];
}

/* n'th within the function */
static struct phx_physical_ep *
phx_k35_physical_endpoint(mx_instance_state_t *is, int nth)
{
  struct phx_fpga_endpoint_pool *pool = &is->phx.pfi->ep_pool;
  return &pool->endpoints[nth + is->phx.pci_info.func * pool->num_function_endpoints];
}

/* K35: DblRx is aliased, per-function */
static phx_counter_t phx_k35_counters[] = {
  { "Cycle Count", PHX_DBL_RX_BASE + PHX_DBL_RX_CYCLE_COUNT,               CNT_ALIASED_HEX,     0},
  { "Driver Uptime", 0,                                                    CNT_DRV_UPTIME, 0},
  { "Counters Uptime", 0,                                                  CNT_UPTIME, 0},
  { "Packets Received", PHX_DBL_RX_BASE + PHX_DBL_RX_PACKET_COUNT,         CNT_ALIASED_64,     0},
  { "Bytes Received", PHX_DBL_RX_BASE + PHX_DBL_RX_BYTE_COUNT,             CNT_ALIASED_64,     0},
  { "Filter Packets Rx", PHX_FLTR_BASE + PHX_FLTR_PKTS_SENT,               CNT_ALIASED,     0},
  { "Filter Bytes Rx", PHX_FLTR_BASE + PHX_FLTR_BYTES_SENT,                CNT_ALIASED,     0},
  { "Packets Drop Filter", PHX_FLTR_BASE + PHX_FLTR_PKTS_DROPPED,          CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "MAC Receive Errors", PHX_FLTR_BASE + PHX_FLTR_PKTS_ABORTED,           CNT_ALIASED,     0},
  { "Packets Filtered", PHX_FLTR_BASE + PHX_FLTR_PKTS_FILTERED,            CNT_ALIASED,     0},
  { "Packets Unfiltered", PHX_FLTR_BASE + PHX_FLTR_PKTS_UNFILTERED,        CNT_ALIASED,     0},
  { "Max Filter Latency", PHX_FLTR_BASE + PHX_FLTR_PRT_MAX_LATENCY,        CNT_ALIASED | COUNTER_TYPE_SNAPSHOT,     0},
  { "Packets Skip Filter HW", PHX_FLTR_BASE + PHX_FLTR_PKTS_SKIPPED_FILTER,   CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Packets Drop Filter HW", PHX_FLTR_BASE + PHX_FLTR_PKTS_DROP_FILTER,   CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Packets Rx Filter", PHX_FLTR_BASE + PHX_FLTR_PKTS_RECEIVED,           CNT_ALIASED,     0},
  { "Packets Drop Incoming", PHX_FLTR_BASE + PHX_FLTR_INC_PACKETS_DROPPED, CNT_ALIASED,     0},
  { "FF Packets Received", PHX_DBL_RX_BASE + PHX_DBL_RX_FF_PACKET_COUNT,          CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "FF Bytes Received", PHX_DBL_RX_BASE + PHX_DBL_RX_FF_BYTE_COUNT,              CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "FD Packets Discarded", PHX_DBL_RX_BASE + PHX_DBL_RX_FD_PACKET_COUNT,         CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "FD Bytes Discarded", PHX_DBL_RX_BASE + PHX_DBL_RX_FD_BYTE_COUNT,             CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "PCIE FIFO empty", PHX_DBL_RX_BASE + PHX_DBL_RX_BP_EMPTY,                     CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "PCIE FIFO < 1/4", PHX_DBL_RX_BASE + PHX_DBL_RX_BP_Q1,                        CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "PCIE FIFO 1/4 - 1/2", PHX_DBL_RX_BASE + PHX_DBL_RX_BP_Q2,                    CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "PCIE FIFO 1/2 - 3/4", PHX_DBL_RX_BASE + PHX_DBL_RX_BP_Q3,                    CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "PCIE FIFO > 3/4", PHX_DBL_RX_BASE + PHX_DBL_RX_BP_Q4,                        CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "PCIE FIFO full", PHX_DBL_RX_BASE + PHX_DBL_RX_BP_FULL,                       CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "Filter Kernel Endpoint", PHX_FLTR_BASE + PHX_FLTR_KERNEL_ENDPOINT,           CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "Filter Lookup Count", PHX_FLTR_BASE + PHX_FLTR_LOOKUP_CNT,                   CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Filter Lookup Fail", PHX_FLTR_BASE + PHX_FLTR_LOOKUP_FAIL,                   CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Filter Lookup Search", PHX_FLTR_BASE + PHX_FLTR_LOOKUP_SEARCH,               CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Filter Add Error", PHX_FLTR_BASE + PHX_FLTR_ADD_ERROR,                       CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "Filter Remove Error", PHX_FLTR_BASE + PHX_FLTR_REMOVE_ERROR,                 CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Packets Received", PHX_FLTR_BASE + PHX_FLTR_RP_PKTS_RECV,      CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Packets Sent", PHX_FLTR_BASE + PHX_FLTR_RP_PKTS_SENT,          CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Bytes Sent", PHX_FLTR_BASE + PHX_FLTR_RP_BYTES_SENT,           CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Packets Dropped", PHX_FLTR_BASE + PHX_FLTR_RP_PKTS_DROPPED,    CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Sent Duplicate", PHX_FLTR_BASE + PHX_FLTR_RP_SENT_DUP,         CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Sent Abort", PHX_FLTR_BASE + PHX_FLTR_RP_SENT_ABORT,           CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Dropped Abort", PHX_FLTR_BASE + PHX_FLTR_RP_DROPPED_ABORT,     CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Dropped Control", PHX_FLTR_BASE + PHX_FLTR_RP_DROPPED_CONTROL, CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Dropped Meta Data", PHX_FLTR_BASE + PHX_FLTR_RP_DROPPED_META,  CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Control Full", PHX_FLTR_BASE + PHX_FLTR_RP_CONTROL_FULL,       CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Receive Abort", PHX_FLTR_BASE + PHX_FLTR_RP_RECV_ABORT,        CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Replay Buffer Meta Error", PHX_FLTR_BASE + PHX_FLTR_RP_META_ERROR,           CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Filter Add Slots",  PHX_FLTR_BASE + PHX_FLTR_ADD_SLOTS,               CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE,     0},
  { "Filter Used Slots",  PHX_FLTR_BASE + PHX_FLTR_FREE_SLOTS,
                 CNT_ALIASED | COUNTER_TYPE_16_HI | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE, 0},
  { "Filter Free Slots",  PHX_FLTR_BASE + PHX_FLTR_FREE_SLOTS,
                 CNT_ALIASED | COUNTER_TYPE_16_LO | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE, 0},
  { "RX Packet Count Limit", PHX_DBL_RX_BASE + PHX_DBL_RX_PACKET_COUNT_LIMIT,     CNT_ALIASED | COUNTER_TYPE_SNAPSHOT  | COUNTER_DISPLAY_VERBOSE,     0},
  { "RX Length Limit", PHX_DBL_RX_BASE + PHX_DBL_RX_LENGTH_LIMIT,                 CNT_ALIASED_HEX | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE, 0},
  { "RX Time Limit", PHX_DBL_RX_BASE + PHX_DBL_RX_TIME_LIMIT,                     CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE,     0},
  { "Packet Overflow", PHX_DBL_RX_BASE + PHX_DBL_RX_PACKET_OVERFLOW,              CNT_ALIASED | COUNTER_DISPLAY_VERBOSE,     0},
  { "Descriptor Overflow", PHX_DBL_RX_BASE + PHX_DBL_RX_DESC_OVERFLOW,            CNT_ALIASED | COUNTER_DISPLAY_VERBOSE,     0},
  { "TX Pause MAC CTRL frames", PHX_MAC_BASE + PHX_MAC_TX_PAUSE_MACCTRL,          CNT_ALIASED | COUNTER_TYPE_SELFCLEAR | COUNTER_DISPLAY_VERBOSE,     0},
  { "RX Pause MAC CTRL frames", PHX_MAC_BASE + PHX_MAC_RX_PAUSE_MACCTRL,          CNT_ALIASED | COUNTER_TYPE_SELFCLEAR | COUNTER_DISPLAY_VERBOSE,     0},
  { "APU Version ", PHX_APU_BASE + PHX_APU_ID_VERSION,                            CNT_ALIASED_HEX | COUNTER_TYPE_SNAPSHOT  | COUNTER_DISPLAY_VB2, 0},
  { "APU Configuration", PHX_APU_BASE + PHX_APU_NUM_CHANNEL,                      CNT_ALIASED_HEX | COUNTER_TYPE_SNAPSHOT  | COUNTER_DISPLAY_VB2, 0},
  { "EP Packets Received", PHX_DBL_RX_BASE + PHX_DBL_RX_PACKET_COUNT_N,           CNT_ENDPOINT | COUNTER_TYPE_64BIT,   PHX_DBL_RX_ENDPOINT_SIZE},
  { "EP Bytes Received", PHX_DBL_RX_BASE + PHX_DBL_RX_BYTE_COUNT_N,               CNT_ENDPOINT | COUNTER_TYPE_64BIT,   PHX_DBL_RX_ENDPOINT_SIZE},
  { "EP Filter Packets Sent", PHX_FLTR_BASE + PHX_FLTR_EP_PKTS_SENT,              CNT_ALIASED_ENDPOINT | COUNTER_DISPLAY_VERBOSE,   8},
  { "EP Filter Bytes Sent", PHX_FLTR_BASE + PHX_FLTR_EP_BYTES_SENT,               CNT_ALIASED_ENDPOINT | COUNTER_DISPLAY_VERBOSE,   8},
  { "EP TX PIO Packets", 0,                                                       CNT_SW_ENDPOINT, 0},
  { "EP TX PIO Bytes", 1,                                                         CNT_SW_ENDPOINT, 0},
  { "EP TX DMA Packets", 2,                                                       CNT_SW_ENDPOINT, 0},
  { "EP TX DMA Bytes", 3,                                                         CNT_SW_ENDPOINT, 0},
  { "EP Host RX Packets", 4,                                                      CNT_SW_ENDPOINT, 0},
  { "EP Host RX Packets with MAC Errors", 5,                                      CNT_SW_ENDPOINT, 0},
  { "EP TX Completions", PHX_REG_TX_BASE + PHX_EP_REG_TX_COMPLETION_COUNT,        CNT_ENDPOINT | COUNTER_TYPE_TX, 0},
  { "EP Data Pages", PHX_DBL_RX_BASE + PHX_DBL_RX_DATA_RING_SIZE,
                 CNT_ENDPOINT | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE,   PHX_DBL_RX_ENDPOINT_SIZE},
  { "EP Desc Pages", PHX_DBL_RX_BASE + PHX_DBL_RX_DESC_RING_SIZE,
                 CNT_ENDPOINT | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE,   PHX_DBL_RX_ENDPOINT_SIZE},
  { "EP Data Consumer Position",  PHX_DBL_RX_BASE + PHX_DBL_RX_DATA_CONSR_POSITION_N,
		 CNT_ENDPOINT | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE, PHX_DBL_RX_ENDPOINT_SIZE },
  { "EP Data Producer Position",  PHX_DBL_RX_BASE + PHX_DBL_RX_DATA_PRODW_POSITION_N,
		 CNT_ENDPOINT | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE, PHX_DBL_RX_ENDPOINT_SIZE },
  { "EP Desc Consumer Position",  PHX_DBL_RX_BASE + PHX_DBL_RX_DESC_CONSR_POSITION_N,
		 CNT_ENDPOINT | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE, PHX_DBL_RX_ENDPOINT_SIZE },
  { "EP Desc Producer Position",  PHX_DBL_RX_BASE + PHX_DBL_RX_DESC_PRODW_POSITION_N,
		 CNT_ENDPOINT | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE, PHX_DBL_RX_ENDPOINT_SIZE },
  { "Rx Sel Packets Rx", PHX_RX_SEL_BASE + PHX_RX_SEL_PKTS_RX,                   CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Rx Sel Packets Dropped", PHX_RX_SEL_BASE + PHX_RX_SEL_PKTS_DROP,            CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Rx Sel Packets Aborted", PHX_RX_SEL_BASE + PHX_RX_SEL_PKTS_ABT,             CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Rx Sel Packets Ok", PHX_RX_SEL_BASE + PHX_RX_SEL_PKTS_SENT,                 CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Rx Sel Packets No Meta", PHX_RX_SEL_BASE + PHX_RX_SEL_PKTS_NO_META,         CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Rx Sel Packets Meta Dropped", PHX_RX_SEL_BASE + PHX_RX_SEL_PKTS_META_DROP,  CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Rx Sel Packets Extra Meta", PHX_RX_SEL_BASE + PHX_RX_SEL_PKTS_EXTRA_META,   CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Rx Sel Packets Ctrl Dropped", PHX_RX_SEL_BASE + PHX_RX_SEL_PKTS_CTRL_DROP,  CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Rx Sel Bytes OK", PHX_RX_SEL_BASE + PHX_RX_SEL_BYTES_SENT,                  CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Rx Sel Packets Bypassed", PHX_RX_SEL_BASE + PHX_RX_SEL_PKTS_BYPASSED,       CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Rx Sel Bytes Bypassed", PHX_RX_SEL_BASE + PHX_RX_SEL_BYTES_BYPASSED,        CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "Rx Sel Bypass Disable", PHX_RX_SEL_BASE + PHX_RX_SEL_BYPASS_DISABLE,        CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "MAC Filter Packets Dropped", K35_MAC_FILTER_BASE + K35_MAC_FILTER_DISCARD_PKTS, CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "MAC Filter Bytes Dropped", K35_MAC_FILTER_BASE + K35_MAC_FILTER_DISCARD_BYTES, CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "MAC Filter Packets FWD", K35_MAC_FILTER_BASE + K35_MAC_FILTER_FWD_PKTS, CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "MAC Filter Bytes FWD", K35_MAC_FILTER_BASE + K35_MAC_FILTER_FWD_BYTES, CNT_ALIASED | COUNTER_DISPLAY_VB2,     0},
  { "PIO CP Read Dword", PHX_BBSYSCTL_BASE + PHX_BBSYSCTL_CP_RD_DW,              CNT_GLOBAL,     0},
  { "PIO CP Read Header", PHX_BBSYSCTL_BASE + PHX_BBSYSCTL_CP_RD_HEADER,         CNT_GLOBAL,     0},
  { "PIO CP Write Dword", PHX_BBSYSCTL_BASE + PHX_BBSYSCTL_CP_WR_DW,             CNT_GLOBAL,     0},
  { "PIO CP Write Header", PHX_BBSYSCTL_BASE + PHX_BBSYSCTL_CP_WR_HEADER,        CNT_GLOBAL,     0},
  { "IRQ Endpoints", PHX_ICU_BASE + PHX_ICU_ENDPOINTS,                           CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "IRQ Vectors Allocated", PHX_ICU_BASE + PHX_ICU_VECTORS_ALLOC,               CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "IRQ Operation Mode", PHX_ICU_BASE + PHX_ICU_OPERATION_MODE,                 CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2,     0},
  { "IRQ Global Mask", PHX_ICU_BASE + PHX_ICU_GLOBAL_MASK,                       CNT_ALIASED_HEX | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2, 0},
  { "IRQ Enable Lo", PHX_ICU_BASE + PHX_ICU_ENABLE_LO,                           CNT_ALIASED_HEX | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2, 0},
  { "IRQ Enable Mid", PHX_ICU_BASE + PHX_ICU_ENABLE_MID,                         CNT_ALIASED_HEX | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2, 0},
  { "IRQ Enable Hi", PHX_ICU_BASE + PHX_ICU_ENABLE_HI,                           CNT_ALIASED_HEX | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2, 0},
#ifdef MYRI_DBL_AB
  { "AB Module ID", PHX_AB_BASE + PHX_AB_MODULE_ID,				 CNT_GLOBAL_HEX | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2, 0 },
  { "AB Unit Count", PHX_AB_BASE + PHX_AB_UNIT_COUNT,				 CNT_GLOBAL | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2, 0 },
  { "AB History Depth", PHX_AB_BASE + PHX_AB_HISTORY,				 CNT_GLOBAL | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2, 0 },
  { "AB Ports", PHX_AB_BASE + PHX_AB_PORTS,					 CNT_GLOBAL | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2, 0 },
  { "AB Global Bypass", PHX_AB_BASE + PHX_AB_GLOBAL_BYPASS,			 CNT_GLOBAL | COUNTER_TYPE_SNAPSHOT, 0 },
  { "AB Bypass", PHX_AB_BASE + PHX_AB_BYPASS,           			 CNT_ALIASED | COUNTER_TYPE_SNAPSHOT, 0 },
  { "AB Auto Reset Seq Enable", PHX_AB_BASE + PHX_AB_AUTORESET_ENABLE,           CNT_ALIASED | COUNTER_TYPE_SNAPSHOT, 0 },
  { "AB Auto Reset Boundary", PHX_AB_BASE + PHX_AB_AUTORESET_BOUND,              CNT_ALIASED | COUNTER_TYPE_SNAPSHOT, 0 },
  { "AB Packet Count", PHX_AB_BASE + PHX_AB_PKT_COUNT,				 CNT_ALIASED, 0 },
  { "AB Packet Passed", PHX_AB_BASE + PHX_AB_PKT_PASS,  			 CNT_ALIASED, 0 },
  { "AB Packet Passed Pre-history", PHX_AB_BASE + PHX_AB_PRE_HISTORY, 		 CNT_ALIASED, 0 },
  { "AB A Packets Passed", PHX_AB_BASE + PHX_AB_PKT_A_PASSED,			 CNT_ALIASED, 0 },
  { "AB A Packets Removed", PHX_AB_BASE + PHX_AB_PKT_A_REMOVED,			 CNT_ALIASED, 0 },
  { "AB B Packets Passed", PHX_AB_BASE + PHX_AB_PKT_B_PASSED,			 CNT_ALIASED, 0 },
  { "AB B Packets Removed", PHX_AB_BASE + PHX_AB_PKT_B_REMOVED,			 CNT_ALIASED, 0 },
  { "AB C Packets Passed", PHX_AB_BASE + PHX_AB_PKT_C_PASSED,			 CNT_ALIASED, 0 },
  { "AB C Packets Removed", PHX_AB_BASE + PHX_AB_PKT_C_REMOVED,			 CNT_ALIASED, 0 },
  { "AB D Packets Passed", PHX_AB_BASE + PHX_AB_PKT_D_PASSED,			 CNT_ALIASED, 0 },
  { "AB D Packets Removed", PHX_AB_BASE + PHX_AB_PKT_D_REMOVED,			 CNT_ALIASED, 0 },
  { "AB Auto Reset Count", PHX_AB_BASE + PHX_AB_AUTORESET_COUNT,                 CNT_ALIASED, 0 },
  { "AB Last Unit", PHX_AB_BASE + PHX_AB_LAST_UNIT,			         CNT_ALIASED | COUNTER_TYPE_SNAPSHOT, 0 },
  { "AB Last Count", PHX_AB_BASE + PHX_AB_LAST_COUNT,			         CNT_ALIASED | COUNTER_TYPE_SNAPSHOT, 0 },
  { "AB Last Seq", PHX_AB_BASE + PHX_AB_LAST_SEQ,			         CNT_ALIASED | COUNTER_TYPE_SNAPSHOT, 0 },
#endif
  { "LTSSM Not L0 state", K35_LTSSM_NOT_L0                   , CNT_GLOBAL | COUNTER_DISPLAY_VB2, 0 },
};

static int
phx_k35_get_counter_definitions(mx_instance_state_t *is, const phx_counter_t **counters,
                                  uint32_t *count)
{
  *counters = phx_k35_counters;
  *count = sizeof(phx_k35_counters) / sizeof(struct phx_counter);
  return 0;
}

static int
phx_k35_counter_accum_index(mx_instance_state_t *is, struct phx_fpga_counter *ctr,
                              int nth_endpoint)
{
  int idx = 0;
  uint32_t bits = ctr->class & (COUNTER_TYPE_ENDPOINT | COUNTER_TYPE_ALIASED |
                                COUNTER_TYPE_GLOBAL);
  if (bits == (COUNTER_TYPE_ENDPOINT | COUNTER_TYPE_ALIASED))
    idx = is->phx.pci_info.func * is->phx.pfi->ep_pool.num_function_endpoints
        + nth_endpoint;
  else if (bits == COUNTER_TYPE_ALIASED)
    idx = is->phx.pci_info.func;
  else if (bits == COUNTER_TYPE_ENDPOINT)
    idx = is->phx.pci_info.func * is->phx.pfi->ep_pool.num_function_endpoints
        + nth_endpoint; /* Same as aliased enpoint */
  return idx;
}

int
phx_k35_set_promisc(mx_instance_state_t *is, int promisc)
{
  phx_k35_mac_set_promisc_mode(is->phx.nic_mem, promisc);
  return 0;
}

int
phx_k35_enable_all_multicast(mx_instance_state_t *is)
{
  phx_k35_mac_set_multicast_mode(is->phx.nic_mem, 1);
  return 0;
}

int
phx_k35_disable_all_multicast(mx_instance_state_t *is)
{
  phx_k35_mac_set_multicast_mode(is->phx.nic_mem, 0);
  return 0;
}

int
phx_k35_remove_filter(mx_instance_state_t *is,  myri_rx_filterq_entry_t *entry)
{
  int i;
  int found = -1;
  phx_mac_filter_t *l2_filt;
  uint32_t high, low;

  /* filters are stored in network order */
  high = htonl(entry->data[0]);
  low = htonl(entry->data[1]);

  /* find this filter, so I know which registers to clear */
  for (i = 0; i < is->phx.num_l2_filters; i++){
    l2_filt = &is->phx.l2_filters[i];
    if (l2_filt->used && (l2_filt->high == high) && (l2_filt->low == low)){
      found = i;
      break;
    }
  }

  /* clear the filter and remove it from the registers */
  if (found >= 0){
    l2_filt = &is->phx.l2_filters[found];
    l2_filt->high = l2_filt->low = 0;
    l2_filt->used = 0;
    phx_k35_set_mac_filter(is->phx.nic_mem, l2_filt->high, l2_filt->low, found);
  }
  return 0;
}

/* If this filter does not exist already, add it */
int
phx_k35_add_filter(mx_instance_state_t *is,  myri_rx_filterq_entry_t *entry)
{
  int i;
  int found = 0;
  int available = -1;
  phx_mac_filter_t *l2_filt;
  uint32_t high, low;
  int rc = -EBUSY;

  /* filters are stored in network order */
  high = htonl(entry->data[0]);
  low = htonl(entry->data[1]);

  /* is this filter already defined */
  for (i = 0; i < is->phx.num_l2_filters; i++){
    l2_filt = &is->phx.l2_filters[i];
    if (l2_filt->used && (l2_filt->high == high) && (l2_filt->low == low)){
      found = i;
      break;
    }
    if (!l2_filt->used && (available < 0))
      available = i;
  }

  /* allocate the available filter and set it up */
  if (!found && (available >= 0)){
    l2_filt = &is->phx.l2_filters[available];
    l2_filt->high = high;
    l2_filt->low = low;   /* we have to move the 2 used bytes up */
    l2_filt->used = 1;
    phx_k35_set_mac_filter(is->phx.nic_mem, l2_filt->high, l2_filt->low, available);
    rc = 0;
  }

  return rc;
}

void
phx_k35_init_board_ops(void)
{
  phx_k35_ops.init                  = phx_k35_init_board;
  phx_k35_ops.map                   = phx_common_map_board;
  phx_k35_ops.unmap                 = phx_common_unmap_board;
  phx_k35_ops.setup_hw_caps         = phx_common_setup_hw_caps;
  phx_k35_ops.enable_interrupt      = phx_common_int_enable;
  phx_k35_ops.disable_interrupt     = phx_common_int_disable;
  phx_k35_ops.park                  = phx_k35_park_board;
  phx_k35_ops.get_counter_definitions  = phx_k35_get_counter_definitions;
  phx_k35_ops.counter_accum_index   = phx_k35_counter_accum_index;
  phx_k35_ops.init_endpoint_pool    = phx_k35_init_endpoint_pool;
  phx_k35_ops.ether_endpoint        = phx_k35_ether_endpoint;
  phx_k35_ops.physical_endpoint     = phx_k35_physical_endpoint;
  phx_k35_ops.sniffer_endpoint      = phx_k35_sniffer_endpoint;
  phx_k35_ops.set_intr_coal         = phx_k35_set_intr_coal;
  phx_k35_ops.set_mac_addr          = phx_k35_set_mac_addr;
  phx_k35_ops.set_promisc           = phx_k35_set_promisc;
  phx_k35_ops.enable_all_multicast  = phx_k35_enable_all_multicast;
  phx_k35_ops.disable_all_multicast = phx_k35_disable_all_multicast;
  phx_k35_ops.remove_filter         = phx_k35_remove_filter;
  phx_k35_ops.add_filter            = phx_k35_add_filter;
}
