/**
 * File:  phx_vitesse.c
 *
 * Desc:  Initialize the vitesse retimer components on the Phoenix board.
 *
 *  Provided by
 *  -----------
 *      BittWare, Inc.
 *      45 South Main St. Suite L100
 *      Concord, NH 03301
 *      Ph:     (603) 226 0404
 *      Fax:    (603) 226 6667
 *      WWW:    http://www.bittware.com
 *      Email:  support@bittware.com
 *
 * Copyright ? 2011, BittWare, Inc.
 *
 * License:
 *
 * The user is hereby granted a non-exclusive license to use
 * and or modify this software provided that it runs on BittWare
 * hardware. Usage of this software on non-BittWare hardware
 * without the express written permission of BittWare is strictly
 * prohibited.
 *
**/

#include "mal_auto_config.h"
#include "mx_arch.h"
#include "mx_misc.h"
#include "mx_instance.h"
#include "mal_stbar.h"
#include "mx_pio.h"
#include "phx_avl.h"
#include "phx_common.h"

/* Temporary until we have a more general logging/verbosity control */
#if MAL_DEBUG
#define _VERBOSE 1
#else
#define _VERBOSE 0
#endif

/* Board Options */
#define A5PL_DUAL	1
#define A5PL_QUAD	2
#define A5PS		3

/* Configuration Options */
#define FIBER_CABLE      0x01
#define COPPER_SHORT     0x02
#define COPPER_LONG      0x03

/* VSC7224 Constants */
#define VSC7224_PAGEREG 0x7F

/* VSC7111 Constants */
#define VSC7111_PAGEREG 0x7F

static void vsc7224_init(char *base, uint32_t qsfp_port, uint32_t cfg_setting, uint32_t channel);
static void vsc7111_init(char *base, uint32_t qsfp_port, uint32_t cfg_setting, uint32_t channel);
static void vsc7111_exit(char *base, uint32_t qsfp_port, uint32_t channel);

int phx_vitesse_init(mx_instance_state_t *is, int first_call)
{
	char *base = (char*)is->phx.i2c_base_mem;
	int board_type;
	uint32_t connection_media;
	uint32_t channel = MX_IS_PORT_ID(is);
	uint32_t qsfp = 1;

	/* what board type are we dealing with? */
	if (is->phx.part_number && strcmp(is->phx.part_number, "09-04642") == 0){
		qsfp = channel + 1;     /* function 0 is qsfp1, function 1 is qsfp2 */
                channel = 0;            /* both ports use channel 0 */
		board_type = A5PL_DUAL;
        } 
	else if (is->phx.part_number && strcmp(is->phx.part_number, "09-04652") == 0){
		board_type = A5PL_QUAD;
	}
	else if (is->phx.part_number && strcmp(is->phx.part_number, "09-04651") == 0){
		board_type = A5PS;
	}
	else {
		MX_WARN(("Vitesse initialization failed, Unknown board type\n"));
                return 0;
	}

	
	if (myri_phx_serdes_mode != 0)
		connection_media = myri_phx_serdes_mode;
	else
		connection_media = COPPER_SHORT;

	/* We have to claim the mutex before using i2c busses */
	do {}
	while (!phx_i2c_mutex_claim(base));

	/* Init Vitesse Devices */
	switch (board_type)
	{
		case A5PL_DUAL:
			MX_INFO(("vsc initialization for A5PL DUAL card, qsfp %d media type %d\n", qsfp, connection_media));
			vsc7224_init(base, qsfp, connection_media, channel);
			vsc7111_exit(base, qsfp, channel);   /* turn off all output channels */
			vsc7111_init(base, qsfp, connection_media, channel);   /* enable the channel I use */
			break;

		case A5PL_QUAD:
			MX_INFO(("vsc initialization for A5PL QUAD card, qsfp %d channel %d media type %d\n", qsfp, channel, connection_media));
			if (first_call){
				vsc7224_init(base, qsfp, connection_media, channel); /* setup once */
				vsc7111_exit(base, qsfp, channel);   /* turn off all output channels */
                        }
			vsc7111_init(base, qsfp, connection_media, channel);    /* enable the channel I use */
			break;

		case A5PS:
			MX_INFO(("vsc initialization for A5PS card, qsfp %d media type %d\n", qsfp,  connection_media));
			vsc7224_init(base, qsfp, connection_media, channel);
			break;

	}

	phx_i2c_mutex_release(base);

	return 0;
}

int phx_vitesse_exit(mx_instance_state_t *is)
{
	char *base = (char*)is->phx.i2c_base_mem;
	int board_type;
	int channel = MX_IS_PORT_ID(is);
	int qsfp = 1;

	/* what board type are we dealing with? */
	if (is->phx.part_number && strcmp(is->phx.part_number, "09-04642") == 0){
		qsfp = channel + 1;     /* function 0 is qsfp1, function 1 is qsfp2 */
                channel = 0;            /* both ports use channel 0 */
		board_type = A5PL_DUAL;
	}
	else if (is->phx.part_number && strcmp(is->phx.part_number, "09-04652") == 0)
		board_type = A5PL_QUAD;
	else if (is->phx.part_number && strcmp(is->phx.part_number, "09-04651") == 0)
		board_type = A5PS;
	else {
		MX_WARN(("Vitesse exit failed, Unknown board type\n"));
                return 0;
	}

	/* We have to claim the mutex before using i2c busses */
	do {}
        while (!phx_i2c_mutex_claim(base));

	/* Shut off Tx Output on Vitesse 7111 Devices */
	switch (board_type)
	{
		case A5PL_DUAL:
			MX_INFO(("vsc exit for A5PL DUAL card\n"));
			vsc7111_exit(base, qsfp, channel);
			break;

		case A5PL_QUAD:
			MX_INFO(("vsc exit for A5PL QUAD card\n"));
			vsc7111_exit(base, qsfp, channel);
			break;
		case A5PS:
			/* for completeness, there is no vsc7111 on this board */
			MX_INFO(("vsc exit for A5PS card\n"));
			break;
	}
	phx_i2c_mutex_release(base);

	return 0;
}
/* Write VSC7224 Register
 * Write Page Register
 * Write Register Address
 * Write Register(s)
 */
static void
vsc7224_wr_reg(char *base, uint32_t avl_base, uint8_t device, uint8_t reg_ofs, uint16_t reg_val)
{
	uint8_t bytes[2];

	bytes[0] = (reg_val >> 8) & 0xff;
	bytes[1] = reg_val & 0xff;
	avl_i2c_write_bytes(base, avl_base, device, reg_ofs, bytes, sizeof(bytes));
#if 0
/* #if _VERBOSE */
	MX_INFO(("%s : WRITE reg %02x = %04x\n",
		 ((avl_base == AVL_BASE_I2C_3) ? "vsc7224-1":"vsc7224-2"), reg_ofs, reg_val));
#endif
}

/* Read VSC7224 Register
 * Write Page Register
 * Write Register Address
 * Read Register(s)
 */
static uint16_t
vsc7224_rd_reg(char *base, uint32_t avl_base, uint8_t device, uint8_t reg_ofs)
{
	uint16_t ret;
	uint8_t bytes[2];

	avl_i2c_read_bytes(base, avl_base, device, reg_ofs, bytes, sizeof(bytes));
	ret = (bytes[0] << 8) | bytes[1];

#if 0
/* #if _VERBOSE */
	MX_INFO(("%s : READ reg %02x = %04x\n",
		 ((avl_base == AVL_BASE_I2C_3) ? "vsc7224-1":"vsc7224-2"), reg_ofs, ret));
#endif
	return ret;
}

/* Write VSC7111 Register
 * Write Page Register
 * Write Register Address
 * Write Register(s)
 */
static void
vsc7111_wr_reg(char *base, uint32_t avl_base, uint8_t device, uint8_t reg_ofs, uint8_t reg_val)
{
	avl_i2c_write_bytes(base, avl_base, device, reg_ofs, &reg_val, 1);
#if 0
/* #if _VERBOSE */
	MX_INFO(("%s : WRITE reg %02x = %02x\n",
		 ((avl_base == AVL_BASE_I2C_2) ? "vsc7111-1":"vsc7111-2"), reg_ofs, reg_val));
#endif
}

/* Init both VSC7224 Devices */
static void
vsc7224_init(char *base, uint32_t qsfp_port, uint32_t cfg_setting, uint32_t channel)
{
	uint8_t i;
	uint32_t i2c_chain;
	uint8_t dev_addr;

	if (qsfp_port == 1) i2c_chain = AVL_BASE_I2C_3;
        else if (qsfp_port == 2) i2c_chain = AVL_BASE_I2C_6;
        else return;

	dev_addr  = DEV_7224;  /* I2C device address (write, +1 for read) */

	/* Set DGMUXCTRL to 0 allowing FPGA to gate link status with LOS */
	vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x40);
	vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x81, 0);


	/* Settings based on connection media */
	switch(cfg_setting)
	{
		case FIBER_CABLE:
			MX_INFO(("Init vsc7224-%d for FIBER_CABLE\n", qsfp_port));
			vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x30);
			for (i=0x80;i<0x86;i++) vsc7224_rd_reg(base, i2c_chain, dev_addr, i);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x80, 0x2841);  /* Mval=D[15:8], Nval=D[7:0] */
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x81, 0x0008);  /* Fval[23:16]=D[7:0] */
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x82, 0x7A00);  /* Fval[15:00]=D[15:0] */
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x83, 0x000F);  /* Rval[23:16]=D[7:0] */
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x84, 0x9C18);  /* Rval[15:00]=D[15:0] */
			for (i=0x80;i<0x86;i++) vsc7224_rd_reg(base, i2c_chain, dev_addr, i);

			/* 
			 * All channels are set equally,
			 * page select Channel Broadcast
			 */
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x7f, 0x0050);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x82, 0x0014);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x89, 0x7F13);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x90, 0x5785);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x92, 0x207F);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x93, 0x2000);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x94, 0x0031);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x9C, 0x0000);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x99, 0x001F);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x9A, 0x000F);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x9B, 0x0004);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0xAA, 0x0888);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0xA8, 0x2408);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0xA9, 0x0060);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x87, 0x4021);
			break;

		case COPPER_SHORT:
			MX_INFO(("Init vsc7224-%d for COPPER_SHORT\n", qsfp_port));
			/* VGASET_EQADCCTRL for each channel is set independently
			 * page select each lane
			 */
			if (qsfp_port == 1){ 
				vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x0000);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x86, 0x3000);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x0001);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x86, 0x3600);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x0002);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x86, 0x3000);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x0003);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x86, 0x3800);
			} else {
				vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x0000);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x86, 0x3a00);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x0001);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x86, 0x4200);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x0002);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x86, 0x3E00);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x0003);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x86, 0x3600);
			}

			vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x30);
			for (i=0x80;i<0x86;i++) vsc7224_rd_reg(base, i2c_chain, dev_addr, i);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x80, 0x2841);  /* Mval=D[15:8], Nval=D[7:0] */
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x81, 0x0008);  /* Fval[23:16]=D[7:0] */
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x82, 0x7A00);  /* Fval[15:00]=D[15:0] */
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x83, 0x000F);  /* Rval[23:16]=D[7:0] */
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x84, 0x9C18);  /* Rval[15:00]=D[15:0] */
			for (i=0x80;i<0x86;i++) vsc7224_rd_reg(base, i2c_chain, dev_addr, i);
			/* 
			 * All channels are set equally,
			 * page select Channel Broadcast
			 */
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x7f, 0x0050);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x82, 0x0014);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x89, 0x7F13);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x90, 0x5785);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x92, 0x207F);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x93, 0x2000);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x94, 0x0031);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x9C, 0x0000);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x99, 0x001F);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x9A, 0x000F);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x9B, 0x0004);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0xAA, 0x0888);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0xA8, 0x2408);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0xA9, 0x0060);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x87, 0x4021);
			break;

		case COPPER_LONG:
			MX_INFO(("Init vsc7224-%d for COPPER_LONG\n", qsfp_port));
			/* VGASET_EQADCCTRL for each channel is set independently
			 * page select each lane
			 */
			if (qsfp_port == 1){ 
				vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x0000);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x86, 0x3600);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x0001);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x86, 0x3600);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x0002);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x86, 0x3c00);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x0003);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x86, 0x3800);
			} else {
				vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x0000);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x86, 0x3800);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x0001);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x86, 0x4800);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x0002);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x86, 0x3E00);
				vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x0003);
			}	vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x86, 0x3600);

			vsc7224_wr_reg(base, i2c_chain, dev_addr, VSC7224_PAGEREG, 0x30);
			for (i=0x80;i<0x86;i++) vsc7224_rd_reg(base, i2c_chain, dev_addr, i);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x80, 0x2841);  /* Mval=D[15:8], Nval=D[7:0] */
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x81, 0x0008);  /* Fval[23:16]=D[7:0] */
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x82, 0x7A00);  /* Fval[15:00]=D[15:0] */
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x83, 0x000F);  /* Rval[23:16]=D[7:0] */
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x84, 0x9C18);  /* Rval[15:00]=D[15:0] */
			for (i=0x80;i<0x86;i++) vsc7224_rd_reg(base, i2c_chain, dev_addr, i);
			/* 
			 * All channels are set equally,
			 * page select Channel Broadcast
			 */
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x7f, 0x0050);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x82, 0x0014);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x89, 0x7b13);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x90, 0x5785);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x92, 0x207F);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x93, 0x2000);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x94, 0x0031);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x9C, 0x0000);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x99, 0x001F);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x9A, 0x000F);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x9B, 0x0004);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0xAA, 0x0888);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0xA8, 0x2408);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0xA9, 0x0060);
			vsc7224_wr_reg(base, i2c_chain, dev_addr, 0x87, 0x4021);
			break;

		default:
			MX_INFO(("Init vsc7224-%d INVALID Config : %d\n", qsfp_port, cfg_setting));
			break;
	}
}

static void
vsc7111_exit(char *base, uint32_t qsfp_port, uint32_t channel)
{
	uint32_t i2c_chain;
	uint8_t dev_addr;

	if (qsfp_port == 1) i2c_chain = AVL_BASE_I2C_2;
        else if (qsfp_port == 2) i2c_chain = AVL_BASE_I2C_5;
        else return;
		
	dev_addr  = DEV_7111;  // I2C device address (write, +1 for read)
        vsc7111_wr_reg(base, i2c_chain, dev_addr, VSC7111_PAGEREG, 0x00);
        vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x59, 0x01);  /* Output OFF */
}

static void
vsc7111_init(char *base, uint32_t qsfp_port, uint32_t cfg_setting, uint32_t channel)
{
	uint32_t i2c_chain;
	uint8_t dev_addr;

	if (qsfp_port == 1) i2c_chain = AVL_BASE_I2C_2;
        else if (qsfp_port == 2) i2c_chain = AVL_BASE_I2C_5;
        else return;
		
	dev_addr  = DEV_7111;  // I2C device address (write, +1 for read)

	vsc7111_wr_reg(base, i2c_chain, dev_addr, VSC7111_PAGEREG, 0x00);

	switch(cfg_setting)
	{
		case FIBER_CABLE:
			MX_INFO(("Init vsc7111-%d for FIBER_CABLE\n", qsfp_port));

			/* VSC7111 Input Settings "6 Gbps or greater, 6-12in PCB" */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x51, 0x1F);  /* Input Gain1 */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x52, 0x1F);  /* Input Gain2 */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x56, 0x00);  /* Output Pre-emphasis 1 */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x57, 0x00);  /* Output Pre-emphasis 2 */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x58, 0x02);  /* Output Level */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x5A, 0x00);  /* ISE 1 Long,  EQ1 long time constant (LTC) */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x5B, 0x00);  /* ISE 1 Short, EQ1 short time constant (LTC) */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x5C, 0x00);  /* ISE 2 Long,  EQ2 long time constant (LTC) */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x5D, 0x00);  /* ISE 2 Short, EQ2 short time constant (LTC) */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x55, 0xFF);  /* LOS threshold = 300mV (max) */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x60, 0x64);  /* EQ ON */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x59, 0x00);  /* Output ON */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x54, 0x02);  /* Input ON */
			break;

		case COPPER_SHORT:
			MX_INFO(("Init vsc7111-%d for COPPER_SHORT\n", qsfp_port));
#if 0
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x51, 0x1F);  /* Input Gain1 */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x52, 0x1F);  /* Input Gain2 */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x56, 0x00);  /* Output Pre-emphasis 1 */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x57, 0x00);  /* Output Pre-emphasis 2 */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x58, 0xB4);  /* Output Level */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x5A, 0x00);  /* ISE 1 Long,  EQ1 long time constant (LTC) */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x5B, 0x00);  /* ISE 1 Short, EQ1 short time constant (LTC) */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x5C, 0x00);  /* ISE 2 Long,  EQ2 long time constant (LTC) */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x5D, 0x00);  /* ISE 2 Short, EQ2 short time constant (LTC) */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x55, 0xFF);  /* LOS threshold = 300mV (max) */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x60, 0xE4);  /* Adaptive EQ off */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x59, 0x00);  /* Output ON */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x54, 0x4E);  /* Input ON, max power/bandwidth */
#else
                        /* Bittware modified values 3/2/2015 */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x55, 0xFF);  /* LOS threshold = 300mV (max) */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x7F, 0x23);  /* Page Reg */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, channel, 0x00);  /* Output ON */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x7F, 0x13);  /* Page Reg */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, channel, 0x4E);  /* Input ON */

			// Group 0 : "1" = / "0" = 0x1F
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x51, 0x07);  /* Input Gain1 */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x52, 0x1F);  /* Input Gain2 */

			// Group 1 & 2 : "1" = / "0" =
			/* Bittware modified values 3/12/15 */
			if (qsfp_port == 1){
				vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x56, 0x28);  /* Output Pre-emphasis 1 */
				vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x57, 0x00);  /* Output Pre-emphasis 2 */
				vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x58, 0xBA);  /* Output Level */
			} else {
				vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x56, 0x08);  /* Output Pre-emphasis 1 */
				vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x57, 0x00);  /* Output Pre-emphasis 2 */
				vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x58, 0xBE);  /* Output Level */
			}

			// Group 3 : "1" = 0xE4 / "0" = 0x64
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x60, 0x64);  /* Adaptive EQ off=0x64, on=0xE4 */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x5A, 0x00);  /* ISE 1 Long,  EQ1 long time constant (LTC) */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x5B, 0x00);  /* ISE 1 Short, EQ1 short time constant (LTC) */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x5C, 0x00);  /* ISE 2 Long,  EQ2 long time constant (LTC) */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x5D, 0x00);  /* ISE 2 Short, EQ2 short time constant (LTC) */
#endif
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x59, 0x00);  /* Output ON */
			break;

		case COPPER_LONG:
			MX_INFO(("Init vsc7111-%d for COPPER_LONG\n", qsfp_port));

			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x51, 0x1F);  /* Input Gain1 */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x52, 0x1F);  /* Input Gain2 */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x56, 0x0F);  /* Output Pre-emphasis 1 */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x57, 0x00);  /* Output Pre-emphasis 2 */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x58, 0xB4);  /* Output Level */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x5A, 0x00);  /* ISE 1 Long,  EQ1 long time constant (LTC) */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x5B, 0x00);  /* ISE 1 Short, EQ1 short time constant (LTC) */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x5C, 0x00);  /* ISE 2 Long,  EQ2 long time constant (LTC) */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x5D, 0x00);  /* ISE 2 Short, EQ2 short time constant (LTC) */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x55, 0xFF);  /* LOS threshold = 300mV (max) */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x60, 0xE4);  /* Adaptive EQ off */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x59, 0x00);  /* Output ON */
			vsc7111_wr_reg(base, i2c_chain, dev_addr, 0x54, 0x4E);  /* Input ON, max power/bandwidth */
			break;

		default:
			MX_INFO(("Init vsc7111-%d INVALID Config : %d\n", qsfp_port, cfg_setting));
			break;
	}

	vsc7111_wr_reg(base, i2c_chain, dev_addr, VSC7111_PAGEREG, 0x00);
}


/*
  This file uses MX driver indentation.

  Local Variables:
  c-file-style:"linux"
  tab-width:8
  indent-tabs-mode:t
  End:
*/
