/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2014-2016 by CSP, Inc.  All rights reserved.                *
 *************************************************************************/

#include "mx_arch.h"
#include "mx_misc.h"
#include "mx_instance.h"
#include "mx_malloc.h"
#include "mx_pio.h"
#include "mal_stbar.h"
#include "myri_version.h"
#include "mal_cpu.h"
#include "phx_hw.h"
#include "phx_common.h"
#include "phx_avl.h"

/* Temporary until we have a more general logging/verbosity control */
#if MAL_DEBUG
#define _VERBOSE 1
#else
#define _VERBOSE 0
#endif

mx_mac_ops_t phx_xilinx_xmac_ops;

void xilinx_xmac_disable(mx_instance_state_t *is)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + PHX_MAC_BASE;
#if _VERBOSE
  MX_INFO(("%s: Turning off MAC tx/rx transfer control\n", __FUNCTION__));
#endif
  PHX_REG_WRITE_4((r + PHX_XLX_MAC_TX_CONTROL), PHX_XLX_TX_DISABLE);
  PHX_REG_WRITE_4((r + PHX_XLX_MAC_RX_CONTROL), PHX_XLX_RX_DISABLE);
}

void xilinx_xmac_enable(mx_instance_state_t *is)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + PHX_MAC_BASE;
#if _VERBOSE
  MX_INFO(("%s: Turning off MAC tx/rx transfer control\n", __FUNCTION__));
#endif
  PHX_REG_WRITE_4((r + PHX_XLX_MAC_TX_CONTROL), PHX_XLX_TX_ENABLE);
  PHX_REG_WRITE_4((r + PHX_XLX_MAC_RX_CONTROL), PHX_XLX_RX_ENABLE);
}

void xilinx_xmac_set_max_frame_length(mx_instance_state_t *is, int mtu)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + PHX_MAC_BASE;
  uint32_t tx_max, rx_max;

#if _VERBOSE
  MX_INFO(("%s:\n", __FUNCTION__));
  tx_max = PHX_REG_READ_4((r + PHX_XLX_MAC_TX_MTU_CONFIG));
  tx_max &= 0x7fff;
  rx_max = PHX_REG_READ_4((r + PHX_XLX_MAC_RX_MTU_CONFIG));
  rx_max &= 0x7fff;
  MX_INFO(("current: tx_max=%u rx_max=%u\n", tx_max, rx_max));
#endif

#ifdef PHX_WORKAROUND_MAC_OVERSIZE_FRAMES
  /* Set lengths to be some large so the NIC never sees
   * oversized frames.
   */
  tx_max = 0x10000 + 10000;  /* enable for MTU reg, and 10000 limit */
  PHX_REG_WRITE_4((r + PHX_XLX_MAC_TX_MTU_CONFIG), tx_max);
  rx_max = 0x10000 + 10000;
  PHX_REG_WRITE_4((r + PHX_XLX_MAC_RX_MTU_CONFIG), rx_max);
#else
  /* "Length" in the Altera MAC covers [Ethernet header ... CRC32].
   * The MAC understands VLAN tagged packets and uses length + 4 when
   * checking the VLAN packet's length.
   *
   * TX length does not affect transmission at all.  It is only used
   * for statistics in the MAC.
   */

  /* Add 14 (ethernet header) and 4 (crc32) to the OS MTU. */
  if (mtu > 1500) {
    mtu |= 0x10000;  /* enable bit for MTU reg */
    PHX_REG_WRITE_4((r + PHX_XLX_MAC_TX_MTU_CONFIG), (mtu + 18));
    PHX_REG_WRITE_4((r + PHX_XLX_MAC_RX_MTU_CONFIG), (mtu + 18));
  }
#endif

#if _VERBOSE
  tx_max = PHX_REG_READ_4((r + PHX_XLX_MAC_TX_MTU_CONFIG));
  tx_max &= 0x7fff;
  rx_max = PHX_REG_READ_4((r + PHX_XLX_MAC_RX_MTU_CONFIG));
  rx_max &= 0x7fff;
  MX_INFO(("current: tx_max=%u rx_max=%u\n", tx_max, rx_max));
#endif
}

void xilinx_xmac_set_promisc_mode(mx_instance_state_t *is, int enable)
{
  /* The low latency MAC is always promiscuous */
}

void xilinx_xmac_set_address(mx_instance_state_t *is, uint8_t mac[6])
{
  /* The low latency MAC has no MAC address */
}

void 
phx_xilinx_init_xmac_ops(void)
{
//  mx_get_counter_definitions_t  *get_counter_definitions;
//  //  mx_counter_accum_index_t      *counter_accum_index;
  phx_xilinx_xmac_ops.mac_set_hwaddr           = xilinx_xmac_set_address;
  phx_xilinx_xmac_ops.mac_set_promisc          = xilinx_xmac_set_promisc_mode;
  phx_xilinx_xmac_ops.mac_enable               = xilinx_xmac_enable;
  phx_xilinx_xmac_ops.mac_disable              = xilinx_xmac_disable;
  phx_xilinx_xmac_ops.mac_set_max_frame_length = xilinx_xmac_set_max_frame_length;
};
