/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2014-2016 by CSP, Inc.  All rights reserved.                *
 *************************************************************************/

#include "mx_arch.h"
#include "mx_misc.h"
#include "mx_instance.h"
#include "mx_malloc.h"
#include "mx_pio.h"
#include "mal_stbar.h"
#include "myri_version.h"
#include "mal_cpu.h"
#include "phx_hw.h"
#include "phx_common.h"
#include "phx_avl.h"

/* FIXME... */
#if MAL_OS_WINDOWS
extern void phx_win_usleep(int usec);
#endif

/* Temporary until we have a more general logging/verbosity control */
#if MAL_DEBUG
#define _VERBOSE 1
#else
#define _VERBOSE 0
#endif

mx_sync_t mx_fpga_mutex;

/* global array of fpga instances */
struct phx_fpga_instance *phx_fpga_instances;
uint32_t phx_fpga_num_instances;

void
phx_pio_memcpy(void *dst, const void *src, int len)
{
  uint8_t *d, *s;
  uint32_t w;
  /* FIXME
   * We cannot have writes > 4B in the current revision.
   * We always have to write 4B at a time, no 1B or 2B writes.
   */
  /* MX_INFO(("%s: dst=%p src=%p len=%d\n", __FUNCTION__, dst, src, len)); */

  d = (uint8_t*)dst;
  s = (uint8_t*)src;
  while (len >= 4) {
    w = (((uint32_t)*(s+0)) << 0) |
        (((uint32_t)*(s+1)) << 8) |
        (((uint32_t)*(s+2)) << 16) |
        (((uint32_t)*(s+3)) << 24);
    *(uint32_t*)d = w;
    d += 4;
    s += 4;
    len -= 4;
    MAL_STBAR();
  }
  if (len > 0) {
    w = (((uint32_t)*(s+0)) << 0);
    if (len > 1)
      w |= (((uint32_t)*(s+1)) << 8);
    if (len > 2)
      w |= (((uint32_t)*(s+2)) << 16);
    *(uint32_t*)d = w;
    MAL_STBAR();
  }

#if 0
  while (len > 0 && ((d & 0x3) != 0 || (s & 0x3) != 0)) {
    *(uint8_t*)d = *(uint8_t*)s;
    d++;
    s++;
    len--;
  }
  while (len >= 4) {
    *(uint32_t*)d = *(uint32_t*)s;
    d += 4;
    s += 4;
    len -= 4;
  }
  while (len > 0) {
    *(uint8_t*)d = *(uint8_t*)s;
    d++;
    s++;
    len--;
  }
#endif
}

#ifdef MYRI_DBL_AB
static int32_t phx_ab_tablex_lookup(mx_instance_state_t *is, uint32_t feed_unit);
#endif

/************
 * Counters *
 ************/
/*
 * There is a global array telling all the names and addresses and types of
 * counters that will be used by myri_counters to display the counts.
 *
 * It was anticipated that the FPGA would DMA the counters to the nic about
 * once a second, but that does not seem realistic.  So for now when someone
 * wants to read the counters we go out and poll the FPGA.
 * There is room for a page of counters to be DMA'ed by the FPGA, this is
 * currently unused.  Code has never been tested, although some place holder code
 * exists.
 *
 * For Fun :
 * Some counters are 32-bit, some are 64 bit.
 * Some are held in the FPGA, some are software.
 * Some register counters are cleared by indirect things such as closing an endpoint.
 * Register counters cannot be changed by writing directly to the registers.
 * Some register counters can be cleared by indirect things such as resetting a core
 *   containing the counters.
 *
 * On the lanai, the concept of clearing counters was done by effectively
 * remembering where the count was when the clear was requested and having counters
 * accumulate from there.  This strategy is also used here.
 *
 * Each item that is counted will have an "accum" value, which is 64 bits,
 * and an "accum_last64" value, also 64 bits.
 *
 *   accum = current value to show in myri_counters
 *   accum_last = value of the counter when "clear" happened -
 *           which may or may not have reset the count to 0.
 *
 * Storage:
 * FIXME
 *
 * Final stray points:
 *
 *  The "stride" for endpoint counters is the physical offset of the registers in
 *  the FPGA's address space, relative to the first endpoint's register.
 *
 *  The seqnum concept is from the lanai and logic is as follows:
 *  The card will DMA up counters occasionally.
 *  A watchdog routine in the kernel occasionally calls a routine to read 32 bit
 *  counters and add them to creat 64 bit versions.
 *  seqnum keeps userspace from intervening and reading the counters when that
 *  kernel update is going on to keep the 64 but values atomic.
 *  DBL on Phoenix is not currently doing the watchdog function since polling
 *  the FPGA is too expensive on latency.    The updates only happen if someone
 *  reads the counters.
 *  Therefore "fake" 64 bit counters may not be accurate.
 */

/*
 * Hardware based clear of the PCIE histogram counters
 * must be called with the counter spinlock already held.
 */
static inline
void
phx_reset_pcie_counters(mx_instance_state_t *is)
{
  uint8_t *r;

  r = (uint8_t *)is->phx.nic_mem + PHX_DBL_RX_BASE + PHX_DBL_RX_BP_RESET;
  if (MX_IS_K35SNF(is)) {
    r += K35SNF_RX_SIZE * is->port_id;
  }
  PHX_REG_WRITE_8(r, 1);
}

int
phx_counters64_get(mx_instance_state_t *is, uint64_t *c_o, int start_off,
                   int num_counters, int endpoint)
{
  struct phx_fpga_instance *fpga = is->phx.pfi;
  uint32_t seqnum;
  uint64_t val;
  int i;
  int ep=0;
  int raw_values=0;
  int offsets=0;
  uint64_t *counters_out = c_o;
  struct phx_fpga_counter *ctr;

  if (endpoint & 0x100)
    raw_values = 1;
  else if (endpoint & 0x200)
    offsets = 1;
  endpoint &= 0xff;

  if ((endpoint >=0) && (endpoint < is->es_count))
    ep = endpoint;

  do {
  try_again:
    c_o = counters_out; /* We may loop, so reset c_o to the base */
    seqnum = fpga->counters.seqnum;
    MAL_READBAR();
    if (unlikely(seqnum & 1)) { /* back off, hit middle of update */
      MAL_CPU_RELAX();
      goto try_again;
    }

    for (i = start_off;
         i < start_off + num_counters && i < fpga->counters.defined_cnt;
         i++) {
      int idx;

      ctr = &fpga->counters.counters[i];
      idx = is->board_ops.counter_accum_index(is, ctr, ep);
      if (offsets){
        val = (uint64_t) ctr->offset + (ep * ctr->stride); /*  offset to the register */
        if (ctr->port_stride)
          val += ctr->port_stride * is->port_id;
      }
      else if (raw_values)
        val = ctr->accum_last_64[idx];
      else
        val = ctr->accum[idx];

      *c_o = val;
      c_o++;
    }
    MAL_READBAR();
  } while (seqnum != fpga->counters.seqnum);
  return 0;
}

/* Accumulate an endpoint software counter. */
static void
phx_update_software_counter(mx_instance_state_t *is, struct phx_fpga_counter *ctr, int endpoint, int is_clear)
{
  struct phx_fpga_instance *fpga = is->phx.pfi;
  mx_endpt_state_t *es;
  volatile uint64_t *ep_tx_counters, *is_tx_counters;
  uint64_t val, val2;
  uint64_t val_prev, val_next;
  const struct phx_physical_ep *phx_ep;
  int idx;
  int sw_index = ctr->offset;    /* this is the index in the soft counters array for SW counters */

  phx_ep = is->board_ops.physical_endpoint(is, endpoint);
  idx = is->board_ops.counter_accum_index(is, ctr, endpoint);

  /* During instance initialization, we do not have ether... */
  val_next = 0;

  /* Special case the kernel ethernet endpoint as it has no endpoint structure. */
  if (phx_ep == is->board_ops.ether_endpoint(is)) {
    ep_tx_counters = (volatile uint64_t*)&is->phx.pfi->counters.ep_tx_stats[phx_ep->tx_ep];
    do {
      val = ep_tx_counters[sw_index];
      val2 = ep_tx_counters[sw_index];
    } while ((val >> 32) != (val2 >> 32));
    val_next = val;
  } else {
    mx_spin_lock(&fpga->counters.es_spinlock);   /* outer spinlock must be held by caller */
    /* Read the accumulator */
    is_tx_counters = (volatile uint64_t*)&fpga->counters.ep_tx_stats[phx_ep->tx_ep];
    val_next = is_tx_counters[sw_index];

    /* Then add the live counter to the above */
    if ((es = fpga->counters.es[phx_ep->tx_ep])) {
      ep_tx_counters = (volatile uint64_t*)&es->flow_window->tx_stats;
      /* Endpoint counters are updated by the user thread, without locking.
       * On 32-bit, update is not atomic.  So read again if we see a carry.
       */
      do {
        val = ep_tx_counters[sw_index];
        val2 = ep_tx_counters[sw_index];
      } while ((val >> 32) != (val2 >> 32));
      val_next += val;
    }
    mx_spin_unlock(&fpga->counters.es_spinlock);
  }

  val_prev = ctr->accum_last_64[idx];
  if (is_clear)
    ctr->accum[idx] = 0;
  else
    ctr->accum[idx] += (val_next - val_prev);
  ctr->accum_last_64[idx] = val_next;
}

static void
phx_counter64_update_single(mx_instance_state_t *is, struct phx_fpga_counter *ctr, int endpoint, int is_clear)
{
  uint8_t *r;
  uint32_t offset;
  uint64_t val_prev, val_current, val_next;
  uint64_t *accum, *accum_last;
  int idx;
  const struct phx_physical_ep *phx_ep;

  /* Special case software counters. */
  if ((ctr->class & CNT_SW_ENDPOINT) == CNT_SW_ENDPOINT) {
    phx_update_software_counter(is, ctr, endpoint, is_clear);
    return;
  }
  if ((ctr->class & COUNTER_TYPE_SW) == COUNTER_TYPE_SW){
     if ((ctr->class & COUNTER_TYPE_UPTIME) == COUNTER_TYPE_UPTIME){
       /* Record the time counters are cleared for myri_counters */
       if (is_clear){
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0)
         struct timespec64 now;
#else
         struct timespec now;
#endif
         mx_getnstimeofday(&now);
         *(ctr->accum) = now.tv_sec;
       }
    }
    return;
  }


  phx_ep = is->board_ops.physical_endpoint(is, endpoint);
  idx = is->board_ops.counter_accum_index(is, ctr, endpoint);
  accum = &ctr->accum[idx];
  accum_last = &ctr->accum_last_64[idx];

  val_prev = *accum_last; /* Last read value */

  /* The current value */
  offset = ctr->offset + (((ctr->class & COUNTER_TYPE_TX) == 0 ?
                           phx_ep->rx_ep : phx_ep->tx_ep) * ctr->stride);

  /* FIXME: When we have HW, fix this calculation. */
  /* Alias stride if specified (i.e. K35SNF) */
  if (ctr->port_stride)
    offset += ctr->port_stride * is->port_id;

  r = (uint8_t *)is->phx.nic_mem + offset;
  if ((ctr->class & COUNTER_TYPE_64BIT)) {
    val_current = PHX_REG_READ_8(r);
    val_next = val_current;
  } else if ((ctr->class & COUNTER_TYPE_48BIT_ROLL)) {
    val_current = PHX_REG_READ_8(r);
    val_current |= ((val_prev >> 48) << 48); /* 16 most significant bits from val_prev */
    if ((val_prev << 16) > (val_current << 16)) {
      /* The 48-bit part has wrapped.  Assume it wrapped once.  As is, we
       * cannot if it has wrapped multiple times.
       */
      val_current += (1ULL << 48);
    }
    val_next = val_current;
  } else if ((ctr->class & COUNTER_TYPE_2_32BIT)) {
    uint32_t lsw, msw;

    /* Read LSW first, which causes HW to ready MSW.  Then read MSW. */
    lsw = PHX_REG_READ_4(r);
    msw = PHX_REG_READ_4((r + 4));
    val_current = (((uint64_t)msw) << 32) | (uint64_t)lsw;
    val_next = val_current;
  } else {
    /* 32-bit counters */
    val_current = (uint64_t) PHX_REG_READ_4(r); /* Fill low 32 bits */
    if ((!is_clear) && (!(ctr->class & COUNTER_TYPE_SNAPSHOT))) {
        /* If this is to clear counters then don't capture prior MSW,
        * which clears it for next run.
        * Snapshot 32-bit counters should never have an MSW
        */
      val_current += (val_prev & 0xffffffff00000000ULL); /* use MSW from last known 64 bit value */
      if (val_current < val_prev) {    /* if count decreased then at least one wrap occurred */
#if _VERBOSE
        MX_INFO((" 32 bit counter %d wrapped, incrementing MSW\n", ctr->id));
#endif
        val_current += 0x100000000ULL; /* FIXME with DMA updates. */
      }
    }
    val_next = val_current;   /* 64-bits for next comparison */
  }

  /* Set the accumulated count correctly */
  if (ctr->class & COUNTER_TYPE_SNAPSHOT) {
    if (ctr->class & COUNTER_TYPE_16_LO)
      *accum = (val_current & 0xffff); /* Low 16 bits */
    else if (ctr->class & COUNTER_TYPE_16_HI)
      *accum = ((val_current & 0xffff0000L) >> 16); /* Upper 16 bits */
    else
      *accum = val_current;
  } else if (ctr->class & COUNTER_TYPE_SELFCLEAR) {
    *accum += val_current; /* Add the current value to accum, as reading register caused it to clear */
    val_next = 0;
  } else {
    /* Add difference to current value.  Need diff because accum might get cleared. */
    *accum += (val_current - val_prev);
  }

  /* Clear applies to all counts except the cycle count */
  if ((is_clear) && (ctr->id != 0))
    *accum = 0;

  *accum_last = val_next; /* Next reading is relative to this */
}

static void
phx_counters64_update_all(mx_instance_state_t *is, int is_clear)
{
  const  phx_counter_t *counters;
  struct phx_fpga_instance *fpga = is->phx.pfi;
  int i;
  uint32_t count;

  is->board_ops.get_counter_definitions(is, &counters, &count);

  mx_spin_lock(&fpga->counters.spinlock);
  fpga->counters.seqnum++;
  MAL_WRITEBAR();

  /* Update all endpoints, global, and this function's counters */
  for (i = 0; i < fpga->counters.defined_cnt; i++) {
    struct phx_fpga_counter *ctr = &fpga->counters.counters[i];
    int ep, ep_cnt;

    /* An endpoint counter has multiple copies */
    if ((ctr->class & COUNTER_TYPE_ENDPOINT))
      ep_cnt = fpga->ep_pool.num_function_endpoints;
    else
      ep_cnt = 1; /* This function's or global counter */

#ifdef DEBUG_COUNTER_ALLOC
    MX_INFO(("Counter = %i, %s ep_cnt = %i\n", i, counters[i].str, ep_cnt));
#endif
    for (ep = 0; ep < ep_cnt; ep++)
      phx_counter64_update_single(is, ctr, ep, is_clear);
  }

  MAL_WRITEBAR();
  if (is_clear)
    phx_reset_pcie_counters(is);   /* reset the PCIE histogram counts */

  fpga->counters.seqnum++;
  mx_spin_unlock(&fpga->counters.spinlock);
}

/* Save the driver uptime for myri_counters */
static void
phx_counters_uptime_init(mx_instance_state_t *is)
{
  const phx_counter_t *counters;
  struct phx_fpga_instance *fpga = is->phx.pfi;
  uint32_t count, i;

  is->board_ops.get_counter_definitions(is, &counters, &count);

  /* Search the phx_*_counters table until we find the driver uptime entry */
  /* and store the current time in seconds there.                          */ 
  for (i = 0; i < count; i++) {
    struct phx_fpga_counter *cnt = &fpga->counters.counters[i];
    if((cnt->class & COUNTER_TYPE_DRV_UPTIME) == COUNTER_TYPE_DRV_UPTIME){
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0)
      struct timespec64 ts;
#else
      struct timespec ts;
#endif
      mx_getnstimeofday(&ts);
      *(cnt->accum) = ts.tv_sec;
      break;
    }
  }
}

/* Fixup the Tx counters which now have a variable stride */
void
phx_counters_tx_fixup(mx_instance_state_t *is)
{
  const phx_counter_t *counters;
  struct phx_fpga_instance *fpga = is->phx.pfi;
  uint32_t count, i;
  uint32_t mask;

  is->board_ops.get_counter_definitions(is, &counters, &count);

  mask = (CNT_ENDPOINT | COUNTER_TYPE_TX);

  /* Search the phx_*_counters table and fixup any Tx endpoint  counters */
  for (i = 0; i < count; i++) {
    struct phx_fpga_counter *cnt = &fpga->counters.counters[i];
    if((cnt->class & mask) == mask){
      cnt->stride = fpga->tx_config.dbltx_core_stride;
    }
  }
}

/* When an endpoint is closed, the counters are automatically reset.
 * This routine grabs them before that happens and set the counters so they
 * make sense to the accumulation code.
 */
void
myri_counters64_ep_close(mx_instance_state_t *is, int ep)
{
  struct phx_fpga_instance *fpga = is->phx.pfi;
  int i;

  mx_spin_lock(&fpga->counters.spinlock);
  fpga->counters.seqnum++;
  MAL_WRITEBAR();

  for (i = 0; i < fpga->counters.defined_cnt; i++) {
    struct phx_fpga_counter *ctr = &fpga->counters.counters[i];
    /* We are looking for 48-bit and 64-bit endpoint counters that are autocleared in HW */
    if (((ctr->class & CNT_ENDPOINT_64) == CNT_ENDPOINT_64) ||
        ((ctr->class & COUNTER_TYPE_48BIT_ROLL) == COUNTER_TYPE_48BIT_ROLL)) {
      /* We are not zero'ing the counter.
       * It is just that the hardware counter resets to zero.
       * So update the current value and zero the "last read" value.
       */
      phx_counter64_update_single(is, ctr, ep, 0);
      ctr->accum_last_64[is->board_ops.counter_accum_index(is, ctr, ep)] = 0;
    }
  }
  MAL_WRITEBAR();
  fpga->counters.seqnum++;
  mx_spin_unlock(&fpga->counters.spinlock);
}

#ifdef MYRI_DBL_AB
/*
 * The a/b interface is always on the FPGA whether compiled in or not,
 * so clearing the counters should not be a problem.
*/
void __phx_ab_counters_clear(mx_instance_state_t *is){

  struct phx_fpga_instance *fpga = is->phx.pfi;
  uint8_t *r;

  mx_spin_lock(&fpga->counters.spinlock);

  r = (uint8_t *)is->phx.nic_mem + PHX_AB_BASE + PHX_AB_RESET_COUNTS;
  PHX_REG_WRITE_4(r, 1);

  MAL_WRITEBAR();
  mx_spin_unlock(&fpga->counters.spinlock);
}
#endif

void
myri_counters64_update(mx_instance_state_t *is)
{
  phx_counters64_update_all(is, 0 /* do not clear */);
}

void
myri_counters64_clear(mx_instance_state_t *is)
{
  phx_counters64_update_all(is, 1 /* clear */);
}

void
myri_counters_uptime_init(mx_instance_state_t *is)
{
  phx_counters_uptime_init(is);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0)
void phx_clk_set(mx_instance_state_t *is, const struct timespec64 *ts)
#else
void phx_clk_set(mx_instance_state_t *is, const struct timespec *ts)
#endif
{
  uint8_t *r;
  uint32_t fsec;
  int i = 0;

  /* While the time storage locations are logically 64b, a 4B write to the LS 
   * stages the lower 32b, while the subsequent 4B write to the MS commits the
   * entire 64b. 
   */

  /* convert fractions of a second into nanoseconds 
   * each second fraction (fsec) is 1/(2^32) seconds.
   * a nanosecond is 1/(10^9) seconds. */
  fsec = (uint32_t)(((uint64_t)ts->tv_nsec * ((uint64_t)1 << 32))
  / (uint64_t) 1000000000);
  
  /* NOTE: Do to an artifact with the write it is necessary to do it twice
   *       to avoid being off by one second. */
  for (i = 0; i < 2; i++) {

    r = (uint8_t *)is->phx.nic_mem + PHX_TIME_LS;
    PHX_REG_WRITE_4(r, fsec);
    
    r = (uint8_t *)is->phx.nic_mem + PHX_TIME_MS;
    PHX_REG_WRITE_4(r, (uint32_t)ts->tv_sec);
  }

}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0)
void phx_clk_set_with_pps(mx_instance_state_t *is, const struct timespec64 *ts)
#else
void phx_clk_set_with_pps(mx_instance_state_t *is, const struct timespec *ts)
#endif
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0)
  struct timespec64 adapter_ts;
  struct timespec64 new_ts;
#else
  struct timespec adapter_ts;
  struct timespec new_ts;
#endif
  int count = 0;
  
 phx_clk_set_pps_again:
  phx_clk_get(is, &adapter_ts);

  new_ts.tv_sec = ts->tv_sec;  

  if (ts->tv_nsec > 800000000 && adapter_ts.tv_nsec < 200000000) {
    /* the hosts nsecs are large, while the adapter's nsecs are small
       which means the adapter has just crossed a second boundry before
       the host.  We need to add one second to the value we set in 
       the adapter */
    new_ts.tv_sec += 1;
  } else if (ts->tv_nsec < 200000000 && adapter_ts.tv_nsec > 800000000) {
    /* the opposite of the above. */
    new_ts.tv_sec -= 1;
  }
   
  /* keep the nsecs the same to minimize effect on phoenix clock */
  new_ts.tv_nsec = adapter_ts.tv_nsec;
  phx_clk_set(is, &new_ts);      
  
  phx_clk_get(is, &new_ts);
  if (adapter_ts.tv_nsec > new_ts.tv_nsec) {
    /* we rolled over sometime between the two adapter clock gets,
       and can't be sure if the change succeeded.  Try again */
    if (count < 5) {
      count++;
      goto phx_clk_set_pps_again;
    } else {
      MX_WARN(("Failed to set clock safely after 5 attempts!\n"));
    }
  }

  /* if (ts->tv_nsec > 800000000 && adapter_ts.tv_nsec < 200000000) { */
  /*   printk("*** adapter was ahead! adding one second\n"); */
  /* } else if (ts->tv_nsec < 200000000 && adapter_ts.tv_nsec > 800000000) { */
  /*   printk("*** host was ahead! subracting subtracting one second\n"); */
  /* }  */

}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0)
void phx_clk_get(mx_instance_state_t *is, struct timespec64 *ts)
#else
void phx_clk_get(mx_instance_state_t *is, struct timespec *ts)
#endif
{
  uint8_t *r;
  uint32_t fsec;

  /* The two registers contain the Time Fractional Seconds and the 
   * Time Integer Seconds. Reading the fractional seconds stages
   * the integer seconds for reading. Therefore the fractional
   * seconds must be read first. */
    
  /* read fractions of a second */
  r = (uint8_t *)is->phx.nic_mem + PHX_TIME_LS;
  fsec = PHX_REG_READ_4(r);

  /* read integer seconds */
  r = (uint8_t *)is->phx.nic_mem + PHX_TIME_MS;
  ts->tv_sec = PHX_REG_READ_4(r);

  /* each second fraction (fsec) is 1/(2^32) seconds.
     a nanosecond is 1/(10^9) seconds. */  
  ts->tv_nsec = (long)(((uint64_t)fsec * (uint64_t)1000000000)/
               (uint64_t)(1ULL<<32));

}


void phx_clk_adjtime(mx_instance_state_t *is, int64_t delta)
{
  
  uint8_t *r;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0)
  struct timespec64 ts;
#else
  struct timespec ts;
#endif
  if (PHX_HAS_FEATURE(is, PHX_FEATURE_TIMEBASE_V2) ||
      delta > 1000000000) {
    phx_clk_get(is, &ts);
    ts.tv_sec  = ts.tv_sec  + delta / 1000000000;
    ts.tv_nsec = ts.tv_nsec + delta % 1000000000;
    phx_clk_set(is, &ts);
    //printk("Adjusted the clock crudely\n");
    
    /*TODO:  Get the fine clock adjustments using the dedicated
      functionality working. 
      
      NOTE:  The fine clock adjustments delivered by Atomic Rules 
      on the 'new' time server has a limitation that it can only
      do 7629 ns at a time, so we may want to continue doing the
      'crude' adjustment above for the larger values of delta to
      avoid iterating many times.
    */
  } else {
    /* Timebase version 1 */
    r = (uint8_t *)is->phx.nic_mem + PHX_TIME_V1_SKIP_NS;
    PHX_REG_WRITE_4(r, (int32_t)delta);
  } 
}

void phx_clk_adjfreq(mx_instance_state_t *is, int32_t ppb)
{
  uint8_t *r;

  /* The new timebase in the firmware had different
     method for adjustment.  Because of reuse of register offsets 
     this function now does nothing if the firmware version is
     not identified as that would lead to bad results */

  if (is->phx.pfi->firmware_rev == 0) {
    MX_WARN(("Can't adjust clock frequency if the firmware version isn't known!\n"));
    return;
  } else if (!PHX_HAS_FEATURE(is, PHX_FEATURE_TIMEBASE_V2)) {
    r = (uint8_t *)is->phx.nic_mem + PHX_TIME_V1_TRIM_PPB;
    PHX_REG_WRITE_4(r, ppb);
  } else {

    /* The frequency adjustment is the following:

       adjusted_increment = nominal_100_MHz_increment + PPB * increment_per_PPB

       where

       nominal_100_MHz_increment = 184467440737.09551616 #   2**64 / 100 MHz
       increment_per_PPB = 184.46744073709551616         #   2**64 / 100 MHz / 1e9

       But we don't want to do floating point in kernel, so we are scaling up
       the values a bit and and hoping for the best.
    */

    int64_t new_inc;
    r = (uint8_t *)is->phx.nic_mem + PHX_TIME_V2_SET_CFREQ_LS;

    new_inc = ((PHX_TIME_V2_INC_NOMINAL_scaled_up 
		+ ((int64_t)PHX_TIME_V2_INC_PER_PPB_scaled_up * ppb)
		+ (PHX_TIME_V2_INC_scale_factor/2)) /*rounding*/
	       / PHX_TIME_V2_INC_scale_factor);
    
    PHX_REG_WRITE_8(r, new_inc);
    //printk("Writing %016llx (%lld) to %p ppb %d\n", new_inc, new_inc, r, ppb);

  }
}

void phx_clk_init(mx_instance_state_t *is, int pps_invert, int disable_pps,
		  int disable_10mhz) {

  uint8_t *r;
  uint32_t regControl;
  uint64_t regGpioOut;

  /* per an email from Bach Long @ Atomic Rules, 1/23/15 : 
     These are timestamp-related registers that need to be
     initialized in each MAC : 

     tx_fns_adjustment_10G (offset 0x408) := 0xD4BC
     tx_ns_adjustment_10G (offset 0x410) := 0xA
     rx_fns_adjustment_10G (offset 0x488) := 0x3062
     rx_ns_adjustment_10G (offset 0x490) := 0xA

     These are static timing adjustment amounts in nanoseconds
     and fractional nanoseconds, calculated according to the
     references [in the email]
  */
  r = (uint8_t *)is->phx.nic_mem + PHX_TIME_TX_FNS_ADJUSTMENT;
  PHX_REG_WRITE_4(r, 0xD4BC);
  r = (uint8_t *)is->phx.nic_mem + PHX_TIME_TX_NS_ADJUSTMENT;
  PHX_REG_WRITE_4(r, 0xA);
  r = (uint8_t *)is->phx.nic_mem + PHX_TIME_RX_FNS_ADJUSTMENT;
  PHX_REG_WRITE_4(r, 0x3062);
  r = (uint8_t *)is->phx.nic_mem + PHX_TIME_RX_NS_ADJUSTMENT;
  PHX_REG_WRITE_4(r, 0xA);

  /* the GPIO registers control the 10MHz input */
  r = (uint8_t *)is->phx.nic_mem + PHX_GPIO_OUT;
  regGpioOut = PHX_REG_READ_8(r);
  if (disable_10mhz) {
    regGpioOut &= ~PHX_TIME_GPIO_OUT_10MHz_Enable;
    MX_INFO(("Disabling 10Mhz input\n"));
  } else {
    regGpioOut |= PHX_TIME_GPIO_OUT_10MHz_Enable;
    MX_INFO(("Enabling 10Mhz input\n"));
  }
  PHX_REG_WRITE_8(r, regGpioOut);

  /* the PPS is controlled by the PHX_TIME_STATUS / CONTROL registers */
  r = (uint8_t *)is->phx.nic_mem + PHX_TIME_CONTROL;
  regControl = PHX_REG_READ_4(r);
  if (disable_pps) {
    regControl |= PHX_TIME_CONTROL_PpsInDisable;
    MX_INFO(("Disabling PPS input\n"));
  } else {
    regControl &= ~PHX_TIME_CONTROL_PpsInDisable;
    MX_INFO(("Enabling PPS input\n"));
    if (!pps_invert) {
      /* The firmware seems to have swapped the pps sense, so if we
 	 we are not asked to invert, set the invert bit on fpga.*/
      regControl |= PHX_TIME_CONTROL_InvertPps;
    } else {
      regControl &= ~PHX_TIME_CONTROL_InvertPps;
      MX_INFO(("Inverting PPS input\n"));
    }
  }

  PHX_REG_WRITE_4(r, regControl); 

  /* Zero out frequency control register */
  phx_clk_adjfreq(is, 0);

}
uint32_t phx_clk_get_pps_status(mx_instance_state_t *is) {
  uint32_t status;
  uint8_t *r;

  r = (uint8_t *)is->phx.nic_mem + PHX_TIME_STATUS;
  status = PHX_REG_READ_4(r);
  
  return status;
}

int phx_clk_is_pps_active(mx_instance_state_t *is) {
  uint32_t status;
  uint32_t control;
  uint8_t *r;

  status = phx_clk_get_pps_status(is);

  r = (uint8_t *)is->phx.nic_mem + PHX_TIME_CONTROL;
  control = PHX_REG_READ_4(r);

  if ((status & PHX_TIME_STATUS_PpsOK) && 
      !(control & PHX_TIME_CONTROL_PpsInDisable)){
    return 1;
  } else {
    return 0;
  }
}

int phx_clk_is_10mhz_active(mx_instance_state_t *is) {
  uint64_t status;
  uint64_t control;
  uint8_t *r;
  
  r = (uint8_t *)is->phx.nic_mem + PHX_GPIO_OUT;
  control = PHX_REG_READ_8(r);
  
  r = (uint8_t *)is->phx.nic_mem + PHX_GPIO_IN;
  status = PHX_REG_READ_8(r);
  
  if ((control & PHX_TIME_GPIO_OUT_10MHz_Enable) &&
      (status & PHX_TIME_GPIO_IN_10MHz_PLL)) {
    return 1;
  } else {
    return 0;
  }
}

void
phx_clk_monitor(mx_instance_state_t *is)
{
  uint32_t new_pps_state;
  uint32_t new_10mhz_state;
  uint64_t host_time;
  struct phx_fpga_instance *fpga = is->phx.pfi;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0)
  struct timespec64 host_ts;
#else
  struct timespec host_ts;
#endif
  mx_spin_lock(&fpga->clk.clk_spinlock);

  /* check to see if we haven't updated the status in more than
     second */

  mx_getnstimeofday(&host_ts);
  host_time = host_ts.tv_sec * 1000000000 + host_ts.tv_nsec;
  /* In case the timer fires slightly early, check that it has 
     been almost a second since the last check instead of exactly one second.*/
  if (host_time - fpga->clk.last_update_host_time < 999000000) {
    /* already updated this recently enough, return now */
    mx_spin_unlock(&fpga->clk.clk_spinlock);
    return;
  } 

  //MX_INFO(("Instance %d checking fpga %d %llu %llu\n", 
  //is->id, fpga->id, host_time, fpga->clk.last_update_host_time));

  new_pps_state = phx_clk_get_pps_status(is) & PHX_TIME_STATUS_MASK;
  new_10mhz_state = phx_clk_is_10mhz_active(is);

  if (new_pps_state != fpga->clk.last_pps_state) {
    
    uint32_t mask = (PHX_TIME_STATUS_PpsLost | PHX_TIME_STATUS_PpsOK);
    //MX_INFO(("PPS state has changed! clock %d old 0x%x new 0x%x\n",
    //        fpga->id, fpga->clk.last_pps_state, new_pps_state));

    if (((fpga->clk.last_pps_state & mask) == PHX_TIME_STATUS_PpsOK) &&
	((new_pps_state & mask) != PHX_TIME_STATUS_PpsOK)) {
      /* we just lost PPS ! */
      fpga->clk.pps_lost_host_time = host_time;
      is->clk_ts_state = MYRI_CLK_TS_INTERNAL_TCXO;
      MX_INFO(("Lost PPS signal on clock %d (0x%x)\n",
	       fpga->id, new_pps_state));
    } else if (((fpga->clk.last_pps_state & mask) != PHX_TIME_STATUS_PpsOK) &&
	       ((new_pps_state & mask) == PHX_TIME_STATUS_PpsOK)) {
      /* we transitioned from bad pps to good pps */
      fpga->clk.pps_good_host_time = host_time;
      
#ifdef SET_FPGA_CLOCK_ON_PPS_RESYNC
      mx_getnstimeofday(&host_ts);
      phx_clk_set_with_pps(is, &host_ts);
      MX_INFO(("PPS now good on clock %d. Set the seconds\n", fpga->id));
#endif

      is->clk_ts_state = MYRI_CLK_TS_EXTERNAL_PPS;
      MX_INFO(("PPS now good on clock %d.\n", fpga->id));
    } else {
      MX_INFO(("PPS state has changed! clock %d old 0x%x new 0x%x\n",
	       fpga->id, fpga->clk.last_pps_state, new_pps_state));
    }
    fpga->clk.last_pps_state = new_pps_state;
  }

  if (new_10mhz_state != fpga->clk.last_10mhz_state) {
    if (new_10mhz_state) {
      MX_INFO(("Detected 10Mhz signal on clock %d. Using it\n",
	       fpga->id));
    } else {
      MX_INFO(("Lost 10Mhz signal on clock %d.\n", fpga->id));
    }
    //MX_INFO(("10Mhz state has changed! clock %d old 0x%x new 0x%x\n",
    //      fpga->id, fpga->clk.last_10mhz_state, new_10mhz_state));
    fpga->clk.last_10mhz_state = new_10mhz_state;
  }

  fpga->clk.last_update_host_time = host_time;

  mx_spin_unlock(&fpga->clk.clk_spinlock);
}

void
phx_sw_counters_ep_open(mx_endpt_state_t *es)
{
  struct phx_fpga_instance *fpga = es->is->phx.pfi;

  /* Remember this endpoint while it is open.
   * myri_counters64_{update,clear} use it to report software counters.
   */
  mx_spin_lock(&fpga->counters.spinlock);
  mx_spin_lock(&fpga->counters.es_spinlock);
  fpga->counters.es[es->phx_ep->tx_ep] = es;
  mx_spin_unlock(&fpga->counters.es_spinlock);
  mx_spin_unlock(&fpga->counters.spinlock);
}

void
phx_sw_counters_ep_close(mx_endpt_state_t *es)
{
  struct phx_fpga_instance *fpga = es->is->phx.pfi;
  phx_endpt_flow_tx_stats_t *tx_stats, *ep_tx_stats;

  mx_spin_lock(&fpga->counters.spinlock);
  mx_spin_lock(&fpga->counters.es_spinlock);

  tx_stats = &fpga->counters.ep_tx_stats[es->phx_ep->tx_ep];
  ep_tx_stats = &es->flow_window->tx_stats;

  /* Accumulate software counters */
  tx_stats->tx_pio_packets += ep_tx_stats->tx_pio_packets;
  tx_stats->tx_pio_bytes += ep_tx_stats->tx_pio_bytes;
  tx_stats->tx_dma_packets += ep_tx_stats->tx_dma_packets;
  tx_stats->tx_dma_bytes += ep_tx_stats->tx_dma_bytes;
  tx_stats->rx_packets += ep_tx_stats->rx_packets;
  tx_stats->rx_packets_mac_error += ep_tx_stats->rx_packets_mac_error;

  /* Remove the endpoint */
  fpga->counters.es[es->phx_ep->tx_ep] = NULL;

  mx_spin_unlock(&fpga->counters.es_spinlock);
  mx_spin_unlock(&fpga->counters.spinlock);
}

void 
phx_get_fpga_ports(mx_instance_state_t *is, myri_get_all_ports_t *units)
{
  struct phx_fpga_instance *fpga = is->phx.pfi;
  int ports, i;
 
  if (fpga){
     ports = fpga->attach_count;
     if ((ports <= 0) || (ports > 4))
	ports = 1; 

     for (i = 0; i < ports; i++){
       units->port_board[i] = fpga->is[i]->id;
     }
  }
  else {
    /* this may not be needed */
    ports = 1;
    units->port_board[0] = is->id;
  }
  units->num_ports = ports;
}

uint32_t
phx_get_dmasend_threshold(mx_instance_state_t *is)
{
  struct phx_fpga_instance *fpga = is->phx.pfi;

  if (fpga)
    return fpga->tx_config.dmasend_threshold;

  return PHX_DMASEND_THRESHOLD_SIZE;   /* phoenix, least risky default */
}

void
phx_attach_fpga_instance(mx_instance_state_t *is)
{
  struct phx_fpga_instance *fpga = NULL;
  int i;

  /* If already attached, return right away, a different
   * instance which is not yet attached may keep on enumerating
   * additional instances
   */
  if (is->phx.pfi)
    return;

  mx_mutex_enter(&mx_fpga_mutex);

  for(i = 0; i < phx_fpga_num_instances; i++) {
    fpga = &phx_fpga_instances[i];
    if (fpga->bus == is->phx.pci_info.bus
        && fpga->dev == is->phx.pci_info.dev)
      break;
  }

  /* The first time we see this board/FPGA. */
  if (i == phx_fpga_num_instances) {
    fpga = &phx_fpga_instances[phx_fpga_num_instances++];
    fpga->id = i;
    fpga->bus = is->phx.pci_info.bus;
    fpga->dev = is->phx.pci_info.dev;
  }

  /* The first time we see this FPGA's function */
  if (fpga->attach_count == 0) {
    fpga->initialized = 0;
    fpga->num_functions = is->phx.num_functions;
#if MAL_OS_WINDOWS
    fpga->avl_initialized = 0;
#endif
    fpga->firmware_rev = 0;
    fpga->eeprom_strings = mx_kmalloc(MX_EEPROM_STRINGS_LEN, MX_NOWAIT);

    /* This sets up endpoint numbers */
    is->board_ops.init_endpoint_pool(is, &fpga->ep_pool);

    /* initialize stuff to keep track of the phx clk state */
    mx_spin_lock_init(&fpga->clk.clk_spinlock, is, -1, "fpga clock spinlock");
    fpga->clk.last_pps_state = phx_clk_get_pps_status(is) & PHX_TIME_STATUS_MASK;
    fpga->clk.last_10mhz_state = phx_clk_is_10mhz_active(is);
    fpga->clk.last_update_host_time = 0;
    fpga->clk.pps_lost_host_time = 0;
    fpga->clk.pps_good_host_time = 0;

    /* Initialize HW hash status */
    fpga->hw_hash.enabled_flags =
        mx_kmalloc(sizeof(uint32_t) * is->num_ports, MX_MZERO);
    fpga->hw_hash.in_use =
        mx_kmalloc(sizeof(int) * is->num_ports, MX_MZERO);
  }

  fpga->is[fpga->attach_count++] = is;
  is->phx.pfi = fpga;
  MX_INFO(("Attach instance to FPGA: is_id=%d pci=%d:%d.%d fpga_id=%d attach_count=%d\n",
           is->id, is->phx.pci_info.bus, is->phx.pci_info.dev,
           is->phx.pci_info.func, fpga->id, fpga->attach_count));
  mx_mutex_exit(&mx_fpga_mutex);
}

void
phx_detach_fpga_instance(mx_instance_state_t *is)
{
  struct phx_fpga_instance *fpga = is->phx.pfi;
  if (fpga) {
    mx_mutex_enter(&mx_fpga_mutex);
    MX_INFO(("Detach instance from FPGA: is_id=%d pci=%d:%d.%d fpga_id=%d attach_count=%d\n",
             is->id, is->phx.pci_info.bus, is->phx.pci_info.dev,
             is->phx.pci_info.func, fpga->id, fpga->attach_count));
    fpga->attach_count--;
    if (fpga->attach_count == 0) {
      if (fpga->eeprom_strings){
        mx_kfree(fpga->eeprom_strings);
        fpga->eeprom_strings = NULL;
      }
      if (fpga->ep_pool.endpoints) {
        mx_kfree(fpga->ep_pool.endpoints);
        fpga->ep_pool.endpoints = NULL;
      }
      if (fpga->hw_hash.enabled_flags) {
        mx_kfree(fpga->hw_hash.enabled_flags);
        fpga->hw_hash.enabled_flags = NULL;
      }
      if (fpga->hw_hash.in_use) {
        mx_kfree(fpga->hw_hash.in_use);
        fpga->hw_hash.in_use = NULL;
      }

      /* Do de-init sequence if any.  There is none right now. */
      fpga->initialized = 0;
      mx_spin_lock_destroy(&fpga->clk.clk_mutex);
    }
    mx_mutex_exit(&mx_fpga_mutex);
  }
}

int
phx_claim_fpga_endpoint(mx_instance_state_t *is, int endpoint_id, int reserved)
{
  int err = 0;
  struct phx_physical_ep *ep;

  if ((endpoint_id < 0) || (endpoint_id >= is->phx.pfi->ep_pool.num_total_endpoints))
    return ENODEV;

  ep = &is->phx.pfi->ep_pool.endpoints[endpoint_id];

  mx_mutex_enter(&mx_fpga_mutex);
  if (ep->owner_is != NULL)
    err = EBUSY;
  else if (ep->reserved && !reserved)
    err = EBUSY;
  else if ((ep->func == -1) || (ep->func == is->phx.pci_info.func)) {
    ep->owner_is = is;
    ep->port_id = is->port_id;
  } else
    err = ENODEV;    /* this device can't access this endpoint */
  /* For K35SNF, set the kernel bit.  For others, it is harmless. */
  if (!err && reserved)
    ep->kernel = 1;
  mx_mutex_exit(&mx_fpga_mutex);
  return err;
}

void
phx_release_fpga_endpoint(mx_instance_state_t *is, int endpoint_id)
{
  struct phx_physical_ep *ep;

  /* Silently ignore invalid endpoint. */
  if ((endpoint_id < 0) || (endpoint_id >= is->phx.pfi->ep_pool.num_total_endpoints))
    return;

  ep = &is->phx.pfi->ep_pool.endpoints[endpoint_id];
  ep->port_id = 0;
  ep->kernel = 0;

  mx_mutex_enter(&mx_fpga_mutex);
  if (ep->owner_is != NULL)
    ep->owner_is = NULL;
  mx_mutex_exit(&mx_fpga_mutex);
}

/* Global n'th endpoint.  It may not belong to the given instance. */
int
phx_fpga_endpoint_owner_instance_id(mx_instance_state_t *is, int endpoint)
{
  mx_instance_state_t *user;
  int id = -1;
  if (endpoint >= 0 && endpoint < is->phx.pfi->ep_pool.num_total_endpoints) {
    mx_mutex_enter(&mx_fpga_mutex);
    if ((user = is->phx.pfi->ep_pool.endpoints[endpoint].owner_is))
      id = user->id;
    mx_mutex_exit(&mx_fpga_mutex);
  }
  return id;
}

int
phx_alloc_fpga_tx_cpage(mx_instance_state_t *is)
{
  struct phx_fpga_instance *fpga = is->phx.pfi;
  int status;
  if (fpga) {
    mx_mutex_enter(&mx_fpga_mutex);
    if (!fpga->tx_cpage.addr) {
      status = mx_alloc_zeroed_dma_page_relax_order(is,
                                                    &fpga->tx_cpage.addr,
                                                    &fpga->tx_cpage.pin);
      if (status) {
        MX_WARN(("Could not allocate the TX completion DMA page\n"));
        fpga->tx_cpage.addr = NULL;
      }
    }
    if (fpga->tx_cpage.addr) {
      is->tx_done.addr = fpga->tx_cpage.addr;
      is->tx_done.pin = fpga->tx_cpage.pin;
      fpga->tx_cpage.ref_count++;
#if _VERBOSE
      MX_INFO(("%s: tx_cpage ref_count=%d\n", __FUNCTION__, fpga->tx_cpage.ref_count));
#endif
      status = 0;
    }
    mx_mutex_exit(&mx_fpga_mutex);
  } else {
    /* This is a bug... */
    MX_WARN(("%s: fpga is not attached.\n", __FUNCTION__));
    status = EINVAL;
  }
  return status;
}

void
phx_free_fpga_tx_cpage(mx_instance_state_t *is)
{
  struct phx_fpga_instance *fpga = is->phx.pfi;
  if (is->tx_done.addr && fpga) {
    mx_mutex_enter(&mx_fpga_mutex);
#if _VERBOSE
    MX_INFO(("%s: tx_cpage ref_count=%d\n", __FUNCTION__, fpga->tx_cpage.ref_count));
#endif
    if (fpga->tx_cpage.addr && --fpga->tx_cpage.ref_count == 0) {
      mx_free_dma_page(is, &fpga->tx_cpage.addr, &fpga->tx_cpage.pin);
      fpga->tx_cpage.addr = NULL;
    }
    mx_mutex_exit(&mx_fpga_mutex);
  }
}

int
phx_alloc_fpga_counters(mx_instance_state_t *is)
{
  struct phx_fpga_instance *fpga = is->phx.pfi;
  int status;

  if (fpga) {
    mx_mutex_enter(&mx_fpga_mutex);

    /* Initialize counters once per FPGA. */
    if (!fpga->counters.es) {
      const phx_counter_t *counters;
      uint64_t *slot;
      uint32_t defined, ep, aliased, aliased_ep, global, total;
      uint32_t i, bits;

      is->board_ops.get_counter_definitions(is, &counters, &defined);
      fpga->counters.defined_cnt = defined;

      /* First count counters per type */
      ep = aliased = aliased_ep = global = 0;
      for (i = 0; i < defined; i++) {
        bits = counters[i].class & (COUNTER_TYPE_ENDPOINT | COUNTER_TYPE_ALIASED |
                                    COUNTER_TYPE_GLOBAL);
        if (bits == (COUNTER_TYPE_ENDPOINT | COUNTER_TYPE_ALIASED))
          aliased_ep++;
        else if (bits == COUNTER_TYPE_ENDPOINT)
          ep++;
        else if (bits == COUNTER_TYPE_GLOBAL)
          global++;
        else if (bits == COUNTER_TYPE_ALIASED)
          aliased++;
        else {
          mal_always_assert(0);
        }
      }

      /* ALIASED means per-function for all board types except K35SNF.
       * For K35SNF, it means per-port.  Since fpga->num_functions == is->num_ports
       * for non-K35SNF, use is->num_ports instead of num_functions.
       */

      total = global + aliased * is->num_ports + ep * fpga->ep_pool.num_total_endpoints +
          aliased_ep * is->num_ports * fpga->ep_pool.num_total_endpoints;

      /* Allocate storage for all counters in the chip + accum_last (x2) */
      fpga->counters.storage = mx_kmalloc((sizeof(uint64_t) * total * 2), MX_MZERO);

#ifdef DEBUG_COUNTER_ALLOC
      MX_INFO(("There are %d total counters\n", total));
      MX_INFO(("There are %d aliased endpoint counters %d ports %d endpoints\n", aliased_ep, fpga->num_functions, fpga->ep_pool.num_total_endpoints));
      MX_INFO(("There are %d endpoint counters %d endpoints\n", ep, fpga->ep_pool.num_total_endpoints));
      MX_INFO(("There are %d aliased counters %d ports\n", aliased, fpga->num_functions));
      MX_INFO(("There are %d global counters\n", global));
      MX_INFO(("Allocated %lu bytes for counter data at %p\n", (sizeof(uint64_t) * total * 2), fpga->counters.storage));
#endif
      /* Counter meta data we use */
      fpga->counters.counters = mx_kmalloc((sizeof(struct phx_fpga_counter) * defined), MX_MZERO);
      if (!fpga->counters.storage || !fpga->counters.counters) {
        status = ENOMEM;
        mx_mutex_exit(&mx_fpga_mutex);
        goto exit;
      }

      /* Go through the counters again, assigning storage slots to each */
      slot = fpga->counters.storage;
      for (i = 0; i < defined; i++) {
        struct phx_fpga_counter *c = &fpga->counters.counters[i];
        int slot_cnt = 0;

        c->id = i;
        c->class = counters[i].class;
        c->offset = counters[i].offset;
        c->stride = counters[i].stride;
        c->port_stride = counters[i].port_stride;

        bits = c->class & (COUNTER_TYPE_ENDPOINT | COUNTER_TYPE_ALIASED |
                           COUNTER_TYPE_GLOBAL);
        if (bits == (COUNTER_TYPE_ENDPOINT | COUNTER_TYPE_ALIASED))
          slot_cnt = fpga->ep_pool.num_total_endpoints * is->num_ports;
        else if (bits == COUNTER_TYPE_ENDPOINT)
          slot_cnt = fpga->ep_pool.num_total_endpoints;
        else if (bits == COUNTER_TYPE_GLOBAL)
          slot_cnt = 1;
        else if (bits == COUNTER_TYPE_ALIASED)
          slot_cnt = is->num_ports;

        c->accum = slot;
        slot += slot_cnt;
        c->accum_last_64 = slot;
        slot += slot_cnt;

#ifdef DEBUG_COUNTER_ALLOC
        MX_INFO(("Counter metadata id=%d class=0x%08x offset=0x%08x stride 0x%x slots=%d\n",
            c->id, c->class, c->offset, c->stride, slot_cnt * 2));
        MX_INFO(("Counter metadata id=%d accum_data  0x%p last_read  0x%p\n",
            c->id, c->accum, c->accum_last_64));
#endif
      }

      /* Software counters */
      fpga->counters.ep_tx_stats = mx_kmalloc(sizeof(phx_endpt_flow_tx_stats_t) *
                                              fpga->ep_pool.num_total_endpoints, MX_MZERO);
      fpga->counters.es = mx_kmalloc(sizeof(mx_endpt_state_t*) *
                                     fpga->ep_pool.num_total_endpoints, MX_MZERO);

#ifdef DEBUG_COUNTER_ALLOC
      MX_INFO(("Allocated room for endpt_flow_tx_stats for %d endpoints\n",
               fpga->ep_pool.num_total_endpoints));
      MX_INFO(("Allocated room for endpt_state pointers for %d endpoints\n",
               fpga->ep_pool.num_total_endpoints));
#endif

      if (fpga->counters.ep_tx_stats == NULL || fpga->counters.es == NULL) {
        /* Do not bother freeing allocated memory.  Let it leak.  In this
         * situation, the machine is already unusable.
         */
        status = ENOMEM;
        mx_mutex_exit(&mx_fpga_mutex);
        goto exit;
      }
      mx_spin_lock_init(&fpga->counters.spinlock, is, -1, "shared counters spinlock");
      mx_spin_lock_init(&fpga->counters.es_spinlock, is, -1, "shared counters endpoint spinlock");
    }
    status = 0;
    fpga->counters.ref_count++;
    mx_mutex_exit(&mx_fpga_mutex);
  } else {
    /* This is a bug... */
    MX_WARN(("%s: fpga is not attached.\n", __FUNCTION__));
    status = EINVAL;
  }

exit:
  return status;
}

void
phx_free_fpga_counters(mx_instance_state_t *is)
{
  struct phx_fpga_instance *fpga = is->phx.pfi;
  if (fpga) {
    mx_mutex_enter(&mx_fpga_mutex);
    if (fpga->counters.es && --fpga->counters.ref_count == 0) {
      mx_kfree(fpga->counters.counters);
      mx_kfree(fpga->counters.storage);
      mx_kfree(fpga->counters.ep_tx_stats);
      mx_kfree(fpga->counters.es);
      mx_spin_lock_destroy(&fpga->counters.spinlock);
      mx_spin_lock_destroy(&fpga->counters.es_spinlock);
      fpga->counters.ep_tx_stats = NULL;
      fpga->counters.es = NULL;
    }
    mx_mutex_exit(&mx_fpga_mutex);
  }
}

/*
 * The eeprom strings on the A5PL/S have the format
 * SN=x\0
 * MAC=x:x:x:x:x:x\0
 * PC=10G-PCIE3-8D-x\0
 * PN=09-046xx\0
 */

int
phx_parse_eeprom_strings(mx_instance_state_t *is)
{
#define MX__NEXT_STRING(p) while(*ptr++)
  char *ptr = is->phx.eeprom_strings;
  char *last_digit;
  int i, hv, lv;

  is->phx.serial = 0;
  is->mac_addr_string = NULL;

  while (*ptr != '\0') {
    if (memcmp(ptr, "SN2=", 4) == 0) {
      is->phx.serial = ptr + 4;
    }
    else if (memcmp(ptr, "SN=", 3) == 0) {
      if (!is->phx.serial)
        is->phx.serial = ptr + 3;
    }
    else if (memcmp(ptr, "MAC=", 4) == 0) {
      char port_id;

      ptr+=4;
      is->mac_addr_string = ptr;
      for (i = 0; i < 6; i++) {
        if ((ptr + 2) - is->phx.eeprom_strings >=  MX_EEPROM_STRINGS_LEN)
          goto abort;

        if (*(ptr+1) == ':') {
          hv = 0;
          lv = mx_digit(*ptr); last_digit = ptr; ptr++;
        }
        else {
          hv = mx_digit(*ptr); ptr++;
          lv = mx_digit(*ptr); last_digit = ptr; ptr++;
        }
        is->perm_mac_addr[i] = is->mac_addr[i] = (hv << 4) | lv;
        ptr++;
      }
      port_id = MX_IS_PORT_ID(is);
      is->perm_mac_addr[5] += port_id;
      is->mac_addr[5] += port_id;
      *last_digit += port_id;
      ptr--;
    }
    else if (memcmp(ptr, "PC=", 3) == 0) {
      is->phx.product_code = ptr + 3;
    }
    else if (memcmp(ptr, "PN=", 3) == 0) {
      is->phx.part_number = ptr + 3;
    }
    else if (ptr[0] < 32 || ptr[0] >= 127) {
      MX_WARN(("skipping invalid eeprom string %s\n", ptr));
    }
    MX__NEXT_STRING(ptr);
  }

  /* Show the mac address, abort if NULL, means eeprom was corrupted or unread */
  if (is->mac_addr_string == NULL)
    goto abort;

  MX_INFO(("%s: MAC address = %s\n", is->is_name, is->mac_addr_string));

  return 0;

abort:
  MX_WARN(("Failed to parse eeprom strings\n"));
  MX_WARN(("strings = %p, ptr = %p\n",is->phx.eeprom_strings, ptr));

  return -1;
}


#if _VERBOSE
#if 0
static void
test_pio(mx_instance_state_t *is)
{
  uint8_t *scratch = (uint8_t*)is->phx.nic_mem + is->phx.pfi->tx_config.dbltx_pio_base; 
  uint8_t d_8b = 0;
  uint16_t d_16b = 0;
  uint32_t d_32b = 0;
  uint64_t d_64b = 0;
  uint8_t buf[64*2] = { 0 };
  uint64_t *aligned_64B;

  aligned_64B = (uint64_t*)(((uintptr_t)buf + 63) & ~63);

  phx_bbsysctl_pio_print_debug_info(is->phx.nic_mem);

  *((uint8_t*)scratch) = d_8b;
  MAL_STBAR();
  MX_INFO(("Did 1B PIO\n"));
  phx_bbsysctl_pio_print_debug_info(is->phx.nic_mem);

  *((uint16_t*)scratch) = d_16b;
  MAL_STBAR();
  MX_INFO(("Did 2B PIO\n"));
  phx_bbsysctl_pio_print_debug_info(is->phx.nic_mem);

  *((uint32_t*)scratch) = d_32b;
  MAL_STBAR();
  MX_INFO(("Did 4B PIO\n"));
  phx_bbsysctl_pio_print_debug_info(is->phx.nic_mem);

  *((uint64_t*)scratch) = d_64b;
  MAL_STBAR();
  MX_INFO(("Did 8B PIO\n"));
  phx_bbsysctl_pio_print_debug_info(is->phx.nic_mem);

  *(((uint64_t*)scratch) + 0) = aligned_64B[0];
  *(((uint64_t*)scratch) + 1) = aligned_64B[1];
  MAL_STBAR();
  MX_INFO(("Did 16B PIO\n"));
  phx_bbsysctl_pio_print_debug_info(is->phx.nic_mem);

  *(((uint64_t*)scratch) + 0) = aligned_64B[0];
  *(((uint64_t*)scratch) + 1) = aligned_64B[1];
  *(((uint64_t*)scratch) + 2) = aligned_64B[2];
  *(((uint64_t*)scratch) + 3) = aligned_64B[3];
  MAL_STBAR();
  MX_INFO(("Did 32B PIO\n"));
  phx_bbsysctl_pio_print_debug_info(is->phx.nic_mem);

  *(((uint64_t*)scratch) + 0) = aligned_64B[0];
  *(((uint64_t*)scratch) + 1) = aligned_64B[1];
  *(((uint64_t*)scratch) + 2) = aligned_64B[2];
  *(((uint64_t*)scratch) + 3) = aligned_64B[3];
  *(((uint64_t*)scratch) + 4) = aligned_64B[4];
  *(((uint64_t*)scratch) + 5) = aligned_64B[5];
  *(((uint64_t*)scratch) + 6) = aligned_64B[6];
  *(((uint64_t*)scratch) + 7) = aligned_64B[7];
  MAL_STBAR();
  MX_INFO(("Did 64B PIO\n"));
  phx_bbsysctl_pio_print_debug_info(is->phx.nic_mem);
}
#endif
#endif

int
phx_common_map_board(mx_instance_state_t *is)
{
#if _VERBOSE
  uint32_t val;
  uint8_t *r;
#endif
  uint32_t sha1_val[4];

  is->phx.nic_mem = mx_map_pci_space(is, 0, 0, is->board_span);
  is->phx.i2c_base_mem = (char *) is->phx.nic_mem + PHX_I2C_BASE;  /* offset to i2c cores */
  if (is->phx.nic_mem == NULL) {
    MX_WARN(("phx_map_board failed\n"));
    return ENODEV;
  }

#if _VERBOSE
  MX_INFO(("Phoenix board mapped to 0x%p\n", (void *)is->phx.nic_mem));
  MX_INFO(("%s %d:Performing sanity checks on phoenix board\n",
	   __FUNCTION__, __LINE__));

  r = (uint8_t *)is->phx.nic_mem + PHX_BBSYSCTL_CONST0;
  MX_INFO(("First sanity check is reading %p\n", r));
  val = PHX_REG_READ_4(r);
  if (val != 0xCAFEF00D) {
    MX_INFO(("First sanity check failed.  Got %08x.\n", val));
    phx_common_unmap_board(is);
    return ENODEV;
  } else {
    MX_INFO(("First sanity check passed.\n"));
  }

  r = (uint8_t *)is->phx.nic_mem + PHX_BBSYSCTL_CONST1;
  MX_INFO(("Second sanity check is reading %p\n", r));
  val = PHX_REG_READ_4(r);
  if (val != 0xDEADBEEF) {
    MX_INFO(("Second sanity check failed.  Got %08x.\n", val));
    phx_common_unmap_board(is);
    return ENODEV;
  } else {
    MX_INFO(("Second sanity check passed.\n"));
  }

  r = (uint8_t *)is->phx.nic_mem + PHX_BBSYSCTL_R4;
  MX_INFO(("Third sanity check is writing and reading %p\n", r));
  PHX_REG_WRITE_4(r, 0xA1B2C3D4);
  val = PHX_REG_READ_4(r);
  if (val != 0xA1B2C3D4) {
    MX_INFO(("Third sanity check failed.  Got %08x.\n", val));
    phx_common_unmap_board(is);
    return ENODEV;
  } else {
    MX_INFO(("Third sanity check passed.\n"));
  }
#endif

  phx_read_bitstream_sha1(is->phx.nic_mem, sha1_val);
  MX_INFO(("SHA1 sum of Altera bitstream is: %08x%08x%08x%08x\n",
           sha1_val[0], sha1_val[1], sha1_val[2], sha1_val[3]));

  /* Read APU configuration from the NIC: channels, endpoints, ports. */
  phx_apu_config(is->phx.nic_mem, &is->phx.apu_config);

#if _VERBOSE
  /* const0 returns 0xbacecace */
  MX_INFO(("SNF const0=0x%x\n", phx_dbl_rx_const0(is->phx.nic_mem)));
#if 0
  /* FIXME - I dont think I can do this anymore */
  test_pio(is);
#endif
#endif

  return 0;
}

void
phx_common_unmap_board(mx_instance_state_t *is)
{
  if (is->phx.nic_mem)
    mx_unmap_io_space(is, is->board_span, (void*)is->phx.nic_mem);
}


void
phx_common_int_enable(mx_instance_state_t *is)
{
  /* MX_WARN(("%s %d: Phoenix: unimpl\n", __FUNCTION__, __LINE__)); */
  /* FIXME.  Use ICU registers. */
}

void
phx_common_setup_hw_caps(mx_instance_state_t *is)
{
  /* MX_WARN(("%s %d: Phoenix: unimpl\n", __FUNCTION__, __LINE__)); */
}

void
phx_common_int_disable(mx_instance_state_t *is)
{
  if (MX_IS_K35SNF(is)) {
    MX_WARN(("BUG: K35SNF should not call phx_common_int_disable\n"));
    return;
  }

  /* Disable every interrupt */
  if (is->phx.nic_mem) {
    int i, vectors;

    vectors = is->phx.pfi->ep_pool.num_function_endpoints + 1;
    for (i = 0; i < vectors; i++)
      phx_icu_disable_interrupt(is, i);
  }
}

void
phx_common_park_board(mx_instance_state_t *is)
{
  if (is->mac_ops.mac_disable)
    is->mac_ops.mac_disable(is);
  phx_vitesse_exit(is);                /* disable the 7111 transmitter if one available */
}

void
phx_common_set_mac_addr(mx_instance_state_t *is, uint8_t mac[6])
{
  is->mac_ops.mac_set_hwaddr(is, mac);   /* set the mac address */
}

/* Phoenix specific endpoint less ioctl.  Nothing for now. */
int
phx_endptless_ioctl(uint32_t cmd, const uaddr_t in, uint32_t privileged,
                    uint32_t is_kernel)
{
#ifdef MYRI_DBL_AB
  int status = 0;
  mx_instance_state_t *is;
  uaddr_t out = in;

  switch (cmd) {

  case MYRI_DBL_GET_TABLEX:
    {
      myri_dbl_ab_tablex_t x; 
      
      status = mx_copyin(in, &x, sizeof(x), is_kernel);
      if (status)
        break;

      is = mx_get_instance(x.board);
      if (!is) {
        status = ENODEV;
        goto abort_with_nothing;
      }

      status = phx_ab_tablex_get(is, &x);
 
      if (status == 0)
        status = mx_copyout(&x, out, sizeof(x), is_kernel);
      mx_release_instance(is);
      break;
    }
    default:
      status = ENOTTY;
  }

abort_with_nothing:
#else
  int status = ENOTTY;
#endif
  return status;
}

#ifdef MYRI_DBL_AB

/******************
 * A/B Statistics *
 ******************/
void __phx_ab_stats_clear(mx_instance_state_t *is, uint32_t tablex_index){

  uint8_t *r;
  int i;

  /* clear the autoreset count for this unit */
  r = (uint8_t *)is->phx.nic_mem + PHX_AB_UNIT_STATS_AUTORESET + (tablex_index * 4);
  PHX_REG_WRITE_4(r, 0);

  /* clear the prehistory for this unit */
  r = (uint8_t *)is->phx.nic_mem + PHX_AB_UNIT_STATS_PRE_HISTORY + (tablex_index * 8);
  PHX_REG_WRITE_8(r, 0);

  /* clear the passed/dropped counters */
  r = (uint8_t *)is->phx.nic_mem + PHX_AB_UNIT_STATS_BASE + (tablex_index * 64);
  for (i = 0; i < 8; i++){ 
    PHX_REG_WRITE_8(r, 0);
    r += 8;
  }

  MAL_WRITEBAR();
}

/* only call with mutex held */
static uint64_t
__phx_ab_autoreset(mx_instance_state_t *is, uint32_t tablex_index){

  uint8_t *r;
  uint64_t rc;

  r = (uint8_t *)is->phx.nic_mem + PHX_AB_UNIT_STATS_AUTORESET + (tablex_index * 4);
  rc = PHX_REG_READ_4(r);

  return rc;
}

/* only call with mutex held */
static uint64_t
__phx_ab_passed(mx_instance_state_t *is, uint32_t tablex_index, int stream){

  uint8_t *r;
  uint64_t rc;

  r = (uint8_t *)is->phx.nic_mem + PHX_AB_UNIT_STATS_BASE + (tablex_index * 64);
  r += (stream * 16);
  rc = PHX_REG_READ_8(r);

  return rc;
}

/* only call with mutex held */
static uint64_t
__phx_ab_removed(mx_instance_state_t *is, uint32_t tablex_index, int stream){

  uint8_t *r;
  uint64_t rc;

  r = (uint8_t *)is->phx.nic_mem + PHX_AB_UNIT_STATS_BASE + (tablex_index * 64);
  r += (stream * 16) + 8;
  rc = PHX_REG_READ_8(r);

  return rc;
}

/* only call with mutex held */
static uint64_t
__phx_ab_pre_history(mx_instance_state_t *is, uint32_t tablex_index){

  uint8_t *r;
  uint64_t rc;

  r = (uint8_t *)is->phx.nic_mem + PHX_AB_UNIT_STATS_PRE_HISTORY + (tablex_index * 8);
  rc = PHX_REG_READ_8(r);

  return rc;
}

void
__phx_ab_reset_seq(mx_instance_state_t *is, uint32_t tablex_index, uint64_t next_seq)
{
  uint8_t *r;

  r = (uint8_t *)is->phx.nic_mem + PHX_AB_UNIT_SEQ_NUMBER + (tablex_index * 8);
  PHX_REG_WRITE_8(r, next_seq);

}

void
__phx_ab_reset_seq_all(mx_instance_state_t *is, uint64_t next_seq)
{
  int i;

  for (i = 0; i < PHX_TABLEX_MAX; i++){
    __phx_ab_reset_seq(is, i, next_seq);
  }
}

void
__phx_ab_reset_seq_feed(mx_instance_state_t *is, uint32_t feed_unit, uint64_t next_seq)
{

  int32_t tablex_index;

  tablex_index = phx_ab_tablex_lookup(is, feed_unit);
  if (tablex_index >= 0)
	__phx_ab_reset_seq(is, tablex_index, next_seq);

}

int
phx_ab_seq_set(mx_endpt_state_t *es, myri_dbl_ab_seq_set_t ss)
{
  mx_mutex_enter(&mx_fpga_mutex);
  if (ss.feed_unit < 0){
    __phx_ab_reset_seq_all(es->is, ss.seq_num);
  } else {
    __phx_ab_reset_seq_feed(es->is, ss.feed_unit, ss.seq_num);
  }
  mx_mutex_exit(&mx_fpga_mutex);

  return 0;

};

static uint64_t
__phx_ab_seq_get(mx_instance_state_t *is, uint32_t tablex_index)
{

  uint8_t *r;
  uint64_t rc;

  r = (uint8_t *)is->phx.nic_mem + PHX_AB_UNIT_SEQ_NUMBER + (tablex_index * 8);
  rc = PHX_REG_READ_8(r);

  return rc;
}

int  
phx_ab_autoreset_init(mx_endpt_state_t *es, myri_dbl_ab_autoreset_t reset)
{
  uint8_t *r;
  mx_instance_state_t *is = es->is;
  uint8_t enable;
  uint16_t bound;

  enable = reset.enable & 1;
  bound = (uint16_t) reset.bound;

  mx_mutex_enter(&mx_fpga_mutex);

  r = (uint8_t *)is->phx.nic_mem + PHX_AB_BASE + PHX_AB_AUTORESET_ENABLE;
  PHX_REG_WRITE_4(r, enable);
  r = (uint8_t *)is->phx.nic_mem + PHX_AB_BASE +  PHX_AB_AUTORESET_BOUND;
  PHX_REG_WRITE_4(r, bound);

  mx_mutex_exit(&mx_fpga_mutex);

  return 0;

};

/**********************
 * TableX handlers    *
 **********************/

static int32_t
phx_ab_tablex_lookup(mx_instance_state_t *is, uint32_t feed_unit)
{
  struct phx_fpga_instance *fpga = is->phx.pfi;
  int32_t match = -1;
  int i;

  for (i = 0; i < PHX_TABLEX_MAX; i++){
    if (fpga->tablex_entries[i].feed_unit == feed_unit){
      match = i;
      break;
    }
  }

  return (match);
}

/*
 * Gets information about the A/B unit, some stored, some from FPGA registers.
 */ 
int  
phx_ab_tablex_get(mx_instance_state_t *is, myri_dbl_ab_tablex_t *get_tablex)
{
  struct phx_fpga_instance *fpga = is->phx.pfi;
  phx_tablex_t *kern_tablex;
  myri_dbl_ab_stream_t *stream;
  int32_t index = -1;
  int reset_seq = 0;
  int i;
  int rc = 0;

  mx_mutex_enter(&mx_fpga_mutex);

  /* two ways to request information, by feed unit, or by index */
#if 1
  if (get_tablex->feed_unit >= 0)
    index = phx_ab_tablex_lookup(is, get_tablex->feed_unit);
  else 
#endif
    index = get_tablex->tablex_index;

  /* this parameter is borrowed by the tool to command resetting the seq number */
  reset_seq = get_tablex->autoreset;

  /* if index is valid, get the stuff */
  if ((index >= 0) && (index < PHX_TABLEX_MAX)){
    get_tablex->tablex_index = index;

    kern_tablex = &fpga->tablex_entries[index];
    get_tablex->feed_unit = kern_tablex->feed_unit;
    get_tablex->pre_history = __phx_ab_pre_history(is, index);
    get_tablex->autoreset = __phx_ab_autoreset(is, index);
    get_tablex->seq = __phx_ab_seq_get(is, index); 
    for (i= 0; i < PHX_STREAM_MAX; i++){
      stream = &get_tablex->streams[i];
      stream->ip_addr = kern_tablex->ip_addr[i];
      stream->port = kern_tablex->port[i];
      stream->status = ((kern_tablex->filters[i] == NULL) ? 0 : 1);  /* 1 = active */
      stream->passed = __phx_ab_passed(is, index, i);
      stream->removed = __phx_ab_removed(is, index, i);
    }
   
    /* similar to myri counters, we report the counts then clear them */
    if (get_tablex->clear != 0)
      __phx_ab_stats_clear(is, index);

    /* this was saved above */
    if (reset_seq != 0)
      __phx_ab_reset_seq(is, index, 0);

  }
  else{
    MX_WARN(("Invalid tablex index or feed unit not found\n"));
    rc = 1;
  }

  mx_mutex_exit(&mx_fpga_mutex);

  return rc;
}

static int
phx_ab_tablex_add(mx_instance_state_t *is, myri_rx_filter_t *f,
                  struct myri_rx_filterq_entry *entry)
{
  struct phx_fpga_instance *fpga = is->phx.pfi;
  uint32_t feed_unit;
  int i, stream, match, avail;
  phx_tablex_t *tablex;

  mx_mutex_enter(&mx_fpga_mutex);

  feed_unit = f->ab.unit_id;    /* unit identifier from feed xml file or list */

  match = avail = -1;

  /* See if there is already an A/B arbiter setup on this feed unit */
  for (i = 0; i < PHX_TABLEX_MAX; i++){
    tablex = &fpga->tablex_entries[i];
    if (tablex->feed_unit == feed_unit){
      match = i;
      break;
    }
    if ((avail<0) && (tablex->feed_unit == -1)){
      avail = i;   /* first available, in case we don't find a match */
    }
  }

  /* if we didn't already have this feed unit, start a new one */
  if (match < 0)
    match = avail;

  if (match < 0)
    goto done;    /* no resource for tablex entry */

  /* add this filter entry */
  tablex = &fpga->tablex_entries[match];

  stream = -1;
  for (i= 0; i < PHX_STREAM_MAX; i++){
    if (tablex->filters[i] == NULL){
      if (tablex->filter_count == 0){
        /* This one will be reused, clear all old info */
        memset(tablex, 0, sizeof(*tablex));
        tablex->feed_unit = feed_unit;
      }
      tablex->filters[i] = entry;
      tablex->filter_count++;
      stream = i;
      tablex->ip_addr[stream] = f->ip_addr;
      tablex->port[stream] = f->port;
      break;
    }
  }

  /* make sure it fit */
  if (stream >= 0){
    f->ab.tablex_id = match;    /* Fill in the tablex number, it is written to the FPGA later */
    f->ab.stream_id = stream;   /* Fill in the stream for the filter, this gets written to the FPGA later */
  }
  else
    match = -1;   /* There was no room in the matching unit, all streams are already used */

#if _VERBOSE
  if (match >= 0)
      MX_INFO(("%s: feed unit %d tablex entry %d stream %c count %d\n", __FUNCTION__,
            feed_unit, match, ('A'+stream), tablex->filter_count ));
#endif

done:
  mx_mutex_exit(&mx_fpga_mutex);
  return match;

}

static void
phx_ab_tablex_remove(mx_instance_state_t *is, struct myri_rx_filterq_entry *entry)
{
  struct phx_fpga_instance *fpga = is->phx.pfi;
  int i, stream, match=-1;
  phx_tablex_t *tablex;

  mx_mutex_enter(&mx_fpga_mutex);

  /* if this entry is in an a/b tablex unit, remove it */
  for (i = 0; ((i < PHX_TABLEX_MAX) && (match < 0)); i++){
    tablex = &fpga->tablex_entries[i];
    for (stream= 0; stream < PHX_STREAM_MAX; stream++){
      if (tablex->filters[stream] == entry){
        tablex->filters[stream] = NULL;
        match = i;
        /* if the table is now empty, mark it free, but keep ip_address/port for stats */
        if (--tablex->filter_count == 0)
          tablex->feed_unit = -1;
        break;
      }
    }
  }

  mx_mutex_exit(&mx_fpga_mutex);
#if _VERBOSE
  if (match >= 0)
      MX_INFO(("%s: tablex entry %d stream %c count %d\n", __FUNCTION__, match, ('A'+stream), tablex->filter_count));
#endif
}

void phx_ab_tablex_init(mx_instance_state_t *is)
{
  struct phx_fpga_instance *fpga = is->phx.pfi;
  int i, stream;
  phx_tablex_t *tablex;

  mx_mutex_enter(&mx_fpga_mutex);

  /* clear all A/B tablex entries */
  for (i = 0; i < PHX_TABLEX_MAX; i++){
    tablex = &fpga->tablex_entries[i];
    for (stream= 0; stream < PHX_STREAM_MAX; stream++){
      tablex->filters[stream] = NULL;
      tablex->ip_addr[stream] = 0;
      tablex->port[stream]= 0;
    }
    tablex->feed_unit = -1;
    tablex->filter_count = 0;
  }

  mx_mutex_exit(&mx_fpga_mutex);
}
#endif /* MYRI_DBL_AB */

/**********************
 * Add/remove filters *
 **********************/

#if MAL_OS_WINDOWS
#pragma pack(push)
#pragma pack(1)
#endif
struct
#if MAL_OS_LINUX
__attribute__ ((__packed__))
#endif
phx_filter {
  uint32_t src_ip;
  uint32_t dst_ip;
  uint16_t src_port;
  uint16_t dst_port;
  uint16_t vlan;
  uint16_t tag;
  uint8_t proto;
  uint8_t add_rm;
  uint8_t endpoint_target;
  uint8_t flag;
  uint16_t ab_tablex_id;
  uint8_t ab_flags;
  uint8_t ab_seqoffset;
  uint8_t ab_seqsize;
  uint8_t ab_cntoffset;
  uint8_t ab_cntsize;
  uint8_t pad[1];
};
#if MAL_OS_WINDOWS
#pragma pack(pop)
#endif

/* in case of a "tx" filter, we need to set the filter to the src */

int
__phx_write_filter_regs(struct phx_nic_mem_global *nic, mx_sync_t *filter_sync,
			myri_rx_filter_t *f, uint32_t endpt, uint8_t flags,
                        phx_public_filter_t *save)
{

  uint8_t *r;
  uint32_t addslots;
  uint32_t slots;
  uint16_t freeslots, usedslots;
  struct phx_filter pf;
  int filter_op_tries = 0;

#ifdef MYRI_DBL_AB
  int do_ab = f->ab.enabled;
#endif

  r = ((uint8_t *)nic) + PHX_FLTR_BASE;

  memset(&pf, 0, sizeof(struct phx_filter));

  /* we are going to ingore the src fields for now as lanai didn't support
     them */

  if (f->txtstamp) {
    pf.src_ip = f->ip_addr;
    pf.src_port = f->port;
  } else {
    pf.src_ip = f->ip_src_addr; /* 0 */
    pf.src_port = 0;
    pf.dst_ip = f->ip_addr;
    pf.dst_port = f->port;
  }
  pf.vlan = f->vlan;
  pf.tag = f->tag;
  switch (f->tcp) {
  case 0:
    /* udp */
    pf.proto = 17;
    break;
  case 1:
    /* tcp */
    pf.proto = 6;
    break;
  case 2:
    /* arista timestamp key frame */
    pf.proto = 253;
    break;
  default:
    MX_WARN(("Invalid filter tcp field\n"));
    return 1;
  }
  if (f->remove) {
    pf.add_rm = 1;
  } else {
    pf.add_rm = 0;
  }
  pf.endpoint_target = endpt;

  pf.ab_flags = 0;  /* no AB by default */
#ifdef MYRI_DBL_AB
  if (do_ab) {
    /* fill in ab fields in pf from f */
    pf.ab_tablex_id = f->ab.tablex_id;
    pf.ab_flags = f->ab.enabled;
    pf.ab_seqoffset = f->ab.seq_ofst + 8;    /* a/b hardware starts counting at udp_header */
    pf.ab_seqsize = f->ab.seq_len;
    if (f->ab.seq_big_endian)
      pf.ab_flags |= AB_SEQ_BIG_ENDIAN;
    pf.ab_cntoffset = f->ab.cnt_ofst + 8;    /* a/b hardware starts counting at udp_header */ 
    pf.ab_cntsize = f->ab.cnt_len;
    if (f->ab.cnt_big_endian)
      pf.ab_flags |= AB_CNT_BIG_ENDIAN;
    pf.ab_flags |= (f->ab.stream_id << 3);
  }
#endif

  pf.flag = flags;
  if (f->txtstamp) {
    pf.flag |= (PHX_FLTR_FLAG_BIT_IGNORE_DST_IP |
		PHX_FLTR_FLAG_BIT_IGNORE_DST_PORT);
  } else {
    if (pf.src_ip != 0) {
      /* request for matching source was made ,
       * e.g IP_ADD_SOURCE_MEMBERSHIP */
      pf.flag |= PHX_FLTR_FLAG_BIT_IGNORE_SRC_PORT;
    } else {
      pf.flag |= (PHX_FLTR_FLAG_BIT_IGNORE_SRC_IP |
	       PHX_FLTR_FLAG_BIT_IGNORE_SRC_PORT);
    }
  }

  if (f->kernel)
    pf.flag |= PHX_FLTR_FLAG_BIT_KERNEL_DUP;

  if (pf.proto == 253)
    pf.flag |= PHX_FLTR_FLAG_BIT_IGNORE_DST_PORT;

 filter_op_try_again:
  mx_mutex_enter(filter_sync);

  addslots = PHX_REG_READ_4((r + PHX_FLTR_ADD_SLOTS));
  slots = PHX_REG_READ_4((r + PHX_FLTR_FREE_SLOTS));
  freeslots = (slots & 0xffff);
  usedslots = ((slots & 0xffff0000) >> 16);

#if _VERBOSE
  MX_INFO(("%s %s filter for ep %d: freeslots=%d usedslots=%d and addslots=%d\n",
           f->remove?"Remove":"Add", f->txtstamp?"TX":"RX", endpt, freeslots, usedslots, addslots));
  MX_INFO(("Filter is for ipaddr %d.%d.%d.%d port %d\n",
           (f->ip_addr & 0xff000000) >> 24,
           (f->ip_addr & 0x00ff0000) >> 16,
           (f->ip_addr & 0x0000ff00) >> 8,
           (f->ip_addr & 0x000000ff),
           f->port));
#ifdef MYRI_DBL_AB
  if (do_ab){
    MX_INFO(("A/B setup: tablex_id %d seq off %d size %d  cnt off %d size %d  flags %02x\n", pf.ab_tablex_id,
          pf.ab_seqoffset,  pf.ab_seqsize,  pf.ab_cntoffset,  pf.ab_cntsize , pf.ab_flags));
  }
#endif
#endif

  if ((addslots == 0) || (freeslots == 0 && ! f->remove)) {
    mx_mutex_exit(filter_sync);
    if (filter_op_tries < 100) {
      filter_op_tries++;
#if MAL_OS_LINUX
      udelay(1);
#else
      phx_win_usleep(1);
#endif
      goto filter_op_try_again;
    }
    MX_WARN(("Filter op %s failed to ep %d, addslots = %d, freeslots = %d attempts = %d\n",
	     f->remove?"Remove":"Add", endpt, addslots, freeslots, filter_op_tries));
    return 1;
  }

  if (save) {
    save->src_ip = pf.src_ip;
    save->dst_ip = pf.dst_ip;
    save->src_port = pf.src_port;
    save->dst_port = pf.dst_port;
    save->vlan = pf.vlan;
    save->tag = pf.tag;
    save->proto = pf.proto;
    save->valid = 1;
    save->endpoint_target = pf.endpoint_target;
    save->flag = pf.flag;
  }

  /* copy the data to the fpga */
  phx_pio_memcpy(r + PHX_FLTR_F_SRC_IP, (void *)&pf, sizeof(pf));
  mx_mutex_exit(filter_sync);

  /* see if anothing has changed since we added our filter */
  addslots = PHX_REG_READ_4((r + PHX_FLTR_ADD_SLOTS));
  slots = PHX_REG_READ_4((r + PHX_FLTR_FREE_SLOTS));
  freeslots = (slots & 0xffff);
  usedslots = ((slots & 0xffff0000) >> 16);
#if _VERBOSE
  MX_INFO(("After filter op: freeslots=%d usedslots=%d and addslots=%d\n",
    freeslots, usedslots, addslots));
#endif

  return 0;
}

int
phx_add_rx_filter(mx_endpt_state_t *es, myri_rx_filter_t *f)
{
  struct myri_rx_filterq_entry *entry;
  int status = 0;
#ifdef MYRI_DBL_AB
  int tablex_id;
  int do_ab = f->ab.enabled;
#endif

  entry = mx_kmalloc(sizeof(*entry),  MX_WAITOK|MX_MZERO);
  if (entry == NULL)
    return ENOMEM;

#ifdef MYRI_DBL_AB
  /* if this filter will use A/B tables to remove redundant packets, map it in */
  if (do_ab){
    tablex_id = phx_ab_tablex_add(es->is, f, entry);
    if (tablex_id < 0){
      mx_kfree(entry);
      return ENOSPC; /* no more tablex entries, they can add it without A/B enabled */
    }
  }
#endif

  mx_mutex_enter(&es->sync);
  status = __phx_write_filter_regs(es->is->phx.nic_mem, &es->is->phx.filter_sync,
				   f, es->endpt, 0, &entry->phx_filt);

  if (status)
    goto abort;

  /* if the f.txtstamp was set, then what we added above was the tx filter only
     and we need to also now add the rx filter */
  if (f->txtstamp) {
    f->txtstamp = 0;
    status = __phx_write_filter_regs(es->is->phx.nic_mem, &es->is->phx.filter_sync,
				     f, es->endpt, 0, &entry->phx_filt2);
    f->txtstamp = 1; /* restore value */
    if (status) {
      /* failed rx filter add, remove the previous filter */
      f->remove = 1;
      __phx_write_filter_regs(es->is->phx.nic_mem, &es->is->phx.filter_sync,
			      f, es->endpt, 0, NULL);
      f->remove = 0; /* restore value before aborting */
      goto abort;
    }
  }
  bcopy(f, &entry->filt, sizeof(*f));
  TAILQ_INSERT_TAIL(&es->rx_filterq, entry, entries);

 abort:
  mx_mutex_exit(&es->sync);
  if (status != 0) {
#ifdef MYRI_DBL_AB
    if (do_ab)
      phx_ab_tablex_remove(es->is, entry);
#endif
    mx_kfree(entry);
  }
  return -status;
}

int
phx_remove_rx_filter(mx_endpt_state_t *es, myri_rx_filter_t *f,
		     struct myri_rx_filterq_entry *entry)
{
  int status = 0;

  mx_mutex_enter(&es->sync);
  if (entry == NULL) {
    TAILQ_FOREACH(entry, &es->rx_filterq, entries) {
      if (entry->filt.ip_addr == f->ip_addr &&
	  entry->filt.port == f->port &&
	  entry->filt.vlan == f->vlan &&
	  entry->filt.tcp == f->tcp &&
	  entry->filt.tag == f->tag) {
	break;
      }
    }
  }
  if (entry == NULL) {
    mx_mutex_exit(&es->sync);
    return EINVAL;
  }

  TAILQ_REMOVE(&es->rx_filterq, entry, entries);
  entry->filt.remove=1;
  status = __phx_write_filter_regs(es->is->phx.nic_mem, &es->is->phx.filter_sync,
				   &entry->filt, es->endpt, 0, NULL);

  if (entry->filt.txtstamp) {
    /* if the txtstamp is set, then we just removed the tx filter above, but now we need
       to remove the rx one as well */
    entry->filt.txtstamp = 0;
    status = __phx_write_filter_regs(es->is->phx.nic_mem, &es->is->phx.filter_sync,
				     &entry->filt, es->endpt, 0, NULL);
  }

  mx_mutex_exit(&es->sync);
#ifdef MYRI_DBL_AB
  phx_ab_tablex_remove(es->is, entry);
#endif
  mx_kfree(entry);
  return (-status);
}

void
phx_get_filter_state(mx_instance_state_t *is, phx_public_filter_state_t *filters)
{
  uint32_t i;
  mx_endpt_state_t *es;
  struct myri_rx_filterq_entry *entry;
  phx_public_filter_t *f;
  uint8_t *r;

  filters->filter_count = 0;
  filters->kernel_endpoint = 0;
  filters->discard_multicast_enabled = 0;

  return;   /* sniffer has no filter unit */

  mx_mutex_enter(&is->sync);
  for (i = 0; i < is->es_count; i++) {
    es = is->es[i];
    if (es) {
      mx_mutex_enter(&es->sync);

      TAILQ_FOREACH(entry, &es->rx_filterq, entries) {
        if (entry->phx_filt.valid) {
          if (filters->filter_count >= PHX_RX_FILTER_CNT)
            break;
          f = &filters->filters[filters->filter_count++];
          *f = entry->phx_filt;
        }

        if (entry->phx_filt2.valid) {
          if (filters->filter_count >= PHX_RX_FILTER_CNT)
            break;
          f = &filters->filters[filters->filter_count++];
          *f = entry->phx_filt2;
        }
      }

      mx_mutex_exit(&es->sync);
    }
  }
  mx_mutex_exit(&is->sync);

  r = ((uint8_t*)is->phx.nic_mem) + PHX_FLTR_BASE;
  filters->kernel_endpoint = PHX_REG_READ_4((r + PHX_FLTR_KERNEL_ENDPOINT));
  filters->discard_multicast_enabled = is->discard_multicast_enabled;
}

void
phx_get_mac_filters(mx_instance_state_t *is, phx_get_mac_filters_t *filters)
{
  uint32_t i;
  phx_mac_filter_t *f;
  uint64_t *mac;

  filters->count = is->phx.num_l2_filters;

  mx_mutex_enter(&is->sync);
  for (i = 0; i < is->phx.num_l2_filters; i++) {
    mac = &filters->mac[i];
    f = &is->phx.l2_filters[i];
    *mac = ((uint64_t)f->high << 32) + f->low;
  }
  mx_mutex_exit(&is->sync);
}

/************************************
 * Read/write various NIC registers *
 ************************************/

static void
phx_dbl_rx_ring_size(struct phx_nic_mem_global *nic, int endpoint,
                     uint32_t data_pages, uint32_t desc_pages)
{
  uint8_t *r = (uint8_t*)nic + PHX_DBL_RX_BASE
      + (PHX_DBL_RX_ENDPOINT_SIZE * endpoint);

  PHX_REG_WRITE_4((r + PHX_DBL_RX_DATA_RING_SIZE), data_pages);
  PHX_REG_WRITE_4((r + PHX_DBL_RX_DESC_RING_SIZE), desc_pages);
}

void
phx_dbl_rx_global_start(mx_instance_state_t *is)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + PHX_DBL_RX_BASE + PHX_DBL_RX_COMMAND;
  if (MX_IS_CHILD(is)) {
    /* FIXME: Implement this */
    return;
  }
  PHX_REG_WRITE_4(r, 0x1);
}

void
phx_dbl_rx_global_stop(mx_instance_state_t *is)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + PHX_DBL_RX_BASE + PHX_DBL_RX_COMMAND;
  if (MX_IS_CHILD(is)) {
    /* FIXME: Implement this */
    return;
  }
  PHX_REG_WRITE_4(r, 0x2);
}

static void
phx_dbl_rx_endpoint_start(struct phx_nic_mem_global *nic, int endpoint)
{
  uint8_t *r = (uint8_t*)nic + PHX_DBL_RX_BASE
      + PHX_DBL_RX_ENDPOINT_COMMAND
      + (PHX_DBL_RX_ENDPOINT_SIZE * endpoint);
#if _VERBOSE
  MX_INFO(("%s: endpoint=%d\n", __FUNCTION__, endpoint));
#endif
  PHX_REG_WRITE_4(r, 0x1);
}

/*
 * This operation will read the producer and consumer data/desc. position
 * and return the register as a 64bit value.  This is faster than 2 reads.
 */
static void
_k35snf_get_page_position(mx_instance_state_t *is, uint32_t position, uint64_t *pcp)
{
  uint8_t *r;

  /* get base address to the Rx Lite ports */
  r = (uint8_t*)is->phx.nic_mem + K35SNF_RX_BASE;
  r += (K35SNF_RX_SIZE * is->port_id);

  /*
   * This will read the consumer position into the lower word (32b)
   * and the producer position in the upper word (32b).
   */
  *pcp = PHX_REG_READ_8((r + position));
}

/*
 * This allows the card to send packets that are buffered in a
 * FIFO beyond the SNF Rx core.
 * The buffer provides elasticity from PCIE stalls.
 * Just give back pages until the producer is expended, to drain
 * these packets.
 */
static void
_k35snf_drain_rx_buffer(mx_instance_state_t *is, int endpoint)
{
  struct dword_to_word_s {
    int32_t l32;
    int32_t h32;
  } prod_cons_position;
  int loop_count;
  int32_t credits;
  uint32_t desc_page_credits, data_page_credits;
  uint32_t desc_prod_last, data_prod_last;
  int done;

  loop_count = 1000;
  desc_prod_last = data_prod_last = 0;

  do {
    done = 1;

    /* get descriptor buffer position */
    _k35snf_get_page_position(is, K35SNF_RX_DESC_CONSR_POSITION, (uint64_t *)&prod_cons_position);
    if((credits = (prod_cons_position.h32 - prod_cons_position.l32)) < 0 )
      credits += is->phx.rx_desc_pages;   /* handle wrap */
    desc_page_credits = (credits > 32 ? 32 : credits);
    if (desc_prod_last != prod_cons_position.h32){
      desc_prod_last = prod_cons_position.h32;
      done = 0;
    }

    /* get data buffer position */
    _k35snf_get_page_position(is, K35SNF_RX_DATA_CONSR_POSITION, (uint64_t *)&prod_cons_position);
    if((credits = (prod_cons_position.h32 - prod_cons_position.l32)) < 0 )
      credits += is->phx.rx_data_pages;   /* handle wrap */
    data_page_credits = (credits > 32 ? 32 : credits);
    if (data_prod_last != prod_cons_position.h32){
      data_prod_last = prod_cons_position.h32;
      done = 0;
    }

    /* give the card some more room to send packets up */
    phx_rx_desc_cons_advance(is, endpoint, desc_page_credits);
    phx_rx_data_cons_advance(is, endpoint, data_page_credits);

    udelay(10);
#if 0
    MX_INFO(("%s: desc position %d data position %d\n", __FUNCTION__, desc_prod_last, data_prod_last));
#endif
    loop_count--;

  } while (!done && loop_count);
}

void
phx_stop_rx_endpoint(mx_instance_state_t *is, int endpoint)
{
  int max_loops;
  int loop_count;
  uint32_t val = 0;
  uint8_t *r = (uint8_t *) is->phx.nic_mem;
  uint8_t *reg_cmd, *reg_flushed;

#if _VERBOSE
  MX_INFO(("%s: endpoint=%d\n", __FUNCTION__, endpoint));
#endif

  if (MX_IS_K35SNF(is)) {
    /*
     * We need to increase the delay by a factor x10 to account for
     * a 1G interface.  We have seen flush timeouts in our testing
     * of 1G.  We could check if the interface is 10G or 1G and 
     * delay appropriately but instead we will take the simplier 
     * approach and delay x10 more, for a total of 10msec.  The
     * is counter based and each pass sleeps 1us.
     */
    max_loops = 1000 * 10; /* Wait for up to 10 msec */

    /* disable the RX MAC for this port */
    k35snf_xmac_rx_disable(is, is->port_id);

    /*
     * If port merging with next_ring FPGA acceleration then we have a potentail
     * issue of leaving bot MACs in the port pair enabled while stopping
     * the ports.  Since our stop/start EP handling is on a port basis
     * we need to combine some of the operations before proceeding and deal
     * deal in a combined stop operation -- basically we need to disable 
     * both ports in the merged pair then disable there seqeuncer and if we
     * don't we could get into an Rx stuck issue with the opposite port
     * receiving while we try to drain the disabled mac we could get pkts on
     * on the other port forwarded into this port even after the drain is 
     * though to have finished and then do a stop and this could still be 
     * getting pkts forwarded from the other port.  So to prevent this we are
     * nowing going to disable the Rx MAC of the other port in the pair not drain
     * it as that will occur when the other port is stopped -- this is OK and
     * allows us to not have to restructure how the drain operation works.
     */
    if (is->phx.port_merge == SNF_PORT_MERGE_MODE_NEXT_RING) {
      k35snf_xmac_rx_disable(is, (is->port_id ^ 1));
    }

    /* allow packets not already pushed to drain naturally */
    _k35snf_drain_rx_buffer(is, endpoint);

    /*
     * With the new next_ring port merge if it is enabled it is enabled
     * on pair of ports: 0,1 and/or 2,3.  So when we go to stop a port
     * we have to disable the next_ring sequencer on the pair before 
     * issuing the stop on a interface or we can get into a stuck
     * condition as the other port than the one serviced in this 
     * invocation can be writing to the FIFO of this interface as its
     * just a ping pong.  With this change this disable will be occur
     * twice per pair -- this is OK.
     */
    if (PHX_HAS_FEATURE(is, PHX_FEATURE_FPGA_MERGE_NEXT_RING)) {
      uint32_t port_id;
      uint32_t ports;
      port_id = (is->port_id & 2);
      for (ports=0; ports<2; ports++, port_id++) {
        r = (uint8_t *) is->phx.nic_mem;
        r += K35SNF_RSS_BASE + (K35SNF_RSS_RESET + (K35SNF_RSS_SIZE * port_id));
        PHX_REG_WRITE_4(r, K35SNF_MERGE_NEXT_RING_DISABLE_SEQUENCER);
        udelay(1);
      }
    }

    /*
     * Now, shut down and flush the endpoint.
     */
    r = (uint8_t *) is->phx.nic_mem;
    r += K35SNF_RX_BASE + (K35SNF_RX_SIZE * is->port_id);
    reg_cmd = r + K35SNF_RX_COMMAND;
    reg_flushed = r + K35SNF_RX_STOP_FLUSHED;
  } else {
    r += PHX_DBL_RX_BASE + (PHX_DBL_RX_ENDPOINT_SIZE * endpoint);
    reg_cmd = r + PHX_DBL_RX_ENDPOINT_COMMAND;
    reg_flushed = r + PHX_DBL_RX_STOP_FLUSHED;
    max_loops = 1000; /* Wait for up to 1 msec */
  }

  /* STOP */
  PHX_REG_WRITE_4(reg_cmd, 0x2);

  /* wait for the RX stopFlushed register to be 1 again */
  loop_count = max_loops;
  do {
    val = PHX_REG_READ_4(reg_flushed);
    if (val != 0) {
      break;
    }
    loop_count--;
#if MAL_OS_LINUX
    udelay(1);
#else
    phx_win_usleep(1);
#endif
  } while (loop_count > 0);

  if (PHX_HAS_FEATURE(is, PHX_FEATURE_FPGA_MERGE_NEXT_RING)) {
   /* Reset the RSS path and the sequencer.  This has to happen after
     * the endpoint is stopped and flushed.
     *
     * Description of bits in the register which controls merge next ring
     * generator.
     *
     * Bit 0: active high. Clears statistics counters.
     *                     Sets all RSS counters to 0.
     * Bit 1: active high. RSS port reset.
     *                     Resets entire port path.
     * Bit 2: active high. Disables next ring seqeuncer. Used for
     *                     debugging purposes only.  Should not be used
     *                     during normal operation.
     *
     * Reset procedure:
     * Bit 0: set it then clear it. (We don't do this. It is not needed.)
     * Bit 1: set it then clear it. (We do this. It will reset the RSS core.)
     * Bit 2: set it then clear it. (We do this when not port merging.)
     */
    r = (uint8_t *) is->phx.nic_mem;
    r += K35SNF_RSS_BASE + (K35SNF_RSS_RESET + (K35SNF_RSS_SIZE * is->port_id));

    PHX_REG_WRITE_4(r, (K35SNF_MERGE_NEXT_RING_RESET_RSS_BLOCK | K35SNF_MERGE_NEXT_RING_DISABLE_SEQUENCER));
    udelay(1);
    PHX_REG_WRITE_4(r, K35SNF_MERGE_NEXT_RING_DISABLE_SEQUENCER);
    udelay(1);
  }

  if (val == 1) {
#if _VERBOSE
    MX_INFO(("RX endpoint %d flushed after %d us.\n", endpoint, max_loops - loop_count));
#endif
  } else if  (val == 0xffffffff) {
    MX_WARN(("The board appears to be in a bad state\n"));
  } else {
    MX_WARN(("RX endpoint is not flushed (max_loops = %d). flush_register=0x%x\n", max_loops, val));
  }
}

void
phx_reset_rx_endpoint(mx_instance_state_t *is, int endpoint)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem;
#if _VERBOSE
  MX_INFO(("%s: endpoint=%d\n", __FUNCTION__, endpoint));
#endif
  if (MX_IS_K35SNF(is)) {
    r += K35SNF_RX_BASE + K35SNF_RX_COMMAND + (K35SNF_RX_SIZE * is->port_id);
  } else {
    r += PHX_DBL_RX_BASE + PHX_DBL_RX_ENDPOINT_COMMAND +
        (PHX_DBL_RX_ENDPOINT_SIZE * endpoint);
  }
  PHX_REG_WRITE_4(r, 0x3);
}

void
phx_rx_data_cons_advance(mx_instance_state_t *is, int endpoint, uint32_t pages)
{
  /* MX_INFO(("%s: endpoint=%d pages=%u\n", __FUNCTION__, endpoint, pages)); */
  PHX_REG_WRITE_4(phx_rx_data_consr_advance(is, endpoint), pages);
}

void
phx_rx_desc_cons_advance(mx_instance_state_t *is, int endpoint, uint32_t pages)
{
  /* MX_INFO(("%s: endpoint=%d pages=%u\n", __FUNCTION__, endpoint, pages)); */
  PHX_REG_WRITE_4(phx_rx_desc_consr_advance(is, endpoint), pages);
}

void
phx_dbl_rx_wake_request(struct phx_nic_mem_global *nic, int endpoint, uint32_t host_count)
{
  uint8_t *r = (uint8_t*)nic + PHX_DBL_RX_BASE
      + (PHX_DBL_RX_ENDPOINT_SIZE * endpoint);
#if 0
  uint32_t nic_count;

  nic_count = PHX_REG_READ_4((r + PHX_DBL_RX_DESC_COUNT));
  MX_INFO(("%s: endpoint=%d host_count=%u nic_count=%u\n", __FUNCTION__,
           endpoint, host_count, nic_count));
#endif
  PHX_REG_WRITE_4((r + PHX_DBL_RX_DESC_WAKE_REQUEST), host_count);
}

uint32_t
phx_dbl_rx_const0(struct phx_nic_mem_global *nic)
{
  uint8_t *r = (uint8_t*)nic + PHX_DBL_RX_BASE + PHX_DBL_RX_CONST0;
  return PHX_REG_READ_4(r);
}

void
phx_dbl_rx_packet_count_limit(mx_instance_state_t *is, uint32_t pkt)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + PHX_DBL_RX_BASE + PHX_DBL_RX_PACKET_COUNT_LIMIT;
  if (MX_IS_CHILD(is)) {
    /* FIXME: Implement this */
    return;
  }
  PHX_REG_WRITE_4(r, pkt);
  /* report value if not default */
  if (pkt != PHX_RX_COALESCE_PKTS)
    MX_INFO(("%s: pkt=0x%x\n", __FUNCTION__, pkt));
}

void
phx_dbl_rx_time_limit(mx_instance_state_t *is, uint32_t nanosecs)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + PHX_DBL_RX_BASE + PHX_DBL_RX_TIME_LIMIT;
  if (MX_IS_CHILD(is)) {
    /* FIXME: Implement this */
    return;
  }
#if _VERBOSE
  MX_INFO(("%s: time(ns)=0x%x\n", __FUNCTION__, nanosecs));
#endif
  PHX_REG_WRITE_4(r, nanosecs);
}

void
phx_dbl_rx_length_limit(mx_instance_state_t *is, uint32_t bytes)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + PHX_DBL_RX_BASE + PHX_DBL_RX_LENGTH_LIMIT;
  if (MX_IS_CHILD(is)) {
    /* FIXME: Implement this */
    return;
  }
#if _VERBOSE
  MX_INFO(("%s: bytes=0x%x\n", __FUNCTION__, bytes));
#endif
  PHX_REG_WRITE_4(r, bytes);
}

void
phx_dbl_rx_print_debug(struct phx_nic_mem_global *nic)
{
  uint8_t *r = (uint8_t*)nic + PHX_DBL_RX_BASE;
  uint32_t cycle_count, max_length_limit;
  uint32_t byte_count, packet_count;
  uint32_t ff_byte_count, ff_packet_count, fd_byte_count, fd_packet_count;
  uint32_t packet_overflow, desc_overflow;
  uint32_t data_consr_pos, data_prodw_pos, desc_consr_pos, desc_prodw_pos;
  int ep;

  cycle_count = PHX_REG_READ_4((r + PHX_DBL_RX_CYCLE_COUNT));
  max_length_limit = PHX_REG_READ_4((r + PHX_DBL_RX_MAX_LENGTH_LIMIT));

  byte_count = PHX_REG_READ_4((r + PHX_DBL_RX_BYTE_COUNT));
  packet_count = PHX_REG_READ_4((r + PHX_DBL_RX_PACKET_COUNT));

  ff_byte_count = PHX_REG_READ_4((r + PHX_DBL_RX_FF_BYTE_COUNT));
  ff_packet_count = PHX_REG_READ_4((r + PHX_DBL_RX_FF_PACKET_COUNT));
  fd_byte_count = PHX_REG_READ_4((r + PHX_DBL_RX_FD_BYTE_COUNT));
  fd_packet_count = PHX_REG_READ_4((r + PHX_DBL_RX_FD_PACKET_COUNT));
  packet_overflow = PHX_REG_READ_4((r + PHX_DBL_RX_PACKET_OVERFLOW));
  desc_overflow = PHX_REG_READ_4((r + PHX_DBL_RX_DESC_OVERFLOW));

  MX_INFO(("DBL RX debug counters:\n"
           "cycle_count=%u max_length_limit=%u\n"
           "byte_count=%u packet_count=%u\n"
           "ff_byte_count=%u ff_packet_count=%u\n"
           "fd_byte_count=%u fd_packet_count=%u\n"
           "packet_overflow=%u desc_overflow=%u\n",
           cycle_count, max_length_limit,
           byte_count, packet_count,
           ff_byte_count, ff_packet_count,
           fd_byte_count, fd_packet_count,
           packet_overflow, desc_overflow));

  /* FIXME.  Print counters for the first 4 endpoints. */
  for (ep = 0; ep < 4; ep++) {
    r = (uint8_t*)nic + PHX_DBL_RX_BASE + (PHX_DBL_RX_ENDPOINT_SIZE * ep);

    data_consr_pos = PHX_REG_READ_4((r + PHX_DBL_RX_DATA_CONSR_POSITION_N));
    data_prodw_pos = PHX_REG_READ_4((r + PHX_DBL_RX_DATA_PRODW_POSITION_N));
    desc_consr_pos = PHX_REG_READ_4((r + PHX_DBL_RX_DESC_CONSR_POSITION_N));
    desc_prodw_pos = PHX_REG_READ_4((r + PHX_DBL_RX_DESC_PRODW_POSITION_N));
    byte_count = PHX_REG_READ_4((r + PHX_DBL_RX_BYTE_COUNT_N));
    packet_count = PHX_REG_READ_4((r + PHX_DBL_RX_PACKET_COUNT_N));

    MX_INFO(("Endpoint %d\n"
             "data_consr_pos=%u data_prodw_pos=%u\n"
             "desc_consr_pos=%u desc_prodw_pos=%u\n"
             "byte_count=%u packet_count=%u\n",
             ep, data_consr_pos, data_prodw_pos,
             desc_consr_pos, desc_prodw_pos,
             byte_count, packet_count));
  }
}

void
phx_dbl_ab_init(struct phx_nic_mem_global *nic)
{
  uint8_t *r;
  uint32_t val;

  /* the AB core will be turned on later if there is an appropriate configuration file */
  r = (uint8_t *)nic + PHX_AB_BASE;
  val = PHX_REG_READ_4(r);
  if ((val & 0xffff0000) != 0x41420000) { 
    MX_INFO(("A/B core not present.  Got %08x.\n", val));
  } else {
    /* Reset the A/B module */
    r = (uint8_t *)nic + PHX_AB_BASE + PHX_AB_GLOBAL_RESET;
    PHX_REG_WRITE_4(r, 0xffffffff);
 
#ifdef MYRI_DBL_AB
    MX_INFO(("Enable A/B globally\n"));
    r = (uint8_t *)nic + PHX_AB_BASE + PHX_AB_GLOBAL_BYPASS;
    PHX_REG_WRITE_4(r, 0);
#else
    MX_INFO(("Disabling A/B globally\n"));
    r = (uint8_t *)nic + PHX_AB_BASE + PHX_AB_GLOBAL_BYPASS;
    PHX_REG_WRITE_4(r, 1);
#endif
  }
}

void
phx_dbl_ab_local_init(struct phx_nic_mem_global *nic)
{
  uint8_t *r;
  uint32_t val;

  /* turn on or off the local AB bypass switch */
  r = (uint8_t *)nic + PHX_AB_BASE;
  val = PHX_REG_READ_4(r);
  if ((val & 0xffff0000) == 0x41420000) {
#ifdef MYRI_DBL_AB
    MX_INFO(("Enable A/B locally\n"));
    r = (uint8_t *)nic + PHX_AB_BASE + PHX_AB_BYPASS;
    PHX_REG_WRITE_4(r, 0);
#else
    MX_INFO(("Disabling A/B locally\n"));
    r = (uint8_t *)nic + PHX_AB_BASE + PHX_AB_BYPASS;
    PHX_REG_WRITE_4(r, 1);
#endif
  }
}

void
phx_dbl_filter_init(mx_instance_state_t *is, uint32_t kernel_endpoint)
{
  uint8_t *r;
  uint32_t id_ver, k_endpoint, filter_flags;

  if (MX_IS_CHILD(is)) {
    /* FIXME: Implement this */
    return;
  }

  r = (uint8_t*)is->phx.nic_mem + PHX_FLTR_BASE;
  /* Remove all filters, reset all counts (takes 16K * 6.8ns) */
  /* NOTE: the PHX_FLTR_SOFT_RESET is now a global hammer 
   so use PORT reset here. */
  PHX_REG_WRITE_4((r + PHX_FLTR_PORT_RESET), 0x3);

#if MAL_OS_LINUX
  udelay(100);
#else
  phx_win_usleep(100);
#endif

  /* probably not needed since it is the default */
  PHX_REG_WRITE_4((r + PHX_FLTR_FILTER_FLAGS), 0x0);

  PHX_REG_WRITE_4((r + PHX_FLTR_KERNEL_ENDPOINT), kernel_endpoint);

  id_ver = PHX_REG_READ_4((r + PHX_FLTR_ID_VERSION));
  filter_flags = PHX_REG_READ_4((r + PHX_FLTR_FILTER_FLAGS));
  k_endpoint = PHX_REG_READ_4((r + PHX_FLTR_KERNEL_ENDPOINT));
#if _VERBOSE
  MX_INFO(("id_vers=0x%x kernel_endpoint=%u filter_flags=0x%x\n",
           id_ver, k_endpoint, filter_flags));
#endif
}

void
phx_dbl_filter_set_kernel_endpoint(mx_instance_state_t *is, uint32_t kernel_endpoint)
{
  uint8_t *r;

#if _VERBOSE
  MX_INFO(("Setting the kernel endpoint. pci_func=%d endpoint=%d\n",
           is->phx.pci_info.func, kernel_endpoint));
#endif
  r = (uint8_t*)(is->phx.nic_mem) + PHX_FLTR_BASE;
  PHX_REG_WRITE_4((r + PHX_FLTR_KERNEL_ENDPOINT), kernel_endpoint);
}

void
phx_dbl_filter_discard_multicast(mx_instance_state_t *is, int discard)
{
  struct phx_nic_mem_global *nic = is->phx.nic_mem;
  uint8_t *r;

  /* check that feature set supports this */
  if (!PHX_HAS_FEATURE(is, PHX_FEATURE_MULTICAST_DISCARD))
       return;

  /* check if driver enable allows discarding */
  if ((1 << is->id) & myri_enable_ipv4_multicast_discard_bitmask)
  {
    r = (uint8_t*)nic + PHX_FLTR_BASE;

    PHX_REG_WRITE_4((r + PHX_FLTR_CLASS_D_DISCARD), discard ? 1 : 0);
    is->discard_multicast_enabled = discard;
#if _VERBOSE
    MX_INFO(("instance %d %s multicast discard.\n", is->id,
	   discard? "enabled":"disabled"));
#endif
  }
}

void
phx_dbl_filter_print_debug(struct phx_nic_mem_global *nic)
{
  uint8_t *r = (uint8_t*)nic + PHX_FLTR_BASE;
  uint32_t pkts_received, bytes_received;
  uint32_t pkts_sent, bytes_sent;
  uint32_t pkts_dropped, pkts_aborted;
  uint32_t pkts_filtered, pkts_unfiltered;
  uint32_t pkts_skipped_filter, pkts_drop_filter;

  pkts_received = PHX_REG_READ_4((r + PHX_FLTR_PKTS_RECEIVED));
  bytes_received = PHX_REG_READ_4((r + PHX_FLTR_BYTES_RECEIVED));
  pkts_sent = PHX_REG_READ_4((r + PHX_FLTR_PKTS_SENT));
  bytes_sent = PHX_REG_READ_4((r + PHX_FLTR_BYTES_SENT));
  pkts_dropped = PHX_REG_READ_4((r + PHX_FLTR_PKTS_DROPPED));
  pkts_aborted = PHX_REG_READ_4((r + PHX_FLTR_PKTS_ABORTED));
  pkts_filtered = PHX_REG_READ_4((r + PHX_FLTR_PKTS_FILTERED));
  pkts_unfiltered = PHX_REG_READ_4((r + PHX_FLTR_PKTS_UNFILTERED));
  pkts_skipped_filter = PHX_REG_READ_4((r + PHX_FLTR_PKTS_SKIPPED_FILTER));
  pkts_drop_filter = PHX_REG_READ_4((r + PHX_FLTR_PKTS_DROP_FILTER));

  MX_INFO(("Filter debug counters:\n"
           "pkts_received=%u bytes_received=%u pkts_sent=%u bytes_sent=%u\n"
           "pkts_dropped=%u pkts_aborted=%u pkts_filtered=%u pkts_unfiltered=%u\n"
           "pkts_skipped_filter=%u pkts_drop_filter=%u\n",
           pkts_received, bytes_received, pkts_sent, bytes_sent,
           pkts_dropped, pkts_aborted, pkts_filtered, pkts_unfiltered,
           pkts_skipped_filter, pkts_drop_filter));
}

void
phx_apu_config(struct phx_nic_mem_global *nic, struct phx_apu_config *conf)
{
  uint8_t *r = (uint8_t*)nic + PHX_APU_BASE;
  uint32_t n;
  uint32_t apu_cfg_reg;

  conf->id_version = PHX_REG_READ_4((r + PHX_APU_ID_VERSION));

  apu_cfg_reg = PHX_REG_READ_4((r + PHX_APU_NUM_CHANNEL));
  conf->num_channel = (apu_cfg_reg & 0xffff) ;
  conf->endpoints = ((apu_cfg_reg & 0xff0000) >> 16);
  conf->ports = ((apu_cfg_reg & 0xff000000) >> 24);

  /* Round up to a power of 2*/
  n = 1;
  while (n < conf->endpoints)
    n = n << 1;
  conf->channels_per_port = 2 * n;

#if _VERBOSE
  MX_INFO(("APU configuration: id_version=0x%x num_channel=%u endpoints=%u"
           " ports=%u channels_per_port=%u\n",
           conf->id_version, conf->num_channel, conf->endpoints, conf->ports,
           conf->channels_per_port));
#endif
}

int
phx_apu_channel_id(mx_instance_state_t *is, int data, int endpoint, int tx)
{
  int id;
  id = data ? 0 : 1;
  if (MX_IS_K35SNF(is)) {
    /* 4 channels per port. 'endpoint' is not used. */
    if (tx)
       id += 2;
  } else
    id += 2 * endpoint;
  return id;
}

void
phx_apu_channel_reset(mx_instance_state_t *is, int channel)
{
  /*
   * Should add a timeout and return an error in production version.
   */

  uint8_t *r;
  uint32_t val;
  int loop_count;

  if (MX_IS_K35SNF(is)) {
    r = (uint8_t*)is->phx.nic_mem + K35SNF_RX_APU_BASE;
    r += K35SNF_RX_APU_CHANNEL_RESET +
        (is->port_id * K35SNF_RX_APU_SIZE) +
        (channel * K35SNF_RX_APU_CHANNEL_SIZE);
  } else {
    r = (uint8_t*)is->phx.nic_mem + PHX_APU_BASE +
        PHX_APU_CHANNEL_RESET +
        (channel * PHX_APU_CHANNEL_SIZE);
  }

#if _VERBOSE
  MX_INFO(("%s: channel=%d\n", __FUNCTION__, channel));
#endif
  PHX_REG_WRITE_4(r, 1);

  loop_count = 1000; /* Wait for up to 1msec */
  do {
#if MAL_OS_LINUX
    udelay(1);
#else
    phx_win_usleep(1);
#endif
    loop_count--;
    val = PHX_REG_READ_4(r);
    if (val == 0xffffffff) /* The board in a bad state */
      break;
  } while (loop_count > 0 && val != 0);
  if (val == 0) {
#if _VERBOSE
    MX_INFO(("Reset channel okay\n"));
#endif
  } else {
    MX_WARN(("Reset channel %d timed out! Reset register=0x%x\n", channel, val));
  }
}

void
phx_apu_channel_start(mx_instance_state_t *is, int channel,
                      uint32_t low, uint32_t high)
{
  uint8_t *r, *reg_low, *reg_high;

  if (MX_IS_K35SNF(is)) {
    /* Each port has its own APU */
    r = (uint8_t*)is->phx.nic_mem + K35SNF_RX_APU_BASE;
    r += (is->port_id * K35SNF_RX_APU_SIZE) +
        (channel * K35SNF_RX_APU_CHANNEL_SIZE);
    reg_low = r + K35SNF_RX_APU_ALPHA_NEXT_LS;
    reg_high = r + K35SNF_RX_APU_ALPHA_NEXT_MS;
  } else {
    r = (uint8_t*)is->phx.nic_mem + PHX_APU_BASE + (channel * PHX_APU_CHANNEL_SIZE);
    reg_low = r + PHX_APU_ALPHA_NEXT_LS;
    reg_high = r + PHX_APU_ALPHA_NEXT_MS;
  }

#if _VERBOSE
  MX_INFO(("%s: channel=%d low=0x%x high=0x%x\n", __FUNCTION__,
           channel, low, high));
#endif
  PHX_REG_WRITE_4(reg_low, low);
  PHX_REG_WRITE_4(reg_high, high);
}

void
phx_apu_channel_print_debug(struct phx_nic_mem_global *nic, int channel)
{
  uint8_t *r = (uint8_t*)nic + PHX_APU_BASE + (channel * PHX_APU_CHANNEL_SIZE);
  uint32_t alpha_fetches, beta_fetches, gamma_fetches;
  uint8_t last_gamma_ch, loop_ch;
  uint32_t alpha_next_ls, alpha_next_ms, alpha_count;
  uint32_t beta_next_ls, beta_next_ms, beta_count;
  uint32_t gamma_next_ls, gamma_next_ms, gamma_count;

  alpha_fetches = PHX_REG_READ_4((r + PHX_APU_ALPHA_FETCHES));
  beta_fetches = PHX_REG_READ_4((r + PHX_APU_BETA_FETCHES));
  gamma_fetches = PHX_REG_READ_4((r + PHX_APU_GAMMA_FETCHES));
  last_gamma_ch = PHX_REG_READ_1((r + PHX_APU_LAST_GAMMA_CH));
  loop_ch = PHX_REG_READ_1((r + PHX_APU_LOOP_CH));
  alpha_next_ls = PHX_REG_READ_4((r + PHX_APU_ALPHA_NEXT_LS));
  alpha_next_ms = PHX_REG_READ_4((r + PHX_APU_ALPHA_NEXT_MS));
  alpha_count = PHX_REG_READ_4((r + PHX_APU_ALPHA_COUNT));
  beta_next_ls = PHX_REG_READ_4((r + PHX_APU_BETA_NEXT_LS));
  beta_next_ms = PHX_REG_READ_4((r + PHX_APU_BETA_NEXT_MS));
  beta_count = PHX_REG_READ_4((r + PHX_APU_BETA_COUNT));
  gamma_next_ls = PHX_REG_READ_4((r + PHX_APU_GAMMA_NEXT_LS));
  gamma_next_ms = PHX_REG_READ_4((r + PHX_APU_GAMMA_NEXT_MS));
  gamma_count = PHX_REG_READ_4((r + PHX_APU_GAMMA_COUNT));

  MX_INFO(("%s: channel=%d\n"
           "alpha_fetches=%u beta_fetches=%u gamma_fetches=%u\n"
           "last_gamma_ch=%u loop_ch=%u\n"
           "alpha_next_ls=0x%x alpha_next_ms=0x%x alpha_count=%u\n"
           "beta_next_ls=0x%x beta_next_ms=0x%x beta_count=%u\n"
           "gamma_next_ls=0x%x gamma_next_ms=0x%x gamma_count=%u\n",
           __FUNCTION__, channel,
           alpha_fetches, beta_fetches, gamma_fetches,
           last_gamma_ch, loop_ch,
           alpha_next_ls, alpha_next_ms, alpha_count,
           beta_next_ls, beta_next_ms, beta_count,
           gamma_next_ls, gamma_next_ms, gamma_count));
}

uint32_t
phx_read_ep_reg_tx_desc_buffer_size(struct phx_nic_mem_endpoint *ep)
{
  uint8_t *nic = (uint8_t*)ep + PHX_EP_REG_TX_DESC_BUFFER_SIZE;
  return PHX_REG_READ_4(nic);
}

void
phx_write_ep_reg_tx_addr_buffer_clear(struct phx_nic_mem_endpoint *ep)
{
  uint8_t *nic = (uint8_t*)ep + PHX_EP_REG_TX_ADDR_BUFFER_CLEAR;
  PHX_REG_WRITE_4(nic, 1);
}

void
phx_write_ep_reg_tx_addr_load(struct phx_nic_mem_endpoint *ep, uint32_t low, uint32_t high)
{
  uint8_t *nic;

  /* MX_INFO(("%s: high=0x%x low=0x%x\n", __FUNCTION__, high, low)); */
  nic = (uint8_t*)ep + PHX_EP_REG_TX_ADDR_LOAD_LSB;
  PHX_REG_WRITE_4(nic, low);
  nic = (uint8_t*)ep + PHX_EP_REG_TX_ADDR_LOAD_MSB;
  PHX_REG_WRITE_4(nic, high);
}

void
phx_write_ep_reg_tx_addr_buffer_finish(struct phx_nic_mem_endpoint *ep)
{
  uint8_t *nic = (uint8_t*)ep + PHX_EP_REG_TX_ADDR_BUFFER_FINISH;
  PHX_REG_WRITE_4(nic, 1);
}

void
phx_write_reg_tx_host_completion(struct phx_nic_mem_global *nic_mem, uint32_t low, uint32_t high)
{
  uint8_t *nic;

  /* MX_INFO(("%s: high=0x%x low=0x%x\n", __FUNCTION__, high, low)); */
  nic = (uint8_t*)nic_mem + PHX_REG_TX_BASE + PHX_REG_TX_HOST_COMPLETION_LSB;
  PHX_REG_WRITE_4(nic, low);
  nic = (uint8_t*)nic_mem + PHX_REG_TX_BASE + PHX_REG_TX_HOST_COMPLETION_MSB;
  PHX_REG_WRITE_4(nic, high);
}

void
phx_tx_enable_auto_read_request(struct phx_nic_mem_global *nic_mem)
{
  uint8_t *r;

  r = (uint8_t*)nic_mem + PHX_REG_TX_BASE + PHX_REG_TX_AUTO_READ_REQUEST;
  PHX_REG_WRITE_4(r, 1);
}

void
phx_tx_numdwperpush(struct phx_nic_mem_global *nic_mem, uint32_t dw)
{
  uint8_t *nic;

  nic = (uint8_t*)nic_mem + PHX_REG_TX_BASE + PHX_REG_TX_NUM_DW_PER_PUSH;
  PHX_REG_WRITE_4(nic, dw);
}

void
phx_tx_threshold(struct phx_nic_mem_global *nic_mem, uint32_t pkts)
{
  uint8_t *nic;

  nic = (uint8_t*)nic_mem + PHX_REG_TX_BASE + PHX_REG_TX_THRESHOLD;
  PHX_REG_WRITE_4(nic, pkts);
}

void
phx_tx_wake_request(struct phx_nic_mem_endpoint *ep, uint32_t host_count)
{
  uint8_t *nic;

  /* MX_INFO(("%s: host_count=%u\n", __FUNCTION__, host_count)); */
  nic = (uint8_t*)ep + PHX_EP_REG_TX_WAKE_REQUEST;
  PHX_REG_WRITE_4(nic, host_count);
}

uint32_t
phx_ep_tx_status(struct phx_nic_mem_endpoint *ep)
{
  uint8_t *nic;

  MX_INFO(("%s:\n", __FUNCTION__));
  nic = (uint8_t*)ep + PHX_EP_REG_TX_STATUS;
  return PHX_REG_READ_4(nic);
}

uint32_t
phx_ep_tx_completion_count(struct phx_nic_mem_endpoint *ep)
{
  uint8_t *nic;

  nic = (uint8_t*)ep + PHX_EP_REG_TX_COMPLETION_COUNT;
  return PHX_REG_READ_4(nic);
}

struct phx_nic_mem_endpoint *
phx_endpoint_nic_mem(mx_instance_state_t *is, int endpoint)
{
  uint8_t *nic = (uint8_t *)is->phx.nic_mem;
  uint32_t base = is->phx.pfi->tx_config.dbltx_core_base;
  uint32_t stride = is->phx.pfi->tx_config.dbltx_core_stride;

  return (struct phx_nic_mem_endpoint*)(nic + base + (endpoint * stride));
}

void *
phx_ep_pio_fifo(mx_instance_state_t *is, int endpoint)
{
  uint8_t *nic = (uint8_t *)is->phx.nic_mem;
  uint32_t base = is->phx.pfi->tx_config.dbltx_pio_base;
  uint32_t stride = is->phx.pfi->tx_config.dbltx_pio_stride;

  return (void *)(nic + base + (endpoint * stride));
}

phx_tx_desc_t *
phx_ep_desc_fifo(struct phx_nic_mem_endpoint *ep)
{
  return (phx_tx_desc_t*)((uint8_t*)ep + PHX_EP_REG_TX_DESCRIPTOR_DATA);
}

void
phx_bbsysctl_pio_print_debug_info(struct phx_nic_mem_global *nic)
{
  uint8_t *r = (uint8_t*)nic;
  uint32_t rd_dw, rd_header, wr_dw, wr_header;

  rd_dw = PHX_REG_READ_4((r+PHX_BBSYSCTL_CP_RD_DW));
  rd_header = PHX_REG_READ_4((r+PHX_BBSYSCTL_CP_RD_HEADER));
  wr_dw = PHX_REG_READ_4((r+PHX_BBSYSCTL_CP_WR_DW));
  wr_header = PHX_REG_READ_4((r+PHX_BBSYSCTL_CP_WR_HEADER));

  MX_INFO(("PIO debug counters: "
           "rd_dw=%u rd_header=%u wr_dw=%u wr_header=%u\n",
           rd_dw, rd_header, wr_dw, wr_header));
}

void
phx_rx_selector_reset(struct phx_nic_mem_global *nic)
{
  uint8_t *r = (uint8_t*)nic + PHX_RX_SEL_BASE;
  uint32_t flags;

  flags = 0x03;   // reset core and counters
#if _VERBOSE
  MX_INFO(("%s: resetting rx selector and loopback \n", __FUNCTION__));
#endif
  PHX_REG_WRITE_4((r + PHX_RX_SEL_LOOPBACK_RESET), flags);
  PHX_REG_WRITE_4((r + PHX_RX_SEL_RESET), flags);
}

void
phx_rx_selector_bypass_disable(struct phx_nic_mem_global *nic, uint32_t flags)
{
  uint8_t *r = (uint8_t*)nic + PHX_RX_SEL_BASE;

#if _VERBOSE
  MX_INFO(("%s: setting %d\n", __FUNCTION__, flags));
#endif
  PHX_REG_WRITE_4((r + PHX_RX_SEL_BYPASS_DISABLE), flags);
}

/* a bit of a hack.  Look into one of the debug registers to get the
   last seqnum used.  Used to initialize the next expected one during
   board init */
uint8_t phx_icu_get_last_interrupt_seqnum(struct phx_nic_mem_global *nic)
{
  uint8_t *r = (uint8_t*)nic;
  uint8_t seqnum;
  uint32_t read;

  read = PHX_REG_READ_4((r + PHX_ICU_BASE + PHX_ICU_LAST_IR_DESC_W3));
  seqnum = (uint8_t)((read >> 24) & 0xff);
  MX_INFO(("Last interrupt seqnum was %d\n", seqnum));
  return seqnum;
}

static void
phx_icu_enable_interrupt_internal(struct phx_nic_mem_global *nic,
                                  uint32_t intr_num)
{
  uint8_t *r = (uint8_t*)nic + PHX_ICU_BASE;
#if MAL_DEBUG
  uint32_t enable_lo, enable_mid, enable_hi;

  enable_lo = PHX_REG_READ_4((r + PHX_ICU_ENABLE_LO));
  enable_mid = PHX_REG_READ_4((r + PHX_ICU_ENABLE_MID));
  enable_hi = PHX_REG_READ_4((r + PHX_ICU_ENABLE_HI));
  MX_INFO(("Enable interrupt. intr_num=%u "
           "enable_lo=0x%x enable_mid=0x%x enable_hi=0x%x\n",
	   intr_num, enable_lo, enable_mid, enable_hi));
#endif

  PHX_REG_WRITE_4((r + PHX_ICU_ENABLE_IR), intr_num);

#if MAL_DEBUG
  /* print out the enabled bits again */
  enable_lo = PHX_REG_READ_4((r + PHX_ICU_ENABLE_LO));
  enable_mid = PHX_REG_READ_4((r + PHX_ICU_ENABLE_MID));
  enable_hi = PHX_REG_READ_4((r + PHX_ICU_ENABLE_HI));

  MX_INFO(("enable_lo=0x%x enable_mid=0x%x enable_hi=0x%x\n",
	   enable_lo, enable_mid, enable_hi));
#endif
}

static void
phx_icu_disable_interrupt_internal(struct phx_nic_mem_global *nic,
                                   uint32_t intr_num)
{
  uint8_t *r = (uint8_t*)nic + PHX_ICU_BASE;

#if _VERBOSE
  MX_INFO(("%s: intr_num=%u\n", __FUNCTION__, intr_num));
#endif
  PHX_REG_WRITE_4((r + PHX_ICU_DISABLE_IR), intr_num);
}

static void
phx_icu_reset_endpt_internal(struct phx_nic_mem_global *nic, uint32_t endpt_num)
{
  uint8_t *r = (uint8_t*)nic + PHX_ICU_BASE;

  PHX_REG_WRITE_4((r + PHX_ICU_EP_RESET), endpt_num);
}

uint8_t
phx_icu_read_link_status(struct phx_nic_mem_global *nic)
{
  uint8_t *r = (uint8_t*)nic + PHX_ICU_BASE;
  uint32_t word;
  struct phx_icu_link_status status;

  word = PHX_REG_READ_4((r + PHX_ICU_LINK_STATUS));
  status = *((struct phx_icu_link_status*)&word);
#if _VERBOSE
  MX_INFO(("link_status. 0x%x 0x%x 0x%x 0x%x\n", status.status, status.intr_req,
           status.unused, status.last_intr_status));
#endif
  return status.status;
}

void phx_icu_request_link_change_interrupt(struct phx_nic_mem_global *nic)
{
  uint8_t *r = (uint8_t*)nic + PHX_ICU_BASE;

  PHX_REG_WRITE_4((r + PHX_ICU_REQ_LINK_IR), 0);
}

void
phx_icu_enable(mx_instance_state_t *is)
{
  uint8_t *r;
  uint32_t *w;
  uint32_t endpoints, global_mask, enable_lo, enable_mid, enable_hi;
  uint32_t vectors_alloc, operation_mode;
  int i, function_endpoints;

  if (MX_IS_K35SNF(is)) {
    MX_WARN(("BUG: K35SNF should not call phx_icu_enable\n"));
    return;
  }

  if (!is->irq_one_to_one_enabled && !is->irq_shared_enabled) {
    MX_INFO(("No IRQs enabled.  ICU is left disabled.\n"));
    return;
  }

  r = (uint8_t*)is->phx.nic_mem;
  r += PHX_ICU_BASE;

  endpoints = PHX_REG_READ_4((r + PHX_ICU_ENDPOINTS));
  vectors_alloc = PHX_REG_READ_4((r + PHX_ICU_VECTORS_ALLOC));
  operation_mode = PHX_REG_READ_4((r + PHX_ICU_OPERATION_MODE));
  global_mask = PHX_REG_READ_4((r + PHX_ICU_GLOBAL_MASK));
  enable_lo = PHX_REG_READ_4((r + PHX_ICU_ENABLE_LO));
  enable_mid = PHX_REG_READ_4((r + PHX_ICU_ENABLE_MID));
  enable_hi = PHX_REG_READ_4((r + PHX_ICU_ENABLE_HI));

#if _VERBOSE
  MX_INFO(("endpoints=%u vectors_alloc=%u operation_mode=%u"
           " global_mask=0x%x enable_lo=0x%x enable_mid=0x%x enable_hi=0x%x\n",
           endpoints, vectors_alloc, operation_mode, global_mask,
           enable_lo, enable_mid, enable_hi));
#endif

  /* Fill in the interrupt descriptor addresses.
   * First, vector 0 (global events).
   */
  w = (uint32_t*)(r + PHX_ICU_DESC_ADDRESS_OFFSET);
  PHX_REG_WRITE_4(w, is->intr.pin.dma.low);
  w++;
  PHX_REG_WRITE_4(w, is->intr.pin.dma.high);
  w++;
  vectors_alloc = 1;
  operation_mode = 0;

  function_endpoints = is->phx.pfi->ep_pool.num_function_endpoints;

  /* We need to write these pages unconditionally even in shared mode as
     the fpga still writes the descriptor to the endpoint specific pages. */
  for (i = 0; i < function_endpoints; i++) {
    nic_dma_addr_t dma = is->ep_irq[i].intr.pin.dma;
    PHX_REG_WRITE_4(w, dma.low);
    w++;
    PHX_REG_WRITE_4(w, dma.high);
    w++;
  }

  if (is->irq_one_to_one_enabled) {
    vectors_alloc += function_endpoints;
    operation_mode = 1;
  }

#if _VERBOSE
  MX_INFO(("Writing vectors_alloc. vectors_alloc=%u\n", vectors_alloc));
#endif
  PHX_REG_WRITE_4((r + PHX_ICU_VECTORS_ALLOC), vectors_alloc);
#if _VERBOSE
  vectors_alloc = PHX_REG_READ_4((r + PHX_ICU_VECTORS_ALLOC));
  MX_INFO(("vectors_alloc=%u\n", vectors_alloc));

  MX_INFO(("Writing operation_mode. mode=%u\n", operation_mode));
#endif
  PHX_REG_WRITE_4((r + PHX_ICU_OPERATION_MODE), operation_mode);
#if _VERBOSE
  operation_mode = PHX_REG_READ_4((r + PHX_ICU_OPERATION_MODE));
  MX_INFO(("operation_mode=%u\n", operation_mode));
#endif

#if _VERBOSE
  MX_INFO(("ICU is enabled.\n"));
#endif
}

void phx_print_intr_desc(phx_intr_t *desc)
{
  MX_INFO(("endpt_bitmap=%x:%x:%x:%x:%x:%x:%x:%x "
           "flags=0x%x seqnum=%u\n",
           desc->endpt_bitmap[0], desc->endpt_bitmap[1],
           desc->endpt_bitmap[2], desc->endpt_bitmap[3],
           desc->endpt_bitmap[4], desc->endpt_bitmap[5],
           desc->endpt_bitmap[6], desc->endpt_bitmap[7],
           desc->flags, desc->seqnum));
}

void phx_read_bitstream_sha1(struct phx_nic_mem_global *nic, uint32_t sha1_val[4])
{
  uint8_t *r;
  uint32_t val, *v;
  int i;

  r = (uint8_t *)nic + PHX_I2C_BASE + 0x180;
  v = sha1_val;
  for (i = 0; i < 4; i ++) {
    val = PHX_REG_READ_4(r);
    /* Byte swap so the caller can printf it and see the bytes in the right order. */
    *v = ((val & 0xff) << 24) |
        (((val >> 8) & 0xff) << 16) |
        (((val >> 16) & 0xff) << 8) |
        ((val >> 24) & 0xff);
    v++;
    r += 4;
  }
}

static void phx_eeprom_read_all(char* base, uint8_t device, char *buf, size_t len, uint32_t xlx_board)
{
        unsigned char data[256], *datap;
        uint16_t reg_ofs;
        uint8_t i;

        /* since reading the eeprom requires i2c, grab the mutex */
	do {} while(!phx_i2c_mutex_claim(base));

        reg_ofs=0;

        if (xlx_board == KCU105_ID_REGVAL){
		/* open the mux on KCU105 */
                avl_i2c_set_mux(base, KCU105_DEV_MUXx4, 4);
		if (device == DEV_SPECS) device = KCU105_DEV_SPECS;
		if (device == DEV_LICENSE) device = KCU105_DEV_LICENSE;
	} else if (xlx_board == K35_ID_REGVAL){
		/* open the mux on K35 */
                avl_i2c_set_mux(base, K35_DEV_MUXx8, 1);
                /* same physical part with 2 byte offset */
		if (device == DEV_SPECS) reg_ofs = 0;
		if (device == DEV_LICENSE) reg_ofs = 256;
		device = K35_DEV_EEPROM;
	}

        for (i=0; i< 16; i++){
                datap = &data[i*16];
                if (xlx_board == K35_ID_REGVAL)
                       at24c64_i2c_read_bytes(base, AVL_BASE_I2C_0, device, reg_ofs, datap, 16); // two byte offset
		else
                       avl_i2c_read_bytes(base, AVL_BASE_I2C_0, device, (uint8_t) reg_ofs, datap, 16);
                reg_ofs += 16;
        }

	/* close any opened muxes */
       if (xlx_board == KCU105_ID_REGVAL){
                avl_i2c_set_mux(base, KCU105_DEV_MUXx4, 0);
        } else if (xlx_board == K35_ID_REGVAL){
                avl_i2c_set_mux(base, K35_DEV_MUXx8, 0);
        }

        phx_i2c_mutex_release(base);

        if (len > 256)
                len = 256;

        memcpy(buf, data, len);
}

void phx_read_specs(char *base, char *specs, size_t len, uint32_t xlx_board)
{
        phx_eeprom_read_all(base, DEV_SPECS, specs, len, xlx_board);
}
void phx_read_license(char* base, char *lic, size_t len, uint32_t xlx_board)
{
        phx_eeprom_read_all(base, DEV_LICENSE, lic, len, xlx_board);
}

static void phx_read_qsfp(char* base, char *buf, size_t len, uint32_t port, uint32_t xlx_board)
{
        unsigned char data[16];

        memset(data, 0, 16);

        /* since reading requires i2c, grab the mutex */
	do {} while(!phx_i2c_mutex_claim(base));

	if (xlx_board == K35_ID_REGVAL){
                avl_i2c_set_mux(base, K35_DEV_MUXx8, (port + 2));
                avl_i2c_read_bytes(base, AVL_BASE_I2C_0, DEV_QSFP, (uint8_t) 0, data, 16);
                avl_i2c_set_mux(base, K35_DEV_MUXx8, 0);
        }

        phx_i2c_mutex_release(base);

        if (len > 16)
                len = 16;

        memcpy(buf, data, len);
}

/* Certain Finisar Transceivers have an internal Marvell PHY */
#define MPHY_CONTROL_REG           0
#define MPHY_1000BT_CONTROL_REG    9
#define MPHY_ENABLE_RX_LOS_REG     24

void phx_reset_marvell_88E1111(mx_instance_state_t *is, uint32_t port_id)
{
  uint8_t *base = (uint8_t*)is->phx.i2c_base_mem;
  uint32_t board_type = 0;
  unsigned char data[2];

  board_type = is->phx.pfi->xilinx_board_type;

  /* since reading requires i2c, grab the mutex */
  do {} while(!phx_i2c_mutex_claim(base));

  /* currently only the K35 supports 1G MAC */
  if (board_type == K35_ID_REGVAL){
    avl_i2c_set_mux(base, K35_DEV_MUXx8, (port_id + 2));

    /* advertise 1G autonegotiation */
    data[0]=0x0F;
    data[1]=0x00;
    avl_i2c_write_bytes(base, AVL_BASE_I2C_0, K35_DEV_MARVELL, MPHY_1000BT_CONTROL_REG, data, 2);

    /* enable LOS led (tied to MAC) */
    data[0]=0x41;
    data[1]=0x08;
    avl_i2c_write_bytes(base, AVL_BASE_I2C_0, K35_DEV_MARVELL, MPHY_ENABLE_RX_LOS_REG, data, 2);

    /* reset the chip - causes 1G negotiation to succeed */
    data[0] = 0x91;
    data[1] = 0x40;
    avl_i2c_write_bytes(base, AVL_BASE_I2C_0, K35_DEV_MARVELL, MPHY_CONTROL_REG, data, 2);

    avl_i2c_set_mux(base, K35_DEV_MUXx8, 0);
  }

  phx_i2c_mutex_release(base);
}

#if MAL_OS_WINDOWS
/* Really hack to read the MAC address from string specs, before we do
 * mx_instance_init.
 */
int
phx_win_init_fpga(mx_instance_state_t *is)
{
  uint32_t version;
  uint32_t board_arch=0;
  int init_avl;
  int attempts = 0;
  uint16_t vendor;
  uint16_t device;

  if (phx_common_map_board(is))
    return -1;
  /* attach_fpga_instance needs board_ops, so set it up first. */
  mx_read_pci_config_16(is, offsetof(mx_pci_config_t, Vendor_ID), &vendor);
  mx_read_pci_config_16(is, offsetof(mx_pci_config_t, Device_ID), &device);
  if (mx_select_board_type(is, vendor, device, is->pci_rev))
    return -1;
  phx_attach_fpga_instance(is);

  mx_mutex_enter(&mx_fpga_mutex);
  init_avl = !is->phx.pfi->avl_initialized;
  if (init_avl) {
    MX_INFO(("Initializing FPGA: start\n"));
    if (is->board_type == MAL_BOARD_TYPE_PHX_K35)
      phx_xilinx_init((char*)is->phx.i2c_base_mem, &version, &board_arch);
    else{
      phx_avl_init((char*)is->phx.i2c_base_mem, &version);
    }
    /* firmware monotonic version for compatibility checks */
    is->phx.pfi->firmware_rev = version;
    is->phx.pfi->xilinx_board_type = board_arch;
    if (board_arch)
      MX_INFO(("Initializing FPGA: end, firmware version %d board arch %c%c%c%c\n", version, 
              ((board_arch & 0xff000000) >> 24), 
              ((board_arch & 0xff0000) >> 16), 
              ((board_arch & 0xff00) >> 8), 
              (board_arch & 0xff)));
    else
      MX_INFO(("Initializing FPGA: end, firmware version %d\n", version)); 

    /* read the eeprom strings only once */
    if (is->phx.pfi->eeprom_strings)
      phx_read_specs((char*)is->phx.i2c_base_mem, is->phx.pfi->eeprom_strings, 256, 
                      is->phx.pfi->xilinx_board_type);

    is->phx.pfi->avl_initialized = 1;
  }
  mx_mutex_exit(&mx_fpga_mutex);

  /* Do not detach from the FPGA until this instance either fails
   * during the initialization or halts.
   */

  is->phx.eeprom_strings = mx_kmalloc(MX_EEPROM_STRINGS_LEN, MX_NOWAIT);
  if ((!is->phx.eeprom_strings) || (!is->phx.pfi->eeprom_strings))
    return -1;

  memcpy(is->phx.eeprom_strings, is->phx.pfi->eeprom_strings, 256);  /* private copy gets modified during parse */

  /* This sets MAC address (is->mac_addr).
   * PHX-310 retry reading if copy if not OK. 
   */
  while (phx_parse_eeprom_strings(is) && (attempts++ < 10)){
    MX_INFO(("FATAL error could not process eeprom strings\n"));
    phx_read_specs((char*)is->phx.i2c_base_mem, is->phx.eeprom_strings, 256, 
                   is->phx.pfi->xilinx_board_type);
  }

  phx_common_unmap_board(is);
  mx_kfree(is->phx.eeprom_strings);
  is->phx.eeprom_strings = NULL;
  return 0;
}
#endif

int
phx_start_dmabench(mx_instance_state_t *is, mx_dmabench_t *x, mx_page_pin_t *pin)
{
  uint8_t *nic;
  uint8_t *r;
  uint32_t size, count, cmd;
  uint32_t t0, t1, dt;
  uint32_t status;

  nic = (uint8_t *)is->phx.nic_mem;

  r = (uint8_t *)nic + PHX_DMABENCH_STATUS;
  status = PHX_REG_READ_4(r);

  /* is the dmabench core available */
  if (status != 0){
        MX_WARN(("DMAbench: bad starting status = 0x%08x\n", status));
        /* confirm we really are accessing this core */
        r = (uint8_t *) nic + ( PHX_DMABENCH_CONST0);
        status = PHX_REG_READ_4(r);
        MX_INFO(("DMAbench const0 %p = 0x%08x\n", r, status));
        r = (uint8_t *) nic + ( PHX_DMABENCH_CONST1);
        status = PHX_REG_READ_4(r);
        MX_INFO(("DMAbench const0 %p = 0x%08x\n", r, status));
	return EBUSY;
  }

  /* error check parameters */
  size = (1 << (x->log_size));
  if (size > 4096) size = 4096;

  count = x->count;
  if (count > 100) count = 100;

  cmd = x->dma_read + 1;
  if (cmd > 2) cmd = 1;

  /* makle sure the timer is running */
  r = (uint8_t *) nic + PHX_DMABENCH_FREE;
  t0 = PHX_REG_READ_4(r);
  t1 = PHX_REG_READ_4(r);
  dt = t1 - t0;

  if (dt == 0){
    MX_INFO(("DMABench freeCount frozen at 0x%x\n",  t0));
   } else {
    MX_INFO(("DMABench observed %d cycles between sequential freeCount calls.\n", dt));
  }
  MX_INFO(("DMAbench: iterations = 0x%08x size= 0x%08x\n", count, size));

  /* DMA bench command support */
  r = (uint8_t *) nic+ PHX_DMABENCH_DMA_LS;
  PHX_REG_WRITE_4(r, pin->dma.low);
  r = (uint8_t *) nic+ PHX_DMABENCH_DMA_MS;
  PHX_REG_WRITE_4(r, pin->dma.high);
  r = (uint8_t *) nic+ PHX_DMABENCH_LAST;
  PHX_REG_WRITE_4(r, 0);
  r = (uint8_t *) nic+ PHX_DMABENCH_EXTENT;
  PHX_REG_WRITE_4(r, size);
  r = (uint8_t *) nic+ PHX_DMABENCH_ITER;
  PHX_REG_WRITE_4(r, count);
  r = (uint8_t *) nic+ PHX_DMABENCH_WRITE_VAL;
  PHX_REG_WRITE_4(r, 0xc0bebeef);
  r = (uint8_t *) nic+ PHX_DMABENCH_COMMAND;
  PHX_REG_WRITE_4(r, cmd);

  return 0;
}

int
phx_finish_dmabench(mx_instance_state_t *is, mx_dmabench_t *x)
{
  uint8_t *nic;
  uint8_t *r;
  uint32_t status;
  uint32_t left;
  uint32_t cycles;

  nic = (uint8_t *)is->phx.nic_mem;

  MX_INFO(("DMAbench: finishing, register ptr 0x%p\n", nic));
  r = (uint8_t *)nic + PHX_DMABENCH_STATUS;
  status = PHX_REG_READ_4(r);

  if (status != 0) {
     MX_WARN(("DMAbench: bad ending status = 0x%08x\n", status));
     /* are there still iterations to go */
     r = (uint8_t *) nic + PHX_DMABENCH_ITER_REM;
     left = PHX_REG_READ_4(r);
     MX_INFO(("DMAbench: cycles left to go = 0x%08x\n", left));
     return EBUSY;
  }

  /* get results of DMAbench */
  r = (uint8_t *) nic + PHX_DMABENCH_ITER_REM;
  left = PHX_REG_READ_4(r);
     MX_INFO(("DMAbench: iterations left to go = 0x%08x\n", left));
  x->count -= left;
     MX_INFO(("DMAbench: count = 0x%08x\n", x->count));

  r = (uint8_t *) nic + PHX_DMABENCH_LAST;
  cycles = PHX_REG_READ_4(r);
     MX_INFO(("DMAbench: cycles = 0x%08x\n", cycles));
  x->cycles = cycles;

  return 0;
}

void phx_i2c_mutex_release(uint8_t *i2c_base)
{
  uint8_t *r = i2c_base + PHX_I2C_MUTEX;
  uint32_t val;

  val = PHX_REG_READ_4(r);
  if (val == PHX_I2C_MUTEX_CLAIM_VAL) {
#if _VERBOSE
    MX_INFO(("%s: release mutex\n", __FUNCTION__));
#endif
    PHX_REG_WRITE_4(r, PHX_I2C_MUTEX_RELEASE_VAL);
  } else {
    MX_WARN(("%s: Mutex wasn't ours! Ignoring\n", __FUNCTION__));
  }
}

int phx_i2c_mutex_claim(uint8_t *i2c_base)
{
  int i;
  uint32_t val;
  uint8_t *r = i2c_base + PHX_I2C_MUTEX;


  for (i = 0; i < 1000; i++) {
    PHX_REG_WRITE_4(r, PHX_I2C_MUTEX_CLAIM_VAL);
    val = PHX_REG_READ_4(r);
    if (val == PHX_I2C_MUTEX_CLAIM_VAL) {
      break;
    }
#if MAL_OS_LINUX
    udelay(1);
#else
    phx_win_usleep(1);
#endif
  }

  if (i >= 1000) {
    MX_WARN(("%s: Breaking lock on I2C mutex after %d attempts!\n", __FUNCTION__, i));
    PHX_REG_WRITE_4(r, (val & 0xffff0000));
    return 0;
  }
#if _VERBOSE
  MX_INFO(("%s: Mutex claimed on attempt %d.\n", __FUNCTION__, i+1));
#endif
  return 1;
}

void
phx_free_pages(mx_instance_state_t *is, phx_pages_t *p)
{
  mx_page_pin_t *pages;
  char *addr;
  int i;

  if (p && p->pages) {
#if MAL_DEBUG
    MX_INFO(("free_phx_pages: is_name=%s pages=%p n_pages=%d\n",
             is->is_name, p->pages, p->n_pages));
#endif
    pages = p->pages;
    for (i = 0; i < p->n_pages; i++) {
      if (pages[i].va) {
        addr = (char*)pages[i].va;
        mx_free_dma_page(is, &addr, &pages[i]);
        memset(&pages[i], 0, sizeof(pages[i]));
      }
    }
    mx_kfree(pages);
    memset(p, 0, sizeof(*p));
  }
}

int
phx_alloc_pages(mx_instance_state_t *is, phx_pages_t *p, int n_pages)
{
  mx_page_pin_t *pages;
  char *addr;
  int i;

#if MAL_DEBUG
  MX_WARN(("alloc_phx_pages: is_name=%s n_pages=%d\n", is->is_name, n_pages));
#endif
  pages = mx_kmalloc(sizeof(mx_page_pin_t) * n_pages, MX_MZERO | MX_NOWAIT);
  if (!pages)
    goto abort;
  p->pages = pages;
  p->n_pages = n_pages;
  for (i = 0; i < n_pages; i++) {
    if (mx_alloc_dma_page(is, &addr, &pages[i]))
      goto abort;
    memset(addr, 0, PAGE_SIZE);
  }
  return 0;

abort:
  phx_free_pages(is, p);
  return -ENOMEM;
}

static void
setup_apu_alpha_beta_pages(phx_rx_ring_t *ring)
{
  int n_beta, n_alpha;
  int i, j, k, rem;
  mx_page_pin_t *page, *alpha_page;
  phx_dma_addr_t *addr;
  phx_apu_flp_t *flp, *prev_flp;

  /* Fill in beta pages */
  n_beta = ring->beta.n_pages;
  page = ring->gamma.pages + ring->preamble_n_pages; /* skip the preamble pages */
  for (i = 0; i < n_beta; i++) {
    addr = (phx_dma_addr_t *)ring->beta.pages[i].va;
    for (j = 0; j < PHX_APU_GAMMA_PAGES_PER_BETA; j++) {
      addr->low = page->dma.low;
      addr->high = page->dma.high;
      addr++;
      page++;
    }
  }

  /* Fill in alpha pages */
  n_alpha = ring->alpha.n_pages;
  page = ring->beta.pages;
  rem = n_beta;
  prev_flp = NULL;
  for (i = 0; rem > 0 && i < n_alpha; i++) {
    uint64_t flp_dma_addr;

    alpha_page = &ring->alpha.pages[i];
    flp = (phx_apu_flp_t *)alpha_page->va;
    flp_dma_addr = (((uint64_t)alpha_page->dma.high) << 32) | (uint64_t)alpha_page->dma.low;
    for (j = 0; rem > 0 && j < PHX_APU_FLPS_PER_ALPHA; j++) {
      for (k = 0; rem > 0 && k < PHX_APU_BETA_PAGES_PER_FLP; k++) {
        flp->beta[k].low = page->dma.low;
        flp->beta[k].high = page->dma.high;
        rem--;
        page++;
      }
      /* Link FLPs.  The previous FLP points to the current FLP. */
      if (prev_flp) {
        prev_flp->next.low = (uint32_t)flp_dma_addr;
        prev_flp->next.high = (uint32_t)(flp_dma_addr >> 32);
      }
      flp->opcode = 0;
      PHX_APU_FLP_OPCODE_SET_NUM_BETA(flp->opcode, k);
      prev_flp = flp;
      flp++;
      flp_dma_addr += sizeof(phx_apu_flp_t);
    }
  }
  /* The last/final FLP points back to the first FLP. */
  alpha_page = &ring->alpha.pages[0];
  prev_flp->next.low = alpha_page->dma.low;
  prev_flp->next.high = alpha_page->dma.high;
}

int
phx_alloc_rx_ring(mx_instance_state_t *is, phx_rx_ring_t *ring,
                  uint64_t size, uint32_t preamble_size)
{
  int n_gamma, n_beta, n_alpha, n_preamble;
  int flp_per_page = PAGE_SIZE / sizeof(phx_apu_flp_t);

  /* size does not include preamble_size. */
  ring->size = size;
  ring->preamble_size = preamble_size;
  n_preamble = preamble_size / PAGE_SIZE;
  n_gamma = size / PAGE_SIZE;
  n_beta = (n_gamma + PHX_APU_GAMMA_PAGES_PER_BETA - 1) / PHX_APU_GAMMA_PAGES_PER_BETA;
  n_alpha = (n_beta + flp_per_page - 1) / flp_per_page;
#if MAL_DEBUG
  MX_INFO(("phx_alloc_rx_ring: is_name=%s pages=%d/%d/%d preamble=%d\n",
           is->is_name, n_gamma, n_beta, n_alpha, n_preamble));
#endif
  /* Sanity check.  Beta pages must be full. */
  if (n_gamma != n_beta * PHX_APU_GAMMA_PAGES_PER_BETA) {
    MX_WARN(("%s: phx_alloc_rx_ring: beta pages are not full. pages=%d/%d/%d"
             " size=%llu\n", is->is_name, n_gamma, n_beta, n_alpha,
             (unsigned long long)size));
    return -EINVAL;
  }

  ring->preamble_n_pages = n_preamble;
#if MAL_OS_WINDOWS
  if (phx_win_alloc_gamma_pages(is, &ring->gamma, n_gamma + n_preamble, &ring->gamma_mdl))
    goto abort;
#else
  if (phx_alloc_pages(is, &ring->gamma, n_gamma + n_preamble))
    goto abort;
#endif
  if (phx_alloc_pages(is, &ring->beta, n_beta))
    goto abort;
  if (phx_alloc_pages(is, &ring->alpha, n_alpha))
    goto abort;
  setup_apu_alpha_beta_pages(ring);
  return 0;

abort:
  /* These check pointers internally.  It is okay to call them without checking
   * NULL pointers.
   */
#if MAL_OS_WINDOWS
  phx_win_free_gamma_pages(is, &ring->gamma, ring->gamma_mdl);
  ring->gamma_mdl = NULL;
#else
  phx_free_pages(is, &ring->gamma);
#endif
  phx_free_pages(is, &ring->beta);
  phx_free_pages(is, &ring->alpha);
  return -ENOMEM;
}

void
phx_free_rx_ring(mx_instance_state_t *is, phx_rx_ring_t *ring)
{
#if MAL_OS_WINDOWS
  phx_win_free_gamma_pages(is, &ring->gamma, ring->gamma_mdl);
  ring->gamma_mdl = NULL;
#else
  phx_free_pages(is, &ring->gamma);
#endif
  phx_free_pages(is, &ring->beta);
  phx_free_pages(is, &ring->alpha);
  ring->size = 0;
}
int phx_sample_hwclock(mx_instance_state_t *is,
                    uint64_t *nic_ticks, uint64_t *host_nsecs,
                    uint64_t *host_nsecs_delay)
{
  uint64_t tsB_nsecs, tsE_nsecs;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0)
  struct timespec64 tsB, tsE, ts;
#else
  struct timespec tsB, tsE, ts;
#endif
  struct phx_fpga_instance *fpga = is->phx.pfi;
  unsigned long flags = 0;
  int attempts = 10;

  if (mx_is_dead(is))
  {
    return -EIO;
  }
  try_again:
    mx_getnstimeofday(&tsB);
    mx_spin_lock_irqsave(&fpga->clk.clk_spinlock, flags);
    phx_clk_get(is, &ts);
    mx_spin_unlock_irqrestore(&fpga->clk.clk_spinlock, flags);
    mx_getnstimeofday(&tsE);

    tsB_nsecs = (uint64_t)mx_timespec_to_ns((const struct timespec *)&tsB); 
    tsE_nsecs = (uint64_t)mx_timespec_to_ns((const struct timespec *)&tsE); 
    *nic_ticks = (uint64_t)mx_timespec_to_ns((const struct timespec *)&ts);
    
   if (tsB_nsecs > tsE_nsecs) 
  {
    static int once = 0;
    
    if (attempts-- > 0)
    {
      goto try_again;
    }
    
    if (!once) 
    {
      MX_WARN(("System clock going backwards? Can't sync host/NIC clocks\n"));
      once = 1;
    }
    return -EIO;
  }

  
  *host_nsecs = (tsB_nsecs + tsE_nsecs) / 2;
  *host_nsecs_delay = tsE_nsecs - tsB_nsecs;
  
  return 0;

}

volatile uint32_t *
phx_rx_desc_consr_advance(mx_instance_state_t *is, int endpoint)
{
  uint8_t *nic = (uint8_t*)is->phx.nic_mem;
  if (MX_IS_K35SNF(is)) {
    return (volatile uint32_t*)
        (nic + K35SNF_RX_BASE + K35SNF_RX_DESC_CONSR_ADVANCE +
         (K35SNF_RX_SIZE * is->port_id));
  } else {
    return (volatile uint32_t*)
        (nic + PHX_DBL_RX_BASE + PHX_DBL_RX_DESC_CONSR_ADVANCE_N +
         (PHX_DBL_RX_ENDPOINT_SIZE * endpoint));
  }
}

volatile uint32_t *
phx_rx_data_consr_advance(mx_instance_state_t *is, int endpoint)
{
  uint8_t *nic = (uint8_t*)is->phx.nic_mem;
  if (MX_IS_K35SNF(is)) {
    return (volatile uint32_t*)
        (nic + K35SNF_RX_BASE + K35SNF_RX_DATA_CONSR_ADVANCE +
         (K35SNF_RX_SIZE * is->port_id));
  } else {
    return (volatile uint32_t*)
        (nic + PHX_DBL_RX_BASE + PHX_DBL_RX_DATA_CONSR_ADVANCE_N +
         (PHX_DBL_RX_ENDPOINT_SIZE * endpoint));
  }
}

int
phx_rx_desc_size(mx_instance_state_t *is)
{
  if (MX_IS_K35SNF(is))
    return (int)sizeof(k35snf_rx_desc_t);
  else
    return (int)sizeof(phx_rx_desc_t);
}

void
phx_start_rx_endpoint(mx_instance_state_t *is, int endpoint, uint32_t data_pages,
                      uint32_t desc_pages)
{
  if (MX_IS_K35SNF(is)) {
    uint8_t *r = (uint8_t *) is->phx.nic_mem;

#if _VERBOSE
    MX_INFO(("k35snf_start_rx_endpoint. port_id=%d endpoint=%d\n",
             is->port_id, endpoint));
#endif

    if (PHX_HAS_FEATURE(is, PHX_FEATURE_FPGA_MERGE_NEXT_RING)) {
      int reset_flag = 0;

      /* 
       * if not using next ring port merge we need to disable it.
       * When enabled it will always perform the next ring merge 
       * and will break the non-merge use cases.
       */
      if (is->phx.port_merge != SNF_PORT_MERGE_MODE_NEXT_RING) {
        reset_flag = K35SNF_MERGE_NEXT_RING_DISABLE_SEQUENCER;
      }

      r += K35SNF_RSS_BASE + (K35SNF_RSS_RESET + (K35SNF_RSS_SIZE * is->port_id));

      PHX_REG_WRITE_4(r, (K35SNF_MERGE_NEXT_RING_RESET_RSS_BLOCK | 
                          K35SNF_MERGE_NEXT_RING_DISABLE_SEQUENCER |
                          K35SNF_MERGE_NEXT_RING_CLEAR_RSS_STATS));
      udelay(1);
      PHX_REG_WRITE_4(r, reset_flag);
      udelay(1);
    }

    r = (uint8_t *) is->phx.nic_mem;
    r += K35SNF_RX_BASE + (K35SNF_RX_SIZE * is->port_id);
    PHX_REG_WRITE_4((r + K35SNF_RX_DATA_RING_SIZE), data_pages);
    PHX_REG_WRITE_4((r + K35SNF_RX_DESC_RING_SIZE), desc_pages);
    PHX_REG_WRITE_4((r + K35SNF_RX_COMMAND), 0x1);

    /* enable the RX MAC for this port */
    k35snf_xmac_rx_enable(is, is->port_id);

    /* Save so we can see the ring sizes when stopping */
    is->phx.rx_desc_pages = desc_pages;
    is->phx.rx_data_pages = data_pages;
  } else {
    phx_dbl_rx_ring_size(is->phx.nic_mem, endpoint, data_pages, desc_pages);
    phx_dbl_rx_endpoint_start(is->phx.nic_mem, endpoint);
  }
}

void *
k35snf_tx_base(mx_instance_state_t *is)
{
  /* SnifferTX register base for this port */
  return (void*)((char*)is->phx.nic_mem +
                 K35SNF_TX_BASE + (is->port_id * K35SNF_TX_SIZE));
}

void
k35snf_tx_stop(mx_instance_state_t *is)
{
  /*
   * We need to increase the delay by a factor x10 to account for
   * a 1G interface.  We have seen flush timeouts in our testing
   * of 1G.  We could check if the interface is 10G or 1G and 
   * delay appropriately but instead we will take the simplier 
   * approach and delay x10 more, for a total of 30msec.
   */
  int msec = 3 * 10;
  int loop_count, max_loops;
  uint32_t val;
  uint8_t *r = (uint8_t*)k35snf_tx_base(is);
  uint8_t *reg_cmd, *reg_flushed;

  reg_cmd = r + K35SNF_TX_COMMAND;
  reg_flushed = r + K35SNF_TX_STOP_FLUSHED;

  PHX_REG_WRITE_4(reg_cmd, 0x2);

  /* Wait for stopFlushed to be 1 */
  max_loops = loop_count = 1000 * msec;
  do {
    val = PHX_REG_READ_4(reg_flushed);
    if (val != 0) {
      break;
    }
    loop_count--;
#if MAL_OS_LINUX
    udelay(1);
#else
    phx_win_usleep(1);
#endif
  } while (loop_count > 0);
  if (val == 1) {
#if _VERBOSE
    MX_INFO(("SnifferTX port %d flushed after %d usec\n",
             is->port_id, max_loops - loop_count));
#endif
  } else if  (val == 0xffffffff) {
    MX_WARN(("The board appears to be in a bad state\n"));
  } else {
    MX_WARN(("SnifferTX still not flushed after %d msec: regsiter=0x%x\n",
	     msec, val));
  }

  /* RESET */
  PHX_REG_WRITE_4(reg_cmd, 0x3);
}

void
k35snf_tx_start(mx_instance_state_t *is, uint32_t compl_low, uint32_t compl_high,
                uint32_t data_pages, uint32_t num_desc)
{
  uint8_t *r = (uint8_t*)k35snf_tx_base(is);
  PHX_REG_WRITE_4((r + K35SNF_TX_COMPL_ADDR_LOW), compl_low);
  PHX_REG_WRITE_4((r + K35SNF_TX_COMPL_ADDR_HIGH), compl_high);
  PHX_REG_WRITE_4((r + K35SNF_TX_DATA_RING_SIZE), data_pages);
  PHX_REG_WRITE_4((r + K35SNF_TX_DESC_RING_SIZE), num_desc);
  PHX_REG_WRITE_4((r + K35SNF_TX_COMMAND), 0x1);

#if _VERBOSE
  MX_INFO(("k35snf_tx_start: compl_addr=0x%08x:0x%08x data_size=%u desc_size=%u\n",
           compl_high, compl_low, data_pages, num_desc));
#endif
}

uint32_t
k35snf_tx_get_scale_factor(mx_instance_state_t *is)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem;
  r += K35SNF_SYS_BASE + K35SNF_SYS_FCSYS0;
  return PHX_REG_READ_4(r);
}

void
k35snf_xmac_rx_disable(mx_instance_state_t *is, uint32_t port_id)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem;

  r = r + K35SNF_XMAC_BASE + (K35SNF_XMAC_SIZE * port_id);
  /* Receiver Disable */
  PHX_REG_WRITE_4((r + K35SNF_XMAC_RX_CONFIG_WORD_1), 0x02000000);
  mx_spin(1000); /* Large gap help 1G finish packet rx */
  /* Receiver Reset */
  PHX_REG_WRITE_4((r + K35SNF_XMAC_RX_CONFIG_WORD_1), (1 << 31));
  mx_spin(10);
}

void
k35snf_xmac_rx_enable(mx_instance_state_t *is, uint32_t port_id)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem;

  r = r + K35SNF_XMAC_BASE + (K35SNF_XMAC_SIZE * port_id);
  /* Receiver Reset */
  PHX_REG_WRITE_4((r + K35SNF_XMAC_RX_CONFIG_WORD_1), (1 << 31));
  mx_spin(10);
  /* Receiver Enable + Disable Length/Type Check */
  PHX_REG_WRITE_4((r + K35SNF_XMAC_RX_CONFIG_WORD_1), (1 << 28) | (1 << 25));
}

void
k35snf_icu_enable(mx_instance_state_t *is)
{
  uint8_t *r;
  uint32_t *w;
  uint32_t endpoints, global_mask, enable, pci_id, intr_interval;
  uint32_t vectors_alloc, operation_mode;
  int i, function_endpoints;

  if (MX_IS_CHILD(is)) {
    MX_WARN(("BUG: A child instance should not call k35snf_icu_enable\n"));
    return;
  }

  if (!is->irq_one_to_one_enabled && !is->irq_shared_enabled) {
    MX_INFO(("No IRQs enabled.  ICU is left disabled.\n"));
    return;
  }

  r = (uint8_t*)is->phx.nic_mem;
  r += K35SNF_ICU_BASE;

  endpoints = PHX_REG_READ_4((r + K35SNF_ICU_ENDPOINTS));
  vectors_alloc = PHX_REG_READ_4((r + K35SNF_ICU_VECTORS_ALLOC));
  operation_mode = PHX_REG_READ_4((r + K35SNF_ICU_OPERATION_MODE));
  global_mask = PHX_REG_READ_4((r + K35SNF_ICU_GLOBAL_MASK));
  enable = PHX_REG_READ_4((r + K35SNF_ICU_ENABLE));
  pci_id = PHX_REG_READ_4((r + K35SNF_ICU_PCI_ID));
  intr_interval = PHX_REG_READ_4((r + K35SNF_ICU_INTR_INTERVAL));

#if MAL_DEBUG
  MX_INFO(("endpoints=%u vectors_alloc=%u operation_mode=%u"
           " global_mask=0x%x enable=0x%x pci_id=0x%x inter_interval=%u\n",
           endpoints, vectors_alloc, operation_mode, global_mask,
           enable, pci_id, intr_interval));
#endif

  /* Fill in the interrupt descriptor addresses */
  w = (uint32_t*)(r + K35SNF_ICU_IDT_ADDRESS);
  PHX_REG_WRITE_4(w, is->intr.pin.dma.low);
  w++;
  PHX_REG_WRITE_4(w, is->intr.pin.dma.high);
  w++;
  vectors_alloc = 1;
  operation_mode = 0;
  function_endpoints = is->phx.pfi->ep_pool.num_function_endpoints;
  for (i = 0; i < function_endpoints; i++) {
    nic_dma_addr_t dma = is->ep_irq[i].intr.pin.dma;
    PHX_REG_WRITE_4(w, dma.low);
    w++;
    PHX_REG_WRITE_4(w, dma.high);
    w++;
  }

  if (is->irq_one_to_one_enabled) {
    vectors_alloc += function_endpoints;
    operation_mode = 1;
  }

  /* Clear all endpoint state */
  for (i = 0; i < function_endpoints; i++) {
    PHX_REG_WRITE_4((r + K35SNF_ICU_EP_RESET), i);
  }

  PHX_REG_WRITE_4((r + K35SNF_ICU_VECTORS_ALLOC), vectors_alloc);
  PHX_REG_WRITE_4((r + K35SNF_ICU_OPERATION_MODE), operation_mode);

#if MAL_DEBUG
  vectors_alloc = PHX_REG_READ_4((r + K35SNF_ICU_VECTORS_ALLOC));
  MX_INFO(("vectors_alloc=%u\n", vectors_alloc));

  operation_mode = PHX_REG_READ_4((r + K35SNF_ICU_OPERATION_MODE));
  MX_INFO(("operation_mode=%u\n", operation_mode));
  MX_INFO(("ICU is enabled.\n"));
#endif
}

static void
k35snf_icu_enable_interrupt_internal(mx_instance_state_t *is, uint32_t intr_num)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + K35SNF_ICU_BASE;
  /* MX_INFO(("k35snf_icu_enable_interrupt: intr_num=%u\n", intr_num)); */
  PHX_REG_WRITE_4((r + K35SNF_ICU_ENABLE_IR), intr_num);
}

static void
k35snf_icu_disable_interrupt_internal(mx_instance_state_t *is, uint32_t intr_num)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + K35SNF_ICU_BASE;
#if _VERBOSE
  MX_INFO(("k35snf_icu_disable_interrupt: intr_num=%u\n", intr_num));
#endif
  PHX_REG_WRITE_4((r + K35SNF_ICU_DISABLE_IR), intr_num);
}

static void
k35snf_icu_reset_endpt_internal(mx_instance_state_t *is, uint32_t endpt_num)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + K35SNF_ICU_BASE;
#if _VERBOSE
  MX_INFO(("k35snf_icu_reset_endpt: endpt_num=%u\n", endpt_num));
#endif
  PHX_REG_WRITE_4((r + K35SNF_ICU_EP_RESET), endpt_num);
}

uint8_t
k35snf_icu_read_link_status(mx_instance_state_t *is)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + K35SNF_ICU_BASE;
  uint32_t word;
  k35snf_icu_link_status_t status;

  word = PHX_REG_READ_4((r + K35SNF_ICU_LINK_STATUS));
  status = *((k35snf_icu_link_status_t*)&word);

#if _VERBOSE
  MX_INFO(("link_status. 0x%x 0x%x 0x%x 0x%x\n", status.linkStatus,
           status.linkChangeIRReq, status.unused, status.linkLastIRStatus));
#endif

  return !!((1 << is->port_id) & status.linkStatus);
}

void
k35snf_icu_request_link_change_interrupt(mx_instance_state_t *is)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + K35SNF_ICU_BASE;
#if _VERBOSE
  MX_INFO(("k35snf_icu_request_link_change_interrupt: port_id=%d\n", is->port_id));
#endif
  PHX_REG_WRITE_4((r + K35SNF_ICU_REQ_LINK_CHANGE_IR), is->port_id);
}

void
k35snf_icu_set_last_link_status_up(mx_instance_state_t *is)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + K35SNF_ICU_BASE;
#if _VERBOSE
  MX_INFO(("k35snf_icu_set_last_link_status_up: port_id=%d\n", is->port_id));
#endif
  PHX_REG_WRITE_4((r + K35SNF_ICU_SET_STATUS_UP), is->port_id);
}

void
k35snf_icu_set_last_link_status_down(mx_instance_state_t *is)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + K35SNF_ICU_BASE;
#if _VERBOSE
  MX_INFO(("k35snf_icu_set_last_link_status_down: port_id=%d\n", is->port_id));
#endif
  PHX_REG_WRITE_4((r + K35SNF_ICU_SET_STATUS_DOWN), is->port_id);
}

uint8_t
k35snf_icu_read_pres_status(mx_instance_state_t *is)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + K35SNF_ICU_BASE;
  uint32_t word;
  k35snf_icu_pres_status_t status;

  word = PHX_REG_READ_4((r + K35SNF_ICU_PRES_STATUS));
  status = *((k35snf_icu_pres_status_t*)&word);

#if _VERBOSE
  MX_INFO(("pres_status. 0x%x 0x%x 0x%x 0x%x\n", status.presStatus,
           status.presChangeIRReq, status.unused, status.presLastIRStatus));
#endif

  return !!((1 << is->port_id) & status.presStatus);
}

void
k35snf_icu_request_pres_change_interrupt(mx_instance_state_t *is)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + K35SNF_ICU_BASE;
#if _VERBOSE
  MX_INFO(("k35snf_icu_request_pres_change_interrupt: port_id=%d\n", is->port_id));
#endif
  PHX_REG_WRITE_4((r + K35SNF_ICU_REQ_PRES_CHANGE_IR), is->port_id);
}

void
k35snf_icu_set_last_pres_status_in(mx_instance_state_t *is)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + K35SNF_ICU_BASE;
#if _VERBOSE
  MX_INFO(("k35snf_icu_set_last_pres_status_in: port_id=%d\n", is->port_id));
#endif
  PHX_REG_WRITE_4((r + K35SNF_ICU_SET_PRES_IN), is->port_id);
}

void
k35snf_icu_set_last_pres_status_out(mx_instance_state_t *is)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem + K35SNF_ICU_BASE;
#if _VERBOSE
  MX_INFO(("k35snf_icu_set_last_pres_status_out: port_id=%d\n", is->port_id));
#endif
  PHX_REG_WRITE_4((r + K35SNF_ICU_SET_PRES_OUT), is->port_id);
}

uint8_t
k35snf_read_pres_status(mx_instance_state_t *is)
{
  uint8_t *r = (uint8_t*)is->phx.i2c_base_mem;
  uint32_t status;

  status = PHX_REG_READ_4((r + K35_SFP_PRESENT));

  return !!((1 << is->port_id) & status);
}

uint64_t
k35snf_xmac_get_link_speed(mx_instance_state_t *is)
{
  uint8_t *r = (uint8_t *)is->phx.nic_mem;
  uint32_t mac_type;

  r += K35SNF_XMAC_BASE + (K35SNF_XMAC_SIZE * is->port_id);
  mac_type = PHX_REG_READ_4((r + K35SNF_XMAC_VERSION));
  if (mac_type == K35SNF_MAC_1G)
    return 1000000000;   /* 10^9 bps */
  else
    return 10000000000; /* 10^10 bps */
}

uint32_t
phx_get_transceiver_type(mx_instance_state_t *is, uint32_t port)
{
  char sfp_id[16];
  uint8_t *r = (uint8_t*)is->phx.i2c_base_mem;
  uint32_t xcvr_type = 0;

  phx_read_qsfp(r, sfp_id, 16, port, is->phx.pfi->xilinx_board_type);

  xcvr_type =  (uint32_t) sfp_id[2];

  return xcvr_type;
}

void
phx_set_serdes_preemphasis(mx_instance_state_t *is, uint32_t port, uint32_t value)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem;

  r += K35_SERDES_PREEMPHASIS_BASE + (K35_SERDES_PREEMPHASIS_SIZE * port);
  PHX_REG_WRITE_4(r, value);
}

void
k35snf_rx_wake_request(mx_instance_state_t *is, uint32_t host_count)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem;

  r += K35SNF_RX_BASE + (K35SNF_RX_SIZE * is->port_id);
  PHX_REG_WRITE_4((r + K35SNF_RX_EP_DESC_WAKE_REQUEST), host_count);
  /*
  MX_INFO(("k35snf_rx_wake_request: port_id=%d host_count=%u\n",
           is->port_id, host_count));
  */
}

uint32_t
k35snf_tx_compl_count(mx_instance_state_t *is)
{
  uint8_t *r = (uint8_t*)k35snf_tx_base(is);
  uint32_t last_completed;

  last_completed = PHX_REG_READ_4((r + K35SNF_TX_COMPL_COUNT));

  return last_completed;
}

void
k35snf_tx_compl_count_ack(mx_instance_state_t *is, uint32_t packets)
{
  uint8_t *r = (uint8_t*)k35snf_tx_base(is);

  /* MX_INFO(("k35snf_tx_compl_count_ack: ack=%u\n", packets)); */
  PHX_REG_WRITE_4((r + K35SNF_TX_COMPL_COUNT_ACK), packets);
}

void
phx_icu_enable_interrupt(mx_instance_state_t *is, uint32_t intr_num)
{
  if (MX_IS_K35SNF(is))
    k35snf_icu_enable_interrupt_internal(is, intr_num);
  else
    phx_icu_enable_interrupt_internal(is->phx.nic_mem, intr_num);
}

void
phx_icu_disable_interrupt(mx_instance_state_t *is, uint32_t intr_num)
{
  if (MX_IS_K35SNF(is))
    k35snf_icu_disable_interrupt_internal(is, intr_num);
  else
    phx_icu_disable_interrupt_internal(is->phx.nic_mem, intr_num);
}

void
phx_icu_reset_endpt(mx_instance_state_t *is, const phx_physical_ep_t *ep)
{
  if (MX_IS_K35SNF(is))
    k35snf_icu_reset_endpt_internal(is, ep->id);
  else
    phx_icu_reset_endpt_internal(is->phx.nic_mem, ep->icu);
}

void
phx_rx_wake_request(mx_instance_state_t *is, int endpoint, uint32_t host_count)
{
  if (MX_IS_K35SNF(is))
    k35snf_rx_wake_request(is, host_count);
  else
    phx_dbl_rx_wake_request(is->phx.nic_mem, endpoint, host_count);
}
