/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2014-2016 by CSP, Inc.  All rights reserved.                *
 *************************************************************************/

#include "mx_arch.h"
#include "mx_misc.h"
#include "mx_instance.h"
#include "mx_malloc.h"
#include "phx_common.h"
#include "phx_ether_common.h"
#include "mal_stbar.h"
#include "mx_pio.h"

#include <linux/kthread.h>
#if PHX_ENABLE_PTP
#include <linux/net_tstamp.h>
#endif

#ifdef HAVE_CSUM_PARTIAL_COPY
#define ENABLE_RX_CSUM 1
#else
#define ENABLE_RX_CSUM 0
#endif

int phx_ether_napi_weight = 128;

module_param(myri_rx_data_ring_mb, int, S_IRUGO | S_IWUSR);
module_param(myri_rx_desc_ring_mb, int, S_IRUGO | S_IWUSR);

/* Spin in xmit() until there are free descriptors and buffers.
 * Useful for pktgen type testing.
 */
static int myri_busy_wait_tx = 0;
module_param(myri_busy_wait_tx, int, S_IRUGO | S_IWUSR);

/* Use interrupt driven ethernet/kernel endpoint. */
static int myri_intr_ether = 1;
module_param(myri_intr_ether, int, S_IRUGO | S_IWUSR);

/* Entirely for testing/debugging. */
static int myri_ether_worker_force_sleep = 0;
module_param(myri_ether_worker_force_sleep, int, S_IRUGO | S_IWUSR);

static void
phx_fixup_os_stats(struct phx_ether *eth)
{
	eth->arch.stats.rx_dropped = eth->stats_dropped_oversized;
	eth->arch.stats.rx_length_errors = eth->stats_dropped_oversized;
	eth->arch.stats.rx_dropped +=
		(eth->ep_tx_stats ? eth->ep_tx_stats->rx_packets_mac_error : 0);
}

/* compatibility cruft for multiple linux versions */

static struct net_device *
phx_netdev_alloc(char *s, int *err, int num)
{
	struct phx_ether *eth;
	struct net_device *dev;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	char name[IFNAMSIZ];

	eth = mx_kmalloc(sizeof(*eth), MX_MZERO|MX_NOWAIT);
	if (!eth) {
		*err = -ENOMEM;
		return NULL;
	}
	strcpy(name, s);
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,17,0)
	dev = alloc_netdev(sizeof(eth), name, ether_setup);
#else
	dev = alloc_netdev(sizeof(eth), name, NET_NAME_UNKNOWN, ether_setup);
#endif
	if (!dev) {
		mx_kfree(eth);
		*err = -ENOMEM;
                return (dev);
        }
	
	*(struct phx_ether **)(netdev_priv(dev)) = eth;
#else
	eth = mx_kmalloc(sizeof(*eth), MX_MZERO|MX_NOWAIT);
	if (!eth) {
		*err = -ENOMEM;
		return NULL;
	}
	dev = dev_alloc(s,err);
	if (!dev) {
		mx_kfree(eth);
		*err = -ENOMEM;
		return (dev);
	}
	dev->priv = eth;
#endif
	*err = 0;
	return (dev);
}

static void
phx_netdev_free(struct net_device *d)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        void *ptr = phx_netdev_priv(d);
	free_netdev(d);
	mx_kfree(ptr);
#else
	mx_kfree(d->priv);
	kfree(d);
#endif
}

static void
print_tx_stuck(struct phx_ether *eth)
{
	phx_ether_tx_t *tx = &eth->myri_phx(tx);
	phx_tx_desc_t *desc;
	phx_ether_tx_queued_packet_t *info;
	uint32_t nic;
	uint32_t cnt = phx_ep_tx_completion_count(eth->ep_nic_mem);

	printk("TX stuck\n");
	printk("host_idx=%u nic_idx=%u sent_comp_page=%u sent_register=%u\n",
	       tx->desc_idx_host, tx->desc_idx_nic,
	       tx->completion->endpoint[eth->myri_phx(ep)->tx_ep].sent, cnt);

	printk("last_tx_wake_request=%u last_tx_completion=%u\n",
	       eth->debug_last_tx_wake_request, eth->debug_last_tx_completion);

	nic = tx->desc_idx_nic;
	while (tx->desc_idx_host != nic) {
		info = &tx->queued[(nic % tx->desc_count)];
		desc = &info->save_desc;
		printk("desc %u: d_off=%u d_len=%u h_len=%u flags=0x%x\n",
		       nic, desc->data_offset, desc->data_length,
		       desc->header_length, desc->flags);
		nic++;
	}
}

static void
k35snf_print_tx_stuck(struct phx_ether *eth)
{
	k35snf_ether_tx_t *tx = &eth->tx.k35snf;
	uint32_t cnt;
#if 0
	k35snf_tx_desc_t *desc;
	uint32_t nic;
	phx_ether_tx_queued_packet_t *info;
#endif
	cnt = k35snf_tx_compl_count(eth->is);

	printk("TX stuck\n");
	printk("host_idx=%u nic_idx=%u sent_register=%u\n",
	       tx->desc_idx_host, tx->desc_idx_nic, cnt);

	printk("last_tx_completion_packets=%u last_tx_completion_bytes=%u\n",
	       eth->debug_last_tx_completion_packets, eth->debug_last_tx_completion_bytes);
#if 0
/* FIXME - the k35snf does not keep a queued array of pointers to descriptors.
 * The descriptor are all in host memory in the descriptor queue...show them from there.
 */
	nic = tx->desc_idx_nic;
	while (tx->desc_idx_host != nic) {
		info = &tx->queued[(nic % tx->desc_count)];
		desc = &info->save_desc;
		printk("desc %u: d_off=%u d_len=%u h_len=%u flags=0x%x\n",
		       nic, desc->data_offset, desc->data_length,
		       desc->header_length, desc->flags);
		nic++;
	}
#endif
}

static int
k35snf_ether_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct phx_ether *eth = phx_netdev_priv(dev);
	k35snf_ether_tx_t *tx = &eth->tx.k35snf;
	k35snf_tx_desc_t *desc;
	uint32_t data_offset, qlen;
	int desc_used, data_used;
	int i, n_frags;
	phx_ether_pkt_frag_t pkt_frags[MAX_SKB_FRAGS + 1];
#if MX_DEBUG_ETHER
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,5,0)
  struct timespec skb_ts;
#else
  struct __kernel_old_timespec skb_ts;
#endif
  uint64_t skb_ts_64;
#endif
  
	if (ep_enter(&eth->arch, EP_FLAGS_TX_ENTER)) {
		eth->arch.stats_no_ep_tx_dropped++;
		dev_kfree_skb_any(skb);
		return NETDEV_TX_OK;
	}
 
	desc_used = (int)tx->desc_idx_host - (int)tx->desc_idx_nic;
	data_used = (int)tx->data_offset_host - (int)tx->data_offset_nic;
	if (desc_used + 1 >= tx->desc_count ||
	    data_used + MX_MAX_ETHER_MTU >= tx->data_ring_size)
		goto abort_busy;
	/* The data ring has at least 9KB of free space.  No need to check. */

	/* Make a list of <addr, len> */
	pkt_frags[0].vaddr = skb->data;
	pkt_frags[0].len = skb_headlen(skb);
	for (i = 0; i < skb_shinfo(skb)->nr_frags; i++) {
		skb_frag_t *frag = &skb_shinfo(skb)->frags[i];

		pkt_frags[i+1].vaddr = skb_frag_address(frag);
		pkt_frags[i+1].len = skb_frag_size(frag);
	}
	n_frags = skb_shinfo(skb)->nr_frags + 1;

	/* Copy the packet into the data ring */
	qlen = K35SNF_ALIGN_TX_PACKET_LENGTH(skb->len);
	data_offset = tx->data_offset_host % tx->data_ring_size;
	k35snf_ether_tx_copy_pkt(eth, pkt_frags, n_frags, skb->len, data_offset);
	tx->data_offset_host += qlen;

	/* Fill in the descriptor */
	desc = k35snf_ether_next_tx_desc(tx);
	desc->delta_ns = 0;
	desc->tag = 0;
	desc->pktlen = skb->len;
	desc->flags = 0;
	desc->seqnum = 0;
	/* Flush every packet */
	MAL_STBAR();
	*tx->prod_doorbell = 1;
 	MAL_STBAR();

#if PHX_ENABLE_PTP
  skb_tx_timestamp(skb); /* support software timestamps */
#if MX_DEBUG_ETHER
  skb_get_timestampns(skb, &skb_ts);
  skb_ts_64 = (uint64_t) skb_ts.tv_sec * 1000000000 + skb_ts.tv_nsec;
  MX_INFO(("SKB timestamp  = %12" PRId64 "\n", skb_ts_64));
#endif

#endif

	tx->desc_idx_host++;
	eth->arch.stats.tx_packets++;
	eth->arch.stats.tx_bytes += skb->len;
	dev_kfree_skb_any(skb);

	ep_leave(&eth->arch, EP_FLAGS_TX_ENTER);
	return NETDEV_TX_OK;

abort_busy:
	netif_stop_queue(dev);
	eth->arch.stats_tx_stop_queue++;
	ep_leave(&eth->arch, EP_FLAGS_TX_ENTER);
	return NETDEV_TX_BUSY;
}

static int
phx_ether_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct phx_ether *eth = phx_netdev_priv(dev);
	phx_ether_tx_t *tx = &eth->myri_phx(tx);
	phx_tx_desc_t *desc;
	phx_ether_tx_queued_packet_t *info;
	int used, i, n_frags;
	uint32_t hlen, dlen, desc_idx_nic;
        uint32_t dmasend_threshold;
	phx_ether_pkt_frag_t pkt_frags[MAX_SKB_FRAGS + 1];
	unsigned long irq_flags;

	if (myri_disable_eth) {
		dev_kfree_skb_any(skb);
		return NETDEV_TX_OK;
	}

try_again:
	desc = tx->aligned_desc;
	desc_idx_nic = tx->desc_idx_nic;

	/* Flow control: descriptor ring */
	used = (int)tx->desc_idx_host - (int)desc_idx_nic;
	if (used >= tx->desc_count)
		goto abort_busy;
#if PHX_WORKAROUND_TX_WAKE_WRAP_AROUND
	/* When we near the wraparound point, wait till the send completion count
	 * catch up and the queue becomes empty.
	 */
	if (tx->desc_idx_host == 0xffffffff && desc_idx_nic != 0xffffffff) {
		int stopped = 0;
		spin_lock_irqsave(&eth->arch.tx_intr_lock, irq_flags);
		if (tx->desc_idx_nic != 0xffffffff) {
			netif_stop_queue(dev);
			stopped = 1;
		}
		spin_unlock_irqrestore(&eth->arch.tx_intr_lock, irq_flags);
		if (stopped)
			return NETDEV_TX_BUSY;
	}
#endif
	info = &tx->queued[tx->desc_idx_host % tx->desc_count];

	/* Make a list of <addr, len> */
	pkt_frags[0].vaddr = skb->data;
	pkt_frags[0].len = skb_headlen(skb);
	for (i = 0; i < skb_shinfo(skb)->nr_frags; i++) {
		skb_frag_t *frag = &skb_shinfo(skb)->frags[i];

		pkt_frags[i+1].vaddr = skb_frag_address(frag);
		pkt_frags[i+1].len = skb_frag_size(frag);
	}
	n_frags = skb_shinfo(skb)->nr_frags + 1;

#if MX_DEBUG_ETHER
	MX_INFO(("%s: skb_len=%d n_frags=%d\n", __FUNCTION__, skb->len, n_frags));
#endif

	/* Header length in the descriptor, and data length in DMA/PIO buffer.
	 *
	 * FIXME.  An apparent bug in the chip.  The header length cannot be
	 * a multiple of 4.  (4n + 2) seems to be okay.
	 *
	 * 56, 52: No packets sent out.  Need a reboot.
	 * 42, 46, 50, 54: okay
	 * 44: No packets sent out.  Need a reboot.
	 */
	hlen = PHX_TX_DESC_HEADER_LEN - 2; /* 54 */
	if (skb_headlen(skb) < hlen)
		hlen = skb_headlen(skb);
	if (skb->len < hlen)
		hlen = skb->len;
	/* Make hlen = 4n + 2 */
	hlen = ((hlen - 2) & ~0x3) + 2;
	dlen = skb->len - hlen;

	desc->header_length = hlen;
	desc->data_length = dlen;
	desc->data_offset = 0;

	/* Copy header bytes into the descriptor */
	phx_ether_copy_pkthdr(pkt_frags, n_frags, desc->header, hlen);

	/* Address the correct outgoing port */
	desc->flags = 0;
	switch (eth->is->phx.pci_info.func)
	{
		case 0: desc->flags |= PHX_TX_DESC_FLAG_SEND_PORT0; break;
		case 1: desc->flags |= PHX_TX_DESC_FLAG_SEND_PORT1; break;
		case 2: desc->flags |= PHX_TX_DESC_FLAG_SEND_PORT2; break;
		case 3: desc->flags |= PHX_TX_DESC_FLAG_SEND_PORT3; break;
        }

#if PHX_ENABLE_PTP
	/* set the timestamp flag if requested */
	if (eth->want_tx_ts && 
	    skb_shinfo(skb)->tx_flags & SKBTX_HW_TSTAMP) {
		desc->flags |= PHX_TX_DESC_FLAG_LOOPBACK_TIMESTAMP;
		skb_shinfo(skb)->tx_flags |= SKBTX_IN_PROGRESS;
		if (eth->arch.tx_ts_skb != NULL) {
			MX_INFO(("Want to timestamp a packet before previous one is received!\n"));
			MX_INFO(("Freeing previous one!\n"));
			dev_kfree_skb_any(eth->arch.tx_ts_skb);
		}
		eth->arch.tx_ts_skb = skb;
	}
#endif

	dmasend_threshold = phx_get_dmasend_threshold(eth->is);
	/* Flow control and data copy.  DMA or PIO? */
	if (skb->len <= dmasend_threshold) { /* Board defined threshold */
		/* PIO buffer.  The packet may not wrap. */
		if (dlen > 0) {
			if (0 != phx_ether_tx_alloc_piobuf(eth, dlen, &desc->data_offset))
				goto abort_busy;
			/* Copy the packet (minus header bytes in the descriptor)
			 * to the PIO buffer on the NIC.
			 */
			phx_ether_tx_copy_pkt_piobuf(eth, pkt_frags, n_frags, hlen, dlen,
						     desc->data_offset);
		}
		/* Set end_offset regardless of dlen.  dlen may be zero for tiny packets. */
		info->end_offset = tx->pio_offset_host;
	}
	else {
		/* DMA buffer.  */
		desc->flags |= PHX_TX_DESC_FLAG_DMA;
		if (dlen > 0) {
			if (0 != phx_ether_tx_alloc_dmabuf(eth, dlen, &desc->data_offset))
				goto abort_busy;
			/* Copy the packet (minus header bytes in the descriptor)
			 * to the DMA buffer in host memory.
			 */
			phx_ether_tx_copy_pkt_dmabuf(eth, pkt_frags, n_frags, hlen, dlen,
						     desc->data_offset);
		}
		info->end_offset = tx->dma_offset_host;
	}
	MAL_STBAR();

	/* Save meta to clean up later on (send completion) */
	info->flags = desc->flags;

	/* Copy descriptor.  FIXME.  Use one of the piocopy routines. */
#if MX_DEBUG_ETHER
	MX_INFO(("tx_desc: d_off=%u d_len=%u h_len=%u flags=0x%x host_idx=%u\n",
		 desc->data_offset, desc->data_length, desc->header_length, desc->flags,
		 tx->desc_idx_host));
#endif
#if PHX_ENABLE_PTP
	skb_tx_timestamp(skb); /* support software timestamps */
#endif
	phx_ether_pio_tx_desc(eth);
	eth->arch.stats.tx_packets++;
	eth->arch.stats.tx_bytes += skb->len;

	/* We have copied the packet to another buffer.  Free the skb right away. */
#if PHX_ENABLE_PTP
	if (!(eth->want_tx_ts && (skb_shinfo(skb)->tx_flags & SKBTX_HW_TSTAMP))){
		dev_kfree_skb_any(skb);
	}
#else
	dev_kfree_skb_any(skb);
#endif


	/* Save the descriptor for later debugging */
	info->save_desc = *desc;

	tx->desc_idx_host++;
#if PHX_WORKAROUND_TX_WAKE_WRAP_AROUND
	/* Deal with the 0th packet without interrupts.  Once we know the send
	 * completion count has wrapped around, we enable the interrupt.
	 */
	if (tx->desc_idx_host == 1) {
		/* Poll till the 0th packet is completed. */
		int loop_count = 10000;
		uint32_t sent;
		while (loop_count > 0) {
			sent = phx_ep_tx_completion_count(eth->ep_nic_mem);
			if (sent >= 1)
				break;
			udelay(1);
			loop_count--;
		}
		if (loop_count == 0) {
			sent = phx_ep_tx_completion_count(eth->ep_nic_mem);
			printk("0th packet is stuck? sent=%u\n", sent);
			/* No clear way out once we get into this situation. */
		}
		phx_tx_wake_request(eth->ep_nic_mem, 0);
	}
#endif
	return NETDEV_TX_OK;

abort_busy:
	if (myri_busy_wait_tx) {
		uint32_t sent;
		int loop_count = 10000;
		while (loop_count > 0 && (sent = tx->completion->endpoint[eth->myri_phx(ep)->tx_ep].sent) == tx->desc_idx_nic) {
			udelay(1);
			loop_count--;
		}
		if (loop_count > 0) {
			phx_ether_clean_sent_packets(eth);
			goto try_again;
		} else {
			print_tx_stuck(eth);
			dev_kfree_skb_any(skb);
			netif_stop_queue(dev);
			return NETDEV_TX_BUSY;
		}
	}

	/* When we stop the tx queue, we need to be certain that the
	 * descriptor queue is full.  A false positive based on stale
	 * index values can hang the tx queue forever.  Because the
	 * descriptor queue is so shallow (9), the following sequence
	 * sometimes happens.
	 *
	 * 1. xmit(): free descriptor?  No.
	 * 2. tx_done(): clean sent packets.
	 * 3. tx_done(): queue_stopped()?  No.
	 * 4. tx_done(): return with the empty descriptor queue.  No more interrupts.
	 * 5. xmit(): stop_queue(), return TX_BUSY.
	 *
	 * So, synchronize around the stop bit...
	 */
	spin_lock_irqsave(&eth->arch.tx_intr_lock, irq_flags);
	used = tx->desc_idx_nic - desc_idx_nic;
	if (!used) {
		/* NIC index has not changed.  So, the descriptor or
		 * data queue is really full.
		 */
		netif_stop_queue(dev);
	}

	if (((int)tx->desc_idx_host - (int)desc_idx_nic) >= tx->desc_count
	    && tx->desc_idx_nic == tx->desc_idx_host) {
		/* We thought the descriptor queue was full, but in fact it is empty. */
		eth->arch.stats_tx_false_fullness++;
	}
	spin_unlock_irqrestore(&eth->arch.tx_intr_lock, irq_flags);
	if (used) {
		/* We've used stale index values.  So check again. */
		eth->arch.stats_tx_try_again++;
		goto try_again;
	} else
		eth->arch.stats_tx_stop_queue++;
	return NETDEV_TX_BUSY;
}


static inline void
phx_ether_vlan_ip_csum(struct sk_buff *skb, u32 hw_csum)
{
	struct vlan_hdr *vh = (struct vlan_hdr *) (skb->data);

	if ((skb->protocol == htons(ETH_P_8021Q)) &&
	    (vh->h_vlan_encapsulated_proto == htons(ETH_P_IP) ||
	     vh->h_vlan_encapsulated_proto == htons(ETH_P_IPV6))) {
		skb->csum = hw_csum;
		skb->ip_summed = CHECKSUM_COMPLETE;
		if (myri_ether_vlan_csum_fixup)
			skb->csum =
				csum_sub(skb->csum,
					 csum_partial(skb->data,
						      VLAN_HLEN, 0));
	}
}

#if PHX_ENABLE_PTP
static inline int
phx_get_ts_info(struct net_device *netdev, struct ethtool_ts_info *info)
{
       struct phx_ether *phx = phx_netdev_priv(netdev);
       struct mx_instance_state *is = phx->is;

       info->so_timestamping =
	       SOF_TIMESTAMPING_TX_SOFTWARE |
         SOF_TIMESTAMPING_RX_HARDWARE |
	       SOF_TIMESTAMPING_RX_SOFTWARE |
	       SOF_TIMESTAMPING_SOFTWARE |
	       SOF_TIMESTAMPING_RAW_HARDWARE ;


       if (is->phx.pfi->ptp_clock)
               info->phc_index = ptp_clock_index(is->phx.pfi->ptp_clock);
       else
               info->phc_index = -1;

       info->tx_types =
               0;
       info->rx_filters =
               (1 << HWTSTAMP_FILTER_NONE) |
               (1 << HWTSTAMP_FILTER_ALL);


       return 0;
}

#endif

#ifdef HAVE_NEW_NAPI
static int
phx_ether_poll(struct napi_struct *napi, int budget)
{
	struct phx_ether_arch *arch = container_of(napi, struct phx_ether_arch, napi);
	struct phx_ether *eth;
	mx_instance_state_t *is;
	phx_rx_desc_t *rx_desc;
	int work_done = 0;
	uint8_t seqnum;

	eth = container_of(arch, struct phx_ether, arch);
	is = eth->is;

	while (work_done < budget) {
		rx_desc = phx_ether_get_rx_desc_from_ring(eth);
		if ((seqnum = *((volatile uint8_t*)&rx_desc->seqnum)) == eth->rx_desc.seqnum) {
			phx_ether_process_rx_desc(eth, rx_desc);
			work_done++;
		}
		else
			break;
	}
	if (work_done < budget) {
		eth->arch.stats_napi_complete++;

		/* No more receive packets.  Request the wake interrupt again. */
		napi_complete(napi);
		phx_rx_wake_request(eth->is, eth->myri_phx(ep)->rx_ep,
				    eth->rx_desc.desc_idx);
	}
	return (work_done);
}

static int
k35snf_ether_poll(struct napi_struct *napi, int budget)
{
	struct phx_ether_arch *arch = container_of(napi, struct phx_ether_arch, napi);
	struct phx_ether *eth;
	mx_instance_state_t *is;
	k35snf_rx_desc_t *rx_desc;
	int work_done = 0;
	uint8_t seqnum;

	if (ep_enter(arch, EP_FLAGS_POLL_ENTER)) {
		arch->stats_no_ep_napi_complete++;
		napi_complete(napi);
		return work_done;
	}

	eth = container_of(arch, struct phx_ether, arch);
	is = eth->is;

	while (work_done < budget) {
		rx_desc = k35snf_ether_get_rx_desc_from_ring(eth);
		if ((seqnum = *((volatile uint8_t*)&rx_desc->seqnum)) == eth->rx_desc.seqnum) {
			k35snf_ether_process_rx_desc(eth, rx_desc);
			work_done++;
		}
		else
			break;
	}
	if (work_done < budget) {
		eth->arch.stats_napi_complete++;

		/* No more receive packets.  Request the wake interrupt again. */
		napi_complete(napi);
		k35snf_rx_wake_request(eth->is, eth->rx_desc.desc_idx);
	}

	ep_leave(arch, EP_FLAGS_POLL_ENTER);
	return (work_done);
}
#endif

static int
phx_ether_set_pause(struct net_device *dev, int pause)
{
	struct phx_ether *eth = phx_netdev_priv(dev);
	int status;

	status = phx_ether_set_pause_common(eth, pause);
	if (status) {
		MX_WARN(("%s: Failed to set flow control mode\n", dev->name));
		return status;
	}
	eth->pause = pause;
	return 0;
}

static int
phx_ether_close(struct net_device *dev)
{
	struct phx_ether *eth;

	eth = phx_netdev_priv(dev);

	/* MX_WARN(("%s: phx_ether_close\n", eth->arch.dev->name)); */

	/* if the device not running give up */
	if (eth->running != MX_ETH_RUNNING &&
	    eth->running != MX_ETH_OPEN_FAILED)
		return -ENOTTY;

	if (myri_disable_eth) {
		netif_tx_disable(dev);
		eth->running = MX_ETH_STOPPED;
		return 0;
	}

	eth->running = MX_ETH_STOPPING;
	netif_tx_disable(dev);
	if (eth->arch.kthread)
		kthread_stop(eth->arch.kthread);
	if (eth->use_interrupt) {
		/* Disable interrupt */
		phx_icu_disable_interrupt(eth->is, eth->myri_phx(ep)->icu);

		/* We do not get an interrupt from the NIC.  Wait a
		 * bit so we process any outstanding interrupts.
		 */
		msleep_interruptible(500);
	}
	eth->running = MX_ETH_STOPPED;
#ifdef HAVE_NEW_NAPI
	napi_disable(&eth->arch.napi);
#endif
	phx_ether_close_common(eth);
	return 0;
}

#if ENABLE_RX_CSUM
/* This function is a copy of phx_copy_from_page_ring(), except csum_partial_copy_nocheck().
 * Need to refactor.  FIXME
 */
static __wsum
phx_ether_copy_rx_packet(phx_ether_pages_t *pages, uint32_t start_off, uint32_t len, char *dst)
{
	mx_page_pin_t *p;
	uint32_t p_idx, p_off, copy_len;
	__wsum csum = 0;

	/* Start page */
	p_idx = (start_off / PAGE_SIZE) % pages->n_pages;
	p_off = (start_off & (PAGE_SIZE-1));

	/* The data may span across multiple pages.  Visit as many pages as required. */
	while (len > 0) {
		p = &pages->pages[p_idx];
		copy_len = PAGE_SIZE - p_off;
		if (copy_len > len)
			copy_len = len;
		csum = csum_partial_copy_nocheck((char*)(p->va + p_off), dst, copy_len, csum);
		dst += copy_len;
		len -= copy_len;
		p_off = 0;
		p_idx = (p_idx + 1) % pages->n_pages;
	}
	return csum;
}
#endif /* ENABLE_RX_CSUM */

void
phx_ether_receive_packet(struct phx_ether *eth, uint16_t pkt_len,
			 uint32_t ts_seconds, uint32_t ts_nanos)
{
	struct net_device *dev = eth->arch.dev;
	struct sk_buff *skb;

#if MX_DEBUG_ETHER
	MX_INFO(("rx packet. off=%u len=%u\n", eth->rx_data.data_off, pkt_len));
#endif

	/* Allocate skb */
	skb = netdev_alloc_skb(dev, pkt_len);
	if (skb) {
		/* Copy the packet into skb */
#if ENABLE_RX_CSUM
		__wsum csum = phx_ether_copy_rx_packet(&eth->rx_data.ring.gamma,
						       eth->rx_data.data_off,
						       pkt_len, skb->data);
		unsigned char * raw = skb->data;
#else
		phx_copy_from_page_ring(&eth->rx_data.ring.gamma,
					eth->rx_data.data_off,
					pkt_len, skb->data);
#endif
		skb_put(skb, pkt_len);
		skb->ip_summed = CHECKSUM_NONE;
		skb->protocol = eth_type_trans(skb, dev);
		skb->dev = dev;

#if PHX_ENABLE_PTP
		if( eth->want_tx_ts &&
		    raw[6] == eth->current_mac[0] &&
		    raw[7] == eth->current_mac[1] &&
		    raw[8] == eth->current_mac[2] &&
		    raw[9] == eth->current_mac[3] &&
		    raw[10] == eth->current_mac[4] &&
		    raw[11] == eth->current_mac[5]) {
			/* loopback packet so check if it is the one we last
			   sent to be timestamped */
			if (pkt_len == eth->arch.tx_ts_skb->len) {
				struct skb_shared_hwtstamps timestamp;
				/* TODO: more comparison that packet received
				   is the correct loopback packet.  If we do
				   more checks, we may have to check for
				   non-linear skbs and handle them properly */
				timestamp.hwtstamp = ktime_set(ts_seconds,
							       ts_nanos);
				skb_tstamp_tx(eth->arch.tx_ts_skb, &timestamp);
				dev_kfree_skb_any(eth->arch.tx_ts_skb);
				eth->arch.tx_ts_skb = NULL;
				/* We don't want to receive this packet so 
				   free skb and return */
				eth->arch.stats_tx_timestamp++;
				dev_kfree_skb_any(skb);
				return;
			}
		}
#endif
#if ENABLE_RX_CSUM
		if (eth->rx_csum_offload) {
			if ((skb->protocol == htons(ETH_P_IP)) ||
			    (skb->protocol == htons(ETH_P_IPV6))) {
				__wsum hdr_csum = csum_partial(raw, ETH_HLEN, 0);
				skb->csum = csum_sub(csum, hdr_csum);
				skb->ip_summed = CHECKSUM_COMPLETE;
			} else {
				__wsum hdr_csum = csum_partial(raw, ETH_HLEN, 0);
				phx_ether_vlan_ip_csum(skb, csum_sub(csum, hdr_csum));
			}
		}
#endif
		eth->arch.stats.rx_packets++;
		eth->arch.stats.rx_bytes += skb->len;

#if PHX_ENABLE_PTP
		if (eth->want_rx_ts) {
			skb_hwtstamps(skb)->hwtstamp = ktime_set(ts_seconds, ts_nanos);
		}
#endif

#ifdef HAVE_NEW_NAPI
#ifdef HAVE_GRO_RECEIVE
		napi_gro_receive(&eth->arch.napi, skb);
#else
		netif_receive_skb(skb)
#endif
#else
		/* Off to the network stack via net softirq */
		netif_rx_ni(skb);
#endif
	}
}

void
phx_ether_tx_done(mx_instance_state_t *is)
{
	struct phx_ether *eth = is->ether;
	struct net_device *dev = eth->arch.dev;
	phx_ether_tx_t *tx = &eth->myri_phx(tx);
	unsigned long flags;
	int wake = 0;

	phx_ether_clean_sent_packets(eth);

	/* If we are in the irq handler, irqsave is not necessary.
	 * Do not bother with optimizing yet.  The fact that we are using
	 * a spin lock is plenty terrible.
	 */
	spin_lock_irqsave(&eth->arch.tx_intr_lock, flags);
	if (netif_queue_stopped(dev)) {
		int used = (int)tx->desc_idx_host - (int)tx->desc_idx_nic;
		/* Restart as soon as we have any free descriptor.
		 * The descriptor queue is only 9 packets, and the default
		 * completion threshold is 8 packets...  FIXME.
		 */
		if (used < tx->desc_count) {
			/* Restart the send queue */
			wake = 1;
		}
	}
	spin_unlock_irqrestore(&eth->arch.tx_intr_lock, flags);
	if (wake) {
		netif_wake_queue(dev);
		eth->arch.stats_tx_wake_queue++;
	}
	if (eth->use_interrupt) {
		eth->debug_last_tx_wake_request = tx->desc_idx_nic;
#if PHX_WORKAROUND_TX_WAKE_WRAP_AROUND
		/* xmit() is waiting for us to catch up.  Now that we have
		 * caught up and the queue is empty.  Do not write to
		 * the wake register as we cannot be sure whether the send completion
		 * count has already wrapped around.  xmit() polls for the completion
		 * for the first packet and then requests the wake interrupt.
		 */
		if (tx->desc_idx_nic != 0xffffffff)
			phx_tx_wake_request(eth->ep_nic_mem, tx->desc_idx_nic);
#else
		phx_tx_wake_request(eth->ep_nic_mem, tx->desc_idx_nic);
#endif
	}
}

static int
phx_eth_worker(void *data)
{
	struct phx_ether *eth = data;
	struct net_device *dev = eth->arch.dev;
	int idle;

	MX_WARN(("%s: the worker thread is running\n", dev->name));
	while (!kthread_should_stop()) {
		/* Poll the receive descriptor ring */
		idle = phx_ether_rx_done(eth->is) > 0 ? 0 : 1;

		/* Send completions. */
		if (!myri_busy_wait_tx)
			phx_ether_tx_done(eth->is);

		/* FIXME.  Link up/down */
		if (myri_ether_worker_force_sleep > 0) {
			msleep_interruptible(myri_ether_worker_force_sleep);
		} else if (idle) {
			/* FIXME.  Wait. */
			msleep_interruptible(1);
		}
	}
	MX_WARN(("%s: the worker thread is stopping\n", dev->name));
	return 0;
}

static int
phx_ether_open(struct net_device *dev)
{
	int error;
	struct phx_ether *eth;

	if (MX_DMA_DBG)
		return -ENOSYS; /* ether code does not update dma bitmap */
	eth = phx_netdev_priv(dev);

	if (eth->running != MX_ETH_STOPPED)
		return 0;
	/* If ethernet endpoint is disabled, pretend that everything suceeds.
	 * Do not call any of the open/alloc/start functions.
	 */
	if (myri_disable_eth) {
		MX_WARN(("%s: ethernet endpoint is disabled\n", dev->name));
		eth->running = MX_ETH_RUNNING;
		netif_wake_queue(dev);
		return 0;
	}

	error = phx_ether_open_common(eth->is, myri_intr_ether, dev->mtu,
				      myri_rx_data_ring_mb, myri_rx_desc_ring_mb);
	if (error) {
		MX_WARN(("%s: ether_open_common failed\n", dev->name));
		goto abort_with_nothing;
	}

#ifdef HAVE_NEW_NAPI
	napi_enable(&eth->arch.napi); /* must happen before any irq */
#endif
	/* flow control */
	phx_ether_set_pause(dev, eth->pause);

	spin_lock_init(&eth->arch.tx_intr_lock);
	if (!eth->use_interrupt) {
		eth->arch.kthread = kthread_create(phx_eth_worker, eth, "myri_ether");
		if (IS_ERR(eth->arch.kthread)) {
			MX_WARN(("%s: unable to create a worker thread\n", dev->name));
			goto abort_with_open;
		}
	}
	if (phx_ether_start_common(eth)) {
		MX_WARN(("%s: unable to enable send/receive on the NIC\n", dev->name));
		goto abort_with_open;
	}

	eth->running = MX_ETH_RUNNING;
	netif_wake_queue(dev);

	phx_ether_link_change_notify(eth->is);
	if (!eth->use_interrupt)
		wake_up_process(eth->arch.kthread);
	else {
		/* Request TX and RX interrupts. */
		phx_rx_wake_request(eth->is,
					eth->myri_phx(ep)->rx_ep,
					eth->rx_desc.desc_idx);
#if !PHX_WORKAROUND_TX_WAKE_WRAP_AROUND
		phx_tx_wake_request(eth->ep_nic_mem, 
				    eth->myri_phx(tx).desc_idx_nic);
#endif
		phx_icu_enable_interrupt(eth->is, eth->myri_phx(ep)->icu);
	}
	return 0;

abort_with_open:
#ifdef HAVE_NEW_NAPI
	napi_disable(&eth->arch.napi);
#endif
	eth->running = MX_ETH_OPEN_FAILED;
	phx_ether_close(dev);

abort_with_nothing:
	eth->running = MX_ETH_STOPPED;
	return error;
}

void
k35snf_ether_receive_packet(struct phx_ether *eth, uint16_t pkt_len, uint64_t ts_nanos)
{
	struct net_device *dev = eth->arch.dev;
	struct sk_buff *skb;

#if MX_DEBUG_ETHER
	MX_INFO(("rx packet. off=%u len=%u\n", eth->rx_data.data_off, pkt_len));
#endif

	/* Allocate skb */
	skb = netdev_alloc_skb(dev, pkt_len);
	if (skb) {
		/* Copy the packet into skb */
#if ENABLE_RX_CSUM
		__wsum csum = phx_ether_copy_rx_packet(&eth->rx_data.ring.gamma,
						       eth->rx_data.data_off,
						       pkt_len, skb->data);
		unsigned char * raw = skb->data;
#else
		phx_copy_from_page_ring(&eth->rx_data.ring.gamma,
					eth->rx_data.data_off,
					pkt_len, skb->data);
#endif
		skb_put(skb, pkt_len);
		skb->ip_summed = CHECKSUM_NONE;
		skb->protocol = eth_type_trans(skb, dev);
		skb->dev = dev;
#if ENABLE_RX_CSUM
		if (eth->rx_csum_offload) {
			if ((skb->protocol == htons(ETH_P_IP)) ||
			    (skb->protocol == htons(ETH_P_IPV6))) {
				__wsum hdr_csum = csum_partial(raw, ETH_HLEN, 0);
				skb->csum = csum_sub(csum, hdr_csum);
				skb->ip_summed = CHECKSUM_COMPLETE;
			} else {
				__wsum hdr_csum = csum_partial(raw, ETH_HLEN, 0);
				phx_ether_vlan_ip_csum(skb, csum_sub(csum, hdr_csum));
			}
		}
#endif
		eth->arch.stats.rx_packets++;
		eth->arch.stats.rx_bytes += skb->len;
#ifdef HAVE_NEW_NAPI
#ifdef HAVE_GRO_RECEIVE
		napi_gro_receive(&eth->arch.napi, skb);
#else
		netif_receive_skb(skb)
#endif
#else
		/* Off to the network stack via net softirq */
		netif_rx_ni(skb);
#endif
	}
}

void
k35snf_ether_tx_done(mx_instance_state_t *is)
{
	struct phx_ether *eth = is->ether;
	struct net_device *dev = eth->arch.dev;
	k35snf_ether_tx_t *tx = &eth->tx.k35snf;
	int wake = 0;

	k35snf_ether_clean_sent_packets(eth);
	if (netif_queue_stopped(dev)) {
		int used = (int)tx->desc_idx_host - (int)tx->desc_idx_nic;
		if (used + 128 < tx->desc_count) {
			/* Restart the send queue */
			wake = 1;
		}
	}
	if (wake) {
		netif_wake_queue(dev);
		eth->arch.stats_tx_wake_queue++;
	}
	/* As long as threshold is << 128 above, TX will not hang.
	 * The data ring is at least 2MB.
	 */
	k35snf_tx_compl_count_ack(eth->is, 64);
}

#if !defined (RHEL_RELEASE_CODE)
void
k35snf_ether_timeout(struct net_device *dev)
{
        struct phx_ether *eth = phx_netdev_priv(dev);

        k35snf_print_tx_stuck(eth);
        if (dev->watchdog_timeo < HZ * 3600)
                dev->watchdog_timeo = 2 * dev->watchdog_timeo;
}
#else
#if (RHEL_RELEASE_CODE < RHEL_RELEASE_VERSION(8,2))
void
k35snf_ether_timeout(struct net_device *dev)
{
        struct phx_ether *eth = phx_netdev_priv(dev);

        k35snf_print_tx_stuck(eth);
        if (dev->watchdog_timeo < HZ * 3600)
                dev->watchdog_timeo = 2 * dev->watchdog_timeo;
}
#else
void
k35snf_ether_timeout(struct net_device *dev, unsigned int txqueue)
{
        struct phx_ether *eth = phx_netdev_priv(dev);

        k35snf_print_tx_stuck(eth);
        if (dev->watchdog_timeo < HZ * 3600)
                dev->watchdog_timeo = 2 * dev->watchdog_timeo;
}
#endif
#endif

static int
k35snf_ether_close(struct net_device *dev)
{
	struct phx_ether *eth;
	struct phx_ether_arch *arch;

	eth = phx_netdev_priv(dev);
	arch = &eth->arch;

	MX_INFO(("%s: k35snf_ether_close\n", arch->dev->name));

	/* if the device not running give up */
	if (eth->running != MX_ETH_RUNNING &&
	    eth->running != MX_ETH_OPEN_FAILED)
		return -ENOTTY;
	eth->running = MX_ETH_STOPPING;
	netif_tx_disable(dev);

	/* Wait till removing completes and then enter CLOSE */
	ep_enter_wait(arch,
		      (EP_FLAGS_REMOVE_ENTER | EP_FLAGS_RETURN_ENTER),
		      EP_FLAGS_CLOSE_ENTER);
	if (!eth->ep.k35snf.stolen) {
		phx_icu_disable_interrupt(eth->is, eth->ep.k35snf.rx->icu);
		phx_icu_disable_interrupt(eth->is, eth->ep.k35snf.tx->icu);
		/* We do not get an interrupt from the NIC.  Wait a
		 * bit so we process any outstanding interrupts.
		 */
		msleep_interruptible(500);
	}

	eth->running = MX_ETH_STOPPED;
#ifdef HAVE_NEW_NAPI
	napi_disable(&eth->arch.napi);
#endif
	k35snf_ether_close_common(eth);
	ep_leave(arch, EP_FLAGS_CLOSE_ENTER);
	return 0;
}

static int
k35snf_ether_open(struct net_device *dev)
{
	int error = 0;
	struct phx_ether *eth;

	eth = phx_netdev_priv(dev);
	if (eth->running != MX_ETH_STOPPED)
		return 0;

	ep_enter_wait(&eth->arch,
		      (EP_FLAGS_REMOVE_ENTER |
		       EP_FLAGS_RETURN_ENTER),
		      EP_FLAGS_OPEN_ENTER);

	error = k35snf_ether_open_common(eth->is, dev->mtu,
					 myri_rx_data_ring_mb, myri_rx_desc_ring_mb);
	if (error)
		goto abort_with_nothing;

#ifdef HAVE_NEW_NAPI
	napi_enable(&eth->arch.napi); /* must happen before any irq */
#endif
	spin_lock_init(&eth->arch.tx_intr_lock);
	if (k35snf_ether_start_common(eth)) {
		MX_WARN(("%s: unable to enable send/receive on the NIC\n", dev->name));
		goto abort_with_open;
	}
	if (eth->ep.k35snf.stolen) {
		ep_lock(&eth->arch);
		ep_flags_set(&eth->arch, EP_FLAGS_REMOVED);
		ep_unlock(&eth->arch);
	}

	eth->running = MX_ETH_RUNNING;
	netif_wake_queue(dev);

	TAILQ_INIT(&eth->ether_mc_filter_list);

	phx_ether_link_change_notify(eth->is);

	/* Request TX and RX interrupts. */
	if (!eth->ep.k35snf.stolen) {
		k35snf_rx_wake_request(eth->is, eth->rx_desc.desc_idx);
		k35snf_tx_compl_count_ack(eth->is, 1);
		phx_icu_enable_interrupt(eth->is, eth->ep.k35snf.rx->icu);
		phx_icu_enable_interrupt(eth->is, eth->ep.k35snf.tx->icu);
	}
	ep_leave(&eth->arch, EP_FLAGS_OPEN_ENTER);
	return 0;

abort_with_open:
#ifdef HAVE_NEW_NAPI
	napi_disable(&eth->arch.napi);
#endif
	eth->running = MX_ETH_OPEN_FAILED;

abort_with_nothing:
	eth->running = MX_ETH_STOPPED;
	ep_leave(&eth->arch, EP_FLAGS_OPEN_ENTER);
	return error;
}

static int
phx_ether_change_mtu (struct net_device *dev, int new_mtu)
{
	struct phx_ether *eth = phx_netdev_priv(dev);

	if ((new_mtu < 68) || (ETH_HLEN + new_mtu > MX_MAX_ETHER_MTU)) {
		MX_PRINT(("%s: new mtu (%d) is not valid\n",
			  dev->name, new_mtu));
		return -EINVAL;
	}
	MX_INFO(("%s: changing mtu from %d to %d\n",
		 dev->name, dev->mtu, new_mtu));
	/* This driver has nothing that depends on the MTU.
	 * All we need is setting the MAC's frame length.
	 */
	phx_ether_set_mtu(eth, new_mtu);
	dev->mtu = new_mtu;
	return 0;
}

static struct net_device_stats *
phx_ether_get_stats(struct net_device *dev)
{
	struct phx_ether *eth = phx_netdev_priv(dev);
	phx_fixup_os_stats(eth);
	return &eth->arch.stats;
}

#if !defined (RHEL_RELEASE_CODE)
void
phx_ether_timeout(struct net_device *dev)
{
	struct phx_ether *eth = phx_netdev_priv(dev);
	mx_instance_state_t *is;

	is = eth->is;
        MX_WARN(("%s %d: Phoenix: unimpl\n", __FUNCTION__, __LINE__));
	print_tx_stuck(eth);
	if (dev->watchdog_timeo < HZ * 3600)
		dev->watchdog_timeo = 2 * dev->watchdog_timeo;
}
#else
#if (RHEL_RELEASE_CODE < RHEL_RELEASE_VERSION(8,2))
void
phx_ether_timeout(struct net_device *dev)
{
	struct phx_ether *eth = phx_netdev_priv(dev);
	mx_instance_state_t *is;

	is = eth->is;
        MX_WARN(("%s %d: Phoenix: unimpl\n", __FUNCTION__, __LINE__));
	print_tx_stuck(eth);
	if (dev->watchdog_timeo < HZ * 3600)
		dev->watchdog_timeo = 2 * dev->watchdog_timeo;
}
#else
void
phx_ether_timeout(struct net_device *dev, unsigned int txqueue)
{
        struct phx_ether *eth = phx_netdev_priv(dev);
        mx_instance_state_t *is;

        is = eth->is;
	MX_WARN(("%s %d: Phoenix: unimpl, TxQueue %d\n", __FUNCTION__, __LINE__, txqueue));
        print_tx_stuck(eth);
        if (dev->watchdog_timeo < HZ * 3600)
                dev->watchdog_timeo = 2 * dev->watchdog_timeo;
}
#endif
#endif

int
phx_ether_set_mac_address (struct net_device *dev, void *addr)
{
	struct sockaddr *sa = (struct sockaddr *) addr;
	struct phx_ether *eth = phx_netdev_priv(dev);

	/* change the dev structure */
	bcopy(sa->sa_data, dev->dev_addr, 6);
	phx_ether_set_mac_address_common(eth, dev->dev_addr);
	return 0;
}

#if LINUX_XX < 50
static int
phx_ether_get_settings(struct net_device *netdev, struct ethtool_cmd *cmd)
{
	/* struct phx_ether *eth = phx_netdev_priv(netdev); */
	cmd->autoneg = AUTONEG_DISABLE;
	cmd->speed = SPEED_10000;
	cmd->duplex = DUPLEX_FULL;
	return 0;
}
#else
static int
phx_ether_get_link_ksettings(struct net_device *netdev,
                            struct ethtool_link_ksettings *cmd)
{
	return 0;
}
#endif

static void
phx_ether_get_drvinfo(struct net_device *netdev,
		     struct ethtool_drvinfo *info)
{
	struct phx_ether *eth = phx_netdev_priv(netdev);
	mx_instance_state_t *is = eth->is;
	
	sprintf(info->driver, "%s ethernet", MYRI_DRIVER_STR);
	sprintf(info->version, "0x%x", MYRI_DRIVER_API_MAGIC);
	strlcpy(info->bus_info, pci_name(is->arch.pci_dev), sizeof (info->bus_info));
}

int
phx_ether_get_coalesce(struct net_device *netdev,
		     struct ethtool_coalesce *coal)
{
	struct phx_ether *eth = phx_netdev_priv(netdev);
	coal->rx_coalesce_usecs = eth->is->intr_coal_usec;
	return 0;
}

int
phx_ether_set_coalesce(struct net_device *netdev,
		     struct ethtool_coalesce *coal)
{
	struct phx_ether *eth = phx_netdev_priv(netdev);
	(void)eth->is->board_ops.set_intr_coal(eth->is, coal->rx_coalesce_usecs);
	return 0;
}

static void
phx_ether_get_pauseparam(struct net_device *netdev,
			struct ethtool_pauseparam *pause)
{
	struct phx_ether *eth = phx_netdev_priv(netdev);

	pause->autoneg = 0;
	pause->rx_pause = eth->pause;
	pause->tx_pause = eth->pause;
}

static int
phx_ether_set_pauseparam(struct net_device *netdev,
			struct ethtool_pauseparam *pause)
{
	struct phx_ether *eth = phx_netdev_priv(netdev);

	if (pause->tx_pause != eth->pause)
		return phx_ether_set_pause(netdev, pause->tx_pause);
	if (pause->rx_pause != eth->pause)
		return phx_ether_set_pause(netdev, pause->rx_pause);
	if (pause->autoneg != 0)
		return -EINVAL;
	return 0;
}


#if PHX_ENABLE_PTP
static int phx_hwtstamp_set(struct net_device *dev, struct ifreq *ifr){
	
	struct hwtstamp_config tsconf;
	struct phx_ether *eth = phx_netdev_priv(dev);

	if (copy_from_user(&tsconf, ifr->ifr_data, sizeof(tsconf))) {
		return -EFAULT;
	}

	if (tsconf.flags) {
		/* reserved for future use */
		return -EINVAL;
	}

	switch (tsconf.tx_type) {
	case HWTSTAMP_TX_ONESTEP_SYNC:
		/* we don't support onestep mode */
		return -ERANGE;
		break;
	case HWTSTAMP_TX_ON:
    eth->want_tx_ts = 0;
    return -EOPNOTSUPP;
		break;
	case HWTSTAMP_TX_OFF:
		eth->want_tx_ts = 0;
		break;
	default:
		return -ERANGE;
	}

	switch (tsconf.rx_filter) {
	case HWTSTAMP_FILTER_NONE:
		eth->want_rx_ts = 0;
		break;
	case HWTSTAMP_FILTER_ALL:
	default:
		eth->want_rx_ts = 1;
		tsconf.rx_filter = HWTSTAMP_FILTER_ALL;
		break;
	}

	if (copy_to_user(ifr->ifr_data, &tsconf, sizeof(tsconf))) {
		return -EFAULT;
	} else {
		return 0;
	}
}

#ifdef HAVE_SIOCGHWTSTAMP
static int phx_hwtstamp_get(struct net_device *dev, struct ifreq *ifr){

       //struct phx_ether *eth = phx_netdev_priv(dev);
       struct hwtstamp_config tsconf;
       struct phx_ether *eth = phx_netdev_priv(dev);

       tsconf.flags = 0;
       if (eth->want_tx_ts) {
         /* SNF does not support TX timestamps. If we get here, 
          * we should make sure it's still set to off. */      
	       tsconf.tx_type = HWTSTAMP_TX_OFF;
       } else {
	       tsconf.tx_type = HWTSTAMP_TX_OFF;
       }

       if (eth->want_rx_ts) {
	       tsconf.rx_filter = HWTSTAMP_FILTER_ALL;
       } else {
	       tsconf.rx_filter = HWTSTAMP_FILTER_NONE;
       }

       return copy_to_user(ifr->ifr_data, &tsconf, sizeof(tsconf)) ?
           -EFAULT:0;
}
#endif
#endif /* PHX_ENABLE_PTP */

static inline int
phx_ether_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
       int error = -EOPNOTSUPP;

       switch (cmd) {

#if PHX_ENABLE_PTP
       case SIOCSHWTSTAMP:
               error = phx_hwtstamp_set(dev, ifr);
               break;
#ifdef HAVE_SIOCGHWTSTAMP
       case SIOCGHWTSTAMP:
               error = phx_hwtstamp_get(dev, ifr);
               break;
#endif
#endif
       /* put link-state checking here */

       default: 
               break;
       }

       return error;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,3,0)

static u32
phx_ether_get_rx_csum(struct net_device *netdev)
{
        struct phx_ether *eth = phx_netdev_priv(netdev);
	return eth->rx_csum_offload ? 1 : 0;
}

static int
phx_ether_set_rx_csum(struct net_device *netdev, u32 csum_enabled)
{
        struct phx_ether *eth = phx_netdev_priv(netdev);
	eth->rx_csum_offload = csum_enabled ? 1 : 0;
	return 0;
}

static u32
phx_ether_get_tx_csum(struct net_device *netdev)
{
	if ((netdev->features & NETIF_F_IP_CSUM) != 0)
		return 1;
	else
		return 0;
}

static int
phx_ether_set_tx_csum(struct net_device *netdev, u32 csum_enabled)
{
	if (csum_enabled)
		netdev->features |= NETIF_F_IP_CSUM;
	else
		netdev->features &= ~NETIF_F_IP_CSUM;
	return 0;
}

#else /* LINUX_VERSION_CODE < KERNEL_VERSION(3,3,0) */

static int phx_ether_set_features(struct net_device *netdev,
				 netdev_features_t features)
{
        struct phx_ether *eth = phx_netdev_priv(netdev);
	if (features & NETIF_F_RXCSUM)
		eth->rx_csum_offload = 1;
	else
		eth->rx_csum_offload = 0;
	return 0;
}
		
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(3,3,0) */

static const char mx_eth_gstrings[][ETH_GSTRING_LEN] = {
	"rx_packets", "tx_packets", "rx_bytes", "tx_bytes", "rx_errors",
	"tx_errors", "rx_dropped", "tx_dropped", "multicast", "collisions",
	"rx_length_errors", "rx_over_errors", "rx_crc_errors",
	"rx_frame_errors", "rx_fifo_errors", "rx_missed_errors",
	"tx_aborted_errors", "tx_carrier_errors", "tx_fifo_errors",
	"tx_heartbeat_errors", "tx_window_errors",
	/* device-specific stats */
	"link_up", "tx_false_fullness", "tx_try_again",
	"tx_stop_queue", "tx_wake_queue", "napi_complete",
	"tx_timestamp", "no_ep_tx_wake_queue", "no_ep_tx_dropped",
	"no_ep_napi_complete"
};

#define MX_ETH_GSTRINGS_LEN  sizeof(mx_eth_gstrings) / ETH_GSTRING_LEN
#define MX_ETH_NET_STATS_LEN 21

static void
phx_ether_get_strings(struct net_device *netdev, u32 stringset, u8 *data)
{
	switch (stringset) {
	case ETH_SS_STATS:
		memcpy(data, *mx_eth_gstrings,
		       sizeof(mx_eth_gstrings));
		break;
	}
}

#ifndef HAVE_SSET_COUNT
static int
phx_ether_get_stats_count(struct net_device *netdev)
{
	return MX_ETH_GSTRINGS_LEN;
}
#else
static int
phx_ether_get_sset_count(struct net_device *netdev, int sset)
{
	switch (sset) {
	case ETH_SS_STATS:
		return MX_ETH_GSTRINGS_LEN;
	default:
		return -EOPNOTSUPP;
	}
}
#endif /* HAVE_SSET_COUNT */

static void
phx_ether_get_ethtool_stats(struct net_device *netdev,
			   struct ethtool_stats *stats, u64 *data)
{
	struct phx_ether *eth = phx_netdev_priv(netdev);
	mx_instance_state_t *is = eth->is;

	int i;
	phx_fixup_os_stats(eth);
	for (i = 0; i < MX_ETH_NET_STATS_LEN; i++)
		data[i] =  ((unsigned long *) &eth->arch.stats)[i];
	data[i++] = is->link_state ? 1:0;
	data[i++] = eth->arch.stats_tx_false_fullness;
	data[i++] = eth->arch.stats_tx_try_again;
	data[i++] = eth->arch.stats_tx_stop_queue;
	data[i++] = eth->arch.stats_tx_wake_queue;
	data[i++] = eth->arch.stats_napi_complete;
	data[i++] = eth->arch.stats_tx_timestamp;
	data[i++] = eth->arch.stats_no_ep_tx_wake_queue;
	data[i++] = eth->arch.stats_no_ep_tx_dropped;
	data[i++] = eth->arch.stats_no_ep_napi_complete;
}

static MX_ETHTOOL_OPS_TYPE mx_ethtool_ops = {
#if LINUX_XX < 50
	.get_settings 			= phx_ether_get_settings,
#else
	.get_link_ksettings             = phx_ether_get_link_ksettings,
#endif
	.get_drvinfo			= phx_ether_get_drvinfo,
	.get_coalesce			= phx_ether_get_coalesce,
	.set_coalesce			= phx_ether_set_coalesce,
	.get_pauseparam			= phx_ether_get_pauseparam,
	.set_pauseparam			= phx_ether_set_pauseparam,
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,3,0)
	.get_rx_csum			= phx_ether_get_rx_csum,
	.set_rx_csum			= phx_ether_set_rx_csum,
	/* get_tx_csum, get_sg and get_tso are set by default since 2.6.24 */
	.get_tx_csum			= phx_ether_get_tx_csum,
	.set_tx_csum			= phx_ether_set_tx_csum,
	.get_sg				= ethtool_op_get_sg,
	.get_tso			= ethtool_op_get_tso,
	.set_sg				= ethtool_op_set_sg,
#endif
	.get_link			= ethtool_op_get_link,
	.get_strings			= phx_ether_get_strings,
#ifdef HAVE_SSET_COUNT
	.get_sset_count			= phx_ether_get_sset_count,
#else
	.get_stats_count		= phx_ether_get_stats_count,
#endif
	.get_ethtool_stats		= phx_ether_get_ethtool_stats,
#if LINUX_VERSION_CODE > KERNEL_VERSION(3,5,0)
#if PHX_ENABLE_PTP
       .get_ts_info                    = phx_get_ts_info,
#endif
#endif

#ifdef NOT_YET
#ifdef HAVE_SET_PHYS_ID
        .set_phys_id                    = phx_ether_set_phys_id,
#else
	.phys_id		        = phx_ether_phys_id,
#endif
#endif
};

static int
phx_ether_init(struct net_device *dev)
{
	return 0;
}

static void
phx_ether_set_multicast_list(struct net_device *dev)
{
	struct phx_ether *eth = phx_netdev_priv(dev);
	phx_ether_set_promisc_common(eth, !!(dev->flags & IFF_PROMISC));

	/* This hack is no longer used, to match the Windows behavior. */
}

void
phx_ether_link_change_notify(mx_instance_state_t *is)
{
	if (is->ether == NULL || is->ether->running != MX_ETH_RUNNING)
		return;
	if (is->link_state)
		netif_carrier_on(is->ether->arch.dev);
	else
		netif_carrier_off(is->ether->arch.dev);
}

#ifdef HAVE_NET_DEVICE_OPS
static const struct net_device_ops phx_netdev_ops = {
	.ndo_init		= phx_ether_init,
	.ndo_open		= phx_ether_open,
	.ndo_stop		= phx_ether_close,
	.ndo_start_xmit		= phx_ether_xmit,
	.ndo_get_stats		= phx_ether_get_stats,
	.ndo_validate_addr	= eth_validate_addr,
#ifndef RHEL_RELEASE_CODE                     /* handle ubuntu case */
	.ndo_change_mtu		= phx_ether_change_mtu,
#else                                         /* handle CentOS case */
#if RHEL_RELEASE_CODE < RHEL_RELEASE_VERSION(7, 4)
        .ndo_change_mtu 	= phx_ether_change_mtu,
#else
#if RHEL_RELEASE_CODE <= RHEL_RELEASE_VERSION(7, 9)
        .extended.ndo_change_mtu = phx_ether_change_mtu,
#else
        .ndo_change_mtu         = phx_ether_change_mtu,
#endif
#endif
#endif /* end if not defined */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,2,0)
	.ndo_set_multicast_list	= phx_ether_set_multicast_list,
#else
	.ndo_set_rx_mode	= phx_ether_set_multicast_list,
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,3,0)
	.ndo_set_features       = phx_ether_set_features,
#endif
	.ndo_set_mac_address	= phx_ether_set_mac_address,
        .ndo_do_ioctl           = phx_ether_ioctl,

	.ndo_tx_timeout		= phx_ether_timeout,
};

static const struct net_device_ops k35snf_netdev_ops = {
	.ndo_init		= phx_ether_init,
	.ndo_open		= k35snf_ether_open,
	.ndo_stop		= k35snf_ether_close,
	.ndo_start_xmit		= k35snf_ether_xmit,
	.ndo_get_stats		= phx_ether_get_stats,
	.ndo_validate_addr	= eth_validate_addr,
#ifndef RHEL_RELEASE_CODE                     /* handle ubuntu case */
	.ndo_change_mtu		= phx_ether_change_mtu,
#else                                         /* handle CentOS case */
#if RHEL_RELEASE_CODE < RHEL_RELEASE_VERSION(7, 4)
        .ndo_change_mtu 	= phx_ether_change_mtu,
#else
#if RHEL_RELEASE_CODE <= RHEL_RELEASE_VERSION(7, 9)
        .extended.ndo_change_mtu = phx_ether_change_mtu,
#else
        .ndo_change_mtu         = phx_ether_change_mtu,
#endif
#endif
#endif /* end if not defined */	
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,2,0)
	.ndo_set_multicast_list	= phx_ether_set_multicast_list,
#else
	.ndo_set_rx_mode	= phx_ether_set_multicast_list,
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,3,0)
	.ndo_set_features       = phx_ether_set_features,
#endif
	.ndo_set_mac_address	= phx_ether_set_mac_address,
        .ndo_do_ioctl           = phx_ether_ioctl,

	.ndo_tx_timeout		= k35snf_ether_timeout,
};
#endif

int 
phx_ether_attach(mx_instance_state_t *is)
{
	struct net_device *dev;
	struct phx_ether *eth;
	int err, i;

	dev = phx_netdev_alloc("eth%d", &err, is->id);
	if (!dev) {
		MX_WARN(("%s: Alloc netdev failed\n", is->is_name));
		return err;
	}
	ether_setup(dev);
	eth = phx_netdev_priv(dev);
	eth->rx_csum_offload = ENABLE_RX_CSUM;
	eth->pause = myri_ether_pause;
	/* setup all the pointers.. */
	eth->board_type = is->board_type;
	eth->is = is;
	eth->arch.dev = dev;
#if PHX_ENABLE_PTP
	eth->arch.tx_ts_skb = NULL;
#endif
	eth->want_tx_ts = 0;
	eth->want_rx_ts = 0;

	TAILQ_INIT(&eth->ether_mc_filter_list);

	MAL_STBAR();
	is->ether = eth;

	phx_ether_link_change_notify(is);
	dev->mtu = MX_DEFAULT_ETHER_MTU - ETH_HLEN;
#ifdef HAVE_NET_DEVICE_OPS
        if (ETHER_K35SNF(eth))
                dev->netdev_ops = &k35snf_netdev_ops;
        else
                dev->netdev_ops = &phx_netdev_ops;
#else
        if (ETHER_K35SNF(eth)) {
                dev->open = k35snf_ether_open;
                dev->stop = k35snf_ether_close;
                dev->hard_start_xmit = k35snf_ether_xmit;
                dev->tx_timeout = k35snf_ether_timeout;
        } else {
                dev->open = phx_ether_open;
                dev->stop = phx_ether_close;
                dev->hard_start_xmit = phx_ether_xmit;
                dev->tx_timeout = phx_ether_timeout;
        }
        dev->get_stats = phx_ether_get_stats;
        dev->change_mtu = phx_ether_change_mtu;
        dev->set_multicast_list = phx_ether_set_multicast_list;
        dev->set_mac_address = phx_ether_set_mac_address;
        dev->init = phx_ether_init;
#endif
	dev->base_addr = pci_resource_start(is->arch.pci_dev, 0);
	/* We only use MSI-X.  Is this correct?  FIXME */
	/* dev->irq = is->arch.irq; */
	dev->watchdog_timeo = HZ * 2;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,16,0)
        SET_ETHTOOL_OPS(dev, &mx_ethtool_ops);
#else   
        netdev_set_default_ethtool_ops(dev, &mx_ethtool_ops);
#endif
	for (i = 0; i < 6; i++) {
		eth->current_mac[i] = dev->dev_addr[i]= is->mac_addr[i];
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
		dev->perm_addr[i]= is->mac_addr[i];
#endif
	}

	dev->features = NETIF_F_HIGHDMA;
#ifdef HAVE_GRO_RECEIVE
	dev->features |= NETIF_F_GRO;
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,3,0)
	dev->features |= NETIF_F_RXCSUM;
	dev->hw_features = dev->features; 
#endif

	SET_NETDEV_DEV(dev, &is->arch.pci_dev->dev);
#ifdef HAVE_NEW_NAPI
	if (ETHER_K35SNF(eth))
		netif_napi_add(dev, &eth->arch.napi, k35snf_ether_poll, phx_ether_napi_weight);
	else
		netif_napi_add(dev, &eth->arch.napi, phx_ether_poll, phx_ether_napi_weight);
#endif

	if ((err = register_netdev(dev))) {
		if (err != 0)
			MX_WARN(("%s:register_netdev failed!! (err = %d)\n", dev->name, err));
		else {
			MX_WARN(("%s:fake page alloc failed!!\n", dev->name));
			err = -ENOMEM;
		}
#ifdef HAVE_NEW_NAPI
		netif_napi_del(&eth->arch.napi);
#endif
		phx_netdev_free(dev);
		is->ether = NULL;
		return err;
	}
        if (strcmp(dev->name, is->is_name) != 0) {
          MX_INFO(("%s: my ether interface name is %s\n", 
		   is->is_name, dev->name));
        }
	is->is_name = dev->name;
	if (ETHER_K35SNF(eth)) {
		spin_lock_init(&eth->arch.ep_lock);
		eth->arch.ep_flags = 0;
	}

	return (0);
}

void
phx_ether_detach(mx_instance_state_t *is)
{
	struct net_device *dev;
	struct phx_ether *eth;

	eth = is->ether;
	if (!eth)
		return;
	dev = eth->arch.dev;
        is->is_name = is->is_name_default;
	unregister_netdev(dev);
	is->ether = 0;
	MAL_STBAR();
	phx_netdev_free(dev);
}

void
phx_ether_watchdog(mx_instance_state_t *is)
{
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)

inline int
_myri_ether_skb_add_data(struct sk_buff *skb, char __user *from, int copy)
{
	int err, off = skb->len;
	char *tail = skb->tail;

	if (skb->ip_summed == CHECKSUM_NONE) {
		uint32_t csum = csum_and_copy_from_user(from, tail, copy, 0, &err);
		if (err)
			return err;
		skb->csum = csum_block_add(skb->csum, csum, off);
	} else {
		err = copy_from_user(tail, from, copy);
		if (err)
			return err;
	}
	skb_put(skb, copy);
	return 0;
}
#endif

int
myri_ether_soft_rx(mx_instance_state_t *is, myri_soft_rx_t *r)
{
	struct sk_buff *skb = NULL;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,0,0)
	char *uaddr;
#else
	struct iov_iter *uaddr;
#endif
	int retval = -EBADF;
	int hlen, copylen, buf, copy;


	if (r->pkt_len > 9018)
		return EINVAL;

	skb = alloc_skb(r->pkt_len, GFP_KERNEL);
	if (skb == NULL) {
		retval = ENOMEM;
		goto bad;
	}

	/* avoid checksum on copyin for simplicity */
	skb->ip_summed = CHECKSUM_UNNECESSARY;
	copylen = r->pkt_len;	
	/* copy header separately, if required */
	buf = 0;
	if ((r->flags &  MYRI_SOFT_RX_CSUM_PART)) {
		hlen = r->hdr_len;
		if (hlen > copylen) {
			retval = EINVAL;
			goto bad;
		}
		while (hlen != 0) {
			copy = min(hlen, (int)r->copy_desc[buf].len);
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,0,0)
			uaddr = (char *)(unsigned long)r->copy_desc[buf].ptr;
#else
			uaddr = (struct iov_iter *)(unsigned long)r->copy_desc[buf].ptr;
#endif
			retval = myri_ether_skb_add_data(skb, uaddr, copy);
			if (retval)
				goto bad;
			hlen -= copy;
			copylen -= copy;
			r->copy_desc[buf].len -= copy;
			if (r->copy_desc[buf].len == 0)
				buf++;
			if (buf > 2) {
				printk("error, too many bufs!\n");
				retval = EINVAL;
				goto bad;
			}
		}
	}

	/* copy the rest of the packet */
	while (copylen != 0) {
		copy = min(copylen, (int)r->copy_desc[buf].len);
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,0,0)
		uaddr = (char *)(unsigned long)r->copy_desc[buf].ptr;
#else
		uaddr = (struct iov_iter *)(unsigned long)r->copy_desc[buf].ptr;
#endif
		retval = myri_ether_skb_add_data(skb, uaddr, copy);
		if (retval)
			goto bad;
		copylen -= copy;
		r->copy_desc[buf].len -= copy;
		if (r->copy_desc[buf].len == 0)
			buf++;
		if (buf > 2) {
			printk("error, too many bufs!\n");
			retval = EINVAL;
			goto bad;
		}
	}

	switch (r->flags & MYRI_SOFT_CSUM_MASK) {

	case  MYRI_SOFT_RX_CSUM_PART:
		skb->ip_summed = CHECKSUM_COMPLETE;
		skb->csum = r->csum;
		break;

	case  MYRI_SOFT_RX_CSUM_GOOD:
	case  MYRI_SOFT_RX_FAKE:
		skb->ip_summed = CHECKSUM_UNNECESSARY;
		break;
	default:
		skb->ip_summed = CHECKSUM_NONE;
		break;
	}


	skb->truesize = r->pkt_len + sizeof (*skb);
	skb->protocol = eth_type_trans(skb, is->ether->arch.dev);
	skb->dev = is->ether->arch.dev;
	netif_rx_ni(skb);
	return (0);

bad:
	if (skb != NULL)
		dev_kfree_skb(skb);
	return (retval);
}

int
phx_ether_set_discard_whitelist(mx_instance_state_t *is,
				phx_class_d_discard_whitelist_t *list)
{
	struct phx_ether *eth = is->ether;
	int i;

	if (ETHER_K35SNF(eth)) /* discard is not in K35SNF */
		return -1;
	if (!PHX_HAS_FEATURE(is, PHX_FEATURE_MULTICAST_DISCARD))
		return -1;
	if (!is->class_d_discard_whitelist)
		is->class_d_discard_whitelist = mx_kmalloc(sizeof(*list),
							   MX_MZERO|MX_WAITOK);
	if (!is->class_d_discard_whitelist)
		return -1;
	/* struct copy the whole list */
	*is->class_d_discard_whitelist = *list;
	if ((1 << is->id) & myri_enable_ipv4_multicast_discard_bitmask) {
		struct myri_rx_filterq_entry *entry;
		int removed;

		MX_INFO(("%s: setting white list filters\n", eth->arch.dev->name));

		/* Remove the old filters */
		removed = 0;
		while (!TAILQ_EMPTY(&eth->ether_mc_filter_list)) {
			entry = TAILQ_FIRST(&eth->ether_mc_filter_list);
			entry->filt.remove = 1;

			__phx_write_filter_regs(is->phx.nic_mem,
						&is->phx.filter_sync,
						&entry->filt,
						eth->myri_phx(ep)->rx_ep,
						PHX_FLTR_FLAG_BIT_IGNORE_VLAN |
						PHX_FLTR_FLAG_BIT_IGNORE_DST_PORT,
						NULL);

			TAILQ_REMOVE(&eth->ether_mc_filter_list, entry, entries);
			mx_kfree(entry);
			removed++;
		}
		MX_INFO(("%s: removed %d previous filters\n", eth->arch.dev->name,
			 removed));

		for (i = 0; i < list->count; i++) {
			int ip;

			/* Not our filter */
			if (memcmp(eth->current_mac, list->entries[i].if_mac, 6))
				continue;

			entry = mx_kmalloc(sizeof(*entry), MX_WAITOK|MX_MZERO);
			if (entry == NULL) {
				MX_WARN(("Failed to allocate filter\n"));
				return -1;
			}

			/* Fill in filter */
			ip = list->entries[i].class_d_ip;
			entry->filt.ip_addr = ntohl(list->entries[i].class_d_ip);
			entry->filt.port = 0; /* don't care */
			entry->filt.vlan = 0;
			entry->filt.tag = 0;
			entry->filt.remove = 0;
			entry->filt.kernel = 0;
			entry->filt.tcp = 0;
			entry->filt.txtstamp = 0;
			entry->filt.ab.enabled = 0;

			MX_INFO(("%s: add white list filter for ip %d.%d.%d.%d\n",
				 eth->arch.dev->name,
				 ((ip >> 0) & 0xff),
				 ((ip >> 8) & 0xff),
				 ((ip >> 16) & 0xff),
				 ((ip >> 24) & 0xff)));

			__phx_write_filter_regs(is->phx.nic_mem,
						&is->phx.filter_sync,
						&entry->filt,
						eth->myri_phx(ep)->rx_ep,
						PHX_FLTR_FLAG_BIT_IGNORE_VLAN |
						PHX_FLTR_FLAG_BIT_IGNORE_DST_PORT,
						NULL);
			TAILQ_INSERT_TAIL(&eth->ether_mc_filter_list, entry, entries);
		}
	} else if (list->count > 0) {
		MX_INFO(("%s: not setting white list filters because"
			 " discard is disabled\n", eth->arch.dev->name));
	}
	return 0;
}

int
k35snf_ether_remove_endpoints(mx_instance_state_t *is)
{
	struct phx_ether *eth = is->ether;
	struct phx_ether_arch *arch;

	if (!eth)
		return -1;
	arch = &eth->arch;

	/* Request TX/RX to stop using the endpoint */
	ep_lock(arch);
	ep_flags_set(arch, EP_FLAGS_REMOVE_ENTER);
	ep_unlock(arch);

	/* Wait till all ENTER flags are clear.  It is harmelss to set
	 * REMOVE_ENTER again.
	 */
	ep_enter_wait(arch,
		      (EP_FLAGS_TX_ENTER |
		       EP_FLAGS_IRQ_ENTER |
		       EP_FLAGS_POLL_ENTER |
		       EP_FLAGS_OPEN_ENTER |
		       EP_FLAGS_CLOSE_ENTER |
		       EP_FLAGS_RETURN_ENTER),
		      EP_FLAGS_REMOVE_ENTER);
	if (!is->ether_is_open) {
		ep_leave(arch, EP_FLAGS_REMOVE_ENTER);
		return -1;
	}

	/* If tx is stopped, let it continue */
	if (netif_queue_stopped(arch->dev)) {
		arch->stats_no_ep_tx_wake_queue++;
		netif_wake_queue(arch->dev);
	}

	/* The ethernet driver is no longer touching the endpoint. */
	phx_icu_disable_interrupt(eth->is, eth->ep.k35snf.rx->icu);
	phx_icu_disable_interrupt(eth->is, eth->ep.k35snf.tx->icu);

	/* There might be interrupts in flight. */
	msleep_interruptible(1);

	phx_stop_rx_endpoint(is, 0);
	phx_reset_rx_endpoint(is, 0);
	phx_apu_channel_reset(is, eth->apu_data_ch);
	phx_apu_channel_reset(is, eth->apu_desc_ch);
	k35snf_tx_stop(is);
	phx_apu_channel_reset(is, eth->tx.k35snf.apu_data_ch);
	phx_apu_channel_reset(is, eth->tx.k35snf.apu_desc_ch);

	phx_release_fpga_endpoint(is, eth->ep.k35snf.rx->id);
	phx_release_fpga_endpoint(is, eth->ep.k35snf.tx->id);

	ep_lock(arch);
	ep_flags_clear(arch, EP_FLAGS_REMOVE_ENTER);
	ep_flags_set(arch, EP_FLAGS_REMOVED);
	eth->ep.k35snf.stolen = 1;
	ep_unlock(arch);
	return 0;
}

void
k35snf_ether_return_endpoint(mx_instance_state_t *is)
{
	struct phx_ether *eth = is->ether;
	struct phx_ether_arch *arch = &eth->arch;
	int tx_claimed, rx_claimed;

	ep_enter_wait(arch,
		      (EP_FLAGS_REMOVE_ENTER |
		       EP_FLAGS_RETURN_ENTER |
		       EP_FLAGS_CLOSE_ENTER |
		       EP_FLAGS_OPEN_ENTER),
		      EP_FLAGS_RETURN_ENTER);
	/* No remove/open/close/return in progress. */
	if (!ep_flags_isset(arch, EP_FLAGS_REMOVED))
		goto exit; /* Never taken the endpoints */
	if (!is->ether_is_open) {
		/* The interface has been closed */
		eth->ep.k35snf.stolen = 0;
		ep_lock(arch);
		ep_flags_clear(arch, EP_FLAGS_REMOVED);
		ep_unlock(arch);
		goto exit;
	}
	/* First grab both endpoints */
	tx_claimed = phx_claim_fpga_endpoint(is, eth->ep.k35snf.tx->id, 1);
	rx_claimed = phx_claim_fpga_endpoint(is, eth->ep.k35snf.rx->id, 1);
	if (tx_claimed || rx_claimed) {
		if (!tx_claimed)
			phx_release_fpga_endpoint(is, eth->ep.k35snf.tx->id);
		if (!rx_claimed)
			phx_release_fpga_endpoint(is, eth->ep.k35snf.rx->id);
		goto exit;
	}
	/* Now re-init the endpoints */
	eth->ep.k35snf.stolen = 0;
	k35snf_ether_start_common(eth);

	ep_lock(arch);
	ep_flags_clear(arch, EP_FLAGS_REMOVED);
	ep_unlock(arch);

	/* Request TX and RX interrupts. */
	k35snf_rx_wake_request(is, eth->rx_desc.desc_idx);
	k35snf_tx_compl_count_ack(is, 1);
	phx_icu_enable_interrupt(is, eth->ep.k35snf.rx->icu);
	phx_icu_enable_interrupt(is, eth->ep.k35snf.tx->icu);

	MX_INFO(("%s: The ethernet interface re-attached to port 0\n",
		 is->is_name));
exit:
	ep_leave(arch, EP_FLAGS_RETURN_ENTER);
}

/*
  This file uses MX driver indentation.

  Local Variables:
  c-file-style:"linux"
  tab-width:8
  indent-tabs-mode:t
  End:
*/
