/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2014-2016 by CSP, Inc.  All rights reserved.                *
 *************************************************************************/

#include "mx_arch.h"
#include "mx_misc.h"
#include "mx_instance.h"
#include "mx_malloc.h"
#include "mx_pio.h"
#include "mal_stbar.h"
#include "myri_version.h"
#include "mal_cpu.h"
#include "phx_hw.h"
#include "phx_common.h"
#include "phx_avl.h"
#include "phx_ether_common.h"

mx_board_ops_t phx_k35snf_ops = { NULL };

static void
k35snf_tx_clear_stats(mx_instance_state_t *is, uint32_t port_id)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem;
  r = r + K35SNF_TX_BASE + (K35SNF_TX_SIZE * port_id);
  PHX_REG_WRITE_4((r + K35SNF_TX_TLP_STATS_CLEAR), 1);
}

static void
k35snf_rx_clear_stats(mx_instance_state_t *is, uint32_t port_id)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem;
  r = r + K35SNF_RX_BASE + (K35SNF_RX_SIZE * port_id);
  PHX_REG_WRITE_4((r + K35SNF_RX_TLP_STATS_CLEAR), 1);
  PHX_REG_WRITE_4((r + K35SNF_RX_PCIBP_RESET), 1);
}

static void
k35snf_xmac_rx_clear_stats(mx_instance_state_t *is, uint32_t port_id)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem;
  r = r + K35SNF_XMAC_RX_BASE + (K35SNF_XMAC_RX_SIZE * port_id) + K35SNF_XMAC_RX_STATS_CLEAR;
  PHX_REG_WRITE_4(r, 1);
}

static int
k35snf_mdio_poll(mx_instance_state_t *is, uint32_t port_id)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem;
  uint32_t value;
  uint32_t ready=0;
  int i = 0;

  r = r + K35SNF_XMAC_BASE + (K35SNF_XMAC_SIZE * port_id);

  while (!ready){
#if MAL_OS_LINUX
    udelay(1);
#else
    phx_win_usleep(1);
#endif
    value = PHX_REG_READ_4((r + K35SNF_XMAC_MDIO_CFG1));
    ready = ((value >> 7) & 1);
    if (++i > 1000){
      return -1;
    }
  }
  return 0;
}

static int
k35snf_1G_mac_init(mx_instance_state_t *is, int port_id, int auto_negotiate)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem;
  uint32_t cmd;
  uint32_t value;
  uint32_t addr = 0;

  /*  enable mdio , clock divider d = 38
   *  produces 156.25/((1+d)*2) ~= 2MHz clock < 2.5 MHz
   */
  r = r + K35SNF_XMAC_BASE + (K35SNF_XMAC_SIZE * port_id);

  value = ((1 << 6) | 38 );
  PHX_REG_WRITE_4((r + K35SNF_XMAC_MDIO_CFG0), (value));

  cmd = (((addr & 0x1f) << 16) |
         ((K35SNF_MDIO_OP_READ & 3) << 14) |
         (1 << 11));
  PHX_REG_WRITE_4((r + K35SNF_XMAC_MDIO_CFG1), (cmd));

  if (k35snf_mdio_poll(is, port_id) < 0) return -1;

  /* read the data */
  value = PHX_REG_READ_4((r + K35SNF_XMAC_MDIO_RX));

  /* turn off the Isolate bit */
  value &= ~(1 << 10);

  /* turn on/off the Auto-Negotiate bit */
  if (auto_negotiate)
    value |= (1 << 12);
  else
    value &= ~(1 << 12);

  /* write the modified data */
  PHX_REG_WRITE_4((r + K35SNF_XMAC_MDIO_TX), (value));

  cmd = (((addr & 0x1f) << 16) |
         ((K35SNF_MDIO_OP_WRITE & 3) << 14) |
         (1 << 11));
  PHX_REG_WRITE_4((r + K35SNF_XMAC_MDIO_CFG1), (cmd));

  if (k35snf_mdio_poll(is, port_id) < 0) return -1;

  return 0;
}

/* See SHQ-160.  Make it larger than Super Jumbo. */
#define K35SNF_MTU 17000

void
k35snf_xmac_init(mx_instance_state_t *is, uint32_t port_id)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem;
  uint32_t i;

  r = r + K35SNF_XMAC_BASE + (K35SNF_XMAC_SIZE * port_id);

  /* Receiver Reset */
  PHX_REG_WRITE_4((r + K35SNF_XMAC_RX_CONFIG_WORD_1), (1 << 31));
  mx_spin(10); /* 10 usec, necessary? */
  /* RX MTU Size + Enable */
  PHX_REG_WRITE_4((r + K35SNF_XMAC_RX_MTU_CONFIG_WORD), ((1 << 16) | K35SNF_MTU));
  /* Flow Control Enable TX and RX: disable both */
  PHX_REG_WRITE_4((r + K35SNF_XMAC_FLOW_CTRL_CONFIG), 0);
  /* Receiver Enable + Disable Length/Type Check */
  /* This now happens as part of opening the endpoint, when SNF RX is stopped */
  /*  PHX_REG_WRITE_4((r + K35SNF_XMAC_RX_CONFIG_WORD_1), (1 << 28) | (1 << 25)); */

  /* Disable flow controls */
  PHX_REG_WRITE_4((r + K35SNF_XMAC_FLOW_CTRL_CONFIG), 0);

  /* Transmitter Reset */
  PHX_REG_WRITE_4((r + K35SNF_XMAC_TX_CONFIG_WORD), (1 << 31));
  mx_spin(10); /* 10 usec, necessary? */
  /* TX MTU Size + Enable */
  PHX_REG_WRITE_4((r + K35SNF_XMAC_TX_MTU_CONFIG_WORD), ((1 << 16) | K35SNF_MTU));
  /* Enable transmit and jumbo frame */
  PHX_REG_WRITE_4((r + K35SNF_XMAC_TX_CONFIG_WORD), (1 << 30) | (1 << 28));

  /* Allow all packets to enter user space by initially disabling L2 filters. */
  r = (uint8_t *) is->phx.nic_mem;
  r = r + K35SNF_XMAC_RX_BASE + (K35SNF_XMAC_RX_SIZE * port_id);

  for (i = 0; i < K35SNF_XMAC_RX_FILTER_MU_COUNT; i++) {
    PHX_REG_WRITE_8((r + K35SNF_XMAC_RX_FILTER_PATTERN_MASK + i * sizeof(k35snf_L2_filter_pattern_mask_t)), (uint64_t) 0);
    PHX_REG_WRITE_8((r + K35SNF_XMAC_RX_FILTER_PATTERN_MASK + i * sizeof(k35snf_L2_filter_pattern_mask_t) + 8), (uint64_t) 0);
  }

  for (i = 0; i < K35SNF_XMAC_RX_FILTER_LT_ENTRY_COUNT; i++)
    PHX_REG_WRITE_4((r + K35SNF_XMAC_RX_FILTER_LOOKUP_TABLE), ((i << 16) | 0x1ff));

  PHX_REG_WRITE_2((r + K35SNF_XMAC_RX_FILTER_MU_ENABLE), (uint16_t) 1);
}

#define PLUG_RJ45  0x22
#define PLUG_COPPER_PIGTAIL 0x21
#define PLUG_FIBER 0x7

void
k35snf_xmac_configure(mx_instance_state_t *is, uint32_t port_id)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem;
  uint32_t version, cap;
  uint32_t connector;
  uint32_t auto_negotiate;
  uint32_t preemphasis=K35_SERDES_10G_FIBER;

  /* If this port has a 1G mac, need to disable isolate bit in phy. */
  r = (uint8_t *) is->phx.nic_mem;
  r = r + K35SNF_XMAC_BASE + (K35SNF_XMAC_SIZE * port_id);

  version = PHX_REG_READ_4((r + K35SNF_XMAC_VERSION));
  cap = PHX_REG_READ_4((r + K35SNF_XMAC_CAPABILITY));

  /* determine the connector type */
  connector = phx_get_transceiver_type(is, port_id);

  /* FIXME - dont support transceivers with no connector type */
  if (connector == 0)
    connector = PLUG_RJ45;

  if (connector == PLUG_RJ45)
    auto_negotiate = 0;
  else
    auto_negotiate = 1;

  /* do any special configuration based on the MAC and transceiver */
  if (version == K35SNF_MAC_1G){
    if (k35snf_1G_mac_init(is, port_id, auto_negotiate))
      MX_WARN(("1000BASE-X MAC port %d: is still isolated\n", port_id));

    if (connector == PLUG_RJ45)
       phx_reset_marvell_88E1111(is, port_id);  /* allow RJ45 to retrain link */

    if ((connector == PLUG_RJ45) || (connector == PLUG_COPPER_PIGTAIL))
      preemphasis=K35_SERDES_1G_COPPER;

    if (connector == PLUG_FIBER)
      preemphasis = K35_SERDES_1G_FIBER;

    phx_set_serdes_preemphasis(is, port_id, preemphasis);
#if MAL_DEBUG
    MX_INFO(("1000BASE-X MAC port %d: version=0x%08x capability=0x%08x conn=0x%x pre=0x%08x\n", port_id,
           version, cap, connector, preemphasis));
#endif
  }

  if (version == K35SNF_MAC_10G){
    if (connector == PLUG_COPPER_PIGTAIL)
      preemphasis=K35_SERDES_10G_COPPER;

    if (connector == PLUG_FIBER)
      preemphasis=K35_SERDES_10G_FIBER;

    phx_set_serdes_preemphasis(is, port_id, preemphasis);
#if MAL_DEBUG
    MX_INFO(("XMAC port %d: version=0x%08x capability=0x%08x conn=0x%x pre=0x%08x\n", port_id,
           version, cap, connector, preemphasis));
#endif
  }
}

/* This routine takes the MAC core out of reset allowing register access */
static void
k35snf_mac_enable(mx_instance_state_t *is, uint32_t port_id)
{
  uint8_t *r;
  uint32_t rx_inhibit;

  /* the inhibit is a bit in a global FPGA register, clear it */
  r = (uint8_t*)is->phx.nic_mem + PHX_I2C_BASE;
  rx_inhibit = PHX_REG_READ_4((r + K35_MAC_RESET));

  /* uninhibit at the gearbox and release rx/tx reset */
  PHX_REG_WRITE_4((r + K35_MAC_RESET), (rx_inhibit & ~(0x10011 << port_id)));
  rx_inhibit = PHX_REG_READ_4((r + K35_MAC_RESET));
}

static void
k35snf_mac_disable(mx_instance_state_t *is, uint32_t port_id)
{
  uint8_t *r;
  uint32_t rx_inhibit;

  /* the inhibit is a bit in a global FPGA register, clear it */
  r = (uint8_t*)is->phx.nic_mem + PHX_I2C_BASE;
  rx_inhibit = PHX_REG_READ_4((r + K35_MAC_RESET));

  /* inhibit at the gearbox and hold rx/tx in reset */
  PHX_REG_WRITE_4((r + K35_MAC_RESET), (rx_inhibit | (0x10011 << port_id)));
  rx_inhibit = PHX_REG_READ_4((r + K35_MAC_RESET));
}

static int
k35snf_set_intr_coal(mx_instance_state_t *is, int microseconds)
{
  /* 4 nsec per cycle */
  uint32_t cycles = ((microseconds * 1000) / 4);
  uint8_t *r = (uint8_t*)is->phx.nic_mem + K35SNF_ICU_BASE;
  r += K35SNF_ICU_INTR_INTERVAL;
  PHX_REG_WRITE_4(r, cycles);
  is->intr_coal_usec = microseconds;
  return is->intr_coal_usec;
}

static void
k35snf_set_rx_timelimit(mx_instance_state_t *is, int limit)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem;
  r += K35SNF_RX_BASE + (K35SNF_RX_SIZE * is->port_id) + K35SNF_RX_TIME_LIMIT;
  PHX_REG_WRITE_4(r, limit);
}

static int
k35snf_init_board(mx_instance_state_t *is)
{
  int i, init_fpga;
  uint32_t version=0;
  uint32_t board_arch=0;
  int attempts = 0;
  uint8_t *r;

  if (!MX_IS_PARENT(is)) {
    MX_WARN(("BUG: A child instance should not call k35snf_init_board\n"));
    return -1;
  }

  mx_mutex_enter(&mx_fpga_mutex);
  init_fpga = !is->phx.pfi->initialized;

  /* Enable all 4 ports, set promisc, MTU, address and so on. */
  for (i = 0; i < is->num_ports; i++) {
    /* stop incoming packets at the MAC before removing inhibit */
    k35snf_xmac_rx_disable(is, i);

    /*
     * Take RSS core out of reset.
     *
     * Bit 0: active high.  Clears statistics counters.  Sets all RSS counters to 0.
     * Bit 1: active high.  RSS port reset.  Resets entire port path.
     * Bit 2: active high.  Disables next ring sequencer.
     *
     * Reset procedure:
     * Bit 0: set it, then clear it.  (We do this.)
     * Bit 1: set it, then clear it.  (We do this.  It brings RSS out of reset.)
     * Bit 2: set it, then leave it.  (We do this, leave it set so that we default the sequencer disabled)
     * */
    r = ((uint8_t *) is->phx.nic_mem) + K35SNF_RSS_BASE + (K35SNF_RSS_RESET + (K35SNF_RSS_SIZE * i));
    
    PHX_REG_WRITE_4(r, (K35SNF_MERGE_NEXT_RING_RESET_RSS_BLOCK |
                        K35SNF_MERGE_NEXT_RING_DISABLE_SEQUENCER |
                        K35SNF_MERGE_NEXT_RING_CLEAR_RSS_STATS));
    udelay(1);
    PHX_REG_WRITE_4(r, K35SNF_MERGE_NEXT_RING_DISABLE_SEQUENCER);
    udelay(1);

    /* board comes up with the MAC in reset and inhibited */
    k35snf_mac_enable(is, i);
    /* Clear RX backpressure and TLP stats */
    k35snf_rx_clear_stats(is, i);
    /* Clear TX TLP stats */
    k35snf_tx_clear_stats(is, i);
    /* Clear XMAC Receiver counters */
    k35snf_xmac_rx_clear_stats(is, i);
    k35snf_xmac_init(is, i);
  }

  k35snf_icu_enable(is);

  k35snf_set_intr_coal(is, myri_intr_coal_delay);
  k35snf_set_rx_timelimit(is, myri_k35snf_rx_timelimit);

  is->link_speed = k35snf_xmac_get_link_speed(is);

  /* Clear the counters for the first time.  We will read all the registers
   * and save them in the process.
   */
  myri_counters64_clear(is);

  if (init_fpga){
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0)
    struct timespec64 ts;
#else
    struct timespec ts;
#endif
    /* read the specs from the EEPROM to get MAC address */
    MX_INFO(("Initializing FPGA: start\n"));

    phx_xilinx_init((char*)is->phx.i2c_base_mem, &version, &board_arch);

    /* firmware monotonic version for compatibility checks */
    is->phx.pfi->firmware_rev = version;
    is->phx.pfi->xilinx_board_type = board_arch;

    /* read the eeprom strings from the onboard eeprom */
    phx_read_specs((char*)is->phx.i2c_base_mem, is->phx.eeprom_strings, 256,
                   is->phx.pfi->xilinx_board_type);

    /* This sets MAC address (is->mac_addr).
     * PHX-310 retry reading if is not OK.
     */
    while (phx_parse_eeprom_strings(is) && (attempts++ < 10)){
      MX_INFO(("Error processing eeprom strings retry %d\n", attempts));
      phx_read_specs((char*)is->phx.i2c_base_mem, is->phx.eeprom_strings, 256,
                     is->phx.pfi->xilinx_board_type);
    }

    MX_INFO(("Initializing FPGA: end, firmware version %d board arch %c%c%c%c\n", version,
              ((board_arch & 0xff000000) >> 24),
              ((board_arch & 0xff0000) >> 16),
              ((board_arch & 0xff00) >> 8),
              (board_arch & 0xff)));

    is->phx.pfi->initialized = 1;

#if PHX_ENABLE_PTP
    /* initialize the PTP clock */
    phx_ptp_init(is->phx.pfi);

    is->phx.pfi->ptp_clock = ptp_clock_register(&is->phx.pfi->ptp_info,
           &is->arch.pci_dev->dev);
    if(IS_ERR(is->phx.pfi->ptp_clock)){
       is->phx.pfi->ptp_clock = NULL;
    }
      phx_k35_check_pps(is);
#endif

    /* Initialze the NIC clock.  The registers are same as PHX. */
    phx_clk_init(is, myri_clk_invert_pps, !myri_clk_enable_pps,
		 !myri_clk_enable_10mhz);

    /* Push the current host time to the NIC.  Afterwards, we do not
     * synchronize the clocks.
     */
    mx_getnstimeofday(&ts);
    if (phx_clk_is_pps_active(is)) {
      phx_clk_set_with_pps(is, &ts);
      MX_INFO(("PPS is active: wrote the seconds to the adapter\n"));
      is->clk_ts_state = MYRI_CLK_TS_EXTERNAL_PPS;
    } else {
      phx_clk_set(is, &ts);
      MX_INFO(("PPS is inactive: wrote the time to the adapter ts=%lld.%ld\n",
               (long long int)ts.tv_sec, ts.tv_nsec));
      is->clk_ts_state = MYRI_CLK_TS_INTERNAL_TCXO;
    }
  }

  /*
   * Transceiver specific MAC setup requires i2c, setup by phx_xilinx_init above.
   */
  for (i = 0; i < is->num_ports; i++) {
    k35snf_xmac_configure(is, i);
  }

  mx_mutex_exit(&mx_fpga_mutex);
  return 0;
}

static int
k35snf_init_endpoint_pool(mx_instance_state_t *is, struct phx_fpga_endpoint_pool *pool)
{
  int i;

  /* FPGA has one function and 4 RX endpoints and 4 TX endpoints.
   * Port 0: kernel and sniffer share the single endpoint.
   * Ports 1-3: only for sniffer
   */
  pool->num_function_endpoints = 4 * 2;
  pool->num_total_endpoints = 4 * 2;
  pool->endpoints = mx_kmalloc(sizeof(struct phx_physical_ep) * pool->num_total_endpoints, MX_MZERO);
  pool->num_function_not_reserved_endpoints = 0;
  if (!pool->endpoints)
    return -1;
  for (i = 0; i < pool->num_total_endpoints; i++) {
    struct phx_physical_ep *ep = &pool->endpoints[i];
    ep->id = i;
    ep->owner_is = NULL;
    ep->reserved = 0;
    if ((i & 1)) {
      /* RX endpoint = port number.  Each port has its own SnifferRX and one
       * RX endpoint (always 0).  So rx_ep is only used to signify the port,
       * not the actual endpoint number when we access SnifferRX registers.
       */
      ep->rx_ep = i / 2;
      ep->tx_ep = INVALID_EP;
    } else {
      ep->rx_ep = INVALID_EP;
      /* TX endpoint = port number.  The same story as RX endpoint. */
      ep->tx_ep = i / 2;
    }
    ep->icu = i + 1;
    ep->func = 0;
    ep->kernel = 0;
  }
  return 0;
}

/* Called once per FPGA */
static void
k35snf_setup_hw_caps(mx_instance_state_t *is)
{
  struct phx_fpga_instance *pfi = is->phx.pfi;

  /* malloc failed earlier on, do nothing. We should not even come to this
   * function if malloc fails, though.
   */
  if (!(pfi->hw_hash.enabled_flags && pfi->hw_hash.in_use))
    return;

  /* next_ring firmware happens to have HW hash too */
  if (PHX_HAS_FEATURE(is, PHX_FEATURE_FPGA_MERGE_NEXT_RING)) {
    pfi->hw_hash.supported_flags = (SNF_RSS_FLAGS_IP |
                                    SNF_RSS_FLAGS_SRC_PORT |
                                    SNF_RSS_FLAGS_DST_PORT |
                                    SNF_RSS_FLAGS_GTP |
                                    SNF_RSS_FLAGS_GRE);
  }

  if (PHX_HAS_FEATURE(is, PHX_FEATURE_FPGA_MERGE_NEXT_RING)) {
    pfi->port_merge.mode_flags = SNF_PORT_MERGE_MODE_NEXT_RING;
  } else {
    pfi->port_merge.mode_flags = SNF_PORT_MERGE_MODE_NONE;
  }
}

/* Returns 1 if next_ring merge is enabled/allowed, 0 otherwise */
static int
k35snf_enable_next_ring_merge(mx_instance_state_t *is, uint32_t portmask)
{
  struct phx_fpga_instance *fpga = is->phx.pfi;
  uint32_t start_id = -1;
  int i;

  /* Find the start id (port 0's id) */
  for (i = 0; i < fpga->attach_count; i++) {
    if (MX_IS_PORT_ID(fpga->is[i]) == 0)
      start_id = fpga->is[i]->id;
  }
  mal_always_assert(start_id != -1);

  /* Nothing to do to enable next_ring in this function. When we enable
   * an rx endpoint, we always enable/reset next_ring id. See phx_start_rx_endpoint.
   */

  /* For now, only allow ports 0 and 1 or 2 and 3. */
  return (is->num_ports == 2 && (0x3 << start_id) == portmask) ||
      (is->num_ports == 4 && (((0x3 << start_id) == portmask) ||
                              (0x3 << (start_id + 2)) == portmask));
}

static void
k35snf_enable_hw_hash(mx_instance_state_t *is, uint32_t rss_flags)
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem;
  int port = MX_IS_PORT_ID(is);
  uint32_t mask = 0xff;
  uint32_t mode = K35SNF_RSS_MODE_LINK;

  if ((rss_flags & SNF_RSS_FLAGS_IP))
    mask &= ~(K35SNF_RSS_MASK_DST_IP | K35SNF_RSS_MASK_SRC_IP);
  if ((rss_flags & SNF_RSS_FLAGS_SRC_PORT))
    mask &= ~(K35SNF_RSS_MASK_SRC_PORT);
  if ((rss_flags & SNF_RSS_FLAGS_DST_PORT))
    mask &= ~(K35SNF_RSS_MASK_DST_PORT);
  if ((rss_flags & SNF_RSS_FLAGS_GTP))
    mask &= ~(K35SNF_RSS_MASK_TEID);

  /* Tunnels need a different mode... */
  if ((rss_flags & (SNF_RSS_FLAGS_GTP | SNF_RSS_FLAGS_GRE)))
    mode = K35SNF_RSS_MODE_TUNNELED;

#if MAL_DEBUG
  MX_INFO(("k35snf_enable_hw_hash: mask=0x%x mode=%d\n", mask, mode));
#endif

  r += K35SNF_RSS_BASE + (K35SNF_RSS_SIZE * port) + K35SNF_RSS_MASK;
  PHX_REG_WRITE_4(r, mask);

  r = (uint8_t*)is->phx.nic_mem;
  r += K35SNF_RSS_BASE + (K35SNF_RSS_SIZE * port) + K35SNF_RSS_SELECT;
  PHX_REG_WRITE_4(r, mode);
}

/* Only port 0 attaches to the kernel as an ethernet port
 */
static struct phx_physical_ep *
k35snf_ether_endpoint(mx_instance_state_t *is)
{
  struct phx_fpga_endpoint_pool *pool = &is->phx.pfi->ep_pool;
  if (MX_IS_CHILD(is))
    return NULL;
  return &pool->endpoints[0];
}

/* Port 0 => 0, port 1 => 1, and so on
 */
static struct phx_physical_ep *
k35snf_sniffer_endpoint(mx_instance_state_t *is)
{
  struct phx_fpga_endpoint_pool *pool = &is->phx.pfi->ep_pool;
  return &pool->endpoints[is->port_id * 2 + 1];
}

/* n'th within the function */
static struct phx_physical_ep *
k35snf_physical_endpoint(mx_instance_state_t *is, int nth)
{
  struct phx_fpga_endpoint_pool *pool = &is->phx.pfi->ep_pool;
  return &pool->endpoints[nth];
}

/* We have only 1 function.  Aliased counters are per-port counters.
 * These per-port counters act like per-endpoint counters, and need strides.
 * So, set per_port_stride accordingly.  FIXME
 *
 * We only have 1 RX and 1 TX endpoints.  So, every non-global counter has
 * TYPE_ALIASED.  We have no TYPE_ENDPOINT counters.
 */
const phx_counter_t phx_k35snf_counters[] = {
  { "Cycle Count", K35SNF_RX_BASE + K35SNF_RX_RX_CYCLE_COUNT,              CNT_ALIASED_HEX, 0, K35SNF_RX_SIZE },
  { "Driver Uptime", 0,                                                    CNT_DRV_UPTIME, 0 },
  { "Counters Uptime", 0,                                                  CNT_UPTIME, 0 },
  { "Packets Received", K35SNF_RX_BASE + K35SNF_RX_PACKET_COUNT,                  CNT_ALIASED_48, 0, K35SNF_RX_SIZE },
  { "Bytes Received", K35SNF_RX_BASE + K35SNF_RX_BYTE_COUNT,                      CNT_ALIASED_48, 0, K35SNF_RX_SIZE },
  { "Packets Finalized", K35SNF_RX_BASE + K35SNF_RX_FF_PACKET_COUNT,              CNT_ALIASED_48, 0, K35SNF_RX_SIZE },
#define RX_COUNTER_TYPE (CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE)
#define RX_VB2_COUNTER_TYPE (CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2)
  { "PCIE FIFO Empty", K35SNF_RX_BASE + K35SNF_RX_PCIBP_EMPTY,                    RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "PCIE FIFO < 1/4", K35SNF_RX_BASE + K35SNF_RX_PCIBP_Q1,                       RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "PCIE FIFO 1/4 - 1/2", K35SNF_RX_BASE + K35SNF_RX_PCIBP_Q2,                   RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "PCIE FIFO 1/2 - 3/4", K35SNF_RX_BASE + K35SNF_RX_PCIBP_Q3,                   RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "PCIE FIFO > 3/4", K35SNF_RX_BASE + K35SNF_RX_PCIBP_Q4,                       RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "PCIE FIFO Full", K35SNF_RX_BASE + K35SNF_RX_PCIBP_FULL,                      RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Time Limit", K35SNF_RX_BASE + K35SNF_RX_TIME_LIMIT,                       RX_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "APU Version ", K35SNF_RX_APU_BASE + K35SNF_RX_APU_ID_VERSION,                CNT_ALIASED_HEX | COUNTER_TYPE_SNAPSHOT  | COUNTER_DISPLAY_VB2, 0, K35SNF_RX_SIZE },
  { "APU Configuration", K35SNF_RX_APU_BASE + K35SNF_RX_APU_NUM_CHANNEL,          CNT_ALIASED_HEX | COUNTER_TYPE_SNAPSHOT  | COUNTER_DISPLAY_VB2, 0, K35SNF_RX_SIZE },
  { "RX Cycle Count", K35SNF_RX_BASE + K35SNF_RX_RX_CYCLE_COUNT,                  RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Data Pages", K35SNF_RX_BASE + K35SNF_RX_DATA_RING_SIZE,                   RX_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Desc Pages", K35SNF_RX_BASE + K35SNF_RX_DESC_RING_SIZE,                   RX_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  /* FIXME: These two read wrong values? */
  { "RX Data Pages Available", K35SNF_RX_BASE + K35SNF_RX_DATA_PG_AVAIL,          RX_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Desc Pages Available", K35SNF_RX_BASE + K35SNF_RX_DESC_PG_AVAIL,          RX_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Data Pages Available Min", K35SNF_RX_BASE + K35SNF_RX_DATA_PG_AVAIL_MIN,  RX_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Desc Pages Available Min", K35SNF_RX_BASE + K35SNF_RX_DESC_PG_AVAIL_MIN,  RX_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Data Consumer Position", K35SNF_RX_BASE + K35SNF_RX_DATA_CONSR_POSITION,  RX_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Data Producer Position", K35SNF_RX_BASE + K35SNF_RX_DATA_PRODW_POSITION,  RX_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Desc Consumer Position", K35SNF_RX_BASE + K35SNF_RX_DESC_CONSR_POSITION,  RX_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Desc Producer Position", K35SNF_RX_BASE + K35SNF_RX_DESC_PRODW_POSITION,  RX_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Descriptors Created", K35SNF_RX_BASE + K35SNF_RX_EP_DESC_COUNT,           RX_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Wake Request", K35SNF_RX_BASE + K35SNF_RX_EP_DESC_WAKE_REQUEST,           RX_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Data TLP Q1", K35SNF_RX_BASE + K35SNF_RX_DATA_TLP_Q1,                     RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Data TLP Q2", K35SNF_RX_BASE + K35SNF_RX_DATA_TLP_Q2,                     RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Data TLP Q3", K35SNF_RX_BASE + K35SNF_RX_DATA_TLP_Q3,                     RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Data TLP Q4", K35SNF_RX_BASE + K35SNF_RX_DATA_TLP_Q4,                     RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Data TLP Full", K35SNF_RX_BASE + K35SNF_RX_DATA_TLP_FULL,                 RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Desc TLP Q1", K35SNF_RX_BASE + K35SNF_RX_DESC_TLP_Q1,                     RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Desc TLP Q2", K35SNF_RX_BASE + K35SNF_RX_DESC_TLP_Q2,                     RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Desc TLP Q3", K35SNF_RX_BASE + K35SNF_RX_DESC_TLP_Q3,                     RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Desc TLP Q4", K35SNF_RX_BASE + K35SNF_RX_DESC_TLP_Q4,                     RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Desc TLP Full", K35SNF_RX_BASE + K35SNF_RX_DESC_TLP_FULL,                 RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Data MWr/s Min", K35SNF_RX_BASE + K35SNF_RX_DATA_TLP_PS_MIN,              RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Data MWr/s Max", K35SNF_RX_BASE + K35SNF_RX_DATA_TLP_PS_MAX,              RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Data Full MWr/s Min", K35SNF_RX_BASE + K35SNF_RX_DATA_TLP_FULL_PS_MIN,    RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Data Full MWr/s Max", K35SNF_RX_BASE + K35SNF_RX_DATA_TLP_FULL_PS_MAX,    RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Data MWr DW/s Min", K35SNF_RX_BASE + K35SNF_RX_DATA_TLP_DW_PS_MIN,        RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Data MWr DW/s Max", K35SNF_RX_BASE + K35SNF_RX_DATA_TLP_DW_PS_MAX,        RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Data MWr PLDW/s Min", K35SNF_RX_BASE + K35SNF_RX_DATA_TLP_PLDW_PS_MIN,    RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Data MWr PLDW/s Max", K35SNF_RX_BASE + K35SNF_RX_DATA_TLP_PLDW_PS_MAX,    RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Desc MWr/s Min", K35SNF_RX_BASE + K35SNF_RX_DESC_TLP_PS_MIN,              RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Desc MWr/s Max", K35SNF_RX_BASE + K35SNF_RX_DESC_TLP_PS_MAX,              RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Desc Full MWr/s Min", K35SNF_RX_BASE + K35SNF_RX_DESC_TLP_FULL_PS_MIN,    RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Desc Full MWr/s Max", K35SNF_RX_BASE + K35SNF_RX_DESC_TLP_FULL_PS_MAX,    RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Desc MWr DW/s Min", K35SNF_RX_BASE + K35SNF_RX_DESC_TLP_DW_PS_MIN,        RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Desc MWr DW/s Max", K35SNF_RX_BASE + K35SNF_RX_DESC_TLP_DW_PS_MAX,        RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Desc MWr PLDW/s Min", K35SNF_RX_BASE + K35SNF_RX_DESC_TLP_PLDW_PS_MIN,    RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
  { "RX Desc MWr PLDW/s Max", K35SNF_RX_BASE + K35SNF_RX_DESC_TLP_PLDW_PS_MAX,    RX_VB2_COUNTER_TYPE, 0, K35SNF_RX_SIZE },
#define TX_COUNTER_64_TYPE (CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_TYPE_64BIT)
#define TX_COUNTER_TYPE (CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VERBOSE)
#define TX_VB2_COUNTER_TYPE (CNT_ALIASED | COUNTER_TYPE_SNAPSHOT | COUNTER_DISPLAY_VB2)
  { "TX Time Limit", K35SNF_TX_BASE + K35SNF_TX_TIME_LIMIT,                       TX_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  /* FIXME: Need to clear these when we close TX endpoints */
  /* 48-bit rolling */
  { "TX Byte Count", K35SNF_TX_BASE + K35SNF_TX_BYTE_COUNT,                       TX_COUNTER_64_TYPE, 0, K35SNF_TX_SIZE },
  /* 48-bit rolling */
  { "TX Packet Count", K35SNF_TX_BASE + K35SNF_TX_PACKET_COUNT,                   TX_COUNTER_64_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Completion Page LSW", K35SNF_TX_BASE + K35SNF_TX_COMPL_ADDR_LOW,          CNT_ALIASED_HEX | TX_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Completion Page MSW", K35SNF_TX_BASE + K35SNF_TX_COMPL_ADDR_HIGH,         CNT_ALIASED_HEX | TX_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Completion Count", K35SNF_TX_BASE + K35SNF_TX_COMPL_COUNT,                CNT_ALIASED | COUNTER_TYPE_SNAPSHOT, 0, K35SNF_TX_SIZE },
  { "TX Completion Count Ack", K35SNF_TX_BASE + K35SNF_TX_COMPL_COUNT_ACK,        TX_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data MRd Q1", K35SNF_TX_BASE + K35SNF_TX_DATA_MRD_Q1,                     TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data MRd Q2", K35SNF_TX_BASE + K35SNF_TX_DATA_MRD_Q2,                     TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data MRd Q3", K35SNF_TX_BASE + K35SNF_TX_DATA_MRD_Q3,                     TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data MRd Q4", K35SNF_TX_BASE + K35SNF_TX_DATA_MRD_Q4,                     TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data MRd Full", K35SNF_TX_BASE + K35SNF_TX_DATA_MRD_FULL,                 TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc MRd Q1", K35SNF_TX_BASE + K35SNF_TX_DESC_MRD_Q1,                     TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc MRd Q2", K35SNF_TX_BASE + K35SNF_TX_DESC_MRD_Q2,                     TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc MRd Q3", K35SNF_TX_BASE + K35SNF_TX_DESC_MRD_Q3,                     TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc MRd Q4", K35SNF_TX_BASE + K35SNF_TX_DESC_MRD_Q4,                     TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc MRd Full", K35SNF_TX_BASE + K35SNF_TX_DESC_MRD_FULL,                 TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data CplD Q1", K35SNF_TX_BASE + K35SNF_TX_DATA_CPLD_Q1,                   TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data CplD Q2", K35SNF_TX_BASE + K35SNF_TX_DATA_CPLD_Q2,                   TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data CplD Q3", K35SNF_TX_BASE + K35SNF_TX_DATA_CPLD_Q3,                   TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data CplD Q4", K35SNF_TX_BASE + K35SNF_TX_DATA_CPLD_Q4,                   TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data CplD Full", K35SNF_TX_BASE + K35SNF_TX_DATA_CPLD_FULL,               TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc CplD Q1", K35SNF_TX_BASE + K35SNF_TX_DESC_CPLD_Q1,                   TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc CplD Q2", K35SNF_TX_BASE + K35SNF_TX_DESC_CPLD_Q2,                   TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc CplD Q3", K35SNF_TX_BASE + K35SNF_TX_DESC_CPLD_Q3,                   TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc CplD Q4", K35SNF_TX_BASE + K35SNF_TX_DESC_CPLD_Q4,                   TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc CplD Full", K35SNF_TX_BASE + K35SNF_TX_DESC_CPLD_FULL,               TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data MRd/s Min", K35SNF_TX_BASE + K35SNF_TX_DATA_MRD_PS_MIN,              TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data MRd/s Max", K35SNF_TX_BASE + K35SNF_TX_DATA_MRD_PS_MAX,              TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data Full MRd/s Min", K35SNF_TX_BASE + K35SNF_TX_DATA_MRD_FULL_PS_MIN,    TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data Full MRd/s Max", K35SNF_TX_BASE + K35SNF_TX_DATA_MRD_FULL_PS_MAX,    TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data MRd DW/s Min", K35SNF_TX_BASE + K35SNF_TX_DATA_MRD_DW_PS_MIN,        TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data MRd DW/s Max", K35SNF_TX_BASE + K35SNF_TX_DATA_MRD_DW_PS_MAX,        TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc MRd/s Min", K35SNF_TX_BASE + K35SNF_TX_DESC_MRD_PS_MIN,              TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc MRd/s Max", K35SNF_TX_BASE + K35SNF_TX_DESC_MRD_PS_MAX,              TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc Full MRd/s Min", K35SNF_TX_BASE + K35SNF_TX_DESC_MRD_FULL_PS_MIN,    TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc Full MRd/s Max", K35SNF_TX_BASE + K35SNF_TX_DESC_MRD_FULL_PS_MAX,    TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc MRd DW/s Min", K35SNF_TX_BASE + K35SNF_TX_DESC_MRD_DW_PS_MIN,        TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc MRd DW/s Max", K35SNF_TX_BASE + K35SNF_TX_DESC_MRD_DW_PS_MAX,        TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data CplD/s Min", K35SNF_TX_BASE + K35SNF_TX_DATA_CPLD_PS_MIN,            TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data CplD/s Max", K35SNF_TX_BASE + K35SNF_TX_DATA_CPLD_PS_MAX,            TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data Full CplD/s Min", K35SNF_TX_BASE + K35SNF_TX_DATA_CPLD_FULL_PS_MIN,  TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data Full CplD/s Max", K35SNF_TX_BASE + K35SNF_TX_DATA_CPLD_FULL_PS_MAX,  TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data CplD DW/s Min", K35SNF_TX_BASE + K35SNF_TX_DATA_CPLD_DW_PS_MIN,      TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data CplD DW/s Max", K35SNF_TX_BASE + K35SNF_TX_DATA_CPLD_DW_PS_MAX,      TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc CplD/s Min", K35SNF_TX_BASE + K35SNF_TX_DESC_CPLD_PS_MIN,            TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc CplD/s Max", K35SNF_TX_BASE + K35SNF_TX_DESC_CPLD_PS_MAX,            TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc Full CplD/s Min", K35SNF_TX_BASE + K35SNF_TX_DESC_CPLD_FULL_PS_MIN,  TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc Full CplD/s Max", K35SNF_TX_BASE + K35SNF_TX_DESC_CPLD_FULL_PS_MAX,  TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc CplD DW/s Min", K35SNF_TX_BASE + K35SNF_TX_DESC_CPLD_DW_PS_MIN,      TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Desc CplD DW/s Max", K35SNF_TX_BASE + K35SNF_TX_DESC_CPLD_DW_PS_MAX,      TX_VB2_COUNTER_TYPE, 0, K35SNF_TX_SIZE },
  { "TX Data Ring Size (4KB Pages)", K35SNF_TX_BASE + K35SNF_TX_DATA_RING_SIZE,   CNT_ALIASED | COUNTER_TYPE_SNAPSHOT, 0, K35SNF_TX_SIZE },
  { "TX Data Ring ConsR Position", K35SNF_TX_BASE + K35SNF_TX_DATA_CONSR_POSITION, CNT_ALIASED | COUNTER_TYPE_SNAPSHOT, 0, K35SNF_TX_SIZE },
  { "TX Desc Ring Size (Descriptors)", K35SNF_TX_BASE + K35SNF_TX_DESC_RING_SIZE,  CNT_ALIASED | COUNTER_TYPE_SNAPSHOT, 0, K35SNF_TX_SIZE },
  { "TX Desc Ring ProdW Position", K35SNF_TX_BASE + K35SNF_TX_DESC_PRODW_POSITION, CNT_ALIASED | COUNTER_TYPE_SNAPSHOT, 0, K35SNF_TX_SIZE },
  { "TX Desc Ring ConsR Position", K35SNF_TX_BASE + K35SNF_TX_DESC_CONSR_POSITION, CNT_ALIASED | COUNTER_TYPE_SNAPSHOT, 0, K35SNF_TX_SIZE },
  /* 2 x 32-bit registers (48 bits valid) */
#define XMAC_RX_COUNTER_TYPE (CNT_ALIASED | COUNTER_DISPLAY_VERBOSE | COUNTER_TYPE_2_32BIT)
  { "XMAC RX pktsRcvd", K35SNF_XMAC_RX_BASE + K35SNF_XMAC_RX_PKTS_RCVD,             XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_RX_SIZE },
  { "XMAC RX pktsAbort", K35SNF_XMAC_RX_BASE + K35SNF_XMAC_RX_PKTS_ABORT,           XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_RX_SIZE },
  { "XMAC RX pktsDropped", K35SNF_XMAC_RX_BASE + K35SNF_XMAC_RX_PKTS_DROPPED,       CNT_ALIASED | COUNTER_TYPE_2_32BIT, 0, K35SNF_XMAC_RX_SIZE },
  { "XMAC RX pktsNoMeta", K35SNF_XMAC_RX_BASE + K35SNF_XMAC_RX_PKTS_NO_META,        XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_RX_SIZE },
  { "XMAC RX pktsNoOutCtrl", K35SNF_XMAC_RX_BASE + K35SNF_XMAC_RX_PKTS_NO_OUT_CTRL, XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_RX_SIZE },
  { "XMAC RX pktsDropAbort", K35SNF_XMAC_RX_BASE + K35SNF_XMAC_RX_PKTS_DROP_ABORT,  XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_RX_SIZE },
  { "XMAC RX pktsSent", K35SNF_XMAC_RX_BASE + K35SNF_XMAC_RX_PKTS_SENT,             XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_RX_SIZE },
  { "XMAC RX bytesSent", K35SNF_XMAC_RX_BASE + K35SNF_XMAC_RX_BYTES_SENT,           XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_RX_SIZE },
  { "XMAC RX SendDup", K35SNF_XMAC_RX_BASE + K35SNF_XMAC_RX_SEND_DUP,               XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_RX_SIZE },
  { "XMAC RX SentWithErr", K35SNF_XMAC_RX_BASE + K35SNF_XMAC_RX_SENT_WITH_ERR,      XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_RX_SIZE },
  { "XMAC RX ExtraMeta", K35SNF_XMAC_RX_BASE + K35SNF_XMAC_RX_EXTRA_META,           XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_RX_SIZE },
  { "XMAC RX CtrlDropped", K35SNF_XMAC_RX_BASE + K35SNF_XMAC_RX_CTRL_DROPPED,       XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_RX_SIZE },
  { "XMAC RX Discarded", K35SNF_XMAC_RX_BASE + K35SNF_XMAC_RX_DISCARDED,            XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_RX_SIZE },
  /* 2 x 32-bit registers (64 bits valid) */
  { "XMAC RX Bytes", K35SNF_XMAC_BASE + K35SNF_XMAC_RX_BYTES,                       CNT_ALIASED | COUNTER_TYPE_2_32BIT, 0, K35SNF_XMAC_SIZE },
  { "XMAC TX Bytes", K35SNF_XMAC_BASE + K35SNF_XMAC_TX_BYTES,                       CNT_ALIASED | COUNTER_TYPE_2_32BIT, 0, K35SNF_XMAC_SIZE },
  { "XMAC RX Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_RX_FRAMES_OK,                  CNT_ALIASED | COUNTER_TYPE_2_32BIT, 0, K35SNF_XMAC_SIZE },
  { "XMAC TX Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_TX_FRAMES_OK,                  CNT_ALIASED | COUNTER_TYPE_2_32BIT, 0, K35SNF_XMAC_SIZE },
  { "XMAC RX Undersized Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_RX_FRAMES_RUNT,     XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC RX Fragmented Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_RX_FRAMES_FRAG,     XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC RX 64B Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_RX_FRAMES_64,              XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC RX 65-127B Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_RX_FRAMES_65_127,      XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC RX 128-255B Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_RX_FRAMES_128_255,    XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC RX 256-511B Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_RX_FRAMES_256_511,    XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC RX 512-1023B Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_RX_FRAMES_512_1023,  XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC RX 1024-Max Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_RX_FRAMES_1024_MAX,   XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC RX Oversized Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_RX_FRAMES_OVERSIZE,  XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC TX 64B Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_TX_FRAMES_64,              XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC TX 65-127B Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_TX_FRAMES_65_127,      XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC TX 128-255B Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_TX_FRAMES_128_255,    XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC TX 256-511B Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_TX_FRAMES_256_511,    XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC TX 512-1023B Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_TX_FRAMES_512_1023,  XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC TX 1024-Max Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_TX_FRAMES_1024_MAX,   XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC TX Oversized Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_TX_FRAMES_OVERSIZE,  XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC RX FCS Errors", K35SNF_XMAC_BASE + K35SNF_XMAC_RX_FCS_BAD,                XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC RX Broadcast Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_RX_BCAST_OK,         XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC RX Multicast Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_RX_MCAST_OK,         XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC RX MAC CTRL Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_RX_MAC_CTRL,          XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC RX Length/Type Range Error", K35SNF_XMAC_BASE + K35SNF_XMAC_RX_LT_RANGE,  XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC RX VLAN Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_RX_VLAN,                  XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC RX Pause Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_RX_PAUSE,                XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC RX MAC Unsupported Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_RX_MAC_UNSUP,  XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC TX Broadcast Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_TX_BCAST_OK,     XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC TX Multicast Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_TX_MCAST_OK,     XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
  { "XMAC TX Underrun Frames", K35SNF_XMAC_BASE + K35SNF_XMAC_TX_UNDERRUN,      XMAC_RX_COUNTER_TYPE, 0, K35SNF_XMAC_SIZE },
#define RSS_COUNTER_TYPE (CNT_ALIASED | COUNTER_DISPLAY_VERBOSE | COUNTER_TYPE_SNAPSHOT)
  { "RSS Mode", K35SNF_RSS_BASE + K35SNF_RSS_SELECT,                                CNT_ALIASED | COUNTER_TYPE_SNAPSHOT, 0, K35SNF_RSS_SIZE },
  { "RSS Mask", K35SNF_RSS_BASE + K35SNF_RSS_MASK,                                  CNT_ALIASED_HEX | COUNTER_TYPE_SNAPSHOT, 0, K35SNF_RSS_SIZE },
  { "RSS Rx Packets", K35SNF_RSS_BASE + K35SNF_RSS_PKT_SENT,                        RSS_COUNTER_TYPE, 0, K35SNF_RSS_SIZE },
  { "RSS Rx Bytes", K35SNF_RSS_BASE + K35SNF_RSS_BYTES_SENT,                        RSS_COUNTER_TYPE, 0, K35SNF_RSS_SIZE },
  { "RSS Dropped Packets", K35SNF_RSS_BASE + K35SNF_RSS_PKT_DROP,                   RSS_COUNTER_TYPE, 0, K35SNF_RSS_SIZE },
  { "RSS Dropped Bytes", K35SNF_RSS_BASE + K35SNF_RSS_BYTES_DROP,                   RSS_COUNTER_TYPE, 0, K35SNF_RSS_SIZE },
  { "RSS Error Packets", K35SNF_RSS_BASE + K35SNF_RSS_PKT_ERR,                      RSS_COUNTER_TYPE, 0, K35SNF_RSS_SIZE },
  { "RSS Error Bytes", K35SNF_RSS_BASE + K35SNF_RSS_BYTES_ERR,                      RSS_COUNTER_TYPE, 0, K35SNF_RSS_SIZE },
  { "RSS Dispatcher Dropped Packets", K35SNF_RSS_BASE + K35SNF_RSS_DISP_PKT_DROP,   RSS_COUNTER_TYPE, 0, K35SNF_RSS_SIZE },
  { "RSS Dispatcher Forwarded Packets", K35SNF_RSS_BASE + K35SNF_RSS_DISP_PKT_FWD,  RSS_COUNTER_TYPE, 0, K35SNF_RSS_SIZE },
  { "LTSSM Not L0 state", K35_LTSSM_NOT_L0                   , CNT_GLOBAL | COUNTER_DISPLAY_VB2, 0, 0 },
};

/* FIXME: RSS registers are available only if PHX_FEATURE_FPGA_MERGE_NEXT_RING */

static int
k35snf_find_counter_idx(uint32_t off)
{
  int len = sizeof(phx_k35snf_counters) / sizeof(struct phx_counter);
  int i;
  for (i = 0; i < len; i++) {
    if (phx_k35snf_counters[i].offset == off)
      return i;
  }
  MX_WARN(("Bug: k35snf_find_counter_idx. offset=0x%x\n", off));
  return 0;
}

static void
k35snf_sniffer_read_stats(mx_instance_state_t *is, phx_snf_stats_t *x)
{
  uint64_t counter[1];
  int ring_drop_idx;
  int drop_idx;
  int error_idx;

  if (PHX_HAS_FEATURE(is, PHX_FEATURE_FPGA_MERGE_NEXT_RING)){
    drop_idx = k35snf_find_counter_idx(K35SNF_RSS_BASE +
                                       K35SNF_RSS_DISP_PKT_DROP);
    error_idx = k35snf_find_counter_idx(K35SNF_RSS_BASE +
                                       K35SNF_RSS_PKT_ERR);
    ring_drop_idx = k35snf_find_counter_idx(K35SNF_RSS_BASE +
                                       K35SNF_RSS_PKT_DROP);
  } else {
    drop_idx = k35snf_find_counter_idx(K35SNF_XMAC_RX_BASE +
                                       K35SNF_XMAC_RX_CTRL_DROPPED);
    error_idx = k35snf_find_counter_idx(K35SNF_XMAC_BASE +
                                        K35SNF_XMAC_RX_FCS_BAD);
    ring_drop_idx = drop_idx;
  }
  myri_counters64_update(is);

  /* 3=Packets Received */
  phx_counters64_get(is, counter, 3, 1, 0);
  x->nic_pkt_recv = counter[0];
  /* 4=Bytes Received */
  phx_counters64_get(is, counter, 4, 1, 0);
  x->nic_bytes_recv = counter[0];
  /* Dropped due to no pages, data or descriptor */
  phx_counters64_get(is, counter, ring_drop_idx, 1, 0);
  x->nic_pkt_no_buffer = counter[0];
  /* Dropped due to NIC unable to push */
  phx_counters64_get(is, counter, drop_idx, 1, 0);
  x->nic_pkt_dropped = counter[0];
  /* Dropped due to bad packet from MAC */
  phx_counters64_get(is, counter, error_idx, 1, 0);
  x->nic_pkt_bad = counter[0];
}

static int
k35snf_get_counter_definitions(mx_instance_state_t *is, const phx_counter_t **counters,
                                   uint32_t *count)
{
  *counters = phx_k35snf_counters;
  *count = sizeof(phx_k35snf_counters) / sizeof(struct phx_counter);
  return 0;
}

static int
k35snf_counter_accum_index(mx_instance_state_t *is, struct phx_fpga_counter *ctr,
                           int nth_endpoint)
{
  int idx = 0;
  uint32_t bits = ctr->class & (COUNTER_TYPE_ENDPOINT | COUNTER_TYPE_ALIASED |
                                COUNTER_TYPE_GLOBAL);
  if (bits == (COUNTER_TYPE_ENDPOINT | COUNTER_TYPE_ALIASED))
    idx = is->port_id * is->phx.pfi->ep_pool.num_function_endpoints
        + nth_endpoint;
  else if (bits == COUNTER_TYPE_ALIASED)
    idx = is->port_id;
  else if (bits == COUNTER_TYPE_ENDPOINT)
    idx = nth_endpoint;
  return idx;
}

static void
k35snf_park_board(mx_instance_state_t *is)
{
  int i;

  for (i = 0; i < is->num_ports; i++) {
    /* Unloading the driver, disable MAC receive so nothing gets into
     * the receive pipeline afterwards. Or it sometimes crashes NIC.
     */
    k35snf_xmac_rx_disable(is, i);
    k35snf_mac_disable(is, i);
  }
}

static void
k35snf_unmap_board(mx_instance_state_t *is)
{
  if (is->phx.nic_mem)
    mx_unmap_io_space(is, is->board_span, (void*)is->phx.nic_mem);
}

static void
k35snf_read_bitstream_sha1(mx_instance_state_t *is, uint32_t sha1_val[4])
{
  uint8_t *r = (uint8_t*)is->phx.nic_mem;
  uint32_t val, *v;
  int i;

  r += PHX_I2C_BASE + 0x180;
  v = sha1_val;
  for (i = 0; i < 4; i ++) {
    val = PHX_REG_READ_4(r);
    /* Byte swap so the caller can printf it and see the bytes in the right order. */
    *v = ((val & 0xff) << 24) |
        (((val >> 8) & 0xff) << 16) |
        (((val >> 16) & 0xff) << 8) |
        ((val >> 24) & 0xff);
    v++;
    r += 4;
  }
}

static int
k35snf_map_board(mx_instance_state_t *is)
{
  uint32_t sha1_val[4];
#if MAL_DEBUG
  uint8_t *base, *r;
  uint32_t val;
#endif

  is->phx.nic_mem = mx_map_pci_space(is, 0, 0, is->board_span);
  /* FIXME: Do we still have this? */
  if (is->phx.nic_mem == NULL) {
    MX_WARN(("phx_map_board failed\n"));
    return ENODEV;
  }
  is->phx.i2c_base_mem = (char *) is->phx.nic_mem + PHX_I2C_BASE;  /* offset to i2c cores */

#if MAL_DEBUG
  MX_INFO(("Board mapped to 0x%p\n", (void *)is->phx.nic_mem));
  MX_INFO(("Sanity checks\n"));

  base = (uint8_t*)is->phx.nic_mem;
  r = base + K35SNF_SYS_BASE + K35SNF_SYS_CONST0;
  MX_INFO(("Read SYS_CONST0 %p\n", r));
  val = PHX_REG_READ_4(r);
  if (val != 0xCAFEF00D)
    goto sanity_fail;
  r = base + K35SNF_SYS_BASE + K35SNF_SYS_CONST1;
  MX_INFO(("Read SYS_CONST1 %p\n", r));
  val = PHX_REG_READ_4(r);
  if (val != 0xDEADBEEF)
    goto sanity_fail;
  r = base + K35SNF_SYS_BASE + K35SNF_SYS_R4;
  MX_INFO(("Write and read SYS_R4 %p\n", r));
  PHX_REG_WRITE_4(r, 0xA1B2C3D4);
  val = PHX_REG_READ_4(r);
  if (val != 0xA1B2C3D4)
    goto sanity_fail;
  r = base + K35SNF_SYS_BASE + K35SNF_SYS_R4;
  MX_INFO(("Write and read SYS_R4 %p\n", r));
  PHX_REG_WRITE_4(r, 0xA1B2C3D4);
  val = PHX_REG_READ_4(r);
  if (val != 0xA1B2C3D4)
    goto sanity_fail;
  r = base + K35SNF_RX_BASE + K35SNF_RX_RX_CONST0;
  MX_INFO(("Read RX_RX_CONST0 %p\n", r));
  val = PHX_REG_READ_4(r);
  if (val != 0xBACECACE)
    goto sanity_fail;
#endif

  k35snf_read_bitstream_sha1(is, sha1_val);
  MX_INFO(("SHA1 sum of Altera bitstream is: %08x%08x%08x%08x\n",
           sha1_val[0], sha1_val[1], sha1_val[2], sha1_val[3]));

  /* K35SNF does not have APU configuration */
  return 0;

#if MAL_DEBUG
sanity_fail:
  MX_INFO(("Failed: read value=0x%08x\n", val));
  k35snf_unmap_board(is);
  return ENODEV;
#endif
}

static void
k35snf_int_enable(mx_instance_state_t *is)
{
  /* Unused */
}

static void
k35snf_int_disable(mx_instance_state_t *is)
{
  int i, vectors;

  if (!MX_IS_PARENT(is)) {
    MX_WARN(("BUG: A child instance should not call k35snf_int_disable."
             " id=%d port_id=%d\n", is->id, is->port_id));
    return;
  }
  /* Disable all interrupts */
  vectors = is->phx.pfi->ep_pool.num_total_endpoints + 1;
  for (i = 0; i < vectors; i++)
    phx_icu_disable_interrupt(is, i);
}

void 
k35snf_set_mac_addr(mx_instance_state_t *is, uint8_t mac[6])
{
#if 0
  MX_INFO(("%s: %02x:%02x:%02x:%02x:%02x:%02x\n", __FUNCTION__,
           mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]));
#endif
  /* MX_INFO(("k35snf_set_mac_addr: unimpl\n")); */
}

int
k35snf_set_promisc(mx_instance_state_t *is, int promisc)
{
  /* The MAC is always promiscuous */
  return 0;
}

int
k35snf_enable_all_multicast(mx_instance_state_t *is)
{
  /* Unused */
  return 0;
}

int
k35snf_disable_all_multicast(mx_instance_state_t *is)
{
  /* Unused */
  return 0;
}

int
k35snf_add_filter(mx_instance_state_t *is, myri_rx_filterq_entry_t *entry)
{
  /* Unused */
  return 0;
}

int
k35snf_remove_filter(mx_instance_state_t *is, myri_rx_filterq_entry_t *entry)
{
  /* Unused */
  return 0;
}


void
phx_k35snf_init_board_ops(void)
{
  phx_k35snf_ops.init                  = k35snf_init_board;
  phx_k35snf_ops.map                   = k35snf_map_board;
  phx_k35snf_ops.unmap                 = k35snf_unmap_board;
  phx_k35snf_ops.enable_interrupt      = k35snf_int_enable;
  phx_k35snf_ops.disable_interrupt     = k35snf_int_disable;
  phx_k35snf_ops.park                  = k35snf_park_board;
  phx_k35snf_ops.get_counter_definitions  = k35snf_get_counter_definitions;
  phx_k35snf_ops.counter_accum_index   = k35snf_counter_accum_index;
  phx_k35snf_ops.init_endpoint_pool    = k35snf_init_endpoint_pool;
  phx_k35snf_ops.ether_endpoint        = k35snf_ether_endpoint;
  phx_k35snf_ops.physical_endpoint     = k35snf_physical_endpoint;
  phx_k35snf_ops.sniffer_endpoint      = k35snf_sniffer_endpoint;
  phx_k35snf_ops.sniffer_read_stats    = k35snf_sniffer_read_stats;
  phx_k35snf_ops.set_intr_coal         = k35snf_set_intr_coal;
  phx_k35snf_ops.set_mac_addr          = k35snf_set_mac_addr;
  phx_k35snf_ops.set_promisc           = k35snf_set_promisc;
  phx_k35snf_ops.setup_hw_caps         = k35snf_setup_hw_caps;
  phx_k35snf_ops.enable_hw_hash        = k35snf_enable_hw_hash;
  phx_k35snf_ops.enable_next_ring_merge= k35snf_enable_next_ring_merge;
  phx_k35snf_ops.enable_all_multicast  = k35snf_enable_all_multicast;
  phx_k35snf_ops.disable_all_multicast = k35snf_disable_all_multicast;
  phx_k35snf_ops.remove_filter         = k35snf_remove_filter;
  phx_k35snf_ops.add_filter            = k35snf_add_filter;
}

