/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2011 by Myricom, Inc.  All rights reserved.          *
 * Copyright 2015 - 2016 by CSP, Inc.  All rights reserved.              *
 *************************************************************************/

#include "mx_arch.h"
#include "mx_instance.h"
#include "phx_hw.h"
#include "phx_common.h"

uint32_t myri_snf_rings = 0;
uint32_t myri_snf_flags = -1;

void myri_dump_snf_state(myri_snf_rx_state_t *snf);


static int snf_rx_sync(myri_snf_rx_state_t *snf, int ring_id,
                       struct snf__rx_sync *rs_update);

static void snf_rx_teardown(mx_endpt_state_t *es);
static int snf_rx_open_handle(mx_endpt_state_t *es, myri_snf_rx_params_t *x);
static int snf_rx_attach(mx_endpt_state_t *es, myri_snf_rx_attach_t *x);
static int snf_pq_attach(mx_endpt_state_t *es, myri_snf_pq_attach_t *x);
static int snf_tx_endpt_enable(mx_endpt_state_t *es, myri_snf_tx_params_t *x);
static int snf_rx_portgroup_del(mx_endpt_state_t *es, mx_endpt_state_t *es_rx_ring);
static struct phx_physical_ep *snf_alloc_tx_phx_endpt(mx_instance_state_t *is);
static void k35snf_tx_endpt_free(mx_endpt_state_t *es);
#if K35SNF_FLUSH_STRATEGY == K35SNF_FLUSH_INTERRUPT
static void k35snf_intr_tx_flush_work(
#if defined(NDIS50_MINIPORT) || defined(NDIS51_MINIPORT)
    NDIS_HANDLE wi,
#endif
    void *arg
#if defined(NDIS60_MINIPORT)
    , NDIS_HANDLE wi
#endif
);
#endif

/* Util. for stats */
static
void
update_and_read_snf_stats(mx_instance_state_t *is, phx_snf_stats_t *x)
{
  if (is->board_ops.sniffer_read_stats)
    is->board_ops.sniffer_read_stats(is, x);
}

/* a = a - b */
static
void
subtract_snf_stats(phx_snf_stats_t *a, phx_snf_stats_t *b)
{
  a->nic_pkt_bad = a->nic_pkt_bad - b->nic_pkt_bad;
  a->nic_pkt_recv = a->nic_pkt_recv - b->nic_pkt_recv;
  a->nic_bytes_recv = a->nic_bytes_recv - b->nic_bytes_recv;
  a->nic_pkt_no_buffer = a->nic_pkt_no_buffer - b->nic_pkt_no_buffer;
  a->nic_pkt_dropped = a->nic_pkt_dropped - b->nic_pkt_dropped;
}

static
int
snf_eop_alloc(mx_endpt_state_t *es, mx_instance_state_t *is, 
              myri_set_endpt_t *set_endpt)
{
  int min_endpt, max_endpt;
  enum myri_endpt_type es_type;
  int endpoint;
  union snf_endpt_state *ess = &es->snf;

  es_type = set_endpt->endpt_type;

  es->phx_ep = NULL; /* initialize to an invalid endpoint */

  if (es_type == MYRI_ES_SNF_TX) {
    struct phx_physical_ep *phx_ep = snf_alloc_tx_phx_endpt(is);
    if (!phx_ep)
      return EBUSY;

    endpoint = phx_ep->id;
    if (is->es[endpoint] != 0) {
      phx_release_fpga_endpoint(is, phx_ep->id);
      return EBUSY;
    }

    if (MX_IS_K35SNF(is)) {
      int rc;
      if ((rc = phx_claim_fpga_endpoint(is, phx_ep->id, 0))) {
        if (is->port_id == 0) {
          if (!k35snf_ether_remove_endpoints(is)) {
            MX_INFO(("Removed the endpoints from the ethernet interface.\n"));
            rc = phx_claim_fpga_endpoint(is, phx_ep->id, 0);
          }
        }
        if (rc) {
          MX_WARN(("SNF port is already in use.\n"));
          return rc;
        }
      }
    }

    es->phx_ep = phx_ep;

    es->session_id = 1;

    ess->es_tx.tx_intr_requested = 0;
    mx_spin_lock_init(&ess->es_tx.tx_lock, is, es->endpt, "es_tx.tx_lock");
#if K35SNF_FLUSH_STRATEGY == K35SNF_FLUSH_INTERRUPT
    MX_LXX_INIT_DELAYED_WORK(&es->arch.snf.tx_flush_work,
                             (void *)k35snf_intr_tx_flush_work, es);
#endif
    is->es[endpoint] = es;
    es->flags |= MX_ES_EARLY_ENABLE;

    /* Save es in the parent to handle TX interrupts */
    if (MX_IS_CHILD(is))
      is->parent_port->es[endpoint] = es;
  }
  else {
    if (es_type == MYRI_ES_SNF_RX) {
      min_endpt = PHX_SNF_MAX_ENDPOINTS;
      max_endpt = 2*PHX_SNF_MAX_ENDPOINTS;
      es->flags |= MX_ES_EARLY_ENABLE;
      es->snf.es_rx.rx_ring_id = SNF_RING_UNUSED;
    }
    else if (es_type == MYRI_ES_SNF_RX_RING) {
      min_endpt = 2*PHX_SNF_MAX_ENDPOINTS;
      max_endpt = 3*PHX_SNF_MAX_ENDPOINTS;
      es->flags |= MX_ES_EARLY_ENABLE;
      es->snf.es_rx.rx_ring_id = SNF_RING_UNUSED;
    }
    else if (es_type == MYRI_ES_SNF_RX_GROUP) {
      min_endpt = 3*PHX_SNF_MAX_ENDPOINTS;
      max_endpt = 4*PHX_SNF_MAX_ENDPOINTS;
      TAILQ_INIT(&es->snf.es_pg.pg_waitq);
    }
    else if (es_type == MYRI_ES_SNF_HANDLE) {
      min_endpt = 4*PHX_SNF_MAX_ENDPOINTS;
      max_endpt = 5*PHX_SNF_MAX_ENDPOINTS;
    }
    else if (es_type == MYRI_ES_SNF_RX_PQ) {
      min_endpt = 5*PHX_SNF_MAX_ENDPOINTS;
      max_endpt = 6*PHX_SNF_MAX_ENDPOINTS;
      es->flags |= MX_ES_EARLY_ENABLE;
    }
    else
      return EINVAL;

    for (endpoint = min_endpt; endpoint < max_endpt; endpoint++) {
      if (is->es[endpoint] == 0)
          break;
    }
    if (endpoint == max_endpt)
      return EBUSY;
    es->flags |= MX_ES_SOFT;
    is->es[endpoint] = es;
  }

  es->endpt = endpoint;
  es->is = is;

  return 0;
}

/* XXX 
 * Sniffer requires early enable, can't support splitting opening
 * from enabling... yet.
 */
static
int
snf_eop_enable(mx_endpt_state_t *es, uaddr_t in)
{
  /* mx_instance_state_t *is = es->is; */
  uaddr_t out = in;
  int status;

  mx_mutex_enter(&es->sync);

  switch (es->es_type) {

  case MYRI_ES_SNF_TX:
  {
    myri_snf_tx_params_t x;
#if !(MAL_OS_LINUX || MAL_OS_FREEBSD || MAL_OS_WINDOWS)
    status = EINVAL;
    break;
#endif

    if ((status = mx_copyin(in, &x, sizeof(x), es->is_kernel)))
      break;
    status = snf_tx_endpt_enable(es, &x);
    if (status)
      break;
    status = mx_copyout(&x, out, sizeof(x), es->is_kernel);
  }
  break;

  case MYRI_ES_SNF_RX_RING:
  {
    myri_snf_rx_attach_t x;
    if ((status = mx_copyin(in, &x, sizeof(x), es->is_kernel)))
      break;
    if ((status = snf_rx_attach(es, &x)))
      break;
    status = mx_copyout(&x, out, sizeof(x), es->is_kernel);
  }
  break;

  case MYRI_ES_SNF_RX:
  {
    myri_snf_rx_params_t x;
    if ((status = mx_copyin(in, &x, sizeof(x), es->is_kernel)))
      break;
    if ((status = snf_rx_open_handle(es, &x)))
      break;
    status = mx_copyout(&x, out, sizeof(x), es->is_kernel);
  }
  break;

  case MYRI_ES_SNF_RX_PQ:
  {
    myri_snf_pq_attach_t x;
    if ((status = mx_copyin(in, &x, sizeof(x), es->is_kernel)))
      break;
    if ((status = snf_pq_attach(es, &x)))
      break;
    status = mx_copyout(&x, out, sizeof(x), es->is_kernel);
  }
  break;

  case MYRI_ES_SNF_HANDLE:
    status = 0;
    break;

  default:
    status = EINVAL;
  }

  mx_mutex_exit(&es->sync);

  return status;
}

#if K35SNF_FLUSH_STRATEGY == K35SNF_FLUSH_INTERRUPT
static
int
k35snf_intr_tx_try_flush(mx_endpt_state_t *es)
{
  struct k35snf__tx_shared_params *sp = &es->snf.es_tx.u.k35snf.sp;
  struct k35snf__tx_shared_state *ss = sp->ss;

  if (es->flags & MX_ES_CLOSING) /* stop trying to flush */
    return 1;

  if (!snf__spin_trylock(&ss->tx_lock))
    return 0;

  k35snf_tx_update_completion(sp);
  /* No packets in progress, and there are queued packets.  So flush now. */
  if (ss->queued_desc && ss->inflight == 0)
    k35snf_tx_flush(sp);
  snf__spin_unlock(&ss->tx_lock);
  return 1;
}

static
void
k35snf_intr_tx_flush_work(
#if defined(NDIS50_MINIPORT) || defined(NDIS51_MINIPORT)
    PNDIS_WORK_ITEM wi,
#endif
    void *arg
#if defined(NDIS60_MINIPORT)
    , NDIS_HANDLE wi
#endif
)
{
  mx_endpt_state_t *es = MX_WORK_STRUCT_ARG_CONTAINER_OF(
      arg, mx_endpt_state_t, arch.snf.tx_flush_work);

#if defined(NDIS50_MINIPORT) || defined(NDIS51_MINIPORT)
  NdisFreeMemory(wi, sizeof (*wi), 0);
#elif defined(NDIS60_MINIPORT)
  NdisFreeIoWorkItem(wi);
#endif

  if (!k35snf_intr_tx_try_flush(es)) {
#ifdef _WIN32 /* Any NDIS */
    InterlockedDecrement(&es->arch.snf.work_cnt);
#endif
    myri_schedule_delayed_work(&es->arch.snf.tx_flush_work, 1);
    return;
  } else
    phx_icu_enable_interrupt(es->is, es->phx_ep->icu);

#ifdef _WIN32 /* Any NDIS */
  KeSetEvent(&es->arch.snf.no_work, 0, FALSE);
  InterlockedDecrement(&es->arch.snf.work_cnt);
#endif
}

static
void
k35snf_handle_tx_intr(mx_endpt_state_t *es)
{
  if (!k35snf_intr_tx_try_flush(es)) {
    /* The user has locked the shared state.  Schedule another check.
     * Until we check the completion page, do not enable the interrupt.
     */
    myri_schedule_delayed_work(&es->arch.snf.tx_flush_work, 0);
  } else {
    /* We have checked the completion page, possibly did a flush, and
     * possibly wrote to ACK as necessary.  So re-enable the interrupt.
     */
    phx_icu_enable_interrupt(es->is, es->phx_ep->icu);
  }
}
#endif /* K35SNF_FLUSH_INTERRUPT */

static
void
snf_eop_intr_is(mx_instance_state_t *is, void *context)
{
  myri_snf_rx_state_t *snf = &is->snf;
  struct myri_snf_waitq waitq;
  struct myri_snf_waiter *waiter;
  int rx_intr_requested;

  if (snf->state == SNF_RX_S_FREE)
    return;

  /* Step 1: If interrupt was requested, save each wakeup for further
   * processing w/o lock held */
  mx_spin_lock(&snf->rx_intr_lock);
  rx_intr_requested = snf->rx_intr_requested;
  snf->rx_intr_requested = 0;
  TAILQ_INIT(&waitq);
  if (rx_intr_requested && !TAILQ_EMPTY(&snf->rx_intr_waitq)) {
    while ((waiter = TAILQ_FIRST(&snf->rx_intr_waitq)) != NULL) {
      TAILQ_REMOVE(&snf->rx_intr_waitq, waiter, link);
      waiter->intr_requested = 0;
      waiter->intr_wake_pending = 1;
      TAILQ_INSERT_TAIL(&waitq, waiter, wake_link);
    }
  }
  mx_spin_unlock(&snf->rx_intr_lock);

  /* Step 2: Do all the (potentially slow) wakes w/o lock */
  TAILQ_FOREACH(waiter, &waitq, wake_link) {
    if (waiter->wait_sync)
      mx_wake(waiter->wait_sync);
    waiter->intr_wake_pending = 0;
  }
}

static
int
snf_eop_intr_endpt(mx_endpt_state_t *es, int flags)
{
  union snf_endpt_state *ess = &es->snf;
  int status = 0;

  if (es->flags & MX_ES_CLOSING) /* stop trying to flush */
    return 0;

  /* MX_INFO(("eop_intr_endpt. es_type=%d\n", es->es_type)); */

  if (es->es_type == MYRI_ES_SNF_TX) {
    if (flags & PHX_INTR_ENDPT_WAKE_TX) {
      int intr_requested;
      mx_spin_lock(&ess->es_tx.tx_lock);
      intr_requested = ess->es_tx.tx_intr_requested;
      ess->es_tx.tx_intr_requested = 0;
      mx_spin_unlock(&ess->es_tx.tx_lock);
#if K35SNF_FLUSH_STRATEGY == K35SNF_FLUSH_INTERRUPT
      if (MX_IS_K35SNF(es->is))
        k35snf_handle_tx_intr(es);
#endif
      if (intr_requested)
        mx_wake(&es->wait_sync);
    }
  }
  else
    status = EINVAL;

  return status;
}

static
void
snf_eop_close(mx_endpt_state_t *es)
{
  mx_instance_state_t *is = es->is;

  /* Special case K35SNF TX to stop interrupts early on, before we clear
   * is->es below.
   */
  if (es->es_type == MYRI_ES_SNF_TX && MX_IS_K35SNF(is)) {
#if K35SNF_FLUSH_STRATEGY == K35SNF_FLUSH_INTERRUPT
    myri_cancel_delayed_work(&es->arch.snf.tx_flush_work);
#endif
    /* Stop the interrupt and let the ongoing one drain */
    phx_icu_disable_interrupt(is, es->phx_ep->icu);
    mx_spin(1000); /* 1msec */
    /* Disable again as the interrupt handler may have enabled it. */
    phx_icu_disable_interrupt(is, es->phx_ep->icu);
  }

  mx_mutex_enter(&is->sync);
  mal_always_assert(es->endpt >= 0 && es->endpt < is->es_count);
  is->es[es->endpt] = 0;
  mx_mutex_exit(&is->sync);

  if (es->es_type == MYRI_ES_SNF_RX || es->es_type == MYRI_ES_SNF_RX_RING || es->es_type == MYRI_ES_SNF_RX_PQ)
    snf_rx_teardown(es);
  else if (es->es_type == MYRI_ES_SNF_TX) {
    if (MX_IS_K35SNF(is)) {
      k35snf_tx_stop(is);
      phx_apu_channel_reset(is, es->snf.es_tx.u.k35snf.apu_data_ch);
      phx_apu_channel_reset(is, es->snf.es_tx.u.k35snf.apu_desc_ch);
      /* Reset ICU too */
      phx_icu_reset_endpt(is, es->phx_ep);
      k35snf_tx_endpt_free(es);
    } else {
      struct es_tx_phx *tx = &es->snf.es_tx.u.phx;
      /* Stop the interrupt first and let the ongoing one drain */
      phx_icu_disable_interrupt(is, es->phx_ep->icu);
      mx_spin(1000); /* 1msec */
#if MAL_OS_WINDOWS
      phx_win_free_gamma_pages(is, &tx->tx_dma_pages, tx->tx_dma_mdl);
      tx->tx_dma_mdl = NULL;
#else
      phx_free_pages(is, &tx->tx_dma_pages);
#endif
    }
    mx_spin_lock_destroy(&es->snf.es_tx.tx_lock);
    /* Clear es in the parent too */
    if (MX_IS_CHILD(is))
      is->parent_port->es[es->endpt] = 0;
    phx_release_fpga_endpoint(is, es->phx_ep->id);

    /* The injection endpoint is closed so accumulate counters. */
    myri_counters64_ep_close(is, es->phx_ep->icu - 1 /* n'th within the function */);
    phx_sw_counters_ep_close(es);

    /* Give the port 0 endpoint back to the ethernet driver */
    if (MX_IS_K35SNF(is) && is->port_id == 0)
      k35snf_ether_return_endpoint(is);
  }
  else if (es->es_type == MYRI_ES_SNF_RX_GROUP)
    snf_rx_portgroup_del(es, NULL);
}

static struct phx_physical_ep *
phx_alloc_tx_endpt(mx_instance_state_t *is)
{
  int i, max_ep = is->phx.pfi->ep_pool.num_total_endpoints;
  for (i = 0; i < max_ep; i++) {
    if (!phx_claim_fpga_endpoint(is, i, 0))
      return &is->phx.pfi->ep_pool.endpoints[i];
  }
  return NULL;
}

static struct phx_physical_ep *
k35snf_alloc_tx_endpt(mx_instance_state_t *is)
{
  struct phx_fpga_endpoint_pool *pool = &is->phx.pfi->ep_pool;
  return &pool->endpoints[is->port_id * 2];
}

static struct phx_physical_ep *
snf_alloc_tx_phx_endpt(mx_instance_state_t *is)
{
  if (MX_IS_K35SNF(is))
    return k35snf_alloc_tx_endpt(is);
  else
    return phx_alloc_tx_endpt(is);
}

static int
phx_tx_endpt_enable(mx_endpt_state_t *es, myri_snf_tx_params_t *x)
{
  int i;
  union snf_endpt_state *ess = &es->snf;
  struct es_tx_phx *tx = &ess->es_tx.u.phx;
  uint32_t off;

  if (es->es_type != MYRI_ES_SNF_TX)
    return EINVAL;

#if MAL_DEBUG
  MX_INFO(("phx_tx_endpt_enable: phx_tx_endpt=%d\n", es->phx_ep->tx_ep));
#endif

  tx->tx_regs = phx_endpoint_nic_mem(es->is, es->phx_ep->tx_ep);
  tx->tx_desc_buf = phx_ep_desc_fifo(tx->tx_regs);
  tx->tx_pio_buf = phx_ep_pio_fifo(es->is, es->phx_ep->tx_ep);
  tx->tx_pio_buf_size = es->is->phx.pfi->tx_config.dbltx_pio_length;

#if MAL_OS_WINDOWS
  if (phx_win_alloc_gamma_pages(es->is, &tx->tx_dma_pages, PHX_TX_DMA_PAGES_PER_ENDPOINT,
                                &tx->tx_dma_mdl)) {
    MX_WARN(("phx_tx_endpt_enable: failed to allocate DMA pages\n"));
    return ENOMEM;
  }
#else
  if (phx_alloc_pages(es->is, &tx->tx_dma_pages, PHX_TX_DMA_PAGES_PER_ENDPOINT)) {
    MX_WARN(("phx_tx_endpt_enable: failed to allocate DMA pages\n"));
    return ENOMEM;
  }
#endif

  /* Push all DMA page addresses to the NIC. */
  phx_write_ep_reg_tx_addr_buffer_clear(tx->tx_regs);
  for (i = 0; i < tx->tx_dma_pages.n_pages; i++) {
    phx_write_ep_reg_tx_addr_load(tx->tx_regs,
                                  tx->tx_dma_pages.pages[i].dma.low,
                                  tx->tx_dma_pages.pages[i].dma.high);
  }
  phx_write_ep_reg_tx_addr_buffer_finish(tx->tx_regs);

  phx_icu_enable_interrupt(es->is, es->phx_ep->icu);

  /* HACK. See phx_dbl_ops.c.  Is this still necessary? */
  {
    uint32_t compl_cnt;
    uint32_t *compl;
    compl_cnt = phx_ep_tx_completion_count(tx->tx_regs);
    compl = (uint32_t *)((uintptr_t)es->is->tx_done.addr);
    compl[es->phx_ep->tx_ep * 2] = compl_cnt;

#if MAL_DEBUG
    MX_INFO(("Setting endpt completion count page addr %p to %d\n", compl, compl_cnt));
#endif
  }

  x->dmasend_threshold = phx_get_dmasend_threshold(es->is);
  x->phx_endpt = es->phx_ep->tx_ep;
  x->phx_port = es->is->phx.pci_info.func;

  /* SNF does not have an ioctl/function equivalent to dbl_get_copyblocks().
   * We pass the relevant mapping info at the time of open+enable.
   */
  off = 0;

  /* A.1: TX endpoint pages */
  x->nic_mem_tx_offset = off;
  x->nic_mem_tx_len = es->is->phx.pfi->tx_config.dbltx_core_length;
  off += x->nic_mem_tx_len;

  /* A.1b: TX PIO endpoint pages. */
  x->nic_mem_tx_pio_offset = off;
  x->nic_mem_tx_pio_len = es->is->phx.pfi->tx_config.dbltx_pio_length;
  off += x->nic_mem_tx_pio_len;

  /* A.2: per-endpoint flow window */
  x->flow_window_offset = off;
  x->flow_window_len = (es->flow_window ? MX_PAGE_SIZE : 0);
  off += x->flow_window_len;

  /* A.3: TX completion counters */
  x->send_compcnt_offset = off;
  x->send_compcnt_len = (es->is->tx_done.addr ? MX_PAGE_SIZE : 0);
  off += x->send_compcnt_len;

  /* A.4: TX DMA buffer */
  x->tx_dmabuf_offset = off;
  x->tx_dmabuf_len = tx->tx_dma_pages.n_pages * PAGE_SIZE;
  off += x->tx_dmabuf_len;

  sprintf(es->opener.user_info, "tx ring %d", es->endpt);

  /* Injection endpoints use software counters */
  phx_sw_counters_ep_open(es);

  /* Pass the board type to the user */
  x->board_type = SNF_BOARD_TYPE_DBL4P;
  return 0;
}

static void
k35snf_tx_endpt_free(mx_endpt_state_t *es)
{
  mx_instance_state_t *is = es->is;
  union snf_endpt_state *ess = &es->snf;
  struct es_tx_k35snf *tx = &ess->es_tx.u.k35snf;

  if (tx->completion.addr)
    mx_free_dma_page(is, &tx->completion.addr, &tx->completion.pin);
  phx_free_rx_ring(is, &tx->data_ring);
  phx_free_rx_ring(is, &tx->desc_ring);
}

static int
k35snf_tx_endpt_enable(mx_endpt_state_t *es, myri_snf_tx_params_t *x)
{
  mx_instance_state_t *is = es->is;
  union snf_endpt_state *ess = &es->snf;
  struct es_tx_k35snf *tx = &ess->es_tx.u.k35snf;
  mx_page_pin_t *page;
  uint32_t off;
  int rc;

  if (es->es_type != MYRI_ES_SNF_TX)
    return EINVAL;

  if ((x->tx_data_ring_len % PHX_APU_RING_SIZE_MULTIPLE) ||
      (x->tx_desc_ring_len % PHX_APU_RING_SIZE_MULTIPLE)) {
    MX_WARN(("%s: k35snf_tx_endpt_enable: desc or data ring size is not"
             " a multiple of 2MB. data_len=%u desc_len=%u\n",
             is->is_name, x->tx_data_ring_len, x->tx_desc_ring_len));
    return EINVAL;
  }

#if MAL_DEBUG
  MX_INFO(("k35snf_tx_endpt_enable: phx_tx_endpt=%d\n", es->phx_ep->tx_ep));
#endif

  k35snf_tx_stop(is); /* Also resets */

  /* Remember the base offset for mmap later on */
  tx->tx_regs = k35snf_tx_base(is);

  /* Allocate the data ring */
  rc = phx_alloc_rx_ring(is, &tx->data_ring, x->tx_data_ring_len, 0);
  if (rc) {
    MX_WARN(("Failed to allocate SNF TX data ring\n"));
    rc = ENOMEM;
    goto exit;
  }

  /* Allocate the descriptor ring */
  rc = phx_alloc_rx_ring(is, &tx->desc_ring, x->tx_desc_ring_len, 0);
  if (rc) {
    MX_WARN(("Failed to allocate SNF TX descriptor ring\n"));
    rc = ENOMEM;
    goto exit;
  }

  /* Completion page */
  rc = mx_alloc_zeroed_dma_page_relax_order(is, &tx->completion.addr,
                                            &tx->completion.pin);
  if (rc) {
    MX_WARN(("Failed to allocate TX completion page\n"));
    rc = ENOMEM;
    goto exit;
  }

  /* Set up APUs */
  tx->apu_data_ch = phx_apu_channel_id(is, 1, 0, 1);
  tx->apu_desc_ch = phx_apu_channel_id(is, 0, 0, 1);
#if MAL_DEBUG
  MX_INFO(("Sniffer TX APU channels: data=%d desc=%d\n",
           tx->apu_data_ch, tx->apu_desc_ch));
#endif
  phx_apu_channel_reset(is, tx->apu_data_ch);
  phx_apu_channel_reset(is, tx->apu_desc_ch);
  /* Pointer to the first page */
  page = tx->data_ring.alpha.pages;
  phx_apu_channel_start(is, tx->apu_data_ch,
                        page->dma.low, page->dma.high);
  page = tx->desc_ring.alpha.pages;
  phx_apu_channel_start(is, tx->apu_desc_ch,
                        page->dma.low, page->dma.high);
#if MAL_DEBUG
  MX_INFO(("Starting SNF TX endpoint\n"));
#endif
  /* Push the completion page address to the NIC */
  k35snf_tx_start(is, tx->completion.pin.dma.low,
                  tx->completion.pin.dma.high,
                  tx->data_ring.gamma.n_pages,
                  tx->desc_ring.size / sizeof(k35snf_tx_desc_t));

  /* Setup the scale factor for injection pacing */
  x->tx_scale_factor = k35snf_tx_get_scale_factor(is);

#if K35SNF_FLUSH_STRATEGY == K35SNF_FLUSH_INTERRUPT
  /* Interrupt-driven flush needs interrupt right away */
  phx_icu_enable_interrupt(is, es->phx_ep->icu);

  {
    /* Set up the shared area */
    struct k35snf__tx_shared_params *sp = &es->snf.es_tx.u.k35snf.sp;

    sp->prod_doorbell = (volatile uint32_t*)
        (tx->tx_regs + K35SNF_TX_DESC_PRODW_ADVANCE);
    sp->compl_ack = (volatile uint32_t*)
        (tx->tx_regs + K35SNF_TX_COMPL_COUNT_ACK);
    sp->compl_page = (volatile k35snf_tx_completion_t*)tx->completion.addr;
    sp->compl_threshold = 128; /* FIXME: Make it tunable */
    /* The user library resets ss later on */
    sp->ss = (struct k35snf__tx_shared_state*)es->flow_window;
  }
#endif

  x->dmasend_threshold = 0;
  x->phx_endpt = es->phx_ep->tx_ep;
  x->phx_port = is->port_id;

  /* Pass mmap offsets to the user */
  off = 0;

  /* A.1: SnifferTX registers */
  x->nic_mem_tx_offset = off;
  x->nic_mem_tx_len = MX_PAGE_SIZE;
  off += x->nic_mem_tx_len;

  /* A.2: per-endpoint flow window */
  x->flow_window_offset = off;
  x->flow_window_len = MX_PAGE_SIZE;
  off += x->flow_window_len;

  /* A.3: completion page */
  x->send_compcnt_offset = off;
  x->send_compcnt_len = MX_PAGE_SIZE;
  off += x->send_compcnt_len;

  /* A.4: data ring */
  x->tx_data_ring_offset = off;
  off += x->tx_data_ring_len;

  /* A.5: descriptor ring */
  x->tx_desc_ring_offset = off;
  off += x->tx_desc_ring_len;

  sprintf(es->opener.user_info, "tx ring %d", es->endpt);

  /* Injection endpoints use software counters */
  phx_sw_counters_ep_open(es);
  /* Pass the board type to the user */
  x->board_type = SNF_BOARD_TYPE_K35SNF;
  return 0;

exit:
  k35snf_tx_endpt_free(es);
  return rc;
}

static int
snf_tx_endpt_enable(mx_endpt_state_t *es, myri_snf_tx_params_t *x)
{
  if (MX_IS_K35SNF(es->is))
    return k35snf_tx_endpt_enable(es, x);
  else
    return phx_tx_endpt_enable(es, x);
}

void 
myri_dump_snf_state(myri_snf_rx_state_t *snf)
{
  int nrings;
  MX_WARN(("attach_mask=%#x, refcount=%d, state=%d, rx_s=%p\n",
           snf->sync_s.attach_mask, snf->refcount, snf->state, snf->rx_s));
  MX_WARN(("desc_ring [size=%lu, gamma.n_pages=%d\n",
           (unsigned long) snf->desc_ring.size, snf->desc_ring.gamma.n_pages));
  MX_WARN(("prehash_desc_ring [size=%lu, gamma.n_pages=%d\n",
           (unsigned long) snf->prehash_desc_ring.size, snf->prehash_desc_ring.gamma.n_pages));
  MX_WARN(("data_ring [size=%lu, gamma.n_pages=%d\n",
           (unsigned long) snf->data_ring.size, snf->data_ring.gamma.n_pages));
  nrings = snf->num_rings_requested;
  if (!nrings)
    nrings = MYRI_SNF_MAX_RINGS;
  MX_WARN(("rings_attached=%d\n", snf->num_rings_attached));
}

static
int
snf_eop_poll_events_pending(mx_endpt_state_t *es)
{
  int revents = 0;
  uint8_t *tx_done_vec;

  if (es->flow_window && es->es_type == MYRI_ES_SNF_TX) {
    tx_done_vec = (uint8_t *)(uintptr_t) es->is->tx_done.pin.va;
    if ((uint8_t) es->flow_window->tx_flow.flow_desc != tx_done_vec[es->phx_ep->tx_ep])
      revents |= MYRI_WAIT_TX_EVENT;
  }
  return revents;
}

static
uint32_t
snf_rx_flow_desc_latest(myri_snf_rx_state_t *snf)
{
  struct snf_sync_state *ss = &snf->sync_s;
  struct snf__rx_state *rx_s;
  int i, first = 1;
  uint32_t desc_cnt, desc_cnt_latest, desc_size;

  desc_cnt_latest = 0;

  mx_spin_lock(&ss->rs_lock);

  desc_size = snf->rx_desc_ring_count;

  SNF_FFS_FOREACH(ss->attach_mask, i) {
    rx_s = SNF__RX_STATE(snf->rx_s, i);
    desc_cnt = (uint32_t)rx_s->evq.evq_desc_cnt;
    if (first) {
	desc_cnt_latest = desc_cnt;
	first = 0;
    }
    else if (desc_cnt - desc_cnt_latest < desc_size)
	desc_cnt_latest = desc_cnt;
  }

  mx_spin_unlock(&ss->rs_lock);

  return desc_cnt_latest;
}

static
uint32_t
snf_intr_rx_group_req(mx_endpt_state_t *es)
{
  struct snf_endpt_portgroup *ep_pg = &es->snf.es_pg;
  struct myri_snf_waiter *waiter;
  uint32_t intr_flags_set = 0;
  uint32_t desc_cnt;
  unsigned long flags;
  myri_snf_rx_state_t *snf;

  flags = 0; /* -Wunused on non-linux */

  if (es->es_type != MYRI_ES_SNF_RX_GROUP)
    return 0;

  TAILQ_FOREACH(waiter, &ep_pg->pg_waitq, pg_link) {
    snf = waiter->snf;
    mx_spin_lock_irqsave(&snf->rx_intr_lock, flags);
    if (snf->rx_intr_requested == 0) {
      desc_cnt = snf_rx_flow_desc_latest(snf);
      phx_rx_wake_request(snf->is, snf->phx_ep->rx_ep, desc_cnt);
      intr_flags_set |= MYRI_WAIT_RX_EVENT;
      snf->rx_intr_requested = 1;
    }
    if (waiter->intr_requested == 0) {
      waiter->intr_requested = 1;
      TAILQ_INSERT_TAIL(&snf->rx_intr_waitq, waiter, link);
    }
    mx_spin_unlock_irqrestore(&snf->rx_intr_lock, flags);
  }

  return intr_flags_set;
}

/* FIXME.  Duplicate code.  snf_eop_poll_wait and snf_intr_rx_group_req. */
static
int
snf_pq_group_wait(mx_endpt_state_t *es, int timeout_ms)
{
  struct snf_endpt_portgroup *ep_pg = &es->snf.es_pg;
  struct myri_snf_waiter *waiter;
  uint32_t desc_cnt;
  unsigned long flags;
  myri_snf_rx_state_t *snf;
  int status;

  if (timeout_ms < 0)
    timeout_ms = MX_MAX_WAIT;

  flags = 0;
  TAILQ_FOREACH(waiter, &ep_pg->pg_waitq, pg_link) {
    snf = waiter->snf;
    mx_spin_lock_irqsave(&snf->rx_intr_lock, flags);
    if (snf->rx_intr_requested == 0) {
      /* This might be stale since we run this code after releasing pq locks.
       * So be it.
       */
      desc_cnt = (uint32_t)snf->pq_s->rw.evq_desc_cnt;
      phx_rx_wake_request(snf->is, snf->phx_ep->rx_ep, desc_cnt);
      snf->rx_intr_requested = 1;
    }
    if (waiter->intr_requested == 0) {
      waiter->intr_requested = 1;
      TAILQ_INSERT_TAIL(&snf->rx_intr_waitq, waiter, link);
    }
    mx_spin_unlock_irqrestore(&snf->rx_intr_lock, flags);
  }
  status = mx_sleep(&es->wait_sync, timeout_ms, MX_SLEEP_INTR);
  return status;
}

static
int
snf_vqwait_req(mx_endpt_state_t *es, int timeout_ms)
{
  myri_snf_rx_state_t *snf = &es->is->snf;
  union snf_endpt_state *ess = &es->snf;
  struct myri_snf_waiter *vq_wait;
  int status;

  if (timeout_ms < 0)
    timeout_ms = MX_MAX_WAIT;

  if (es->es_type == MYRI_ES_SNF_RX_GROUP)
    vq_wait = &ess->es_pg.vq_waitsync;
  else
    vq_wait = &ess->es_rx.vq_waitsync;

  /* FIXME.  Use a lighter spinlock instead of heavy mutex. */
  mx_mutex_enter(&snf->sync);
  if (vq_wait->intr_requested == 0) {
    vq_wait->intr_requested = 1;
    TAILQ_INSERT_TAIL(&snf->vq_waitq, vq_wait, link);
  }

  mx_mutex_exit(&snf->sync);
  status = mx_sleep(&es->vq_sync, timeout_ms, MX_SLEEP_INTR);
  return status;
}

static
void
snf_vqwake_req(myri_snf_rx_state_t *snf, int locked)
{
  struct myri_snf_waiter *vq_waiter;

  if (!locked)
   mx_mutex_enter(&snf->sync);

  while ((vq_waiter = TAILQ_FIRST(&snf->vq_waitq)) != NULL) {
    vq_waiter->intr_requested = 0;
    if (vq_waiter->wait_sync)
      mx_wake(vq_waiter->wait_sync);
    TAILQ_REMOVE(&snf->vq_waitq, vq_waiter, link);
  }

  if (!locked)
    mx_mutex_exit(&snf->sync);
}

static
uint32_t
snf_intr_req(mx_endpt_state_t *es, uint32_t intr_flags)
{
  myri_snf_rx_state_t *snf = &es->is->snf;
  union snf_endpt_state *ess = &es->snf;
  unsigned long flags;
  uint32_t intr_flags_set = 0;
  uint32_t evq_desc_cnt;

  flags = 0; /* -Wunused on non-linux */

  /* Ensure only one interrupt request is in flight */
  if (intr_flags & MYRI_WAIT_RX_EVENT) {

    mx_spin_lock_irqsave(&snf->rx_intr_lock, flags);

    if (es->es_type == MYRI_ES_SNF_RX_PQ)
      evq_desc_cnt = (uint32_t)ess->es_rx.pq_s->rw.evq_desc_cnt;
    else
      evq_desc_cnt = (uint32_t)ess->es_rx.rx_s->evq.evq_desc_cnt;
    if (snf->rx_intr_requested == 0) {
      phx_rx_wake_request(es->is, snf->phx_ep->rx_ep, evq_desc_cnt);
      intr_flags_set |= MYRI_WAIT_RX_EVENT;
      TAILQ_INIT(&snf->rx_intr_waitq);
      snf->rx_intr_requested = 1;
    }
    if (ess->es_rx.rx_waitsync.intr_requested == 0) {
      ess->es_rx.rx_waitsync.intr_requested = 1;
      TAILQ_INSERT_TAIL(&snf->rx_intr_waitq, &ess->es_rx.rx_waitsync, link);
    }
    mx_spin_unlock_irqrestore(&snf->rx_intr_lock, flags);
  }

  if ((intr_flags & MYRI_WAIT_TX_EVENT) && es->flow_window) {
    mx_spin_lock_irqsave(&ess->es_tx.tx_lock, flags);

    if (ess->es_tx.tx_intr_requested == 0) {
      if (MX_IS_K35SNF(es->is)) {
#if K35SNF_FLUSH_STRATEGY == K35SNF_FLUSH_CHUNK
        phx_icu_enable_interrupt(es->is, es->phx_ep->icu);
#endif
      } else {
        uint32_t flow_val = es->flow_window->tx_flow.flow_desc;
        phx_tx_wake_request(phx_endpoint_nic_mem(es->is, es->phx_ep->tx_ep), flow_val);
      }
      ess->es_tx.tx_intr_requested = 1;
      intr_flags_set |= MYRI_WAIT_TX_EVENT;
    }
    mx_spin_unlock_irqrestore(&ess->es_tx.tx_lock, flags);
  }

  return intr_flags_set;

  if (intr_flags_set) { /* Requested at least one interrupt, flush doorbell write */
    MAL_STBAR();
  }
  return intr_flags_set;
}

static
void
snf_eop_poll_prep_wait(mx_endpt_state_t *es)
{
  /* Only TX interrupt for now */
  snf_intr_req(es, MYRI_WAIT_TX_EVENT);
}

static
int
snf_eop_poll_wait(mx_endpt_state_t *es, int timeout_ms, int events)
{
  int status;

  if (timeout_ms < 0)
    timeout_ms = MX_MAX_WAIT;

  if (timeout_ms == 0) {
    /* We're not going to block below in sleep, but claim the interrupt if it
     * is available.  This is mostly useful when poll_wait is called with
     * timeout_ms==0 when the endpoint is used with poll() */
    status = mx_sleep(&es->wait_sync, 0, MX_SLEEP_INTR);
    if (status != EAGAIN) {
      MX_WARN(("unexpected return from poll_wait: %d\n", status));
      return status;
    }
  }
  else {
    if (events) {
      if (es->es_type == MYRI_ES_SNF_RX_GROUP) 
        snf_intr_rx_group_req(es);
      else
        snf_intr_req(es, events);
    }
    status = mx_sleep(&es->wait_sync, timeout_ms, MX_SLEEP_INTR);

    return status;
  }

  return status;
}

static int
k35snf_eop_mmap(mx_endpt_state_t *es, unsigned long req,
                unsigned long *off, void **kva,
                int *mem_type, int *readonly, mx_page_pin_t **pin)
{
  void *tmpkva;
  mx_instance_state_t *is = es->is;
  struct es_tx_k35snf *tx = &es->snf.es_tx.u.k35snf;
  int index;

  /* mapping A.1: SnifferTX registers */
  tmpkva = k35snf_tx_base(is);
  *off += MX_PAGE_SIZE;
  if (req < *off) {
    *kva = tmpkva;
    *mem_type = MX_MEM_SRAM;
    *readonly = 0;
    return 0;
  }

  /* mapping A.2: kernel per-endpoint flow page */
  *off += MX_PAGE_SIZE;
  if (req + MX_PAGE_SIZE == *off) {
    *kva = es->flow_window;
    *pin = &es->flow_window_pin;
    *mem_type = MX_MEM_HOSTMEM;
    *readonly = 0;
    return 0;
  }

  /* mapping A.3: completion page */
  *off += MX_PAGE_SIZE;
  if (req + MX_PAGE_SIZE == *off) {
    *kva = tx->completion.addr;
    *pin = &tx->completion.pin;
    *mem_type = MX_MEM_HOSTMEM;
    *readonly = 0;
    return 0;
  }

  /* mapping A.4: data ring */
  index = (req - *off) / MX_PAGE_SIZE;
  *off += (unsigned long)tx->data_ring.size;
  if (req < *off) {
    *mem_type = MX_MEM_HOSTMEM;
    *kva = (void *)(uintptr_t)tx->data_ring.gamma.pages[index].va;
    *pin = &tx->data_ring.gamma.pages[index];
    *readonly = 0;
    return 0;
  }

  /* mapping A.5: descriptor ring */
  index = (req - *off) / MX_PAGE_SIZE;
  *off += (unsigned long)tx->desc_ring.size;
  if (req < *off) {
    *mem_type = MX_MEM_HOSTMEM;
    *kva = (void *)(uintptr_t)tx->desc_ring.gamma.pages[index].va;
    *pin = &tx->desc_ring.gamma.pages[index];
    *readonly = 0;
    return 0;
  }

  return ENOENT;
}

static
int
snf_eop_mmap(mx_endpt_state_t *es, unsigned long req,
             unsigned long *off, void **kva,
	     int *mem_type, int *readonly, mx_page_pin_t **pin)
{
  void *tmpkva;
  mx_instance_state_t *is = es->is;
  int index;

  if (MX_IS_K35SNF(is) && es->es_type == MYRI_ES_SNF_TX)
    return k35snf_eop_mmap(es, req, off, kva, mem_type, readonly, pin);

  if (es->es_type == MYRI_ES_SNF_TX) {
    union snf_endpt_state *ess = &es->snf;

    /* mapping A.1: TX endpoint pages */
    if (es->is->board_span) {
      tmpkva = (void *) ((char *) phx_endpoint_nic_mem(es->is, es->phx_ep->tx_ep) + (req - *off));
      *off += es->is->phx.pfi->tx_config.dbltx_core_length;

      if (req < *off) {
        *kva = tmpkva;
        *mem_type = MX_MEM_SRAM;
        *readonly = 0;
        return 0;
      }

      /* mapping A.?: TX PIO endpoint pages */
      tmpkva = (void *) ((char *) phx_ep_pio_fifo(es->is, es->phx_ep->tx_ep) + (req - *off));
      *off += es->is->phx.pfi->tx_config.dbltx_pio_length;
      if (req < *off) {
        *kva = tmpkva;
        *mem_type = MX_MEM_SRAM;
        *readonly = 0;
        return 0;
      }
    }

    /* mapping A.2: kernel per-endpoint flow page */
    *off += es->flow_window ? MX_PAGE_SIZE : 0;
    if (req + MX_PAGE_SIZE == *off && es->flow_window) {
      *kva = es->flow_window;
      *pin = &es->flow_window_pin;
      *mem_type = MX_MEM_HOSTMEM;
      *readonly = 0;
      return 0;
    }

    /* mapping A.3: TX completion counters */
    if (is->tx_done.addr) {
      *off += is->tx_done.addr ? MX_PAGE_SIZE : 0;
      if (req + MX_PAGE_SIZE == *off && is->tx_done.addr) {
        *kva = is->tx_done.addr;
        *pin = &is->tx_done.pin;
        *mem_type = MX_MEM_HOSTMEM;
        *readonly = 1;
        return 0;
      }
    }

    /* mapping A.4: TX DMA buffer */
    if (ess->es_tx.u.phx.tx_dma_pages.n_pages) {
      index = (req - *off) / PAGE_SIZE;
      *off += (unsigned long)ess->es_tx.u.phx.tx_dma_pages.n_pages * PAGE_SIZE;
      if (req < *off) {
        *mem_type = MX_MEM_HOSTMEM;
        *kva = (void *)(uintptr_t)ess->es_tx.u.phx.tx_dma_pages.pages[index].va;
        *pin = &ess->es_tx.u.phx.tx_dma_pages.pages[index];
        *readonly = 0;
        return 0;
      }
    }

    return ENOENT;
  }

  if (es->es_type == MYRI_ES_SNF_RX_PQ) {
    /* See myri_snf_rx_pq_get_kva_maps().  S.1 PQ */
    if (is->snf.pq_s) {
      unsigned long rel_off = req - *off;
      tmpkva = (void *)((char *)is->snf.pq_s + rel_off);
      *off += is->snf.pq_s_length;
      if (req < *off) {
        *kva = tmpkva;
        *mem_type = MX_MEM_HOSTMEM;
        *readonly = 1;
        return 0;
      }
    }

    /* mapping S.2 PQ: snf prehash desc addrs */
    if (is->snf.prehash_desc_ring.size != 0) {
      index = (req - *off) / PAGE_SIZE;
      *off += (unsigned long)is->snf.prehash_desc_ring.size;
      if (req < *off) {
        *mem_type = MX_MEM_HOSTMEM;
        *kva = (void *)(uintptr_t)is->snf.prehash_desc_ring.gamma.pages[index].va;
        *pin = &is->snf.prehash_desc_ring.gamma.pages[index];
        *readonly = 0;
        return 0;
      }
    }
  } else {
    /* mapping A.2: the kernel window (jiffies, see mx_kernel_window_t) */
    *off += is->kernel_window ? MX_PAGE_SIZE : 0;
    if (req + MX_PAGE_SIZE == *off && is->kernel_window) {
      *kva = is->kernel_window;
      *mem_type = MX_MEM_HOSTMEM;
      *readonly = 1;
      return 0;
    }

    /* mapping S.1: snf rx state */
    if (is->snf.rx_s) {
      unsigned long rel_off = req - *off;
      int ring_req = rel_off / SNF__RX_STATE_SIZE;
      tmpkva = (void *)((char *)is->snf.rx_s + rel_off);
      *off += is->snf.rx_s_length;
      if (req < *off) {
        *kva = tmpkva;
        *mem_type = MX_MEM_HOSTMEM;
        /* Only the actual ring endpoint can map its state read+write */
        if (es->es_type == MYRI_ES_SNF_RX_RING && es->snf.es_rx.rx_ring_id == ring_req)
          *readonly = 0;
        else
          *readonly = 1;
        return 0;
      }
    }
  }

  /* mapping S.2 or S.3 PQ: snf desc addrs */
  if (is->snf.desc_ring.size != 0) {
    index = (req - *off) / PAGE_SIZE;
    *off += (unsigned long)is->snf.desc_ring.size;
    if (req < *off) {
      *mem_type = MX_MEM_HOSTMEM;
      *kva = (void *)(uintptr_t)is->snf.desc_ring.gamma.pages[index].va;
      *pin = &is->snf.desc_ring.gamma.pages[index];
      if (es->es_type == MYRI_ES_SNF_RX_PQ) {
        /* PQ software hashing modifies the ring.  But we no longer have meta data
         * space in the descriptor where we can write our own data.
         * Each virtual ring needs its own descriptor ring now.
         */
        *readonly = 0;
      } else
        *readonly = 1;
      return 0;
    }
  }

  /* mapping S.3 or S.4 PQ: snf data addrs
   * The data ring contains the preamble pages at the front.  Special case them.
   * Preamble + data pages are all in ring.gamma.
   */
  if (is->snf.data_ring.size != 0) {
    /* DBL RX pages */
    index = (req - *off) / PAGE_SIZE;
    *off += (unsigned long)is->snf.data_ring.size + is->snf.data_ring.preamble_size;
    if (req < *off) {
      *mem_type = MX_MEM_HOSTMEM;
      *kva = (void *)(uintptr_t)is->snf.data_ring.gamma.pages[index].va;
      *pin = &is->snf.data_ring.gamma.pages[index];
      *readonly = 0;
      return 0;
    }
  }

  return ENOENT;
}

/* This function runs with snf->sync locked (i.e. mx_mutex_enter) */
static int
myri_snf_start_stop(mx_endpt_state_t *es, int start)
{
  mx_instance_state_t *is = es->is;
  myri_snf_rx_state_t *snf = &is->snf;
  int endpt, promisc;

  if (MX_IS_K35SNF(is)) {
    if (start) {
      update_and_read_snf_stats(is, &snf->stats_at_start);
      if (PHX_HAS_FEATURE(is, PHX_FEATURE_FPGA_MERGE_NEXT_RING)) {
        /* the RSS registers are reset in the start_rx_endpt call below */
        snf->stats_at_start.nic_pkt_no_buffer = 0;
        snf->stats_at_start.nic_pkt_dropped = 0;
        snf->stats_at_start.nic_pkt_bad = 0;
      }
      phx_start_rx_endpoint(is, snf->phx_ep->rx_ep,
                            snf->data_ring.gamma.n_pages - snf->data_ring.preamble_n_pages,
                            snf->desc_ring.gamma.n_pages);
      phx_icu_enable_interrupt(is, snf->phx_ep->icu);
    } else {
      phx_icu_disable_interrupt(is, snf->phx_ep->icu);
      phx_stop_rx_endpoint(is, snf->phx_ep->rx_ep);
    }
    return 0;
  }

  mx_spin_lock(&snf->promisc_lock);
  if (start) {
    /* Switch the kernel endpoint to the sniffer endpoint so that it
     * gets all receive packets.
     */
    endpt = snf->phx_ep->rx_ep;
    promisc = 1;
    update_and_read_snf_stats(is, &snf->stats_at_start);
    snf->promisc_flags |= SNF_RX_PROMISC_FLAG_SNF_OWNED;
  } else {
    /* Restore the original kernel endpoint */
    endpt = is->board_ops.ether_endpoint(is)->rx_ep;
    snf->promisc_flags &= ~SNF_RX_PROMISC_FLAG_SNF_OWNED;
    /* Also restore the kernel's promiscuous mode. Do not clear the request
     * flag as it is sticky. Only eth (phx_snf_set_ether_promisc) can clear it.
     */
    promisc = ((snf->promisc_flags & SNF_RX_PROMISC_FLAG_REQUEST_ENABLE) != 0);
  }

  phx_dbl_filter_set_kernel_endpoint(is, endpt);

  /* Also enable promiscuous mode to disable MAC filtering */
  is->board_ops.set_promisc(is, promisc);
  mx_spin_unlock(&snf->promisc_lock);
  return 0;
}

static
int
myri_snf_rx_detach(mx_endpt_state_t *es, int from_teardown)
{
  myri_snf_rx_state_t *snf = &es->is->snf;
  union snf_endpt_state *ess = &es->snf;
  mx_endpt_state_t *es_pg;
  int ring_id = ess->es_rx.rx_ring_id;
  unsigned long flags;
  int wake_pending;
  uint32_t ring_id_mask = (1u << ring_id);
  int wakeup_waiters = 0;
  int update_tail_now = 0;

  flags = 0;  /* useless initialization to pacify -Wunused */

#if MAL_DEBUG
  MX_INFO(("snf_rx_detach. ring_id=%d from_teardown=%d es_type=%d\n", ring_id,
           from_teardown, es->es_type));
#endif

  if (es->es_type != MYRI_ES_SNF_RX_RING)
    return EINVAL;

  /* ring_id may be -1 (UNUSED).  For example, the app detaches the ring via
   * SNF_RX_DETACH and then closes the ioctl fd.  It is harmless as we check
   * if ring_id is valid below.
   */
  /* mal_assert(ring_id >= 0); */

  /* HACK.  Remove the vq waiter if pending.  See snf_vqwait_req.
   * The caller already locked sync.
   */
  if (ess->es_rx.vq_waitsync.intr_requested) {
    /* MX_INFO(("snf_rx_detach: remove vq_waitsync\n")); */
    ess->es_rx.vq_waitsync.intr_requested = 0;
    TAILQ_REMOVE(&snf->vq_waitq, &ess->es_rx.vq_waitsync, link);
    if (ess->es_rx.vq_waitsync.wait_sync)
      mx_wake(ess->es_rx.vq_waitsync.wait_sync);
  }

  /* Fault tolerant pq lock.  If this endpoint is the owner, then release it. */
  mx_spin_lock(&snf->pq_lock);
  if (snf->pq_locked_es == es) {
    snf->pq_locked_es = NULL;
    wakeup_waiters = 1;
  }
  mx_spin_unlock(&snf->pq_lock);

  /* Then wake up all waiters in case any one is sleeping forever. */
  if (wakeup_waiters) {
    /* MX_INFO(("Snf ring closed without unlocking pq. Release it and wake up waiters.\n")); */
    snf_vqwake_req(snf, 1);
  }

  mx_spin_lock(&snf->sync_s.rs_lock);
  if (snf->sync_s.attach_mask & ring_id_mask && 
      es->snf.es_rx.rx_ring_id != SNF_RING_UNUSED) {
    /* Ensure future updates don't consider this ring but only if the user
     * cleanly requested to be detached.  If the detach comes as part of the
     * normal close path, the ring may or may not have been detached */
    if (!from_teardown)
      snf->sync_s.detach_mask |= ring_id_mask;
    snf->sync_s.attach_mask &= ~ring_id_mask;
    es->is->kernel_window->attach_mask = snf->sync_s.attach_mask;
    snf->num_rings_attached--;
    /* See if this ring's tail is very close to the physical ring's tail.
     * If they are close, then this ring has been preventing the physical
     * ring from getting free space. So, should update the flow control
     * info now. Because the ring is now detached, we will ignore its tail
     * pointer (i.e. consumed the entire ring) and will likely return free
     * space back to the NIC. Otherwise, all the other rings may wait forever
     * for new packets, which never come because the desc/data rings are full.
     * The magic number 16 is from snf_rx_sync.
     */
    update_tail_now =
        (snf->sync_s.r_rs[es->snf.es_rx.rx_ring_id].rs_data_tail >=
	             snf->sync_s.rs.rs_data_tail) &&
        (snf->sync_s.r_rs[es->snf.es_rx.rx_ring_id].rs_data_tail -
	             snf->sync_s.rs.rs_data_tail <= (16ull << SNF_VPAGE_SHIFT));
    /* JFM */
    mx_spin_unlock(&snf->sync_s.rs_lock);

    /* DUP_MA state */
    if (es->snf.es_rx.app_ring_info) {
      struct snf_dup_ma_ring_state *rs = es->snf.es_rx.app_ring_info;

      snf->app_info[rs->app_info_idx].num_rings_attached--;
      snf->app_info[rs->app_info_idx].attach_mask &= ~(1 << rs->hash_ring_id);
      rs->app_info_idx = -1;
      rs->port_ring_id = 0;
      rs->hash_ring_id = 0;
      es->snf.es_rx.app_ring_info = NULL;
    }

    /* Prevent further enqueues */
    snf->refcount--;
    snf->ring_map_ep[ring_id] = NULL;

    mx_spin_lock_irqsave(&snf->rx_intr_lock, flags);
    if (ess->es_rx.rx_waitsync.intr_requested) {
      TAILQ_REMOVE(&snf->rx_intr_waitq, &ess->es_rx.rx_waitsync, link);
      ess->es_rx.rx_waitsync.intr_requested = 0;
      if (TAILQ_EMPTY(&snf->rx_intr_waitq))
        snf->rx_intr_requested = 0;
    }
    wake_pending = ess->es_rx.rx_waitsync.intr_wake_pending;
    mx_spin_unlock_irqrestore(&snf->rx_intr_lock, flags);

    /* If this thread was waiting on an interrupt along with others, it's
     * possible that it is in the process of being awakened.  If this is the
     * case, we give the kernel a chance to complete its wakeups */
    if (wake_pending) {
      wake_pending = 2;
      do {
        mx_sleep(&es->sync, MX_SMALL_WAIT, MX_SLEEP_NOINTR);
      } while (ess->es_rx.rx_waitsync.intr_wake_pending && --wake_pending >= 0);
      if (wake_pending == 0)
        MX_WARN(("%s: ring %d unable to clear wake_pending\n",
                  snf->is->is_name, ring_id));
    }
    es_pg = es->snf.es_rx.rx_pg;
    if (es_pg != NULL) {
      es->snf.es_rx.rx_pg = NULL;
      snf_rx_portgroup_del(es_pg, es);
    }
  }
  else
    mx_spin_unlock(&snf->sync_s.rs_lock);
  es->snf.es_rx.rx_ring_id = SNF_RING_UNUSED;

  /* Update flow control at the end, if necessary. */
  if (update_tail_now)
     snf_rx_sync(snf, 0, NULL);
  return 0;
}

static int
myri_snf_rx_get_kva_maps(const mx_endpt_state_t *es, myri_snf_rx_attach_t *x)
{
  uint32_t offc;
  uint64_t off;
  const myri_snf_rx_state_t *snf = &es->is->snf;

  /* See phx_dbl_ops.c:dbl_get_copyblocks() */
  offc = 0;

  /* A.2: kernel window */
  x->kernel_window_offset = offc;
  x->kernel_window_len = (es->is->kernel_window ? MX_PAGE_SIZE : 0);
  offc += x->kernel_window_len;

  /* Sniffer mappings */
  off = (uint64_t) offc;
  x->vstate_all_offset = off;           /* S.1 */
  x->vstate_all_len = snf->rx_s_length;
  off += snf->rx_s_length;

  x->rx_desc_ring_offset = off;         /* S.2 */
  x->rx_desc_ring_len = snf->desc_ring.size;
  off += x->rx_desc_ring_len;

  /* The data ring length includes the preamble.  The user mmaps the whole
   * ring [preamble pages, DBL data pages] in a contiguous region.
   */
  x->rx_data_ring_offset = off;         /* S.3 */
  x->rx_data_ring_len = snf->data_ring.size + snf->data_ring.preamble_size;
  off += x->rx_data_ring_len;

  return 0;
}

static int
myri_snf_rx_pq_get_kva_maps(const mx_endpt_state_t *es, myri_snf_pq_attach_t *x)
{
  uint64_t off;
  const myri_snf_rx_state_t *snf = &es->is->snf;

  /* Sniffer mappings */
  off = 0;                              /* S.1 PQ */
  x->pq_state_offset = off;
  x->pq_state_len = snf->pq_s_length;
  off += snf->pq_s_length;

  /* Prehash descriptor ring */
  x->prehash_desc_ring_offset = off;         /* S.2 PQ */
  x->prehash_desc_ring_len = snf->prehash_desc_ring.size;
  off += x->prehash_desc_ring_len;

  /* Descriptor and data rings, same as myri_snf_rx_get_kva_maps */
  x->rx_desc_ring_offset = off;         /* S.3 PQ */
  x->rx_desc_ring_len = snf->desc_ring.size;
  off += x->rx_desc_ring_len;

  x->rx_data_ring_offset = off;         /* S.4 PQ */
  x->rx_data_ring_len = snf->data_ring.size + snf->data_ring.preamble_size;
  off += x->rx_data_ring_len;

  return 0;
}

static
int
snf_rx_attach_endpt(myri_snf_rx_state_t *snf, mx_endpt_state_t *es, int ring_id, int app_id)
{
  union snf_endpt_state *ess = &es->snf;
  struct snf_sync_state *ss = &snf->sync_s;
  struct snf__rx_state *rx_s;
  uint32_t ring_id_mask;
  int rc = 0;
  int hash_ring_id;
  struct snf_dup_ma_ring_state *rs = NULL;

#if MAL_DEBUG
  MX_INFO(("snf_rx_attach_endpt. ring_id=%d rings_requested=%d\n", ring_id,
           snf->num_rings_requested));
#endif

  mal_assert(es->es_type == MYRI_ES_SNF_RX_RING);

  mx_spin_lock(&snf->sync_s.rs_lock);

  /* DUP_MA mode.  ring_id is the user request (i.e. per-application id).
   * We always allocate per-port id regardless of ring_id >= 0.  So save it
   * for later use.  Ugly...
   */
  hash_ring_id = ring_id;
  if ((snf->params.open_flags & MYRI_SNF_F_RX_DUP_MA) && ring_id >= 0)
    ring_id = -1; /* force to find a free slot */

  if (ring_id < 0) { /* Look for next free ring id */
    for (ring_id = 0; ring_id < snf->num_rings_requested; ring_id++) {
      ring_id_mask = (1u << ring_id);
      if ((ss->attach_mask & ring_id_mask) == 0)
	break;
    }
    if (ring_id == snf->num_rings_requested) {
      rc = EBUSY;
      goto bail;
    }
  }
  /* FIXME.  What if ring_id >= 0 and that id is already taken? */

  /* DUP_MA mode.  Now do per-application ring id allocation. */
  if ((snf->params.open_flags & MYRI_SNF_F_RX_DUP_MA)) {
    struct snf_dup_ma_app_state *as = NULL;
    int i, pid = (int)mx_kgettgpid(); /* tgid, not pid */

    /* Find the application */
    for (i = 0; i < MYRI_SNF_MAX_RINGS; i++) {
      if (snf->app_info[i].app_id == app_id) {
        as = &snf->app_info[i];
        break;
      }
    }
    /* This is a programming error on the user's part.  Should have opened
     * the port by now.
     */
    if (as == NULL) {
      MX_WARN(("SNF rx attach: cannot find the application/process."
        " app_id=%d pid=%d\n", app_id, pid));
      rc = EINVAL;
      goto bail;
    }
    /* Do we have free slots? */
    if (as->num_rings_attached >= as->num_rings_requested) {
      MX_WARN(("SNF rx attach: out of free rx rings. app_id=%d pid=%d rings_attached=%d"
        " rings_requested=%d\n",
        app_id, pid, as->num_rings_attached, as->num_rings_requested));
      rc = EBUSY;
      goto bail;
    }
    /* If requested a specific id, see if it is already taken. */
    if (hash_ring_id >= 0) {
      if (hash_ring_id >= as->num_rings_requested) {
        MX_WARN(("SNF rx attach: invalid ring id. app_id=%d pid=%d ring_id=%d"
          " rings_requested=%d\n",
          app_id, pid, hash_ring_id, as->num_rings_requested));
        rc = EINVAL;
        goto bail;
      }
      if ((as->attach_mask & (1 << hash_ring_id))) {
        MX_WARN(("SNF rx attach: the requested ring id is already taken."
          " app_id=%d pid=%d ring_id=%d rings_attached=%d rings_requested=%d\n",
          app_id, pid, hash_ring_id, as->num_rings_attached, as->num_rings_requested));
        rc = EBUSY;
        goto bail;
      }
    }
    else {
      for (hash_ring_id = 0; hash_ring_id < as->num_rings_requested;
           hash_ring_id++) {
        if ((as->attach_mask & (1 << hash_ring_id)) == 0)
          break;
      }
      if (hash_ring_id == as->num_rings_requested) {
        MX_WARN(("SNF rx attach: out of free rx rings. app_id=%d pid=%d rings_attached=%d"
          " rings_requested=%d\n",
          app_id, pid, as->num_rings_attached, as->num_rings_requested));
        rc = EBUSY;
        goto bail;
      }
    }

    for (i = 0; i < MYRI_SNF_MAX_RINGS; i++) {
      if (snf->app_ring_info[i].app_info_idx == -1) {
        rs = &snf->app_ring_info[i];
        rs->app_info_idx = as - snf->app_info;
        rs->port_ring_id = ring_id;
        rs->hash_ring_id = hash_ring_id;
        as->num_rings_attached++;
        as->attach_mask |= (1 << hash_ring_id);
        break;
      }
    }
    if (rs == NULL) {
      /* This should not happen */
      MX_WARN(("SNF rx attach: out of free application ring slots."
        " app_id=%d pid=%d rings_attached=%d rings_requested=%d\n",
        app_id, pid, as->num_rings_attached, as->num_rings_requested));
      rc = EBUSY;
      goto bail;
    }
  }

  ring_id_mask = (1u << ring_id);
  rx_s = SNF__RX_STATE(snf->rx_s, ring_id);
  ss->attach_mask |= ring_id_mask;
  /* Provide a copy of the mask for userspace apps, for convenience */
  es->is->kernel_window->attach_mask = ss->attach_mask;
  snf->num_rings_attached++;
  snf->refcount++;
  snf->ring_map_ep[ring_id] = es;

  ess->es_rx.app_ring_info = rs;
  ess->es_rx.rx_ring_id = ring_id;
  ess->es_rx.rx_s = rx_s;
  ess->es_rx.rx_waitsync.wait_sync = &es->wait_sync;
  ess->es_rx.rx_waitsync.intr_requested = 0;
  ess->es_rx.rx_waitsync.snf = snf;

  ess->es_rx.vq_waitsync.wait_sync = &es->vq_sync;
  ess->es_rx.vq_waitsync.snf = snf;
  ess->es_rx.vq_waitsync.intr_requested = 0;

  sprintf(es->opener.user_info, "rx ring %d", ring_id);

  bzero(&rx_s->recvq.stats, sizeof(struct snf__recvq_stats));

  /* If the id_mask is non-zero, we're reattching to a ring that had valid live
   * state. The ring is no longer orphaned and we want to prevent the sync code
   * from reclaiming (and clearing) the state of a ring that is no longer
   * detached with valid live state. */
  if ((snf->sync_s.detach_mask & ring_id_mask) != 0) {
    snf->sync_s.detach_mask &= ~ring_id_mask;
    rx_s->recvq.rq_is_reattach = 1;
  }
  else {
    uint32_t descq_elems;
    struct snf__rx_sync *rs = &ss->rs;

    bzero(rx_s, sizeof(struct snf__rx_state));

    ss->r_rs[ring_id] = ss->rs; /* deep copy */

    descq_elems = snf->rx_desc_ring_count;

    rx_s->evq.evq_seqnum  = (rs->rs_desc_cnt % PHX_RX_DESC_SEQNUM_MAX) + 1;
    rx_s->evq.evq_next_idx = rs->rs_desc_cnt % descq_elems;
    rx_s->evq.evq_last_idx = descq_elems - 1;
    rx_s->evq.evq_evinfo_idx = -1;
    rx_s->evq.evq_desc_cnt = rs->rs_desc_cnt;
    rx_s->evq.evq_data_head = rs->rs_data_head;

    rx_s->recvq.rq_data_off = rs->rs_data_head % snf->params.rx_data_ring_size;
    rx_s->recvq.rq_data_head = rs->rs_data_head;
    rx_s->recvq.rq_data_tail = rs->rs_data_head;
    rx_s->recvq.rq_data_tail_last = rs->rs_data_head;
    rx_s->recvq.rq_data_size = snf->params.rx_data_ring_size;
    rx_s->recvq.rq_ev_ts_last = 1;
  }
  memset(&rx_s->recvq.stats, 0, sizeof(struct snf__recvq_stats));

bail:
  mx_spin_unlock(&snf->sync_s.rs_lock);
  return rc;
}

static
int
snf_pq_attach_endpt(myri_snf_rx_state_t *snf, mx_endpt_state_t *es)
{
  union snf_endpt_state *ess = &es->snf;
  struct snf_sync_state *ss = &snf->sync_s;
  struct snf__pq_state *pq_s;
  int rc = 0;

  mal_assert(es->es_type == MYRI_ES_SNF_RX_PQ);

  ess->es_rx.pq_s = snf->pq_s;
  ess->es_rx.rx_s = snf->rx_s;

  ess->es_rx.rx_waitsync.snf = snf;
  ess->es_rx.rx_waitsync.wait_sync = &es->wait_sync;
  ess->es_rx.rx_waitsync.intr_requested = 0;

  sprintf(es->opener.user_info, "pq processor");

  if (snf->pq_refcount > 1) {
    /* MX_INFO(("pq_attach_endpt. re-attach.\n")); */
    return 0;
  }
  
  mx_spin_lock(&snf->sync_s.rs_lock);
  
  pq_s = snf->pq_s;

  {
    uint32_t descq_elems;
    struct snf__rx_sync *rs = &ss->rs;

    bzero(pq_s, sizeof(struct snf__pq_state));

    descq_elems = snf->rx_desc_ring_count;

    pq_s->rw.evq_prehash_seqnum  = (rs->rs_desc_cnt % PHX_RX_DESC_SEQNUM_MAX) + 1;
    pq_s->rw.rq_prehash_data_off = rs->rs_data_head % snf->params.rx_data_ring_size;
    pq_s->rw.evq_next_prehash_idx = rs->rs_desc_cnt % descq_elems;
    pq_s->rw.evq_desc_cnt = rs->rs_desc_cnt;
    pq_s->rw.evq_ts_last = 1; /* FIXME.  This might be wrong? */
    pq_s->evq_last_prehash_idx = descq_elems - 1;

    /* FIXME.  We cannot get these from snf__rx_sync.
     * We should not re-initialize every time like this. */
  }

  mx_spin_unlock(&snf->sync_s.rs_lock);

  /* MX_INFO(("pq_attach_endpt\n")); */
  return rc;
}

static int
snf_rx_attach(mx_endpt_state_t *es, myri_snf_rx_attach_t *x)
{
  myri_snf_rx_state_t *snf = &es->is->snf;
  int ring_id = x->ring_id;
  int rc = 0;
  struct snf_dup_ma_app_state *as = NULL;

  if (es->es_type != MYRI_ES_SNF_RX_RING)
    return EINVAL;

  mx_mutex_enter(&snf->sync);

#if MAL_DEBUG
  MX_INFO(("%s: es=%d ring_id=%d rings_requested=%d rings_attached=%d\n",
           __FUNCTION__, es->endpt, ring_id, snf->num_rings_requested,
           snf->num_rings_attached));
#endif

  /* Get the easy ones out of the way first */
  if (ring_id >= (int) snf->num_rings_requested) {
    rc = EINVAL;
    goto bail;
  }
  if (snf->num_rings_attached == snf->num_rings_requested) {
    rc = EBUSY;
    goto bail;
  }

  rc = myri_snf_rx_get_kva_maps(es, x);
  if (rc) {
    MX_WARN(("Failed to obtain kva mappings: %d\n", rc));
    goto bail;
  }

  rc = snf_rx_attach_endpt(snf, es, ring_id, x->params.app_id);
  if (rc) {
    MX_WARN(("Failed to attach to ring %d with err=%d\n", ring_id, rc));
    goto bail;
  }

  x->params = snf->params; /* Copy back out to attaching user */
  x->ring_id = es->snf.es_rx.rx_ring_id;
  x->port_ring_id = x->ring_id; /* By default, hash ring id == port ring id */
  myri_clksync_copy_nticks_params(&es->is->kernel_window->clksync, &x->nticks);

  /* DUP_MA mode.  Override parameters with per-application ones */
  if ((snf->params.open_flags & MYRI_SNF_F_RX_DUP_MA)) {
    struct snf_dup_ma_ring_state *rs = es->snf.es_rx.app_ring_info;

    as = &snf->app_info[rs->app_info_idx];
    x->ring_id = rs->hash_ring_id; /* The application sees per-app ring id */
    x->params.num_rings = as->num_rings_requested;
    x->params.rss_flags = as->rss_flags;
    if (MAL_DEBUG) {
      MX_INFO(("rx_attach: copy per-app parameters. app_info_idx=%d app_id=%d num_rings=%d"
        " rss_flags=%d ring_id=%d port_ring_id=%d\n",
        rs->app_info_idx, as->app_id, x->params.num_rings, x->params.rss_flags,
        x->ring_id, x->port_ring_id));
    }
  }

  /* Tell the app that it can use HW hash */
  if ((as && as->using_hw_hash) || (!as && snf->using_hw_hash)) {
    x->params.rss_flags |= SNF_RSS_FLAGS_HW;
  }

bail:
  mx_mutex_exit(&snf->sync);
  return rc;
}

static int
snf_pq_attach(mx_endpt_state_t *es, myri_snf_pq_attach_t *x)
{
  myri_snf_rx_state_t *snf = &es->is->snf;
  int rc = 0;
  if (es->es_type != MYRI_ES_SNF_RX_PQ)
    return EINVAL;

  mx_mutex_enter(&snf->sync);

  rc = myri_snf_rx_pq_get_kva_maps(es, x);
  if (rc) {
    MX_WARN(("Failed to obtain kva mappings: %d\n", rc));
    goto bail;
  }

  rc = snf_pq_attach_endpt(snf, es);
  if (rc) {
    MX_WARN(("Failed to attach PQ, err=%d\n", rc));
    goto bail;
  }

  x->params = snf->params; /* Copy back out to attaching user */
  snf->pq_refcount++;
  snf->refcount++;

bail:
  mx_mutex_exit(&snf->sync);
  return rc;
}

/* HACK.  Clean up interrupt request.  Much of this code is from myri_snf_rx_detach.
 * This function stays until we remove the concept of PQ endpoints.
 */
static
void
snf_pq_detach(mx_endpt_state_t *es)
{
  myri_snf_rx_state_t *snf = &es->is->snf;
  union snf_endpt_state *ess = &es->snf;
  unsigned long flags;
  int wake_pending;

  flags = 0;  /* useless initialization to pacify -Wunused */

  /* MX_INFO(("snf_pq_detach\n")); */
  if (es->es_type != MYRI_ES_SNF_RX_PQ) {
    MX_WARN(("PQ detach called with non RX_PQ endpoint\n"));
    return;
  }
  /* MX_INFO(("intr_requested=%d ring_id=%d\n", ess->es_rx.rx_waitsync.intr_requested, ess->es_rx.rx_ring_id)); */

  mx_spin_lock_irqsave(&snf->rx_intr_lock, flags);
  if (ess->es_rx.rx_waitsync.intr_requested) {
    TAILQ_REMOVE(&snf->rx_intr_waitq, &ess->es_rx.rx_waitsync, link);
    ess->es_rx.rx_waitsync.intr_requested = 0;
    if (TAILQ_EMPTY(&snf->rx_intr_waitq))
      snf->rx_intr_requested = 0;
  }
  wake_pending = ess->es_rx.rx_waitsync.intr_wake_pending;
  mx_spin_unlock_irqrestore(&snf->rx_intr_lock, flags);

  /* If this thread was waiting on an interrupt along with others, it's
   * possible that it is in the process of being awakened.  If this is the
   * case, we give the kernel a chance to complete its wakeups */
  if (wake_pending) {
    wake_pending = 2;
    do {
      mx_sleep(&es->sync, MX_SMALL_WAIT, MX_SLEEP_NOINTR);
    } while (ess->es_rx.rx_waitsync.intr_wake_pending && --wake_pending >= 0);
    if (wake_pending == 0)
      MX_WARN(("%s: pq unable to clear wake_pending\n",
                snf->is->is_name));
  }

  snf->pq_refcount--;
  snf->refcount--;
}

static
void
snf_rx_teardown(mx_endpt_state_t *es)
{
  myri_snf_rx_state_t *snf = &es->is->snf;
  mx_instance_state_t *is = es->is;

  mx_mutex_enter(&snf->sync);
  if (es->es_type == MYRI_ES_SNF_RX_RING)
    myri_snf_rx_detach(es, 1);
  if (es->es_type == MYRI_ES_SNF_RX_PQ) {
    mal_always_assert(snf->pq_refcount > 0);
    snf_pq_detach(es);
    if (MAL_DEBUG) {
      MX_INFO(("rx_teardown. pq. seq=%d ev_idx=%d\n",
             snf->pq_s->rw.evq_prehash_seqnum, snf->pq_s->rw.evq_next_prehash_idx));
    }
  } else if (es->es_type == MYRI_ES_SNF_RX && es->snf.es_rx_h.rx_is_attached) {
    struct snf_dup_ma_app_state *app;
    int release_hw_hash = 0;

    es->snf.es_rx_h.rx_is_attached = 0;
    snf->refcount--;
    /* DUP_MA mode.  If all handles to this app are closed, clear the app
     * information too.
     */
    if ((app = es->snf.es_rx_h.app_info)) {
      release_hw_hash = app->using_hw_hash;
      es->snf.es_rx_h.app_info = NULL;
      app->refcount--;
      if (app->refcount == 0) {
        app->app_id = -1;
        app->using_hw_hash = 0;
      }
    } else {
      release_hw_hash = snf->using_hw_hash;
    }

    /* Release HW hash module */
    if (release_hw_hash) {
      int i, port;
      struct phx_fpga_instance *fpga;

      mx_mutex_enter(&mx_fpga_mutex);

      fpga = is->phx.pfi;
      SNF_FFS_FOREACH(snf->params.aggr_portmask, i) {
        int j;
        for (j = 0; j < fpga->attach_count; j++) {
          if (fpga->is[j]->id == i) {
            port = MX_IS_PORT_ID(fpga->is[j]);
            mal_always_assert(fpga->hw_hash.in_use[port] > 0);
            if (--fpga->hw_hash.in_use[port] == 0)
              fpga->hw_hash.enabled_flags[port] = 0;
#if MAL_DEBUG
            MX_INFO(("release HW hash: port=%d in_use=%d\n",
                     port, fpga->hw_hash.in_use[port]));
#endif
          }
        }
      }

      mx_mutex_exit(&mx_fpga_mutex);
    }
  }

  if (snf->refcount == 0) {
    int i;

    snf->using_hw_hash = 0;
    snf->num_rings_attached = 0;
    snf->num_rings_requested = 0;
    snf->sync_s.detach_mask = 0;
    snf->sync_s.attach_mask = 0;
    snf->owner_pid = -1;
    if (snf->state == SNF_RX_S_STARTED)
      myri_snf_start_stop(es, 0);
    if (snf->state != SNF_RX_S_FREE) {
      /* Stop the interrupt first and let the ongoing one drain */
      phx_icu_disable_interrupt(is, snf->phx_ep->icu);
      mx_spin(1000); /* 1msec */

      mx_spin_lock_destroy(&snf->sync_s.rs_lock);
      for (i = 0; i < MYRI_SNF_MAX_RINGS; i++)
        mx_spin_lock_destroy(&snf->sync_s.r_rs_lock[i]);
      snf->state = SNF_RX_S_FREE;

      /* Stop the receive ring */
      phx_stop_rx_endpoint(is, snf->phx_ep->rx_ep);
      /* FIXME: 10msec (magic number) so the endpoint really becomes inactive. */
      mx_spin(10000);

      /* The sniffer endpoint is closed so accumulate counters.
       * Have to do this before we reset the endpoint, which clears
       * counters.
       */
      myri_counters64_ep_close(is, snf->phx_ep->icu - 1 /* n'th within the function */);

      phx_reset_rx_endpoint(is, snf->phx_ep->rx_ep);

      phx_apu_channel_reset(is, snf->apu_data_ch);
      phx_apu_channel_reset(is, snf->apu_desc_ch);
      /* Reset ICU too */
      phx_icu_reset_endpt(is, snf->phx_ep);

      phx_release_fpga_endpoint(is, snf->phx_ep->id);

      /* Give the port 0 endpoint back to the ethernet driver */
      if (MX_IS_K35SNF(is) && is->port_id == 0)
        k35snf_ether_return_endpoint(is);
    }
    if (snf->rx_s)
      mx_kfree(snf->rx_s);
    if (snf->pq_s)
      mx_kfree(snf->pq_s);
    snf->pq_s = NULL;
    snf->rx_s = NULL;
    snf->rx_s_length = 0;
    phx_free_rx_ring(is, &snf->data_ring);
    phx_free_rx_ring(is, &snf->desc_ring);
    phx_free_rx_ring(is, &snf->prehash_desc_ring);

    /* Reset DUP_MA state as well */
    memset(snf->app_info, 0, sizeof(snf->app_info));
    memset(snf->app_ring_info, 0, sizeof(snf->app_ring_info));
    for (i = 0; i < MYRI_SNF_MAX_RINGS; i++) {
      snf->app_info[i].app_id = -1;
      snf->app_info[i].refcount = 0;
      snf->app_info[i].using_hw_hash = 0;
      snf->app_ring_info[i].app_info_idx = -1;
    }
  }
  mx_mutex_exit(&snf->sync);
}

static int
myri_snf_rx_open_get_params(mx_endpt_state_t *es, myri_snf_rx_params_t *x)
{
  mx_instance_state_t *is = es->is;
  myri_snf_rx_state_t *snf = &is->snf;
  unsigned long meta_size;

  if (snf->state != SNF_RX_S_FREE) {
    if (!es->is_kernel && (uint32_t)mx_kgetpid() != snf->owner_pid && 
        !(snf->params.open_flags & MYRI_SNF_F_PSHARED)) {
      MX_WARN(("SNF can't share rings between processes unless SNF_F_PSHARED "
               "mode is used (owner is PID %d, request from PID %d)\n", 
               snf->owner_pid, mx_kgetpid()));
      return EACCES;
    }
    /* Deny mixed aggregation and non-aggregation modes.  Also check
     * the portmask so we the port does not belong to two different
     * aggregation/port groups.
     */
    if (((snf->params.open_flags ^ x->open_flags) &
         MYRI_SNF_F_AGGREGATE_PORTMASK)) {
      MX_WARN(("SNF cannot open the port in both aggregation and"
          " non-aggregation mode at the same time.\n"));
      return EINVAL;
    }
    if ((snf->params.open_flags & MYRI_SNF_F_AGGREGATE_PORTMASK) &&
      snf->params.aggr_portmask != x->aggr_portmask) {
      MX_WARN(("SNF port cannot belong to different port aggregation"
          " groups.\n"));
      return EINVAL;
    }
    *x = snf->params;
    return 0;
  }

  if (myri_snf_rings != 0)
    x->num_rings = myri_snf_rings;

  if (myri_snf_flags != -1)
    x->open_flags = myri_snf_flags;

  /* If there's only one ring, ignore all dup/copy requests */
  if (x->num_rings == 1)
    x->open_flags &= ~(MYRI_SNF_F_RX_COPY|MYRI_SNF_F_RX_DUP);

  x->rx_map_flags = 0;

  if (x->num_rings > 1)
    x->rx_map_flags |= MYRI_SNF_RXMAP_SHARED;

  /* DUP_MA mode.  Implicitly carries SHARED mode so multiple processes
   * can share the port.
   */
  if ((x->open_flags & MYRI_SNF_F_RX_DUP_MA)) {
    x->open_flags |= MYRI_SNF_F_PSHARED;
    x->rx_map_flags |= MYRI_SNF_RXMAP_SHARED;
  }

  /* Validate that sizes are a multiple of 2 MB */
  meta_size = (unsigned long)PHX_APU_RING_SIZE_MULTIPLE;

  if ((unsigned long)x->rx_data_ring_size & (meta_size - 1) ||
      (unsigned long)x->rx_desc_ring_size & (meta_size - 1)) {
    MX_WARN(("Desc or Data ring size not a multiple of 2MB\n"));
    return EINVAL;
  }
  if (x->num_rings > MYRI_SNF_MAX_RINGS) {
    MX_WARN(("Unsupported number of rings: %d\n", x->num_rings));
    return EINVAL;
  }

  x->rx_data_preamble_size = SNF_ALIGNUP(SNF_MTUWRAP_SIZE, PAGE_SIZE);
  /* Save the board type in params for the user lib */
  if (is->board_type == MAL_BOARD_TYPE_PHX_K35SNF)
    x->board_type = SNF_BOARD_TYPE_K35SNF;
  else
    x->board_type = SNF_BOARD_TYPE_DBL4P;

  return 0;
}

/* DUP_MA mode.  When the application/process first opens the port handle, we
 * call this function to save its app_id, number of rings and such.
 */
static int
myri_snf_app_join(mx_endpt_state_t *es, myri_snf_rx_params_t *x,
  myri_snf_rx_params_t *x_orig, struct snf_dup_ma_app_state **app)
{
  mx_instance_state_t *is = es->is;
  myri_snf_rx_state_t *snf = &is->snf;
  int status = 0;

  *app = NULL;
  if (MAL_DEBUG) {
    MX_INFO(("SNF DUP_MA mode application/process joins\n"));
  }

  /* All processes must also be using DUP_MA mode. */
  if (!(x->open_flags & MYRI_SNF_F_RX_DUP_MA)) {
    MX_WARN(("SNF port is in DUP_MA mode. The joining process must also use"
        " DUP_MA mode.\n"));
    status = EINVAL;
  }
  else {
    /* Remember this process */
    struct snf_dup_ma_app_state *s = NULL;
    int app_id, i;

    app_id = x_orig->app_id;
    if (MAL_DEBUG) {
      int pid = (int)mx_kgettgpid();
      MX_INFO(("DUP_MA process joining the port. app_id=%d pid=%d\n", app_id, pid));
    }
    for (i = 0; i < MYRI_SNF_MAX_RINGS; i++) {
      if (snf->app_info[i].app_id == app_id) {
        /* The process has opened the port before.  It should not open
         * again, but is harmless.
         */
        s = &snf->app_info[i];
        *app = s;

        /* Pass the application's original parameters */
        x->num_rings = s->num_rings_requested;
        x->rss_flags = s->rss_flags;
        goto exit;
      }
    }

    /* A new application/process.  Save its params. */
    for (i = 0; i < MYRI_SNF_MAX_RINGS; i++) {
      if (snf->app_info[i].app_id == -1) {
        /* A free slot. Only need to save num_rings and rss_flags as these may
         * be different from the port's.  Each application may have its own
         * num_rings and rss_flags too.  All the other parameters are inherited
         * from the port's.
         */
        s = &snf->app_info[i];
        s->app_id = app_id;
        s->num_rings_requested = x_orig->num_rings; /* Not the port's param */
        s->rss_flags = x_orig->rss_flags; /* Not the port's param */
        s->num_rings_attached = 0;
        s->attach_mask = 0;
        *app = s;
        break;
      }
    }
    if (s == NULL) {
      int pid = (int)mx_kgettgpid();
      status = EBUSY;
      MX_WARN(("SNF app join: out of free application slots. app_id=%d pid=%d\n", app_id, pid));
    }
  }

exit:
  return status;
}

static int
myri_snf_rx_join(mx_endpt_state_t *es)
{
  int status = 0;
  /* Nothing special to do yet */
  return status;
}

static int
phx_snf_rx_enable(mx_endpt_state_t *es, const myri_snf_rx_params_t *x)
{
  mx_instance_state_t *is = es->is;
  myri_snf_rx_state_t *snf = &is->snf;
  unsigned long flags;
  int status = 0;
  int rc, reserved;
  struct phx_nic_mem_global *nic;
  struct mx_instance_phx_state *phx;
  struct phx_physical_ep *phx_ep;
  mx_page_pin_t *page;

  flags = 0;  /* useless initialization to pacify -Wunused */

  if (x->num_rings > MYRI_SNF_MAX_RINGS) {
    MX_WARN(("Exceeded maximum number of %d rings\n", MYRI_SNF_MAX_RINGS));
    return ENXIO;
  }
  phx_ep = is->board_ops.sniffer_endpoint(is);
  reserved = !(MX_IS_K35SNF(is));
  if ((rc = phx_claim_fpga_endpoint(is, phx_ep->id, reserved))) {
    /* K35SNF port 0 is apparently taken by the ethernet driver */
    if (MX_IS_K35SNF(is) && is->port_id == 0) {
      if (!k35snf_ether_remove_endpoints(is)) {
        MX_INFO(("Removed the endpoints from the ethernet interface.\n"));
        rc = phx_claim_fpga_endpoint(is, phx_ep->id, reserved);
      }
    }
    if (rc) {
      MX_WARN(("SNF port is already in use.\n"));
      return rc;
    }
  }

  nic = is->phx.nic_mem;

  /* We are committed to this sniffer ring.  Assign the phoenix endpoint to it. */
  phx = &is->phx;
  snf->phx_ep = phx_ep;

  /* Per-endpoint flow control registers, now that we know the endpoint */
  snf->desc_ring_consume = phx_rx_desc_consr_advance(is, snf->phx_ep->rx_ep);
  snf->data_ring_consume = phx_rx_data_consr_advance(is, snf->phx_ep->rx_ep);

  TAILQ_INIT(&snf->rx_intr_waitq);
  TAILQ_INIT(&snf->vq_waitq);
  snf->rx_intr_requested = 0;

  /* Allocate the RX data ring */
  status = phx_alloc_rx_ring(is, &snf->data_ring,
                             x->rx_data_ring_size, x->rx_data_preamble_size);
  if (status) {
    MX_WARN(("Failed to allocate snf data copyblock\n"));
    rc = ENOMEM;
    goto exit;
  }

  /* Allocate the RX descriptor */
  status = phx_alloc_rx_ring(is, &snf->desc_ring, x->rx_desc_ring_size, 0);
  if (status) {
    MX_WARN(("Failed to allocate snf desc copyblock\n"));
    rc = ENOMEM;
    goto exit;
  }

  /* Allocate the prehash RX descriptor ring.  This is not a DBL ring and
   * is not pushed to the NIC.  So, alpha and beta pages are wasted.
   * Something to fix later on when we have a lot of free cycles.
   */
  status = phx_alloc_rx_ring(is, &snf->prehash_desc_ring,
                             x->rx_desc_ring_size, 0);
  if (status) {
    MX_WARN(("Failed to allocate snf prehash desc copyblock\n"));
    rc = ENOMEM;
    goto exit;
  }

  /* Allocate meta data buffers */
  snf->rx_s_length =
      SNF_ALIGNUP(MYRI_SNF_MAX_RINGS * SNF__RX_STATE_SIZE, PAGE_SIZE);
  snf->rx_s = mx_kmalloc(snf->rx_s_length, MX_WAITOK|MX_MZERO);
  if (snf->rx_s == NULL) {
    MX_WARN(("Failed to alloc rx state page\n"));
    rc = ENOMEM;
    goto exit;
  }

  snf->pq_s_length = SNF_ALIGNUP(sizeof(struct snf__pq_state), PAGE_SIZE*2);
  snf->pq_s = mx_kmalloc(snf->pq_s_length, MX_WAITOK|MX_MZERO);
  if (snf->rx_s == NULL) {
    MX_WARN(("Failed to alloc pq state page\n"));
    rc = ENOMEM;
    goto exit;
  }

  /* Now push the rings to the NIC (i.e. set up APUs) */
  snf->apu_data_ch = phx_apu_channel_id(is, 1, snf->phx_ep->rx_ep, 0);
  snf->apu_desc_ch = phx_apu_channel_id(is, 0, snf->phx_ep->rx_ep, 0);
#if MAL_DEBUG
  MX_INFO(("Sniffer APU channels: data=%d desc=%d\n",
           snf->apu_data_ch, snf->apu_desc_ch));
#endif
  phx_apu_channel_reset(is, snf->apu_data_ch);
  phx_apu_channel_reset(is, snf->apu_desc_ch);
  /* Pointer to the first page */
  page = snf->data_ring.alpha.pages;
  phx_apu_channel_start(is, snf->apu_data_ch,
                        page->dma.low, page->dma.high);
  page = snf->desc_ring.alpha.pages;
  phx_apu_channel_start(is, snf->apu_desc_ch,
                        page->dma.low, page->dma.high);

#if MAL_DEBUG
  MX_INFO(("Starting DBL RX endpoint. phx_rx_endpt=%d\n", snf->phx_ep->rx_ep));
#endif

  /* K35SNF start the endpoint when the app does SNF_START (snf_start_stop) */
  if (!MX_IS_K35SNF(is)) {
    phx_start_rx_endpoint(is, snf->phx_ep->rx_ep,
                          snf->data_ring.gamma.n_pages - snf->data_ring.preamble_n_pages,
                          snf->desc_ring.gamma.n_pages);
    /* Now, the NIC is fetching ring pages and ready to receive. */

    /* Enable the interrupt too */
    phx_icu_enable_interrupt(is, snf->phx_ep->icu);
  }

  /* Used to implement the fault tolerant pq lock */
  mx_spin_lock_init(&snf->pq_lock, snf->is, 0, "snf pq lock");
  snf->pq_locked_es = NULL;

  snf->promisc_flags = 0;
  mx_spin_lock_init(&snf->promisc_lock, snf->is, 0, "snf promisc lock");

  memset(&snf->sync_s, 0, sizeof(struct snf_sync_state));
  mx_spin_lock_init(&snf->sync_s.rs_lock, snf->is, 0, "snf flow lock");
  {
    int i;
    for (i = 0; i < MYRI_SNF_MAX_RINGS; i++)
      mx_spin_lock_init(&snf->sync_s.r_rs_lock[i], snf->is, 0,
                        "snf per-ring flow lock");
  }

  /* The DBL RX endpoint has full flow control credit (all pages) upon reset.
   * Do not write to flow control registers.
   */

  snf->sync_s.detach_mask = 0;
  snf->owner_pid = (int)mx_kgettgpid();
  snf->params = *x; /* Save params, these are the default now */
  snf->num_rings_requested = snf->params.num_rings;
  snf->rx_desc_ring_count = (uint32_t)(snf->params.rx_desc_ring_size / phx_rx_desc_size(is));

  snf->num_rings_attached = 0;
  snf->state = SNF_RX_S_INIT;

  /* Initialize DUP_MA state */
  if ((x->open_flags & MYRI_SNF_F_RX_DUP_MA)) {
    int i;

    if (MAL_DEBUG) {
      MX_INFO(("Enabling snf port using MYRI_SNF_F_RX_DUP_MA. pid=%d app_id=%d rings=%d\n",
        snf->owner_pid, x->app_id, x->num_rings));
    }

    /* DUP_MA mode.  The very first open of the port sets the flag.
     * We do not know how many applications/processes are going to attach
     * to the port.  So, override and set the number of rings to max.
     */
    snf->num_rings_requested = MYRI_SNF_MAX_RINGS;

    memset(snf->app_info, 0, sizeof(snf->app_info));
    memset(snf->app_ring_info, 0, sizeof(snf->app_ring_info));
    for (i = 0; i < MYRI_SNF_MAX_RINGS; i++) {
      snf->app_info[i].app_id = -1;
      snf->app_ring_info[i].app_info_idx = -1;
    }
  }
  return 0;

exit:
  phx_release_fpga_endpoint(is, snf->phx_ep->id);
  return rc;
}

static int
snf_rx_sync(myri_snf_rx_state_t *snf, int ring_id, struct snf__rx_sync *rs_update)
{
  struct snf_sync_state *ss = &snf->sync_s;
  int idx_n = -1, i;
  int status = 0;
  uint64_t da = 0;
  const uint64_t da_min_update = (16ull << SNF_VPAGE_SHIFT);
  const uint64_t da_max = snf->params.rx_data_ring_size;
  const uint64_t de_max = snf->rx_desc_ring_count;
  uint64_t data_tail, min_data_tail = 0;
  int desc_shift;

#if !MAL_CPU_x86_64
  /* On 64-bit, we read rs.rs_data_tail and rs.rs_desc_cnt without locking */
  mx_spin_lock(&ss->rs_lock);
#endif
  /*
   * The function can be used to force a synchronization (rs_update == NULL)
   * but if not, make sure the ring is valid before we go on using its index.
   *
   * As means of precaution, we also validate the input data from rs_update.
   *
   * We set 'min_update' as the minimum amount of bytes we want to increment
   * before looking through tail pointers and eventually updating the true
   * desc/data heads.
   */
  if (rs_update) {
    uint64_t da = rs_update->rs_data_tail - ss->rs.rs_data_tail;
    if (ring_id < 0 || ring_id >= MYRI_SNF_MAX_RINGS || 
	((ss->attach_mask & (1u<<ring_id)) == 0)) {
      status = EINVAL;
      goto done;
    }
    if (da > da_max || (rs_update->rs_desc_cnt - ss->rs.rs_desc_cnt > de_max)) {
      status = EINVAL;
      goto done;
    }
#if MAL_CPU_x86_64
    /* rs_data_head and rs_desc_cnt need to be in sync and get updated
     * atomically. Use fine grained lock to avoid contension when the caller
     * just wants to store the latest pointers.
     */
    mx_spin_lock(&ss->r_rs_lock[ring_id]);
#endif
    ss->r_rs[ring_id].rs_desc_cnt = rs_update->rs_desc_cnt;
    ss->r_rs[ring_id].rs_data_head = rs_update->rs_data_head;
    ss->r_rs[ring_id].rs_data_tail = rs_update->rs_data_tail;
#if MAL_CPU_x86_64
    mx_spin_unlock(&ss->r_rs_lock[ring_id]);
#endif
    if (da < da_min_update)
      goto done;
  }

#if MAL_CPU_x86_64
  /* If another thread is already updating the NIC tail pointer, no need
   * to wait so return right away.
   */
  if (!mx_spin_trylock(&ss->rs_lock))
    return 0;
#endif

  /*
   * Find the next smallest increment that moves the data tail
   */
  idx_n = -1;
  {
    uint64_t da_min = da_max;

    SNF_FFS_FOREACH(ss->attach_mask, i) {
      /* 64-bit: no need to lock to read the tail pointer. We only need to
       * read this one 8B integer.
       */
      data_tail = ss->r_rs[i].rs_data_tail;
      da = data_tail - ss->rs.rs_data_tail;
      if (da > da_max) {
	idx_n = -1;
	break;
      }
      else if (da <= da_min) {
	idx_n = i;
	da_min = da;
        min_data_tail = data_tail;
      }
    }
  }

  /*
   * If a valid increment is found and it warrants at least "da_min_update" bytes,
   * update the new tail pointer.
   */
  if (idx_n >= 0 && da >= da_min_update) {
    uint32_t new_desc_pages_consumed;
    uint32_t new_data_pages_consumed;
    uint64_t de, de_min = de_max;
    uint64_t desc_cnt, min_desc_cnt = 0, min_data_head;

    ss->rs.rs_data_tail = min_data_tail;

    /* 
     * For the data ring, recently detached rings may refer to data that is now
     * stale.  If that is the case, make sure they are not soon reopened to
     * point to data entries that will soon be released back to the NIC. */ 
    if (ss->detach_mask) {
      SNF_FFS_FOREACH(ss->detach_mask, i) {
	da = ss->r_rs[i].rs_data_tail - ss->rs.rs_data_tail;
	if (da > da_max)
	  ss->detach_mask &= ~(1u<<i);
      }
    }

    /*
     * Since we'll be updating the data ring, go ahead and update the latest
     * descriptor ring value while we're there 
     */
    idx_n = -1;
    de_min = de_max;
    SNF_FFS_FOREACH(ss->attach_mask, i) {
#if MAL_CPU_x86_64
      /* Need to read desc_cnt and data_head atomically */
      mx_spin_lock(&ss->r_rs_lock[i]);
#endif
      desc_cnt = ss->r_rs[i].rs_desc_cnt;
      min_data_head = ss->r_rs[i].rs_data_head;
#if MAL_CPU_x86_64
      mx_spin_unlock(&ss->r_rs_lock[i]);
#endif
      de = desc_cnt - ss->rs.rs_desc_cnt;
      if (de > de_max) {
	idx_n = -1;
	break;
      }
      else if (de <= de_min) {
	idx_n = i;
	de_min = de;
        min_desc_cnt = desc_cnt;
      }
    }
    
    if (idx_n >= 0) {
      /* 
       * The descriptor ring can now be advanced and we also save where the
       * data header ring happened to be for that descriptor count.  This is
       * useful when a new ring needs to join and consume from the same ring 
       */
      ss->rs.rs_desc_cnt = min_desc_cnt;
      ss->rs.rs_data_head = min_data_head;

      /* Ditto for descriptors and detach_mask (see above) */
      if (ss->detach_mask) {
	SNF_FFS_FOREACH(ss->detach_mask, i) {
	  de = ss->r_rs[i].rs_desc_cnt - ss->rs.rs_desc_cnt;
	  if (de > de_max) {
	    ss->detach_mask &= ~(1u<<i);
	  }
	}
      }
    }

    /*
     * At least data flow warrants an update and we take whatever desc cnt we
     * happened to figure out above as a result of advancing at least
     * da_min_update bytes in the data queue.
     *
     * Phoenix.  This is the only place where we update the flow control registers.
     * ss->rs contains the latest consumer pointers.  We need to compute deltas since
     * the last update and push those to the NIC.
     */
    desc_shift = (MX_IS_K35SNF(snf->is)) ? SNF_K35SNF_DESC_SHIFT : SNF_DESC_SHIFT;
    new_desc_pages_consumed = (ss->rs.rs_desc_cnt >> desc_shift) -
        (ss->rs_last_update.rs_desc_cnt >> desc_shift);
    new_data_pages_consumed = SNF_DATA_VPAGES(ss->rs.rs_data_tail) -
        SNF_DATA_VPAGES(ss->rs_last_update.rs_data_tail);
    if (new_desc_pages_consumed > 0) {
      *snf->desc_ring_consume = new_desc_pages_consumed;
      MAL_STBAR();
      ss->rs_last_update.rs_desc_cnt += (new_desc_pages_consumed << desc_shift);
#if 0
      MX_INFO(("new_desc_pages_consumed=%u desc_cnt=%u latest=%u\n",
               new_desc_pages_consumed,
               (uint32_t)ss->rs_last_update.rs_desc_cnt,
               (uint32_t)ss->rs.rs_desc_cnt));
#endif
    }
    if (new_data_pages_consumed > 0) {
      *snf->data_ring_consume = new_data_pages_consumed;
      MAL_STBAR();
      ss->rs_last_update.rs_data_tail += (new_data_pages_consumed << SNF_VPAGE_SHIFT);
#if 0
      MX_INFO(("new_data_pages_consumed=%u data_tail=%u latest=%u\n",
               new_data_pages_consumed,
               (uint32_t)ss->rs_last_update.rs_data_tail,
               (uint32_t)ss->rs.rs_data_tail));
#endif
    }
  }

#if MAL_CPU_x86_64
  mx_spin_unlock(&ss->rs_lock);
done:
#else
done:
  mx_spin_unlock(&ss->rs_lock);
#endif
  return status;
}

static int
snf_rx_open_handle(mx_endpt_state_t *es, myri_snf_rx_params_t *x)
{
  mx_instance_state_t *is = es->is;
  myri_snf_rx_state_t *snf = &is->snf;
  int status = 0;
  myri_snf_rx_params_t x_orig;
  struct snf_dup_ma_app_state *app = NULL;
  uint32_t port_merge_flags = SNF_PORT_MERGE_MODE_NONE;

  if (es->es_type != MYRI_ES_SNF_RX)
    return EINVAL;
  x_orig = *x; /* Save the original parameters for DUP_MA mode. */

  mx_mutex_enter(&snf->sync);
  /* First do all parameter checking where failures won't consume any SNF
   * resources */
  if ((status = myri_snf_rx_open_get_params(es, x))) {
    mx_mutex_exit(&snf->sync);
    return status;
  }

  if (snf->state == SNF_RX_S_FREE) { /* First Sniffer opener */
    if ((status = phx_snf_rx_enable(es, x)))
      goto fail;
  } else {
    if ((status = myri_snf_rx_join(es)))
      goto fail;
  }

  /* DUP_MA mode.  Save the application/process info. */
  if ((snf->params.open_flags & MYRI_SNF_F_RX_DUP_MA)) {
    if ((status = myri_snf_app_join(es, x, &x_orig, &app)))
      goto fail;
    es->snf.es_rx_h.app_info = app;
    app->refcount++;

    /* A one process may open the handle multiple times.  On the second open
     * and later, it passes blank parameters and expects to retrieve
     * parameters via the kernel.  Probably an overkill.  This usage is
     * highly unlikely.  The user must pass in the correct app_id in any case.
     */
    x->num_rings = app->num_rings_requested;
    x->app_id = x_orig.app_id; /* open_get_params clobbers this... */
    sprintf(es->opener.user_info, "app_id=%d rx handle (%d rings)",
        x->app_id, x->num_rings);
  }
  else {
    sprintf(es->opener.user_info, "rx handle (%d %s rings)", 
        snf->num_rings_requested,
        (snf->params.open_flags & MYRI_SNF_F_PSHARED) ? "shared" : "private");
  }
  snf->refcount++;
  es->snf.es_rx_h.rx_is_attached = 1;

  /* Enable HW hash if possible. We check all ports this app is going to use
   * and increment in_use on each.
   */
  if (x->num_rings > 1 && x->rss_flags &&
      (x->rss_flags & is->phx.pfi->hw_hash.supported_flags) == x->rss_flags) {
    int i, use_hw_hash, port, num_ports, idx[SNF_PORTGROUP_MAX_MEMBERS];
    uint32_t enabled_flags, requested_flags;
    struct phx_fpga_instance *fpga;

    requested_flags = app ? app->rss_flags : snf->params.rss_flags;
#if MAL_DEBUG
    MX_INFO(("try to use hw hash: portmask=0x%x requested_flags=0x%x\n",
             snf->params.aggr_portmask, requested_flags));
#endif
    /* We may have to visit multiple ports and FPGAs in case of port merging.
     * So grab the global FPGA mutex. It is okay to be slow here.
     */
    mx_mutex_enter(&mx_fpga_mutex);

    /* Visit all ports we are going to use and check if we can use HW hash
     * on all of them.
     */
    num_ports = 0;
    use_hw_hash = 1;
    fpga = is->phx.pfi;
    SNF_FFS_FOREACH(snf->params.aggr_portmask, i) {
      int j;

      for (j = 0; j < fpga->attach_count; j++) {
        if (fpga->is[j]->id == i) {
          port = MX_IS_PORT_ID(fpga->is[j]);
          idx[num_ports++] = j;
          enabled_flags = fpga->hw_hash.enabled_flags[port];
          if (!(!enabled_flags || enabled_flags == requested_flags))
            use_hw_hash = 0;
          break;
        }
      }
      /* To make life easier, if the port group spans multiple FPGAs, do not
       * use HW hash.
       */
      if (j == fpga->attach_count) {
        use_hw_hash = 0;
        break;
      }
    }

    /* Visit ports and enable HW hash this time */
    if (use_hw_hash) {
      for (i = 0; i < num_ports; i++) {
        mx_instance_state_t *port_is = fpga->is[idx[i]];

        port = MX_IS_PORT_ID(port_is);
        enabled_flags = fpga->hw_hash.enabled_flags[port];
        if (!enabled_flags) {
          /* HW hash is not enabled, so in_use must be 0 */
          mal_always_assert(fpga->hw_hash.in_use[port] == 0);
          fpga->hw_hash.enabled_flags[port] = requested_flags;
          /* Program the module */
          port_is->board_ops.enable_hw_hash(port_is, requested_flags);
        }
        fpga->hw_hash.in_use[port]++;
#if MAL_DEBUG
        MX_INFO(("use HW hash: port=%d in_use=%d\n",
                 port, fpga->hw_hash.in_use[port]));
#endif
      }

      /* TODO: Ugly, the same information gets stored in different places
       * depending on app_id vs. single app.
       */
      if (app)
        app->using_hw_hash = 1;
      else
        snf->using_hw_hash = 1;
    }

    mx_mutex_exit(&mx_fpga_mutex);
  }


  /* See if we can use next_ring based merging */
  if (((snf->params.open_flags & (MYRI_SNF_F_AGGREGATE_PORTMASK |
                                  MYRI_SNF_F_NO_NEXT_RING_MERGE)) ==
       MYRI_SNF_F_AGGREGATE_PORTMASK) &&
      (is->phx.pfi->port_merge.mode_flags & SNF_PORT_MERGE_MODE_NEXT_RING)) {
    if (is->board_ops.enable_next_ring_merge(is, snf->params.aggr_portmask)) {
      port_merge_flags = SNF_PORT_MERGE_MODE_NEXT_RING;
    }
  }

  is->phx.port_merge = port_merge_flags;
  x->port_merge_flags = port_merge_flags;

 fail:
  mx_mutex_exit(&snf->sync);
  if (status)
    snf_rx_teardown(es);
  return status;
}

static
int
snf_rx_portgroup_add(mx_endpt_state_t *es_pg, mx_endpt_state_t *es,
  myri_snf_port_group_t *param)
{
  struct snf_endpt_portgroup *ep_pg = &es_pg->snf.es_pg;
  int status;
  int port_id;
  struct myri_snf_waiter *waiter;

  mx_mutex_enter(&es_pg->sync);
  port_id = es->is->id;
  if (es_pg->es_type == MYRI_ES_SNF_RX_GROUP) {
    /* FIXME  We use pg's vq wait to wait on the port...
     * This is probably not the right place to initialize waitsync.
     */
    if (ep_pg->vq_waitsync.wait_sync == NULL) {
      ep_pg->vq_waitsync.wait_sync = &es_pg->vq_sync;
      ep_pg->vq_waitsync.snf = &es_pg->is->snf;
      ep_pg->vq_waitsync.intr_requested = 0;
    }

    waiter = &ep_pg->pg_waiters[port_id];
    if (waiter->wait_sync != NULL) {
      status = EBUSY;
    }
    else {
      waiter->wait_sync = &es_pg->wait_sync;
      waiter->intr_requested = 0;
      waiter->snf = &es->is->snf;
      TAILQ_INSERT_TAIL(&ep_pg->pg_waitq, waiter, pg_link);
      es->snf.es_rx.rx_pg = es_pg;
      /* es is referencing es_pg, save a back pointer to es so portgroup_del
       * can cut the reference when the portgroup closes.
       */
      waiter->pg_attached_es_rx = es;
      ep_pg->evq_base[port_id] = param->evq_base;
      status = 0;
    }
  }
  else
    status = EINVAL;
  mx_mutex_exit(&es_pg->sync);

  return status;
}

/* 
 * For removal, if es_rx_ring == NULL, remove all refs.
 */
static
int
snf_rx_portgroup_del(mx_endpt_state_t *es, mx_endpt_state_t *es_rx_ring)
{
  struct snf_endpt_portgroup *ep_pg = &es->snf.es_pg;
  struct myri_snf_waiter *waiter;
  myri_snf_rx_state_t *snf;
  int i;
  unsigned long flags;
  int wake_pending;

  flags = 0;  /* useless initialization to pacify -Wunused */

  if (es->es_type != MYRI_ES_SNF_RX_GROUP)
    return EINVAL;

  if (es_rx_ring != NULL) {
    i = es_rx_ring->is->id;
    waiter = &ep_pg->pg_waiters[i];
    if (waiter->wait_sync == NULL)
      return EINVAL;
  }
  else {
    waiter = TAILQ_FIRST(&ep_pg->pg_waitq);
  }

  mx_mutex_enter(&es->sync);

  while (waiter != NULL) {
    if (waiter->wait_sync != NULL) {
      snf = waiter->snf;
      ep_pg->port_mask &= ~(1 << snf->is->id);
      mx_spin_lock_irqsave(&snf->rx_intr_lock, flags);
      if (waiter->intr_requested) {
        TAILQ_REMOVE(&snf->rx_intr_waitq, waiter, link);
        waiter->intr_requested = 0;
      }
      wake_pending = waiter->intr_wake_pending;
      mx_spin_unlock_irqrestore(&snf->rx_intr_lock, flags);
      if (wake_pending) {
        wake_pending = 2;
        do {
          mx_sleep(&es->sync, MX_SMALL_WAIT, MX_SLEEP_NOINTR);
        } while (waiter->intr_wake_pending && --wake_pending >= 0);
        if (wake_pending == 0)
          MX_WARN(("%s: unable to clear wake_pending in portgroup\n",
                  snf->is->is_name));
      }
      waiter->wait_sync = NULL;
      /* Remove the reference from the ring if any, so we do not call
       * portgroup_del when the portgroup endpoint is already gone.
       */
      if (waiter->pg_attached_es_rx) {
        mal_always_assert(
            waiter->pg_attached_es_rx->es_type == MYRI_ES_SNF_RX_RING);
        waiter->pg_attached_es_rx->snf.es_rx.rx_pg = NULL;
      }
    }
    if (es_rx_ring != NULL)
      break;
    else
      waiter = TAILQ_NEXT(waiter, pg_link);
  }

  /* Wake up vq waiters on the group.  FIXME */
  if (es_rx_ring == NULL) {
    union snf_endpt_state *ess = &es->snf;
    if (ess->es_pg.vq_waitsync.intr_requested) {
      ess->es_pg.vq_waitsync.intr_requested = 0;
      snf = &es->is->snf;
      TAILQ_REMOVE(&snf->vq_waitq, &ess->es_pg.vq_waitsync, link);
      if (ess->es_pg.vq_waitsync.wait_sync)
        mx_wake(ess->es_pg.vq_waitsync.wait_sync);
    }
    ep_pg->port_mask = 0;
  }

  mx_mutex_exit(&es->sync);
  return 0;
}

static
int
snf_rx_portgroup_join(mx_endpt_state_t *es_pg, myri_snf_port_group_t *param)
{
  mx_instance_state_t *is;
  mx_endpt_state_t *es_rx_ring;
  int status;
  uint32_t port_id, ring_id;

  if (es_pg->es_type != MYRI_ES_SNF_RX_GROUP)
    return EINVAL;

  port_id = param->port_id;
  ring_id = param->ring_id;
  if (ring_id >= MYRI_SNF_MAX_RINGS || port_id >= SNF_PORTGROUP_MAX_MEMBERS)
    return EINVAL;

  if (es_pg->is->id != port_id) {
    is = mx_get_instance(port_id);
    if (is == 0)
      return ENODEV;
  }
  else
    is = es_pg->is;

  es_rx_ring = is->snf.ring_map_ep[ring_id];
  if (es_rx_ring == NULL) {
    if (es_pg->is->id != port_id)
      mx_release_instance(is);
    return ENODEV;
  }
  (void) mx_get_endpoint(is, es_rx_ring->endpt);

#ifdef _LINUX_UIDGID_H
  if (!uid_eq(es_rx_ring->euid,es_pg->euid)) status = EPERM;
#else
  if (es_rx_ring->euid != es_pg->euid) status = EPERM;
#endif
  else if (es_rx_ring->snf.es_rx.rx_pg != NULL)
    status = EBUSY; /* already in a port group */
  else
    status = snf_rx_portgroup_add(es_pg, es_rx_ring, param);

  mx_put_endpoint(es_rx_ring);
  if (es_pg->is->id != port_id)
    mx_release_instance(is);

  return status;
}

static
int
snf_eop_ioctl(mx_endpt_state_t *es, uint32_t cmd, const uaddr_t in)
{
  uaddr_t out = in;
  int status = ENOTTY;
  myri_snf_rx_state_t *snf = &es->is->snf;

  if (mx_is_dead(es->is)) {
    MX_WARN(("Dead board %d, ignoring ioctl\n", es->is->id));
    return EIO;
  }

  /* Lock the pg's port, instead of all ports in the pg. */
#define LOCK_PG 1

  if (cmd == MYRI_SNF_PQ_GROUP_LOCK) {
    myri_snf_pq_group_lock_t param;
    struct myri_snf_waiter *waiter;
    myri_snf_rx_state_t *pg_snf;
    struct snf_endpt_portgroup *ep_pg;
    int locked, wait;
    unsigned long flags;

    flags = 0; /* -Wunused on non-linux */

    if ((status = mx_copyin(in, &param, sizeof(param), es->is_kernel)))
      return status;

    /* Endpoint is the ring group. */
    ep_pg = &es->snf.es_pg;

    /* The first port stores group wide variables... */
    pg_snf = snf;

  pq_group_lock_tryagain:
    locked = 0; /* 1 = locked a port, 2 = locked by another */
    wait = 0; /* 1 = wait on pq, 2 = wait on vq */

#if LOCK_PG
    mx_spin_lock_irqsave(&pg_snf->pq_lock, flags);
#else
    /* Acquire all spin locks */
    TAILQ_FOREACH(waiter, &ep_pg->pg_waitq, pg_link) {
      snf = waiter->snf;
      mx_spin_lock(&snf->pq_lock);
    }
#endif

    /* Check each port.  FIXME.  Make a faster scan. */
    TAILQ_FOREACH(waiter, &ep_pg->pg_waitq, pg_link) {
      int portnum;

      snf = waiter->snf;
      portnum = snf->is->id;
      /* Skip if we are not supposed to check this port */
      if ((param.u.portmask & (1 << portnum)) == 0)
        continue;

      if (snf->pq_locked_es == NULL) {
        /* Now check if the ring is non-empty */
        uint32_t seqnum;

        if (MX_IS_K35SNF(snf->is)) {
          k35snf_rx_desc_t *evq_base;
          evq_base = (k35snf_rx_desc_t*)ep_pg->evq_base[portnum];
          seqnum = (evq_base + snf->pq_s->rw.evq_next_prehash_idx)->seqnum;
        } else {
          phx_rx_desc_t *evq_base;
          evq_base = (phx_rx_desc_t*)ep_pg->evq_base[portnum];
          seqnum = (evq_base + snf->pq_s->rw.evq_next_prehash_idx)->seqnum;
        }
        if (seqnum == snf->pq_s->rw.evq_prehash_seqnum) {
          /* Non-empty.  Lock the port and we are done. */
          snf->pq_locked_es = es;
          param.u.portnum = portnum;
          /* Remember the port number so we can unlock it later */
          ep_pg->pq_locked_port = snf;
          locked = 1;
          break;
        }
        else {
          /* Empty.  Continue to the next port. */
        }
      }
      else {
        /* Locked by another thread.  Skip */
        locked = 2;
      }
    }

    /* See if we want to wait on pq or vq */
    if (locked != 1 && param.timeout_ms != 0) {
      if (locked == 2 || pg_snf->pq_group_pq_waiter) {
        /* Another thread is hashing or is already waiting on pq.
         * Wait on vq.
         */
        wait = 2;
      }
      else {
        /* Wait on pq */
        wait = 1;
        pg_snf->pq_group_pq_waiter = 1;
      }
    }

#if LOCK_PG
    mx_spin_unlock_irqrestore(&pg_snf->pq_lock, flags);
#else
    /* Release all spin locks */
    TAILQ_FOREACH(waiter, &ep_pg->pg_waitq, pg_link) {
      snf = waiter->snf;
      mx_spin_unlock(&snf->pq_lock);
    }
#endif

    /* Did we lock any port? */
    if (locked == 1) {
      /* Tell the user which port is locked */
      status = mx_copyout(&param, out, sizeof(param), es->is_kernel);
      if (status) {
        /* FIXME. Not sure what we should do */
      }
      status = 0;
    }
    else if (wait != 0) {
      if (wait == 1) {
        /* PQ */
        status = snf_pq_group_wait(es, param.timeout_ms);
        /* Do not bother locking.  Worst, all other threads wait on vq
         * without anyone to wake them up.  The current thread would likely hash
         * the pq again and wake up vq waiters.  Or, if it closes unexpectedly,
         * we unconditionally wake up all waiters.
         */
        pg_snf->pq_group_pq_waiter = 0;
        if (status == 0) {
          /* If we wake up early, try locking a port again to avoid coming into
           * the kernel right afterwards.  But this time, use timeout=0 so
           * we do not sleep repeatedly.
           */
          param.timeout_ms = 0;
          goto pq_group_lock_tryagain;
        }
      }
      else {
        /* VQ.  Wait on the first port for now.  es is the ring group
         * and IS on the first port.
         */
        status = snf_vqwait_req(es, param.timeout_ms);
      }
      if (status == EINTR) {
        /* timeout = -1.  EINTR is the expected return code. */
        /* status = EAGAIN; */
      }
      else if (status != EAGAIN)
        status = EBUSY;
    }
    else {
      /* Could not lock any port and did not go to sleep either. */
      status = EAGAIN;
    }
    return status;
  }
  else if (cmd == MYRI_SNF_PQ_GROUP_UNLOCK) {
    struct snf_endpt_portgroup *ep_pg;
    struct snf__pq_state_rw pq_rw;
    myri_snf_rx_state_t *pg_snf;
    unsigned long flags;

    flags = 0; /* -Wunused on non-linux */

    if (in != 0) {
      if ((status = mx_copyin(in, &pq_rw, sizeof(pq_rw), es->is_kernel)))
        return status;
    }

    pg_snf = snf; /* Group's first port */
    ep_pg = &es->snf.es_pg;
    snf = ep_pg->pq_locked_port;

#if LOCK_PG
    mx_spin_lock_irqsave(&pg_snf->pq_lock, flags);
#else
    /* Only need to acquire the port's spinlock */
    mx_spin_lock(&snf->pq_lock);
#endif

    /* Update the shared pq_s. */
    if (in != 0)
      memcpy(&snf->pq_s->rw, &pq_rw, sizeof(pq_rw));
    snf->pq_locked_es = NULL;
    status = 0;

#if LOCK_PG
    mx_spin_unlock_irqrestore(&pg_snf->pq_lock, flags);
#else
    mx_spin_unlock(&snf->pq_lock);
#endif

    /* Wake up vq waiters.  VQ waiters are on the first port. */
    snf_vqwake_req(pg_snf, 0);
  }

  if (cmd >= MYRI_SNF_PQ_TRYLOCK && cmd <= MYRI_SNF_PQ_UNLOCK_WAKE) {
    struct snf__pq_state_rw pq_rw;
    unsigned long flags;

    flags = 0; /* -Wunused on non-linux */

    if (cmd == MYRI_SNF_PQ_UNLOCK_WAKE && in != 0) {
      if ((status = mx_copyin(in, &pq_rw, sizeof(pq_rw), es->is_kernel)))
        return status;
    }

    mx_spin_lock_irqsave(&snf->pq_lock, flags);
    if (cmd == MYRI_SNF_PQ_TRYLOCK || cmd == MYRI_SNF_PQ_TRYLOCK_WAIT) {
      if (!snf->pq_locked_es) {
        snf->pq_locked_es = es;
        status = 0;
      }
      else
        status = EBUSY;
    }
    else {
      snf->pq_locked_es = NULL;
      status = 0;
      /* Update the shared pq_s */
      if (cmd == MYRI_SNF_PQ_UNLOCK_WAKE && in != 0)
        memcpy(&es->is->snf.pq_s->rw, &pq_rw, sizeof(pq_rw));
    }
    mx_spin_unlock_irqrestore(&snf->pq_lock, flags);

    if (cmd == MYRI_SNF_PQ_TRYLOCK_WAIT && status == EBUSY) {
      /* Trylock and then wait for vq in one syscall */
      uint32_t x;
      status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
      if (!status) {
        if (x == 0)
          status = EAGAIN;
        else {
          status = snf_vqwait_req(es, x);
          if (status == EINTR) {
            /* timeout = -1.  EINTR is the expected return code. */
            /* status = EAGAIN; */
          }
          else if (status != EAGAIN)
            status = EBUSY;
        }
      }
      /* If copyin fails? */      
    }
    else if (cmd == MYRI_SNF_PQ_UNLOCK_WAKE) {
      /* Unlock and wake vq waiters in one syscall */
      snf_vqwake_req(&es->is->snf, 0);
    }
    return status;
  }

  switch (cmd) {

  /* FIXME.  phx_dbl_ops.c has the identical code. */
  case MX_GET_MAX_SEND_HANDLES:
    {
      int cnt = phx_read_ep_reg_tx_desc_buffer_size(es->snf.es_tx.u.phx.tx_regs);
      status = mx_copyout(&cnt, out, sizeof(cnt), es->is_kernel);
    }
    break;

  case MYRI_SNF_RX_START:
  case MYRI_SNF_RX_STOP:
    if (es->es_type != MYRI_ES_SNF_RX) {
      status = EINVAL;
      break;
    }
    mx_mutex_enter(&snf->sync);
    if (cmd == MYRI_SNF_RX_START) {
      if (snf->state != SNF_RX_S_STARTED) {
        status = myri_snf_start_stop(es, 1);
        snf->state = SNF_RX_S_STARTED;
      } else {
        status = 0;
      }
    }
    else {
      if (snf->state == SNF_RX_S_STARTED) {
        status = myri_snf_start_stop(es, 0);
        snf->state = SNF_RX_S_STOPPED;
      } else {
        status = 0;
      }
    }
    mx_mutex_exit(&snf->sync);
    break;

  case MYRI_SNF_WAIT:
    {
      uint32_t x;
      status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
      if (status)
        break;

      if (es->es_type == MYRI_ES_SNF_TX)
        status = snf_eop_poll_wait(es, x, MYRI_WAIT_TX_EVENT);
      else if (es->es_type == MYRI_ES_SNF_RX_RING ||
               es->es_type == MYRI_ES_SNF_RX_GROUP ||
               es->es_type == MYRI_ES_SNF_RX_PQ)
        status = snf_eop_poll_wait(es, x, MYRI_WAIT_RX_EVENT);
      else
        status = EINVAL;
    }
    break;

  case MYRI_SNF_STATS:
    {
      phx_snf_stats_t x;
      update_and_read_snf_stats(es->is, &x);
      subtract_snf_stats(&x, &snf->stats_at_start);
      status = mx_copyout(&x, out, sizeof(x), es->is_kernel);
    }
    break;

  case MYRI_SNF_RX_SYNC:
    {
      struct snf__rx_sync x;

      if (es->es_type != MYRI_ES_SNF_RX_RING) {
        status = EINVAL;
        break;
      }
      status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
      if (status)
	break;
      status = snf_rx_sync(snf, es->snf.es_rx.rx_ring_id, &x);
    }
    break;


  case MYRI_SNF_RX_GROUP_JOIN:
    {
      myri_snf_port_group_t x;

      status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
      if (status == 0)
        status = snf_rx_portgroup_join(es, &x);
    }
    break;

  case MYRI_SNF_RX_DETACH:
    mx_mutex_enter(&snf->sync);
    status = myri_snf_rx_detach(es, 0);
    mx_mutex_exit(&snf->sync);
    break;

#if MAL_OS_LINUX || MAL_OS_FREEBSD || MAL_OS_WINDOWS
  case MYRI_SNF_SOFT_RX:
    {
      myri_soft_rx_t x;
      status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
      if (status)
       break;
      status = myri_ether_soft_rx(es->is, &x);
    }
    break;
#endif /* MAL_OS_LINUX */
  }

  return status;
}

myri_endpt_ops_t phx_snf_ops;

void
phx_snf_init_ops(void)
{
  bzero(&phx_snf_ops, sizeof(phx_snf_ops));
  phx_snf_ops.flags = MYRI_EOP_F_POLL;
  phx_snf_ops.alloc = snf_eop_alloc;
  phx_snf_ops.enable = snf_eop_enable;
  phx_snf_ops.ioctl = snf_eop_ioctl;
  phx_snf_ops.mmap  = snf_eop_mmap;
  phx_snf_ops.close = snf_eop_close;
  phx_snf_ops.interrupt_is = snf_eop_intr_is;
  phx_snf_ops.interrupt_ep = snf_eop_intr_endpt;
  phx_snf_ops.poll_events_pending = snf_eop_poll_events_pending;
  phx_snf_ops.poll_prep_wait = snf_eop_poll_prep_wait;
  phx_snf_ops.poll_wait = snf_eop_poll_wait;
}

void
phx_snf_set_ether_promisc(mx_instance_state_t *is, int enable)
{
  myri_snf_rx_state_t *snf = &is->snf;

  /* If SNF is not using MAC, change the mode right away. Otherwise, save
   * the request bit so myri_snf_start_stop can set the mode later on.
   */
  mx_spin_lock(&snf->promisc_lock);
  /* Clear the eth request flag to avoid leaving the flag set forever */
  snf->promisc_flags &= ~SNF_RX_PROMISC_FLAG_REQUEST_ENABLE;
  if ((snf->promisc_flags & SNF_RX_PROMISC_FLAG_SNF_OWNED)) {
    if (enable)
      snf->promisc_flags |= SNF_RX_PROMISC_FLAG_REQUEST_ENABLE;
  } else {
    is->board_ops.set_promisc(is, enable);
  }
  mx_spin_unlock(&snf->promisc_lock);
}
